Macintosh Drag and Drop for MCL
version 1.5.5

Dan S. Camper
lordgrey@apple.com

Made-up version numbers with real release dates:

     Version 1.0:      05/24/94
     Version 1.1:      01/20/95
     Version 1.1.1:    06/15/95
     Version 1.1.2:    09/25/95
     Version 1.1.3:    11/10/95
     Version 1.5:      05/30/96
     Version 1.5.1:    06/04/96
     Version 1.5.2:    06/13/96
     Version 1.5.3:    07/05/96
     Version 1.5.4:    09/12/96
     Version 1.5.5:    09/13/96  (yep -- a bug fix!)

-----------------------------------------------------------------------------
 What is this?

The module drag-and-drop.lisp provides an MCL interface to the Macintosh
Drag Manager, available on System 7.0 through 7.1.3 via extensions and
built into System 7.5 and later versions.  Included are several examples that
show off the module's features and, in the process, demonstrate how to use
many of the methods and functions.

Also present is the Read Me document included with the Macintosh Drag and Drop
Developer's Kit. This document contains system requirements, installation
instructions, etc. and will answer basic questions regarding the technology.
Further Drag Manager documentation (and possibly the SDK) is available from
Apple's ftp sites and from APDA.

Recommended reading includes the _Drag Manager Programmers Guide_, which is
available with the Macintosh Drag Manager stuff from Apple's ftp sites, or
the Drag Manager section of the QuickView programmer's database available
on the ETO CD-ROM (subscriptions available through APDA).  This file does
not attempt to explain the Drag Manager, just my implementation of it.

The rest of this file documents every method, function, macro and global
variable exported from the Drag and Drop module.  A simple list of those
symbols is included, quasi-categorized by function, followed by more detailed
item-by-item descriptions roughly categorized into "receiving a drag,"
"tracking a drag," "initiating a drag" and the omnipresent "miscellaneous."
At the end of this file are descriptions of the included examples and a few
programming tips.

You can skip from major section to major section by searching for ''.

If you used previous versions of this drag and drop code, you'll be in for a
bit of rewriting in implementing this module.  Significant changes have been
made in almost every method, and many of the older methods have been renamed
for consistency reasons.  Not to mention the fact that the overall architecture
has changed from window-based classes to view-based classes.  I can only hope
that the additional features and bug fixes in this release make the work
worthwhile!

If you have any questions, problems or comments regarding this module, please
feel free to contact me at the email address listed above.

Cheers,

DSC

09/10/96

-----------------------------------------------------------------------------
 Compatibility

This module and the enclosed examples are compatible with Macintosh Common
Lisp 2.0 and higher.  As of this writing, the latest release of MCL was
version 4.0b1.

-----------------------------------------------------------------------------
 Preliminary Information

All of the Drag Manager's behavior is linked to views within MCL.  In order
to make a view "drag-aware" it must inherit from the 'drag-view-mixin class.
This class provides the framework for interacting with the Drag Manager and
also performs much of the work involved with dealing with Drag Manager data
structures.

Note that previous versions of this module had a window-based mixin class for
implementing drag behavior.

This module's implementation of the Drag Manager's interface is heavily
layered in certain places.  For instance, to handle dropped data you may
choose to write specialized methods that handle only the data itself, or you
may specialize "lower-level" methods that handle the entire receiving
process.  Most of the implementation has been constructed and exported to
allow the maximum amount of method specialization.

The drag-and-drop.lisp module, when evaluated, pushes :drag-and-drop onto
the *features* variable in MCL.  This means that you can use the #+ and #-
compiler macros in your code to selectively tune-in or -out Drag Manager
calls.  In addition, the module calls (provide :drag-and-drop), so you can
use #'require in your other code modules.

-----------------------------------------------------------------------------
 Exported Symbols, almost by category (sort of, maybe)

Class Definitions:          drag-view-mixin

Environmental Stuff :       *drag-manager-present-p*
                            drag-manager-present-p
                            init-drag-and-drop
                            drag-window-p
                            dragging-now-p
                            drag-copy-requested-p
                            *drag-error-stream*

Drag Introspection/Info:    drag-get-drag-reference
                            drag-get-source-view
                            drag-get-current-view
                            
                            drag-get-attributes
                            drag-left-sender-view-p
                            drag-left-sender-window-p
                            drag-within-sender-view-p
                            drag-within-sender-window-p
                            drag-within-sender-application-p
          
Mouse Information:          drag-mouse-original-position
                            drag-mouse-drop-position
                            drag-mouse-position
                            drag-mouse-view
                            
                            drag-set-mouse-position
          
Keyboard Information:       drag-get-keyboard-modifiers
                            drag-command-key-p
                            drag-shift-key-p
                            drag-control-key-p
                            drag-option-key-p
                            drag-caps-lock-key-p

Drag Item Information:      drag-count-items
                            drag-get-item-reference-number

Item Flavor Information:    drag-flavor-exists-p
                            drag-count-flavors
                            drag-get-flavor-type
                            drag-get-flavor-type-list
                            drag-get-flavor-flags
                            drag-get-flavor-size

Starting a Drag:            drag-selection-p
                            drag-make-auto-scroll-region
                            drag-cursor
          
                            drag-setup-new-drag
                            drag-handle-new-drag
                            drag-cancel-new-drag
                            drag-cleanup-new-drag

Adding Items to a Drag:     drag-add-drag-contents
                            drag-add-item-flavor
                            drag-promise-item-flavor
                            drag-add-hfs-flavor
                            drag-promise-hfs-flavor
                            drag-add-mcl-object-flavor
                            
                            drag-create-item-bounds
                            drag-te-hilite-region

Fulfilling Promised Items:  drag-fulfill-promise
                            drag-set-item-flavor
                            drag-set-hfs-flavor
                            
Tracking a Drag:            drag-tracking-enter-handler
                            drag-tracking-enter-view
                            drag-tracking-in-view
                            drag-tracking-leave-view
                            drag-tracking-leave-handler
                            
                            drag-allow-copy-within-view-p
                            drag-allow-move-within-view-p
                            
                            drag-show-drag-hilite
                            drag-hide-drag-hilite
                            with-saved-drag-hilite
                            drag-make-drag-hilite-region
                            
                            drag-scroll-view
                            with-drag-scroll
                            
                            drag-draw-caret-p
                            drag-draw-caret
                            drag-caret-position-from-mouse

Receiving a Drag:           drag-accepted-flavor-list

                            drag-allow-drop-p
                            drag-allow-dropped-flavor-p
                            
                            drag-receive-drop
                            drag-receive-dropped-item
                            drag-receive-dropped-flavor
                            
                            drag-receive-drop-setup
                            drag-receive-drop-cleanup

Getting Flavor Data:        drag-get-flavor-data
                            drag-get-hfs-flavor
                            drag-get-promised-hfs-flavor
                            
                            with-drag-flavor

Iterator Macros:            with-each-drag-item
                            with-each-drag-item-reversed
                            with-each-drag-flavor

Drop Location Management:   drag-get-drop-location-descriptor
                            drag-get-drop-location-as-path
                            
                            drag-set-drop-location-descriptor
                            drag-set-drop-location-from-alias
                            drag-set-drop-location-from-path
                            
                            drag-get-target-description
          
Miscellaneous:              drag-zoom-rectangle
                            drag-zoom-rectangle-points
                            drag-zoom-region
                            
                            drag-proxy

-----------------------------------------------------------------------------
 Preparing a view for drag and drop:  The drag-view-mixin class

Each view that will either receive dropped data or allow dragging from it
must inherit from the 'drag-view-mixin class.

There are five additional initialization arguments that go along with this
class:

     :drag-accepted-flavor-list   - A list of drag flavors that the view
                                    normally accepts during a drop operation
     :drag-allow-move-p           - Boolean indicating whether a selection
                                    within the view can remain in the view
                                    and change locations during a drag
     :drag-allow-copy-p           - Boolean indicating whether the user can
                                    request a copy of the data within the
                                    same view by holding down the Option
                                    key during the drag
     :drag-auto-scroll-p          - Boolean indicating whether the view
                                    should scroll when a drag originates
                                    from it and the mouse moves to an edge
                                    of the view
     :drag-scroll-amount          - A point specifying the number of pixels
                                    to scroll in horizontal and vertical
                                    directions, if needed for auto-scrolling.

The first four of the initialization arguments default to nil, the last to
(make-point 12 12).

The first three arguments have corresponding reader methods which may be
specialized to provide dynamic behaviors:

     #'drag-accepted-flavor-list
     #'drag-allow-move-within-view-p
     #'drag-allow-copy-within-view-p

Further information regarding the use of these initialization arguments and
related methods can be found below, buried in their relevant sections.

-----------------------------------------------------------------------------
 Receiving drops in a view

Here are the requirements for allowing a data to be dropped into an MCL
view, before even processing the dropped data:

  1) The view must inherit from 'drag-view-mixin class.
  2) The view must "approve" of the data being dropped by returning true
     through the #'drag-allow-drop-p and #'drag-allow-dropped-flavor-p
     methods (more information is below).
  3) If the drag originated from this same view within MCL, then one of the
     two following conditions must also be met:
     
     a) The method #'drag-allow-move-within-view-p returns true and the user
        is not holding down the Option key (requesting a copy of the data).
     b) The method #'drag-allow-copy-within-view-p returns true and the user
        is holding down the Option key to indicate a request for a copy of
        the dragged data.

The first and third requirements can be met through class definitions and
initialization arguments, so there is no real code to write.  You can,
however, specialize both #'drag-allow-move-within-view-p and
#'drag-allow-copy-within-view-p if you need to be able to enable or disable
those features dynamically.  The specialization will override whatever the
initialization arguments indicated.  The default version of both of these
methods is to return whatever the corresponding initialization arguments
are.

The second requirement can be met by either defining a list of acceptable drag
flavors in the :drag-accepted-flavor-list for the view, providing a
specialization for the #'drag-accepted-flavor-list that returns the same
kind of list, or by overriding #'drag-allow-drop-p and/or
#'drag-allow-dropped-flavor-p methods.  Please see the descriptions below
for information regarding these methods.

Allowing a drop into a view usually triggers a visual feedback to the user.
Unless the user is dragging only within the source view, the area that can
receive a drop shows a "drag hilite" -- a gray border just inside the area.
The hilite is create by a call to #'drag-make-drag-hilite-region and is
shown and hidden by calls to #'drag-show-drag-hilite and
#'drag-hide-drag-hilite respectively.  All three of these methods are called
automatically by the MCL drag handlers and may be specialized.

In some cases it is necessary to provide visual feedback to the user
indicating exactly where in the view the dropped data will be inserted,
such as with a view containing text.  The method #'drag-draw-caret-p
allows you to specify if an insertion caret needs to be drawn at all,
#'drag-caret-position-from-mouse allows you to position the caret based
on the current mouse position, and #'drag-draw-caret performs the actual
drawing.  The MCL tracking handlers call these methods as needed and even
support automatic blinking of the insertion caret, at the interval specified
by the user in their control panel, if the caret stays in one place too long.

If the drop is performed then the next step is to actually receive the data.
This is typically done by specializing any one of the following three
methods:  #'drag-receive-drop, #'drag-receive-dropped-item, or
#'drag-receive-dropped-flavor.  These methods are, by default, called in that
order; at the very least, you must specialize #'drag-receive-dropped-flavor.

Much of the data will arrive in the form of Macintosh pointers, and it is
up to the application to coerce the information into a form it can use.
Many methods are included here for that purpose, as well as some that
provide some of the coercion for you (FSSpec records to Lisp pathnames,
for example).

A number of other methods are provided that allow you to inspect the dropped
data, iterate, etc..  See the relevant sections below.


 Exported methods relating to receiving drops & processing drag items

 drag-accepted-flavor-list ((view drag-view-mixin))

    This method should return either nil or a list of drag flavors that the
    view will accept in a drop operation.  The default version returns the
    value originally given to the :drag-accepted-flavor-list initialization
    argument for the view.
    
    A returned list of drag flavors should be ordered by priority, with the
    most important (or easiest to manipulate, or whatever) being first.
    
    The default MCL drag handlers use the results of this method in two ways:
    1) to determine if an attempted drop operation should be permitted (and
    hence if any visual feedback should be displayed), and 2) to determine
    the priority order for processing a drag item's flavors.
    
    Returns:  nil or an ordered list of drag flavors acceptable to a view

 drag-allow-copy-within-view-p ((view drag-view-mixin))

    If this method returns t then the user is allowed to copy data within
    the view by dragging it while holding down the Option key.  It is
    called by the MCL drag handlers to determine visual feedback for the
    drag and to determine if an attempted drop operation should succeed.
    
    The method #'drag-copy-requested-p is used to determine if the user has
    requested a copy of the drag data.
    
    The default version of this method simply returns the value originally
    given to the :drag-allow-copy-p initialization argument for the view.
    You may override it to allow dynamic enabling/disabling of this feature.
    
    Returns:  t or nil

 drag-allow-drop-p ((view drag-view-mixin))

    This method is called when the MCL drag handlers determine that it is
    possible for a drop to occur in the view (based on the results of
    #'drag-allow-copy-within-view-p, #'drag-allow-move-within-view-p and
    some environmental states) and it is now up to the application itself
    to pass or fail the drop.
    
    The default version of this method iterates through each drag item and
    each flavor in each item, calling #'drag-allow-dropped-flavor-p to
    determine if the view would accept a particular flavor.  If at least
    one flavor is accepted for each item, then the drop as a whole is
    accepted and #'drag-allow-drop-p returns t.
    
    You may override this method to provide more sophisticated tests to
    determine the acceptance for a particular drop.  For instance, a certain
    view may be able to accept only one drag item at a time in addition to
    any flavor requirements, or that a Finder file has a particular creator
    and file type.
    
    If this method is overridden, care should be taken that Macintosh Human
    Interface Guidelines are adhered to.  As noted in the previous paragraph,
    at least one flavor from each item must be acceptable before the any of
    the items are accepted -- all or none, basically.
    
    This method is used to both determine visual feedback to the user and to
    determine if an attempted drop operation should succeed.
    
    Returns:  t or nil

 drag-allow-dropped-flavor-p ((view drag-view-mixin) (flavor keyword))

    This method returns a boolean indicating whether a particular drag flavor
    will be accepted by the view during a drop operation.  It is called by
    the default version of #'drag-allow-drop-p.  Note that the 'flavor keyword
    is actually an OSType.
    
    The default version of this method compares the provided flavor against
    the results from #'drag-accepted-flavor-list; if the flavor is on the
    list then t is returned, otherwise nil.
    
    Returns:  t or nil

 drag-allow-move-within-view-p ((view drag-view-mixin))

    If this method returns t then the user is allowed to move data within
    the view by dragging it.  It is called by the MCL drag handlers to
    determine visual feedback for the drag and to determine if an attempted
    drop operation should succeed.
    
    The default version of this method simply returns the value originally
    given to the :drag-allow-move-p initialization argument for the view.
    You may override it to allow dynamic enabling/disabling of this feature.
    
    Returns:  t or nil

 drag-count-flavors ((item-reference integer))

    If called during a drag, this method returns the number of flavors
    stored in the given drag item.
    
    Returns:  non-negative integer count of flavors if called during a drag;
              nil, if called outside of a drag

 drag-count-items ()

    If called during a drag, this function returns the number of drag items
    stored the by the sending application.
    
    Returns:  non-negative integer count of drag items if called during a drag;
              nil, if called outside of a drag

 drag-flavor-exists-p ((item-reference integer) (flavor keyword))

    This method simply returns a boolean indicating whether the given
    flavor is actually available in the given item.
    
    Returns:  t if the flavor exists in the item cited by item-reference;
              nil if the flavor does not exist or if the call was outside
              of a drag operation

 drag-get-flavor-data ((item-reference integer) (flavor keyword))

    This method returns the data stored as the given flavor within the
    given item.  Two different types of return values are possible:
    1) if the data was passed through the Drag Manager then the return value
    will be a Macintosh pointer, sized correctly, pointing to that data;
    2) if the data is an internal MCL object, stored with a call to the
    #'drag-add-mcl-object-flavor method, then the return value will be that
    object.
    
    If the return value is a Macintosh pointer then it must be eventually
    disposed of with a call to #_DisposePtr.
    
    Returns:  a Macintosh pointer or MCL object if successful, nil if the
              call was made outside a drag operation

 drag-get-flavor-flags ((item-reference integer) (flavor keyword))

    Method returns the Drag Manager flags associated with the given flavor
    within the given drag item.  The flags are interpreted into MCL keywords,
    as follows:
    
            Drag Manager Constant             MCL Keyword
            ---------------------             -----------
            #$flavorSenderOnly                :internal-only
            #$flavorNotSaved                  :not-saved
            #$flavorSenderTranslated          :sender-translated
            #$flavorSystemTranslated          :system-translated

    If the Drag Manager returns a flag untranslatable by MCL, then that
    flag -- an integer -- will be returned instead of a keyword.  These
    flags are the only ones defined as of this writing.
    
    Returns:  keyword indicating the flavor flag or an integer if the
              flavor does have a flag but cannot be translated to a keyword,
              if successful; nil if the call is made outside of a drag
              operation

 drag-get-flavor-size ((item-reference integer) (flavor keyword))

    Returns the size, in bytes, of the data stored as a particular flavor
    within the given item.  Note that if the flavor was promised, this
    will cause the sending application to create and send the data, just
    so the size can be determined.
    
    Returns:  if successful, a non-negative integer representing the size,
              in bytes, the data stored as the given flavor in the given
              item; nil if the call was made outside a drag operation

 drag-get-flavor-type ((item-reference integer) optional (nth-flavor 1))

    Method returns the OSType of the nth-flavor'd flavor within the given
    drag item.  The index is considered one-based, rather than zero-based.
    
    Returns:  OSType of the found flavor if successful; nil if the flavor
              was not found or the call was made outside of a drag operation

 drag-get-flavor-type-list ((item-reference integer))

    Method returns a list of OSTypes indicating the flavors available within
    the given drag item.  The list is in the same relative order as the
    item's flavors.
    
    Returns:  if successful, an ordered list of flavors (as OSTypes) found
              within the drag item; nil if the item was not found or if the
              call was made outside a drag operation

 drag-get-hfs-flavor ((item-ref-or-macptr integer) (hfs-flavor keyword)
                       &optional (resolve-aliases-p nil))
  
  drag-get-hfs-flavor ((item-ref-or-macptr macptr) (hfs-flavor t)
                       &optional (resolve-aliases-p nil))

    This method extracts HFS flavor data and returns the information as
    multiple values, suitable for multiple-value-bind and friends.  The
    values are:
    
            Lisp pathname to the HFS object
            Finder File Creator, as an OSType
            Finder File Type, as an OSType
            Finder Flags, as an integer
    
    If the first argument is an integer, it is assumed to be an item
    reference and the method will request the HFS flavor data from the
    Drag Manager before extracting the information from it.
    
    If the first argument is a Macintosh pointer, it is assumed to be
    a pointer to the HFS flavor data itself and the information will be
    extracted directly from that memory location.  In this case the
    'hfs-flavor argument is ignored.
    
    If 'resolve-aliases-p is t and the Lisp pathname points to an alias,
    the alias will be resolved and the resulting pathname returned instead.
    
    Returns:  if successful, multiple values representing the HFS flavor
              information; nil if the flavor doesn't exist or a drag
              operation is not in progress

 drag-get-item-reference-number (&optional (nth-item 1))

    Function returns the item reference number (a unique ID) for the nth-item
    in a drag.  The index is considered one-based, rather than zero-based.
    
    Returns:  integer representing the reference number of the item if
              called during a drag; nil if called outside of a drag

 drag-get-promised-hfs-flavor ((item-ref-or-ptr integer)
                                (promised-flavor keyword))
  
  drag-get-promised-hfs-flavor ((item-ref-or-ptr t)
                                (promised-flavor keyword))

    This method extracts promised HFS flavor data and returns the information
    as multiple values, suitable for multiple-value-bind and friends.  The
    values are:
    
            Finder File Creator, as an OSType
            Finder File Type, as an OSType
            Finder Flags, as an integer
            Flavor of HFS file specification, as an OSType
    
    If the first argument is an integer, it is assumed to be an item
    reference and the method will request the promised HFS flavor data from
    the Drag Manager before extracting the information from it.
    
    If the first argument is a Macintosh pointer, it is assumed to be
    a pointer to the promised HFS flavor data itself and the information will
    be extracted directly from that memory location.  In this case the
    'promised-flavor argument is ignored.
    
    Given the flavor of the HFS file specification, you can then request that
    flavor to receive a Macintosh FSSpec record pointing to the HFS object
    with a call to #'drag-get-flavor-data (or by using the #'with-drag-flavor
    macro).
    
    See the #'drag-resolve-promised-hfs-flavor method automatically
    fetching the promised HFS flavor, reading the corresponding file
    specification, and converting that specification to a Lisp pathname.
    
    Returns:  if successful, multiple values representing the HFS flavor
              information; nil if the flavor doesn't exist or a drag
              operation is not in progress

 drag-make-drag-hilite-region ((view drag-view-mixin)
                                &optional
                                (topleft nil) (bottomright nil))

    Method creates a Macintosh region that the Drag Manager will use to
    compute the drag hilite, the visual indication that a user can drop
    something into the view.  If 'topleft and 'bottomright are defined,
    they must be points in the view-window's coordinate system that
    define this view's drop area.  The optional arguments are not
    used then suitable points will be generated that define the view's
    borders.
    
    This method is called by the MCL drag handlers while tracking a drag.
    
    Returns:  a Macintosh region

 drag-receive-drop ((view drag-view-mixin))

    This method is called by the MCL drag handlers when a user drops data into
    a drag-aware view.  It will not be called if the attempted drop failed
    any of the contexts checks (#'drag-allow-copy-within-view-p,
    #'drag-allow-move-within-view-p, #'drag-allow-drop-p or
    #'drag-allow-dropped-flavor-p).
    
    This method is responsible for processing all of the items included in the
    drag, which means that one flavor from each item must be inserted into
    the view.  The default version calls #'drag-receive-dropped-item with each
    drag item in order; if #'drag-receive-dropped-item does not return either
    t or #$noErr then all further processing is stopped.
    
    If this method returns anything other than t or #$noErr then the drop
    will be rejected and the Drag Manager will provide the "zoom back"
    animation, notifying the user of the rejection.
    
    Returns:  t if successful; otherwise the error from the last call to
              #'drag-receive-dropped-item

 drag-receive-drop-cleanup ((view drag-view-mixin))
    
    This method is called by the MCL drag handlers after all data from a drop
    into the drag-aware view has been processed.  It will be called regardless
    of the outcome of the drop, even if there has been an error.
    
    The default version of the method does nothing.
    
    Returns:  Undefined.

 drag-receive-drop-setup ((view drag-view-mixin))
    
    This method is called by the MCL drag handlers when a user drops data into
    a drag-aware view, and before the handlers call the methods to process
    that data.  Specializations can use this method to prepare the view for
    the incoming data (removing selection highlights, etc.).
    
    The default version of the method does nothing.
    
    Returns:  Undefined.

 drag-receive-dropped-flavor ((view drag-view-mixin) (flavor keyword)
                               (data-ptr macptr) (data-size integer)
                               (item-reference integer))

    Method receives a dropped flavor and is responsible for performing the
    actual insertion into the view.  If #'drag-receive-drop and
    #'drag-receive-dropped-item have not been overridden -- meaning that
    this method will be called because it is the default action -- then
    it must be specialized.  Otherwise, the view would never receive any
    dropped data.
    
    Returns:  t if successful, nil if #'drag-receive-dropped-item should
              extract another flavor for another try, anything else
              if there was an error

 drag-receive-dropped-item ((view drag-view-mixin) (item-reference integer))

    This method is responsible for processing a single dropped item, and is
    called from #'drag-receive-drop.  It's default behavior is to compare
    the item's flavors against the flavors the view will accept (the results
    from #'drag-accepted-flavor-list), then call #'drag-receive-dropped-flavor
    with that flavor data.
    
    If the method receives a nil from #'drag-receive-dropped-flavor then it
    will obtain the next-highest priority flavor and make another attempt.
    The method will repeat the attempts through all of the flavors available
    in the drag item, providing they are also acceptable to the view.
    
    Returns:  the result from the last call to #'drag-receive-dropped-flavor

 drag-resolve-promised-hfs-flavor ((item-reference integer)
                                    (promised-hfs-flavor keyword)
                                    &optional (resolve-aliases-p nil))

    This method extracts the promised HFS flavor data, determines the flavor
    of the FSSpec that would point to the HFS object, extracts that FSSpec
    and resolves it into a pathname.  It returns the following values:
    
            Lisp pathname to the HFS object
            Finder File Creator, as an OSType
            Finder File Type, as an OSType
            Finder Flags, as an integer
    
    If 'resolve-aliases-p is t and the Lisp pathname points to an alias, that
    alias will be resolved and the resulting pathname returned instead.
    
    Returns:  if successful, multiple values representing the HFS flavor
              information; nil if the flavor doesn't exist or a drag
              operation is not in progress

 drag-set-drop-location-descriptor (ae-drop-descriptor)

    This function takes as an argument a record of type :AEDesc and sets
    the current drop location to that record.  A copy of the descriptor
    record is stored; callers are free to dispose of the AEDesc if needed.
    
    The typical use for this function while receiving dropped data is if
    a needed flavor is a promised HFS file.  In that case, this function
    should be called before requesting the HFS flavor (which is when the
    sending application would create the file, if it needs to be created)
    in order to define a location to store the file.  See the
    #'drag-set-drop-location-from-path function for an easier way to set the
    drop location to a Finder folder.
    
    Returns:  the passed :AEDesc record if successful, nil if the function
              was called outside of a drag operation

 drag-set-drop-location-from-alias (alias-handle)

    This function requires an actual alias handle, not a record of type
    :AliasHandle.  It converts the alias to an Apple Event Descriptor
    record and calls #'drag-set-drop-location-descriptor to set the drop
    location.
    
    Tip:  If the calling function is using an alias stored in a :AliasHandle
    record (possibly defined within an rlet), then wrap the record in a
    %get-ptr call to pass the alias to #'drag-set-drop-location-from-alias.
    As an example:
         (rlet ((alias-rec :AliasHandle))
           ;; generate a valid alias
           (drag-set-drop-location-from-alias (%get-ptr alias-rec)))
    
    The typical use for this function while receiving dropped data is if
    a needed flavor is a promised HFS file.  In that case, this function
    should be called before requesting the HFS flavor (which is when the
    sending application would create the file, if it needs to be created)
    in order to define a location to store the file.  See the
    #'drag-set-drop-location-from-path function for an easier way to set the
    drop location to a Finder folder.
    
    Returns:  the original alias handle passed if successful, nil otherwise

 drag-set-drop-location-from-path (path)

    Function accepts a pathname or string representing a pathname, then
    converts it to an alias and calls #'drag-set-drop-location-from-alias
    with the result.
    
    The typical use for this function while receiving dropped data is if
    a needed flavor is a promised HFS file.  In that case, this function
    should be called before requesting the HFS flavor (which is when the
    sending application would create the file, if it needs to be created)
    in order to define a location to store the file.
    
    Returns:  the original path passed if successful, nil otherwise

 with-drag-flavor ((item-reference flavor-type
                   flavor-data-symbol flavor-size-symbol)
                   &body body)

    Convenience macro for extracting data stored in the given flavor within
    the given item.  The macro takes care of memory management, including
    the disposal of any created Macintosh pointers.
    
    The 'flavor-data-symbol is bound to the Macintosh pointer or MCL object,
    and 'flavor-size-symbol is bound to the size of the pointer.  If
    'flavor-data-symbol is an MCL object, the size will always be zero.
    
    If there is a valid object with that flavor, then both symbols will be
    bound and the body code executed.
    
    Returns:  t if there was valid data stored in the given flavor within
              the given item; nil otherwise.

 with-each-drag-flavor ((item-reference flavor-type-symbol
                        flavor-data-symbol flavor-size-symbol)
                        &body body)

    Convenience macro for iterating through all the flavors stored within
    the given drag item.
    
    For each flavor: 'flavor-type-symbol is bound to the OSType of the
    flavor; 'flavor-data-symbol is bound to the actual data; and
    'flavor-size-symbol is bound to the size of data in 'flavor-data-symbol.
    The body code is then executed.
    
    If 'flavor-data-symbol is bound to an MCL object, then 'flavor-size-symbol
    will always be zero.
    
    The method will take care of memory management, including the disposal of
    any created Macintosh pointers.
    
    Returns:  t if there are flavors to iterate through within the given
              drag item; nil if there are no flavors

 with-each-drag-item ((item-reference-symbol) &body body)

    This is an iteration macro that binds each item's reference number
    in turn to the provided symbol and then executes the body code.
    
    Returns:  t if where were a positive number of drag items;
              nil otherwise

 with-each-drag-item-reversed ((item-reference-symbol) &body body)
    
    Same as #'with-each-drag-item macro except the drag items are traversed
    in reversed order.
    
    Returns:  t if where were a positive number of drag items;
              nil otherwise

-----------------------------------------------------------------------------
 Tracking a drag operation

Once a drag has started, the Drag Manager is in control and all the other
applications involved with the drag basically just respond to messages.  There
are five basic messages that come from the Drag Manager:  Entering a window,
leaving a window, entering a new tracking handler routine, exiting that
handler routine, and tracking while within a window.  MCL's drag handlers
change the window-based messages to view-based -- it's much more flexible and,
after all, windows are views anyway.  The "tracking within a view" message
is called periodically (actually fairly rapidly) while a user is within a
single view, while the others are called only when events warrant them.

There are some things to do to have the application respond in ways consistent
with the Macintosh Human Interface Guidelines, though.  Upon entering a new
view, for instance, the drag hilite should appear if that view would accept
a drop (see #'drag-view-show-hilite and #'drag-view-hide-hilite).  Insertion
carets need to be drawn if necessary (see #'drag-draw-caret-p,
#'drag-draw-caret, and #'drag-caret-position-from-mouse).  Just for examples.

In MCL, all the methods named with the prefix "drag-tracking-" are called
almost directly from the Drag Manager.  They may be overridden or specialized,
but most of them actually perform necessary behaviors, consistent with the
Interface Guidelines.  If they are overridden, please ensure that the basic
functions are duplicated.  See below for more information on these handlers.


 Exported methods relating to drag tracking and visual indicators

 drag-caret-position-from-mouse ((view drag-view-mixin)
                                  local-mouse-position)

    Method interprets the current position into a position the drag insertion
    caret should be drawn at.  The insertion caret may be "pinned" to certain
    locations, such as between characters or at the end of a line in a text
    view or between cells in a grid view.  Override this method if the view
    needs to adjust the drawn caret position to something other than the
    mouse position.
    
    'local-mouse-position is a point in the view's local coordinate system.
    
    The default version of this method simply returns the 'local-mouse-position
    value.  It is called by the MCL drag handlers while tracking a drag.
    
    Returns:  a point in the view's local-coordinate system indicating where
              the drag insertion caret should be drawn

 drag-draw-caret ((view drag-view-mixin) local-position shown-p)

   This method performs the actual drawing and hiding of the drag insertion
   caret within the view.  It is called by MCL's drag handlers in response
   to user activities and system time (to make the caret blink).
   
   The 'local-position argument is a point in the view's local coordinates
   indicating where the insertion caret should be drawn or hidden.  The
   'shown-p argument indicates whether to show or hide the caret.
   
   When 'shown-p is t, 'local-position will be the result of a call to
   #'drag-caret-position-from-mouse; if 'shown-p is nil then 'local-position
   will be the point that the caret was last shown at.
   
   Returns:  Undefined

 drag-draw-caret-p ((view drag-view-mixin))

    This method returns a boolean indicating whether a drag insertion caret
    should be drawn in the view at that point in time.  The method will be
    called only if the drop is legal (as per #'drag-allow-copy-within-view-p,
    #'drag-allow-move-within-view-p, #'drag-allow-drop-p and/or
    #'drag-allow-dropped-flavor-p).
    
    Returns:  t if the drag insertion caret should be draw, nil otherwise

 drag-hide-drag-hilite ((view drag-view-mixin))

    Method hides the hilite around a view indicating that the view would
    accept a drop in the current location.  See #'drag-show-drag-hilite.
    The method is called by the MCL drag handlers while tracking a drag.
    
    Returns:  t

 drag-scroll-view ((view view) (direction-or-point keyword))
  
  drag-scroll-view ((view view) (direction-or-point integer))

  drag-scroll-view ((view fred-mixin) (direction-or-point keyword))

    This method is responsible for scrolling the view during automatic
    scrolling, if the option is enabled via the :drag-auto-scroll-p
    initialization argument for the view, and the drag originated within
    this same view.
    
    The 'direction-or-point argument, when it is a keyword, always has one of
    the following four values:  :up, :down, :left, or :right.  This keyword
    indicates the direction the cursor moved, not the direction the view
    should scroll (ie, if the cursor moves up then the method should scroll
    the view down).    
    
    The actual amount to scroll in any direction is set by the
    :drag-scroll-amount initialization argument for the view, which defaults
    to (make-point 12 12).  Automatically scrolling, driven by the
    #'drag-tracking-in-view handler, uses this specialization.
    
    If 'direction-or-point argument is a point the view is scrolled by the
    amount specified.
    
    The method specialized on a view uses MCL's #'set-view-scroll-position
    to perform the actual scrolling.  This method exists only for views
    and subclasses of views, however; most text views in MCL do not support
    this method.  Therefore there is a version specialized on fred-mixin
    that performs a similar function.
    
    This method is called by the MCL drag handlers while tracking a drag.
    
    Returns:  Undefined

 drag-show-drag-hilite ((view drag-view-mixin)
                         &optional
                         (topleft nil) (bottomright nil))

    This method is responsible for displaying the gray border around a drop-
    aware view.  The method will be called only when it is appropriate to
    display the hilite, as per Apple's Human Interface Guidelines. It is
    called by the MCL drag handlers while tracking a drag.
    
    The optional 'topleft and 'bottomright arguments, if defined, should
    be points that define the view's drop area, in the view-window's
    coordinate scheme.
    
    This method calls #'drag-make-drag-hilite-region with the passed
    arguments in order to create the hilite region.
    
    Returns:  t

 drag-tracking-enter-handler ((window window))

    This method is called by the Drag Manager whenever the mouse moves into a
    new window.  The Drag Manager actually allows multiple tracking handlers
    per window, but MCL uses only one and so this method is functionally
    equivalent to #'drag-tracking-enter-view when the view is the window
    itself.
    
    The default version of this method does nothing.
    
    Returns:  Undefined

 drag-tracking-enter-view ((view drag-view-mixin))

    This method is called when the mouse moves into this view from another.
    
    The default version resets some internal data structures and ensures that
    if the previous view had a drag hilite shown, that hilite is taken down.
    
    Returns:  Undefined

 drag-tracking-in-view ((view drag-view-mixin))

    This method is called periodically while a user is dragging within a single
    MCL view.  The default version performs two important functions:  It
    initiates scrolling of the view if the mouse is within the scrolling region
    (see #'drag-make-auto-scroll-region and #'drag-scroll-region) and it
    manages the drag insertion caret (see #'drag-draw-caret-p,
    #'drag-draw-caret, and #'drag-caret-position-from-mouse).
    
    Returns:  Undefined

 drag-tracking-leave-handler ((window window))

    Method is called when the mouse is about to exit the window.  The default
    version does nothing.
    
    Returns:  Undefined

 drag-tracking-leave-view ((view drag-view-mixin))

    Method is called whenever the mouse is about to leave this view for
    another.  The default version helps ensure that the drag hilite is hidden,
    if shown.
    
    Returns:  Undefined

 with-drag-scroll ((delta-h &optional (delta-v nil)) &body body)
    
    This macro implements the Drag Manager's optimized repainting routines
    for the visible drag hilite.  If the view supports actual scrolling,
    typically via MCL's #'set-view-scroll-position, then use this wrapper
    macro to make the scrolling as smooth as possible.
    
    Both arguments should the number of pixels you will scroll the view,
    in each direction.  If 'delta-v is nil, 'delta-h is assumed to be a
    point.
    
    Returns:  the value of the last form in body

 with-saved-drag-hilite ((view) &body body)
    
    This macro hides the drag hilite in the view if one is shown, executes
    body as an explicit progn, then restores the drag hilite if necessary.
    This technique is used to scroll many different views in MCL, where the
    view doesn't actually scroll but is repainted instead.
    
    Returns:  the result of the last form in body

-----------------------------------------------------------------------------
 Initiating drags from a view

There are only two requirements for allowing a drag operation to begin:  The
view must inherit from 'drag-view-mixin, and the method #'drag-selection-p
must return true.  For a quick visual indication that a particular portion
of a view is draggable, the application can supply a different cursor while
the mouse is over that area; see the #'drag-cursor method.

At the Toolbox level, the Drag Manager sees a call to #_NewDrag, followed at
some point later by a call to #_TrackDrag.  Between these two calls, the
originating application is expected to populate the drag with items (the things
that will be dragged), provide a region that will follow the mouse to indicate
those items being dragged, and optionally install a callback routine that will
handle the delivery of promised data to the destination application.

The MCL handlers take care of the setup and disposal of the Drag Manager
information through calls to #'drag-setup-new-drag, #'drag-handle-new-drag,
#'drag-cancel-new-drag and #'drag-cleanup-new-drag.  They also always install
the callback routine for delivery of promised data, so that feature will always
be available if needed.

The handlers call #'drag-add-drag-contents with the originating view as the
argument, in order to allow your application to add drag items and their drag
regions.  In fact, from the application's viewpoint, the call to
#'drag-add-contents may very well be the first indication that a drag has
started.


 Exported methods relating to drag initiation and adding items to a drag

 drag-add-drag-contents ((view drag-view-mixin))

    This method is called when the user begins a drag, and it's purpose is to
    allow the application to add items and their respective drag regions to
    the drag.  You can cancel the entire drag by returning nil.
    
    See the following methods concerning adding items:  #'drag-add-item-flavor,
    #'drag-add-hfs-flavor, #'drag-add-mcl-object-flavor,
    #'drag-promise-item-flavor and #'drag-promise-hfs-flavor.
    
    See these methods for adding the drag hilite regions (the little gray
    region that moves around with the mouse):  #'drag-create-item-bounds and
    #'drag-te-hilite-region.
    
    Note that when this method is called the drag has not yet been started.
    Drag Manager environmental functions, such as those that return information
    regarding the keyboard and mouse, are not yet active.  If you need
    environmental information, use MCL's (or the Toolbox's) usual routines to
    to obtain it.
    
    The default version of this method simply returns nil.
    
    Returns:  t, to indicate that items and their regions have been added;
              nil to request that the drag gesture be canceled.

 drag-add-hfs-flavor ((item-reference integer) (flavor keyword) path
                       &optional (flavor-flag nil))

    This is a convenience method for adding HFS paths to a drag.  The 'path
    argument may be either a Lisp pathname or a string.  It is converted
    to a special :HFSFlavor record the Drag Manager uses for citing HFS
    objects, then the converted record is passed to #'drag-add-item-flavor.
    
    Returns:  t if successful, nil otherwise

 drag-add-item-flavor ((item-reference integer) (flavor keyword)
                        (item-ptr macptr) (item-size integer)
                        &optional (flavor-flag nil))

    Call this method to place an item into the current drag.  Here is a rundown
    of the arguments:

            item-reference   - A number that is unique to this item within this
                               particular drag.
            flavor           - An OSType describing the "flavor" of the data
                               being stored for this item.  An item may have
                               multiple flavors, but each flavor OSType must
                               be unique within that item.
            item-ptr         - A Macintosh pointer containing the data to be
                               placed within the drag.
            item-size        - The size of the Mac pointer, in bytes.
            flavor-flag      - A single keyword indicating additional flavor
                               attributes.
    
    The 'flavor-flag attribute uses keywords instead of the Drag Manager's
    sometimes-cryptic numbering scheme.  Here is the list of keywords as
    well as the constants they match up against:
    
            Drag Manager Constant             MCL Keyword
            ---------------------             -----------
            #$flavorSenderOnly                :internal-only
            #$flavorNotSaved                  :not-saved
            #$flavorSenderTranslated          :sender-translated
            #$flavorSystemTranslated          :system-translated
    
    Note that if you use the :internal-only flag then the drag item flavor
    will not be visible to outside applications, only to the application
    that originated the drag.
    
    The Drag Manager makes a copy of the data within the pointer passed through
    this method, so the caller is free to dispose of the pointer if necessary.
    
    Returns:  t if successful, nil otherwise

 drag-add-mcl-object-flavor ((item-reference integer) (flavor keyword) thing)

    This method can be used to "pass" an MCL object from one view to another
    through the Drag Manager, without having to translate the object into a
    Macintosh pointer and then translate it back.  The 'thing argument should
    will be the MCL object, and it can be virtually anything.  This method
    will place the object in an internal table and deliver to the destination
    view on demand.
    
    Due to the nature of the data being passed around with this method, this
    flavor is not permitted to be seen outside the MCL application.  The
    method purposely scopes the flavor flag to :internal-only.
    
    Returns:  t if successful, nil otherwise

 drag-cancel-new-drag ((view drag-view-mixin) eventrecord)

    This method is called if a drag initiated from within MCL is canceled
    in any way.  The primary purpose of the drag is to possibly force the
    acceptance of the drag anyway, even if the destination application
    has rejected it.
    
    One example of this is dragging items onto application or alias icons in
    the Finder.  The Finder does understand dropping items into folders, the
    Trash icon, and the Desktop, but will summarily reject attempts to drop
    anything onto an application icon.  But that is potentially useful --
    opening a dragged GIF file in another application by dragging it out of
    the window is pretty cool.  One specialization of this method could
    then check to see if the drop occurred on the right kind of Finder icon
    (by checking the drop location) and, if it is right, continue to process
    the drag anyway.
    
    If a specialization of this method returns a non-nil value then this method
    will also provide a "zoom back" animation of the drag region, indicating
    to the user that no, the drag really was accepted.  We were just kidding.
    
    The 'eventrecord argument will contain the standard Macintosh :EventRecord
    record that was used to start the entire drag process.
    
    Returns:  Undefined

 drag-cleanup-new-drag ((view drag-view-mixin) eventrecord)

    This method is called after the entire drag process is over and provides
    an opportunity for the application to clean up any data structures that
    may have been created explicitly for the drag.  The default version
    of this method does nothing.
    
    The 'eventrecord argument will contain the standard Macintosh :EventRecord
    record that was used to start the entire drag process.
    
    Returns:  Undefined

 drag-create-item-bounds ((view simple-view) (item-reference integer)
                           (item-region macptr) &optional (outline-p t))

    This method is used to generate a Macintosh region that the Drag Manager
    will place under the mouse during the drag, providing a visual cue to
    the user of what they're dragging.  Add only one region per drag item,
    even if there are multiple flavors for that item.
    
    Note that at least one flavor of this drag item must have already
    been provided to the Drag Manager before you can call this method.  The
    Drag Manager checks the 'item-reference value to see if it is a valid
    item, and that is determined by the presence of at least one flavor.
    
    The 'outline-p argument, if set to t, will cause the method to post
    an outline of the region instead of the solid region that is normally
    generated.  In many cases, a solid region is the easiest to generate for
    a drag item, but the outlined region is what should be used by the
    Drag Manager.
    
    Returns:  t if successful, nil otherwise

 drag-cursor ((view drag-view-mixin))

    The sole purpose for this method is to provide a cursor that will be
    displayed when the mouse is over a draggable item in a view.  Other factors
    (#'drag-selection-p and some environment states) determine whether to
    change the cursor or not.  The default version of this method returns an
    arrow cursor, which will probably serve for most circumstances.
    
    This method is called by the MCL event system while MCL is the frontmost
    application.
    
    Returns:  a cursor

 drag-fulfill-promise ((view drag-view-mixin) (item-reference integer)
                        (flavor keyword) (target-description t))

    Method is called when the destination application requests a flavor of
    data that was only promised, not actually added to the drag.  If any
    data is promised then this method must be specialized to provide the
    correct behavior (the default version does nothing).  See the methods
    #'drag-set-item-flavor and #'drag-set-hfs-flavor for information on
    providing promised data.
    
    If the method returns anything other than t or #$noErr then it is
    considered an error and the drag will be canceled.
    
    The typical implementation for this method is to specialize both the 'view
    and 'flavor arguments.  The 'flavor argument is specialized using 'eql;
    this allows the programmer to break each "flavor fulfillment" into separate
    methods and possibly be able to combine functions among views (since
    it's more likely that different views will need to handle the same data
    rather than the other way around).
    
    Returns:  t or #$noErr if successful, anything else to indicate an error
              
 drag-get-drop-location-as-path ()

    This function calls #'drag-get-drop-location-descriptor and, if possible,
    translates the drop location into a valid Lisp pathname.
    
    Returns:  the drop location as a pathname if successful, nil
              otherwise

 drag-get-drop-location-descriptor ()

    This function returns the Apple Event descriptor (an AEDesc record)
    describing the drop location.  Callers of this function should not
    dispose of either the pointer or call #_AEDisposeDesc; the descriptor
    is tracked by MCL's drag handlers and will be disposed of when it is
    no longer valid.
    
    Returns:  an :AEDesc record if the drop location has been defined,
              nil otherwise

 drag-get-target-description (&optional (resolve-aliases-p nil))

    This function attempts to summarize the location of a drop.  If it is
    successful, it returns two values in a consed list; here are the
    possible values of both the 'car and 'cdr:

            CAR                CDR
            :same-view         the view that is receiving the drop
            :same-window       the view that is receiving the drop
            :same-application  the view that is receiving the drop
            :finder-trash      pathname to local Trash
            :finder-file       pathname to Finder file or folder
    
    If 'resolve-aliases-p is t, the cdr of the returning value is a
    pathname, and that pathname is an alias, then the method will resolve
    the alias and return the resulting pathname instead.
    
    This function is automatically called, with 'resolve-aliases-p set to nil,
    and the results provided as an argument to #'drag-fulfill-promise.

    Returns:  a consed list describing the drop location if successful, or
              nil otherwise

 drag-handle-new-drag ((view drag-view-mixin) eventrecord)

    This method is called after #'drag-setup-new-drag.  It turns over mouse
    management to the Drag Manager and then calls the cleanup methods
    (#'drag-cancel-new-drag and #'drag-cleanup-new-drag) as needed.
    
    The 'eventrecord argument will contain the standard Macintosh :EventRecord
    record that was used to start the entire drag process.
    
    Returns:  t

 drag-make-auto-scroll-region ((view drag-view-mixin)
                                &optional
                                (topleft nil) (bottomright nil))

    The auto-scroll-region is a region that, when the mouse moves into it,
    will cause the view to scroll in that direction.  It is basically a
    thick border around the inside edge of the view.  Note that only the
    view that initiates a drag may automatically scroll.
    
    The MCL handlers call this method automatically if the view has been
    initialized with :drag-auto-scroll-p set to t.  This default version of
    this method calls #'drag-make-drag-hilite-region to obtain a region
    defining the draggable portion of the view, then generates a border
    six pixels thick.
    
    The mouse-in-region test occurs within #'drag-tracking-in-view calls.
    
    Returns:  a region defining the auto-scroll trigger

 drag-promise-hfs-flavor ((item-reference integer) (file-creator keyword)
                           (file-type keyword) (flavor keyword)
                           (promised-flavor keyword)
                           &key
                           (flavor-flag nil)
                           (finder-flags nil))

    This method allows you to specify that you will create an HFS object on
    demand, and provides the destination application some hints regarding
    what that object will look like.
    
    The 'flavor argument should contain the OSType describing this particular
    flavor, where you're promising the HFS object.  The 'promised-flavor
    argument should contain the OSType that will eventually contain an FSSpec
    record to the HFS object, once that object is made.  A destination
    application will request the 'promised-flavor flavor, if it demands the
    HFS object.  A specialization of #'drag-fulfill-promise, with the 'flavor
    argument scoped with the (eql :|xxxx|) format, is recommended to handle
    this case.
    
    The 'finder-flags argument should be either a single keyword describing a
    Finder flag or a list of keywords.  Follows is a mapping between the
    keywords and the Finder constants used for the flags:
    
            Finder Constant                   MCL Keyword
            ---------------------             -----------
            #$kHasBeenInited                  :has-been-inited
            #$kIsStationery                   :is-stationery
            #$kNameLocked                     :name-locked
            #$kHasBundle                      :has-bundle
            #$kIsInvisible                    :is-invisible
            #$kIsAlias                        :is-alias
    
    The method posts two flavors to the drag item:  1) The flavor containing
    the Finder information and the promised flavor, and 2) a promised flavor
    with no data to back it up.
    
    Returns:  t if successful, nil otherwise

 drag-promise-item-flavor ((item-reference integer) (flavor keyword)
                            &optional (flavor-flag nil))

    Use this method notify the Drag Manager that a certain flavor of the
    item is available, and the data will be delivered on demand.
    
    The 'flavor-flag attribute uses keywords instead of the Drag Manager's
    sometimes-cryptic numbering scheme.  Here is the list of keywords as
    well as the constants they match up against:
    
            Drag Manager Constant             MCL Keyword
            ---------------------             -----------
            #$flavorSenderOnly                :internal-only
            #$flavorNotSaved                  :not-saved
            #$flavorSenderTranslated          :sender-translated
            #$flavorSystemTranslated          :system-translated
    
    Returns:  t if successful, nil otherwise

 drag-selection-p ((view drag-view-mixin) mouse-position)

    This method is used to determine if the mouse is currently over a draggable
    item.  The 'mouse-position argument will be a point indication the mouse
    location in local view coordinates.
    
    What is draggable and what is not is entirely application-defined.  The
    default version of this method returns nil.
    
    Returns:  t if the mouse is currently over a draggable item, nil otherwise

 drag-set-hfs-flavor ((item-reference integer) (flavor keyword) path)

    This is a convenience function for setting a promised HFS flavor by
    using either a Lisp pathname or a string instead of an :FSSpec
    record.  The method does the conversion and calls #'drag-set-item-flavor
    with the results.
    
    Returns:  t if successful, nil otherwise

 drag-set-item-flavor ((item-reference integer) (flavor keyword)
                        (item-ptr macptr) (item-size integer)
                        &optional (offset 0))

    Call this method to provide data that was only promised earlier.  The
    required arguments are the same as for #'drag-add-item-flavor.  The
    optional 'offset argument allows you to install an item "piecemeal"
    into the drag, through repetitive calls to this method.  Most applications
    won't need to use the offset argument.
    
    Returns:  t if successful, nil otherwise

 drag-setup-new-drag ((view drag-view-mixin) eventrecord)

    This method is called when the low-level MCL event system has detected
    the beginnings of a drag gesture.  The 'eventrecord argument will
    contain the standard Macintosh :EventRecord record that was used to
    determine this call.
    
    If this method returns nil then the drag will be canceled.
    
    Returns:  if the drag should continue, t; otherwise nil

 drag-te-hilite-region ((view simple-view) &optional (region (new-region)))

    This method allows easy extraction of the visible hilite region used
    by the Macintosh TextEdit Toolbox routines.  MCL's 'text-edit-dialog-item
    uses TextEdit, while most (if not all) of the other text views don't.
    
    If the 'region argument is already defined to be a region by the caller,
    then the caller is required to eventually dispose of the return region
    via a call to #'dispose-region.
    
    Returns:  A region showing the visible portion of a TextEdit selection.

-----------------------------------------------------------------------------
 Miscellaneous Methods, Functions and Globals

 *drag-error-stream*

    Errors that occur while processing callbacks from the Drag Manager are
    diverted to this stream and _not_ allowed to break into the debugger.
    This is due to the way the Drag Manager manages system processes during
    a drag -- it completely overrides the Process Manager, and Doing The
    Unexpected (eg, going to a debugger) is bad for the system's health.
    Moving error messages to this stream while gracefully returning to the
    Drag Manager seems to be the safest compromise.
    
    This variable is initially bound to *error-output*.

 *drag-manager-present-p*

    This boolean is set by the function #'drag-manager-present-p and it
    indicates whether the Drag Manager is present and available on the system.

 drag-caps-lock-key-p (&optional (when :now))

    Possible values for 'when are:
    
           :before   - the modifiers before the drag started
           :now      - the current modifiers
           :after    - the modifiers after the drop occurred

    Returns:  the state of the Caps Lock key on the keyboard at the specified
              time, or nil if no drag operation is active

 drag-command-key-p (&optional (when :now))

    Possible values for 'when are:
    
           :before   - the modifiers before the drag started
           :now      - the current modifiers
           :after    - the modifiers after the drop occurred

    Returns:  the state of the Command key on the keyboard at the specified
              time, or nil if no drag operation is active

 drag-control-key-p (&optional (when :now))

    Possible values for 'when are:
    
           :before   - the modifiers before the drag started
           :now      - the current modifiers
           :after    - the modifiers after the drop occurred

    Returns:  the state of the Control key on the keyboard at the specified
              time, or nil if no drag operation is active

 drag-copy-requested-p ()

    Function to determine if the user is requesting a copy of the drag data by
    holding down the Option key.  As per the Human Interface Guidelines, the
    user can either hold down the key at the start of the drag or at the time
    of the drop.
    
    Note that if this function is called during a drag, before the drop
    action occurs, then it will return true only if the user had started the
    drag action with the Option key down.  It could still be a copy operation,
    as the user could have the Option key down at the time of the drop.
    
    Returns:  t if the user is holding down the option key, nil otherwise

 drag-get-attributes ()

    Function returns the attributes of the current drag as an integer.
    
    Returns:  the current drag attributes as an integer, or nil if no drag
              operation is active

 drag-get-drag-reference ()

    Returns:  returns an integer representing the reference for the current
              drag, or nil if no drag is occurring

 drag-get-current-view ()

    Returns:  the view the mouse is currently over if a drag is occurring,
              nil if the mouse is not over an MCL view or there is no drag
              going on

 drag-get-keyboard-modifiers (&optional (when :now))

    Function returns the keyboard drag modifiers as an integer, at the time
    specified in the option 'when argument.  Possible values for 'when are:
    
           :before   - the modifiers before the drag started
           :now      - the current modifiers
           :after    - the modifiers after the drop occurred
    
    Returns:  an integer indicating the state of the keyboard modifiers at
              the specified time, or nil if no drag operation is active

 drag-get-source-view ()

    Returns:  the view that initiated the current drag, or nil if either the
              drag did not begin within MCL or no drag is occurring

 drag-left-sender-view-p ()

    Returns:  t if a drag operation originated within MCL and the user has
              ever left the source view, nil otherwise

 drag-left-sender-window-p ()

    Returns:  t if a drag operation originated within MCL and the user has
              ever left the source window, nil otherwise

 drag-manager-present-p ()

    Function makes the necessary gestalt calls to determine if the Drag
    Manager is installed and available on the system.
    
    Returns:  t or nil, indicating the presence of the Drag Manager

 drag-mouse-drop-position ((view drag-view-mixin) &optional (pinned-p nil))

  drag-mouse-drop-position ((view simple-view) &optional (pinned-p nil))

  drag-mouse-drop-position ((view NULL) &optional (pinned-p nil))

    Method returns the mouse coordinates of where a drag ended.  If the method
    is called before the drop occurred, the results are unpredictable.
    
    The 'drag-view-mixin specialization will attempt to return the position
    cited by the method #'drag-caret-position-from-mouse instead of the actual
    mouse position, under the assumption that that is more likely to be used.
    If #'drag-caret-position-from-mouse did not return a value however, then
    the actual mouse position will be used.  The mouse position will be
    returned in the view's local coordinate system.
    
    The 'simple-view specialization returns the mouse position in the view's
    local coordinate system.
    
    The NULL specialization returns the mouse position in global screen
    coordinates.
    
    In all three versions, the 'pinned-p optional argument indicates whether
    the method should return the pinned version of the mouse coordinates,
    which could have been altered via a call to #'drag-set-mouse-position,
    or the actual coordinates.  The MCL handlers use the actual mouse
    position in all their calls.
    
    Returns:  the position of the mouse at the time of the drop, in local or
              global coordinates

 drag-mouse-original-position ((view simple-view))

  drag-mouse-original-position ((view NULL))

    Method returns the mouse coordinates of where a drag began. If a view is
    provided, the coordinates are returned in that view's local  coordinate
    system; if NULL is used instead, the mouse is returned in global
    coordinates.
    
    Returns:  the mouse coordinates of a where drag begin, in either global
              coordinates or a in a view's local coordinates

 drag-mouse-position ((view simple-view) &optional (pinned-p nil))

  drag-mouse-position ((view NULL) &optional (pinned-p nil))

    Method returns the current location of the mouse.  If a view is provided
    then the position will be in the view's local coordinate system.  If NULL
    is used then the position will be in global screen coordinates.
    
    In both versions, the 'pinned-p optional argument indicates whether
    the method should return the pinned version of the mouse coordinates,
    which could have been altered via a call to #'drag-set-mouse-position,
    or the actual coordinates.  The MCL handlers use the actual mouse
    position in all their calls.
    
    Returns:  the current position of the mouse during a drag in local or
              global screen coordinates, or nil if no drag operation is
              active

 drag-mouse-view ((window window) &optional (pinned-p nil))

    The 'pinned-p optional argument indicates whether the method should
    return the pinned version of the mouse coordinates, which could have been
    altered via a call to #'drag-set-mouse-position, or the actual
    coordinates.  The MCL handlers use the actual mouse position in all their
    calls.
    
    Returns:  the view in the window the mouse is currently over, or nil
              if the mouse is outside the window or there is no drag operation
              active

 drag-option-key-p (&optional (when :now))

    Possible values for 'when are:
    
           :before   - the modifiers before the drag started
           :now      - the current modifiers
           :after    - the modifiers after the drop occurred

    Returns:  the state of the Option key on the keyboard at the specified
              time, or nil if no drag operation is active

 drag-proxy ((view simple-view))

    This method allows a view to designate another view as the actual
    recipient of the Drag Manager's attention.  The default version of this
    method returns the value of the 'view argument.
    
    The typical use for this method is to designate a superview (eg, the
    view's window) as the drag source.  Note that you can specialize this
    method on a non-drag-aware view, in order to "hook" the Drag Manager
    into the performing drag operations.  Most applications won't need to
    use this method.
    
    Returns:  a view that should handle the interface to the Drag Manager

 drag-set-mouse-position ((view NULL) (new-position integer))
  
  drag-set-mouse-position ((view simple-view) (new-position integer))
    
    This method allows you to set the pinned mouse position, which is used
    to compute the location of the item dragging region and also the views
    that are "under" the mouse.  It is most useful for constraining mouse
    movement to certain locations, such as within a certain window.
    
    The 'new-position argument should be a point given in the 'view argument's
    coordinate system -- local to a 'simple-view or a global coordinate if
    the view argument is nil.
    
    Returns:  the 'new-position value if successful, nil otherwise

 drag-shift-key-p (&optional (when :now))

    Possible values for 'when are:
    
           :before   - the modifiers before the drag started
           :now      - the current modifiers
           :after    - the modifiers after the drop occurred

    Returns:  the state of the Shift key on the keyboard at the specified
              time, or nil if no drag operation is active

 drag-window-p ((window-or-wptr window))

  drag-window-p ((window-or-wptr macptr))

    Given either a window object or a window pointer, this method returns
    t if the window contains a view that is drag-aware.
    
    Returns:  t if window contains a window that is drag-aware, nil otherwise
    
    Returns:  t if initialization was successful, nil otherwise

 drag-within-sender-application-p ()

    Returns:  t if there is an active drag operation and the user is currently
              dragging with the originating application, nil otherwise

 drag-within-sender-view-p ()

    Returns:  t if there is an active drag operation and the user is currently
              dragging with the originating view, nil otherwise

 drag-within-sender-window-p ()

    Returns:  t if there is an active drag operation and the user is currently
              dragging with the originating window, nil otherwise

 drag-zoom-rectangle ((source-rect macptr) (dest-rect macptr)
                       &key
                       (zoom-steps 12) zoom-speed)

    Method provides animated "zooming" of rectangles, like what happens if
    a drag is rejected.
    
    'source-rect is a :Rect that determines the starting point of the zoom.
    
    'dest-rect is a :Rect that determines the ending point of the zoom.
    
    'zoom-steps determines the number of rectangles to draw between the
    source and destination during the animation.
    
    'zoom-speed provides apparent speeding up or slowing down of the animated
    rectangles as they near the destination.  Possible values for this
    argument are:  :accelerate, :decelerate, and nil (no speed change).
    
    Returns:  t

 drag-zoom-rectangle-points ((source-topleft integer)
                              (source-bottomright integer)
                              (dest-topleft integer)
                              (dest-bottomright integer)
                              &key (zoom-steps 12) zoom-speed)

    Method is same as #'drag-zoom-rectangle but the source and destination
    are defined with points instead of rectangles.
    
    'source-topleft is a point defining the upper-left corner of the source
    rectangle.
    
    'source-bottomright is a point defining the lower-right corner of the
    source rectangle.
    
    'dest-topleft is a point defining the upper-left corner of the
    destination rectangle.
    
    'dest-bottomright is a point defining the lower-right corner of the
    destination rectangle.
    
    'zoom-steps determines the number of rectangles to draw between the
    source and destination during the animation.
    
    'zoom-speed provides apparent speeding up or slowing down of the animated
    rectangles as they near the destination.  Possible values for this
    argument are:  :accelerate, :decelerate, and nil (no speed change).
    
    Returns:  t

 drag-zoom-region ((region macptr) (distance integer)
                    &key
                    (zoom-steps 12) zoom-speed)

    Method performs a function similar to #'drag-zoom-rectangle but allows
    you to define a source region and a distance to move, rather than a
    source and destination rectangle.
    
    'region is the region that will be zoomed.
    
    'distance is a point describing the direction to move the region.
    
    'zoom-steps determines the number of rectangles to draw between the
    source and destination during the animation.
    
    'zoom-speed provides apparent speeding up or slowing down of the animated
    rectangles as they near the destination.  Possible values for this
    argument are:  :accelerate, :decelerate, and nil (no speed change).
    
    Returns:  t

 dragging-now-p ()

    Function returns t if MCL knows that a drag is occurring.
    
    Returns:  t if a drag is occurring, nil otherwise

 init-drag-and-drop ()

    Function that performs all the necessary initializations for MCL drag and
    drop.  The function is called whenever the module is loaded, compiled or
    executed (via 'eval-when).  In addition, the function is pushed onto
    *lisp-startup-functions*.

-----------------------------------------------------------------------------
 Description of included examples

 drag-aware-list.lisp
  
  This is a simple example of making a list view in MCL drag-aware.  This
  view accepts text flavors and items from other drag-aware lists.  You
  can drop items into specific locations in the list (as indicated by the
  drag insertion caret), move items around, and copy items by holding down
  the Option key during the drag.  Dynamic scrolling is also enabled.

 dynamic-views.lisp
  
  Have you ever wanted to be able to move around user controls in a _live_
  application, because they weren't positioned quite right?  This example
  gives you some an idea of what it would take to make that kind of thing
  work.
  
  Once you evaluate the file, evaluate #'dynamic-views to display a sample
  window containing a bunch of different user controls.  All of the controls
  are standard MCL controls, subclassed to allow easier coding, but they have
  no real data or actions to back them up.  You can play with them, though.
  To move a view, simply hold down both the Option and Command keys and start
  dragging.
  
  The example does not allow you to drop views outside of the original window,
  nor does it save the new locations of any repositioned views.  Perhaps an
  enterprising person will add those features later.  There are a number of
  improvements that could be made to this example, but it at least gives you
  some more code to look at.
  
  If this module is executed in MCL 3.9, an icon dialog item will appear; it
  will not appear under older versions of MCL because the associated class is
  not included in the standard application image.

 fred-drag.lisp
  
  This module gives drag and drop capability to editor windows in MCL.  After
  evaluating the file you have six exported functions available:
  
      #'fred-d             - Brings up a drag-aware fred editor window
      #'enable-fred-d      - Alters the default editor window so that all new
                             editor windows are drag-aware
      #'disable-fred-d     - Changes the default editor window back to the
                             MCL default
      #'listener-d         - Brings up a drag-aware listener window
      #'enable-listener-d  - Alters the default listener window so that all new
                             listener windows are drag-aware
      #'disable-listener-d - Changes the default listener window class back to
                             the MCL default
  
  Drag-aware windows can receive standard ASCII text, a text file from the
  Finder, or a promised text file (this usually comes from another application,
  such as Find File).  There is a catch, though:  Most applications that
  promised files usually also deliver text, and there is no clear-cut way
  to determine which flavor the user wants or which flavor is most appropriate.
  Therefore, the following rule applies to dropping items into editor and
  Listener windows:  If the Control key on the keyboard is up, then the window
  will accept only TEXT and HFS flavors; if the Control key is down, the
  window accepts only promised HFS and "real" HFS flavors (no text).
  
  Dragging text within these windows without holding down any modifier keys
  always produces a move, even between two different panes.  The key is that the
  drag occurred within the same window.
  
  If you hold down the Option key while dragging, the text is always copied.
  
  When dragging outside MCL, the editor windows normally promise only a text
  flavor.  However, if you hold down the Control a the start of the drag,
  an HFS file (with the same file creator and type as an MCL text file) will
  be promised.  This type of drag can be dropped onto application and applet
  icons in the Finder.  If you do, a temporary file will be created on the
  desktop (containing the text selected for the drag), the application or
  applet is (optionally) launched, told to open the temporary file, then
  brought to the foreground.  The temporary file is not automatically deleted.
  
  As per the Human Interface Guidelines, it is not legal to drop text within
  itself (within the original selection).  You can, however, drag text from
  one pane into a second pane showing that same text and it works (so much
  for guidelines...).
  
  Automatic scrolling is enabled for the pane originating the drag.

  The fred-drag.lisp module, when evaluated, pushes :drag-aware-fred onto the
  *features* variable in MCL.  This means that you can use the #- and #+
  compiler macros in your code to selectively tune-in or -out certain calls
  based on the environment (such as a call to #'enable-fred-d).
  
  If you want to build an MCL image that contains the "basic" drag and drop
  functionality, follow these steps:
  
     1) Launch your MCL with the "default" configuration.  This is usually
        without your init.lisp file, so move that out of the MCL folder
        before launching.
     
     2) Load the "drag-and-drop.lisp" module, or open the file and evaluate
        its contents.
     
     3) Load the "fred-drag.lisp" module, or open the file and evaluate
        its contents.
     
     4) In the Listener, activate the drag-aware editor windows by evaluating
        (enable-fred-d).  Activate the drag-aware Listener by evaluating
        (enable-listener-d).
     
     5) Save your image to disk by evaluating
        (save-application "MCL-d":init-file "init") -- substitute a better
        filename than "MCL-d" though.

 Netscape Info
  
  A simple view that allows only inbound dragging from Netscape 2.0 or higher.
  It will accept either links from an HTML page, or Bookmark items.  The view
  will show the URL and title for each link; for bookmarks, the view will
  show only the URL (since that's all that is passed through the Drag Manager).
  
  If you hold down the Command key during the drag, the example will prompt
  Netscape into downloading the associated file for the drag items into your
  Desktop folder.  The Drag Manager prevents Netscape from sending the request,
  so it is not possible to wait around for the file before exiting the Drag
  Manager callback chain.  If you need to process the file, you'll have to
  spin off another process (or 'eval-enqueue a function) to watch the
  destination Finder folder for the new file and then process it once it
  appears.
  
  Netscape has an "interesting" way of placing drag data into a drag.  I
  wouldn't recommend it, though.

-----------------------------------------------------------------------------
 Programming Tips

 One of the most useful tools for both debugging data placed into a drag and
  to inspect other applications' drag flavors is DragPeeker.  DragPeeker
  displays a window that you can drag anything into, and it displays as much
  information about each drag item and flavor that it can.  The app is
  available as part of the Macintosh Drag and Drop developer's kit through APDA
  and from Apple's ftp sites.

 When dealing with multiple items in a view (such as items in a list view), be
  careful how the items are determined during a drag.  In most cases, each
  item will be a separate drag item, both incoming and outgoing.  Therefore,
  you should _not_ use something like the item's position in a list as the
  'item-reference argument to Drag Manager calls -- chaos will ensue.  Consider the
  situation of users selecting multiple disjoint items in a list, then dragging
  them to the Finder Trash.  Since each item is dealt with separately, there is
  no reliable way to determine the correct item to delete from the view if they
  were "tagged" using positional item references.  Assign unique IDs or something
  to each item, then go to the trouble of searching for the items using those
  unique IDs each time you need to.

 The Drag Manager takes over the Process Manager during a drag.  It performs
  full contexts switches, which is usually fine because that is what happens
  when the user switches applications.  However, that means you cannot switch
  applications yourself during drag operations and, more importantly, debugging
  gets to be a real problem (since debuggers typically break the program chain
  in order to, well, debug).
  
  Errors within MCL methods and functions that are executed during Drag Manager
  callbacks are Real Problems.  Dropping into the debugger can mess the system
  up so bad you freeze, either right then or sometime later.
  
  To help overcome that problem, the MCL callback routines are all buffered
  with 'handler-case macros that divert all errors to the stream bound to
  *drag-error-stream* and then silence the error to avoid a program break.
  So while you don't have access to the debugger, you don't crash and you
  usually get to see the error message (if the error allows itself to be sent
  to the stream via a #'format call).
  
  You may want to bind *drag-error-stream* to nil, or to a logging file, for
  production applications.

 Due to Lisp's non-local exit feature, it is really a bad idea to throw to a
  tag outside the calling chain during code that has been called by the Drag
  Manager (eg, #'drag-fulfill-promise and anything called by it).  Otherwise,
  the original function called by the Drag Manager may never return and that
  is not a pretty sight.

 Contrary to what the Drag Manager's programmer's manual implies, it is most
  definitely not okay to call any routine that may call #_WaitNextEvent,
  #_GetNextEvent or any similar event-handling toolbox routines while the Drag
  Manager is busy requesting data from you that you promised.  Within this
  module, that means anything that comes through the #'drag-fulfill-promise
  calling chain.  If you do inadvertently call #_WaitNextEvent or its friends
  you will likely freeze your system.  Avoid displaying interactive dialog boxes
  or allowing the system to do so.  Remember that those dialogs can appear at
  times you wouldn't normally think of, such as during alias resolution (in order
  to get to a file located on an unmounted file server).
  
  I was bitten by this problem and it took me a very painful week to solve it.
  In my specific case, I was attempting to solicit information from a remote
  application via Apple Events during #'drag-fulfill-promise.  It turns out
  that #_AESend (the Toolbox routine that actually sends an event) calls
  #_WaitNextEvent if the caller wants to wait for the reply.  75% of the time
  my Mac would freeze solid; the rest of the time it worked beautifully.  The
  difference depended on whether there were pending events for other
  applications -- they wanted CPU time, and attempted a context switch, which
  the Drag Manager had disabled by virtue of its existence.  Bad move.

 This one may be just me, but:  The only time I've been able to successfully
  promise HFS objects is if they are the very first flavor promised for a
  given drag item.  In any position other than the first, Strange Things
  occur.  Files get created and all that, but the destination application
  has no clue that the promised file was actually made and so cannot act on it.
  I've had this problem in two different languages and three different
  development environments, so I don't think the problem is in the compiler.