;;; make-font-sampler.lisp
;;; Wednesday May 29, 2002 SVS
;;; No warranties; use at your own risk.

;;; Makes an editor window that has a sample of all the fonts in the system for MCL.
;;; Adds an item called "Font Sampler" to your Edit menu.
;;; Very handy in OSX because of all the fonts therein.

(in-package :ccl)

(export '(make-font-sampler)) ; so people can call it from the listener too

(defparameter *fwd-double-quote-char* #\322)
(defparameter *bwd-double-quote-char* #\323)
(defparameter *fwd-single-quote-char* #\324)
(defparameter *bwd-single-quote-char* #\325)

(defun make-font-sampler (&optional (fontsize 12))
  "Make a new fred window and fill it with examples of each font in the system."
  (let ((w (make-instance *default-editor-class* :window-title "Font Sampler")))
    (flet ((print-sample-line (fontname)
           (format w "The rain in Spain prefers ~C~A~C over ~F, don~Ct you think?!~%~%"
              *fwd-double-quote-char*
              fontname
              *bwd-double-quote-char*
              pi
              *bwd-single-quote-char*)))
    (ed-set-view-font w (list "Monaco" 9 :srcor :plain '(:color-index 0)))
    (format w "This page was generated by the MCL expression (make-font-sampler ~D), ~%~
               where ~D is the point-size of the samples desired.~%~
               __________________________________________________~%~%" fontsize fontsize)

    (format w "Your System font is ~S:~%" (sys-font-spec)) 
    (ed-set-view-font w *sys-font-spec*) ; just use the cached value
    (print-sample-line (car *sys-font-spec*))
    (terpri w)
    (dolist (font *font-list*)
      (ed-set-view-font w (list "Monaco" 9 :srcor :plain '(:color-index 0)))
      (format w "This is ~S ~D:~%" font fontsize)
      (ed-set-view-font w (list font fontsize :srcor :plain '(:color-index 0)))
      (print-sample-line font))
    (fred-update w)
    (window-set-not-modified (window-key-handler w)))))

;;; Would be nice to have a submenu so you could give it a size argument. Let somebody else do it.
(when (not (find-menu-item (edit-menu) "Font Sampler"))
  (add-menu-items (edit-menu)
                  (make-instance 'window-menu-item
                    :menu-item-title "Font Sampler"
                    :menu-item-action
                    #'(lambda (w)
                        (declare (ignore w)) ; always make a new window
                        (make-font-sampler)))))


(format t "~%You should now have an item called \"Font Sampler\" to your Edit menu.")
(values)
