;;;-*-Mode: LISP; Package: CCL -*-

(in-package :ccl)

;; ANSI compatible make-load-form and make-load-form-saving-slots (for PPC only)
;; Add make-load-form methods for standard-object, structure-object, and class

#|
;; if you are happy with the status quo - do nothing
;; else load/require ansi-make-load-form

Make sure that all your make-load-form methods accept the optional environment arg.
Also make sure that all calls to make-load-form-saving-slots obey the
new argument conventions.

N.B. CLIM contains 32 make-load-form-methods that do not accept the optional environment arg
If you want to use clim, ansi-make-load-form won't work with it.
CLIM now fixed
|#

;;; akh add defvar for *the-...


#+ignore
(eval-when (:load-toplevel :execute :compile-toplevel)
  (when (find-package "CLIM")
    (error "CLIM and ansi-make-load-form are incompatible")))

#+ppc-target
(eval-when (:load-toplevel :execute)
  (fmakunbound 'make-load-form))




#+ppc-target
(let ((*warn-if-redefine* nil)
      (*warn-if-redefine-kernel* nil))

(defvar *the-definition-environment* nil)

(defgeneric make-load-form (object &optional env))

;; from l1-clos
(defun make-load-form-saving-slots (object &key (slots nil slots-p) environment)
  (let* ((class (class-of object))
         (class-name (class-name class))
         (structurep (structurep object))
         (sd (and structurep (require-type (or (and environment
                                                    (let ((defenv (definition-environment environment)))
                                                      (and defenv
                                                           (%cdr (assq class-name (defenv.structures defenv))))))
                                               (gethash class-name %defstructs%))
                                               'vector))))
    (unless (or structurep
                (standard-instance-p object))
      (%badarg object '(or standard-object structure-object)))
    (if slots-p
      (dolist (slot slots)
        (unless (slot-exists-p object slot)
          (error "~s has no slot named ~s" object slot)))
      (setq slots
            (if structurep
              (let ((res nil))
                (dolist (slot (sd-slots sd))
                  (unless (fixnump (car slot))
                    (push (%car slot) res)))
                (nreverse res))
              (mapcar 'slot-definition-name
                      (class-instance-slots (class-of object))))))
    (values
     (let* ((form (gethash class-name *make-load-form-saving-slots-hash*)))
       (or (and (consp form)
                (eq (car form) 'allocate-instance)
                form)
           (setf (gethash class-name *make-load-form-saving-slots-hash*)
                 `(allocate-instance (find-class ',class-name)))))
     ;; initform is NIL when there are no slots
     (when slots
       `(%set-slot-values
         ',object
         ',slots
         ',(let ((temp #'(lambda (slot)
                           (if (slot-boundp object slot)
                             (slot-value object slot)
                             (%unbound-marker-8)))))
             (declare (dynamic-extent temp))
             (mapcar temp slots)))))))

;; existing MCL methods redefined to accept and ignore the optional environment argument

(defmethod make-load-form ((v vinsn-template) &optional env)
  (declare (ignore env))
  (make-load-form-saving-slots v))

(defmethod make-load-form ((p ppc::ppc-opcode) &optional env)
  (declare (ignore env))
  (make-load-form-saving-slots p))

(defmethod make-load-form ((e lexical-environment) &optional env)
  (declare (ignore env))
  nil)

(defmethod make-load-form ((o ppc::ppc-operand) &optional env)
  (declare (ignore env))
  (make-load-form-saving-slots o))

(defmethod make-load-form ((c ctype) &optional env)
  (declare (ignore env))
  `(specifier-type ',(type-specifier c)))

(defmethod make-load-form ((sld shared-library-descriptor)  &optional env)
  (declare (ignore env))
  `(get-shared-library-descriptor ,(sld.name sld)))

(defmethod make-load-form ((slep shared-library-entry-point) &optional env)
  (declare (ignore env))
  `(load-shared-library-entry-point ,(slep.name slep) ,(sld.name (slep.sld slep))))

(defmethod make-load-form ((hash hash-table) &optional env)
  (declare (ignore env))
  (let ((rehashF (function-name (nhash.rehashF hash)))
        (keytransF (nhash.keytransF hash))
        (compareF (nhash.compareF hash))
        (vector (nhash.vector hash))
        (count (nhash.count hash))
        (locked-additions (nhash.locked-additions hash)))
    (flet ((convert (f)
             (if (or (fixnump f) (symbolp f))
               `',f
               `(symbol-function ',(function-name f)))))
      (values
       `(%cons-hash-table
         nil nil nil nil ,(nhash.grow-threshold hash) ,(nhash.rehash-ratio hash) ,(nhash.rehash-size hash))
       `(%initialize-hash-table ,hash ',rehashF ,(convert keytransF) ,(convert compareF)
                                ',vector ,count ',locked-additions)))))

;;;;;;;;;;;;;;;;;;;

;; new ansi methods that type error vs no-applicable-method


(defmethod make-load-form ((object standard-object) &optional environment)
  (declare (ignore environment))
  (error (make-condition 'type-error :datum object :expected-type (cons 'member (find-make-load-form-specializers)))))

(defmethod make-load-form ((object structure-object) &optional environment)
  (declare (ignore environment))
  (error (make-condition 'type-error :datum object :expected-type (cons 'member (find-make-load-form-specializers)))))

(defun find-make-load-form-specializers ()
  (let* ((methods (%gf-methods (fboundp 'make-load-form)))
         (specs (mapcan #'(lambda (m)
                            (let ((name (class-name (car (method-specializers m)))))
                              (when (not (memq name '(standard-object structure-object t)))
                                (list name))))
                        methods)))
    
    specs))

;;; condition same as standard-object - n'est-ce pas

;; this makes little sense to me - if the class isn't defined how can there be an instance of the class.
;; I suppose it could be a compile-time-defclass but getting one's hands on same is rather obscure.
(defmethod make-load-form ((object class) &optional environment)
  (let* ((name (class-name object))
         (class (find-class name nil environment))) 
    (if (eq class object) ; I don't know why it ought to be eq - why not just class exists?
      (values `(class-cell-find-class (load-time-value (find-class-cell ',name t)) t) nil)
      (error "Can't make load form for ~S" object))))

;;;;;;;;;;;;;;;;;;;



;; from nfcomp
(defun fasl-scan-user-form (form &optional (env *the-definition-environment*))
  (multiple-value-bind (load-form init-form) (if t ;(memq :new-make-load-form *features*)
                                               (make-load-form form env)
                                               (make-load-form form))
    (labels ((simple-load-form (form)
               (or (atom form)
                   (let ((function (car form)))
                     (or (eq function 'quote)
                         (and (symbolp function)
                              ;; using fboundp instead of symbol-function
                              ;; see comments in symbol-function
                              (or (functionp (fboundp function))
                                  (eq function 'progn))
                              ;; (every #'simple-load-form (cdr form))
                              (dolist (arg (cdr form) t)
                                (unless (simple-load-form arg)
                                  (return nil))))))))
             (load-time-eval-form (load-form form type)
               (cond ((quoted-form-p load-form)
                      (%cadr load-form))
                     ((self-evaluating-p load-form)
                      load-form)
                     ((simple-load-form load-form)
                      `(,cfasl-load-time-eval-sym ,load-form))
                     (t (multiple-value-bind (lfun warnings)
                                             (or
                                              (gethash load-form *make-load-form-hash*)
                                              (fcomp-named-function `(lambda () ,load-form) nil env))
                          (when warnings
                            (cerror "Ignore the warnings"
                                    "Compiling the ~s ~a form for~%~s~%produced warnings."
                                    'make-load-form type form))
                          (setf (gethash load-form *make-load-form-hash*) lfun)
                          `(,cfasl-load-time-eval-sym (funcall ,lfun)))))))
      (declare (dynamic-extent #'simple-load-form #'load-time-eval-form))
      (let* ((compiled-initform
              (and init-form (load-time-eval-form init-form form "initialization")))
             (info (list (load-time-eval-form load-form form "creation")
                         T              ; scanning-p
                         nil            ; referenced-p
                         compiled-initform  ;initform-info
                         )))
        (puthash form *fasdump-hash* info)
        (fasl-scan-form (%car info))
        (setf (cadr info) nil)        ; no longer scanning load-form
        (when init-form
          (fasl-scan-form compiled-initform))))))


;; from nfcomp
#|  ; change is therein
(defun %compile-file (src output-file verbose print load features
                          save-local-symbols save-doc-strings save-definitions force target 
			        &aux orig-src)
  (setq orig-src (merge-pathnames src))
  (let* ((output-default-type 
          #-ppc-target
          (if (eq target :m68k) *.fasl-pathname* *.pfsl-pathname*)
          #+ppc-target *.fasl-pathname*))
    (setq src (fcomp-find-file orig-src))
    (let* ((newtype (pathname-type src)))
      (when (and newtype (not (pathname-type orig-src)))
        (setq orig-src (merge-pathnames orig-src (make-pathname :type newtype :defaults nil)))))
    (setq output-file (merge-pathnames (if output-file ; full-pathname in case output-file is relative
                                         (full-pathname (merge-pathnames output-file output-default-type) :no-error nil) 
                                         output-default-type)
                                       orig-src))
    (when (physical-pathname-p orig-src) ; only back-translate to things likely to exist at load time
      (setq orig-src (back-translate-pathname orig-src '("home" "ccl"))))
    (let* ((*fasl-non-style-warnings-signalled-p* nil)
           (*fasl-warnings-signalled-p* nil))
      (when (and (not force)
                 (probe-file output-file)
                 (neq (mac-file-type output-file) (if (eq target :m68k) :fasl :pfsl)))
        (unless (y-or-n-dialog (format nil
                                       "Compile destination ~S is a ~A file!  Overwrite it?"
                                       output-file (mac-file-type output-file)))
          (return-from %compile-file nil)))
      (let* ((*features* (append (if (listp features) features (list features)) *features*))
             (*fasl-deferred-warnings* nil) ; !!! WITH-COMPILATION-UNIT ...
             (*fasl-save-local-symbols* save-local-symbols)
             (*fasl-save-doc-strings* save-doc-strings)
             (*fasl-save-definitions* save-definitions)
             (*fcomp-warnings-header* nil)
             (*compile-file-pathname* orig-src)
             (*compile-file-truename* src)
             (*package* *package*)
             (*readtable* *readtable*)
             (*compile-print* print)
             (*compile-verbose* verbose)
             (*fasl-target* target)
             (*compile-time-symbol-macros* *compile-time-symbol-macros*)
             (defenv (new-definition-environment))
             ;(*the-definition-environment* defenv)
             (lexenv (new-lexical-environment defenv))
             (*the-definition-environment* lexenv))
        #-ppc-target (when (eq target :ppc) (pushnew :ppc-target *features*))
        (let ((forms nil))
          (let* ((*outstanding-deferred-warnings* (%defer-warnings nil)))
            (rplacd (defenv.type defenv) *outstanding-deferred-warnings*)
            (setq forms (fcomp-file src orig-src lexenv))
            (setf (deferred-warnings.warnings *outstanding-deferred-warnings*) 
                  (append *fasl-deferred-warnings* (deferred-warnings.warnings *outstanding-deferred-warnings*))
                  (deferred-warnings.defs *outstanding-deferred-warnings*)
                  (append (defenv.defined defenv) (deferred-warnings.defs *outstanding-deferred-warnings*)))
            (when *compile-verbose* (fresh-line))
            (multiple-value-bind (any harsh) (report-deferred-warnings)
              (setq *fasl-warnings-signalled-p* (or *fasl-warnings-signalled-p* any)
                    *fasl-non-style-warnings-signalled-p* (or *fasl-non-style-warnings-signalled-p* harsh))))
          (fasl-scan-forms-and-dump-file forms output-file)))
      (when load (load output-file :verbose (or verbose *load-verbose*)))
      (values (truename (pathname output-file)) 
              *fasl-warnings-signalled-p* 
              *fasl-non-style-warnings-signalled-p*))))
|#
)

#+ppc-target
(pushnew :ansi-make-load-form *features*)
#+ppc-target
(pushnew :ansi-cl *features*)
#+ppc-target
(provide :ANSI-MAKE-LOAD-FORM)

