;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;    Mirus Controls.lisp
;;    Copyright 1998-1999 The MIRUS Group
;;    By: Chris Young
;;
;;    Provided "as is" and free of charge. The MIRUS Group assumes no responsibility
;;    for any harm that may come from using this code.
;;


;; This file restores some but not all color functionality for dialog-items in OS 8.5
;; Some "genius" at Apple thought it was cool to make the long existing color functionality 
;; do nothing in the presence of the appearance manager and themes.
;; (If I want blue text in a dialog item then I want blue text and I don't care what your melody du jour may be)
;; Another 8.5 problem is that the outer frame for the default button persists even when MCL doesn't
;; draw it - so it remains if one changes the default button. This file fixes that too. 
;; see the method set-default-button.

;;The original version from Mirus has been modified by akh@tiac.net to do radio buttons as well and dtrt when mouse click in an
;; item and then drag outside etc, etc. Doesn't add new classes - the originals should work. Thanks Chris.
;; Colors that work for button-dialog-item in OS 8.5 are :frame :text and :body - the body color is a subrectangle within
;;    the button.The :frame color does not affect the bigger outline for a default button in 8.5 whereas it does in prior OS versions.
;;   for check-box-dialog-item :frame and :text
;;   for radio-button-dialog-item :frame and :text
;;   for scroll-bar-dialog-item just :frame 
;; We didn't (won't) fold this into the release because it may have some performance impact when you don't
;;  use colors. I'm not sure whether the impact would be significant. Nah - just include it.
;; It's also possible that this will cease to work in some OS version newer than 8.5.
;; Maybe the old color functionality will even return in some future OS version. 
;; The basic approach here is that there are a bunch of :after methods that overdraw the colored parts with
;; desired colors after the OS has drawn them as it sees fit in it's brain damaged state.

;; akh redraw-color-dialog-item for button - use needs-redraw vs. dont-throb - two separate issues
;;---------- 4.4b2
;; 05/25/01 see "sick of" re osx for check-box and radio-button - 
;; do some with buttons too - kind of ugly - multiline text is also busted in OSX
;; 05/12/01 akh lose the set-default-button redef if carbon
;; 12/06/99 akh check-box and radio-button don't turn text black during #_trackcontrol
;; 10/12/99 forget with-focused-dialog item (it's useless re clipping) - use with-font-focused-view instead
;; 10/10/99 akh button-dialog-item with body color is prettier, use appearance-available-p vs 8.5
;; ----------- 4.3f1c1
;; 02/24/99 akh some fixnum decls, tweak for some positions to do "round up" vs truncate, do 8.5-or-later-p once at startup
;; 02/20/99 akh fix for #\newline - don't ignore color, fix for item-height not = text-height - text is centered vertically
;; 02/19/99 akh fix for big fonts, if text contains #\newline ignore color

(in-package :ccl)

(defvar *check-box-size* #@(16 16))
#| ;; function of default-font can't be right?
  (let* ((d (make-instance 'check-box-dialog-item :dialog-item-text ""))
         (size (view-default-size d)))
    (make-point (point-v size)(point-v size))))
|#

(defvar *radio-button-size* #@(16 16))
#|
  (let* ((d (make-instance 'radio-button-dialog-item :dialog-item-text ""))
         (size (view-default-size d)))
    (make-point (point-v size)(point-v size))))
|#

(defvar *8.5-or-later-p* nil)

(let ((*warn-if-redefine* nil)
      (*warn-if-redefine-kernel* nil))

;; could do the gestalt call once at startup? - OK
#+ignore
(defun software-version-bits ()
  (let ((sysv (gestalt "sysv")))
    (values (ldb (byte 4 8) sysv)
            (ldb (byte 4 4) sysv)  
            (ldb (byte 4 0) sysv))))

#+ignore
(defun software-version-greater= (major minor &optional sub)
  (declare (fixnum major minor))
  (multiple-value-bind (smajor sminor ssub) (software-version-bits)
    (declare (fixnum smajor sminor ssub))
    (or (> smajor major)
        (and (= smajor major)(or (> sminor minor)(and (= sminor minor)(or (null sub)(>= ssub sub))))))))

;; really should use appearance-available-p?
#+ignore
(defun 8.5-or-later-p ()
  *8.5-or-later-p*)

(eval-when (:compile-toplevel :execute)
(defmacro without-the-text-if-osx (item &rest body)
  (let ((item-var (gensym)))
    `(let ((,item-var ,item))
       (if (and (osx-p)(part-color ,item-var :text))
         (let ((pos (view-position ,item-var)))
           (rlet ((rect :rect :topleft pos
                        :bottom (+ (point-v pos)(point-v (view-size ,item-var)))
                        :right (+ (point-h pos) (dialog-item-width-correction ,item-var))))
             ;; don't draw the damn text at all - sick of figuring out where they put it
             (with-clip-rect-intersect rect
               ,@bodY)))
         ,@body)))))
     
;; OR this for all of em - replace existing method
(defmethod view-click-event-handler ((item control-dialog-item) where  &aux ok)
  (let ((handle (dialog-item-handle item)))
    (setq ok (#_TrackControl handle where (%null-ptr))))  
  (if (neq ok 0) (dialog-item-action item)(progn (redraw-color-dialog-item item))))

(defmethod view-click-event-handler ((item button-dialog-item) where
                                     &aux ok)
  (let ((handle (dialog-item-handle item)))
    (setq ok (#_TrackControl handle where (%null-ptr))))
  (when (dont-throb item)(maybe-draw-default-button-outline item))
  (if (neq ok 0) (dialog-item-action item)(progn (redraw-color-dialog-item item))))


#-carbon-compat
(defmethod set-default-button ((dialog window) new-button)
  (let ((default-button (%get-default-button dialog)))
    (unless (eq default-button new-button)
      (without-interrupts
       (when default-button
         (invalidate-view-border default-button t)
         (when (appearance-available-p)  ;; nothing to do with color - magic to tell OS to stop drawing the damn border
           (erase-button-border default-button)))
       (setf (%get-default-button dialog) new-button)
       (when new-button         
         (invalidate-view-border new-button)))))
  new-button)


)  ;; end let

(defmethod view-draw-contents ((item check-box-dialog-item))
  (when (installed-item-p item)
    (without-the-text-if-osx item (call-next-method))))

(defmethod check-box-check ((item check-box-dialog-item))
  (setf (check-box-checked-p item) t)
  (when (installed-item-p item)
    (with-focused-dialog-item (item)
      (without-the-text-if-osx item
        (#_setcontrolvalue (dialog-item-handle item) 1)))))
      

(defmethod check-box-uncheck ((item check-box-dialog-item))
  (setf (check-box-checked-p item) nil)
  (when (installed-item-p item)
    (with-focused-dialog-item (item)
      (without-the-text-if-osx item
        (#_setcontrolvalue (dialog-item-handle item) 0)))))

(defmethod view-draw-contents ((item radio-button-dialog-item))
  (when (installed-item-p item)
    (without-the-text-if-osx item
      (call-next-method))))

(defmethod dialog-item-action ((item button-dialog-item))
  (let* ((body-color (part-color item :body))
         (text-color (part-color item :text))
         (frame-color (part-color item :frame)))                                      
    (when (and (or body-color text-color frame-color (needs-redraw item))(appearance-available-p))
      (redraw-color-dialog-item item))
    (call-next-method)))

(defmethod dialog-item-enable :after ((item button-dialog-item))
  (when (installed-item-p item)
    (redraw-color-dialog-item item)))

(defmethod view-click-event-handler ((item check-box-dialog-item) where)
  (declare (ignore where))
  (without-the-text-if-osx item
    (call-next-method)))

(defmethod view-click-event-handler ((item radio-button-dialog-item) where)
  (declare (ignore where))
  (without-the-text-if-osx item
    (call-next-method)))
  
;; redefs of things in dialogs.lisp

(defmethod radio-button-push ((item radio-button-dialog-item)
                              &aux (cluster (radio-button-cluster item)))
  (let ((my-container (view-container item))
        (handle (dialog-item-handle item)))
    (when my-container
      (do-dialog-items (other-item my-container 'radio-button-dialog-item)
        (if (neq other-item item)
          (if (eq (radio-button-cluster other-item) cluster)
            (radio-button-unpush other-item))))
      (when (and handle (installed-item-p item))
        (radio-button-really-push item)))
    (setf (radio-button-pushed-p item) t)))

(defmethod radio-button-really-push ((item radio-button-dialog-item))
  (with-focused-dialog-item (item (view-container item))
    (without-the-text-if-osx item
      (#_setcontrolvalue (dialog-item-handle item) 1))))
      

(defmethod radio-button-unpush ((item radio-button-dialog-item))
  (let ((handle (dialog-item-handle item)))
    (when (and handle (installed-item-p item))
      (with-focused-dialog-item (item)
        (without-the-text-if-osx item
          (#_setcontrolvalue (dialog-item-handle item) 0))))
    (setf (radio-button-pushed-p item) nil)))

; Color Check Box
; These methods provide for color text for check box dialog items under the MacOS Platinum appearance
; Manager, since the stupid appearance manager turns off all the colors. Note that the check box itself
; remains black. Only the text color is changed. (Now the frame color may also change)
; This was written for our convenience, and I haven't had time
; to do everything to make the dialogs completely the way they were before. But for most purposes, I expect this
; should be fine. It's been tested under a few conditions, but I cannot promise that there might not be some
; instance where it might fail.


(defmethod view-draw-contents :after ((item control-dialog-item))
  (redraw-color-dialog-item item))
    
(defmethod redraw-color-dialog-item ((item check-box-dialog-item))
  (when (and (appearance-available-p)(dialog-item-enabled-p item))
    (let ((text-color (part-color item :text))
          (frame-color (part-color item :frame)))
      (when (or text-color frame-color)
        (with-font-focused-view item  ;; with-focused-dialog-item gets the font right but doesn't clip, with-focused-view doesn't do font but does clip
          (let ((size (view-size item))
                (position #@(0 0))) ;(view-position item)))
            (when (and text-color )
              (let* ((cb-size *check-box-size*))      
                (multiple-value-bind (ff ms)(view-font-codes item)
                  (multiple-value-bind (a d w l)(font-codes-info ff ms)
                    (declare (ignore w d l))
                    (declare (fixnum a))
                    (let* ((text-height (font-codes-string-height (dialog-item-text item) ff ms))
                           (height-delta (half-round-up (- (the fixnum (point-v size)) text-height)))
                           (text-pos (make-point (+ (the fixnum (point-v cb-size)) 2)
                                                 (+  a height-delta))))
                      (declare (fixnum height-delta text-height))
                      (with-fore-color text-color 
                        (draw-multi-line-text item text-pos)))))))
            (when (and frame-color )
              (let* ((cb-size *check-box-size*)
                     (delta (half-round-up (the fixnum (- (the fixnum (point-v size))(the fixnum (point-v cb-size)))))))
                ;(truncate (the fixnum (- (the fixnum (point-v size))(the fixnum (point-v cb-size)))) 2)))
                (setq position (make-point 0 delta))
                (rlet ((rect :rect))
                  (rset rect :rect.topleft position)
                  (rset rect  :rect.botright (add-points position cb-size))
                  (with-fore-color frame-color
                    (#_insetrect rect 2 2)
                    (#_framerect rect)))))))))))

(defmethod check-box-check :after ((item check-box-dialog-item))
  (let ((text-color (part-color item :text))
        (frame-color (part-color item :frame)))
    (when (and (or frame-color text-color) (appearance-available-p))
      (redraw-color-dialog-item item))))

(defmethod check-box-uncheck :after ((item check-box-dialog-item))
  (let ((text-color (part-color item :text))
        (frame-color (part-color item :frame)))
    (when (and (or frame-color text-color) (appearance-available-p))
      (redraw-color-dialog-item item))))

(defmethod dialog-item-enable :after ((item check-box-dialog-item))
  (when (installed-item-p item)
    (redraw-color-dialog-item item)))
          
;;; color radio button

(defmethod redraw-color-dialog-item ((item radio-button-dialog-item))
  (when (and (dialog-item-enabled-p item)(appearance-available-p))
    (let ((text-color (part-color item :text))
          (frame-color (part-color item :frame)))
      (when (or text-color frame-color)
        (with-font-focused-view item
          (let* ((size (view-size item))
                 (position #@(0 0)))
            (when (and text-color )
              (let* ((rb-size *radio-button-size*))      
                (multiple-value-bind (ff ms)(view-font-codes item)
                  (multiple-value-bind (a d w l)(font-codes-info ff ms)
                    (declare (ignore w d l))
                    (declare (fixnum a))
                    (let* ((text-height (font-codes-string-height (dialog-item-text item) ff ms))
                           (height-delta (half-round-up (- (the fixnum (point-v size)) text-height)))
                           (text-pos (make-point (+  (the fixnum (point-v rb-size)) 2)
                                                 (+  a height-delta))))
                      (declare (fixnum text-height height-delta))
                      (with-fore-color text-color
                        (draw-multi-line-text item text-pos)))))))
            (when (and frame-color )
              (let* ((rb-size *radio-button-size*)
                     (delta (half-round-up (- (point-v size)(point-v rb-size))))) ; (truncate (- (point-v size)(point-v rb-size)) 2)))
                (setq position (add-points position (make-point 0 delta)))
                (rlet ((rect :rect))
                  (rset rect :rect.topleft position)
                  (rset rect  :rect.botright (add-points position rb-size))
                  (with-fore-color frame-color
                    (#_insetrect rect 2 2)
                    (#_frameoval rect)))))))))))

(defmethod radio-button-push :after ((item radio-button-dialog-item))
  (let ((text-color (part-color item :text))
        (frame-color (part-color item :frame)))
    (when (and (or frame-color text-color) (appearance-available-p))
      (redraw-color-dialog-item item))))

(defmethod radio-button-unpush :after ((item radio-button-dialog-item))
  (let ((text-color (part-color item :text))
        (frame-color (part-color item :frame)))
    (when (and (or frame-color text-color) (appearance-available-p))
      (redraw-color-dialog-item item))))

(defmethod dialog-item-enable :after ((item radio-button-dialog-item))
  (when (installed-item-p item)
    (redraw-color-dialog-item item)))  


; Color Button
; These methods provide for color for the body of button dialog items under the MacOS Platinum appearance
; Manager. In this case, the color of the button isn't really changed. What happened is a colored rectangle is
; drawn over the body of the button. This way, the button retains it's pseudo-3D look, but is still able to
; communicate color information. The button itself is supposed to look like the types of buttons on some types of
; equiptment which had colors painted on them. I like to think it looks reasonably attractive and yet stays
; consonant with the Platinum look. This was written for our convenience, and I haven't had time
; to do everything to make the dialogs completely the way they were before. But for most purposes, I expect this
; should be fine. It's been tested under a few conditions, but I cannot promise that there might not be some
; instance where it might fail.
 

#|
(defmethod redraw-color-dialog-item ((item button-dialog-item)) 
  (when (and (dialog-item-enabled-p item)(appearance-available-p))
    (let* ((body-color (part-color item :body))
           (text-color (part-color item :text))
           (frame-color (part-color item :frame)))      
      (when (and (or body-color text-color frame-color))
        (let* ((size (view-size item))
               (position (view-position item))
               (v-adjust (ash (point-v size) -2)) ;(truncate (point-v size) 4))
               (h-adjust v-adjust) ; (truncate (point-v size) 4))
               (top (+ (point-v position) v-adjust))
               (bottom (- (+ (point-v position)
                             (point-v size))
                          v-adjust))
               (left (+ (point-h position) h-adjust))
               (right (- (+ (point-h position)
                            (point-h size))
                         h-adjust))
               ) 
          (declare (fixnum v-adjust h-adjust left right))
          (when (and body-color )
            (simple-color-fill-rect  body-color
                             left top right bottom))
          (when frame-color
            (with-item-rect (rect item)
              (with-fore-color frame-color
                (#_frameroundrect rect 8 8))))
          (when (and (or body-color text-color))
            (multiple-value-bind (ff ms)(view-font-codes item)
              (multiple-value-bind (a d w l)(font-codes-info ff ms)
                (declare (ignore w d l))
                (declare (fixnum a))
                ;; we redraw the text in the right color precisely (i hope) on top of the text
                ;; that was previously drawn in the wrong color or that may have been wiped out by
                ;; the body color above
                ;; cant do multi-line because the rom center justifies in this case - phooey
                ;; ok we fixed  it
                (let* (;(str-width (font-codes-string-width (dialog-item-text item) ff ms))
                       ;(width-delta (- (point-h size) str-width))
                       (text-height (font-codes-string-height (dialog-item-text item) ff ms))
                       (height-delta (half-round-up (- (the fixnum (point-v size)) text-height))) ;; was truncate .. 2
                       (text-pos (make-point (+ (point-h position)) 
                                             (+ (the fixnum (point-v position)) a height-delta))))

                  (declare (fixnum height-delta text-height)) 
                  (progn ;with-back-color body-color
                  (with-fore-color text-color
                    (draw-multi-line-centered-text item text-pos))))))))
      ))))
|#


(defmethod needs-redraw ((item button-dialog-item))
  (and (osx-p)
       (let ((text (dialog-item-text item)))
         (and text (position #\newline text)))))

;; this makes the colored body box at least as big as the text and draws text with back-color - much prettier
(defmethod redraw-color-dialog-item ((item button-dialog-item))
  (when (and (dialog-item-enabled-p item)(appearance-available-p))
    (let* ((body-color (part-color item :body))
           (text-color (part-color item :text))
           (frame-color (part-color item :frame))
           (needs-redraw-p (needs-redraw item)))      
      (when (or body-color text-color frame-color needs-redraw-p)
        
        (with-font-focused-view item
          (multiple-value-bind (ff ms)(view-font-codes item)
            (let* ((size (view-size item))
                   (text-height (font-codes-string-height (dialog-item-text item) ff ms))
                   (height-delta (half-round-up (- (the fixnum (point-v size)) text-height)))
                   ) 
              (declare (fixnum text-height height-delta))
              (when (or body-color needs-redraw-p)
                (let* ((v-adjust (ash (point-v size) -2)) ;(truncate (point-v size) 4))
                       (h-adjust v-adjust)
                       (text-width (font-codes-string-width-with-eol (dialog-item-text item) ff ms))
                       (width-delta (ash (- (the fixnum (point-h size)) text-width) -1))
                       (box-width (max text-width (- (point-h size) (ash h-adjust 1))))
                       (box-height (max text-height (- (point-v size) (ash v-adjust 1))))
                       (top (min v-adjust height-delta))
                       (bottom (+ top box-height))
                       (left  (min h-adjust width-delta))
                       (right (+ left box-width)))
                  (declare (fixnum v-adjust h-adjust text-width width-delta box-width box-height left right top bottom))
                  (simple-color-fill-rect  (or body-color *lighter-gray-color*)
                                           left top right bottom)))
              (when frame-color
                (rlet ((rect :rect :topleft 0 :botright size))  ;item-rect (rect item)
                  (with-fore-color frame-color
                    (#_frameroundrect rect 8 8))))
              (when (or body-color text-color needs-redraw-p)
                (multiple-value-bind (a d w l)(font-codes-info ff ms)
                  (declare (ignore w d l))
                  (declare (fixnum a))
                  ;; we redraw the text in the right color precisely (i hope) on top of the text
                  ;; that was previously drawn in the wrong color or that may have been wiped out by
                  ;; the body color above
                  (let* ((text-pos (make-point 0
                                               (+  a height-delta))))
                    (with-back-color (or body-color *lighter-gray-color*)
                      (with-fore-color (or text-color *black-color*)
                        (draw-multi-line-centered-text item text-pos)))))))
            ))))))
;; N positive fixnum
(defun half-round-up (n)
  (declare (fixnum n))
  (ash (the fixnum (1+ n)) -1))



(defun simple-color-fill-rect (color left top right bottom)
  ;(declare (ignore item))
  (rlet ((rect :rect))
    (rset rect :rect.topleft (make-point left top))
    (rset rect  :rect.botright (make-point right bottom))
    (with-fore-color color
      (#_paintrect rect))))



;; for other subclasses - eg Eric Russells appearance manager stuff
(defmethod redraw-color-dialog-item ((item control-dialog-item))
  nil)

#-carbon-compat
(defmethod erase-button-border ((item button-dialog-item))
  ;; hocus pocus
  (with-focused-dialog-item (item)
    (with-item-rect (rect item)
      (without-interrupts
       (rlet ((ps :penstate))
         (#_getpenstate ps)
         (#_insetrect rect -4 -4)
         ;; there is some OS magic going on: pensize has to be 3 3 and have to do frameroundrect (framerect doesn't do the trick)
         (#_pensize 3 3)
         (with-fore-color (or (slot-value (view-window item) 'back-color) *white-color*)
           (#_frameroundrect rect 16 16)
           (#_setpenstate ps)))))
    ;; above wipes out frame color - so get it back if need be
    (redraw-color-dialog-item item)))



(defmethod view-activate-event-handler :after ((item scroll-bar-dialog-item))
  (redraw-color-dialog-item item))

(defmethod redraw-color-dialog-item ((item scroll-bar-dialog-item))
  (let ((frame-color (part-color item :frame)))
    (when (and frame-color (appearance-available-p))
      (with-focused-dialog-item (item)
        (with-item-rect (rect item)
          (with-fore-color frame-color
            (#_framerect rect)))))))

(defmethod view-click-event-handler :after ((item scroll-bar-dialog-item) where)
  (declare (ignore where))
  (redraw-color-dialog-item item))

;; the innards of view-draw-text for 3d-button
(defun draw-multi-line-text (item text-pos)
  (let* ((curstr (dialog-item-text item))
         (eol (position #\newline curstr))
         (length (length curstr))
         (hpos (point-h text-pos)))
    (let* ((v-delta (multiple-value-call 'font-codes-line-height (view-font-codes item)))
           (vpos (point-v text-pos))
           (pos 0))
      (declare (fixnum v-delta vpos pos))
      (loop
        (#_moveto :word hpos :word vpos)
        (with-pstrs ((button-string curstr pos (or eol length)))
          (#_drawstring button-string))
        (when (null eol)(return))
        (setq vpos (+ vpos v-delta))
        (setq pos  (1+ eol))
        (setq eol (position #\newline curstr :start pos))
        ))))

;; not used today
#|
(defmacro preserving-text-mode (&body body)
  (let  ((old-mode (gensym)))
    `(let ((,old-mode #+carbon-compat (require-trap #_getporttextmode (%getport))
                      #-carbon-compat (rref (%getport) :windowrecord.txmode)))
       ,@body
       (require-trap #_textmode ,old-mode))))
|#

(defun draw-multi-line-centered-text (item text-pos)
  (multiple-value-bind (ff ms)(view-font-codes item) 
    (let* ((curstr (dialog-item-text item))
           (eol (position #\newline curstr))
           (width (point-h (view-size item)))
           (hpos (point-h text-pos))
           (str-length (length curstr)))
      (declare (fixnum width str-length hpos))
      (let ((v-delta (font-codes-line-height ff ms))
            (vpos (point-v text-pos))
            (pos 0)) 
        (declare (fixnum v-delta vpos pos))
        (loop
          (let* ((text-width (font-codes-string-width curstr ff ms pos (or eol str-length)))
                 (width-delta (max 0 (ash (the fixnum (- width text-width)) -1)))) ; (truncate (the fixnum (- width text-width)) 2))))
            (declare (fixnum text-width width-delta))
            (#_moveto :word (the fixnum (+ hpos width-delta)) :word vpos)
            (with-pstrs ((button-string curstr pos (or eol str-length)))
              (#_drawstring button-string))
            (when (not eol)(return))
            (setq vpos (+ vpos v-delta))
            (setq pos (1+ eol))
            (setq eol (position #\newline curstr :start pos))))))))

;; not used currently
(defun font-codes-string-width-and-height (string ff ms)
  (multiple-value-bind (width nlines)(font-codes-string-width-with-eol string ff ms)
    (values width (the fixnum (* (the fixnum nlines) (the fixnum (font-codes-line-height ff ms)))))))

(defun font-codes-string-height (string ff ms)
  ;; this counts a trailing eol but I think the OS does too
  (let ((nlines (count #\newline string)))
    (declare (fixnum nlines))
    (the fixnum (* (the fixnum (1+ nlines))(the fixnum (font-codes-line-height ff ms))))))
  

#+ignore
(def-ccl-pointers 8.5-p ()
  (setq *8.5-or-later-p*
        (software-version-greater= 8 5)))

;; blech - text still in wrong place
(def-ccl-pointers disizes ()
  (if (osx-p)
    (progn (setq *check-box-size* #@(18 18))
           (setq *radio-button-size* #@(18 18)))
    (progn (setq *check-box-size* #@(16 16))
           (setq *radio-button-size* #@(16 16)))))


NIL
        
 
#|   

(defun ctest ()
;; in osx it doesnt work to set-part-color from nil to something after the fact for check-box and radio
(setq w (make-instance 'window :window-title "Colors"
                       :window-show nil
                       :view-subviews 
                       (list (setq d(make-instance 'button-dialog-item :dialog-item-text "foo
baqres"
                                                   :default-button t :view-size #@(60 44)))
                             (setq c (make-instance 'check-box-dialog-item :dialog-item-text "FFF
tbd"
                                                    :view-font '("chicago" 24))))))

(set-part-color d :body *blue-color*)
(set-part-color d :text *pink-color*)
(set-part-color d :frame  *orange-color*)
(set-part-color c :text *green-color*)

(setq e (make-instance 'radio-button-dialog-item :view-container w :dialog-item-text "asd
fqq" :view-font '("chicago" 24)))

(set-part-color e :text *green-color*)
(set-part-color e :frame *green-color*)

(set-part-color c :frame *green-color*)
(window-show w))
(ctest)
|#





