; -*- package: io -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;           Universita' degli Studi di Roma Tre
;
;       Dipartimento di Informatica e Automazione
;
;            Via della Vasca Navale, 84 00146
;
;                 Rome, Italy
;
;
; Project: PLASM
;
; Identification: Package Input&Output
;
;
;
;  Version: 1.0
;  Date: 6 Oct 1998
;
;
;
;  Author: Stefano Portuesi e Stefano Fracesi
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;

(provide 'addvrml)

(inpackage IO)

;---------------------------------------------------
;     HELP SYSTEM
;---------------------------------------------------
(defun HELP-VRML ()
  (format t "~%----------------------------------------------")
  (format t "~%---  HELP SYSTEM FOR EXPORTING VRML FILES  ---")
  (format t "~%----------------------------------------------~%")
  (format t "~%\"ASSIGNCOLOR:<polyhedral_complex,color>\" --> polyhedral_complex")
  (format t "~%     Where color is \"<<diffuse_R,diffuse_G,diffuse_B>,")
  (format t "~%                       <specular_R,specular_G,specular_B>,")
  (format t "~%                       <ambient_intensity>,")
  (format t "~%                       <emissive_R,emissive_G,emissive_B>,")
  (format t "~%                       <shininess>,")
  (format t "~%                       <transparency>>\"~%")
  (format t "~%\"RGBCOLOR:<Red,Green,Blue>\" --> color~%")
   (format t "~%\"ASSIGNTEXTURE:<polyhedral_complex,texture>\" --> polyhedral_complex")
  (format t "~%     Where texture is \"<url,")
  (format t "~%                        <repeatS, repeatT>,")
  (format t "~%                        <translationS, translationT>,")
  (format t "~%                        <rotation>,")
  (format t "~%                        <scaleS, scaleT>,")
  (format t "~%                        <centerS, centerT>>~%")
  (format t "~%       and url is a string representing a local or www filename~%")
  (format t "~%\"SIMPLETEXTURE:url\" --> texture~%")
  (format t "~%\"ASSIGNSPOTLIGHT:<pol,<appearance,geometry>\" --> polyhedral_complex")
  (format t "~%\"ASSIGNPOINTLIGHT:<pol,<appearance,geometry>\" --> polyhedral_complex")
  (format t "~%\"ASSIGNDIRLIGHT:<pol,<appearance,geometry>\" --> polyhedral_complex")
  (format t "~%     Where appereance is \"<color,intensity,ambientIntensity,on>\"")
  (format t "~%         and geometry is \"<location,direction,attenuation,")
  (format t "~%                           radius,beamWidth,cutOffAngle>\"~%")
  (format t "~%(io::SET-DEFAULT-LIGHTS T|nil)  sets default scene lights (default T)~%")
  (format t "~%\"ASSIGNCAMERA:<pol_complex,camera>\" --> pol_complex")
  (format t "~%\"ASSIGNORTOCAMERA:<pol_complex,camera>\" --> pol_complex")
  (format t "~%     Where camera is \"<position,orientation,fieldOfView,(unused),description>\"")
  (format t "~%                        (VRML-2)")
  (format t "~%     Where camera is \"<position,orientation,height,focal-distance,(unused)>\"")
  (format t "~%                        (VRML-1)")


  )

;---------------------------------------------------------------------------
; new node re-use (re-use only if geometry and appereance are the same)
; equality test on lists
(defun listeql (a b)
  (if (and (listp a) (listp b) (eql (length a) (length b)))
      (every #'listeql a b)
    (eql a b)))

;equality on (polyhedral-complex,apppearance)
(defun pol-app= (a b)
  (and (eq (first a) (first b))
       (appearance= (second a) (second b))))

;---------------------------------------------------------------------------

(defvar *APPEARANCE-STACK*)

(defvar *STRICT-VRML-1* nil)

(defun SET-STRICT-VRML-1 (bool)
  (declare (special *STRICT-VRML-1*))
  (setf *STRICT-VRML-1* bool))

(defun SET-POL-COLOR (pol-complex color)
  (let ((name (gensym "COLOR-NODE-"))
        (pol-node (kernel::make-pol-node pol-complex)))
    (kernel::set-node-id pol-node name)
    (setf (get (kernel::get-node-id pol-node) 'color) color)
    (list pol-node)
))

(defun SET-POL-TEXTURE (pol-complex texture)
  (let ((name (gensym "TEXTURE-NODE-"))
        (pol-node (kernel::make-pol-node pol-complex)))
    (kernel::set-node-id pol-node name)
    (setf (get (kernel::get-node-id pol-node) 'texture) texture)
    (list pol-node)
))

(defun SET-POL-APPEARANCE (pol-complex color texture)
  (let ((name (gensym "APPEREANCE-NODE-"))
        (pol-node (kernel::make-pol-node pol-complex)))
    (kernel::set-node-id pol-node name)
    (setf (get (kernel::get-node-id pol-node) 'color) color)
    (setf (get (kernel::get-node-id pol-node) 'texture) texture)
    (list pol-node)
))

(defun RESET-APPEARANCE ()
  (declare (special *APPEARANCE-STACK*))
  (setf *APPEARANCE-STACK* nil))

(defun GET-POL-NODE-COLOR (pol-node)
  (get (kernel::get-node-id pol-node) 'color))

(defun GET-POL-NODE-TEXTURE (pol-node)
  (get (kernel::get-node-id pol-node) 'texture))

(defun GET-POL-NODE-APPEARANCE (pol-node)
  (let ((color (get (kernel::get-node-id pol-node) 'color))
        (texture (get (kernel::get-node-id pol-node) 'texture)))
    (if (or color texture) (list color texture) nil)
))

(defun color? (color)
  (if (and (listp color) (eql (length color) 6))
      (if (and (eql (length (nth 0 color)) 3)
               (eql (length (nth 1 color)) 3)
               (eql (length (nth 2 color)) 1)
               (eql (length (nth 3 color)) 3)
               (eql (length (nth 4 color)) 1)
               (eql (length (nth 5 color)) 1))
          T nil)
  nil))

(defun texture? (tex)
  (if (and (listp tex) (eql (length tex) 6))
      (if (and (stringp (nth 0 tex))
               (eql (length (nth 1 tex)) 2)
               (eql (length (nth 2 tex)) 2)
               (eql (length (nth 3 tex)) 1)
               (eql (length (nth 4 tex)) 2)
               (eql (length (nth 5 tex)) 2))
               T nil)
  nil))

(defun appearance? (app)
  (and (color? (first app)) (texture? (second app))))

(defun appearance= (a b)
  (and (color= (first a) (first b))
       (texture= (second a) (second b))))

(defun color= (a b)
  (listeql a b))

(defun texture= (a b)
  (and (string= (first a) (first b))
       (listeql (cdr a) (cdr b))))

(defun INSERT-INTO-APPEARANCE-STACK (appearance)
  (declare (special *APPEARANCE-STACK*))

  (setf *APPEARANCE-STACK* (cons appearance *APPEARANCE-STACK*)))

(defun REMOVE-FROM-APPEARANCE-STACK ()
  (declare (special *APPEARANCE-STACK*))
  (setf *APPEARANCE-STACK* (rest *APPEARANCE-STACK*)))

(defun  GET-FROM-APPEARANCE-STACK()
  (declare (special *APPEARANCE-STACK*))
  (list (loop for i in *APPEARANCE-STACK*
          when (first i) return (first i))
        (loop for i in *APPEARANCE-STACK*
                   when (second i) return (second i))))

; appereance node re-use

(defvar *JUST-SEEN-APPEARANCE*)

(defun RESET-JUST-SEEN-APPEARANCE ()
  (declare (special *JUST-SEEN-APPEARANCE*))
  (setf *JUST-SEEN-APPEARANCE* nil))

(defun appearance-just-written-for-this (appearance)
  (declare (special *JUST-SEEN-APPEARANCE*))
  (cdr (assoc appearance *JUST-SEEN-APPEARANCE* :test #'appearance=)))

(defun set-as-appearance-just-written (appereance name)
   (declare (special *JUST-SEEN-APPEARANCE*))
   (setq *JUST-SEEN-APPEARANCE* (acons appereance name *JUST-SEEN-APPEARANCE*))
   name)

;lights

(defun light? (light)
  (if (and (listp light) (eql (length light) 2))
      (let ((appear (first light))
            (geom (second light)))
        (and
         (if (nth 0 appear)
             (eql (length (nth 0 appear)) 3) T)
         (if (nth 1 appear)
             (eql (length (nth 1 appear)) 1) T)
         (if (nth 2 appear)
             (eql (length (nth 2 appear)) 1) T)
         (if (nth 3 appear)
             (eql (length (nth 3 appear)) 1) T)
         (if (nth 0 geom)
             (eql (length (nth 0 geom)) 3) T)
         (if (nth 1 geom)
             (eql (length (nth 1 geom)) 3) T)
         (if (nth 2 geom)
             (eql (length (nth 2 geom)) 3) T)
         (if (nth 3 geom)
             (eql (length (nth 3 geom)) 1) T)
         (if (nth 4 geom)
             (eql (length (nth 4 geom)) 1) T)
         (if (nth 5 geom)
             (eql (length (nth 5 geom)) 1) T)))
    nil))

(defvar DEFAULT-LIGHT T)

(defun SET-DEFAULT-LIGHTS (boole)
  (declare (special DEFAULT-LIGHT))
  (setf DEFAULT-LIGHT boole))

(defun SET-POL-LIGHT (pol-complex light)
  (let ((name (gensym "LIGHT-NODE-"))
        (pol-node (kernel::make-pol-node pol-complex)))
    (kernel::set-node-id pol-node name)
    (setf (get (kernel::get-node-id pol-node) 'light) light)
    (list pol-node)
))

(defun GET-POL-NODE-LIGHT (pol-node)
  (get (kernel::get-node-id pol-node) 'light))

;camera

(defun SET-POL-CAMERA (pol-complex camera)
  (let ((name (gensym "CAMERA-NODE-"))
        (pol-node (kernel::make-pol-node pol-complex)))
    (kernel::set-node-id pol-node name)
    (setf (get (kernel::get-node-id pol-node) 'camera) camera)
    (list pol-node)
))

(defun GET-POL-NODE-CAMERA (pol-node)
  (get (kernel::get-node-id pol-node) 'camera))

(defun camera? (camera)
  (if (listp camera)
        (and
         (if (nth 0 camera)
             (eql (length (nth 0 camera)) 3) T)
         (if (nth 1 camera)
             (eql (length (nth 1 camera)) 4) T)
         (if (nth 2 camera)
             (eql (length (nth 2 camera)) 1) T)
         (if (nth 3 camera)
             (eql (length (nth 3 camera)) 1) T)
         (if (nth 4 camera)
             (stringp (nth 4 camera)) T))
    nil))
;----------------------------------------------------

;----------------------------------------------------
; slightly modified for set-unset default light
(defun WRITE-OVERTURE-VRML-FILE (infostring outfile)
  (cond
   ((VRML-VER-1?)
    (format outfile "#VRML V1.0 ascii~%~%")
    (format outfile "Separator {~%")
    (format outfile "  Info { string \"~A\" }~%" infostring)
    (format outfile "  DEF BackgroundColor Info { string \".3 .7 .9\" }~%")
    (light-description outfile)
    (format outfile "
  PerspectiveCamera {
    position        0 0 4
    orientation     0 0 1  0
    focalDistance   10.84
  }~%"))
   ((VRML-VER-2?)
    (format outfile "#VRML V2.0 utf8~%~%")
    (format outfile "Group {~%")
    (format outfile " children [~%")
    (format outfile " WorldInfo {~%")
    (format outfile "   info [  \"~A\" ]~%" infostring)
    (format outfile " }~%")
    (format outfile "
  Background  {
     skyColor    .3 .7 .9
     groundColor .3 .7 .9
  }~%")
    (if DEFAULT-LIGHT (light-description outfile))
    )))

;write light descriptor
(defun WRITE-LIGHT (light outfile)
  (let (
        (type (first light))
        (appereance (second light))
        (geometry (third light)))

    (cond
     ((eql type 0)
      (format outfile (ins-tab "" (tab-count) "SpotLight {~%"))
      (if geometry
          (progn
            (if (nth 0 geometry)
                (format outfile (ins-tab "" (tab-count)
                                         "  location ~{ ~F~}~%")
                        (nth 0 geometry)))
            (if (nth 1 geometry)
                (format outfile (ins-tab "" (tab-count)
                                         "  direction ~{ ~F~}~%")
                        (nth 1 geometry)))
            (if (and (nth 2 geometry) (VRML-VER-2?))
                (format outfile (ins-tab "" (tab-count)
                                         "  attenuation ~{ ~F~}~%")
                        (nth 2 geometry)))

            (if (and (nth 3 geometry) (VRML-VER-2?))
                (format outfile (ins-tab "" (tab-count)
                                         "  radius ~{ ~F~}~%")
                        (nth 3 geometry)))
            (if (and (nth 4 geometry) (VRML-VER-2?))
                (format outfile (ins-tab "" (tab-count)
                                         "  beamWidth ~{ ~F~}~%")
                        (nth 4 geometry)))
            (if (nth 5 geometry)
                (format outfile (ins-tab "" (tab-count)
                                         "  cutOffAngle ~{ ~F~}~%")
                        (nth 5 geometry))))))
     ((eql type 2)
       (format outfile (ins-tab "" (tab-count) "DirectionalLight {~%"))
       (if geometry
          (progn
            (if (nth 1 geometry)
                (format outfile (ins-tab "" (tab-count)
                                         "  direction ~{ ~F~}~%")
                        (nth 1 geometry))))))
     ((eql type 1)
       (format outfile (ins-tab "" (tab-count) "PointLight {~%"))
       (if geometry
          (progn
            (if (nth 0 geometry)
                (format outfile (ins-tab "" (tab-count)
                                         "  location ~{ ~F~}~%")
                        (nth 0 geometry)))
            (if (and (nth 2 geometry)  (VRML-VER-2?))
                (format outfile (ins-tab "" (tab-count)
                                         "  attenuation ~{ ~F~}~%")
                        (nth 2 geometry)))
            (if (nth 3 geometry)
                (format outfile (ins-tab "" (tab-count)
                                         "  radius ~{ ~F~}~%")
                        (nth 3 geometry)))))))
    (if appereance
        (progn
          (if (nth 0 appereance)
              (format outfile (ins-tab "" (tab-count)
                                       "  color ~{ ~F~}~%")
                      (nth 0 appereance)))
          (if (nth 1 appereance)
              (format outfile (ins-tab "" (tab-count)
                                       "  intensity ~{ ~F~}~%")
                      (nth 1 appereance)))
          (if (and (nth 2 appereance) (VRML-VER-2?))
              (format outfile (ins-tab "" (tab-count)
                                       "  ambientIntensity ~{ ~F~}~%")
                      (nth 2 appereance)))
          (if (nth 3 appereance)
              (format outfile (ins-tab "" (tab-count)
                                       "  on ~A~%")
                      (if (eql (nth 3 appereance) 0) "FALSE" "TRUE")))))
  (format outfile (ins-tab "" (tab-count)"}~%") )))


;write camera descriptor
(defun WRITE-CAMERA (camera outfile)
  (let (
        (type (first camera))
        (data (second camera)))
    (if (VRML-VER-1?)
        (if (eql type 0)
            (format outfile (ins-tab "" (tab-count)
                                 "PerspectiveCamera {~%"))
          (format outfile (ins-tab "" (tab-count)
                                 "OrthographicCamera {~%")))
      (format outfile (ins-tab "" (tab-count)
                                 "Viewpoint { ~%")))

    (if (nth 0 data)
        (format outfile (ins-tab "" (tab-count)
                                 "  position ~{ ~F~}~%")
                (nth 0 data)))
    (if (nth 1 data)
        (format outfile (ins-tab "" (tab-count)
                                 "  orientation ~{ ~F~}~%")
                (nth 1 data)))
    (if (nth 2 data)
        (if (VRML-VER-1?)
            (format outfile
                    (if (eql type 0)
                        (ins-tab "" (tab-count) "  heightAngle ~{ ~F~}~%")
                      (ins-tab "" (tab-count) "  height ~{ ~F~}~%"))
                    (nth 2 data))
          (format outfile (ins-tab "" (tab-count) "  fieldOfView ~{ ~F~}~%")
                  (nth 2 data))))
    (if (VRML-VER-1?)
        (if (nth 3 data)
            (format outfile (ins-tab "" (tab-count)
                                     "  focalDistance ~{ ~F~}~%")
                    (nth 3 data))))
    (if (VRML-VER-2?)
        (format outfile (ins-tab "" (tab-count)
                                 "  description \"~A\"~%")
                (if (nth 4 data) (nth 4 data)
                  (gensym "ViewPoint"))))

    (format outfile (ins-tab "" (tab-count)"}~%") )))

;--------------------------------------------------------------
;modified to omogenize vrml-1 behavior to vrml-2
(defun WRITE-OVERTURE-VRML-STRUCTURE (name outfile &key (open-group nil) (plasm-name? nil))
  (inc-tab-counter)
  (cond
   ((VRML-VER-1?)
    (format outfile
            (ins-tab "" (tab-count) "DEF ~A Separator {~%")
            name)
;    (rivedere materiali vrml-1)
    )
;    (if plasm-name? (WRITE-MATERIAL-APPEARANCE-NODE outfile)))
   ((and open-group (VRML-VER-2?))
    (format outfile
            (ins-tab ""
                     (tab-count)
                     "DEF ~A Group { children [~%") name))
   ((VRML-VER-2?)
    (format outfile
            (ins-tab "" (tab-count) "DEF ~A ") ; what follows this is dependent
                                               ; from matrix transform of Plasm node
            name))))
;---------------------------------------------------------
; changed to reset new global variables
(defun SET-RESET-ENVIRONMENT (&optional matrix-dim)
    (CREATE-HASH-TABLE-PLASM-NAME)
    (RESET-APPEARANCE-NO-WRITTEN)
    (RESET-NAMES)
    (RESET-APPEARANCE)
    (RESET-JUST-SEEN-APPEARANCE)
    ;(RESET-CTM matrix-dim)
    (reset-tab-counter)
    (reset-hash-table-name))
;---------------------------------------------------------
;changed for new node re-use and added materials and lights
(defun WRITE-POL-COMPLEX-TO-VRML (dag outfile &key (name-to-use nil)

                                                  (plasm-name? nil))

  ; new node re-use: reuse only if object and appereance are the same
  (cond
   ((get-hash-table-name (list dag (get-from-appearance-stack)))
    ; if it returns T then this pol-complex was just reached

    (write-use-node (get-hash-table-name (list dag (get-from-appearance-stack))) outfile))

   ((polyhedra-complex? dag)
    (let ((name (set-as-just-seen (list dag (get-from-appearance-stack))
                                  ;if dag is plasm named object name will be its plasm name

                                  :name name-to-use
                                  ;if dag comes from a skeleton operation name must be
                                  ; the name of its origin
                                  :use-name-as-radix (not name-to-use))))

      ; obsoleted by new material system (FP)
      ;(insert-into-names (get-hash-table-plasm-name dag))

      (if (= 1 (length dag)) ; dag is a pol-complex with a single pol node
        (write-pol-complex-to-vrml (first dag)
                                   outfile
                                   :name-to-use name
                                   :plasm-name? (get-hash-table-plasm-name dag))
        (progn
          (write-overture-vrml-structure name
                                         outfile
                                         :open-group T
                                         :plasm-name?
                                         (get-hash-table-plasm-name dag))
          (mapcar  #'(lambda (node)
                       (write-pol-complex-to-vrml
                        node
                        outfile))
                   dag)
          (write-closure-vrml-structure name outfile :close-group T))
        )
      ; obsoleted by new material system (FP)
      ;(if (get-hash-table-plasm-name dag)
      ;  (remove-from-names))))
      ))

   ((polyhedral-node? dag)
    (let ((pol (get-node-pol-complex dag))
          (matrix (get-node-matrix dag))
          (node-rep (get-node-rep dag))
          (appearance (get-pol-node-appearance dag))
          (name (if name-to-use
                  (set-as-just-seen (list dag (get-from-appearance-stack))
                                    :name name-to-use) ; this names a polyhedral node
                  (set-as-just-seen (list dag (get-from-appearance-stack))
                                    :name 'PART
                                    :use-name-as-radix T))))

      ;new apperance system
      (declare (special *GENERAL-MATERIAL-NODE*))
      (if appearance
          (if (or (vrml-ver-2?) (not *STRICT-VRML-1*))
              (insert-into-appearance-stack appearance)
            (if (first appearance)
                (WRITE-APPEARANCE-NODE (first appearance) (second appearance)  outfile )
              (WRITE-APPEARANCE-NODE *GENERAL-MATERIAL-NODE* (second appearance)  outfile ))))

      (if (polyhedron? pol)
        (write-vrml-polyhedron-representation pol
                                              matrix
                                              name
                                              outfile)

       ;(polyhedra-complex? pol) predicate returns T
        (progn
          (write-overture-vrml-structure name
                                         outfile
                                         :open-group nil
                                         :plasm-name? plasm-name?)


          ;(set-current-transformation-matrix (if (eql node-rep 'kernel::on-faces)
          ;                                     (mat-inv matrix)
          ;                                     matrix))

          (write-transformation-vrml-structure matrix node-rep outfile)

          ;if it's writing for ver 2.0 after a transform it needs a children node opening
          (if (vrml-ver-2?) (format outfile (ins-tab "" (tab-count) "children [~%")))
          ;check if there is a camera  and write it
          (if (get-pol-node-camera dag)
              (write-camera (get-pol-node-camera dag) outfile))
          ;check if there is some light and write it
          (if (get-pol-node-light dag)
              (write-light (get-pol-node-light dag) outfile))
          (write-pol-complex-to-vrml pol outfile)
          (write-closure-vrml-structure name outfile :close-group nil)

          ;this function closes the children node (only for V2.0)

          ;(set-current-transformation-matrix (if (eql node-rep 'kernel::on-faces)
          ;                                     matrix
          ;                                     (mat-inv matrix)))
          ))
      (if (and (get-pol-node-appearance dag)
               (or (vrml-ver-2?) (not *STRICT-VRML-1*)))
          (remove-from-appearance-stack))
      ))))


;--------------------------------------------------------------------
;new matrial system (FP) (not complete, work in progress)

(setf *GENERAL-MATERIAL-NODE* '((0.8 0.8 0.8) (0 0 0) (0.2) (0 0 0) (0.2) (0)))

(defun WRITE-OVERTURE-VRML-POLYHEDRON (polname outfile)
  (cond
   ((vrml-ver-1?)
    (format outfile
            (ins-tab "" (tab-count) "DEF ~A Separator { #Polyhedron~%")
            polname)
    ;(if (oiv-format?) (write-shapehints-node outfile))
    (inc-tab-counter)
    ;modificata ulteriormente da v 3.0
    (if (not *STRICT-VRML-1*) (write-material-appearance-node outfile))
    )
   ((vrml-ver-2?)
    (format outfile
            (ins-tab "" (tab-count) "DEF ~A Shape { #Polyhedron~%")
            polname)
    (inc-tab-counter)
    (write-material-appearance-node outfile)
    (format outfile (ins-tab "" (tab-count) "geometry~%"))))
  (inc-tab-counter))

(defun WRITE-MATERIAL-APPEARANCE-NODE (outfile)
   (declare (special *GENERAL-MATERIAL-NODE*))
   (let ((appearance (get-from-appearance-stack)))
     (if (first appearance)
       (write-appearance-node (first appearance) (second appearance) outfile)
       (write-appearance-node *GENERAL-MATERIAL-NODE* (second appearance) outfile))))

(defun WRITE-APPEARANCE-NODE (material texture outfile)
  (cond
   ((vrml-ver-1?)
    (if material
        (let ((name-mat (appearance-just-written-for-this (list material nil))))
          (if name-mat (format outfile (ins-tab "" (tab-count) "USE ~A ~%") name-mat)
            (progn
              (format outfile (ins-tab "" (tab-count) "DEF ~A Material {~%")
                      (set-as-appearance-just-written (list material nil) (gensym "MATERIAL")))
	   (format outfile (ins-tab "" (tab-count) "    diffuseColor~{ ~F~}~%") (first material))
	   ;(format outfile (ins-tab "" (tab-count) "    ambientIntensity~{ ~F~}~%") (third material))
	   (format outfile (ins-tab "" (tab-count) "    specularColor~{ ~F~}~%") (second material))
	   (format outfile (ins-tab "" (tab-count) "    emissiveColor~{ ~F~}~%") (fourth material))
	   (format outfile (ins-tab "" (tab-count) "    shininess~{ ~F~}~%") (fifth material))
	   (format outfile (ins-tab "" (tab-count) "    transparency~{ ~F~}~%") (sixth material))
	   (format outfile (ins-tab "" (tab-count) "  }~%"))
              ))))
    (if texture
        (let ((name-tex (appearance-just-written-for-this (list nil texture))))
          (if name-tex
              (progn
                (format outfile (ins-tab "" (tab-count) "USE ~A ~%") name-tex)
                (format outfile (ins-tab "" (tab-count) "USE ~A-TRANSFORM ~%") name-tex))
            (let ((name-tex (set-as-appearance-just-written (list nil texture) (gensym "TEXTURE"))))
              (format outfile (ins-tab "" (tab-count) "DEF ~A Texture2 {~%") name-tex)

              (format outfile (ins-tab "" (tab-count) "  filename \"~A\" ~%") (first texture))
              (format outfile (ins-tab "" (tab-count) "  wrapS ~A ~%")
                      (if (eql (first (second texture)) 0) "CLAMP" "REPEAT"))
              (format outfile (ins-tab "" (tab-count) "  wrapT ~A ~%")
                      (if (eql (second (second texture)) 0) "CLAMP" "REPEAT"))
              (format outfile (ins-tab "" (tab-count) "  }~%"))
              (format outfile (ins-tab "" (tab-count) "DEF ~A-TRANSFORM Texture2Transform  {~%") name-tex)

              (format outfile (ins-tab "" (tab-count) "  translation~{ ~F~}~%") (third texture))
              (format outfile (ins-tab "" (tab-count) "  rotation~{ ~F~}~%") (fourth texture))
              (format outfile (ins-tab "" (tab-count) "  scaleFactor~{ ~F~}~%") (fifth texture))
              (format outfile (ins-tab "" (tab-count) "  center~{ ~F~}~%") (sixth texture))
              (format outfile (ins-tab "" (tab-count) "}~%"))
              )))))

    ((vrml-ver-2?)

     (let ((name (appearance-just-written-for-this (list material texture))))
       (if name (format outfile (ins-tab "" (tab-count) "appearance USE ~A ~%") name)
         (progn
           (format outfile (ins-tab "" (tab-count) "appearance DEF ~A Appearance {~%")
                   (set-as-appearance-just-written (list material texture) (gensym "APPEREANCE")))
           (format outfile (ins-tab "" (tab-count) "  material Material {~%"))
           (format outfile (ins-tab "" (tab-count) "    diffuseColor~{ ~F~}~%") (first material))
           (format outfile (ins-tab "" (tab-count) "    ambientIntensity~{ ~F~}~%") (third material))
           (format outfile (ins-tab "" (tab-count) "    specularColor~{ ~F~}~%") (second material))
           (format outfile (ins-tab "" (tab-count) "    emissiveColor~{ ~F~}~%") (fourth material))
           (format outfile (ins-tab "" (tab-count) "    shininess~{ ~F~}~%") (fifth material))
           (format outfile (ins-tab "" (tab-count) "    transparency~{ ~F~}~%") (sixth material))
           (format outfile (ins-tab "" (tab-count) "  }~%"))
           (if texture
               (progn
                 (format outfile (ins-tab "" (tab-count) "  texture ImageTexture {~%"))
                 (format outfile (ins-tab "" (tab-count) "    url \"~A\" ~%") (first texture))
                 (format outfile (ins-tab "" (tab-count) "    repeatS ~A ~%")
                         (if (eql (first (second texture)) 0) "FALSE" "TRUE"))
                 (format outfile (ins-tab "" (tab-count) "    repeatT ~A ~%")
                         (if (eql (second (second texture)) 0) "FALSE" "TRUE"))
                 (format outfile (ins-tab "" (tab-count) "  }~%"))
                 (format outfile (ins-tab "" (tab-count) "  textureTransform TextureTransform {~%"))
                 (format outfile (ins-tab "" (tab-count) "    translation~{ ~F~}~%") (third texture))
                 (format outfile (ins-tab "" (tab-count) "    rotation~{ ~F~}~%") (fourth texture))
                 (format outfile (ins-tab "" (tab-count) "    scale~{ ~F~}~%") (fifth texture))
                 (format outfile (ins-tab "" (tab-count) "    center~{ ~F~}~%") (sixth texture))
                 (format outfile (ins-tab "" (tab-count) "  }~%"))
                 ))
           (format outfile (ins-tab "" (tab-count) "}~%"))
        ))))))

; end new material system
;---------------------------------------------------

(defun RESET-HASH-TABLE-NAME ()
  (declare (special *JUST-SEEN-OBJECT*))
  (setf *JUST-SEEN-OBJECT* nil))

(defun SET-HASH-TABLE-NAME (obj name)
  (declare (special *JUST-SEEN-OBJECT*))
  (setq *JUST-SEEN-OBJECT* (acons obj name *JUST-SEEN-OBJECT*))
  name)

;the test should be changed in something more specific and faster
(defun GET-HASH-TABLE-NAME (obj)
  (declare (special *JUST-SEEN-OBJECT*))
  (cdr (assoc obj *JUST-SEEN-OBJECT* :test #'pol-app=)))

(defun SET-AS-JUST-SEEN (obj &key (name 'model) (use-name-as-radix nil))
  (declare (special *JUST-SEEN-OBJECT*))
  (let ((plasm-name (GET-HASH-TABLE-PLASM-NAME (first obj))))
        (set-hash-table-name obj
                             (cond
                              (plasm-name (gensym (string plasm-name )))
                              (T (if use-name-as-radix
                                     (gensym (string (if name name 'MODEL)))
                                   (string name)))))))

