; -*- package: plasm -*-
;==========================================================================
;
;                           ANIM.LSP File
;
;==========================================================================

(provide 'anim)
(inpackage plasm)


(export '(
          
          ))


(defun EXECUTE-CLISP-SHELL (&rest strings)
  (let (exit
        (standard-output *standard-output*)
        (error-output *error-output*)
        (command-string (eval `(concatenate 'string ,@strings " > /dev/null"))))
    (setf *standard-output* nil
          *error-output* nil)
    (setf exit (shell command-string))
    (setf *standard-output* standard-output
          *error-output* error-output)
    exit))


(defvar *PLASM-TEMP-DIRECTORY* nil)

(defun SET-*PLASM-TEMP-DIRECTORY* (&optional dir) 
  (declare (special *PLASM-TEMP-DIRECTORY*))
  (setf *PLASM-TEMP-DIRECTORY*
        (if dir
          dir
          (make-project-directory "Please, type a directory to save Plasm temporary files [default ~A]:"
                                :optional-name "PLASM-TEMP")))

  (create-project-directory *PLASM-TEMP-DIRECTORY*)
  )




(defun HOW-MANY-IS-THERE-INSIDE (obj lisp-form)
  "it returns the element equal to the first argument if it exists in the lisp-form argument"
  (if (atom lisp-form)
      (if (equal obj lisp-form)
        1
        0)
   (apply (function +)
          (mapcar (lambda (x)
                    (how-many-is-there-inside obj x))
                  lisp-form))))


(defun IS-THERE-INSIDE (obj lisp-form)
  "it returns the element equal to the first argument if it exists in the lisp-form argument"
  (cond
   ((and (atom lisp-form)
         (equal obj lisp-form)))
   ((consp lisp-form)
    (let ((found (is-there-inside obj (first lisp-form))))
      (if found
        T
        (is-there-inside obj (rest lisp-form)))))))


(defun NO-OBJ-INSIDE (obj-list form)
  "It return nil if there isn't any of obj-list members inside form parameter"
  (declare (special form))
  (not (eval (cons 'or
              (mapcar (lambda (obj)
                         (is-there-inside obj form))
                      obj-list)))))
                  



(defun UKPOL-VERTICES (pol-complex)
  (first (unmake-pol-complex pol-complex)))



(defun MATERIAL-NODE-POL-COMPLEX (dag)
  (cond
   ((polyhedra-complex? dag)
    (cond
     ((not dag) nil) ; void-pol-complex
     ((MATERIAL-NODE-POL-COMPLEX (first dag)))
     ((MATERIAL-NODE-POL-COMPLEX (rest dag)))))
   
   ((and
     (polyhedral-node? dag)
     ;material in appearance? 
     (first (io::get-pol-node-appearance dag)))
    (return-material-node (first (io::get-pol-node-appearance dag))))
   ((and
     (polyhedral-node? dag)
     (not (polyhedron? (get-node-pol-complex dag))))
    (MATERIAL-NODE-POL-COMPLEX (get-node-pol-complex dag)))))


        
	


(defun RETURN-MATERIAL-NODE (material)
  (list
   (third material) ;ambient-color
   (first material)  ;diffuse-color
   (second material) ;specular-color
   (fourth material) ;emissive-color
   (fifth material) ;shininess
   (sixth material) ;transparency
   ))
   



(defun UKPOL-MATERIAL (pol-complex)
  "It returns the material property found with a deep search into pol-complex"
  ;the function must return the following data structure
  ;picking the first material node into the pol-complex argument
;  (list 
;   ;ambient-color
;   '(0.2 0.2 0.2)
;   ;diffuse-color
;   '(0.8 0.8 0.8)
;   ;specular-color
;   '(0 0 0)
;   ;emissive-color
;   '(0 0 0) 
;   ;shininess
;   0.2
;   ;transparency
;   0)
  (material-node-pol-complex pol-complex))


(defun MOVE-MESH (ROWS-COLS- POLS- TIMES-)
  (declare (special *PRINT-A-SEQUENCE-ON-A-LINE*))
  (let ((vertPerRow (first rows-cols-))
        (vertPerCol (second rows-cols-))
        (pol-vertices-list (mapcar (function ukpol-vertices) pols-))
        (pol-material-list (mapcar (function ukpol-material) pols-))
        (mesh-number (length pols-))
        (outstring  (make-string-output-stream)))
    (if (and (apply (function =) (mapcar (function length) pol-vertices-list))
             (= (* vertPerRow vertPerCol) (length (first pol-vertices-list))))
      (progn
        (quad-mesh-open-file outstring)
        (quad-mesh-states outstring)
        (quad-mesh-vertices outstring mesh-number vertPerRow vertPerCol)
        (dotimes (i mesh-number)
          (format outstring "# mesh number : ~A~%~%" (1+ i))
          (format outstring "# start time~%~% ~A~%~%" (nth i times-))
          (let ((pol-material (nth i pol-material-list))
                (pol-vertices (nth i pol-vertices-list)))
            (quad-mesh-material outstring 
                                :ambient-color (first pol-material)
                                :diffuse-color (second pol-material)
                                :specular-color (third pol-material)
                                :emissive-color (fourth pol-material)
                                :shininess (fifth pol-material)
                                :transparency (sixth pol-material)
                                )
            (format outstring "# mesh vertices~%~%")
            (dotimes (i (* vertPerRow vertPerCol))
              (format outstring
                      *PRINT-A-SEQUENCE-ON-A-LINE*
                      (nth i pol-vertices))))
          (format outstring "~%~%"))
        (format outstring "# end quad mesh file~%")
        (get-output-stream-string outstring))
      (plasm-error 7010 vertPerRow vertPerCol))))
    
              
(defun RANGE-0-1-P (item)
  (and (numberp item)
       (>= item 0)
       (<= item 1)))

(defun TRIPLE-0-1-P (lista)
  (and (consp lista)
       (= 3 (length lista))
       (every (function range-0-1-p) lista)))



(defstruct RGBCOLOR (r 0.0) (g 0.0) (b 0.0))

(defun MKCOLOR (pairs)
  (let ((color (make-rgbcolor))
        code)
    (dolist (pair pairs color)
      (setq code (string-upcase (first pair)))
      (cond
       ((and (string= "R" code)
             (>= (second pair) 0.0)
             (<= (second pair) 1.0))
        (setf (rgbcolor-r color) (second pair)))
       ((and (string= "G" code)
             (>= (second pair) 0.0)
             (<= (second pair) 1.0))
        (setf (rgbcolor-g color) (second pair)))
       ((and (string= "B" code)
             (>= (second pair) 0.0)
             (<= (second pair) 1.0))
        (setf (rgbcolor-b color) (second pair)))
       (T (plasm-error 8000 'mkcolor pair))))))



;called from MK_RGBCOLOR Plasm function and MKCOLOR
(defun MKCOLOR (triple &optional just-color?)
  (if (and (consp triple)
           (every (function range-0-1-p) triple))
    (if just-color?
      (let ((color (make-rgbcolor)))
        (setf (rgbcolor-r color) (first triple))
        (setf (rgbcolor-g color) (second triple))
        (setf (rgbcolor-b color) (third triple))
        color)
      (let ((color (make-rgbcolor))
            (material (make-material)))
        (setf (rgbcolor-r color) (first triple))
        (setf (rgbcolor-g color) (second triple))
        (setf (rgbcolor-b color) (third triple))
        (setf (material-ambient-color material) color)
        (setf (material-specular-color material) color)
        ))
    (plasm-error 8000 
                 (if just-color?
                   'mkcolor
                   'mk_rgbcolor)
                 triple)))

(defstruct MATERIAL
           (ambient-color (make-rgbcolor :r 0.2 :g 0.2 :b 0.2))
           (diffuse-color (make-rgbcolor :r 0.8 :g 0.8 :b 0.8))
           (specular-color (make-rgbcolor :r 0.0 :g 0.0 :b 0.0))
           (emissive-color (make-rgbcolor :r 0.0 :g 0.0 :b 0.0))
           (shiness 0.2)
           (transparency 0.0))

(defun MKMATERIAL (pairs)
  (let ((material (make-material))
        code)
    (dolist (pair pairs material)
      (setq code (string-upcase (first pair)))
      (cond
       ((and (string= "AMBIENT" code)
             (rgbcolor-p (second pair)))
        (setf (material-ambient-color material) (second pair)))
       ((and (string= "DIFFUSE" code)
             (rgbcolor-p (second pair)))
        (setf (material-diffuse-color material) (second pair)))
       ((and (string= "SPECULAR" code)
             (rgbcolor-p (second pair)))
        (setf (material-specular-color material) (second pair)))
       ((and (string= "EMISSIVE" code)
             (rgbcolor-p (second pair)))
        (setf (material-emissive-color material) (second pair)))
       ((and (string= "TRANSPARENCY" code)
             (>= (second pair) 0.0)
             (<= (second pair) 1.0))
        (setf (material-transparency material) (second pair)))
       ((and (string= "SHINESS" code)
             (>= (second pair) 0.0)
             (<= (second pair) 1.0))
        (setf (material-shiness material) (second pair)))
       (T (plasm-error 8000 'MKMATERIAL pair))))))




(defstruct TEXTURE
           (filename nil)
           (wraps 0) ; 0=CLAMP 1=REPEAT
           (wrapt 0) ; 0=CLAMP 1=REPEAT
           (translation '(0.0 0.0))
           (rotation 0.0)
           (scaleFactor '(0.0 0.0))
           (center '(0.0 0.0)))

(defun MKTEXTURE (pairs)
  (make-texture))

(defstruct APPEARANCE (material (make-material))
                      (texture (make-texture)))



(defun MKAPPEARANCE (seq)
  (let ((appearance (make-appearance))
        texture
        material)
    (if (listp seq)
      (progn
        (setq texture (cond 
                  ((texture-p (first seq)) 0)
                  ((texture-p (second seq)) 1)))
        (setq material (cond 
                   ((material-p (first seq)) 0)
                   ((material-p (second seq)) 1)))
        (if texture
          (setf (appearance-texture appearance) (nth texture seq)))
        (if material
          (setf (appearance-material appearance) (nth material seq)))
        appearance)
      (plasm-error 8000 'MKAPPEARANCE seq))))




(defun PRINT-TEXTURE-VRML1 (out texture-struct)
  (if (texture-filename texture-struct)
    (progn
      (format out "Texture2 {~%")
      (format out "  filename \"~A\" ~%" (texture-filename texture-struct))
      (format out"  wrapS ~A ~%" 
              (if (eql (texture-wraps texture-struct) 0) "CLAMP" "REPEAT"))
      (format out"  wrapT ~A ~%" 
              (if (eql (texture-wrapt texture-struct) 0) "CLAMP" "REPEAT"))
      (format out "  }~%")

      (format out "Texture2Transform  {~%")
      (format out "  translation ~{ ~F~}~%" (texture-translation texture-struct))
      (format out "  rotation  ~F ~%" (texture-rotation texture-struct))
      (format out "  scaleFactor ~{ ~F~}~%" (texture-scaleFactor texture-struct))
      (format out "  center ~{ ~F~}~%") (texture-center texture-struct))
      (format out "}~%")))



(defun MAKE-MESH (appearance grid)
  (if (and (appearance-p appearance)
           (test-grid-mesh grid))
    (list appearance grid)
    (plasm-error 'mkmesh appearance grid)))


(defun TEST-GRID-MESH (grid)
  (let (false)
    (and (consp grid)
         (dolist (i grid t)
           (if (consp i)
             (dolist (j i t)
               (if (and (consp j)
                        (every (function numberp) j))
                 t
                 (setq false t)))
             (setq false t)))
         (apply (function =) (mapcar (function length) grid))
         (not false))))





(defun GET-GRID-MESH (mesh)
  (second mesh))


(defun GET-APPEARANCE-MESH (mesh)
  (first mesh))




(defun MOVE-MESH (ROWS-COLS- POLS- TIMES-)
  (declare (special *PRINT-A-SEQUENCE-ON-A-LINE*))
  (let ((vertPerRow (first rows-cols-))
        (vertPerCol (second rows-cols-))
        (pol-vertices-list (mapcar (function ukpol-vertices) pols-))
        (pol-material-list (mapcar (function ukpol-material) pols-))
        (mesh-number (length pols-))
        (outstring  (make-string-output-stream)))
    (if (and (apply (function =) (mapcar (function length) pol-vertices-list))
             (= (* vertPerRow vertPerCol) (length (first pol-vertices-list))))
      (progn
        (quad-mesh-open-file outstring)
        (quad-mesh-states outstring)
        (quad-mesh-vertices outstring mesh-number vertPerRow vertPerCol)
        (dotimes (i mesh-number)
          (format outstring "# mesh number : ~A~%~%" (1+ i))
          (format outstring "# start time~%~% ~A~%~%" (nth i times-))
          (let ((pol-material (nth i pol-material-list))
                (pol-vertices (nth i pol-vertices-list)))
            (quad-mesh-material outstring 
                                :ambient-color (first pol-material)
                                :diffuse-color (second pol-material)
                                :specular-color (third pol-material)
                                :emissive-color (fourth pol-material)
                                :shininess (fifth pol-material)
                                :transparency (sixth pol-material)
                                )
            (format outstring "# mesh vertices~%~%")
            (dotimes (i (* vertPerRow vertPerCol))
              (format outstring
                      *PRINT-A-SEQUENCE-ON-A-LINE*
                      (nth i pol-vertices))))
          (format outstring "~%~%"))
        (format outstring "# end quad mesh file~%")
        (get-output-stream-string outstring))
      (plasm-error 7010 vertPerRow vertPerCol))))
    



(defun MOVE-MESH (ROWS-COLS- MESHES- TIMES-)
  (declare (special *PRINT-A-SEQUENCE-ON-A-LINE*))
  (let* ((vertPerRow (first rows-cols-))
         (vertPerCol (second rows-cols-))
         (numvertices (* vertPerRow vertPerCol))
         (mesh-number (length meshes-))
         (mesh-grid-list (mapcar (function get-grid-mesh) meshes-))
         (mesh-appearance-list (mapcar (function get-appearance-mesh) meshes-))
        
         (outstring  (make-string-output-stream)))
    (if (and (apply (function =) (mapcar (lambda (x) (length (apply (function append) x))) 
                                         mesh-grid-list))
             (= numvertices (length (apply (function append) (first mesh-grid-list)))))
      (progn
        (quad-mesh-open-file outstring)
        (quad-mesh-states outstring)
        (quad-mesh-vertices outstring mesh-number vertPerRow vertPerCol)
        (dotimes (i mesh-number)
          (format outstring "# mesh number : ~A~%~%" (1+ i))
          (format outstring "# start time~%~% ~A~%~%" (nth i times-))
          (let ((mesh-vertices  (apply (function append) (nth i mesh-grid-list))))
            (mesh-appearance outstring 
                             (nth i mesh-appearance-list)
                             )
            (format outstring "# mesh vertices~%~%")
            (dotimes (i numvertices)
              (format outstring
                      *PRINT-A-SEQUENCE-ON-A-LINE*
                      (nth i mesh-vertices))))
          (format outstring "~%~%"))
        (format outstring "# end quad mesh file~%")
        (get-output-stream-string outstring))
      (plasm-error 7010 vertPerRow vertPerCol))))
    
(defun QUAD-MESH-P (arg)
  (and (stringp arg)
       (>= (length arg) LEN-*QUAD-MESH-FILE-OVERTURE*)
       (string= *QUAD-MESH-FILE-OVERTURE*
                (subseq arg 0 (length *QUAD-MESH-FILE-OVERTURE*)))))


(defconstant *QUAD-MESH-FILE-OVERTURE* "# quad mesh file")

(defconstant LEN-*QUAD-MESH-FILE-OVERTURE* 
  (length *QUAD-MESH-FILE-OVERTURE*))


(defun QUAD-MESH-OPEN-FILE (out)
  (declare (special *QUAD-MESH-FILE-OVERTURE*))
  (format out
          "~A~%~%"
          *QUAD-MESH-FILE-OVERTURE*))

(defun QUAD-MESH-MATERIAL (out 
                           &key 
                           (ambient-color '(0.2 0.2 0.2))
                           (diffuse-color '(0.8 0.8 0.8))
                           (specular-color '(0 0 0))
                           (emissive-color '(0 0 0)) 
                           (shininess 0.2)
                           (transparency 0)
                           )
    (format out
            "# mesh material~%
 ~{~,2F ~}  # ambient color 
 ~{~,2F ~}  # diffuse color 
 ~{~,2F ~}  # specular color 
 ~{~,2F ~}  # emissive color 
 ~,2F             # shininess 
 ~,2F             # transparency ~%~%"
            
            ambient-color 
            diffuse-color 
            specular-color 
            emissive-color 
            shininess 
            transparency))

(defun MESH-APPEARANCE (out appearance)
  (let ((material (appearance-material appearance)))
    (format out
            "# mesh material~%
 ~{~,2F ~}  # ambient color 
 ~{~,2F ~}  # diffuse color 
 ~{~,2F ~}  # specular color 
 ~{~,2F ~}  # emissive color 
 ~,2F             # shininess 
 ~,2F             # transparency ~%~%"
            
            (rgbcolor-to-list (material-ambient-color material))
            (rgbcolor-to-list (material-diffuse-color  material))
            (rgbcolor-to-list (material-specular-color  material))
            (rgbcolor-to-list (material-emissive-color  material))
            (material-shiness  material)
            (material-transparency material))))


(defun RGBCOLOR-TO-LIST (rgbcolor)
  (list (rgbcolor-r rgbcolor)
        (rgbcolor-g rgbcolor)
        (rgbcolor-b rgbcolor)))

(defun QUAD-MESH-STATES (out &key (startstate 0) (stopstate 0))
  (format out
          "# states~%
 ~A ~A   # initial and final state (0=off, -1=on)~%~%"

          startstate stopstate))


(defun QUAD-MESH-VERTICES (out mesh-number vertPerRow vertPerCol)
  (format out 
"# mesh data~%
 ~A      # number of mesh~%
 ~A  ~A  # unsigned integer~%~%"
 mesh-number vertPerRow  vertPerCol  ))








(setf *MOTION-FILE-HEAD* "#!plasm v0.1 ascii~%~%")


(setf *MOTION-FILE-IV-FILE-NAME* "######~%# Inventor File Name~%~%~T ~A~%~%")

(setf *MOTION-FILE-REMARK-TRANSFORM*
"######
# Transformation
#
#   Transformation Type
#   x, y, z -- rotation about the x, y, z axis
#         r -- rotation about an arbitrary axis 
#         t -- translation         
#         s -- scaling
#         m -- matrix ---- FOR THE MOMENT NOT IMPLEMENTED~%~%")

(setf *MOTION-FILE-PARAMETERS-NUMBER* "# Param Number~%~%~T~A~%~%")


(setf *PRINT-A-SEQUENCE-ON-A-LINE* "~{ ~,4F ~}~%")


(setf *MOTION-FILE-PARAMETERS-NAME-TYPE-HEADER* 
      "# Name (used in Open Inventor text format file)  Type~%~%")


(defun WRITE-LABELS (labels outfile)
  (declare (special *PRINT-A-SEQUENCE-ON-A-LINE*
                    *MOTION-FILE-PARAMETERS-NAME-TYPE-HEADER*))
  (format outfile
            *MOTION-FILE-PARAMETERS-NAME-TYPE-HEADER*)
  (dotimes (i (length labels))
    (format outfile
            *PRINT-A-SEQUENCE-ON-A-LINE*
            (list (make-label (third (nth i labels)))
                  (first (nth i labels)))))
      (format outfile "~%~%"))

 


(setf *MOTION-FILE-MOTION-REMARK*

"######
# Motion

#############################################################
# for rotation about the x, y, z axis one param is needed: 
#   the alpha param
#
# for rotation about an arbitrary axis four param is needed: 
#   the alpha param and the x y and z axis components
#
# for translation and scaling three param is needed:
#   the x y and z components
#
# MATRIX NOT IMPLEMENTED
#
#############################################################
#
# Example
#
#  time   rotz   rotz   rotz   rot(1,1,1)    translation
#  1.0    0.0    0.0    0.0    0.0  1 1 1    0    0    0
#  3.0    45.0   0.0    0.0    10.0 1 1 1    1    1    1 
#  5.0    90.0   45.0   0.0    20.0 1 1 1    2    2    2 
#  7.0    90.0   90.0   0.0    30.0 1 1 1    3    3    3 
#  9.0    90.0   90.0   90.0   40.0 1 1 1    4    4    4 
#  12.0   90.0   -45.0  90.0   45.0 1 1 1    5    5    5 
#############################################################~%~%")


(setf *MOTION-FILE-STATES-NUMBER* "# Number of state~%~%~T~A~%~%")


(setf *MOTION-FILE-MOTION-VALUES-HEADER* "# time param0 param1 ......~%~%")


(defun WRITE-MOTION-VALUES (motion-values time-data outfile)
  (declare (special *PRINT-A-SEQUENCE-ON-A-LINE*
                    *MOTION-FILE-MOTION-VALUES-HEADER*))
  (format outfile
            *MOTION-FILE-MOTION-VALUES-HEADER*)
  (dotimes (i (length motion-values))
    (format outfile " ~,4F " (nth i time-data))
    (mapcar (lambda (x)
              (format outfile
                      " ~{ ~,4F ~} "
                      x))
            (insert-zero  (nth i motion-values)))
    (format outfile " ~%"))
  (format outfile "~%~%"))


(defun INSERT-ZERO (single-motion-data)
  (declare (special motion-labels))
  (let ((rebuild-single-motion-data '()))
    (dotimes (i (length motion-labels) (reverse rebuild-single-motion-data))
      (setf rebuild-single-motion-data 
            (cons (build-for-all
                   (first (nth i motion-labels))
                   (second (nth i motion-labels))
                   (remove nil (mapcar (lambda (true-false value)
                                         (if (= 1 true-false)
                                           value
                                           nil))
                                       (third (nth i motion-labels))
                                       single-motion-data)))
                  rebuild-single-motion-data)))))


(defun BUILD-FOR-ALL (type axe provide-values)
  (cond
   ((or (char= type #\x)
          (char= type #\y)
          (char= type #\z))
    provide-values)
    ((or (char= type #\s)
          (char= type #\t))
     (let* ((axes (if (consp axe)
                   axe
                   (list axe)))
            (axe-value-pair (mapcar (function list)
                                          axes 
                                          provide-values)))
       (dotimes (i 3 (mapcar (function second)
                             (sort axe-value-pair '< :key 'first)))
         (if (not (assoc (1+ i) axe-value-pair))
           (setf axe-value-pair
                 (cons (list (1+ i) 0)
                       axe-value-pair))))))))
         
                   
                                 
                                        


(setf *MOTION-FILE-END* "#~%# end of file~%")




(defun WRITE-MOTION-FILE (outfile geometry-filename labels move-data time-data)
  (declare (special 
            *MOTION-FILE-HEAD*
            *MOTION-FILE-IV-FILE-NAME*
            *MOTION-FILE-REMARK-TRANSFORM*
            *MOTION-FILE-PARAMETERS-NUMBER*
            *MOTION-FILE-PARAMETERS-NAME-TYPE-HEADER*
            *MOTION-FILE-MOTION-REMARK*
            *MOTION-FILE-STATES-NUMBER*
            *MOTION-FILE-MOTION-VALUES-HEADER*
            *MOTION-FILE-END*))
  (format outfile *MOTION-FILE-HEAD*)
  (format outfile *MOTION-FILE-IV-FILE-NAME* (concatenate 'string "actors/" geometry-filename))
  (format outfile *MOTION-FILE-REMARK-TRANSFORM*)
  (format outfile *MOTION-FILE-PARAMETERS-NUMBER* 
          (length labels))
  (write-labels labels outfile)
  (format outfile *MOTION-FILE-MOTION-REMARK*)
  ;(format outfile *MOTION-FILE-STATES-NUMBER* (length move-data))
  ;(format outfile *MOTION-FILE-MOTION-VALUES-HEADER*)
  (write-motion-values ;(shift-right-all
                           move-data ;) 
                           time-data
                           outfile)
  (format outfile *MOTION-FILE-END*))





(defun SHIFT-RIGHT-ALL (move-data)
  (dotimes (i (length move-data) move-data)
    (setf (nth i move-data)
          (cons (first (last (nth i move-data))) (butlast (nth i move-data))))))



(defun RESOLVE-ANIMATION (move-function move-data time-data)
  (let* (motion-filename 
         geometry-filename 
         plasm-function-struct 
         plasm-function-variables ; it is a list of list of names
         plasm-function-variables-list ; it is a plain list of variable names
         where-bindind-form  ; where-name LET binding part  
         move-function-body 
         middle-body 
         motion-labels 
         )
    (declare (special 
                      plasm-function-struct
                      plasm-function-variables
                      plasm-function-variables-list
                      where-bindind-form  
                      move-function-body 
                      middle-body 
                      motion-labels
                      ))

    ;correctness testing follow
    (if (or  (not (symbolp move-function))
             (not (bound-function-in-user-fenv  move-function :call-error nil)))
      (plasm-error 7004 move-function)
      )
    
    (setf plasm-function-struct (bound-function move-function))
    (setf plasm-function-variables (plasm-function-used-formal-pars 
                                    plasm-function-struct))
    (setf plasm-function-variables-list (first plasm-function-variables))

    (if (< 1 (length (plasm-function-formal-pars plasm-function-struct)))
      (plasm-error 7001  (plasm-function-name plasm-function-struct)))

    (if (not (= (length plasm-function-variables-list)
                (length (first move-data))))
      (plasm-error 7003 
                   (plasm-function-name plasm-function-struct)
                   (length plasm-function-variables-list)
                   (length (first move-data))))

    ; formal control finished

    (IO::SET-RESET-ENVIRONMENT)
    (setf move-function-body (plasm-function-body plasm-function-struct))
    (setf middle-body (take-middle-body move-function-body))

    ;now where-bindind-form contains where-name and formal parameters
    (setf where-bindind-form 
          (remove-formal-parameters where-bindind-form
                                    plasm-function-variables-list))
    ;now where-bindind-form contains only where-name 


    
    ; set names for exporting files
    (setf geometry-filename (prompt-for-a-new-file "test" ".iv"))
    (with-open-file (geometry-outfile geometry-filename 
                                    :direction :output
                                    :if-exists :overwrite
                                    :if-does-not-exist :create)
      (write-iv-geometry-file geometry-outfile
                           (file-namestring geometry-filename)
                           ))

    ;motion-labels are now built reading the plasm function body
    ; we've to check if they are semantically correct

    (build-motion-format-labels)

    (setf motion-filename (prompt-for-a-new-file "test" ".motion"))
    (with-open-file (motion-outfile motion-filename 
                                    :direction :output
                                    :if-exists :overwrite
                                    :if-does-not-exist :create)
      (write-motion-file motion-outfile 
                         (file-namestring geometry-filename)
                         motion-labels
                         move-data
                         time-data
                         ))

    (format nil 
            "~%Geometry inside ~A function can be animated.~%"
            (plasm-function-name plasm-function-struct))
    (values (file-namestring motion-filename)
            move-function)
    ))


(defun RESOLVE-ANIMATION (move-function move-data time-data)
  (let* (
         (motion-outfile  (make-string-output-stream))
         (geometry-outfile (make-string-output-stream))
         iv-string
         motion-string
         motion-filename 
         geometry-filename 
         plasm-function-struct 
         plasm-function-variables ; it is a list of list of names
         plasm-function-variables-list ; it is a plain list of variable names
         where-bindind-form  ; where-name LET binding part  
         move-function-body 
         middle-body 
         motion-labels 
         )
    (declare (special 
                      plasm-function-struct
                      plasm-function-variables
                      plasm-function-variables-list
                      where-bindind-form  
                      move-function-body 
                      middle-body 
                      motion-labels
                      ))

    ;correctness testing follow
    (if (or  (not (symbolp move-function))
             (not (bound-function-in-user-fenv  move-function :call-error nil)))
      (plasm-error 7004 move-function)
      )
    
    (setf plasm-function-struct (bound-function move-function))
    (setf plasm-function-variables (plasm-function-used-formal-pars 
                                    plasm-function-struct))
    (setf plasm-function-variables-list (first plasm-function-variables))

    (if (< 1 (length (plasm-function-formal-pars plasm-function-struct)))
      (plasm-error 7001  (plasm-function-name plasm-function-struct)))

    (if (not (= (length plasm-function-variables-list)
                (length (first move-data))))
      (plasm-error 7003 
                   (plasm-function-name plasm-function-struct)
                   (length plasm-function-variables-list)
                   (length (first move-data))))

    ; formal control finished


    (if (not (plasm-function-motion-labels plasm-function-struct))
      (progn
        (IO::SET-RESET-ENVIRONMENT)
        (setf move-function-body (plasm-function-body plasm-function-struct))
        (setf middle-body (take-middle-body move-function-body))
        
        ;now where-bindind-form contains where-name and formal parameters
        (setf where-bindind-form 
              (remove-formal-parameters where-bindind-form
                                        plasm-function-variables-list))
        ;now where-bindind-form contains only where-name 

        ; set names for exporting files
        (setf geometry-filename (concatenate 'string 
                                             (string-downcase (string move-function))
                                             ".iv"))
        (write-iv-geometry-file geometry-outfile
                                geometry-filename
                                )
        ;motion-labels are now built reading the plasm function body
        ; we've to check if they are semantically correct

        (build-motion-format-labels)
        (setf  (plasm-function-motion-labels plasm-function-struct) motion-labels))

     (progn
       (setf geometry-filename (concatenate 'string 
                                            (string-downcase (string move-function))
                                            ".iv"))
       (setf motion-labels (plasm-function-motion-labels plasm-function-struct))))

    (write-motion-file motion-outfile 
                       geometry-filename
                       motion-labels
                       move-data
                       time-data
                       )

    (setf iv-string (get-output-stream-string geometry-outfile))
    (setf motion-string (get-output-stream-string motion-outfile))
    
    (set-file-slot-and-save-in-file plasm-function-struct iv-string)

    (list (plasm-function-file  plasm-function-struct)  
          motion-string
         geometry-filename)
    ))


(defun SET-FILE-SLOT-AND-SAVE-IN-FILE (plasm-function-struct iv-string)
  (declare (special *PLASM-TEMP-DIRECTORY*))
  (if (not *PLASM-TEMP-DIRECTORY*)
    (set-*PLASM-TEMP-DIRECTORY*))
  (setf (plasm-function-file  plasm-function-struct)
        (make-pathname :directory (pathname-directory *PLASM-TEMP-DIRECTORY*)
                       :name (string-downcase (plasm-function-name plasm-function-struct))
                       :type "iv"))
  (with-open-file (out-geometry (plasm-function-file  plasm-function-struct)
                                :direction :output
                                :if-exists :overwrite
                                :if-does-not-exist :create)
    (format out-geometry "~A" iv-string)))


(defun MAKE-PROJECT-DIRECTORY (
              &optional
              (format-string
               #+:APPLE "Type a directory to contain animation project files."
               #+:CLISP "~%Type a directory to contain animation project files [default ~A]:"
               )
              &key (optional-name "untitled")
              )
"It returns a directory string"
  (let ((user-dir "")
        (default-dir  (concatenate 'string 
                               (directory-namestring 
                                #+:CLISP (user-homedir-pathname)
                                #+:APPLE (choose-file-default-directory)
                                )
                               optional-name
                               #+:CLISP "/"
                               #+:APPLE ":"
                               )))
    #+:APPLE 
    (setf user-dir
          (get-string-from-user  format-string  :initial-string default-dir))
    #+:CLISP
    (progn
      (format t format-string default-dir)
      (setf user-dir (read-line)))
    (if (string= "" user-dir)
      default-dir
      (correct-pathname-string user-dir))))


(defun CORRECT-PATHNAME-STRING (path-string)
  (let ((loc-path-string path-string))
    (if (char= #\"
               (char loc-path-string 0))
      (setf loc-path-string
            (subseq loc-path-string 1 (1- (length loc-path-string)))))
    (if (char= 
         #+:CLISP #\/
         #+:APPLE #\: 
         (char loc-path-string (1- (length loc-path-string))))
      loc-path-string 
      (concatenate 'string
                   loc-path-string  
                   #+:CLISP "/" 
                   #+:APPLE ":"
                   ))))

(defun MAKE-DIRECTORY-NAMESTRING (arg)
"It returns a directory namestring from or a pathname
or an ABSOLUTE directory list either a list of string"
  (cond
    ((pathnamep arg)
     (directory-namestring arg))
    ((equal (first arg) :ABSOLUTE)
     (directory-namestring (make-pathname
                            :directory arg)))
    ((and (consp arg)
          (every (function stringp) arg))
     (directory-namestring (make-pathname
                            :directory `(:ABSOLUTE ,@arg))))))


(defun CREATE-PROJECT-DIRECTORY (dir-namestring)
"It creates a directory which name string is its argument
(if that directory does exist it will not take any effect.)."
  #+:APPLE
  (create-directory dir-namestring
                  :if-exists nil)
  #+:CLISP 
;  (if (not (eq 0 (execute-clisp-shell "touch " dir-namestring)))
;    (execute-clisp-shell "mkdir " dir-namestring))
   (if (not (directory (pathname dir-namestring)))
     (execute-clisp-shell "mkdir " dir-namestring))

  dir-namestring)



(defun WRITE-STRING-IN-FILE (content-string &key 
                                          (directory  nil)
                                          (type "iv")
                                          (filename (string (gensym "iv-file")))
                                          (file-path nil))
  (declare (special *PLASM-TEMP-DIRECTORY*))
  (if (not *PLASM-TEMP-DIRECTORY*)
    (set-*PLASM-TEMP-DIRECTORY*))
  (if (not directory)
        (setf directory *PLASM-TEMP-DIRECTORY*))

;if file-path is not nil content-string will be the content of file-path
  (let ((path (cond
                (file-path)
                (T (make-pathname :directory (pathname-directory directory)
                             :name filename
                             :type type)))))
    (with-open-file (outfile path 
                             :direction :output
                             :if-exists :overwrite
                             :if-does-not-exist :create)
      (format outfile "~A" content-string))
    path))


(defun FRAMES (pol-complex t1-t2-pair &optional (name (string (gensym "frames"))))
  (list (write-string-in-file (export-iv-stream pol-complex))
        (first t1-t2-pair)
        (second t1-t2-pair)))

(defun CAMERA (states times)
  (declare (special PLASM::*CAMERA-TAG*))
  (list
   PLASM::*CAMERA-TAG*
   states
   times))



(defun ANIMATION-TO (file objs-list)
  (animation objs-list file))

(defun SET-TEMP-PLASM-DIR (dir)
 (SET-*PLASM-TEMP-DIRECTORY* dir))


(defun ANIMATION (objs-list &optional dir-from-animation-to)
  (let* ((project-directory (if dir-from-animation-to
                              (if (string=
                                   #+:APPLE ":"
                                   #+:CLISP "/" 
                                   (subseq dir-from-animation-to
                                           (1- (length dir-from-animation-to))))
                                dir-from-animation-to
                                (concatenate 'string
                                             dir-from-animation-to
                                             #+:APPLE ":"
                                             #+:CLISP "/" ))
                              (make-project-directory)))
         (abs-prj-directory (pathname-directory project-directory))
         (animation-filename  (first (last abs-prj-directory)))
         background-section
         background-directory
         mesh-section
         mesh-directory
         cameras-section
         cameras-directory
         cs-section
         cs-directory
         actors-section
         actors-directory
         frames-section
         frames-directory
         )


    (if (funcall 
         #+:APPLE #'directoryp
         #+:CLISP #'directory
         (make-pathname :directory (butlast abs-prj-directory)))
      (create-project-directory project-directory)
      (plasm-error 7006
                   (namestring (make-pathname :directory (butlast abs-prj-directory)))
                   (namestring (make-pathname :directory  abs-prj-directory))))
      (setf cs-directory  
            (make-directory-namestring 
             (append abs-prj-directory
                     '("cs"))))
      (setf  background-directory
            (make-directory-namestring 
             (append abs-prj-directory
                     '("background"))))
      (setf  mesh-directory
            (make-directory-namestring 
             (append abs-prj-directory
                     '("mesh"))))
      (setf frames-directory
            (make-directory-namestring 
             (append abs-prj-directory
                     '("frames"))))
       (setf actors-directory
             (make-directory-namestring 
              (append abs-prj-directory
                     '("actors"))))
       (setf cameras-directory
            (make-directory-namestring 
             (append abs-prj-directory
                     '("cameras"))))

       (mapcar 'create-project-directory
               (list
                background-directory
                mesh-directory
                cameras-directory
                cs-directory
                actors-directory
                frames-directory
                ))
  
      (dolist (obj objs-list)
        (cond
          ((not-void-polyhedra-complex? obj)  ;go in background section
           (push (write-string-in-file (export-iv-stream obj)
                                       :directory background-directory)
                 background-section))
                 
          ((quad-mesh-p obj)  ;go in mesh section
           (push (write-string-in-file obj
                                       :directory mesh-directory
                                       :type "qm"
                                       :filename (string (gensym "mesh-file")))
                 mesh-section))
                 
          ((plasm::geometry-motion-pair obj) ;go in cs and actors section
           (let ((file (make-pathname :directory (pathname-directory actors-directory)
                                      :name (pathname-name (third obj))
                                      :type (pathname-type (third obj)))))
             (if (not (probe-file file))
               (progn
                 (push file actors-section)
                 #+:APPLE (namestring (copy-file (first obj) file :if-exists :supersede))
                 #+:CLISP (progn
                            (execute-clisp-shell
                             "cp "
                             (namestring (first obj))
                             " "
                             (namestring file))
                            (namestring file))))
             (push (write-string-in-file (second obj)
                                         :directory cs-directory
                                         :type "cs")
                   cs-section)))
          ((plasm::frames-p obj) ;go in frame section
           (let ((file (namestring (make-pathname :directory (pathname-directory frames-directory)
                                      :name (pathname-name (first obj))
                                      :type "iv"))))
             (push (list
                    #+:APPLE (namestring (copy-file (first obj) file :if-exists :supersede))
                    #+:CLISP (progn
                               (execute-clisp-shell
                                "cp "
                                (namestring (first obj))
                                " "
                                file)
                               file)
                    (second obj)
                    (third obj))
                   frames-section)))
          ((plasm::camera-p obj)
           (let ((file (make-pathname :directory (pathname-directory cameras-directory)
                                      :name (pathname-name (string (gensym "camera-")))
                                      :type "cam")))
             (write-string-in-file (let ((stream-string (make-string-output-stream)))
                                     (build-camera-file-content (second obj) (third obj) stream-string)
                                     (get-output-stream-string stream-string))
                                   :file-path file)
             (push (namestring file)
                   cameras-section)))
          ))

      (with-open-file (project-outfile (make-pathname :directory abs-prj-directory 
                                                      :name animation-filename
                                                      :type "prj") 
                                      :direction :output
                                      :if-exists :overwrite
                                      :if-does-not-exist :create)
        (format project-outfile "[options]~%")
        (format project-outfile " WindowTitle=~A~%" animation-filename)
        (format project-outfile " BackgroundColor=0.3 0.7 0.9~%" )
        (format project-outfile " FramesPerSecond=18~%")
        (format project-outfile " TimeUnit=1.0~%" )

        ;(format project-outfile "[actors]~%")
        ;(dolist (item actors-section)
        ;  (format project-outfile " actors/~A.iv ~%" (pathname-name item)))
                  
 
        (format project-outfile "[cs]~%")
        (dolist (item cs-section)
          (format project-outfile " cs/~A.cs ~%" (pathname-name item)))

        (format project-outfile "[background]~%")
        (dolist (item background-section)
          (format project-outfile " background/~A.iv ~%" (pathname-name item)))

        (format project-outfile "[mesh]~%")
        (dolist (item mesh-section)
          (format project-outfile " mesh/~A.qm ~%" (pathname-name item)))

        (format project-outfile "[frames]~%")
        (dolist (item frames-section)
          (format project-outfile " frames/~A.iv ~A ~A ~%" 
                  (pathname-name (first item))
                  (second item)
                  (third item)))

        (format project-outfile "[cameras]~%")
        (dolist (item cameras-section)
          (format project-outfile " cameras/~A.cam ~%" (pathname-name item)))

        (format t "~%~%An animation project file set is in ~A directory.~%~%"
                (make-directory-namestring  abs-prj-directory))

        (values)

    
        )))

  


(defun BUILD-CAMERA-FILE-CONTENT (states times stream)
  (format stream "[default]~%")
  (format stream "#aspectRatio=SO_ASPECT_SQUARE~%")
  (format stream "aspectRatio=SO_ASPECT_VIDEO~%")
  (format stream "#aspectRatio=SO_ASPECT_HDTV~%")
  (format stream "heightAngle= ~,6F~%" (/ pi 4))
  (format stream "~%[data]~%")
  (dotimes (i (length states))
    (format stream
            "~,6F  ~{~,6F ~} ~%"
            (nth i times)
            (nth i states))))

  

(defun WRITE-IV-GEOMETRY-FILE (geometry-outfile geometry-filename)
"This function generates an Open Inventor text format file. This function uses extensively
the environment of its calling function RESOLVE-ANIMATION."
  (declare (special where-bindind-form 
                    plasm-function-struct
                    plasm-function-variables 
                    where-bindind-form  
                    move-function-body 
                    middle-body 
                    motion-labels
                    
                    ))
  ;(format T "~%~A~%" middle-body)
  (let ((*PARTIAL-TRANSFORM-FUNCTION-APPLICATIONS* 0))
    (declare (special *PARTIAL-TRANSFORM-FUNCTION-APPLICATIONS*))
    (WRITE-OVERTURE-IV-FILE geometry-outfile)
    (write-its-iv-node geometry-outfile middle-body)
    (WRITE-CLOSURE-IV-FILE geometry-outfile)
    ))






(defun PLASM-TRAP ()
  (plasm-error 7000))


(defun WRITE-ITS-IV-NODE (geometry-outfile form)
 ; (format t "~% WRITE-ITS-IV-NODE calling with ~A ~%" form)
  (cond
   ((xstruct-reduce-function-p form)
    (write-separator-overture geometry-outfile)
    (process-xstruct-argument geometry-outfile (third form)) ; takes out reduce-function
                                                                   ; and (quote xstruct) elem.
    (write-separator-close geometry-outfile)
    )

   ((property-expression-p form)
    (write-property-iv-node geometry-outfile (evaluate-this-form form)))

   ((independent-transformation-expression-p form)
    (write-independent-transformation-iv-node geometry-outfile form))

   ((transformation-expression-p form)
    (write-transformation-iv-node geometry-outfile form))

   ((take-me-form-for-where-end-symbol-p form)
    (parse-take-me-form-for-where-end-symbol geometry-outfile form))


   ((eval-general-expression geometry-outfile form))

   (T (plasm-trap))
   )
 ; (format t "~% WRITE-ITS-IV-NODE called with ~A ~%" form)
  )




(defun TAKE-ME-FORM-FOR-WHERE-END-SYMBOL-P (obj)
"This function check if obj is the lisp expression used to take the value of a where-end symbol
 defined in a plasm function"
   (and (ispair obj)
        (eq 'TAKE-ME (first obj))))


(defun TRANSFORMATION-CLOSURE-P (obj)
  "At the moment this function considers any closure a transformation function.
   We need just to apply this closure to a void polyhedron and then export it"
  (functionp obj))




(defun WRITE-TRANSFORMATION-FROM-PARTIAL-FUNCTION (geometry-outfile partial-function)
  (declare (special *PARTIAL-TRANSFORM-FUNCTION-APPLICATIONS*))
  (incf *PARTIAL-TRANSFORM-FUNCTION-APPLICATIONS*)
  (IO::WRITE-TRANSFORMATION (reduce-function partial-function 
                                        (build-pol-complex '((0 0 0)) '((0)) '((0))))
                            geometry-outfile))



(defun EVAL-GENERAL-EXPRESSION (geometry-outfile form)
  (declare (special plasm-function-variables-list))
  (if (no-obj-inside plasm-function-variables-list form)
    (let ((value-form (evaluate-this-form form)))
      (cond
       ((not-void-polyhedra-complex? value-form)
        (format geometry-outfile "~A" (generate-pol-complex-string-iv-format value-form))
        T)
       ((transformation-closure-p value-form)
        (write-transformation-from-partial-function geometry-outfile value-form)
        T)
       (nil)
       ))
    nil))


(defun RETURN-LOCATED-POLYHEDRON-VALUE (pol)
  (declare (special *PARTIAL-TRASNFORM-FUNCTIONS*))
  (if *PARTIAL-TRASNFORM-FUNCTIONS*
    (reduce-function 
     (reduce-function 'comp (reverse *PARTIAL-TRASNFORM-FUNCTIONS*))
     pol)
    pol))




(defun PARSE-take-me-form-for-where-end-symbol (geometry-outfile form)
;form is something like (TAKE-ME <A SYMBOL>) !
;(assoc (second form) where-bindind-form) returns something like (QUOTE <A LISP ESPRESSION>) !
    (declare (special where-bindind-form
            ))
    (WRITE-ITS-IV-NODE geometry-outfile (second (second (assoc (second form) where-bindind-form)))))




(defun EVALUATE-THIS-FORM (form)
  (declare (special where-bindind-form
            ))
  (eval `(LET ,where-bindind-form
           (declare (special ,@(mapcar (function first) where-bindind-form)))
           ,form)))


(defun XSTRUCT-REDUCE-FUNCTION-P (form)
  (and (function-calling-form-p form)
       (equal 'xstruct (second (second form)))))


(defun PROCESS-XSTRUCT-ARGUMENT (geometry-outfile form)
  (cond
   ((and (consp form)
         (equal 'list (first form)))
    (mapcar (lambda (x) 
              (write-its-iv-node geometry-outfile x))
            (rest form)))
    (T (plasm-trap))
  ))


(defun REMOVE-FORMAL-PARAMETERS (where-bindind-form
                                plasm-function-variables)
  (if (member (first (first where-bindind-form)) plasm-function-variables)
    (remove-formal-parameters (rest where-bindind-form)
                                    plasm-function-variables)
    where-bindind-form))

(defun TAKE-MIDDLE-BODY (body-form)
  (first 
   (remove-type-checking 
   (take-let-form 
   (rest body-form) ; body form always starts  with PROGN
                 ))))

(defun REMOVE-TYPE-CHECKING (body-form)
  (if (equal (first (first body-form)) 'TYPE-TEST-FUNCTION)
    (remove-type-checking (rest body-form))
    body-form))

(defun TAKE-LET-FORM (body-form)
  (declare (special where-bindind-form))
  (if (equal 'let (first (first body-form)))
    (progn 
      (setf where-bindind-form (second (first body-form)))
      (rest ;remove DECLARE SPECIAL
       (rest ; remove variables bindings
        (rest ; remove LET
         (first body-form) ;LET form
         ))))
    (take-let-form (rest body-form))))

  

(defvar *PROPERTY-NODES* 
        '(
          COLOR
          MATERIAL
          LIGHT
          CAMERA

          ; and whatever you need
          ))

(defvar *TRANSFORMATION-NODES* 
        '(
          TRANSLATE
          R
          S
          H
          MAT

          ; and whatever you need
          ))


(defun FUNCTION-CALLING-FORM-P (form)
  (and (consp form)
       (equal (first form) 'reduce-function)))


(defun PROPERTY-EXPRESSION-P (form)
  (declare (special *PROPERTY-NODES*
                    plasm-function-variables-list
                    ))
  (cond
   ((and (consp form)
         (function-calling-form-p form)
         (no-obj-inside plasm-function-variables-list form)
         (member (second (second form)) ; the function name is quoted.
                 *PROPERTY-NODES*)
         )
    form)))

(defun TRANSFORMATION-EXPRESSION-P (form)
  (declare (special *TRANSFORMATION-NODES*
                    plasm-function-variables-list
                    ))
  (cond
   ((and (consp form)
         (function-calling-form-p form)
         (function-calling-form-p (second form))

         ;the following is a limit to remove
         ;(no-obj-inside plasm-function-variables-list form)

         (member (second (second (second form))) *TRANSFORMATION-NODES*))
    form)
   ))

(defun INDEPENDENT-TRANSFORMATION-EXPRESSION-P (form)
  (declare (special *TRANSFORMATION-NODES*
                    plasm-function-variables-list
                    ))
  (cond
   ((and (consp form)
         (function-calling-form-p form)
         (function-calling-form-p (second form))
         (no-obj-inside plasm-function-variables-list form)
         (member (second (second (second form))) *TRANSFORMATION-NODES*))
    form)
   ))


(defun WRITE-TRANSFORMATION-IV-NODE (geometry-outfile form)
  (declare (special plasm-function-variables motion-labels))
  (let ( 
  ;------------------------------------------------------------------------ 
  ; now it relies on a simple affine trasformation expressions
  ;------------------------------------------------------------------------
        (transform-type (second (second (second form))))
        (transform-axes (evaluate-this-form (third (second form))))
        (variables-occurence (collect-variables-occurence (third form)))
        )
    (cond
     ((rotation-transform-p transform-type)
      (write-rotation-label geometry-outfile
                            (make-label variables-occurence)
                            (make-rotation-label (build-rotation-axe transform-axes)))
      (setf motion-labels (cons (list (make-rotation-label (build-rotation-axe transform-axes))
                                      transform-axes
                                      variables-occurence)
                                
                                motion-labels)))
     ((translate-transform-p transform-type)
      (write-translate-label geometry-outfile
                             (make-label variables-occurence))
      (setf motion-labels (cons (list #\t 
                                      transform-axes
                                      variables-occurence)
                                motion-labels)))
     ((scale-transform-p transform-type)
      (write-scale-label geometry-outfile
                            (make-label variables-occurence))
      (setf motion-labels (cons (list #\s
                                      transform-axes
                                      variables-occurence)
                              motion-labels)))
      )))


(defun MAKE-ROTATION-LABEL (axe)
  (cond
   ((= 1 (nth 0 axe)) #\x)
   ((= 1 (nth 1 axe)) #\y)
   ((= 1 (nth 2 axe)) #\z)))


(defun WRITE-TRANSLATE-LABEL (geometry-outfile label)
  (format geometry-outfile
          "~%DEF ~A Translation { }~%"
          label))

(defun WRITE-SCALE-LABEL (geometry-outfile label)
  (format geometry-outfile
          "~%DEF ~A Scale { }~%"
          label))

(defun WRITE-ROTATION-LABEL (geometry-outfile label axe)
  (format geometry-outfile
          "~%DEF ~A RotationXYZ { axis ~A }~%"
          label (char-upcase axe)))

    

(defun WRITE-INDEPENDENT-TRANSFORMATION-IV-NODE (geometry-outfile form)
  (declare (special plasm-function-variables))
  (let (
        (transform-type (second (second (second form))))
        (transform-axes (evaluate-this-form (third (second form))))
        (transform-value (evaluate-this-form (third form)))
        )
    ;(format t "~%transform-type = ~A~%" transform-type)
    ;(format t "~%transform-axes = ~A~%" transform-axes)
    ;(format t "~%transform-value = ~A~%" transform-value)
    (cond
     ((rotation-transform-p transform-type)
      (write-sotransform-node geometry-outfile
       :rotation-axe
       (build-rotation-axe transform-axes)
       :rotation-value transform-value))
     ((translate-transform-p transform-type)
      (write-sotransform-node geometry-outfile
       :translation
       (build-translate-scale-values transform-axes transform-value 0)))
     ((scale-transform-p transform-type)
      (write-sotransform-node geometry-outfile
       :scale-factor
       (build-translate-scale-values transform-axes transform-value 1)))
     )))

(defun BUILD-ROTATION-AXE (rotation-plasm-parameters)
  (if (and (= 2 (length rotation-plasm-parameters))
           (every 'numberp rotation-plasm-parameters)
           (not (= (nth 0 rotation-plasm-parameters)
                   (nth 1 rotation-plasm-parameters))))
    (mapcar (lambda (x) (if x 0 1))
                     (mapcar (lambda (item) (member  item rotation-plasm-parameters :test #'eq))
                             '(1 2 3)))
    (plasm-error 7004 'R rotation-plasm-parameters)))
  


(defun BUILD-TRANSLATE-SCALE-VALUES (axes values default)
  (let* ((axes-l (if (consp axes) axes (list axes)))
         (values-l (if (consp values) values (list values)))
         (x-axe (position 1 axes-l))
         (y-axe (position 2 axes-l))
         (z-axe (position 3 axes-l)))
    (list
     (if x-axe (nth x-axe values-l) default)
     (if y-axe (nth y-axe values-l) default)
     (if z-axe (nth z-axe values-l) default))))



(defun WRITE-SOTRANSFORM-NODE (geometry-outfile
                               &key 
                               (translation '(0 0 0))
                               (rotation-axe '(0 0 1))
                               (rotation-value 0)
                               (scale-factor '(1 1 1))
                               (scale-orientation-axe '(0 0 1))
                               (scale-orientation-value 0)
                               (center '(0 0 0))
                               (name nil))
  (format geometry-outfile 
          (concatenate 'string
                       (if name 
                         (format nil "DEF ~A " name)
                         "")
"Transform {
   translation       ~{~,6F ~}
   rotation          ~{ ~A ~} ~,6F
   scaleFactor       ~{~,6F ~}
   scaleOrientation  ~{~,6F ~} ~A
   center            ~{~,6F ~}
 }
")
          translation
          rotation-axe 
          rotation-value
          scale-factor
          scale-orientation-axe
          scale-orientation-value
          center))
    
                                    

(defun ROTATION-TRANSFORM-P (label)
  (equal label 'R))

(defun TRANSLATE-TRANSFORM-P (label)
  (equal label 'TRANSLATE))

(defun SCALE-TRANSFORM-P (label)
  (equal label 'S))

    


(defun COLLECT-VARIABLES-OCCURENCE (form)
  (declare (special plasm-function-variables-list))
  (let (res)
    (dolist (i plasm-function-variables-list (reverse res))
      (setf res (cons (HOW-MANY-IS-THERE-INSIDE i form) res)))))


(defun MAKE-LABEL (occurence-list)
  (declare (special plasm-function-variables-list))
  (eval 
   (append '(concatenate 'string)
           (mapcar (lambda (x y)
                     (if (= 0 x)
                       ""
                       (string y)))
                   occurence-list
                   plasm-function-variables-list))))
           

(defun BUILD-MOTION-FORMAT-LABELS ()
  (declare (special plasm-function-struct motion-labels))
  (let ((sum (eval `(mapcar (function +)
                            ,@(mapcar (lambda (x) (list 'quote (third x)))
                                      motion-labels))))
        ordered-motion-labels)
    (if (or (not (every (lambda (x) (= 1 x)) sum))
            (some (lambda (x) (= 0 x)) sum))
      (plasm-error 7005 (plasm-function-name plasm-function-struct)))
    (setf ordered-motion-labels
          (sort  motion-labels
                 (lambda (x y)
                   (< (position 1 (third x))
                      (position 1 (third y))))))
    (setf motion-labels ordered-motion-labels)))
