; -*- package: plasm -*-
;---------------------------------------------------------------------------
;                    PLASMERROR.LISP FILE
;---------------------------------------------------------------------------

(provide 'error)
(inpackage plasm)


(defun PLASM-ERROR (code  &rest args)
  (declare (special *DEBUGGING* **DEFINITION-EVALUATING**))
  
  #+:APPLE
  (window-select (find-window "listener"))
  
  
  (if **DEFINITION-EVALUATING** (reset-environment-definition))
  

  (case code
    
    ( 0 nil)
    ( 1 (format  t "~%~%Attempt to take functional value of symbol ~A !~%~%"
		(print-plasm (first args))))
    ( 2 (format  t "~%~%Attempt to apply the function ~A to the function ~A !~
		~%Do you need COMPOSITION ?~%~%" 
	       (print-plasm (first args)) (print-plasm (second args))))
    ( 3 (format  t "~%~%~
		Affine transformation ~A needs a sequence of integers and ~
		a sequence of real.~%On the contrary, YOU have dispatched ~
		me ~A and ~A !~%~%" 
		 (print-plasm (first args)) (print-plasm (second args))
		 (print-plasm (third args))))
    ( 4 (format  t "~%~%Too much argument to function ~A !~%~%" 
		 (print-plasm (first args))))
    ( 5 (format  t "~%~%Function ~A does not exist !~%~%" 
		 (print-plasm (first args))))
    ( 6 (format  t "~%~%Attempt to take the ~A^ element of the sequence ~A !~%~%" 
	       (print-plasm (first args)) (print-plasm (second args))))
    ( 7 (format  t "~%~%Function ~A can't be applied to ~A argument !~%~%"
	       (print-plasm (first args)) (print-plasm (second args))))
    (10 (format  t "~%~%I can't bind ~A to ~A !~%~%"
	       (print-plasm (first args)) (print-plasm (second args))))
    (100 (format  t "~%~%CHECKER function failed for ~A~%~%" 
		  (print-plasm (first args))))
    (101 (format  t "~%~%Duplicated name in parameter sequence : ~A .~%~%" 
		(print-plasm (first args))))
    (102 (format  t "~%~%Duplicated name in specifiers sequence : ~A .~%~%" 
		(print-plasm (first args))))
    (103 (format  t "~%~%Duplicated name in default sequence : ~A .~%~%" 
		(print-plasm (first args))))
    (104 (format  t "~%~%Duplicated name in local functions : ~A .~%~%" 
		(print-plasm (first args))))
    (105 (format  t "~%~%There are names duplicated either in parameter(s) - ~A - ~
		 or local function(s) - ~A -. ~%~%" 
		(print-plasm (first args))
		(print-plasm (second args))))
    (106 (format  t "~%~%There exists name(s) for default parameter(s) ~A ~
		 without the matching parameter(s): ~A .~%~%" 
		(print-plasm (first args)) (print-plasm (second args))))
    (200 (format  t "~%~%The argument ~A is not of type ~A .~%~%" 
		(print-plasm (first args)) (print-plasm (second args))))
    ((201 202) (format  t "~%~%~A is not long ~A as declared in definition ~
			   of function ~A.~%~%" 
		(print-plasm (first args)) 
		(print-plasm (second args))
		(print-plasm (third args))))

    (300 (format  t "~%~%~A is a body for the function? I do not think at all?~%~%" 
		(print-plasm (first args))))

    (301 (format  t "~%~%~A is a no good function name! Please rename your function.~%~%" 
		(print-plasm (first args))))

    (403 (format  t 
                  "~%~%You have typed an open remark. Please, check the following Plasm code:~%~%~A~%~%~%" 
                  (print-plasm (first args))))

    (600 (format  t "~%~%The function ~A cannot work on ~A as argument !~%~%"
		(print-plasm (first args)) (print-plasm (second args)) ))

    (601 (format  t "~%~%The function ~A does not use a function as type check espression !~%~%"
		(print-plasm (first args)) ))

    (602 (format  t "~%~%The function ~A cannot work on ~{ (~A) ~} with one/or all of its argument/s !~%~%"
		(print-plasm (first args)) (mapcar #'print-plasm (second args)) ))

    (700 (format  t "~%~%The function CAMERA cannot work with ~A and/or ~A !~%~%"
		(print-plasm (first args)) (print-plasm (second args))))

    (1000 (format  t "~%~%Function ~A can't be applied to ~A argument !~%~%"
	       (print-plasm (second args)) (print-plasm (first args))))


    (2000 (format  t "~%~%The argument ~A is not valid to STRUCT operator !~%~%"
		(print-plasm (first args))))
    (2001 (format  t "~%~%The sequence argument to a STRUCT function has ~
		  polyhedra complexes of different dimensions!~%~%"))
    (2002 (format  t "~%~%The argument ~A is not valid for an export operation!~%~%"
		   (print-plasm (first args))))
    (6000 (format  t "~%~%The function TRANS can be applied to matrix only, not to value ~A !~%~%"
		   (print-plasm (first args))))

    ;7000 series: error for animation Plasm features 
    (7000 (format  t "~%~%Sorry, you've used unimplemented features for MOVE function!~%~%"))
 
    (7001 (format  t "~%~%The function MOVE can be applied to curried function ~A !~%~%"
		   (print-plasm (first args))))
    (7002 (format  t "~%~%Function ~A must be used inside a Plasm Function Definition !~%~%"
	           (print-plasm (first args))))
    (7003 (format  t "~%~%The configuration space dimension (that is ~A) of the scene given by Plasm function~%
~A is not equal to ~A (that is the # of values provided)!~%~%"
                   (print-plasm (second args)) (print-plasm (first args)) (print-plasm (third args))))
    (7004 (format  t "~%~%You ca not use the affine transformation ~A applied to this/these ~A!~%~%"
	           (print-plasm (first args)) (print-plasm (second args))))
    (7005 (format  t "~%~%Please check the use of formal parameters inside function ~A!~%
 Probably missing or multiple use of some variable:  error for animation request.~%"
	           (print-plasm (first args))))
    (7006 (format t "~%~%Please, check if directory ~A exists. If not, directory you have typed (~A) can not be created.~%"
                  (first args) (second args)))

    (7007 (format t "~%~%A MOVEMESH function has been applied to ~A while it waited for a pair of positive integers.~%"
                  (print-plasm (first args))))

    (7008 (format t "~%~%A MOVEMESH function has been applied to ~A while it waited for a sequence of polyhedral complexes.~%"
                  (print-plasm (first args))))
    (7009 (format t "~%~%A MOVEMESH function has been applied to ~A while it waited for a sequence of numbers.~%"
                  (print-plasm (first args))))

    (7010 (format t "~%~%A MOVEMESH function has been applied to <~A,~A> while at least one of polyhedral complexes ~
in the sequence doesn't have ~A vertices, that is, the sequence of polyhedra cannot represent an animated mesh.~%"
                  (first args) (second args) (* (first args) (second args))))

    (8000 (format t "~%~%The function ~A cannot be applied to the pair ~A.~%"
                  (print-plasm (first args)) (print-plasm (second args))))


    (t (format  t "error code : ~A  with these args : ~A~%~%" code args))
     )

  ;(if function (help-plasm-function function))
  
  (if *DEBUGGING* 
    (break "~A" """")
    (abort ))
    )
