; -*- package: plasm -*-
;---------------------------------------------------------------------------
;                    EVAL_P.LISP FILE
;---------------------------------------------------------------------------

(provide 'eval_p)
(inpackage plasm)


;-------------------------Form to eliminate-----------------------------------
(mapcar #'fmakunbound 
	'(FIRST-SON SECOND-SON THIRD-SON FOURTH-SON
	  FIFTH-SON SIXTH-SON SEVENTH-SON  EIGHTH-SON))
;--------------------------------------------------------------------------



;==================  FUNCTIONS ON SYNTAX-TREE  =============================
; A syntax-tree is of this type : 
;           (NODE SONS)
; where 
;           NODE == (DIGIT NON-TERMINAL)
; and 
;           SONS == (FIRST-SON SECOND-SON......) 
; while 
;           FIRST-SON, SECOND-SON..... 
; are syntax-tree. 
;===========================================================================




;---------------------------------------------------------------------------
;Given a syntax (sub)tree this function returns his sons. 

(defun PICK-UP-SONS (syntax-tree)
  (second syntax-tree))

;--------------------------------------------------------------------------
;This  function returns the 'i'-th son of a syntax-tree 'tree'.

(defun PICK-UP-SON-NUM (i tree)
  (let ((son (nth (1- i) (pick-up-sons tree))))
    (if son 
      son
      (plasm-error 5 i (length (pick-up-sons tree))))))

;------------------------------------------------------------------------
;These are the constants which values are the symbols for NON-TERMINALs. 

(defconstant *TOKENID* 'TOKENID)
(defconstant *SEQUENCE* 'SEQUEN)
(defconstant *APPLICAT* 'APPLIC)
(defconstant *CONSTR* 'CONSTR)
(defconstant *COMPOSE* 'COMPOS)
(defconstant *INFEXPR* 'INFEXPR)
(defconstant *VALUE* 'VALUE)
(defconstant *EXPRES* 'EXPRES)
(defconstant *DEFIN* 'DEFIN)

;-------------------------------------------------------------------------
;
;         SCRIPT ::= EXPRES , SCRIPT
;         SCRIPT ::= DEFIN , SCRIPT
;         SCRIPT ::= EXPRES 
;         SCRIPT ::= DEFIN




(defun EVAL-SCRIPT (tree) 
  (if tree
      (eval-script1 tree)))


(defun EVAL-SCRIPT1 (tree)
  (let ((sons (pick-up-sons tree)))
    (if (= 1 (length sons))
      (eval-defin-or-expres (first sons))
      (progn
	(eval-defin-or-expres (first sons))
	(eval-script1 (third sons))))
    (values)))






(defun EVAL-DEFIN-OR-EXPRES (tree)
  (let ((root (first tree)))
    (cond
     ((is-there-in *EXPRES* root)
      (show-expression-value (eval-expres tree)))
     ((is-there-in *DEFIN* root)
      (show-definition-value
       (eval-defin tree))))))

;---------------------------------------------------------------------------
;This function eval an EXPRESS syntax-tree checking what kind of expression 
;it is and by calling the rigth function. We have some rules that tell us
;what expression have to manage. In the COND form there is the filter that 
;choose the right eval function. 

(defun EVAL-EXPRES (tree-expr)
  (let* ((tree (pick-up-son-num 1 tree-expr))
	 (root (first tree)))
    (cond ((is-there-in *TOKENID* root)  (eval-tokenid tree))
	  ((is-there-in *SEQUENCE* root) (eval-sequence tree))
	  ((is-there-in *APPLICAT* root) (eval-applic tree))
	  ((is-there-in *CONSTR* root)   (eval-constr tree))
	  ((is-there-in *COMPOSE* root)  (eval-compos tree))
	  ((is-there-in *INFEXPR* root)  (eval-infexpr tree))
	  ((is-there-in *VALUE* root)    (eval-value tree)))))

;---------------------------------------------------------------------------
;IS-THERE-IN function is a predicate :it returns T if his first argument is 
;member of his second !
;~~~~~~

(defun IS-THERE-IN (item pair-list)
  (eq item (second pair-list)))



;---------------------------------------------------------------------------
;This function takes a syntax-tree which node is *TOKENID* and returns the
;symbol substituted by LEX function with a token. 

    
(defun EVAL-TOKENID (tree &optional (take-memory T))
  (let* ((token (take-leaf-tokenid-tree tree)) 
	 (sym-name (get-token-value token)))
    (cond
     ((not sym-name) sym-name) ; sym-name is nil/false
     ((eq sym-name T) sym-name) ; sym-name is true
     ((eq sym-name 'df) sym-name) ; sym-name is true
     ((and (definition-p)
	   take-memory
	   (symbolp sym-name))
      (processed-name sym-name))
     ((definition-p) sym-name)
     ((symbolp sym-name) (set-and-or-return-value sym-name))
     (T sym-name))))


; TOKENID syntax tree is a special syntax tree 
; (see GIVE-THEM-A-FATHER function in file 'syntax-tree.lisp')
(defun TAKE-LEAF-TOKENID-TREE (tree)
  (second tree))


;--------------------------------------------------------------------------
;GET-TOKEN-VALUE function takes a string three chars long and returns a
;symbol that match the string substituded by LEX function in the script with
;the token which radix is the argument string. 
; Some PLASM functions have names equal to LISP functions. 
; Then, a translation of these symbol names is here made to eliminate these
; collisions. Moreover, some symbols in sym-string are caught because they
;can't be used with READ-FROM-STRING lisp function. To these sym-strings is
;to be given a special symbol whitout 'read' it. 


(defun GET-TOKEN-VALUE (token)
  (if (symbolp token)
    (let ((sym-string (symbol-name token)))
      (cond
       ((is-skeleton sym-string) 
	(reduce-function-args 'SKELETON (read-from-string (subseq sym-string 1))))
       ((is-sel sym-string)
	(reduce-function-args 'SEL (read-from-string (subseq sym-string 1))))
       (T token)))
    token))






(defun IS-SEL (string-)
  (and (> (length string-) 1)
       (or (char= #\S (char string- 0))
	   (char= #\s (char string- 0)))
       (integerpos-p (read-from-string (subseq string- 1)))))



(defun IS-SKELETON (string-)
  (and 
   (char= #\@ (char string- 0))
   (integernotneg-p  (read-from-string (subseq string- 1)))))


(defun INTEGERPOS-P (arg)
  (and (integerp arg)
       (> arg 0)))

(defun INTEGERNOTNEG-P (arg)
  (and (integerp arg)
       (>= arg 0)))


;-------------------------------------------------------------------------
;This function returns T if his argument is a string that start and end 
;with a quote and is long more than 2 chars ; otherwise returns NIL. 

(defun IS-USER-STRING (string-)
  (and (> (length string-) 1)
       (char= #\' (char string- 0))
       (char= #\' (char string- 
			(1- (length string-))))))

;-------------------------------------------------------------------------
;Here you are the SEQUENCE construct interpreter !The sequence rule is
; sequence ::= < { expression }* >. There are more than one rule used by
;automata to recognize a sequence. Everyone will be shown in the following.
;A sequence will be always translated in a list which elements are evaluated
;expressions.

(defun EVAL-SEQUENCE (tree)
  (let ((sons (pick-up-sons tree)))
    (if (= 2 (length sons))
	;=== we have an empty sequence ===
	'()
	;=== we have a not empty sequence ===
	(let* ((seqg-tree (second sons)) ;=== SEQG ::= SEQNOEM  rule ! ===
	       (son-seqg (pick-up-son-num 1 seqg-tree)))
	  (eval-seqnoem son-seqg)))))


;-------------------------------------------------------------------------
;In the PLASM language if something can be generated by the rule
;                expression {,  expression }*
;it will be caught by the SEQNOEM rule of PLASM grammar.The value of such
;construct is the list made of evaluated expression.
;There are two rules :
;   SEQNOEM ::= EXPRESS 
;   SEQNOEM ::= EXPRESS,  SEQNOEM         
;We must catch these two rules !

(defun EVAL-SEQNOEM (tree)
  (let ((sons (pick-up-sons tree)))
    (if (= 1 (length sons))
	;=== we have SEQNOEM ::= EXPRESS ===
	(let* ((value-expr (eval-expres (first sons))))
	  (list value-expr))
	;=== we have SEQNOEM ::= EXPRESS,  SEQNOEM  ===
	(let* ((expr-son (first sons))
	       (value-expr (eval-expres expr-son))
	       (value-seqnoem (eval-seqnoem (third sons))))
	    (cons value-expr value-seqnoem)))))


;---------------------------------------------------------------------------
;  Rule to detect     COMPOS ::= EXPRES ~ EXPRESS
;                                  /\        /\
;                                  ||        ||
;  in code we've                 expres2   expres1



(defun EVAL-COMPOS (tree)
  (let* ((expres1-value (eval-expres (pick-up-son-num 3 tree)))
	 (expres2-value (eval-expres (pick-up-son-num 1 tree))))
    (eval-compos-after expres2-value expres1-value)))





(defun EVAL-COMPOS-AFTER (expres2-value expres1-value)
  (declare (special local-function-names))
    (cond  ;there are these possible cases  :
     ((definition-p) ; it is defining a new function
      `(reduce-function 
	'comp
	(list 
	,(cond
	  ((atomic-p expres2-value) expres2-value)
	  (T (make-auto-eval-form expres2-value)))
	,(cond
	 ((atomic-p expres1-value) expres1-value) ; arg is a parameter name
	 (T (make-auto-eval-form expres1-value))))))        ; arg is a cons

     ;  f ~ g ... and we've to return a new function 


      ; the two or's can be substituted by (bound-function-p expres1-value)
     ((and (bound-function-p expres1-value)  
	   (bound-function-p expres2-value))
      (reduce-function-args 
       'comp 
       (list expres2-value  expres1-value)))
	 
     ; expres2-value or expres1-value  is not a functional value, then we have :
     ; 1. an error 
     ; 2. something else not in my mind, at the moment I hope !

     ( T (plasm-error 1  expres1-value expres2-value))))









;+++++++++++++++++++++++++



;---------------------------------------------------------------------------
;This function returns a keylist valid for fenv !It takes two arguments that
;must be symbols bound in fenv, and build a lisp form able to compute the
;application of function g on his argument and then the application of the
;function f on this result. 
;This lisp form is what we need in PLASM while we are interpreting a script
;of this kind :   PrimaFunzione ~ SecondaFunzione ~ TerzaFunzione   !!!!

 



(defun COMPOSE-FORM (f g) (reduce-function-args 'comp (list f g)))



;EXAMPLE OF USE
;
;In fenv there are these bindings :
;
;   (f (x1) (function1 (function2 x1 (+ 2 x1))))
;
;                     and
;
;   (g (x1 x2) (Function34 x1 (funcall x2)))      . 
;
;If we call (COMPOSE-FORM 'f 'g) we'll get this lisp form :
;
;     (***COMP123 (x1 x2) (let ((**x24 (g x1 x2))) (f **x24)))    . 
;     --------------------------------------------------------
;
;It should be clear that function f have to take exactly one argument, the 
;one returned from g !!!!
;---------------------------------------------------------------------------


;---------------------------------------------------------------------------
;Rules to detect  SPECSEQ ::= { SEQNOEM }
;                 SPECSEQ ::= { }



(defun EVAL-SPECSEQ (tree)
  (let ((sons (pick-up-sons tree)))
    (if (= 2 (length sons))
      ; we have the rule SPECSEQ ::= { }
      nil 
      ; we've the rule SPECSEQ ::= { SEQNOEM }
      (let ((val (eval-seqnoem (pick-up-son-num 2 tree))))
	(if (= 1 (length val)) ;if SEQNOEM is made of an element then
			       ; we return an element (not a list of
			       ; an element
	  (first val)
	  val)))))




;--------------------------------------------------------------------------
;  Rule to detect   SPECIF ::= TOKENID SPECSEQ

(defun EVAL-SPECIF (tree)
  (let* ((sons (pick-up-sons tree))
	 (func-name (eval-tokenid (first sons)))
	 (specif-list-value (eval-specseq (second sons))))
    (reduce-function-args func-name specif-list-value)))



;--------------------------------------------------------------------------
; Rule to detect   CONSTR ::= [ SEQNOEM ]
;
;  [f,g,h]:x == <f:x,g:x,h:x>
;
;where f, g and h are to be bound symbols with functional values. 
;We can solve this behaviour by putting in default environment a high order
;function that takes one argument (list of functions) and returns a new
;function of one argument. The returned function will work on his only
;argument making a list of values. We can see this behaviour like a reverse
;sense of MAPCAR LISP form. 

(defun EVAL-CONSTR (tree)
  (let* ((seqnoem-tree (pick-up-son-num 2 tree))
	 (list-of-funs (eval-seqnoem seqnoem-tree)))
    (reduce-function-args 'CONS list-of-funs)))



;--------------------------------------------------------------------------
;Rule to detect   APPLIC ::= EXPRES : EXPRES
;                              /\        /\
;                              ||        ||
;  in code we've             expres2 : expres1

(defun EVAL-APPLIC (tree )
  (let* ((expres1-value (eval-expres (pick-up-son-num 3 tree)))
	 (expres2-value (eval-expres (pick-up-son-num 1 tree))))
    ;(if (and (bound-function-p expres2-value)
    ;         (not (bound-function-p expres1-value )))
	(reduce-function-args expres2-value expres1-value)
    ;    (plasm-error 2 expres2-value expres1-value))
))

;--------------------------------------------------------------------------
;  Rule to detect  VALUE ::= ( EXPRES )

(defun EVAL-VALUE (tree)
   (eval-expres (pick-up-son-num 2 tree)))


;--------------------------------------------------------------------------
;  Rule to detect INFEXPR ::= EXPRES TOKENID EXPRES 
;
;  Tokenid is to be a valid symbol for a function taking a pair. 
;Then that syntax is equivalent to this application :
;         TOKENID : < EXPRES1,  EXPRES2 >

(defun EVAL-INFEXPR (tree)
  (let* ((fun (eval-tokenid (pick-up-son-num 2 tree)))
	 (firstop (eval-expres (pick-up-son-num 1 tree) ))
	 (secondop (eval-expres (pick-up-son-num 3 tree))))
    (reduce-function-args fun (list firstop secondop))))


;--------------------------------------------------------------------------
;  Rule to detect INFEXPR ::= EXPRES EXPRES EXPRES 
;
;  EXPRES in the middle has to be  a function taking a pair. 
;Then that syntax is equivalent to this application :
;        mid-EXPRES : < EXPRES1,  EXPRES2 >

(defun EVAL-INFEXPR (tree)
  (let* ((fun (eval-expres (pick-up-son-num 2 tree)))
	 (firstop (eval-expres (pick-up-son-num 1 tree) ))
	 (secondop (eval-expres (pick-up-son-num 3 tree))))
    (reduce-function-args fun (list firstop secondop))))



;--------------------------------------------------------------------------

(defun MAKE-GENERAL-LAMBDA-FORM (a-plasm-func-struct)
  (let ((parameters (plasm-function-formal-pars a-plasm-func-struct)))
    (if parameters
      (let* ((pars (reverse parameters)) 
	     (lambda-function (make-lambda-function 
			       (first pars) 
			       (plasm-function-body a-plasm-func-struct))))
	(dolist (par (rest pars) (eval lambda-function))
	  (setf lambda-function (make-lambda-function par lambda-function))))
      (eval (plasm-function-body a-plasm-func-struct)))))



(defun MAKE-LAMBDA-FUNCTION (arg body)
  (if arg
    `(function (lambda (,arg) ,body))
    `(function (lambda () ,body))))


#+:APPLE
(defun SET-ANONYMOUS-FUNCTION-VALUE (funstructure)
  (setf (plasm-function-value funstructure)
	(make-general-lambda-form funstructure)))

#+:RS6000
(defun SET-ANONYMOUS-FUNCTION-VALUE (funstructure)
  (let ((val (make-general-lambda-form funstructure)))
    (setf (plasm-function-value funstructure)
      (if (functionp val)
	(compile nil (function-lambda-expression val))
	val))))

 #+:clisp
(defun SET-ANONYMOUS-FUNCTION-VALUE (funstructure)
  (let ((val (make-general-lambda-form funstructure)))
    (setf (plasm-function-value funstructure)
      (if (functionp val)
	(compile nil (function-lambda-expression val))
	val))))

      



;--------------------------------------------------------------------------
;REDUCE FUNCTION-ARG is the high order function constructor :
;it returns either a name bound in fenv, that is a new function taking more 
;zero parameters, or a value returned from a reduced function became a 
;constant (thus is a function with zero parameters !). 
;
; (reduce-function-args Fname Arg ... ) ==> NewFname
;
; if Fname function definition likes 
;
;      Fname : Arg -> (OthersArg -> "VALUE")
;
; (reduce-function-args Fname Arg... ) ==> "VALUE"
;
; if Fname function definition likes
;
;      Fname : Arg -> "VALUE"




;+++++++++++++++++++++


(defvar *NOT-GENERABLE* '%%%np) ;nobody can generate this symbol


(defun REDUCE-FUNCTION-ARGS (func &optional (arg *NOT-GENERABLE*))   
   (declare (special *NOT-GENERABLE* param+specif+local-names))
   (cond
    
    ((definition-p)
     `(reduce-function 
       ,(cond
	 ((consp func) func)
	 ((member func param+specif+local-names) func)
	  (T `',func))
       ,(cond
	 ((member arg param+specif+local-names) `,arg) ; arg is a parameter name
	 ((symbolp arg) `',arg)                  ; arg is a plasm-function
	 (T (make-auto-eval-form arg)))))        ; arg is a cons


   (T (reduce-function func arg))))

(defun REDUCE-FUNCTION-ARGS (func &optional (arg *NOT-GENERABLE*))   
   (declare (special *NOT-GENERABLE* local-function-names))
   (cond
    
    ((definition-p)
     `(reduce-function 
       ,(cond
	 ((atomic-p func) func)
	  (T (make-auto-eval-form func)))
       ,(cond
	 ((atomic-p  arg) arg); arg is a parameter name
	 (T (make-auto-eval-form arg)))))       ; arg is a cons


   (T (reduce-function func arg))))




;************************************************************************


(defun REDUCE-FUNCTION (func &optional (arg *NOT-GENERABLE*)
					    &key (run t))
   (declare (special *NOT-GENERABLE*))
  (let* ((lisp-function (if (functionp func) func))

	 (function-def (if (not lisp-function) 
			 (bound-function func))))
    (cond

     (lisp-function)

     ((and function-def
	     (plasm-function-value function-def))
      (setf lisp-function (plasm-function-value function-def)))

     (function-def 
      (setf lisp-function (set-anonymous-function-value function-def))))

    (cond
     ((and (functionp lisp-function) (equal arg *NOT-GENERABLE*))
      (if run
	(funcall lisp-function)
	lisp-function)) 
     ((and (functionp lisp-function) 
	   (not (equal arg *NOT-GENERABLE*)))
      (funcall lisp-function arg)) ; must be always runned if there are arg
     (T lisp-function))))  ; it is always a function if it comes from
			   ; a definition without parameters




;-------------------------------------------------------------------------




;====================================================================

(defun TRACE-PLASM () 
  (declare (special *PLASM-TRACE*))
  (setf *PLASM-TRACE* T))

(setf *PLASM-TRACE* nil)

(defun UNTRACE-PLASM ()
  (declare (special *PLASM-TRACE*))
  (setf *PLASM-TRACE* nil))

(defun DEBUG-PLASM ()
(declare (special *DEBUGGING*))
  (setf *DEBUGGING* T))

(setf *DEBUGGING* nil)

(defun NO-DEBUG-PLASM ()
(declare (special *DEBUGGING*))
  (setf *DEBUGGING* nil))


;====================================================================
