; -*- package: plasm -*-
;==========================================================================
;
;                           FENVSTRUCT.LISP File
;
;==========================================================================
; ATTENTION: If you want to modify a plasm predefined function please see
; at the function QUOTE_

(provide 'fenvs)
(inpackage plasm)

;----------------------------------------------------
; Affine Transformation names 

(setf *SCALING* 'S)
(setf *ROTATION* 'R)
(setf *TRANSLATION* 'TR)
(setf *SHEARING* 'H)
(setf *MAT* 'MAT)
(setf *AFFINE-TRANSFORMATION-NAMES* 
      (list *SCALING* *ROTATION* *TRANSLATION* *SHEARING*))

;----------------------------------------------------

(setf **USER-FENV** (make-hash-table))

;-------------------------------------------------------------------------
;This variable will be a constant when it will be completed. 
;**DEFAULT-FENV** is the primitive functional environment. In a new PLASM
; SESSION the functional environment contains the only functions defined in
;**DEFAULT-FENV** constant. 

(setf **DEFAULT-FENV** (make-hash-table))


(defun MAKE-PLASM-FUNCTION-FROM-SIMPLE-CONS (name-formal-pars-let-vars-body)
  (make-plasm-function 
   :name (first name-formal-pars-let-vars-body)
   :formal-pars (second name-formal-pars-let-vars-body)
   :body (third name-formal-pars-let-vars-body)))

(defun MAKE-PLASM-FUNCTION-FROM-SIMPLE-CONS (simple-cons)
  (make-plasm-function 
   :name (first simple-cons)
   :formal-pars (second simple-cons)
   :body (third simple-cons)
   :function-description (fourth simple-cons)))


(defun SET-FUNCTION-DESC (func-string desc-string)
  (let ((sym-fun (intern (string-upcase func-string))))
    (if (bound-function-p sym-fun)
      (progn
        (setf (plasm-function-function-description  (bound-function sym-fun))
            desc-string)
        sym-fun)
      (plasm-error 1 sym-fun))))


(defun HELP-PLASM-FUNCTION (func &optional returnit)
  (if (and (symbolp func)
           (bound-function-p func))
    (visualize-help-function (bound-function func))
    (cond 
      (returnit func)
      (T (values)))))


(defun VISUALIZE-ALL-HELP-FUNCTION ()
  (declare (special **DEFAULT-FENV**))
  (maphash (lambda (k d) (visualize-help-function d)) **DEFAULT-FENV**)
  (values))

(defun VISUALIZE-HELP-FUNCTION (plasm-func-struct)
  #+:CLISP
  (format t "~%=============~%The function ~A has the following description:~%~
  ~A~%~
============="
          (plasm-function-name plasm-func-struct)
          (if (plasm-function-function-description plasm-func-struct)
            (plasm-function-function-description plasm-func-struct)
            " - No description has been set! -"))
  #+:APPLE
  (message-dialog
   (format nil "The function ~A has the following description:~%~A"
           (plasm-function-name plasm-func-struct)
           (if (plasm-function-function-description plasm-func-struct)
             (plasm-function-function-description plasm-func-struct)
             " - No description has been set! -")
           )
   :size (make-point 400 300))
   (values))

;------------------------------------
;  i simboli usati nella slot FORMAL-PARS non possono essere
;simboli generabili dall'utente per definire le sue funzioni
;  PERCHE? non lo ricordo piu'!



(mapcar #'(lambda (fun)
	    (declare (special **DEFAULT-FENV**))
	    (INSERT-IN-DEFAULT-FENV fun))
	(mapcar #'make-plasm-function-from-simple-cons

		'(
      


(SET_FUNCTION_DESC 
 (FUN- DESC-) 
 (SET-FUNCTION-DESC FUN- DESC-) 
 "This function is used to set a descrition for a function by the language. It's use is: 
SET_FUNCTION_DESC : 'function_name' : 'a text description'."
)
(HELP_FUNCTION (FUN-) (HELP-PLASM-FUNCTION FUN- T))
      ;==================================================================
      ;    HIGH ORDER FUNCTIONS

(CONS (FUNS-LIST- ARG-)
      (LET (RES)
	(DOLIST (FUN FUNS-LIST- RES)
	  (SETF RES (APPEND RES (LIST (REDUCE-FUNCTION-ARGS FUN ARG-)))))))
(AA (F ARG-)
    (LET (RES)
      (DOLIST (ELEM ARG- RES)
	(SETF RES (APPEND RES (LIST (REDUCE-FUNCTION-ARGS F ELEM)))))))
(LIFT (F- FUNS-LIST- X-)
      (REDUCE-FUNCTION-ARGS
       F-
       (LET (RES)
	 (DOLIST (FUN FUNS-LIST- RES)
	   (SETF RES (APPEND RES (LIST (REDUCE-FUNCTION-ARGS FUN X-))))))))
(C (F- X- Y-) (REDUCE-FUNCTION-ARGS F- (LIST X- Y-)))

(COMP (FUNS-LIST- ARG-)
      (DOLIST (FUN (REVERSE FUNS-LIST-) ARG-)
	(SETF ARG- (REDUCE-FUNCTION-ARGS FUN ARG-))))

(IF (PRED-AND-TWO-FUNS- ARG-)
  (LET ((PRED (FIRST PRED-AND-TWO-FUNS-))
	(THENOP (SECOND PRED-AND-TWO-FUNS-))
	(ELSEOP (THIRD PRED-AND-TWO-FUNS-)))
    (IF (REDUCE-FUNCTION-ARGS PRED ARG-)
      (REDUCE-FUNCTION-ARGS THENOP ARG-)
      (REDUCE-FUNCTION-ARGS ELSEOP ARG-))))
(K (X- Y-) (PROG2 Y- X-))

(INSL (FUN- LIST-)
      (COND ((NULL LIST-) (PLASM-ERROR 3000 'INSL NIL)) 
	    (T (INSL-LISP FUN- LIST-))))
(INSR (FUN- LIST-)
      (COND ((NULL LIST-) (PLASM-ERROR 3000 'INSR NIL)) 
	    (T (INSR-LISP FUN- LIST-))))

(TREE (FUN- LIST-)
      (COND ((NULL LIST-) (REDUCE-FUNCTION-ARGS FUN- LIST-)) 
	    (T (TREE-LISP FUN- LIST-))))
(MERGE (FUN- LIST-)
       (IF (AND (OR (FUNCTIONP FUN-) (BOUND-FUNCTION-P FUN-))
		(LISTP (FIRST LIST-))
		(LISTP (SECOND LIST-)))
	 (MERGE-LISP FUN- LIST-)
	 (PLASM-ERROR 7 'MERGE LIST-)))
(APPLY (FUN-ARG-SEQ-)
       (IF (ISPAIR FUN-ARG-SEQ-)
	 (REDUCE-FUNCTION-ARGS (FIRST FUN-ARG-SEQ-) (SECOND FUN-ARG-SEQ-))
	 (PLASM-ERROR 3000 FUN-ARG-SEQ-)))


(CASE (PRED-FUN-S ARG)
 (DOTIMES (I (LENGTH PRED-FUN-S) NIL)
  (LET ((PRED-FUN (NTH I PRED-FUN-S)))
   (IF (REDUCE-FUNCTION (FIRST PRED-FUN) ARG)
       (RETURN (REDUCE-FUNCTION (SECOND PRED-FUN) ARG))))))

;====================================================================
;  AFFINE TRANSFORMATIONS
;TRANSLATION

(TRANSLATE (SPEC- PAR- POL-)
	   (LET ((PAR-LIST (IF (ATOM PAR-) (LIST PAR-) PAR-))
		 (SPEC-LIST (IF (ATOM SPEC-) (LIST SPEC-) SPEC-)))
	     (DECLARE (SPECIAL *TRANSLATION*))
	     (IF (OR (/= (LENGTH SPEC-LIST) (LENGTH PAR-LIST))
		     (NOT (EVAL (CONS 'AND (MAPCAR #'INTEGERP SPEC-LIST)))))
	       (PLASM-ERROR 3 'T SPEC-LIST PAR-LIST)
	       (MAKE-MULTIPLE-AFFINE-TRANSFORMATION
		(SORT (MAPCAR #'(LAMBDA (X Y) (LIST X Y)) SPEC-LIST PAR-LIST)
		      #'<
		      :KEY
		      #'CAR)
		POL-
		*TRANSLATION*))))
(S (SPEC- PAR- POL-)
   (LET ((PAR-LIST (IF (ATOM PAR-) (LIST PAR-) PAR-))
	 (SPEC-LIST (IF (ATOM SPEC-) (LIST SPEC-) SPEC-)))
     (DECLARE (SPECIAL *SCALING*))
     (IF (OR (/= (LENGTH SPEC-LIST) (LENGTH PAR-LIST))
	     (NOT (EVAL (CONS 'AND (MAPCAR #'INTEGERP SPEC-LIST)))))
       (PLASM-ERROR 3 'S SPEC-LIST PAR-LIST)
       (MAKE-MULTIPLE-AFFINE-TRANSFORMATION
	(SORT (MAPCAR #'(LAMBDA (X Y) (LIST X Y)) SPEC-LIST PAR-LIST) #'< :KEY #'CAR)
	POL-
	*SCALING*))))

(MAT (MAT- POL-)
     (LET ()
       (DECLARE (SPECIAL *MAT*))
       (IF (AND (CONSP MAT-)
		(EVERY #'CONSP MAT-)
		(AND (> (LENGTH MAT-) 1) (APPLY #'/= (MAPCAR #'LENGTH MAT-))))
	 (PLASM-ERROR 3 'MAT MAT-)
	 (MAKE-MULTIPLE-AFFINE-TRANSFORMATION MAT- POL- *MAT*))))

(R (SPEC- PAR- POL-)
   (LET ((PAR-LIST (IF (ATOM PAR-) (LIST PAR-) PAR-))
	 (SPEC-LIST (IF (AND (CONSP SPEC-) (ATOM (FIRST SPEC-))) (LIST SPEC-) SPEC-)))
     (DECLARE (SPECIAL *ROTATION*))
     (IF (OR (/= (LENGTH SPEC-LIST) (LENGTH PAR-LIST))
	     (NOT (EVAL (CONS 'AND
			      (MAPCAR #'(LAMBDA (PAIR-LIST)
					  (AND (= 2 (LENGTH PAIR-LIST))
					       (INTEGERP (FIRST PAIR-LIST))
					       (INTEGERP (SECOND PAIR-LIST))))
				      SPEC-LIST)))))
       (PLASM-ERROR 3 'R SPEC-LIST PAR-LIST)
       (MAKE-MULTIPLE-AFFINE-TRANSFORMATION
	(MAPCAR #'(LAMBDA (X Y) (LIST X Y)) SPEC-LIST PAR-LIST)
	POL-
	*ROTATION*))))
(H (SPEC- PAR- POL-)
   (LET ((SPEC-- (IF (EVERY #'NUMBERP SPEC-) (LIST SPEC-) SPEC-)))
     (DECLARE (SPECIAL *SHEARING*))
     (COND ((AND (= 1 (LENGTH SPEC--)) (CONSP PAR-) (EVERY #'NUMBERP PAR-))
	    (LIST 'H (LIST (FIRST (FIRST SPEC--)) (CDR (FIRST SPEC--)) PAR-)))
	   ((AND (= (LENGTH SPEC--) (LENGTH PAR-))
		 (EVERY #'CONSP PAR-)
		 (EVAL (CONS 'AND
			     (MAPCAR #'(LAMBDA (LIST-) (EVERY #'NUMBERP LIST-))
				     PAR-))))
	    (MAKE-MULTIPLE-AFFINE-TRANSFORMATION
	     (MAPCAR #'LIST (MAPCAR #'FIRST SPEC--) (MAPCAR #'CDR SPEC--) PAR-)
	     POL-
	     *SHEARING*))
	   (T (PLASM-ERROR 3 'H SPEC- PAR-)))))
(STRUCT (SEQ-) (GIVE-ME-POLYHEDRAL-COMPLEX-OF-PLASM-STRUCTURE SEQ-))


;==============================
; Animation extension

;(XSTRUCT (SEQ-) (GIVE-ME-POLYHEDRAL-COMPLEX-OF-PLASM-STRUCTURE SEQ-))
(XSTRUCT (SEQ-) SEQ-)
(MOVE (MOVE-FUNCTION MOVE-DATA TIME-DATA) 
      (RESOLVE-ANIMATION MOVE-FUNCTION MOVE-DATA TIME-DATA))

(SET_TEMP_DIR (FILE-) (SET-TEMP-PLASM-DIR FILE-))
(ANIMATION (SEQ-) (ANIMATION SEQ-))

(ANIMATION_TO (FILE- SEQ-) (ANIMATION-TO FILE- SEQ-))

(FRAME (POL-COMPLEX- T1T2-)
 (FRAMES POL-COMPLEX- T1T2-))


(MKMESH$ (APPEARANCE- GRID-) (MAKE-MESH APPEARANCE- GRID-) "IT IS AN INTERNAL FUNCTION!")


(MKTEXTURE (PAIR-SEQ-)
           (MKTEXTURE PAIR-SEQ-)
"
please, now use MKTEXTURE:<>!

MKTEXTURE:<
            <'FILE',\"FILENAME\">,
            <'WRAPS',O \| 1>,
            <'WRAPT',O \| 1>,
            <'TRANSLATION', \"A PAIR OF REAL\" >,
            <'ROTATION',\"A REAL\">,
            <'SCALEFACTOR', \"A PAIR OF REAL\" >,
            <'CENTER', \"A PAIR OF REAL\" >,
           > -> \"A TETXURE\"
WHERE ANY PROPERTY (BUT FILE) TO MKTEXTURE IS OPTIONAL
(IT WILL USE DEFAULT VALUES)
")


(MKCOLOR (TRIPLE)
         (MKCOLOR TRIPLE T)
"MKCOLOR:<
          \"A REAL\",
          \"A REAL\",
          \"A REAL\"
         > -> \"A COLOR\"

WHERE ANY \"A REAL\" IS RANGING BETWEEN 0 AND 1.")



(MK_RGBCOLOR (TRIPLE)
         (MKCOLOR TRIPLE)
"MK_RGBCOLOR:<
          \"A REAL\",
          \"A REAL\",
          \"A REAL\"
         > -> \"A STANDARD MATERIAL\"

WHERE ANY \"A REAL\" IS RANGING BETWEEN 0 AND 1.")

(MKMATERIAL (PAIR-SEQ-)
            (IF (AND (CONSP PAIR-SEQ-)
                     (EVERY (LAMBDA (PAIR)
                              (AND (CONSP PAIR)
                                   (ISPAIR PAIR)
                                   (STRINGP (FIRST PAIR))
                                   (OR (NUMBERP (SECOND PAIR))
                                       (RGBCOLOR-P (SECOND PAIR)))))
                            PAIR-SEQ-))
              (MKMATERIAL PAIR-SEQ-)
              (PLASM-ERROR 8000 'MKMATERIAL PAIR-SEQ-))
"MKMATERIAL:<
             <'AMBIENT',\"A COLOR\">,
             <'DIFFUSE',\"A COLOR\">,
             <'SPECULAR',\"A COLOR\">,
             <'EMISSIVE',\"A COLOR\">,
             <'SHINESS',\"A REAL\">,
             <'TRANSPARENCY',\"A REAL\">,
            > -> \"A MATERIAL\"
WHERE ANY PAIR IS OPTIONAL
(IT WILL USE DEFAULT VALUES)
")


(MKAPPEARANCE (SEQ-)
            (IF (LISTP SEQ-)
              (MKAPPEARANCE SEQ-)
              (PLASM-ERROR 8000 'MKAPPEARANCE SEQ-))
"MKAPPEARANCE:<
             \"A TEXTURE\",
             \"A MATERIAL\">,
            > -> \"AN APPEARANCE\"
WHERE BOTH ARGUMENTS ARE OPTIONAL
(IT WILL USE DEFAULT VALUES)
")


;(MOVEMESH (ROWS-COLS- POLS- TIMES-)
;          (COND
;           ((NOT (ISPAIR ROWS-COLS-)) 
;            (PLASM-ERROR 7007 ROWS-COLS-))
;           ((OR (NOT (NUMBERP (FIRST ROWS-COLS-)))
;                (NOT (> (FIRST ROWS-COLS-) 0)))
;            (PLASM-ERROR 7007 ROWS-COLS-))
;           ((OR (NOT (NUMBERP (SECOND ROWS-COLS-)))
;               (NOT (> (SECOND ROWS-COLS-) 0)))
;            (PLASM-ERROR 7007 ROWS-COLS-))
;           ((OR (NOT (CONSP POLS-))
;                (NOT (EVERY (FUNCTION NOT-VOID-POLYHEDRA-COMPLEX?) POLS-)))
;            (PLASM-ERROR 7008 POLS-))
;           ((OR (NOT (CONSP TIMES-))
;                (NOT (EVERY (FUNCTION NUMBERP) TIMES-)))
;            (PLASM-ERROR 7009 TIMES-))
;           (T (MOVE-MESH ROWS-COLS- POLS- TIMES-))))


(MOVEMESH (ROWS-COLS- MESHES- TIMES-)
          (COND
           ((NOT (ISPAIR ROWS-COLS-)) 
            (PLASM-ERROR 7007 ROWS-COLS-))
           ((OR (NOT (NUMBERP (FIRST ROWS-COLS-)))
                (NOT (> (FIRST ROWS-COLS-) 0)))
            (PLASM-ERROR 7007 ROWS-COLS-))
           ((OR (NOT (NUMBERP (SECOND ROWS-COLS-)))
               (NOT (> (SECOND ROWS-COLS-) 0)))
            (PLASM-ERROR 7007 ROWS-COLS-))
           ((OR (NOT (CONSP MESHES-))
                (NOT (EVERY (FUNCTION MESH-P) MESHES-)))
            (PLASM-ERROR 7008 MESHES-))
           ((OR (NOT (CONSP TIMES-))
                (NOT (EVERY (FUNCTION NUMBERP) TIMES-)))
            (PLASM-ERROR 7009 TIMES-))
           (T (MOVE-MESH ROWS-COLS- MESHES- TIMES-))))





(CAMERA (STATES TIMES)
 (IF (AND (LISTP STATES)
          (LISTP TIMES)
          (= (LENGTH STATES)
             (LENGTH TIMES))
          (EVERY (LAMBDA (X) (AND (NUMBERP 9) (= 9 X)))
                 (MAPCAR (FUNCTION LENGTH)
                         STATES)))
   (CAMERA STATES TIMES)
   (PLASM-ERROR 700 STATES TIMES)))
      
(COLOR (SEQ-) 
       (COND 
        ((EVERY (FUNCTION NUMBERP) SEQ-)
         (CONS "COLOR" SEQ-))
        (T (PLASM-ERROR 7 'COLOR SEQ-))))

(XS (AXES VALUES) (LIST 'XSCALE AXES VALUES))
(XT (AXES VALUES) (LIST 'XTRANS AXES VALUES))
(XR (PLANE VALUES) (LIST 'XROTATE PLANE VALUES))
(XM (MATRIX) (LIST 'XMATRIX MATRIX))
(DIRECTION (VALUES) (LIST DIRECTION VALUES))


;=================================================================
(ISPOL (POLCOMPL-) (POLYHEDRA-COMPLEX? POLCOMPL-))
(ISPOLDIM (DIM-LIST- POLCOMPL-)
	  (AND (POLYHEDRA-COMPLEX? POLCOMPL-)
	       (= (SECOND DIM-LIST-) (GET-NPOL-COMPLEX POLCOMPL-))
	       (= (FIRST DIM-LIST-) (GET-DPOL-COMPLEX POLCOMPL-))))

(DIM (POL-) (GET-DPOL-COMPLEX POL-))
(RN (POL-) (GET-NPOL-COMPLEX POL-))

(EMBED (UP- POL-) (EMBED-POL-COMPLEX POL- UP-))
 
(QUOTE (LISTA) (QUOTE-PLASM LISTA))
(PQUOTE (LISTA) (PQUOTE-PLASM LISTA))

(GRID (LISTA) (GRID-PLASM LISTA))
(SQUARE_HOLE (ARGS) (KERNEL::SQUARE-HOLE ARGS))

(POLYGON (ARGS) (KERNEL::POLYGON ARGS))
;=================================================================

(ISINTPOS (X-) (AND (INTEGERP X-) (> X- 0)) "It returns TRUE if its argument is a positive integer number")

(ISINT (X-) (INTEGERP X-))

(ISREALPOS (X-) (AND (NUMBERP X-) (> X- 0)))
(ISREAL (X-) (NUMBERP X-))
(ISSEQ (X-) (LISTP X-))
(ISNULL (X-) (AND (LISTP X-) (NULL X-)))
(TT (X-) (PROGN X- T))
(ISPAIR (X-) (ISPAIR X-))

(ISFUN (F-) (IF (OR (BOUND-FUNCTION-P F-) (FUNCTIONP F-)) T NIL))
(ISSTRING (STRING--) (STRINGP STRING--))
(ISBOOL (X-) (OR (EQ X- NIL) (EQ X- T)))

(INTMIN (X-) (FLOOR X-))

(INTMAX (X-) (CEILING X-))

(ABS (X-) (ABS X-))
;=================================================================
(MINUS (ARG-) 
       (cond
        ((numberp arg-)
         (FUNCALL #'- ARG-))
        ((and (listp arg-)
              (every (function numberp) arg-))
         (APPLY #'- ARG-))
        ((AND (LISTP ARG-)
             (= 2 (LENGTH ARG-))
             (NOT-VOID-POLYHEDRA-COMPLEX? (FIRST ARG-))
             (NOT-VOID-POLYHEDRA-COMPLEX? (SECOND ARG-)))
         (kernel::_difference arg-))
        (T (plasm-error 7 '- (print-plasm arg-)))))
(SUB (PAIR-)
     (COND ((ISPAIR PAIR-) (APPLY #'- PAIR-))
	   ((NUMBERP PAIR-) (FUNCALL #'- PAIR-))
	   (T (PLASM-ERROR 7 'SUB (print-plasm PAIR-)))))

(ADD (ARG-) 
    (cond
        ((NUMBERP ARG-) ARG-)
        ((every (function numberp) arg-)(APPLY #'+ ARG-))
        ((AND (LISTP ARG-) 
              (= 2 (LENGTH ARG-))
              (NOT-VOID-POLYHEDRA-COMPLEX? (FIRST ARG-))
              (NOT-VOID-POLYHEDRA-COMPLEX? (SECOND ARG-))
         )
         (kernel::_union arg-)
        )
      (T (plasm-error 7 '+ (print-plasm arg-)))
    )
)
(DIV (PAIR-) (IF (ISPAIR PAIR-) 
	       (APPLY #'/ PAIR-)
	       (PLASM-ERROR 7 'DIV PAIR-)))

(MUL (SEQ-) (IF (LISTP SEQ-) (APPLY #'* SEQ-) (PLASM-ERROR 7 'MUL SEQ-)))

;=================================================================
(LT (X1- X2-) (< X2- X1-))
(GT (X1- X2-) (> X2- X1-))
(LE (X1- X2-) (<= X2- X1-))

(GE (X1- X2-) (>= X2- X1-))
(EQ (X-)
    (LET ((LIST- (IF (LISTP X-) X- (PLASM-ERROR 7 'EQ X-))))
      (IF (>= (LENGTH LIST-) 2)
	(IF (REDUCE #'(LAMBDA (X Y) (IF (EQUALP X Y) X)) LIST-) T)
	T)))

(NEQ (X-) (NOT (REDUCE-FUNCTION-ARGS 'EQ X-)))

(AND (SEQ-) (IF (CONSP SEQ-) 
	      (IF (MEMBER NIL SEQ-) NIL T) (PLASM-ERROR 3001 'AND)))
(R_AND (FUNS- ARG-)
       (IF (AND (LISTP FUNS-) (EVERY #'BOUND-FUNCTION-P FUNS-))
	 (DOLIST (I FUNS- T) (IF (NOT (REDUCE-FUNCTION-ARGS I ARG-)) (RETURN NIL)))
	 (PLASM-ERROR 3001 'R_AND)))
(OR (SEQ-)
    (IF (LISTP SEQ-)
      (COND ((NULL SEQ-) NIL) (T (EVAL (CONS 'OR SEQ-))))
      (PLASM-ERROR 3001 'OR)))
(R_OR (FUNS- ARG-)
      (IF (AND (LISTP FUNS-) (EVERY #'BOUND-FUNCTION-P FUNS-))
	(DOLIST (I FUNS- NIL) (IF (REDUCE-FUNCTION-ARGS I ARG-) (RETURN T)))
	(PLASM-ERROR 3001 'R_OR)))

(NOT (ARG-) (NOT ARG-))

;=================================================================

(LEN (X-) (LENGTH X-))

(LIST (X-) (LIST X-))
(FIRST (X-) (FIRST X-))

(LAST (X-) (FIRST (LAST X-)))
(TAIL (X-) (CDR X-))
(TAILR (X-) (BUTLAST X-))

(REVERSE (X-) (REVERSE X-))

(SEL (I* SEQ*) 
     (IF (AND (INTEGERP I*)
	      (> I* 0)
	      (CONSP SEQ*)
              (<= I* (LENGTH SEQ*)))
       (NTH (1- I*) SEQ*)
       (PLASM-ERROR 6 I* SEQ*)))

;(CAT (LISTA-)
; (IF (LISTP LISTA-)
;  (COND
;   ((EVERY #'NOT-VOID-POLYHEDRA-COMPLEX? LISTA-)
;    (REDUCE #'CAT-POL-COMPLEX LISTA- :FROM-END T))
;   ((EVERY #'LISTP LISTA-) (APPLY #'APPEND LISTA-)))
;  (PLASM-ERROR 2000 'CAT LISTA-)))

(CAT (LISTA-)
 (IF (LISTP LISTA-)
  (COND
   ((EVERY #'NOT-VOID-POLYHEDRA-COMPLEX? LISTA-)
    (REDUCE #'CAT-POL-COMPLEX LISTA- :FROM-END T))
   ((EVERY #'LISTP LISTA-) (APPLY #'APPEND LISTA-))
   (T (PLASM-ERROR 1000 LISTA- 'CAT)))
  (PLASM-ERROR 1000 LISTA- 'CAT)))

(AL (SEQ-)
    (IF (AND (ISPAIR SEQ-)
             (LISTP (SECOND SEQ-)))
      (LET ((X (FIRST SEQ-)) (Y-SEQ (SECOND SEQ-))) (CONS X Y-SEQ))
      (PLASM-ERROR 1000  SEQ- 'AL)))

(AR (SEQ-)
    (IF (AND (ISPAIR SEQ-)
             (LISTP (FIRST SEQ-)))
      (LET ((X-SEQ (FIRST SEQ-)) (Y (SECOND SEQ-))) (APPEND X-SEQ (LIST Y)))
      (PLASM-ERROR 1000 SEQ- 'AR)))

(DISTL (SEQ-)
       (IF (AND (ISPAIR SEQ-) (LISTP (SECOND SEQ-)))
	 (LET ((X (FIRST SEQ-)) (Y-SEQ (SECOND SEQ-)) (RES NIL))
	   (DOTIMES (I (LENGTH Y-SEQ) (REVERSE RES))
	     (SETQ RES (CONS (LIST X (FIRST Y-SEQ)) RES))
	     (SETQ Y-SEQ (CDR Y-SEQ))))
	 (PLASM-ERROR 1000 SEQ- 'DISTL)))

(DISTR (SEQ-)
  (IF (AND (ISPAIR SEQ-) (LISTP (FIRST SEQ-)))
    (LET ((X-SEQ (FIRST SEQ-)) (Y (SECOND SEQ-)) (RES NIL))
      (DOTIMES (I (LENGTH X-SEQ) (REVERSE RES))
	(SETQ RES (CONS (LIST (FIRST X-SEQ) Y) RES))
	(SETQ X-SEQ (CDR X-SEQ))))
    (PLASM-ERROR 1000 SEQ- 'DISTR)))
(TRANS (LIST-)
       (IF (AND (CONSP LIST-) (EVERY #'LISTP LIST-) (APPLY #'= (MAPCAR #'LENGTH LIST-)))
	 (EVAL (APPEND '(MAPCAR #'LIST)
		       (MAPCAR #'(LAMBDA (X) (LIST* 'QUOTE (LIST X))) LIST-)))
	 (PLASM-ERROR 6000 LIST-)))
(FROMTO (LISTA-)
	(LET ((S (FIRST LISTA-)) (E (SECOND LISTA-)))
	  (IF (AND (INTEGERP S) (INTEGERP E))
	    (FROMTO2 S E NIL)
	    (PLASM-ERROR 7 'FROMTO (LIST S E)))))
(INTSTO (S)
	(IF (INTEGERP S)
	  (REDUCE-FUNCTION-ARGS 'FROMTO (LIST 1 S))
	  (PLASM-ERROR 7 'INTSTO S)))


;===================================================================

(DOUBLE-DIESIS (N- EXPR-)
	       (IF (AND (INTEGERP N-) (> N- 0) (LISTP EXPR-))
		 (APPLY #'APPEND 
			(MAKE-SEQUENCE 'LIST N- :INITIAL-ELEMENT EXPR-))))

;(DIESIS (N- EXPR-)
;        (IF (AND (INTEGERP N-) (> N- 0))
;            (MAKE-SEQUENCE 'LIST N- :INITIAL-ELEMENT EXPR-)
;            (PLASM-ERROR 1000 N- "#")))


;====================================================================
(ID (X-) X-)

(TO_ROBLAN (OBJ--)
	   (COND ((NOT-VOID-POLYHEDRA-COMPLEX? OBJ--) (PROGN (TO-ROBLAN OBJ-- NIL) OBJ--))
		 ((FUNCTION-WITHOUT-PAR-P OBJ--)
		  (PROGN
		    (TO-ROBLAN (SET-AND-OR-RETURN-VALUE OBJ--)
			       (MAKE-CALLING-TREE-OF-FUN OBJ--))
		    OBJ--))
		 (T (PLASM-ERROR 7 'TO_ROBLAN OBJ--))))
(SYM (PLASM-STRING-) (READ-FROM-STRING PLASM-STRING- 1))
(SIGNAL (MESSAGES-SEQUENCE-)
	(LET* ((DIRETTIVA "")
	       (FORMAT-STRING
		(DOTIMES (I (LENGTH MESSAGES-SEQUENCE-) DIRETTIVA)
		  (SETF DIRETTIVA (CONCATENATE 'STRING DIRETTIVA "~A ")))))
	  (PLASM-ERROR 2345
		       (EVAL (LIST* 'FORMAT
				    (LIST* NIL
					   (LIST* FORMAT-STRING
						  MESSAGES-SEQUENCE-)))))))
(PRINT (ARG) (PROGN (FORMAT T "~%~%~A~%~%" (PRINT-PLASM ARG)) ARG))

(EXPORT (POL-) (if (NOT-VOID-POLYHEDRA-COMPLEX? POL-)
                 (progn (EXPORT-POL-COMPLEX-PLASM POL-)
                        POL-)
                 (PLASM-ERROR 7 'EXPORT POL-)))
(EXPORT_SO (POL-) (if (NOT-VOID-POLYHEDRA-COMPLEX? POL-)
                    (progn (EXPORT-POL-COMPLEX-PLASM POL- nil nil)
                           'no-echo)
                    (PLASM-ERROR 7 'EXPORT POL-)))
(VIEW (POL-) (PROGN (VIEW-POL-COMPLEX POL-)
                    "File last-view.pol contains the output to view."))


(VRML_VER (VER-) (SET-VRML-VER VER-))

(EXPORT_IV (POL-) (if (NOT-VOID-POLYHEDRA-COMPLEX? POL-)
                      (PROGN
                        (EXPORT-IV-FILE POL-)
                        POL-)
                      (PLASM-ERROR 7 'EXPORT_IV POL-)))


(EXPORT_VRML (POL-) (if (NOT-VOID-POLYHEDRA-COMPLEX? POL-)
                      (PROGN
                        (EXPORT-POL-COMPLEX-TO-VRML-FILE-FORMAT-FIRST POL-)
                        POL-)
                      (PLASM-ERROR 7 'EXPORT_VRML POL-)))
(EXPORT_VRML_SO (POL-) (IF (NOT-VOID-POLYHEDRA-COMPLEX? POL-)
                         (PROGN (EXPORT-POL-COMPLEX-TO-VRML-FILE-FORMAT-FIRST POL- nil nil)
                                'no-echo)
                         (PLASM-ERROR 7 'EXPORT_VRML_SO POL-)))
(EXPORT_NRB (POL-) (IF (NOT-VOID-POLYHEDRA-COMPLEX? POL-)
                     (PROGN (EXPORT-POL-COMPLEX-PLASM-NRB POL-)
                            POL-)
                     (PLASM-ERROR 7 'EXPORT_NRB POL-)))
(EXPORT_TO (FILE-NAME- POL-) (progn (EXPORT-POL-COMPLEX-PLASM POL- FILE-NAME-)
                                    POL-))

(EXPORT_TO_ROBLAN (LEVEL- FILE-NAME- POL-)
 (EXPORT-POL-COMPLEX-ROBLAN POL- FILE-NAME- LEVEL-))
(IMPORT (POL-NAME-STRING-)
	(IMPORT-POL-COMPLEX-PLASM (READ-FROM-STRING POL-NAME-STRING-)))

(LAST_VALUE NIL NIL)


;##################################################################
; MATH FUNCTIONS
;##################################################################

(PI NIL (EVAL PI))

(** (X-) (APPLY #'EXPT X- ))

(EXP (X-) (EXP (COERCE X- 'DOUBLE-FLOAT)))

(LN (X-) (LOG (COERCE X- 'DOUBLE-FLOAT)))
(SIGN (N-)
      (IF (NUMBERP N-)
	(SIGNUM N-)
	(PLASM-ERROR 2000 'SIGN N-)))

(SQRT (X-) (IF (>= X- 0) (SQRT (COERCE X- 'DOUBLE-FLOAT)) (PLASM-ERROR 5000 'SQRT X-)))
 
(COS (X-) (COS (COERCE X- 'DOUBLE-FLOAT)))
(ACOS (X-) (ACOS (COERCE X- 'DOUBLE-FLOAT)))
(SIN (X-) (SIN (COERCE X- 'DOUBLE-FLOAT)))
(ASIN (X-) (ASIN (COERCE X- 'DOUBLE-FLOAT)))
(TAN (X-) (TAN (COERCE X- 'DOUBLE-FLOAT)))
(ATAN (Z-) (LET ((Y (FIRST Z-)) (X (SECOND Z-))) (ATAN Y X)))

(INV (MAT-)
     (IF (AND (CONSP MAT-) (EVERY #'CONSP MAT-) (APPLY #'/= (MAPCAR #'LENGTH MAT-)))
       (PLASM-ERROR 3 'INV MAT-)
       (LET* ((MATRIX-DIM (LENGTH MAT-))
	      (MATRIX (MAKE-ARRAY (LIST MATRIX-DIM MATRIX-DIM) :INITIAL-CONTENTS MAT-))
	      (MATRIX-INV (MATLIB:MAT-INV MATRIX)))
	 (BIDIM-ARRAY-TO-LIST-OF-LIST MATRIX-INV))))

;##################################################################
; GEOMETRIC FUNCTIONS
;##################################################################

(JOIN (POL-COMPLEXES-)
      (COND ((EVERY #'NOT-VOID-POLYHEDRA-COMPLEX? POL-COMPLEXES-)
	     (APPLY #'JOIN-POL-COMPLEX POL-COMPLEXES-))
	    ((NOT-VOID-POLYHEDRA-COMPLEX? POL-COMPLEXES-)
	     (JOIN-POL-COMPLEX POL-COMPLEXES-))
	    (T (PLASM-ERROR 600 'JOIN POL-COMPLEXES-))))
(SIMPLEX (DIM-)
	 (IF (AND (INTEGERP DIM-) (> DIM- -1))
	   (BUILD-POL-COMPLEX-SIMPLEX DIM-)
	   (PLASM-ERROR 600 'SIMPLEX DIM-)))
;(CUBOID (LIST--)
;	(IF (AND (LISTP LIST--)
;		 (EVERY #'(LAMBDA (OBJ) (AND (NUMBERP OBJ) (> OBJ 0))) LIST--))
;	  (LET* ((DIM (LENGTH LIST--))
;		 RES
;		 (POL-COMPLEX (BUILD-POL-COMPLEX-CUBE DIM))
;		 (PAR-TO-MULTIPLE-TRANSFORMATION
;		  (DOTIMES (I (LENGTH LIST--) RES)
;		    (SETF RES (APPEND RES (LIST (LIST (1+ I) (NTH I LIST--))))))))
;	    (MAKE-MULTIPLE-AFFINE-TRANSFORMATION
;	     PAR-TO-MULTIPLE-TRANSFORMATION
;	     POL-COMPLEX
;	     'S))
;	  (PLASM-ERROR 601 'CUBOID LIST--)))


(CUBOID (LIST--)
	(IF (AND (LISTP LIST--)
		 (EVERY #'(LAMBDA (OBJ) (AND (NUMBERP OBJ) (> OBJ 0))) LIST--))
	  (LET* ((DIM (LENGTH LIST--))
		 RES
		 (THE-CUBE (BUILD-CUBE DIM))
                 (THE-COMPLEX (MAKE-EMPTY-POLYHEDRA-COMPLEX))
		 (COORD-VALUE-PAIRS
		  (DOTIMES (I (LENGTH LIST--) RES)
		    (SETF RES (APPEND RES (LIST (LIST (1+ I) (NTH I LIST--))))))))
	    (PUT-NODE-IN-POL-COMPLEX (MAKE-POL-NODE THE-CUBE
                                                    (BUILD-SCALE-MATRIX 
                                                     COORD-VALUE-PAIRS
                                                     'ON-FACES
                                                     (1+ DIM)))
	                             THE-COMPLEX)
            THE-COMPLEX)
	  (PLASM-ERROR 601 'CUBOID LIST--)))


(OCUBOID (borders-- dimension2--)
	(IF (AND (LISTP borders--)
		 (EVERY #'(lambda (obj) (and (listp obj) (= (length obj) 2)
					     (EVERY #'(LAMBDA (OBJ)(or (eq 1 OBJ)(eq OBJ 0)))
						    obj)))
			borders--)
		 (LISTP dimension2--)
		 (EVERY #'(LAMBDA (OBJ) (AND (NUMBERP OBJ) (> OBJ 0))) dimension2--))
	  (LET* ((DIM (LENGTH borders--))
		 RES
		 (POL-COMPLEX (COERCE-POL-COMPLEX 
			       (BUILD-POL-COMPLEX-CUBE DIM borders--) 'ON-VERTICES))
		 (PAR-TO-MULTIPLE-TRANSFORMATION
		  (DOTIMES (I (LENGTH borders--) RES)
		    (SETF RES (APPEND RES (LIST (LIST (1+ I) (NTH I dimension2--))))))))
	    (MAKE-MULTIPLE-AFFINE-TRANSFORMATION
	     PAR-TO-MULTIPLE-TRANSFORMATION
	     POL-COMPLEX
	     'S))
	  (PLASM-ERROR 601 'CUBOID (list borders-- dimension2--))))
(MKPOL (SEQ-)
       (let ((seq- (IO::RIPULISCI-MAKE-POL seq-)))
	(BUILD-POL-COMPLEX
	 (FIRST SEQ-)
	 (MAPCAR #'(LAMBDA (SEQ) (MAPCAR #'1- SEQ)) (SECOND SEQ-))
	 (MAPCAR #'(LAMBDA (SEQ) (MAPCAR #'1- SEQ)) (THIRD SEQ-))
	 'ON-VERTICES)))
(UKPOL (POL-)
       (IF (NOT-VOID-POLYHEDRA-COMPLEX? POL-)
	 (LET ((VERT-CELL-POL (UNMAKE-POL-COMPLEX POL-)))
	   (LIST (FIRST VERT-CELL-POL)
		 (MAPCAR #'(LAMBDA (LISTA) (MAPCAR #'(LAMBDA (INDEX) (1+ INDEX)) LISTA))
			 (SECOND VERT-CELL-POL))
		 (MAPCAR #'(LAMBDA (LISTA) (MAPCAR #'(LAMBDA (INDEX) (1+ INDEX)) LISTA))
			 (THIRD VERT-CELL-POL))))
	 (PLASM-ERROR 600 'UKPOL POL-)))

(MAP (FUN- POL-) (MAP-ON-POL-COMPLEX FUN- POL-))

(FIT (POL-) (FIT-POL-COMPLEX POL-))
(&& (SPEC- POLS-SEQ-)
    (LET* ((SPEC-- (IF (EVERY #'CONSP SPEC-) SPEC- (LIST SPEC-)))
	   (ARG-FOR-INT-EXTR (LIST (FIRST POLS-SEQ-) (SECOND POLS-SEQ-) SPEC--)))
      (APPLY #'INTERSECTION-OF-EXTRUSIONS-ON-COMPLEXES ARG-FOR-INT-EXTR)))

(POWER (ARG-SEQ-)
       (COND ((AND (CONSP ARG-SEQ-) (EVERY #'NUMBERP ARG-SEQ-)) (APPLY #'* ARG-SEQ-))
	     ((AND (CONSP ARG-SEQ-)
		   (EVERY #'POLYHEDRA-COMPLEX? ARG-SEQ-)
		   (= 2 (LENGTH ARG-SEQ-)))
	      (LET ((POL1 (FIRST ARG-SEQ-)) (POL2 (SECOND ARG-SEQ-)))
		(POWER POL1 POL2)))
	     ((AND (AND (LISTP (FIRST ARG-SEQ-))
			(EVERY #'(LAMBDA (X) (AND (LISTP X) (EVERY #'NUMBERP X)))
			       (FIRST ARG-SEQ-)))
		   (AND (LISTP (SECOND ARG-SEQ-))
			(EVERY #'(LAMBDA (X) (AND (LISTP X) (EVERY #'NUMBERP X)))
			       (SECOND ARG-SEQ-))))
	      (LET ((MAT1
		     (MAKE-ARRAY (LIST (LENGTH (FIRST ARG-SEQ-))
				       (LENGTH (FIRST (FIRST ARG-SEQ-))))
				 :INITIAL-CONTENTS
				 (FIRST ARG-SEQ-)))
		    (MAT2
		     (MAKE-ARRAY (LIST (LENGTH (SECOND ARG-SEQ-))
				       (LENGTH (FIRST (SECOND ARG-SEQ-))))
				 :INITIAL-CONTENTS
				 (SECOND ARG-SEQ-))))
		(ARRAY->LIST (MATLIB:MAT-PRODUCT MAT1 MAT2))))
	     (T (PLASM-ERROR 600 'POWER ARG-SEQ-))))
(SKELETON (N- POLCOMPLEX-)
	  (IF (AND (INTEGERP N-) (> N- -1) (POLYHEDRA-COMPLEX? POLCOMPLEX-))
	    (SKELETON POLCOMPLEX- N-)))

(BOUNDARY (POL-COMPLEX-) (PLASM-ERROR 3000 "Sorry: no implemented operator."))

(MIN (COORD-LIST- POL-)
     (LET ((COORD (IF (LISTP COORD-LIST-) COORD-LIST- (LIST COORD-LIST-))))
       (COND ((AND (EVERY #'INTEGERP COORD) (POLYHEDRA-COMPLEX? POL-))
	      (LET ((MINS- (MINS COORD POL-)))
		(IF (= 1 (LENGTH MINS-)) (NTH 0 MINS-) MINS-)))
	     (T
	      (PLASM-ERROR 3000
			   (FORMAT NIL
				   "~A and ~A are not valid arguments to MIN."
				   COORD
				   POL-))))))

(MAX (COORD-LIST- POL-)
     (LET ((COORD (IF (LISTP COORD-LIST-) COORD-LIST- (LIST COORD-LIST-))))
       (COND ((AND (EVERY #'INTEGERP COORD) (POLYHEDRA-COMPLEX? POL-))
	      (LET ((MAXS- (MAXS COORD POL-)))
		(IF (= 1 (LENGTH MAXS-)) (NTH 0 MAXS-) MAXS-)))
	     (T
	      (PLASM-ERROR 3000
			   (FORMAT NIL
				   "~A and ~A are not valid arguments to MAX."
				   COORD
				   POL-))))))

(SIZE (COORD-LIST- POL-)
      (LET ((COORD (IF (LISTP COORD-LIST-) COORD-LIST- (LIST COORD-LIST-))))
	(COND ((AND (EVERY #'INTEGERP COORD) (POLYHEDRA-COMPLEX? POL-))
	       (LET ((SIZES
		      (APPLY #'(LAMBDA (MINS MAXS) (MAPCAR #'- MAXS MINS))
			     (MINS-MAXS COORD POL-))))
		 (IF (= 1 (LENGTH SIZES)) (NTH 0 SIZES) SIZES)))
	      (T
	       (PLASM-ERROR 3000
			    (FORMAT NIL
				    "~A and ~A are not valid arguments to SIZE."
				    COORD
				    POL-))))))

(MED (COORD-LIST- POL-)
     (LET ((COORD (IF (LISTP COORD-LIST-) COORD-LIST- (LIST COORD-LIST-))))
       (COND ((AND (EVERY #'INTEGERP COORD) (POLYHEDRA-COMPLEX? POL-))
	      (LET ((MEDS
		     (APPLY #'(LAMBDA (MINS MAXS)
				(MAPCAR #'(LAMBDA (X Y) (/ (+ X Y) 2)) MAXS MINS))
			    (MINS-MAXS COORD POL-))))
		(IF (= 1 (LENGTH MEDS)) (NTH 0 MEDS) MEDS)))
	     (T
	      (PLASM-ERROR 3000
			   (FORMAT NIL
				   "~A and ~A are not valid arguments to MED."
				   COORD
				   POL-))))))

;##################################################################
; BOOL FUNCTIONS
;##################################################################
(UNION (POLS-PAIR-)
    (IF (AND (LISTP POLS-PAIR-)
             (= 2 (LENGTH POLS-PAIR-))
             (NOT-VOID-POLYHEDRA-COMPLEX? (FIRST POLS-PAIR-))
             (NOT-VOID-POLYHEDRA-COMPLEX? (SECOND POLS-PAIR-)))
      (kernel::_union POLS-PAIR-)
      (PLASM-ERROR 600 'UNION (print-plasm POLS-PAIR-))))
(INTERSECTION (POLS-PAIR-)
    (IF (AND (LISTP POLS-PAIR-)
             (= 2 (LENGTH POLS-PAIR-))
             (NOT-VOID-POLYHEDRA-COMPLEX? (FIRST POLS-PAIR-))
             (NOT-VOID-POLYHEDRA-COMPLEX? (SECOND POLS-PAIR-)))
      (kernel::_intersection POLS-PAIR-)
      (PLASM-ERROR 600 'INTERSECTION (print-plasm POLS-PAIR-))))

(DIFFERENCE (POLS-PAIR-)
    (IF (AND (LISTP POLS-PAIR-)
             (= 2 (LENGTH POLS-PAIR-))
             (NOT-VOID-POLYHEDRA-COMPLEX? (FIRST POLS-PAIR-))
             (NOT-VOID-POLYHEDRA-COMPLEX? (SECOND POLS-PAIR-)))
      (kernel::_difference POLS-PAIR-)
      (PLASM-ERROR 600 'DIFFERENCE (print-plasm POLS-PAIR-))))
(SDIFFERENCE (POLS-PAIR-)
    (IF (AND (LISTP POLS-PAIR-)
             (= 2 (LENGTH POLS-PAIR-))
             (NOT-VOID-POLYHEDRA-COMPLEX? (FIRST POLS-PAIR-))
             (NOT-VOID-POLYHEDRA-COMPLEX? (SECOND POLS-PAIR-)))
      (kernel::_sdifference POLS-PAIR-)
      (PLASM-ERROR 600 'XOR (print-plasm POLS-PAIR-))))
(XOR (POLS-PAIR-)
    (IF (AND (LISTP POLS-PAIR-)
             (= 2 (LENGTH POLS-PAIR-))
             (NOT-VOID-POLYHEDRA-COMPLEX? (FIRST POLS-PAIR-))
             (NOT-VOID-POLYHEDRA-COMPLEX? (SECOND POLS-PAIR-)))
      (kernel::_sdifference POLS-PAIR-)
      (PLASM-ERROR 600 'XOR (print-plasm POLS-PAIR-))))

(& (POLS-PAIR-)
    (IF (AND (LISTP POLS-PAIR-)
             (= 2 (LENGTH POLS-PAIR-))
             (NOT-VOID-POLYHEDRA-COMPLEX? (FIRST POLS-PAIR-))
             (NOT-VOID-POLYHEDRA-COMPLEX? (SECOND POLS-PAIR-)))
      (kernel::_intersection POLS-PAIR-)
      (PLASM-ERROR 600 'INTERSECTION (print-plasm POLS-PAIR-))))

;(| (POLS-PAIR-)
;    (IF (AND (LISTP POLS-PAIR-)
;             (= 2 (LENGTH POLS-PAIR-))
;            (NOT-VOID-POLYHEDRA-COMPLEX? (FIRST POLS-PAIR-))
;             (NOT-VOID-POLYHEDRA-COMPLEX? (SECOND POLS-PAIR-)))
;      (kernel::_union POLS-PAIR-)
;      (PLASM-ERROR 600 'UNION (print-plasm POLS-PAIR-))))

(^ (POLS-PAIR-)
    (IF (AND (LISTP POLS-PAIR-)
             (= 2 (LENGTH POLS-PAIR-))
             (NOT-VOID-POLYHEDRA-COMPLEX? (FIRST POLS-PAIR-))
             (NOT-VOID-POLYHEDRA-COMPLEX? (SECOND POLS-PAIR-)))
      (kernel::_sdifference POLS-PAIR-)
      (PLASM-ERROR 600 'XOR (print-plasm POLS-PAIR-))))

;##################################################################
; APPEARANCE FUNCTIONS
;##################################################################


(ASSIGNAPPEARANCE (ARG-) 
           (if (and (io::texture? (third ARG-)) (io::color? (second ARG-)) 
                    (not-void-polyhedra-complex? (first ARG-)))
		  (IO::SET-POL-APPEARANCE (FIRST ARG-) (SECOND ARG-) (third ARG-))
	      (PLASM-ERROR 600 'ASSIGNAPPEARANCE (print-plasm ARG-)))
"ASSIGNAPPEARANCE:<polyhedral_complex,color,texture> --> polyhedral_complex
     Where color is <<diffuse_R,diffuse_G,diffuse_B>,
                     <specular_R,specular_G,specular_B>,
                     <ambient_intensity>,
                     <emissive_R,emissive_G,emissive_B>,
                     <shininess>,<transparency>>
       and texture is <url,
                      <repeatS, repeatT>,
                      <translationS, translationT>,
                      <rotation>, <scaleS, scaleT>,
                      <centerS, centerT>>
       and url is a string representing a local or www filename.
"
)

(ASSIGNCOLOR (ARG-)
           (if (and (io::color? (second ARG-)) (not-void-polyhedra-complex? (first ARG-)))
		(IO::SET-POL-APPEARANCE  (FIRST ARG-) (SECOND ARG-) nil)
	      (PLASM-ERROR 600 'ASSIGNCOLOR (print-plasm ARG-)))
"ASSIGNCOLOR:<polyhedral_complex,color> --> polyhedral_complex
     Where color is <<diffuse_R,diffuse_G,diffuse_B>,
                     <specular_R,specular_G,specular_B>,
                     <ambient_intensity>,
                     <emissive_R,emissive_G,emissive_B>,
                     <shininess>,
                     <transparency>>"           
           )

(ASSIGNTEXTURE (ARG-) 
	       (if (and (io::texture? (second ARG-)) (not-void-polyhedra-complex? (first ARG-)))
	      (IO::SET-POL-APPEARANCE  (FIRST ARG-) nil (SECOND ARG-))
	      (PLASM-ERROR 600 'ASSIGNTEXTURE (print-plasm ARG-)))
"ASSIGNTEXTURE:<polyhedral_complex,texture> --> polyhedral_complex
     Where texture is <url,
                      <repeatS, repeatT>,
                        <translationS, translationT>,
                        <rotation>,
                        <scaleS, scaleT>,
                        <centerS, centerT>>
       and url is a string representing a local or www filename."
)



(ASSIGNSPOTLIGHT (ARG-)
		 (if (and (polyhedra-complex? (first ARG-)) (io::light? (second arg-)))	 
		     (IO::SET-POL-LIGHT (FIRST ARG-) 
					(LIST 0
					      (FIRST (SECOND ARG-)) 
					      (SECOND (SECOND ARG-)) ))
		   (PLASM-ERROR 600 'ASSIGNSPOTLIGHT (print-plasm ARG-)))
"ASSIGNSPOTLIGHT:<pol,<appearance,geometry> --> polyhedral_complex"
)
		   

(ASSIGNPOINTLIGHT (ARG-)
		 (if (and (not-void-polyhedra-complex? (first ARG-)) (io::light? (second arg-)))	
	(IO::SET-POL-LIGHT (FIRST ARG-) 
                           (LIST  1
                                 (FIRST (SECOND ARG-)) 
                                 (SECOND (SECOND ARG-)) ))
		   (PLASM-ERROR 600 'ASSIGNPOINTLIGHT (print-plasm ARG-)))
"ASSIGNPOINTLIGHT:<pol,<appearance,geometry> --> polyhedral_complex"
)

(ASSIGNDIRLIGHT (ARG-)
		 (if (and (not-void-polyhedra-complex? (first ARG-)) (io::light? (second arg-)))
	(IO::SET-POL-LIGHT (FIRST ARG-) 
                           (LIST 2
                                 (FIRST (SECOND ARG-)) 
                                 (SECOND (SECOND ARG-)) )))
		   (PLASM-ERROR 600 'ASSIGNPOINTLIGHT (print-plasm ARG-))
"ASSIGNDIRLIGHT:<pol,<appearance,geometry> --> polyhedral_complex
     Where appereance is <color,intensity,ambientIntensity,on>
         and geometry is <location,direction,attenuation,
                           radius,beamWidth,cutOffAngle>"
)

(ASSIGNCAMERA (ARG-)
              (if (and (not-void-polyhedra-complex? (first ARG-))
                       (io::camera? (second arg-)))
                  (IO::SET-POL-CAMERA (FIRST ARG-)
                                      (LIST 0 (SECOND ARG-))) 
                (PLASM-ERROR 600 'ASSIGNCAMERA (print-plasm ARG-)))
"ASSIGNCAMERA:<pol_complex,camera> --> pol_complex"
)

(ASSIGNORTOCAMERA (ARG-)
                  (if (and (not-void-polyhedra-complex? (first ARG-))
                           (io::camera? (second arg-)))
                      (IO::SET-POL-CAMERA (FIRST ARG-)
                                          (LIST 1 (SECOND ARG-))) 
                    (PLASM-ERROR 600 'ASSIGNCAMERA (print-plasm ARG-)))
"ASSIGNORTOCAMERA:<pol_complex,camera> --> pol_complex
     Where (VRML-2) camera is <position,orientation,fieldOfView,(unused),description>
     Where (VRML-1) camera is <position,orientation,height,focal-distance,(unused)>"
)

;##################################################################
;##################################################################


)))


(defun FROMTO2 (s e res)
  (if (> s e)
    res
    (fromto2 s (1- e) (cons e res))))


(defun ISPAIR (obj)
  (and (consp obj)
       (first obj)
       (second obj)
       (not (third obj))
       T))

(defun ISPAIR (obj)
  (and (consp obj)
       (= 2 (length obj))))



(defun INSL-LISP (fun list-)
  (cond
   ((not (consp list-))
    (plasm-error 1000 list- "INSR:any-PLASM-function"))
   (T (reduce 
       #'(lambda (x y) (funcall 
		      (if (functionp fun)
			fun 
			(set-and-or-return-value fun t)) 
		      (list x y)))
       list-
       :from-end nil))))


(defun INSR-LISP (fun list-)
  (cond
   ((not (consp list-))
    (plasm-error 1000 list- "INSL:any-PLASM-function"))
   (T (reduce 
       #'(lambda (x y) (funcall 
			(if (functionp fun)
			  fun 
			  (set-and-or-return-value  fun t)) 
			(list x y)))
       list-
       :from-end T))))


(defun TREE-LISP (fun list-)
  (tree-lisp1 (if (functionp fun)
		fun 
		(set-and-or-return-value  fun t))
	      (if (consp list-)
		list-
		(plasm-error 1000 list- "TREE:any-PLASM-function"))))


(defun TREE-LISP1 (fun list-)
  (cond
   ((not (second list-)) (first list-))
   (T (let ((k (truncate (/ (length list-) 2))))
	(funcall
	 fun
	 (list (tree-lisp1 fun (subseq list- 0 k))
	       (tree-lisp1 fun (subseq list- k))))))))


(defun MERGE-LISP (fun list-)
  (merge-lisp1
   (if (functionp fun)
     fun 
     (set-and-or-return-value  fun t))
   list-))

(defun MERGE-LISP1 (fun- list-)
  (cond 
   ((null (first list-)) (second list-))
   ((null (second list-)) (first list-))
   ((funcall
     fun-
     (list (first (first list-))
	   (first (second list-))))
    (cons
     (first (first list-)) 
     (merge-lisp1
      fun-
      (list 
       (cdr (first list-))
       (second list-)))))
   (T
    (cons
     (first (second list-))
     (merge-lisp
      fun-
      (list 
       (first list-)
       (cdr (second list-))))))))

		

(defun MY-EQUAL (objects)
	(if (every #'stringp objects)
   (apply #'string= objects)
   (apply #'equal objects)))

(defun QUOTE-PLASM (sequence-sizes)
  (let ((unit-cube (coerce-pol-complex (build-cube 1) 'on-vertices))
	(trans-par 0)
	(pol-complex nil))
    (dolist (size sequence-sizes (reverse (coerce-pol-complex pol-complex 'on-faces)))
      (if (> size 0)
	(put-node-in-pol-complex-as-first
	 (make-pol-node unit-cube
			(make-array (list 2 2)
				    :initial-contents
				    (list (list 1 0)
					  (list trans-par  size))))
	 pol-complex))
      (setf trans-par (+ trans-par (abs size))))))

(defun GRID-PLASM (sequence)
  (let ((vprec 0)
	(vlist '((0)))
	 clist
	 plist)
    (dotimes (i (length sequence))
      
(print vprec)      
      
      (if (>= (nth i sequence) 0)
	(progn
	  (setq vlist (append vlist (list (list (+ vprec (nth i sequence)))))) 
	  (setq clist (append clist (list (list (1- (1- (length vlist))) 
						(1- (length vlist))))))
	  (setq plist (append plist (list (1- (length clist)))))
	  (setq vprec (+ vprec (nth i sequence))))
	(progn
	  (setq vlist (append vlist (list (list (+ vprec (abs (nth i sequence)))))))
	  (setq vprec (+ vprec (abs (nth i sequence)))))))
  (kernel::build-pol-complex vlist clist (list plist) 'ON-VERTICES)
  ))
	  
	

(defun PQUOTE-PLASM (sequence)
  (let ((vprec 0)
	(vlist '((0)))
	 clist
	 plist)
    (dotimes (i (length sequence))
      
      
      (if (>= (nth i sequence) 0)
	(progn
	  (setq vlist (append vlist (list (list (+ vprec (nth i sequence)))))) 
	  (setq clist (append clist (list (list (1- (1- (length vlist))) 
						(1- (length vlist))))))
	  (setq plist (append plist (list (1- (length clist)))))
	  (setq vprec (+ vprec (nth i sequence))))
	(progn
	  (setq vlist (append vlist (list (list (+ vprec (abs (nth i sequence)))))))
	  (setq vprec (+ vprec (abs (nth i sequence)))))))
  (kernel::build-pol-complex vlist clist (list plist) 'ON-VERTICES)
  ))
	  
	

;----------------------------------------------------------------------------------------------
; ATTENZIONE: QUESTA FUNZIONE RINOMINA UNA FUNZIONE MA DEVE ESSERE USATA CON GRANDE ACCORTEZZA.
; UN SUO USO SPREGIUDICATO PUO' ESSERE FATTO SOLO SE VIENE APPLICATA A FUNZIONI CHE NON VENGONO
; CHIAMATE DA NESSUNA ALTRA FUNZIONE! 

(defun RENAME-A-USER-PLASM-FUNCTION  (old new) 
  (declare (special **USER-FENV**))
  (let ((fun (bound-function old)))
    (remove-from-user-fenv  old)
    (setf (plasm-function-name fun) new)
    (insert-in-user-fenv fun)))
;----------------------------------------------------------------------------------------------

(format t "~%~%~%~%Plasm is growing on Plasm ... ~%~%~%")

; attiva la possibilita' di usare identificatori extra: si puo' usare il carattere $

     ;--------------------
     ;--------------------
      (provide-extra-id)
     ;--------------------
     ;--------------------

(mapcar #'eval-plasm
	'("def as (fun::isfun) = (comp~[k:cons,aa]):fun"
	  
	  "def ac (fun::isfun) = (comp ~ [K:comp,aa]):fun"              
	  
	  "def isseqof (predicate::IsFun)  (arg::tt) = 
if:<and~[isseq,not~IsNull],
    and ~ aa:predicate,
    k:false>:arg"
 
	  "def raise (op::IsFun) (arg::tt) = if:<isseqof:isfun,lift:op,op>:arg"

	  "def raised_plus = raise:add"

	  "def raised_sub = raise:sub"
	  "def raised_power = raise:power"
	  
	  "def raised_div = raise:div"
	  
	  "def diesis (n::IsIntPos) (obj::tt) = (aa:s2 ~ distr) :<intsto:n,  obj>"

;"Def Quote (x :: isSeq) =
;
;   (STRUCT ~ CAT ~
;
;     AA:(IF:<isRealPos , [ STRUCT ~ [S:1,(K~SIMPLEX):1] , T:1 ] , [ T:1 ~ - ] >)):x"


"DEF LOCATE (pol::IsPol; i::IsIntPos; distances::IsSeqOf:IsReal)  =
 (STRUCT ~ CAT ~ DISTR):<AA:(T:i):distances, pol>"

"DEF BOX (index_s::TT) (pol::IsPol) =
   
    STRUCT:<translation:mins, CUBOID:sizes>
 
  WHERE 
   mins = MIN:index_s:pol ,
   sizes = SIZE:index_s:pol, 
   translation = t:(if:<isseq,intsto~len,k:1>:index_s)
  END"

"DEF ALIGN (inf::TT) (pol1,pol2::IsPol) = (STRUCT ~ fun):<pol1,pol2>
  WHERE 
   fun = IF:<IsSeqOf:IsSeq,COMP~AA:align$,align$>:inf 
  END"

"DEF align$ (index,pos1,pos2::TT) (pol1,pol2::IsPol) =
  
    <pol1, STRUCT:<T:index:trans_val, pol2>>

  WHERE 
   trans_val = SUB:<APPLY:<pos1,index>:pol1 ,APPLY:<pos2,index>:pol2> 
  END"

"DEF TOP (Pol1, Pol2 ::IsPol) = 
   ALIGN:<<3,MAX,MIN>,<1,MED,MED>,<2,MED,MED>>:<Pol1,Pol2> "

"DEF BOTTOM (Pol1, Pol2 ::IsPol) = 
  
 ALIGN:<<3,MIN,MAX>,<1,MED,MED>,<2,MED,MED>>:<Pol1,Pol2>"


"DEF LEFT (Pol1, Pol2 ::IsPol) = 
    ALIGN:<<1,MIN,MAX>,<3,MIN,MIN>>:<Pol1,Pol2>"
"DEF RIGHT (Pol1, Pol2 ::IsPol) = 
   ALIGN:<<1,MAX,MIN>,<3,MIN,MIN>>:<Pol1,Pol2> "

"DEF UP (Pol1, Pol2 ::IsPol) =

   ALIGN:<<2,MAX,MIN>,<3,MIN,MIN>>:<Pol1,Pol2> "

"DEF DOWN (Pol1, Pol2 ::IsPol) = 
 ALIGN:<<2,MIN,MAX>,<3,MIN,MIN>>:<Pol1,Pol2>"

"DEF split_2pi$ (i::IsIntPos) = 
  (AA:* ~ DISTR):<FROMTO:<0,i- 1>, 2 * PI/i>"

"DEF cylinder (radius,height::IsRealPos; steps::IsIntPos) =

     circle * quote:<height>    

 WHERE
  
  map = AA:[K:radius * COS, K:radius * SIN ] ,

  interval = split_2pi$:steps ,
  indexs_cell = INTSTO:steps,

  circle = mkpol:<map:interval,
		      <indexs_cell>,
		      <<1>>>

END"

"DEF subseq$ (i_j::(and ~ [and ~aa:isint, ispair]); seq::isseq) =
  (cons~aa:sel~..):i_j:seq"

"DEF subseq = c:subseq$"

	  
"Def North  = [[Max:1,Max:2],[Min:1,Max:2]] "
     
"Def South  = [[Min:1,Min:2],[Max:1,Min:2]] "

"Def West   = [[Min:1,Max:2],[Min:1,Min:2]] "
"Def East    = [[Max:1,Min:2],[Max:1,Max:2]]"

"Def pipe (f,g::IsFun) =  struct ~ [s1, trans_function, s2]

 where

    vect_diff = aa:- ~ trans,
    vect_sum  = aa:+ ~ trans,
    vect_perp  = [s2, - ~ s1],
    norm = aa:sign,
 
    points_a = f ~ s1 ,

    points_b = g ~ s2 ,
    a1 = s1 ~ points_a, 
    a2 = s2 ~ points_a,
   
    b1 = s1 ~ points_b, 
    b2 = s2 ~ points_b,
    
    va = vect_diff ~ [a2,a1],
    va_perp = norm ~ vect_perp ~ va,
    a3 = vect_sum ~ [a1 , va_perp],
 
    vb = vect_diff ~ [b1,b2],
    vb_perp = norm ~ vect_perp ~ vb,
    b3 =  vect_sum ~ [b2 , vb_perp],

    m1  = al ~ [k:<1,1,1>, id] ~ trans ~ [a1, a2, a3] ,
    m2  = al ~ [k:<1,1,1>, id] ~ trans ~ [b2, b1, b3] ,
    
    trans_matrix  = * ~ [m1, inv ~ m2] ,
    
    trans_function = mat ~ trans_matrix 
end"

"DEF ! = pipe"
	  
"DEF CAT_POL = CAT~AL~[K:<>, ID]"

; funzioni di utilita' per i materiali


"DEF TEST_RGBCOLOR_$ (seq::TT) = (AND~[IsSeq,EQ~[LEN,k:3]]):seq"

"DEF RGBCOLOR(colour::TEST_RGBCOLOR_$)=<colour,colour,<0.2>,<0,0,0>,<0.2>,<0>>;"

"SET_FUNCTION_DESC:RGBCOLOR:
'
RGBCOLOR:<Red,Green,Blue> --> color
'"


"DEF MKMESH (APPEARANCE::TT) (GRID::ISSEQOF:(ISSEQOF:(ISSEQOF:ISREAL)) ) = 
   MKMESH$:APPEARANCE:GRID"



"DEF SIMPLETEXTURE(name::IsString)=<name,<0,0>,<0,0>,<0>,<1,1>,<0,0>>"
"SET_FUNCTION_DESC:SIMPLETEXTURE:'SIMPLETEXTURE:url --> texture'"




))
; disattiva la possibilita' di usare identificatori extra

     ;--------------------
     ;--------------------
      (inpackage plasm)  ;eval-plasm function sets *package* to user package after every run!
      (do-not-provide-extra-id)
     ;--------------------
     ;--------------------

      

;-----------------------------------------------------------------
; la forma seguente opera un cambiamento di nome alle funzioni primitive plasm definite
; in plasm per le quali al momento della loro definizione non e' stato possibile usare
; il nome desiderato

(mapcar #'rename-a-user-plasm-function 
	'(raised_plus raised_sub raised_power raised_div)
	'(+ - * /))
;----------------------------------------------------------

(mapcar #'(lambda (name)
	    (insert-in-default-fenv (bound-function name)))
	(let (listname)
	  (declare (special **USER-FENV**))
	  (maphash #'(lambda (name d) 
		       (declare (ignore d))
		       (setf listname (cons name listname)))
		   **USER-FENV**)
	  listname))
	

(clrhash  **USER-FENV**)
(RESET-*SET-WHEN-YOU-CAN*) ; do not record calling to anything 



