; -*- package: plasm -*-
;---------------------------------------------------------------------------
;                    INOUT.LISP FILE
;---------------------------------------------------------------------------

(provide 'inout)
(inpackage plasm)

(export '(read-plasm-script))

(defun SHOW-DEFINITION-VALUE (arg) 
  (if arg
    (format t "~%===================~%~
	       Function ~A defined.~%===================~%" arg))
  (values))

(defun SHOW-EXPRESSION-VALUE (arg)
  (declare (special user::LAST_VALUE))
  (setq user::LAST_VALUE arg)
  (let ((fun (bound-function 'last_value)))
    (setf (plasm-function-value fun)
	  arg))
  (format t "~%~%")
  (format t "~A" (print-plasm arg))
  (format t "~%~%")
  (values))


(defun PRINT-PLASM (arg)
  (with-output-to-string   (stringa)
    (cond
     ((eq arg 'no-echo)
      nil)
     ((geometry-motion-pair arg)
      (format stringa "A pair geometry/motion data."))

     ((camera-p arg)
      (format stringa "A camera data set."))

     ((frames-p arg)
      (format stringa "A frame data set."))

     ((RGBCOLOR-p arg)
      (format stringa "A color."))

     ((material-p arg)
      (format stringa "A material."))

     ((texture-p arg)
      (format stringa "A texture."))

     ((appearance-p arg)
      (format stringa "An appearance."))

     ((mesh-p arg)
      (format stringa "A mesh data set."))

     ((quad-mesh-p arg)
      (format stringa "An animated mesh data set."))

     ((not-void-polyhedra-complex? arg)
      (format stringa "A-Polyhedral-Complex{~D,~D}"
	      (get-dpol-complex arg) (get-npol-complex arg) ))

     ((consp arg)
      (format stringa "~A" '<)
      (format stringa "~A"
                (eval (append
                    '(concatenate 'string)
                            (mapcar #'PRINT-PLASM (put-comma arg)))))
      (format stringa "~A" '>))


     ((null arg) (format stringa "<>"))

     ((functionp arg)
      (format stringa "~A " "An-Anonymous-Function"))
     ((eq #\, arg) (format stringa ", "))

     ((eq T arg) (format stringa "True "))


     ((stringp arg) (format stringa "'~A'" arg))

     ((atom arg) (format stringa "~A" arg))
     (T (format stringa "~A" arg)))))


(defun GEOMETRY-MOTION-PAIR (arg)
  (and (consp arg)
           (= 3 (length arg))
           (pathnamep (first arg))
           (stringp  (second arg))
           (stringp (third arg))
        ;   (string-equal "#Inventor V2.1 ascii" (first arg) :end2 20)
           (string-equal "#!plasm v0.1 ascii" (second arg) :end2 18)))


(defun MESH-P (arg)
  (and (ispair arg)
       (appearance-p (first arg))
       (consp (second arg)))) ; (second arg) is a sequence of sequence of reals




(defun FRAMES-P (arg)
  (and (consp arg)
       (= 3 (length arg))
       (pathnamep (first arg))
       (numberp  (second arg))
       (numberp (third arg))))


(defvar *CAMERA-TAG* "camera")

(defun CAMERA-P (arg)
  (declare (special *CAMERA-TAG*))
  (and (consp arg)
       (= 3 (length arg))
       (and (stringp (first arg))
            (string= (first arg) *CAMERA-TAG*))
       (every (lambda (x)
               (and (consp x)
                    (every (function numberp)
                           x)))
              (second arg))
       (every (function numberp) (third arg))))


;(defun PRINT-PLASM (arg)
;  (cond
;   ((not-void-polyhedra-complex? arg)
;    (format T "A-Polyhedral-Complex{~D,~D}"
;	    (get-dpol-complex arg) (get-npol-complex arg)))
;  
;   ((consp arg)
;    (format T "~A" '<)
;    (print-plasm (first arg))
;    (dolist (item (rest arg) (values))
;	    (format T "~A " #\,)
;	    (print-plasm item))
;    (format T "~A" '>))
;   
;   
;   ((null arg) (format T "<>"))
;   
;   ((functionp arg)
;    (format T "~A " "An-Anonymous-Function"))
;   
;   ((eq T arg) (format T "True "))
;   
;   ((stringp arg) (format T "'~A'" arg))
;   
;   ((atom arg) (format T "~A" arg))
;  
;   (T (format T "~A" arg)))
;  (values))

(defun PUT-COMMA (arg)
  (if (second arg)
    (cons (first arg)
	  (put-comma1 (rest arg)))
    arg))

(defun PUT-COMMA (arg)
  (if (> (length arg) 1)
    (cons (first arg)
	  (put-comma1 (rest arg)))
    arg))

(defun PUT-COMMA1 (arg)
  (cond
   ((null arg) nil)
   ((not (second arg))
    (cons #\, arg))
   (T (cons #\,
	    (cons (first arg)
		  (put-comma1 (rest arg)))))))

(defun PUT-COMMA1 (arg)
  (cond
   ((null arg) nil)
   ((not (> (length arg) 1))
    (cons #\, arg))
   (T (cons #\,
	    (cons (first arg)
		  (put-comma1 (rest arg)))))))


;----------------------------------------------------
(setf *PATH-TO-FENVS* 
      (concatenate 'string
		   user::**plasm-directory
		   #+:APPLE ":"
		   #+:UNIX "/"
		   #+:DOS "\\"
		   ;;#+:RS6000 "/"
		   "user-fenv"
		   #+:APPLE ":"
		   #+:UNIX "/"
		   #+:DOS "\\"
		   ;;#+:RS6000 "/"
                   ))

(defun SAVE-USER-FENV (&optional filename)
  (declare (special **USER-FENV** *PATH-TO-FENVS* ))
  (let ((filename (if filename 
		    filename
		    (prompt-for-a-new-file
		     *PATH-TO-FENVS* 
		     "untitled.fen"))))
    (if (not (string-equal (pathname-type filename) "fen"))
      (setq filename (concatenate 'string (pathname-host  filename)
				  (pathname-device filename)
				  (pathname-directory filename)
				  (pathname-name filename)
				  ".fen")))
    (with-open-file (outfile filename
			     :direction :output
			     :if-exists :overwrite
			     :if-does-not-exist :create)
      (prin1 (get-all-entries-from-temp-fenv nil) outfile))
     t))

(defun READ-USER-FENV (&optional filename)
  (declare (special *PATH-TO-FENVS* ))
  (if (not filename) 
    (setq filename (prompt-for-an-existing-file
		    *PATH-TO-FENVS* )))
  (with-open-file (infile filename
			  :direction :input)
      (let ((funs-list (read infile)))
	(if (every #'plasm-function-p funs-list)
	  (mapcar #'insert-in-user-fenv funs-list)
	  (plasm-error 500 funs-list))
	T)))

;---------------------------------------------------------
; It is the path to plasm script files.
(setf *PATH-TO-FOLDER-PLASM_SCRIPTS* (concatenate 'string
				    user::**plasm-directory
				    #+:APPLE ":"
				    ;;#+:RS6000 "/"
				    #+:UNIX "/" 
				    #+:DOS "\\"
				    "PLASM_SCRIPTS"
				    #+:UNIX "/" 
				    #+:DOS "\\"
				    #+:APPLE ":"
				    ;;#+:RS6000 "/"
                                    ))

;---------------------------------------------------------
; READ-PLASM-SCRIPT reads a PLASM script from a file.
; The file can contain one or more LISP forms: every string will be evaluated
; as PLASM script, while others LISP forms will be evaluated in the standard
; way.

(defun READ-PLASM-SCRIPT (&optional filename)
  (declare (special *PATH-TO-FOLDER-PLASM_SCRIPTS*))
  (if (not filename) 
    (setq filename (prompt-for-an-existing-file
		    *PATH-TO-FOLDER-PLASM_SCRIPTS*
		    )))
  (with-open-file (infile filename
			  :direction :input)
    (let (obj)
      (loop
	(if (setf obj (read  infile nil nil nil))
	  (progn
	    (format T "~%~%~A~%~%" obj)
	    (new-eval obj))
	  (return (values)))))))


;---------------------------------------------------------
; It exports a polyhedral complex in a minerva format and
; returns  the polyhedral complex argument or
; nil if the argument is not a polyhedral complex.
; It is called by the EXPORT PLASM function.


;it has been modified to be used with plasm function export_so, which exports
; a polcomplex to *standard-output* stream

(defun EXPORT-POL-COMPLEX-PLASM (pol-complex &rest file-name)
  (if (not-void-polyhedra-complex? pol-complex)
    (cond
     ((= 2 (length file-name))
      (if (and (= 3 (get-npol-complex pol-complex))
		 (= 3 (get-dpol-complex pol-complex)))
	  (export-pol-complex (skeleton pol-complex 2) nil nil)
	  (export-pol-complex pol-complex nil nil)))
     (T (let* ((file-name (first file-name))
             (filename (if file-name
		         (let* ((dir-name (directory-namestring file-name))
			        (name-string (file-namestring file-name))
			        (empty-dir (string= dir-name 
					            #+:RS6000 "./" 
					            #+:APPLE ""
					            #+:CLISP ""))
			        (Full-dir (if (and (not empty-dir)
					           #+:APPLE (directoryp dir-name)
					           #+:RS6000 (XTK::DIRECTORYP dir-name)
					           #+:CLISP (DIRECTORYP dir-name))
				            dir-name)))
		           (make-pathname  :directory (first (directory full-dir))
				           :name  name-string))
		         (prompt-for-a-new-file "untitled" ".pol"))))
	(format t "~%~%Exporting ~A to file \"~A\" ..." 
                (print-plasm pol-complex) filename)
	(if (and (= 3 (get-npol-complex pol-complex))
		 (= 3 (get-dpol-complex pol-complex)))
	  ;;(pol-complex-to-winged-rep-to-file pol-complex filename)
	  (export-pol-complex (skeleton pol-complex 2) filename)
	  (export-pol-complex pol-complex filename))
	(format t "done.~%~%")
	pol-complex)))
    (plasm-error 2002 pol-complex)))


(defun IMPORT-POL-COMPLEX-PLASM (name-pol-complex)
  (let ((filename (prompt-for-an-existing-file
		    ))
	pol-complex)
    (with-open-file (infile filename
			    :direction :input)
      (setq pol-complex (read infile nil nil nil))
      (setq pol-complex
	    (cond
	     ((not-void-polyhedra-complex? pol-complex)
	      pol-complex)
	     
	     ((simplexn-format-p pol-complex)
	      (eval pol-complex))
	     
	     ((minerva-format-p pol-complex)
	      (rebuild-from-minerva pol-complex))
	     
	     (T (plasm-error 10000 'not-pol-complex pol-complex) )))
      
      (insert-in-user-fenv 
       (make-plasm-function  
	:name name-pol-complex
	:value pol-complex)))))

	
(defun SIMPLEXN-FORMAT-P (arg)
  (and (consp arg)
       (eq 'list (first arg))))

(defun MINERVA-FORMAT-P (arg)
  (cond
   (arg nil)))
;-----------------------------------------------------------------------------
;La seg. funzione e' chiamata dalla funzione PLASM di nome
;EXPORT_NRB e chiama la funzione del simplexn EXPORT-POL-COMPLEX-NRB
;che esporta i punti di controllo dei simplessi cubici di un poliedro
;complesso.
;Se i simplessi cubici non sono stati calcolati, deve restituire un errore.
;    
; by Marco Conti
;-----------------------------------------------------------------------------
(defun EXPORT-POL-COMPLEX-PLASM-NRB (pol-complex &optional file-name)
  (if (not-void-polyhedra-complex? pol-complex)
    (let ((filename (if file-name
		   (let* ((dir-name (directory-namestring file-name))
			  (name-string (file-namestring file-name))
			  (empty-dir (string= dir-name 
					      #+:RS6000 "./" 
					      #+:APPLE ""
					      #+:CLISP ""))
			  (Full-dir (if (and (not empty-dir)
					     #+:APPLE (directoryp dir-name)
					     #+:RS6000 (XTK::DIRECTORYP dir-name)
					     #+:CLISP (DIRECTORYP dir-name))
				      dir-name )))
		     (make-pathname  :directory (first (directory full-dir))
				     :name  name-string))
		   (prompt-for-a-new-file "untitled" ".pol"))))
	(format t "~%~%Exporting ~A to file \"~A\" ..." 
		  (print-plasm pol-complex) filename)
	(io:export-pol-complex-nrb pol-complex filename)
	(format t "done.~%~%")
	pol-complex)
      (plasm-error 2002 pol-complex)))



