; -*- package: plasm -*-
;======================================================================
;                   THIS IS FILE lex.lisp
;======================================================================

(provide 'lex)
(inpackage plasm)


;it makes the bar a plain character.
; bar is used for intersection plasm operator


(set-syntax-from-char #\| #\a)

;the syn-tree-3 function switchs between the *PLASM-READTABLE* and a
; standard lisp readtable

(setf *PLASM-READTABLE* *READTABLE*)


;----------------------------------------------------------------------------------
; chars returned by a calling  such a  (code-char i), with 'i' less or equal to 32,
; are special: they can't be 'plasm chars'!

(defun NOT-IMPORTANT-CHAR (chars-array index)
  (<= (char-code (aref chars-array index)) 32))
;----------------------------------------------------------------------------------


(defun IS-LETTER (a-char)
  (or (and (char>= a-char #\A) (char<= a-char #\Z))
      (and (char>= a-char #\a) (char<= a-char #\z))))

(defun  IS-DIGIT (a-char)
  (and (char>= a-char #\0) 
       (char<= a-char #\9)))

(defun IS-UNDERSCORE (a-char)
  (char= #\_ a-char))

(defun IS-EXTRA-CHAR (a-char)
  (char= #\$ a-char))

(setf *OTHER-CHARS* '(#\^ #\! #\?))

(defun IS-GOOD-CHAR-FOR-IDENTIFIER (a-char)
  (declare (special *OTHER-CHARS*))
  (member a-char *OTHER-CHARS*))

(defun IS-NUM (a-char b-char)
   (or (is-digit a-char)
       (and (is-sign a-char)
            b-char
            (is-digit b-char))))

(defun IS-NEG-NUM (a-char b-char)
   (or (is-digit a-char)
       (and (is-minus-sign a-char)
            b-char
            (is-digit b-char))))

(defun IS-SIGN (a-char)
  (or (char= #\+ a-char)
      (char= #\- a-char)))

(defun IS-MINUS-SIGN (a-char)
  (char= #\- a-char))


(defun BEGIN-REMARK-P (array index)
  (char= (aref array index) #\%))


;-----------------------------------------------------------------------
;this function picks up a list of chars and converts it to a string

(defun COERCE-STRING (list-of-chars)
  (let* ((len (length list-of-chars))
         (new-string (make-string len)))
    (dotimes (i len new-string)
      (setf (aref new-string i) (nth i list-of-chars)))))          

;------------------------------------------------------------------------
;PICK-ID reads from string 'stringa' a sequence of char of 'stringa'
;until it does not reach a char not allowed for identificators.It returns
;a string value.Tehre is a call to this function if beginner identificator
;was just recognized.

(defun PICK-ID (array-script index index-end)
  (declare (special *EXTRA-ID*))
   (let ((out index)
         (pointed-char #\Null))
     (loop 
       (setq out (1+ out))
       (if (or (= out index-end)
               (progn
                 (setq pointed-char (aref array-script out))
                 nil) ; nil result can't allow for an out from the OR macro 
               (not (or (is-digit pointed-char)
                        (is-letter pointed-char)
                        (is-underscore pointed-char)
                        (IS-GOOD-CHAR-FOR-IDENTIFIER pointed-char)
                        (and *EXTRA-ID*
                             (is-extra-char pointed-char)))))
         (return (- out index))))))


;----------------------------------------------------------------------------------
; Consente di considerare come identifier anche sequenze contenenti un carattere
; extra ($). E' utile quando si vuole definire una nuova funzione da inserire come 
; default nell'ambiente funzionale che utilizza pero' altre funzioni che non devono
; essere visibili all'utente. Quest'ultime possono essere mascherate utilizzando 
; nel loro nome il carattere extra.

(setf *EXTRA-ID* nil)

(defun PROVIDE-EXTRA-ID ()
  (declare (special *EXTRA-ID*))
  (setf *EXTRA-ID* t))

(defun DO-NOT-PROVIDE-EXTRA-ID ()
  (declare (special *EXTRA-ID*))
  (setf *EXTRA-ID* nil))



;--------------------------------------------------------------------------
;It returns a string which first and last char is #\' .

(defun PICK-USER-STRING (array-script index index-end)
  (let ((scan-index index))
     (loop 
       (incf scan-index)

       (cond 
        ((= scan-index index-end) 
         (plasm-error 100000 array-script index index-end scan-index))

        ((and (char= #\' (aref array-script scan-index))
              (< (1+ scan-index) index-end)
              (char= #\' (aref array-script (1+ scan-index))))
         (incf scan-index 2)) ; === inside a user string can be used ===
       ; === the delimiter char #\' typing it ===
       ; === twice. 'pippo e'' maldestro' is  ===
       ; === a user string !                  ===
       ((char= #\' (aref array-script scan-index)) 
        (return (- scan-index index -1)))))))

;--------------------------------------------------------------------------
;The two following func. cuts from head string an integer - signed or not -
;if 'array-script' starts with an integer in 'index' position, otherwise returns 0.

(defun PARSE-INTEGER-STRING (array-script index index-end)
  (cond
     ((and (is-sign (aref array-script index))
           (< (1+ index) index-end) 
           (is-digit (aref array-script (1+ index))))
      (1+ (parse-not-signed-int-string array-script (1+ index) index-end 0)))
     ( T ; === not signed integer ===
       (parse-not-signed-int-string array-script index index-end 0))))



(defun PARSE-NOT-SIGNED-INT-STRING (array-script index index-end res)
  (cond 
     ((= index index-end)
      res)
     ((is-digit (aref array-script index))
      (parse-not-signed-int-string array-script (1+ index) index-end (1+ res)))
     (T  
      res)))

;----------------------------------------------------------------------
;It extracts from 'script' a number. A number rule is: 

;   <number> :: [sign] {digit}+ [ . {digit}+ ] [ E [sign] {digit}+ ] 

;When this function is called a beginner number was just recognized .

(defun PICK-NUM (array-script index index-end)
  (let ((scan-index index))
    (incf scan-index (parse-integer-string array-script index index-end))

    (if (and (< (1+ scan-index) index-end) ; are there at least two others chars?
             (char= #\. (aref array-script scan-index))
             (is-digit (aref array-script (1+ scan-index))))
        ; === there is a fractional part ===
          (incf scan-index 
                (1+ (parse-integer-string array-script (1+ scan-index) index-end)))
          ) ; now index  points to a non digit
            ; the frac part do not consider decimal
            ; point, so we must add 2!

    (if (and (< (1+ scan-index) index-end) ; are there at least two others chars?
             (char= #\E (aref array-script scan-index))
             (or (is-digit (aref array-script (1+ scan-index)))
                 (is-sign (aref array-script (1+ scan-index)))))
        (incf scan-index
              (1+ (parse-integer-string array-script (1+ scan-index) index-end))))

    (- scan-index index)))
 



;---------------------------------------------------------------------------

(defun PICK-CHIOCCIOLA (array-script index index-end)
  (let* ((index-plus-1 (1+ index))
         (b-char (if (< index-plus-1 index-end)
                   (aref array-script index-plus-1))))
    (if (and b-char (is-digit b-char))
      2
      1)))


;------------------------------------------------------------------

(defun SKIP-TO-SECOND-% (array index end-index)
  (loop
    (setq index (1+ index))
    (cond
     ((= index end-index)
      (plasm-error 403 "You have typed a open remark!"))
     ((char= #\% (aref array index)) 
      (return index)))))
    
(defun SKIP-TO-SECOND-% (array index end-index)
  (let ((started-in index))
    (loop
     (setq index (1+ index))
     (cond
       ((= index end-index)
        (plasm-error 403 (subseq array started-in end-index)))
       ((char= #\% (aref array index)) 
        (return index))))))
    

;------------------------------------------------------------------
;This is the function that takes a PLASM script and transform it in
;a tokened script. All IDENTIFIERs, NUMBERs, PRIMITIVE-OPERATORs will be
;substituted with a token.

(defun LEX (array-script)
  (let* (
         (*END-INDEX* (length array-script)) 
         ;*END-INDEX* index does not exist in array-script array
         
         (*ACTUAL-INDEX* -1)
         (tokened-script '())
         (elem-length     nil)
        )
    (loop
      (incf *ACTUAL-INDEX*)
      (if (= *ACTUAL-INDEX* *END-INDEX*)
        (return (reverse tokened-script))
        (cond

         ((not-important-char array-script *ACTUAL-INDEX*))

         ((begin-remark-p array-script *ACTUAL-INDEX*)
          (setq *ACTUAL-INDEX* 
                (skip-to-second-% array-script *ACTUAL-INDEX* *END-INDEX*)))

         ((setq elem-length (detect-elem array-script *ACTUAL-INDEX* *END-INDEX*))
          (setq tokened-script
                (cons (give-me-token-of 
                       (subseq array-script 
                               *ACTUAL-INDEX* (+ *ACTUAL-INDEX* elem-length)))
                      tokened-script))
          (setq *ACTUAL-INDEX* (+ *ACTUAL-INDEX* elem-length -1)))

         (T ; === the *ACTUAL-INDEX*-th char of 'script' is not the  
            ;     first of something recognized as token associable ===
          (setq tokened-script
                (cons (aref array-script *ACTUAL-INDEX*)
                      tokened-script))))))))


(defun LEX (array-script *ACTUAL-INDEX* *END-INDEX*)
  ;*END-INDEX* index does not exist in array-script array
  (let* (
         (elem-length     nil)
        )
    (loop
      (if (= *ACTUAL-INDEX* *END-INDEX*)
        (return nil)
        (cond

         ((not-important-char array-script *ACTUAL-INDEX*))

         ((begin-remark-p array-script *ACTUAL-INDEX*)
          (setq *ACTUAL-INDEX* 
                (skip-to-second-% array-script *ACTUAL-INDEX* *END-INDEX*)))

         ((setq elem-length (detect-elem array-script *ACTUAL-INDEX* *END-INDEX*))
          (return (list (give-me-token-of 
                         (subseq array-script 
                                 *ACTUAL-INDEX* (+ *ACTUAL-INDEX* elem-length)))
                      (+ *ACTUAL-INDEX* elem-length))))

         (T ; === the *ACTUAL-INDEX*-th char of 'script' is not the  
            ;     first of something recognized as token associable ===
          (return (list (aref array-script *ACTUAL-INDEX*)
                        (1+ *ACTUAL-INDEX*))))))
      (incf *ACTUAL-INDEX*))))

;--------------------------------------------------------------------
;It tries to recognize an element that must be substitute with a token.
;It returns a string if it has succefull, otherwise NIL .


(defun DETECT-ELEM  (array-script index index-end)
  (let ((first-char (aref array-script index))
        (second-char (if (< (1+ index) index-end)
                         (aref array-script (1+ index)))))
     (cond
       ((or (is-letter first-char)
            (IS-GOOD-CHAR-FOR-IDENTIFIER first-char))
        (pick-id array-script index index-end))

       ;=== numbers can be typed with minus sign and without sign. note that  
       ;=== "- 3" is not a negative number while "-3" is a negative number
       ;=== positive numbers can't be typed with a plus sign. note that "+3"  
       ;=== is equivalent to "op 3", where "op" is in this case "+".  
       ((is-neg-num first-char second-char)
        (pick-num array-script index index-end))

       ((my-char= #\+ first-char second-char) ; === is it ++ op ?
          ;"++"
        2)

       ((char= #\+ first-char) ; === is it + op ?
          ;"+"
        1)

       ((char= #\- first-char) ; === is it - op ?
          ;"-"
        1)

       ((my-char= #\* first-char second-char) ; === is it ** op ?
          ;"**"
        2)

       ((char= #\* first-char) ; === is it * op ?
          ;"*"
        1)

       ((char= #\/ first-char) ; === is it / op ?
          ;"/"
        1)

       ((my-char= #\' first-char second-char ) ; === is it QUOTE-PLASM op ?
          ;"''"
        2)

       ((char= #\' first-char) (pick-user-string array-script index index-end))

       ((my-char= #\& first-char second-char) ; === is it && op ?
         ; "&&"
        2)

       ((char= #\& first-char) ; === is it & op ?
         ; "&"
        1)

       ((my-char= #\| first-char second-char) ; === is it "||"  op ?
         ; "||"
        2)

       ((char= #\| first-char) ; === is it "|"  op ?
         ; "|"
        1)

       ((my-char= #\# first-char second-char) ; is it ## op ?
         ; "##"
        2)

       ((char= #\# first-char) ; is it # op ?
         ; "#"
        1)

       ((my-char= #\. first-char second-char) ; === is it .. op ?
         ; ".."
        2)

       ((char= #\@ first-char) ; === is it SKELETON op ?
        (pick-chiocciola array-script index index-end)))))

;--------------------------------------------------------------------------
;

(defun MY-CHAR= (&rest list-of-chars)
  (if (member nil list-of-chars)
    nil
  (apply #'char= list-of-chars)))

;--------------------------------------------------------------------------
;This function returns the token that can substitute part of PLASM script.


(defun GIVE-ME-TOKEN-OF (sym-string)
  (cond
   ((equal sym-string "#") 'DIESIS) ; function 'equal' is faster than 'string=' !
   ((equal sym-string "##") 'DOUBLE-DIESIS)
   ((equal sym-string "..") 'FROMTO)
   ((char= #\' (char sym-string 0)) ; sym-string is a plasm string
    (subseq sym-string 1 (1- (length sym-string)))) 
   (T
    (let ((sym-name (read-from-string sym-string)))
      (cond
       
       ((equal sym-name 'TRUE) 'T)
       ((equal sym-name 'FALSE) 'NIL)     
       ((equal sym-name 'T) 'TRANSLATE)     
       ((equal sym-name 'DEF) #\0)
       ((equal sym-name 'DEFAULT) #\1)
       ((equal sym-name 'END) #\2)
       ((equal sym-name 'WHERE) #\3)

       (T sym-name))))))



                      
;-----------------------------------------------------------------------
;this function picks up a string and converts it to an array of chars

;(defun COERCE-CHARS-ARRAY (stringa len)
;  (let ((array-chars (make-array len
;                                 :element-type 'character
;                                 :initial-element #\Null)))
;    (dotimes (i len array-chars)
;      (setf (aref array-chars i) (char stringa i)))))

;------------------------------------------------------------------------


;(defun TAKE-CHARS (elem-length array index)
;  (let ((res '())
;        (last (+ elem-length index -1)))
;    (loop
;      (if (< last index)
;        (return res))
;      (setq res (cons (aref array last) res))
;      (decf last))))
