; -*- package: io -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;           Universita' degli Studi di Roma Tre
;
;       Dipartimento di Informatica e Automazione
;
;            Via della Vasca Navale, 84 00146
;
;                 Rome, Italy
;
;
;
;                      CAD Group
;
;
;
; Project: PLASM
;
; Identification: Package Input&Output
;
;
;
;  Version: 1.0
;  Date: 11 August 1998
;
;
;
;  Author: M. Vicentino
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
(provide 'p_iv)

(inpackage IO)

(export '(
          GENERATE-POL-COMPLEX-STRING-IV-FORMAT
          WRITE-OVERTURE-IV-FILE
          WRITE-CLOSURE-IV-FILE
          SPOTLIGHT-IV
          DIRECTIONALLIGHT-IV
          WRITE-SEPARATOR-OVERTURE
          WRITE-SEPARATOR-CLOSE
          WRITE-PROPERTY-IV-NODE
          EXPORT-IV-FILE
          EXPORT-IV-STREAM
	  ))


;----------------------------------------------------

(defun WRITE-PROPERTY-IV-NODE (outfile evaluated-property)
  (cond
   ((string= "COLOR" (first evaluated-property))
    (write-color-iv-node outfile (rest evaluated-property)))
   ))

(defun WRITE-COLOR-IV-NODE (outfile values)
  (format outfile 
          "~%    Material {
	diffuseColor	~{ ~A ~}
    }~%"
          values))


(defun WRITE-SEPARATOR-OVERTURE (outfile)
  (format outfile "Separator {~%")
  )

(defun WRITE-SEPARATOR-CLOSE (outfile)
  (format outfile "~%}~%")
  )

(defun WRITE-SHAPEHINTS-NODE (outfile)
  (format outfile "ShapeHints {	vertexOrdering	CLOCKWISE    }~%"))

(defun WRITE-OVERTURE-IV-FILE (outfile)
  (format outfile "#Inventor V2.1 ascii~%~%")
  (format outfile "Separator {~%")
  (write-shapehints-node outfile)
  )


(defun WRITE-CLOSURE-IV-FILE (outfile)
  (format outfile "~%}~%") ; the end of the separator opened by WRITE-OVERTURE-IV-FILE function
  (format outfile "#Inventor V2.1 ascii File End~%"))


(defun SPOTLIGHT-IV (outfile &optional (location '(0 1 0)) 
                         (direction '(0 -1 0)) 
                         (intensity 1) (cutOffAngle 1.9))
  (format outfile 
  "SpotLight {      # Light from above
    location      ~{~A ~}
    direction     ~{~A ~}
    intensity     ~A
    cutOffAngle   ~A
  }~%" location direction intensity cutOffAngle))

(defun DIRECTIONALLIGHT-IV (outfile &optional (direction   '(0 0 -1))
                                    (intensity   1.0))
  (format outfile 
          "DirectionalLight {
    direction   ~{~A ~}  
    intensity   ~A
  }~%"
          direction intensity))

;----------------------------------------------------------------------

(defvar *OIV* nil)
(defun RESET-OIV-FORMAT ()
  (declare (special *OIV*))
  (setf *OIV* NIL))

(defun SET-OIV-FORMAT ()
  (declare (special *OIV*))
  (setf *OIV* T))

(defun OIV-FORMAT? ()
  (declare (special *OIV*))
  *OIV*)

;---------------------------------------------------

(defun GENERATE-POL-COMPLEX-STRING-IV-FORMAT (pol-complex &optional stream)
  (let ((*N-SPACE* (get-npol-complex pol-complex))
        (old-vrml-ver-setting-form `(SET-VRML-VER ,(if (not (VRML-VER-1?)) 1 2)))
        (double-face-old-status-setting-form `(set-doubled-faces ,(doubled-faces?)))
        (old-oiv-setting-form (if (oiv-format?)
                                '(SET-OIV-FORMAT)
                                '(RESET-OIV-FORMAT)))
        (string-stream (cond
                        (stream)
                        ((make-string-output-stream)))))
    (declare (special *N-SPACE*))
    (set-vrml-ver 1)
    (do-not-export-with-doubled-faces)
    (set-oiv-format)  
    (format string-stream "Separator {~%")
    ;(set-vrml-ver 1)
    ;(set-oiv-format)
    (write-pol-complex-to-vrml  pol-complex string-stream)
    (format string-stream "}~%")
    (eval old-oiv-setting-form)
    (eval old-vrml-ver-setting-form)
    (eval double-face-old-status-setting-form)
    (if (not stream)
      (get-output-stream-string string-stream))
  ))






(defun EXPORT-IV-FILE (pol-complex &rest streams)
"This function generates an Open Inventor text format file for 
its polyhedral complex argument."
  (let ((filename (or (first streams)
                      (prompt-for-a-new-file 
                       "untitled" ".iv"))))
    (with-open-file (outfile filename 
                             :direction :output
                             :if-exists :supersede
                             :if-does-not-exist :create)
      (format t "~%Exporting~%   \"~A\" ~%  to Open Inventor file ~%     \"~A\" ..~%"
              (plasm::print-plasm pol-complex) filename)
      (IO::SET-RESET-ENVIRONMENT)
      (WRITE-OVERTURE-IV-FILE outfile)
      (generate-pol-complex-string-iv-format pol-complex outfile)
      (WRITE-CLOSURE-IV-FILE outfile)
      (format t "... done.~%")
      )))
;Please, if you have modified the previous function you have to modify the nextone accordingly!

(defun EXPORT-IV-STREAM (pol-complex  &optional stream)
"This function generates an Open Inventor text format file for 
its polyhedral complex argument writing on argument stream."
   (let ((out (if stream stream (make-string-output-stream))))
      (IO::SET-RESET-ENVIRONMENT)
      (WRITE-OVERTURE-IV-FILE out)
      (generate-pol-complex-string-iv-format pol-complex out)
      (WRITE-CLOSURE-IV-FILE out)
      (if (not stream) 
        (get-output-stream-string out))))








