; -*- package: plasm -*-
;==========================================================================
;
;                           MISCELLANOUS.LISP File
;
;==========================================================================


(provide 'p_misc)
(require 'up_file)

(inpackage plasm)

(export '(
	  save-user-fenv
	  read-user-fenv
	  ))





(defun IS-PLASM-AFFINE-TRANSFORMATION (obj)
  (declare (special *AFFINE-TRANSFORMATION-NAMES*))
  (and (consp obj)
       (member (first obj) *AFFINE-TRANSFORMATION-NAMES*)))



(defun IS-PLASM-AFFINE-TRANSFORMATION (obj)
  (if obj
    (or (functionp obj)
	(bound-function-p obj))))



(defun CAN-IT-BE-IN-PLASM-STRUCTURE (obj)
  (or (is-plasm-affine-transformation obj)
      ;(polyhedron? obj)
      (not-void-polyhedra-complex? obj)
      ;(is-plasm-structure obj)

      ))




(defun IS-PLASM-STRUCTURE (obj)
  (if (consp obj)
    (not (member nil (mapcar #'can-it-be-in-plasm-structure obj)))))


(defun IS-PLASM-STRUCTURE (obj)
  (and (consp obj)
       (not (member nil (mapcar #'can-it-be-in-plasm-structure obj)))
       (are-pol-complexes-of-same-ndim obj)))


(defun ARE-POL-COMPLEXES-OF-SAME-NDIM (plasm-structure-elements)
  (let* ((pol-complexes-in-structure
	  (get-pol-complexes-in-plasm-structure plasm-structure-elements))
	 (n-dims (mapcar 'get-npol-complex pol-complexes-in-structure)))
    (if (= (apply 'max n-dims) (apply 'min n-dims))
      T
      (plasm-error 2001))))


(defun GET-POL-COMPLEXES-IN-PLASM-STRUCTURE (seq)
  (remove nil
	  (mapcar #'(lambda (seq-elem)
		      (if (not-void-polyhedra-complex? seq-elem)
			seq-elem))
		  seq)))

   


(defun MAKE-MULTIPLE-AFFINE-TRANSFORMATION (pair-coord-value-list polcomplex type-trans)
  (declare (special *SCALING* *SHEARING* *ROTATION* *TRANSLATION* *MAT*))
  (let ((funtocall 
	 (cond
	  ((eq type-trans *SCALING*)
	   'scale-pol-complex-plasm)
	  ((eq type-trans *SHEARING*)
	   'shear-pol-complex)
	  ((eq type-trans *ROTATION*)
	   'rotate-pol-complex)
	  ((eq type-trans *TRANSLATION*)
	   'translate-pol-complex-plasm))))
    (cond
     ((eql funtocall 'rotate-pol-complex)
      
      (dolist (coord-value pair-coord-value-list polcomplex)
        (setf polcomplex
              (funcall #'rotate-pol-complex
                       (1- (first (first coord-value)))
                       (1- (second (first coord-value)))
                       (second coord-value)
                       polcomplex))))
     (funtocall
      
      (funcall funtocall
               pair-coord-value-list
               polcomplex))
    ((eq type-trans *MAT*)
     
     (let* ((dim-matrix (1+ (get-npol-complex polcomplex)))
            (dim-rows (length pair-coord-value-list))
            (dim-cols (length (first pair-coord-value-list)))
            (matrix (matlib::mat-identity dim-matrix)))
       (dotimes (i dim-rows) 
         (dotimes (j dim-cols)
           (setf  (aref  matrix i j) 
                  (nth j (nth i pair-coord-value-list ; in questo caso sara'
                                                      ; una lista di liste
                              )))))
       (general-transform polcomplex (matlib::mat-inv matrix))))
    (T (plasm-error 2000 'unknown-transformation)))))


(defun APPLY-AFFINE-TRANSFORMATIONS (affine-transformations pol-complex)
  (if pol-complex ; pol-complex can be nil if there are affine transformation
		  ; as last elements of a PLASM structure
    (dolist (onetransf (reverse affine-transformations) pol-complex)
      (setf pol-complex
	    (make-multiple-affine-transformation
	     (rest onetransf) ; transformation parameters
	     pol-complex
	     (first onetransf) ; transformation type
	     )))))


(defun APPLY-AFFINE-TRANSFORMATIONS (affine-transformations pol-complex)
  (if pol-complex ; pol-complex can be nil if there are affine transformation
		  ; as last elements of a PLASM structure
    (dolist (onetransf (reverse affine-transformations) pol-complex)
      (setf pol-complex
	    (reduce-function-args onetransf pol-complex)))))




(defun GIVE-ME-POLYHEDRAL-COMPLEX-OF-PLASM-STRUCTURE (plasm-struct)
  (if (is-plasm-structure plasm-struct)
      (let* ((reps  (mapcar 
		     #'(lambda (pol) 
			 (if (not (is-plasm-affine-transformation pol))
			     (get-pol-complex-rep pol)))
		     plasm-struct))
	     (rep  (some  #'(lambda (x) x) reps)))
	(dotimes (i (length reps))
	  (if (and (not (eq  rep (nth i reps))) (nth i reps))
	      (setf (nth i plasm-struct)
		(coerce-pol-complex (nth i plasm-struct) rep))))
	(give-me-polyhedral-complex-of-plasm-structure1 plasm-struct))
    (plasm-error 2000 plasm-struct)))

   
   
(defun GIVE-ME-POLYHEDRAL-COMPLEX-OF-PLASM-STRUCTURE1 (plasm-struct)
  (cond
   ((is-plasm-affine-transformation (first plasm-struct))
    (let* ((sequential-transf (pick-up-seq-transf plasm-struct))
	   (rest-struct (nthcdr (length sequential-transf) plasm-struct)))
      (apply-affine-transformations 
       sequential-transf 
       (give-me-polyhedral-complex-of-plasm-structure1 rest-struct))))
   ((not-void-polyhedra-complex? (first plasm-struct))
    (let* ((sequential-complex (pick-up-seq-complex plasm-struct))
	   (pol-complex (give-me-polyhedral-complex-of-plasm-structure1 
						(nthcdr (length sequential-complex) plasm-struct))))
	(dolist (pol (reverse sequential-complex) pol-complex)
		(put-node-in-pol-complex-as-first (make-pol-node pol) pol-complex))))))
   


(defun PICK-UP-SEQ-TRANSF (plasm-struct)
  (let ((i 0))
    (loop (if (is-plasm-affine-transformation (nth i plasm-struct))
	    (setf i (1+ i))
	    (return)))
    (subseq plasm-struct 0 i)))



(defun PICK-UP-SEQ-COMPLEX (plasm-struct)
  (let ((i 0))
    (loop (if (not-void-polyhedra-complex? (nth i plasm-struct))
	    (setf i (1+ i))
	    (return)))
    (subseq plasm-struct 0 i)))


(defun NOT-VOID-POLYHEDRA-COMPLEX? (pol-complex)
  (and pol-complex (polyhedra-complex? pol-complex)))




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; 

(defun MAKE-CALLING-TREE-OF-FUN (f)
  (make-calling-tree-of-fun2 f nil))


;
; questa funzione restituisce l'albero delle chiamate a funzioni non predefinite
; per la valutazione della funzione f
; - per evitare loops e' necessario che della funzione f siano noti quali sono
;  i suoi progenitori, cioe' le funzioni che 
;

(defun MAKE-CALLING-TREE-OF-FUN2 (f progenitori-di-f)
  (let ((called-functions 
	 (set-difference
	  (plasm-function-called-function (bound-function f))
	  progenitori-di-f)))
    (if called-functions
      (list f (mapcar #'(lambda (arg)
			(make-calling-tree-of-fun2 arg (cons f progenitori-di-f)))
		      called-functions))
      f)))



(defun MAKE-CALLING-TREE-OF-FUN (f)
  (remove nil
	  (mapcar #'(lambda (arg) (if (function-without-par-p arg) arg))
		  (remove-duplicates (make-calling-tree-of-fun2 f nil)))))


(defun MAKE-CALLING-TREE-OF-FUN2 (f progenitori-di-f)
  (let ((called-functions 
	 (set-difference
	  (plasm-function-called-function (bound-function f))
	  progenitori-di-f)))
    (if called-functions
      (cons f
	     
	    (apply 'append   (mapcar #'(lambda (arg)
			(make-calling-tree-of-fun2 arg (cons f progenitori-di-f)))
		      called-functions)))
      (list f))))

;--------------------------------------------------------------
; trasforma un array bidimensionale in una lista di liste .
; l'i-esimo elemento della lista e' l'i-esima riga dell'array

(defun BIDIM-ARRAY-TO-LIST-OF-LIST (bidim-array)
  (let* ((array-dims (array-dimensions bidim-array)) 
	 (list-of-list (make-sequence 'list (first array-dims) :initial-element nil)))
       (dotimes (i (first array-dims) list-of-list) 
	 (let (row)
	   (dotimes (j (second array-dims) (setf (nth i list-of-list) (reverse row)))
	     (setf row
		   (cons (aref bidim-array i j) row)))))))          


;--------------------------------------------------------------
;  pol-complex-b diventa fratello dell'ultimo poliedro in una visita in profondita'
; del complesso poliedrale pol-complex-a

(defun CAT-POL-COMPLEX (pol-complex-a pol-complex-b)
  (if (polyhedron? pol-complex-a)
    (append (list (make-pol-node pol-complex-a))
	    pol-complex-b)
    (append (butlast pol-complex-a)
	    (let ((first-last (first (last pol-complex-a))))
	      (list
	       (make-pol-node (cat-pol-complex (get-node-pol-complex first-last)
					       pol-complex-b)
			      (get-node-matrix first-last)
			      (get-node-id first-last)))))))



;--------------------
; This function try to calculate how many functions were used by a CONS plasm functional
; to obtain the a plasm function whose symbol is "plasm-function-symbol".
; If this symbol is the value of a CONS plasm application it should return a right value.

(defun HOW-MANY-FUNCTION-FOR-THIS-CONS? (plasm-function-symbol)
  (if (and (bound-function-p plasm-function-symbol)
	   (let ((let-fun (plasm-function-let-vars 
			   (bound-function plasm-function-symbol))))
	     (and (= 3 (length let-fun))
		  (= 2 (length (third let-fun)))
		  (listp (second (second (third let-fun)))))))

    (length (second (second (third (plasm-function-let-vars 
			   (bound-function plasm-function-symbol))))))
    (plasm-error 8000 "Bad use of a MAP function! " )))


;--------------
(defun NORMALIZE-LIST-OF-NUMBERS (seq)
"It takes a list of number and returns a list of number with norma equal to one"
  (let* ((norma (apply (function +) (mapcar (lambda (x) (* x x)) seq))))
    (mapcar (lambda (x) 
              (/ x norma))
            seq)))

; (normalize-list-of-numbers '(1 2 3 4))
                    


;-------------------------------------------------------------------------
(defun SCALE-POL-COMPLEX-PLASM (i-scale-factor-pairs pol-complex)

  (let* ((n (get-npol-complex pol-complex))
	 (rep (get-pol-complex-rep pol-complex))
	 )
    (general-transform pol-complex 
                       (build-scale-matrix i-scale-factor-pairs
                                           rep
                                           (1+ n)))))


(defun TRANSLATE-POL-COMPLEX-PLASM (i-translate-factor-pairs pol-complex)
  (let* ((n (get-npol-complex pol-complex))
	 (rep (get-pol-complex-rep pol-complex)))
    (general-transform pol-complex 
                       (build-translate-matrix i-translate-factor-pairs
                                               rep
                                               (1+ n)))))


(defun BUILD-TRANSLATE-MATRIX (i-translate-factor-pairs rep n)
  (let ((matrix (kernel::mat-identity  n)))
    (if (kernel::symbol-equal rep 'on-faces)
      (dolist (coord-value i-translate-factor-pairs matrix)
        (setf (aref matrix (first coord-value) 0)
              (- (second coord-value))))
      (dolist (coord-value i-translate-factor-pairs matrix)
        (setf (aref matrix (first coord-value) 0)
              (second coord-value))))))

(defun BUILD-SCALE-MATRIX (i-scale-factor-pairs rep n)
  (let ((matrix (kernel::mat-identity  n)))
    (if (kernel::symbol-equal rep 'on-faces)
      (dolist (coord-value i-scale-factor-pairs matrix)
        (setf (aref matrix (first coord-value) (first coord-value))
              (/ (second coord-value))))
      (dolist (coord-value i-scale-factor-pairs matrix)
        (setf (aref matrix (first coord-value) (first coord-value))
              (second coord-value))))))



;(eval-plasm "struct:<s:<1,2,3>:<5,5,5>,cuboid:<1,1,1>>")