; -*- package: io -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;           Universita' degli Studi di Roma Tre
;
;       Dipartimento di Informatica e Automazione
;
;            Via della Vasca Navale, 84 00146
;
;                 Rome, Italy
;
;
;
;                      CAD Group
;
;
;
; Project: PLASM
;
; Identification: Package Input&Output
;
;
;
;  Version: 2.0
;  Date: 25 Feb 1998
;  Last fixed bug: Now it makes fine polygons collecting point along triangle adjacences
;
;
;
;  Author: M. Vicentino
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
(provide 'p_vrml)

(inpackage IO)

(export '(
           EXPORT-POL-COMPLEX-TO-VRML-FILE-FORMAT-FIRST
           SET-VRML-VER
	  ))

;---------------------------------------------------
(defvar *DOUBLED-FACES* T)
(defun SET-DOUBLED-FACES (boole)
  (declare (special *DOUBLED-FACES*))
  (setf *DOUBLED-FACES* boole))
(defun DOUBLED-FACES? ()
  (declare (special *DOUBLED-FACES*))
  *DOUBLED-FACES*)

(defun EXPORT-WITH-DOUBLED-FACES ()
  (set-doubled-faces T))

(defun DO-NOT-EXPORT-WITH-DOUBLED-FACES ()
  (set-doubled-faces nil))
;---------------------------------------------------
(defvar *APPEARANCE* nil)
(defun RESET-APPEARANCE-NO-WRITTEN ()
  (declare (special *APPEARANCE*))
  (setf *APPEARANCE* NIL))

(defun SET-APPEARANCE-JUST-WRITTEN ()
  (declare (special *APPEARANCE*))
  (setf *APPEARANCE* T))
(defun APPEARANCE-JUST-WRITTEN ()
  (declare (special *APPEARANCE*))
  *APPEARANCE*)

;---------------------------------------------------
(defvar *NAMES* )
(defvar *CURRENT-NAME*)

(defun RESET-NAMES ()
  (declare (special *NAMES* *CURRENT-NAME*))
  (setf *NAMES* NIL)
  (setf *CURRENT-NAME* nil))

(defun INSERT-INTO-NAMES (name)
  (declare (special *NAMES*))
  (if name
    (progn
      (setf *NAMES* (cons (concatenate 'string 
                                        (string name)
                                        "-MATERIAL")
                          *NAMES*))
      (return-name))))
(defun REMOVE-FROM-NAMES ()
  (declare (special *NAMES* *CURRENT-NAME*))
  (setf *NAMES* (rest *NAMES*))
  (setf *CURRENT-NAME* (rest *CURRENT-NAME*)))

(defun RETURN-NAME ()
  (declare (special *NAMES*))
  (first *NAMES*))

(defun appearance-just-written-for-this (name)
  (declare (special *CURRENT-NAME*))
  (member name *CURRENT-NAME* :test 'string=))

(defun set-as-appearance-just-written (name)
  (declare (special *CURRENT-NAME*))
  (setf *CURRENT-NAME* (cons name *CURRENT-NAME*)))

;---------------------------------------------------
(defvar *CMT*)
(defun RESET-CTM (dim)
  (declare (special *CTM*))
  (setf *CTM* (make-idmat dim)))

(defun SET-CURRENT-TRANSFORMATION-MATRIX (matrix)
  (declare (special *CTM*))
  (setf *CTM* (mat-product *CTM* matrix)))

(defun RETURN-CTM ()
  (declare (special *CTM*))
  *CTM*)
;----------------------------------------------------------------------

(defvar *VRML-VER* 2)

(defun SET-VRML-VER (ver)
  (declare (special *VRML-VER*))
  (setf *VRML-VER* ver))
(defun VRML-VER-1? ()
  (declare (special *VRML-VER*))
  (eq *VRML-VER* 1))

(defun VRML-VER-2? ()
  (declare (special *VRML-VER*))
  (eq *VRML-VER* 2))

;---------------------------------------------------

(defvar *JUST-SEEN-OBJECT*)
(defun RESET-HASH-TABLE-NAME ()
  (declare (special *JUST-SEEN-OBJECT*))
  (setf *JUST-SEEN-OBJECT* (make-hash-table :test #'eq)))
(defun SET-HASH-TABLE-NAME (pol-complex name)
  (declare (special *JUST-SEEN-OBJECT*))
  (setf (gethash pol-complex *JUST-SEEN-OBJECT*)
        name))
(defun GET-HASH-TABLE-NAME (pol-complex)
  (declare (special *JUST-SEEN-OBJECT*))
  (gethash pol-complex *JUST-SEEN-OBJECT*))


(defun SET-AS-JUST-SEEN (obj &key (name 'model) (use-name-as-radix nil))
  (declare (special *JUST-SEEN-OBJECT*))
  (set-hash-table-name obj
                       (cond
                        ((GET-HASH-TABLE-PLASM-NAME obj))
                        (T (if use-name-as-radix
                             (gensym (string (if name name 'MODEL)))
                             (string name))))))

;---------------------------------------------------------------------------

; PLASM::**USER-FENV** is the USER PLASM FUNCTIONAL ENVIRONMENT
; PLASM::PLASM-FUNCTION-VALUE is the lisp function getting the slot value VALUE
;  of a PLASM FUNCTION REPRESENTATION (a LISP STRUCTURE)
(defvar *PLASM-POL-COMPLEX-NAMES*)
(defun RESET-HASH-TABLE-PLASM-NAME ()
  (declare (special *PLASM-POL-COMPLEX-NAMES*))
  (setf *PLASM-POL-COMPLEX-NAMES* (make-hash-table :test #'eq)))
(defun CREATE-HASH-TABLE-PLASM-NAME ()
    (declare (special PLASM::**USER-FENV**))
    (reset-hash-table-plasm-name)
    (maphash #'(lambda (key value) (set-if-pol-complex key value))
             PLASM::**USER-FENV**))
(defun SET-IF-POL-COMPLEX (name plasm-structure-value)
  (let ((plasm-value (plasm::plasm-function-value plasm-structure-value)))
    (if (and plasm-value
             (polyhedra-complex? plasm-value))
      (set-hash-table-plasm-name plasm-value name))))

(defun SET-HASH-TABLE-PLASM-NAME (pol-complex plasm-name)
  (declare (special *PLASM-POL-COMPLEX-NAMES*))
  (setf (gethash pol-complex *PLASM-POL-COMPLEX-NAMES*)
        plasm-name))

(defun GET-HASH-TABLE-PLASM-NAME (pol-complex)
  (declare (special *PLASM-POL-COMPLEX-NAMES*))
  (gethash pol-complex *PLASM-POL-COMPLEX-NAMES*))

;---------------------------------------------------------------------------

;----------------------------------------------------

(defvar *TAB-COUNT*)

(defvar *DELTA-TAB* 2)

(defun TAB-COUNT ()
  (declare (special *TAB-COUNT*))
  *TAB-COUNT* )

(defun RESET-TAB-COUNTER ()
  (declare (special *TAB-COUNT*))
  (setf *TAB-COUNT* 0))
(defun INC-TAB-COUNTER ()
  (declare (special *TAB-COUNT* *DELTA-TAB*))
  (setf *TAB-COUNT* (+ *TAB-COUNT* *DELTA-TAB*)))

(defun DEC-TAB-COUNTER ()
  (declare (special *TAB-COUNT* *DELTA-TAB*))
  (setf *TAB-COUNT* (- *TAB-COUNT* *DELTA-TAB*)))

(defun INS-TAB (string_before tabulate &rest strings_after)
  (apply #'concatenate
         (list 'string
                string_before
                ;"~"
                ;(integer->string tabulate)
                ;"T"
                (apply #'concatenate
                       (cons 'string strings_after)))))

;----------------------------------------------------

(defun IS-NUMBER-ONE (number)
  (= 1 number))

(defun IS-IDENTITY-MAT (mat)
  (let* ((rows-cols (array-dimensions mat))
         (rows (first rows-cols))
         (cols (second rows-cols)))
    (block out
      (dotimes (i rows t)
        (dotimes (j cols t)
          (cond
           ((and (not (= i j))
                 (= 0 (aref mat i j))))
           ((and (= i j)
                 (= 1 (aref mat i j))))
           (t (return-FROM OUT nil))))))))



(defun TAKE-ALFA (axis matrix)
  (let* ((i (first-zero axis))
         (j (+ (1+ i) (first-zero (subseq axis (1+ i))))) ; i < j is always true !
         (alfa (acos (aref matrix (1+ i) (1+ i))))
         (sin-sign (signum (asin (aref matrix (1+ i) (1+ j)))))
         )
     (if (= sin-sign -1)
      alfa
      (- alfa))))


(defun FIRST-ZERO (alist)
  (block out
    (dotimes (i (length alist))
      (if (= 0 (nth i alist))
        (return-from out i)))))

(defvar *EPSILON-ROTATION* 1.0E-6)

(defun TAKE-ROTATIONAL-AXIS (matrix)
  (declare (special *N-SPACE* *EPSILON-ROTATION*))
  (let (axis
        (matrix-dim (1- (first (array-dimensions matrix)))))
    (cond
     ((= 2 matrix-dim) ; pol-complex is in xy plane !

;      matrix
;
;   ((1.0        0.0              0.0        )
;    (0.0 (aref matrix 1 1) (aref matrix 1 2))
;    (0.0 (aref matrix 2 1) (aref matrix 2 2)))
;
      (setq axis
            (if (and (< (abs (- (aref matrix 1 1) (aref matrix 2 2))) *EPSILON-ROTATION*)
                     (< (abs (- (abs (aref matrix 1 2)) (abs (aref matrix 2 1)))) *EPSILON-ROTATION*)
                     (or  (not (= 0 (aref matrix 1 2)))
                          ; then (= 0 (aref matrix 1 2)) is True !!
                          (= -1 (aref matrix 1 1))))
              '(0 0 1)
              '(0 0 0))))
     ((= 3 matrix-dim)
      (setq axis
            (list
             (if (and (= 0 (aref matrix 2 1))
                      (= 0 (aref matrix 3 1)))
               1
               0)
             (if (and (= 0 (aref matrix 1 2))
                      (= 0 (aref matrix 3 2)))
               1
               0)
             (if (and (= 0 (aref matrix 1 3))
                      (= 0 (aref matrix 2 3)))
               1
               0)))))
    (if (> (apply #'+ axis) 1)
      '(0 0 0) ; matrix is not a rotation!
      axis
      )))
;----------------------------------------------------------------------

(defun WRITE-OVERTURE-VRML-FILE (infostring outfile)
  (cond
   ((VRML-VER-1?)
    (format outfile "#VRML V1.0 ascii~%~%")
    (format outfile "Separator {~%")
    (format outfile "  Info { string \"~A\" }~%" infostring)
    (format outfile "  DEF BackgroundColor Info { string \".3 .7 .9\" }~%")
    (light-description outfile)
    (format outfile "
  PerspectiveCamera {
    position        0 0 4
    orientation     0 0 1  0
    focalDistance   10.84
  }~%"))
   ((VRML-VER-2?)
    (format outfile "#VRML V2.0 utf8~%~%")
    (format outfile "Group {~%")
    (format outfile " children [~%")
    (format outfile " WorldInfo {~%")
    (format outfile "   info [  \"~A\" ]~%" infostring)
    (format outfile " }~%")
    (format outfile "
  Background  { 
     skyColor    .3 .7 .9
     groundColor .3 .7 .9
  }~%")
    (light-description outfile)
    )))


(defun LIGHT-DESCRIPTION (outfile)
  (format outfile "
  SpotLight {      # Light from above
    location      0 1 0
    direction     0 -1 0
    intensity     1
    cutOffAngle   1.9
  }
  DirectionalLight {
    direction   0 0 -1  # Light shining from viewer into scene
    intensity   1.0
  }~%"))

(defun WRITE-CLOSURE-VRML-FILE (outfile)
  (cond
   ((VRML-VER-1?)
    (format outfile "}~%")
    (format outfile "#VRML 1.0 File End~%"))
   ((VRML-VER-2?)
    (format outfile " ]~%")
    (format outfile "}~%")
    (format outfile "#VRML 2.0 File End~%"))))

;-------------------------------------------------------------------------------

(defun WRITE-MATERIAL-APPEARANCE-NODE (outfile)
  (declare (special *GENERAL-MATERIAL-NODE*))
  (if (not (oiv-format?))
    (if (return-name)
      (cond
       ((appearance-just-written-for-this (return-name))
        (write-use-instance-material-appearance-node (return-name) outfile))
       (T 
        (set-as-appearance-just-written (return-name))
        (write-general-material-node (return-name) outfile)
        ))
      (if (appearance-just-written)
        (write-use-general-material-node outfile)
        (progn
          (write-general-material-node *GENERAL-MATERIAL-NODE* outfile)
          (set-appearance-just-written))))))
  
(setf *GENERAL-MATERIAL-NODE* "Generalmaterial")

(defun WRITE-USE-GENERAL-MATERIAL-NODE (outfile)
  (declare (special *GENERAL-MATERIAL-NODE*))
  (if (not (oiv-format?))
    (write-use-instance-material-appearance-node *GENERAL-MATERIAL-NODE* outfile)))

(defun WRITE-GENERAL-MATERIAL-NODE (general-name outfile)
  (if (not (oiv-format?))
    (cond
     ((vrml-ver-1?)
      (format outfile (ins-tab "" (tab-count) "DEF ~A Material {~%") general-name)
      (format outfile (ins-tab "" (tab-count) "  diffuseColor  0.8 0.8 0.8  # can be edited to change object color~%"))
      (format outfile (ins-tab "" (tab-count) "  specularColor 0.1 0.1 0.1  # idem~%"))
      (format outfile (ins-tab "" (tab-count) "}~%")))
     ((vrml-ver-2?)
      (format outfile (ins-tab "" (tab-count) "appearance DEF ~A Appearance {~%") general-name)
      (format outfile (ins-tab "" (tab-count) "  material Material {~%"))
      (format outfile (ins-tab "" (tab-count) "    diffuseColor 0.8 0.8 0.8~%"))
      (format outfile (ins-tab "" (tab-count) "    ambientIntensity 0.2~%"))
      (format outfile (ins-tab "" (tab-count) "    specularColor 0.0 0.0 0.0~%"))
      (format outfile (ins-tab "" (tab-count) "    emissiveColor 0.0 0.0 0.0~%"))
      (format outfile (ins-tab "" (tab-count) "    shininess 0.2~%"))
      (format outfile (ins-tab "" (tab-count) "    transparency 0.0~%"))
      (format outfile (ins-tab "" (tab-count) "  }~%"))
      (format outfile (ins-tab "" (tab-count) "}~%"))))))

(defun WRITE-USE-INSTANCE-MATERIAL-APPEARANCE-NODE (name outfile)
  (if (not (oiv-format?))
    (cond
     ((vrml-ver-1?)
      (format outfile (ins-tab "" (tab-count) "USE ~A ~%") name))
     ((vrml-ver-2?)
      (format outfile (ins-tab "" (tab-count) "appearance USE ~A ~%") name)))))

;-------------------------------------------------------------------------------






(defun WRITE-USE-NODE (name outfile)
  ;this node is VRML version independent
  (inc-tab-counter)
  (format outfile (ins-tab "" (tab-count) "USE ~A ~%") name)
  (dec-tab-counter))

;---------------------------------------------------------------------------

(defun WRITE-OVERTURE-VRML-STRUCTURE (name outfile &key (open-group nil) (plasm-name? nil))
  (inc-tab-counter)
  (cond
   ((VRML-VER-1?)
    (format outfile
            (ins-tab "" (tab-count) "DEF ~A Separator {~%")
            name)
    (if plasm-name? (WRITE-MATERIAL-APPEARANCE-NODE outfile)))
   ((and open-group (VRML-VER-2?))
    (format outfile
            (ins-tab ""
                     (tab-count)
                     "DEF ~A Group { children [~%") name))
   ((VRML-VER-2?)
    (format outfile
            (ins-tab "" (tab-count) "DEF ~A ") ; what follows this is dependent
                                               ; from matrix transform of Plasm node
            name))))

(defun WRITE-CLOSURE-VRML-STRUCTURE (name outfile &key (close-group nil))
  (cond
   ((VRML-VER-1?)
    (format outfile (ins-tab "" (tab-count) "} #~A~%") name))
   ((and close-group (VRML-VER-2?)) ; a Group children node have to be closed
    (format outfile (ins-tab "" (tab-count) "] } # ~A ~%") name))
   ((VRML-VER-2?) ; a Transform  children node have to be closed
    (format outfile (ins-tab "" (tab-count) "] } #~A~%") name)))
  (dec-tab-counter))

;---------------------------------------------------------------------------

(defun WRITE-TRANSFORMATION-VRML-STRUCTURE (mat rep outfile)
  (declare (special *N-SPACE*))
  (if (is-identity-mat mat)
    (if (vrml-ver-2?) (format outfile "Group {~%"))
    (let ((matrix (if (eql rep 'kernel::on-faces)
                    (mat-inv mat)
                    mat))
           (matrix-dim (1- (first (array-dimensions mat)))))
      (let* ((axis (take-rotational-axis matrix))
             (rot? (> (apply #'+ axis) 0)))
        (inc-tab-counter)
        (format outfile (ins-tab "" (tab-count) "Transform {~%"))
        (inc-tab-counter)
        ; translation parameters
        (let ((trans-par (list (aref matrix 1 0)
                               (if (>= matrix-dim 2)
                                 (aref matrix 2 0)
                                 0)
                               (if (= matrix-dim 3)
                                 (aref matrix 3 0)
                                 0))))
          (if (not (eval `,(cons 'and `,(mapcar #'zerop trans-par))))
            (format outfile
                    (ins-tab "" (tab-count) "translation ~{~,6F ~}~%")
                    trans-par)))

        ; scale parameters
        (if (not rot?)
          (let ((scale-pars (list (aref matrix 1 1)
                                  (if (>= matrix-dim 2)
                                    (aref matrix 2 2)
                                    1)
                                  (if (= matrix-dim 3)
                                    (aref matrix 3 3)
                                    1))))
            (if (not (eval `,(cons 'and `,(mapcar #'is-number-one scale-pars))))
              (cond
               ((vrml-ver-1?)
                (format outfile
                        (ins-tab "" (tab-count) "scaleFactor ~{~,6F ~}~%")
                        scale-pars))
               ((vrml-ver-2?)
                (format outfile
                        (ins-tab "" (tab-count) "scale ~{~,6F ~}~%")
                        scale-pars))))))

        ; rotation parameters
        (if rot?
          (progn
            (format outfile (ins-tab "" (tab-count) "rotation "))
            (format outfile " ~{~A ~} ~,6F ~%"  axis (take-alfa axis matrix))))
        (dec-tab-counter)
        (if (vrml-ver-1?) ; when it's writing for version 2.0 here we're in a
                          ; Transform node and it must open a children node in
                          ; in the following code
          (format outfile (ins-tab "" (tab-count) "}~%")))
        (dec-tab-counter)))))

;----------------------------------------------------------------------

(defun SET-RESET-ENVIRONMENT (&optional matrix-dim)
    (CREATE-HASH-TABLE-PLASM-NAME)
    (RESET-APPEARANCE-NO-WRITTEN)
    (RESET-NAMES)
    ;(RESET-CTM matrix-dim)
    (reset-tab-counter)
    (reset-hash-table-name))


(defun EXPORT-POL-COMPLEX-TO-VRML-FILE-FORMAT-FIRST
                          (pol-complex
                           &rest streams
                          )
  (if (or (not streams)
          (= 1 (length streams)))
    ; then streams stands for "export it on a file!"
   (let ((filename (or (first streams)
                                 (prompt-for-a-new-file 
                                                 "untitled" ".wrl"))))	
    (with-open-file (outfile filename 
                             :direction :output
                             :if-exists :supersede
                             :if-does-not-exist :create)
      (export-pol-complex-to-vrml-file-format pol-complex outfile filename)))
    ; then streams stands for "export it on standard output!"
     (let ((outstring (make-string-output-stream)))
        (export-pol-complex-to-vrml-string-stream-format  pol-complex outstring)
        (format t "~A" (get-output-stream-string outstring)))
    ))

;Please, you have to pay attention if changes on the following function are needed.: you have to
; modify accordingly its following function too.

(defun EXPORT-POL-COMPLEX-TO-VRML-FILE-FORMAT (pol-complex filestream filename)
  (let* ((*N-SPACE* (get-npol-complex pol-complex)))
    (declare (special *N-SPACE*))

    (format t "~%Filename is ~A~%" filename)

    (SET-RESET-ENVIRONMENT (1+ (get-npol-complex pol-complex)))
      (format t "~%Exporting~%   \"~A\" ~%  to VRML file ~%     \"~A\" ..~%"
              (plasm::print-plasm pol-complex) filename)
      (write-overture-vrml-file (pathname-name filename) filestream)

      (write-pol-complex-to-vrml  pol-complex filestream)
      (write-closure-vrml-file filestream)
      (format t "... done.~%")))

;Please, read the last comment above!

(defun EXPORT-POL-COMPLEX-TO-VRML-STRING-STREAM-FORMAT (pol-complex string-stream)
  (let* ((*N-SPACE* (get-npol-complex pol-complex)))
    (declare (special *N-SPACE*))
    (SET-RESET-ENVIRONMENT (1+ (get-npol-complex pol-complex)))
      
      (write-overture-vrml-file "standard output" string-stream)
      (write-pol-complex-to-vrml  pol-complex string-stream)
      (write-closure-vrml-file string-stream)
     )) 





;----------------------------------------------------------------------


(defun WRITE-POL-COMPLEX-TO-VRML (dag outfile &key (name-to-use nil)
                                                    (plasm-name? nil))
  (cond
   ((get-hash-table-name dag) ; if it returns T then this pol-complex was just reached
    (write-use-node (get-hash-table-name dag) outfile))
   ((polyhedra-complex? dag)
    (let ((name (set-as-just-seen dag 
                                  ;if dag is plasm named object name will be its plasm name
                                  
                                  :name name-to-use
                                  ;if dag comes from a skeleton operation name must be 
                                  ; the name of its origin
                                  :use-name-as-radix (not name-to-use))))

      (insert-into-names (get-hash-table-plasm-name dag))
      (if (= 1 (length dag)) ; dag is a pol-complex with a single pol node
        (write-pol-complex-to-vrml (first dag)
                                   outfile
                                   :name-to-use name
                                   :plasm-name? (get-hash-table-plasm-name dag))
        (progn
          (write-overture-vrml-structure name
                                         outfile
                                         :open-group T
                                         :plasm-name?
                                         (get-hash-table-plasm-name dag))
          (mapcar  #'(lambda (node)
                       (write-pol-complex-to-vrml
                        node
                        outfile))
                   dag)
          (write-closure-vrml-structure name outfile :close-group T)))
      (if (get-hash-table-plasm-name dag) 
        (remove-from-names))))

   ((polyhedral-node? dag)
    (let ((pol (get-node-pol-complex dag))
          (matrix (get-node-matrix dag))
          (node-rep (get-node-rep dag))
          (name (if name-to-use
                  (set-as-just-seen dag :name name-to-use) ; this names a polyhedral node
                  (set-as-just-seen dag
                                    :name 'PART
                                    :use-name-as-radix T))))
      (if (polyhedron? pol)
        (write-vrml-polyhedron-representation pol
                                              matrix
                                              name
                                              outfile)
       ;(polyhedra-complex? pol) predicate returns T
        (progn
          (write-overture-vrml-structure name
                                         outfile
                                         :open-group nil
                                         :plasm-name? plasm-name?)
          
          ;(set-current-transformation-matrix (if (eql node-rep 'kernel::on-faces)
          ;                                     (mat-inv matrix)
          ;                                     matrix))
          (write-transformation-vrml-structure matrix node-rep outfile)
          
          ;if it's writing for ver 2.0 after a transform it needs a children node opening
          (if (vrml-ver-2?) (format outfile (ins-tab "" (tab-count) "children [~%")))
          
          (write-pol-complex-to-vrml pol outfile)
          (write-closure-vrml-structure name outfile :close-group nil)
          ;this function closes the children node (only for V2.0)
          
          ;(set-current-transformation-matrix (if (eql node-rep 'kernel::on-faces)
          ;                                     matrix
          ;                                     (mat-inv matrix)))
          ))))))




;------------------------------------------------

(defun WRITE-TRANSFORMATION (dag outfile)
  (cond
   ((polyhedra-complex? dag)
    (mapcar  #'(lambda (node)
                 (write-transformation
                  node
                  outfile))
             dag))

   ((polyhedral-node? dag)
    (let ((pol (get-node-pol-complex dag))
          (matrix (get-node-matrix dag))
          (node-rep (get-node-rep dag))
          )
      (if (polyhedron? pol)
        nil
        (progn
          (write-transformation-vrml-structure matrix node-rep outfile)
          (write-transformation pol outfile)))))))





;---------------------------------------------------------------------------

;(defun get-pol-complex-vertices-and-faces (node old-mat vertices faces)




(defun WRITE-VRML-POLYHEDRON-REPRESENTATION (pol matrix name outfile)
  (let* ((dim (get-npol pol))
         (vertices (make-hash-table))
         (faces (make-hash-table))
         )
    (if (= 3 (get-dpol pol)) ; pol is solid!
      (let ((pol-complex (make-empty-polyhedra-complex )))
        (put-node-in-pol-complex (make-pol-node pol matrix)
                                 pol-complex)
        (setf pol-complex (skeleton  pol-complex 2))
        (if (symbol-equal (get-pol-complex-rep pol-complex) 'on-faces)
          (setf pol-complex (coerce-pol-complex pol-complex 'on-vertices)))
        (let* ((n (get-npol-complex pol-complex))
	       (id-mat (mat-identity (1+ n))))
          (mapcar #'(lambda (node)
		      (get-pol-complex-vertices-and-faces node id-mat vertices faces))
	          pol-complex)
          ;(format t ".")
          (write-overture-vrml-polyhedron name outfile)
          (write-vertices-faceset-vrml-format
           vertices
           faces
           2 ; in this case dimension is for sure 2!
           outfile)
          (write-closure-vrml-polyhedron outfile)))
      (let ((pol-rep (get-pol-rep pol)))
        (format t ".")
        (write-overture-vrml-polyhedron name outfile)
        (get-pol-vertices-and-convex-faces (if (eq pol-rep 'kernel::on-vertices)
                                      pol
                                      (coerce-polyhedron pol 'kernel::on-vertices))
                                    (if (eq pol-rep 'kernel::on-vertices)
                                      matrix
                                      (mat-inv matrix))
                                    vertices
                                    faces)
        (write-vertices-faceset-vrml-format vertices faces dim outfile)
        (write-closure-vrml-polyhedron outfile)))))


(defun WRITE-VRML-POLYHEDRON-REPRESENTATION (pol matrix name outfile)
  (let* ((dim (get-npol pol))
         (vertices (make-hash-table))
         (faces (make-hash-table))
         )
    (if (= 3 (get-dpol pol)) ; pol is solid!
      (let ((pol-complex (make-empty-polyhedra-complex )))
        (put-node-in-pol-complex (make-pol-node pol matrix)
                                 pol-complex)
        (setf pol-complex (skeleton  pol-complex 2))
        (if (symbol-equal (get-pol-complex-rep pol-complex) 'on-faces)
          (setf pol-complex (coerce-pol-complex pol-complex 'on-vertices)))
        (let* ((n (get-npol-complex pol-complex))
	       (id-mat (mat-identity (1+ n))))
          (mapcar #'(lambda (node)
		      (get-pol-complex-vertices-and-faces-vrml node id-mat vertices faces))
	          pol-complex)
          ;(format t ".")
          (write-overture-vrml-polyhedron name outfile)
          (write-vertices-faceset-vrml-format
           vertices
           faces
           2 ; in this case dimension is for sure 2!
           outfile)
          (write-closure-vrml-polyhedron outfile)))
      (let ((pol-rep (get-pol-rep pol)))
        (format t ".")
        (write-overture-vrml-polyhedron name outfile)

        (get-pol-vertices-and-convex-faces (if (eq pol-rep 'kernel::on-vertices)
                                      pol
                                      (coerce-polyhedron pol 'kernel::on-vertices))
                                    (if (eq pol-rep 'kernel::on-vertices)
                                      matrix
                                      (mat-inv matrix))
                                    vertices
                                    faces)

        (write-vertices-faceset-vrml-format vertices faces dim outfile)

        (write-closure-vrml-polyhedron outfile)))))

(defun WRITE-OVERTURE-VRML-POLYHEDRON (polname outfile)
  (cond
   ((vrml-ver-1?)
    (format outfile
            (ins-tab "" (tab-count) "DEF ~A Separator { #Polyhedron~%")
            polname)
    ;(if (oiv-format?) (write-shapehints-node outfile))
    (inc-tab-counter)
    (write-material-appearance-node outfile))
   ((vrml-ver-2?)
    (format outfile
            (ins-tab "" (tab-count) "DEF ~A Shape { #Polyhedron~%")
            polname)
    (inc-tab-counter)
    (write-material-appearance-node outfile)
    (format outfile (ins-tab "" (tab-count) "geometry~%"))))
  (inc-tab-counter))


(defun WRITE-CLOSURE-VRML-POLYHEDRON (outfile)
  (dec-tab-counter)
  (if (vrml-ver-2?) (dec-tab-counter))
  ;close a Separator/Shape node
  (format outfile (ins-tab "" (tab-count) "} #Polyhedron~%")))

(defun WRITE-POINT-FIELD (vertices outfile)
    (let* ((len (length vertices))
           (len-1 (1- len))
           (format-string (ins-tab "" (tab-count) "~{~,6F ~}")))
      (format outfile (ins-tab "" (tab-count) "point [~%"))

      (dotimes (i len)
        (format outfile
                format-string
                (subseq (append (nth i vertices) '(0 0 0)) 0 3))
        ; questa rende il vertice sempre triimensionale !!!
        (if (< i len-1)
          (format outfile ",~%")
          (format outfile "~%")))
      (format outfile (ins-tab "" (tab-count) "]~%"))))

(defun WRITE-COORDINDEX-NODE-1D (adj outfile)
    (let* ((len (length adj))
           (len-1 (1- len))
           (format-string (ins-tab "" (tab-count) "~{ ~A, ~} -1")))
      (format outfile (ins-tab "" (tab-count) "coordIndex [~%"))
      (dotimes (i len)
        (format outfile
                format-string
                (subseq             ;
                 (mapcar #'1- (nth i adj))
                 0                  ;
                 2
                 )                ;
                )
        (if (< i len-1)
          (format outfile ",~%")
          (format outfile "~%")))
      (format outfile (ins-tab "" (tab-count) "]~%"))))


(defun WRITE-COORDINDEX-NODE-1D (adj outfile)
    (let* ((len (length adj))
           (len-1 (1- len))
           (format-string (ins-tab "" (tab-count) "~{ ~A, ~} -1")))
      (format outfile (ins-tab "" (tab-count) "coordIndex [~%"))
      (dotimes (i len)
        (format outfile
                format-string
                (mapcar #'1- 
                        (subseq 
                         (nth i adj)
                         0
                         2)))
        (if (< i len-1)
          (format outfile ",~%")
          (format outfile "~%")))
      (format outfile (ins-tab "" (tab-count) "]~%"))))




(defun WRITE-COORDINDEX-NODE-2D (polygons outfile)
  (let* ((len (length polygons))
         (len-1 (1- len))
         (format-string (ins-tab "" (tab-count) "~{ ~A, ~} -1 ")))
    (format outfile (ins-tab "" (tab-count) "coordIndex [~%"))
    (dotimes (i len)
      (format outfile format-string (mapcar #'1- (nth i polygons)))
      (if (< i len-1)
        (format outfile ",~%")
        (format outfile "~%")))
    
    
    ;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    ; replica delle celle convesse bidimensionali per evitare le trasparenze
    (if (doubled-faces?)
      (progn
        (format outfile ", ")
        (dotimes (i len)
          (mapcar #'(lambda (number)
                      ;(format outfile "~T~T  ")
                      (format outfile " ~A, " number))
                  ;reverses the order of polygon vertices
                  (reverse (mapcar #'1- (nth i polygons))))
          (if (< i len-1)
            (format outfile " -1, ~%")
            (format outfile " -1 ~%")))))
    ;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    
    
    (format outfile (ins-tab "" (tab-count) "]~%"))))






(defun WRITE-COORDINDEX-NODE-2D (polygons outfile)
  (let* ((len (length polygons))
         (len-1 (1- len))
         (format-string (ins-tab "" (tab-count) "~{ ~A, ~} -1 ")))
    (format outfile (ins-tab "" (tab-count) "coordIndex [~%"))
    (dotimes (i len)
      (format outfile format-string (mapcar #'1- (nth i polygons)))
      (if (< i len-1)
        (format outfile ",~%")
        (format outfile "~%")))


;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; replica delle celle convesse bidimensionali per evitare le trasparenze
    (if (doubled-faces?)
      (progn
        (format outfile ", ")
        (dotimes (i len)
          (mapcar #'(lambda (number)
                      ;(format outfile "~T~T  ")
                      (format outfile " ~A, " number))
                  ;reverses the order of polygon vertices
                  (reverse (mapcar #'1- (nth i polygons))))
          (if (< i len-1)
            (format outfile " -1, ~%")
            (format outfile " -1 ~%")))))
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    (format outfile (ins-tab "" (tab-count) "]~%"))))


(defun WRITE-COORDINDEX-NODE-2D (polygons outfile)
  (let* ((len (length polygons))
         (len-1 (1- len))
         (format-string (ins-tab "" (tab-count) "~{ ~A, ~} -1 ")))
    (format outfile (ins-tab "" (tab-count) "coordIndex [~%"))
    (dotimes (i len)
      (format outfile format-string (mapcar #'1- (nth i polygons)))
      (if (< i len-1)
        (format outfile ",~%")
        (format outfile "~%")))
    
    
    ;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    ; replica delle celle convesse bidimensionali per evitare le trasparenze
    (if (doubled-faces?)
      (progn
        (format outfile (ins-tab "" (tab-count) ",#reverse faces~%"))
        (dotimes (i len)
	  (format outfile format-string 
		  (reverse (mapcar #'1- (nth i polygons))))
	   (if (< i len-1)
	       (format outfile ",~%")
	     (format outfile "~%")))))
    ;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    
    
    (format outfile (ins-tab "" (tab-count) "]~%"))))


(defun WRITE-VERTICES-FACESET-VRML-FORMAT (vertices faces dim outfile)
  (let* ((clear-w-rep (ripulisci-vrml (crea-w-rep vertices faces)))
         (vertices (first clear-w-rep))
         (adj (second clear-w-rep))
         (faceset-node-string (if (= 2 dim)
                                "IndexedFaceSet"
                                "IndexedLineSet"))
         )

         
    (cond
     ((vrml-ver-1?)
      (format outfile (ins-tab "" (tab-count) "Coordinate3 {~%"))
      (inc-tab-counter)
      (write-point-field vertices outfile)
      (dec-tab-counter)
      (format outfile (ins-tab "" (tab-count) "} #Coordinate3 ~%"))
      (format outfile (ins-tab ""
                               (tab-count)
                               faceset-node-string
                               " {~%"))
      (inc-tab-counter)
      (cond
       ((= 2 dim)
        (write-coordindex-node-2d  adj outfile))
       (T
        (write-coordindex-node-1d adj outfile)))

      (dec-tab-counter)
      (format outfile (ins-tab "" (tab-count)
                               "} #"
                               faceset-node-string
                               " ~%")))
     ((vrml-ver-2?)
      (format outfile (ins-tab ""
                               (tab-count)
                               faceset-node-string
                               " {~%"))
      (inc-tab-counter)
      (format outfile (ins-tab "" (tab-count) "coord Coordinate {~%"))
      (inc-tab-counter)
      (write-point-field vertices outfile)
      (dec-tab-counter)
      (format outfile (ins-tab "" (tab-count) "} #Coordinate ~%"))

      (cond
       ((= 2 dim)
        (write-coordindex-node-2d adj outfile))
       (T
        (write-coordindex-node-1d adj outfile)))

      (if (and 
           (= 2 dim)
           (not (doubled-faces?)))
        (format outfile (ins-tab "" (tab-count) "solid FALSE~%")))

      (dec-tab-counter)
      (format outfile (ins-tab "" (tab-count)
                               "} #"
                               faceset-node-string
                               " ~%"))))))
;-------------------------------------------------------------




; GET-POLYGONS receives a list of one or more of (v1 v2 v3 adj1 adj2 adj3)
; and returns a list of one or more (v1 v2 ... vn), that is a list of polygons.
;For every 6-upla it calls GET-CYCLE.

(defun GET-POLYGONS (triang-adjs)
  (declare (special triang-adjs))
  (let (polygons
        polygon
        (flagged (make-hash-table)))
    (declare (special flagged polygon))
    (dotimes (i (length triang-adjs) polygons) 
      (if (not (gethash (nth i triang-adjs) flagged))
        (progn
          (setf polygon '()) 
          (get-cycle 0 i)
          (setf polygons (cons (reverse polygon) polygons)))))))


;GET-CYCLE receives a pointer to the last visited triangle  and to the current one.
;If the current one triangle has just visited it exits,
;otherwise it flags the current triangle. Then if the current edge (current vertex with its next)
;is on external edge of the polyhedron GET-CYCLE writes in the variable polygons (non local 
;variable) the current vertex of current triangle, otherwise GET-CYCLE calls itself with previous
; equal to the index at the current triangle and i equal to its adjacent one 
;(found on its adj list).

(defun GET-CYCLE (previous i)
  (declare (special flagged triang-adjs polygon))
  (let (kk before)
    (if (not (gethash (nth i triang-adjs) flagged))          ; if not flagged
      (progn
        (setf (gethash (nth i triang-adjs) flagged) T)       ; flag it !
        ;     (make-a-permutation previous (nth i triang-adjs)) ; make a permutation?
        (setf before (find-kk-element-constructor previous (nth i triang-adjs)))
        (dotimes (j 3)
          (setf kk (mod (+ before j) 3))
          (if (= 0 (nth (+ 3 kk) (nth i triang-adjs)))
            (setf polygon (cons (nth kk (nth i triang-adjs)) polygon))
            (get-cycle i (1- (nth (+ 3 kk) (nth i triang-adjs))))))))))


(defun GET-CYCLE (previous i)
  (declare (special flagged triang-adjs polygon))
  (let (kk before)
    (if (not (gethash (nth i triang-adjs) flagged))          ; if not flagged
      (progn
        (setf (gethash (nth i triang-adjs) flagged) T)       ; flag it !
        ;     (make-a-permutation previous (nth i triang-adjs)) ; make a permutation?
        (setf before (find-kk-element-constructor previous (nth i triang-adjs)))
        (dotimes (j 3)
          (setf kk (if before
                     (mod (+ before (1- j)) 3)
                     j))
          (if (= 0 (nth (+ 3 kk) (nth i triang-adjs)))
            (setf polygon (cons (nth kk (nth i triang-adjs)) polygon))
            (get-cycle i (1- (nth (+ 3 kk) (nth i triang-adjs))))))))))




(defun FIND-KK-ELEMENT-CONSTRUCTOR (previous i-triang-adjs)
  (cond
  ; ((= previous (nth 3 i-triang-adjs))
  ;  0)
   ((= previous (nth 4 i-triang-adjs))
    1)
   ((= previous (nth 5 i-triang-adjs))
    2)
   (T 0)))

(defun FIND-KK-ELEMENT-CONSTRUCTOR (previous i-triang-adjs)
  (cond
   ((= previous (nth 3 i-triang-adjs))
    0)
   ((= previous (nth 4 i-triang-adjs))
    1)
   ((= previous (nth 5 i-triang-adjs))
    2)
   (T nil)))





(defun TAKE-DISTINCT-AND-ORDER-THEM (polygons-list)
  (sort (remove-duplicates (apply #'append polygons-list)) #'<))

(defun MAKE-HASHTABLE-OLD-NEW-INDEX (old-index-list)
  (let ((hashtable-old-new (make-hash-table )))
    (dotimes (i (length old-index-list) hashtable-old-new)
      (setf (gethash (nth i old-index-list) hashtable-old-new)
            (1+ i)))))

(defun CHANGE-OLD-INDEX-TO-NEW (polygons hashtable-old-new)
  (mapcar #'(lambda (polygon)
              (mapcar #'(lambda (key) (gethash key hashtable-old-new))
                      polygon))
          polygons))

(defun BUILD-NEW-POLYGONS (old-points-list polygons)
  (let* ((needed-vertices (take-distinct-and-order-them polygons))
         (hashtable-old-new-index (make-hashtable-old-new-index needed-vertices))
         new-points-list)
    (list
     (dotimes (i (length old-points-list) (reverse new-points-list))
       (if (gethash (1+ i) hashtable-old-new-index)
         (setf new-points-list (cons (nth i old-points-list) new-points-list))))
     (change-old-index-to-new polygons hashtable-old-new-index))))
      




;-------------------------------------------------------------------------------

(defun GET-POL-VERTICES-AND-CONVEX-FACES (pol mat vertices faces)
  (let* ((n (get-npol pol))
	 (nn (array-dimension mat 0))
	 (zeros (make-list (- nn n 1) :initial-element 0))
	 (vertices-list (get-vlist pol))
	 (cells-list (get-clist pol))
	 (vert-num (hash-table-count vertices))
	 (face-num (hash-table-count faces))
	 (verts-adiac (make-hash-table :test 'equal))
	 )
    
    (cond
     ((= n 0)
      (mapcar #'(lambda (cell)
		  (let* ((cell-vertices (get-cell-vertices cell))
			 (v1 (make-array 
			      nn
			      :initial-contents           
			      (append '(1)
				      (find-vcoord (first cell-vertices) vertices-list)
				      zeros)))
			 
			 )
		    (setf (gethash (incf vert-num) vertices)
			  (rest (coerce (normalize-omo-coord (mat-vet-product-extended
					 mat v1)) 'list)))
		    
		    (setf (gethash (incf face-num) faces) 
			  (list vert-num
				vert-num
				vert-num 0 0 0))))
	      cells-list))
     ((= n 1)
      (mapcar #'(lambda (cell)
		  (let* ((cell-vertices (get-cell-vertices cell))
			 (v1 (make-array 
			      nn
			      :initial-contents           
			      (append '(1)
				      (find-vcoord (first cell-vertices) vertices-list)
				      zeros)))
			 (v2 (make-array 
			      nn
			      :initial-contents           
			      (append '(1)
				      (find-vcoord (second cell-vertices) vertices-list)
				      zeros))))
		    (setf (gethash (incf vert-num) vertices)
			  (rest  (coerce (normalize-omo-coord (mat-vet-product-extended
					 mat v1)) 'list)))
		    (setf (gethash (incf vert-num) vertices)
			  (rest (coerce (normalize-omo-coord (mat-vet-product-extended
					 mat v2)) 'list)))
		    (setf (gethash (incf face-num) faces) 
			  (list vert-num
				(1- vert-num)
				vert-num 0 0 0))
		    
		    
		    ))
	      cells-list))
     
     ((= n 2)
      (mapcar #'(lambda (cell)
		    (let* ((old-vert-num (1+ vert-num))
			   (old-face-num (1+ face-num))
			   (cell-vertices (get-cell-vertices cell))
			   (cells-num (length cell-vertices))
			   (cell-vert-coord 
			    (mapcar #'(lambda (vid)
					(find-vcoord vid vertices-list))
				    cell-vertices))
			   (center (let ((x 0)(y 0))
				     (mapcar #'(lambda (xy)
						 (setf x (+ x (first xy))
						       y (+ y (second xy))))
					     cell-vert-coord)
				     (list (/ x cells-num) (/ y cells-num))))
			   (new-coord (mapcar #'(lambda (coords)
						  (list (- (first coords)
							   (first center))
							(- (second coords)
							   (second center))))
					      cell-vert-coord))
			   (phases (mapcar #'(lambda (coords)
					       (phase (complex (first coords)(second coords))))
					   new-coord))
			   (phases&coord (mapcar 'cons phases cell-vert-coord))
			   (sorted-coord 
			    (mapcar #'cdr (sort phases&coord #'< :key #'first)))
			   
			   
			   ; a questo punto i vertici della cella [ 2D !] sono ordinati in
			   ; base alla fase rispetto al centro della cella
			   
			   (world-coords
			    (mapcar #'(lambda (vet)   (normalize-omo-coord
					(mat-vet-product-extended
					 mat vet)))
				    (mapcar #'(lambda (coord)
						(make-array 
						 nn
						 :initial-contents           
						 (append '(1)
							 coord
							 zeros)))
					    sorted-coord)))
			   (world-center  (normalize-omo-coord (mat-vet-product-extended
					  
					  mat 
					  (make-array 
					   nn
					   :initial-contents           
					   (append '(1)
						   center
						   zeros)))))
			   )
		      ;Pone nella tab. hash le coordinate del centro della cella
		      ;(setf (gethash (incf vert-num) vertices)
			;    (rest (coerce world-center 'list)))
		      ;pone nella tavola hash tutti i vertici
		      (dotimes (i cells-num)
			(setf (gethash (incf vert-num) vertices)
			      (rest (coerce (nth i world-coords) 'list))))
		      
		      
		      ;ciclo di gestione delle adiacenze
		      
		      (dotimes (i cells-num)
			(setf (gethash (list
					(gethash (+ old-vert-num 1 (mod (+ i 1) cells-num)) vertices)
					(gethash (+ old-vert-num 1  i ) vertices)
					)
				       verts-adiac)
			      (list (1+ (+ i face-num)) 1))
			)
		      
		      ;pone nella tab. hash dei triangoli tutti i triangoli ottenuti da
		      ; punto centrale + facet. cells-num e' il numero di vertici.
                      (let ((polygon '()))
                        (dotimes (i cells-num)
                          (push (+ old-vert-num  i ) polygon)
			  )
                        (setf (gethash (incf face-num) faces) 
                              polygon
                              )                        )
		      ))
	      
              cells-list)))))


(defun GET-POL-COMPLEX-VERTICES-AND-FACES-VRML (node old-mat vertices faces)
  (let ((pol (get-node-pol-complex node))
	(matrix (get-node-matrix node)))
    (if (polyhedron? pol)
      (get-pol-vertices-and-convex-faces pol (mat-product-extended old-mat matrix)
                                  vertices faces)
      (mapcar  #'(lambda (node)
		 (get-pol-complex-vertices-and-faces-vrml ; ex ...-xxx 
		  node 
		  (mat-product-extended old-mat matrix)
		  vertices faces))
	      pol))))


(defun RIPULISCI-VRML (lista-w-rep) ;  it calls aggiorna-indici
  (let* ((triangoli (second lista-w-rep))
	 (vertici-ord-ind (sort (insert-indice (first lista-w-rep))
				'vert-min
				:key 'car))
	 (v (make-array (+ 1 (length vertici-ord-ind))
			:element-type 'integer)))
    (list (new-lista vertici-ord-ind v)
	  (aggiorna-indici-vrml triangoli v))))
;-----
(defun AGGIORNA-INDICI-VRML (triangoli v) 
  (let* ((half  (length (first triangoli)) )) ; deleted division by two
    (declare (special half))
    (mapcar #'(lambda (a-triangle)
		(append
		 (mapcar #'(lambda (index) (aref v index))
			 (subseq a-triangle 0 half))
		 (subseq a-triangle half)))
	    triangoli)))



(defun AGGIORNA-INDICI-VRML (polygons v) 
  (mapcar #'(lambda (a-triangle)
              (mapcar #'(lambda (index) (aref v index))
                      a-triangle))
              
	    polygons))
