; PLaSM Enhancer (I hope!) by Claudio Morgia 1998

; This patch enhances the PLaSM interpreter with a package

; system.

; Actually, it can be loaded over a previous PLaSM image memory

; without any modification to the underlying interpreter.

; At the CLisp prompt, issue the command '(load-file "scheme.lsp" :compiling T)'

; and the interpreter will be patched.

;

; Author:       Morgia Claudio

; E-Mail:       morgia@dia.uniroma3.it

; Developed on: Common Lisp Version 97.12.06 Revision 9

;               SuSE Linux 5.3

;               Intel Platform

;

; Status:       alpha

; Revision:     0.4

; Last-Modify:  4-11-1998

;

; Note: See the file documentation on how to use the new commands.

;       Currently tested only on Common Lisp, OS Linux, Intel Platform

; Bugs: Probably too much, currently on heavy test (sorry, my CLisp

;       doesn't have a debugger!)

; Docs: Currently, technical documentation is on the road.

 

(inpackage plasm)



; Dichiarazione delle variabili globali



(proclaim '(special **NAME-SPACES** **ACTUAL-SPACE** **OLD-TERM**

		    **SOCK** **STREAM-SOCK** **ACTUAL-BRIDGE**))



; Definizione valori iniziali delle variabili globali



(setq **NAME-SPACES** (make-hash-table)

      **ACTUAL-SPACE** nil

      **OLD-TERM** nil

      **ACTUAL-BRIDGE** 1

      **SOCK** nil

      **STREAM-SOCK** nil

      )



; Normalmente l'interprete PLaSM memorizza le definizioni delle funzioni,

; insieme ai loro valori precalcolati, in una tavola hash memorizzata nella

; variabile globale **USER-FENV**.

; Con questa modifica si introduce una nuova tavola hash **NAME-SPACES**

; indicizzata secondo il nome del particolare spazio dei nomi e contenente

; a sua volta strutture del tipo 'namespace' composte dai seguenti slot:

; - script  contiene una lista di stringhe, ognuna delle quali e' una linea

;           dello script che ha generato questo 'namespace'

; - space   e' la forma lisp che rappresenta la valutazione dello script di

;           cui sopra

; - exports contiene la lista dei simboli di questo spazio dei nomi che

;           vengono resi visibili all'esterno.



(defstruct namespace

  (script nil)

  (space (make-hash-table))

  (exports nil)

)



; Nome:      MAKE-EXT (string string) = string

; Status:    Internal

; Parametri: nome file in forma di stringa

;            ext  estensione da appendere o sostituire

; Ritorna:   Stringa con estensione sostituita



(defun MAKE-EXT (filename ext)

  (if (not (string-equal (pathname-type filename) ext))

      (setq filename (namestring 

		      (make-pathname 

		       :host (pathname-host  filename)

		       :device (pathname-device filename)

		       :directory (pathname-directory filename)

		       :name (pathname-name filename)

		       :type ext)

		      )

	    )

    )

  filename

)



; Nome:      PLASM-ERROR

; Sintassi:  PLASM-ERROR (number obj)

; Status:    Internal Patched

; Paramatri: code Codice di errore numerico

;            args Lista argomenti

;

; Introdotti nuovi codici di errore da 3000 a 3003

; Nota che il caso con 'code=T' non stampa il codice ma True essendo il codice=T

; probabilmente si voleva specificare un caso di default come con cond.

; La struttura con case comporta che per aggiungere nuovi codici di errore occorre

; modificare la funzione originale.



(defun PLASM-ERROR (code &rest args)

  (declare (special *DEBUGGING* **DEFINITION-EVALUATING**))

  #+:APPLE

  (window-select (find-window "listener"))



  (if **DEFINITION-EVALUATING** (reset-environment-definition))



  (case code

    ( 0 nil)

    ( 1 (format  t "~%~%Attempt to take functional value of symbol ~A !~%~%"

		 (print-plasm (first args))))

    ( 2 (format  t "~%~%Attempt to apply the function ~A to the function ~A !~

		~%Do you need COMPOSITION ?~%~%" 

		 (print-plasm (first args)) (print-plasm (second args))))

    ( 3 (format  t "~%~%~

		Affine transformation ~A needs a sequence of integers and ~

		a sequence of real.~%On the contrary, YOU have dispatched ~

		me ~A and ~A !~%~%" 

		 (print-plasm (first args)) (print-plasm (second args))

		 (print-plasm (third args))))

    ( 4 (format  t "~%~%Too much argument to function ~A !~%~%" 

		 (print-plasm (first args))))

    ( 5 (format  t "~%~%Function ~A does not exist !~%~%" 

		 (print-plasm (first args))))

    ( 6 (format  t "~%~%Attempt to take the ~A^ element of the sequence ~A !~%~%" 

		 (print-plasm (first args)) (print-plasm (second args))))

    ( 7 (format  t "~%~%Function ~A can't be applied to ~A argument !~%~%"

		 (print-plasm (first args)) (print-plasm (second args))))

    (10 (format  t "~%~%I can't bind ~A to ~A !~%~%"

		 (print-plasm (first args)) (print-plasm (second args))))

    (100 (format  t "~%~%CHECKER function failed for ~A~%~%" 

		  (print-plasm (first args))))

    (101 (format  t "~%~%Duplicated name in parameter sequence : ~A .~%~%" 

		  (print-plasm (first args))))

    (102 (format  t "~%~%Duplicated name in specifiers sequence : ~A .~%~%" 

		  (print-plasm (first args))))

    (103 (format  t "~%~%Duplicated name in default sequence : ~A .~%~%" 

		  (print-plasm (first args))))

    (104 (format  t "~%~%Duplicated name in local functions : ~A .~%~%" 

		  (print-plasm (first args))))

    (105 (format  t "~%~%There are names duplicated either in parameter(s) - ~A - ~

		 or local function(s) - ~A -. ~%~%" 

		  (print-plasm (first args))

		  (print-plasm (second args))))

    (106 (format  t "~%~%There exists name(s) for default parameter(s) ~A ~

		 without the matching parameter(s): ~A .~%~%" 

		  (print-plasm (first args)) (print-plasm (second args))))

    (200 (format  t "~%~%The argument ~A is not of type ~A .~%~%" 

		  (print-plasm (first args)) (print-plasm (second args))))

    ((201 202) (format  t "~%~%~A is not long ~A as declared in definition ~

			   of function ~A.~%~%" 

			(print-plasm (first args)) 

			(print-plasm (second args))

			(print-plasm (third args))))

    (300 (format  t "~%~%~A is a body for the function? I do not think at all?~%~%" 

		  (print-plasm (first args))))

    (301 (format  t "~%~%~A is a no good function name! Please rename your function.~%~%" 

		  (print-plasm (first args))))

    (600 (format  t "~%~%The function ~A cannot work on ~A as argument !~%~%"

		  (print-plasm (first args)) (print-plasm (second args)) ))

    (1000 (format  t "~%~%Function ~A can't be applied to ~A argument !~%~%"

		   (print-plasm (second args)) (print-plasm (first args))))

    (2000 (format  t "~%~%The argument ~A is not valid to STRUCT operator !~%~%"

		   (print-plasm (first args))))

    (2001 (format  t "~%~%The sequence argument to a STRUCT function has ~

		  polyhedra complexes of different dimensions!~%~%"))

    (2002 (format  t "~%~%The argument ~A is not valid for an export operation!~%~%"

		   (print-plasm (first args))))

    (3000 (format t "~%Argument to IMPORT or SYMEXP must be a list.~%"))

    (3001 (format t "~%The list that's the argument of IMPORT must be composed of symbols or list, the latter being composed of two elements: alias and package name.~%"))

    (3002 (format t "The list that's the argument of SYMEXP must be composed of symbols.~%")) 

    (3003 (format t "You're trying to reference the unexported (private) symbol '~A'~%" (first args)))

    (t (format  t "error code : ~A  with these args : ~A~%~%" code args))

    )



  (if *DEBUGGING* 

      (break "~A" """")

    (abort )

    )

)



; Nome:     READ-PLASM-SCRIPT

; Sintassi: READ-PLASM-SCRIPT (<STREAM>) = string

; Status:   Internal

; Parametri:Deve essere un input-stream

;

; Legge dallo stream linea per linea, concatena le linee separandole con un newline

; e le restituisce sotto forma di stringa.

; Notare l'uso della funzione 'reduce', normalmente usata per applicare operatori

; binari, qui usata tramite una forma lambda binaria appositamente costruita.

; Il costrutto di loop viene largamente usato.



(defun READ-PLASM-SCRIPT (stream )

  (let (

	(linea nil)

	(sep (make-string 1 :initial-element #\Newline))

	)

    (reduce #'(lambda 

		(x y) 

		(concatenate 'string x sep y) 

		)

	    (loop 

	      while (setq linea (read-line stream nil nil))

	      collect linea

	      )

	    )

    )

)



; Internal

; Obsolete

; Precompilatore, probabilmente ormai inutile.



(defun PSMCOMPILE (file)

  (let (

	(tmp-space **ACTUAL-SPACE**)

	)

    (new-space 'just-to-compile)

    (change-space 'just-to-compile)

    (with-open-file (ifile (make-ext file "psm") :direction :input)

		    (read-actual-space file nil :compiled nil))

    (with-open-file (ifile file :direction :output)

		    (save-actual-space file :compiled T))

    (change-space tmp-space)

    (rm-space 'just-to-compile)

    )

)



; Nome:     ALL-PACKS

; Sintassi: ALL-PACKS (stream) = T

; Status:   External

; Note:     Se stream=T esce su *standard-output*

;

; Produce sullo stream indicato la lista degli spazi dei nomi attualmente

; caricati sul server.



(defun ALL-PACKS (stream)

  (declare (special **NAME-SPACES**))

  (maphash #'(lambda (key val) (declare (ignore val)) (if key

				   (format stream "~A~%" key)

				 )

	       )

	   **NAME-SPACES**

	   )

  T

)



; Nome:     GET-HOSTNAME

; Sintassi: GET-HOSTNAME = string

; Status:   Internal

; Note:     Unix-Only

;

; Produce in uscita una stringa contenente il nome dell'host locale su cui gira

; questo processo.

; Valido solo su UNIX POSIX compliant.

; Note: esiste un modo piu' generale?



(defun GET-HOSTNAME ()

  (nth-value 0 (read-line (run-program "hostname" :output :stream )))

)



; Nome:     SPLIT-SCRIPT

; Sintassi: SPLIT-SCRIPT (string) = list of strings

; Status:   Internal

;

; La stringa rappresenta uno script PLaSM da separare per linee.

; Da notare che con linee non si intendono le singole linee di testo ma

; linee di comandi complete, ovvero una o piu' linee fino ad un punto e virgola.

; Funzione ricorsiva, costruisce la lista ponendo come primo elemento la prima

; linea del parametro locale 'script' e come resto l'esecuzione della funzione

; stessa su cio' che rimane della stringa.

; Notare l'uso del costrutto 'let*' che esegue le definizioni non in parallelo,

; utile quando un parametro dipende dal precedente.



(defun SPLIT-SCRIPT ( script )

  (let* (

	 (script (remove #\Newline script))

	 (pos (position #\; script))

	 )

    (if pos 

	(append (list (subseq script 0 (+ 1 pos)))

		(split-script (subseq script (+ 1 pos)))

		)

      nil

      )

    )

)



; Nome:     NAME-FROM-DEF

; Sintassi: NAME-FROM-DEF (string) = string | nil

; Status:   Internal

;

; Se la stringa inizia con 'DEF' (non preceduta da spazi, tornerebbero utili le

; espressioni regolari, il trattamento del testo in CLisp non e' dei migliori)

; estrae il nome della funzione in base al criterio che il nome e' compreso

; tra il primo spazio della stringa ('DEF nome') ed il carattere '='.

; Infine si elimina dal risultato eventuali spazi precedenti o seguenti.



(defun NAME-FROM-DEF (str)

  (if (string-lessp "DEF" str)

      (remove #\Space 

	      (subseq str 

		      (position-if #'alphanumericp str :start (position #\Space str)) 

		      (position #\= str)

		      )

	      )

    nil

    )

)



; Nome:     REMOVE-TRAILING-SPACES

; Sintassi: REMOVE-TRAILING-SPACE (string) = string

; Status:   Internal

;

; Elimina eventuali spazi presenti dopo il punto e virgola finale.

; Se non e' presente un punto e virgola finale, rimuove gli spazi in coda e ne

; inserisce uno (;) d'"ufficio".



(defun REMOVE-TRAILING-SPACES (str)

  (let (

	(pos (position #\; str :from-end T))

	)



    (if (not pos)

	(progn

	  (setq pos (position-if-not #'(lambda (x) (char-equal x #\Space)) 

				     str 

				     :from-end T))

	  (concatenate 'string (subseq str 0 (+ 1 pos)) ";")

	  )

      (subseq str 0 (+ pos 1))

      )

    )

)



; Nome:     PREPARE-ACTUAL-SPACE

; Sintassi: PREPARE-ACTUAL-SPACE = string

; Status:   Internal

;

; Come tutte le funzioni che contengono la parola ACTUAL nel nome, fa riferimento

; allo spazio dei nomi identificato dal valore della variabile globale **ACTUAL-SPACE**.

;

; Questa e' una funzione fondamentale poiche' prepara lo spazio attuale per essere

; trasmesso o salvato e lo fa processando le linee di testo che sono state inserite

; mano a mano.

; Fondamentalmente ne semplifica le componenti, eliminando definizioni ridondanti,

; fondendo le importazioni e le esportazioni e separando le definizioni dalle

; esecuzioni.

;

; La forma principale e' un ciclo che analizza le linee che compongono lo script:

; se sono "IMPORTazioni" ne estrae la lista dei pacchetti e la fa confluire in

; un'unica lista;

; se sono definizioni le raccoglie insieme;

; se sono "esportazioni" di simboli, le salta perche' verrano ricavate dall'apposita

; lista presente in ciascuna struttura di tipo 'namespace';

; ogni altra linea viene considerata una valutazione e confluisce nell'apposita lista.

;

; Nella clausola 'finally' vengono operati alcuni aggiustamenti come:

; - costruzione stringa finale per clausola "IMPORT"

; - rimozione definizioni duplicate e sovradefinite

; 

; All'uscita dal loop c'e' un uso massiccio dei costrutti 'reduce' e 'concatenate'

; per la generazione finale dello script che ha queste caratteristiche:

; - una sola linea (sempre iniziale, se presente) di direttive di importazione

; - definizioni senza duplicazioni

; - valutazioni senza duplicazioni

; - una sola linea (sempre finale, se presente) di direttive di esportazione di

;   simboli

;

; Questa struttura facilita eventuali parsing dal lato server e facilita a noi la

; gestione di questo gruppo di estensioni.



(defun PREPARE-ACTUAL-SPACE ()

  (declare (special **ACTUAL-SPACE** **NAME-SPACES**))

  (let* (

	(impstr (make-string 0))

	(defs '())

	(evals '())

	(space (gethash **ACTUAL-SPACE** **NAME-SPACES**))

	(script (namespace-script space))

	(exports (namespace-exports space))

	(tmpstr)

	)



    (loop for linea in script do

      (cond (

	     (string-lessp "IMPORT" linea)

	     (setq impstr (concatenate 'string 

				       impstr 

				       (subseq linea 

					       (+ 1 (position #\< linea)) 

					       (position #\> linea :from-end T)

					       )

				       ","

				       )

		   )

	     (remove linea script :test #'string-equal :count 1)

	     )



	    (

	     (string-lessp "DEF" linea)

	     (setq defs (append defs (list linea)))

	     )



	    (

	     (string-lessp "SYMEXP" linea)

	     )



	    (

	     T

	     (setq evals (append evals linea))

	     )

	    

	    )

      finally

      (if (> 0 (length impstr))

	  (setq impstr (concatenate 'string 

				    "IMPORT:<" 

				    (remove #\, impstr :from-end T :count 1 )

				    ">;~%"))

	)

      (remove-duplicates defs :test #'(lambda ( x y ) (string-equal (name-from-def x)

								    (name-from-def y)

								    )

					)

			 )

      )



    (setq impstr (concatenate 'string 

			      impstr

			      (reduce #'(lambda (x y ) (concatenate 'string x "~%" y)) defs :initial-value "")

			      (reduce #'(lambda (x y ) (concatenate 'string x y)) evals :initial-value "")))



    (setq tmpstr (reduce #'(lambda (x y ) (concatenate 'string x "," y)) exports :initial-value ""))



    (if (> (length tmpstr) 0)

	(setq impstr (concatenate 'string impstr "SYMEXP:<" tmpstr ">;"))

      )

    impstr

    )

)



; Nome:     CHECK-ACTUAL-SPACE

; Sintassi: CHECK-ACTUAL-SPACE = nil | T

; Status:   Internal

;

; Questa funzione esegue un controllo di integrita' dello spazio dei nomi attuale

; nell'unica maniera possibile per essere sicuri del controllo:

;

; - viene assemblato tramite 'prepare-actual-space' lo script equivalente che ha

;   generato il namespace attuale;

; - viene creato un namespace temporaneo con nome 'just-to-test-and-not-to-use'

;   riservato (collisioni difficili su questo nome :-) ), ci si trasferisce sullo

;   stesso e si valuta lo script risulato di 'prepare-actual-space'

; - se la forma lisp generata e' "EQuivalente" (nel senso LISP) a quella del

;   namespace da controllare ritorna T altrimenti nil

; - in ogni caso viene rimosso in namespace temporaneo



(defun CHECK-ACTUAL-SPACE ()

  (declare (special **USER-FENV**))

  (let (

	(script (prepare-actual-space))

	(oldspace **ACTUAL-SPACE**)

	(fenv)

	)



    (new-space 'just-to-test-and-not-to-use)

    (change-space 'just-to-test-and-not-to-use)

    (eval-plasm script)

    (setq fenv **USER-FENV**)

    (change-space oldspace)

    (prog1

	(equal **USER-FENV** fenv)

      (rm-space 'just-to-test-and-not-to-use)

      )

    )

)



; Nome:      SAVE-ACTUAL-SPACE

; Sintassi:  SAVE-ACTUAL-SPACE ( string &key bool bool) = nothing

; Status:    External

; Parametri: name-of-space nome dello spazio dei nomi (pacchetto)

;            readable      se la forma lisp generata deve essere leggibile

;                          da read

;            compiled      richiesta o meno, se possibile, il salvataggio,

;                          oltre allo script testuale, anche della forma

;                          precompilata

;

; Anche in questa occasione si usa 'prepare-actual-space' per ottenere uno

; script equivalente ma semplificato.

; Le estensioni standard per i files sono ".psm" per gli script e ".fen"

; per le forme precompilate.

; Eventuali files preesistenti con lo stesso nome verranno sostituiti

; senza alcun preavviso.

 

(defun SAVE-ACTUAL-SPACE (name-of-space &key 

					((:readable readable) T) 

					((:compiled compiled) nil))

  (declare (special **USER-FENV**))

  (let* (

	(old-var *print-readably*)

	(*print-readably* readable)

	(name-of-space (string-downcase name-of-space))

	(psmname (make-ext name-of-space "psm"))

	(fenname (make-ext name-of-space "fen"))

	)



    (with-open-file (outfile psmname

			     :direction :output

			     :if-exists :overwrite

			     :if-does-not-exist :create)

		    (format outfile (prepare-actual-space))

		    )



    (if compiled

	(with-open-file (outfile fenname

				 :direction :output

				 :if-exists :overwrite

				 :if-does-not-exist :create)

			(prin1 **USER-FENV** outfile)

			)

      )



    (setq *print-readably* old-var)

    )

)



; Nome:      READ-ACTUAL-SPACE

; Sintassi:  READ-ACTUAL-SPACE ( stream stream &key bool) = T

; Status:    Internal

; Parametri: psmstream     stream da cui prelevare lo script

;            fenstream     stream da cui prelevare la forma precompilata

;                          (se richiesto con 'compiled' e se possibile)

;            compiled      richiesta o meno, se possibile, di caricare anche

;                          la forma precompilata e quindi di NON procedere

;                          alla valutazione dello script

;

; Note:      fenstream puo' essere 'nil' se contestualmente e' stato passato 'nil'

;            come valore per 'compiled', ma se compiled e' non nil e fenstream e' nil

;            viene emesso un warning che evidenzia il fatto che si procedera'

;            comunque alla valutazione dello script.

;            Di per se non e' un errore ma e' una avvisaglia che le funzioni di

;            livello superiore hanno erroneamente settato 'compiled', anche se

;            non vi erano i presupposti: questo puo' precedere una perdita di

;            integrita' del sistema.

;

; Note2:     Il front-end naturale per questa funzione e' 'LOAD-PACK' oppure

;            'DOWNLOAD-ACTUAL-SPACE' secondo se il meccanismo di risoluzione e'

;            basato su filesystem o networked bridge.



(defun READ-ACTUAL-SPACE (psmstream fenstream &key ((:compiled compiled) nil))

  (declare (special **USER-FENV** ))

  (let (

	(script nil)

	)



    (if (and compiled (not fenstream))

	(progn

	  (format t "Warning: compiled script unavailable for loading...evaluating the real script~%")

	  (setq compiled nil)

	  )

      )

    

    (if (not compiled)

	(progn

	  (setq **USER-FENV** (make-hash-table))

	  (syn-tree-3 (read-plasm-script psmstream))

	  )

      (progn

	(setf **USER-FENV** (read fenstream)



	      script (read-plasm-script psmstream)



	      (namespace-script (gethash **ACTUAL-SPACE** **NAME-SPACES**))

	      (split-script script)

	      

	      script

	      (subseq script (+ (search "SYMEXP:" script) (length "SYMEXP:")))

	      

	      script 

	      (subseq script 0 (- (position #\; script) 1))

	      

	      (namespace-exports (gethash **ACTUAL-SPACE** **NAME-SPACES**))

	      (read (make-string-input-stream 

		     (map 'string #'(lambda (x) (cond (

						       (char-equal x #\<)

						       #\(

						       )

						      

						      (

						       (char-equal x #\>)

						       #\)

						       )

						      

						      (

						       (char-equal x #\,)

						       #\Space

						       )

						      

						      (

						       T

						       x

						       )

						      )

				      )

			  script

			  )

		     )

		    )

	      

	      )

	)

      )

    )

  T

)



; Nome:      NEW-SPACE

; Sintassi:  NEW-SPACE (string) = T 

; Status:    External

; Parametri: nome nome da assegnare al namespace da creare

;

; Con questa funzione si puo' creare un nuovo namespace, se non ne esiste gia' uno

; con lo stesso nome.



(defun NEW-SPACE (nome)

  (declare (special **NAME-SPACES**))

   (if (not (gethash nome **NAME-SPACES**))

        (setf (gethash nome **NAME-SPACES**) (make-namespace))

	)

  T

)



; Nome:      RM-SPACE

; Sintassi:  RM-SPACE (string) = nil | T 

; Status:    External

; Parametri: nome nome del namespace da eliminare

;

; L'eliminazione e' possibile se il namespace in questione non e' quello attuale

; oppure se esiste.



(defun RM-SPACE (nome)

  (declare (special **NAME-SPACES**))

  (if (eq nome **ACTUAL-SPACE**)

      nil

    (remhash nome **NAME-SPACES**)

    )

)



; Nome:      RENAME-SPACE

; Sintassi:  RENAME-SPACE (string string) = nil | T 

; Status:    External

; Parametri: oldnome nome del namespace da rinominare

;            newname nuovo nome da assegnare

;

; Note:      secondo l'implementazione LISP sottostante l'operazione puo' essere

;            leggera o anche molto pesante, secondo se l'interprete LISP esegue

;            le copie profonde o tramite puntatori.

;            Common Lisp su Intel procede per puntatori.



(defun RENAME-SPACE (oldname newname)

  (declare (special **NAME-SPACES**))

  (let (

	(old (gethash oldname **NAME-SPACES**))

	)

    (if old

	(progn

	  (setf (gethash newname **NAME-SPACES**) old)

	  (remhash oldname **NAME-SPACES**)

	  T

	  )

      nil

      )

    )

)



; Nome:      CHANGE-SPACE

; Sintassi:  CHANGE-SPACE (string) = nil | T 

; Status:    External

; Parametri: nome nome del namespace su cui portarsi, rendendolo corrente

; Note:      la sicurezza e la verifica dei permessi viene svolta dal bridge

;            e non dal lato CLisp/PLaSM

;

; Il cambio di contesto viene eseguito salvando l'attuale **USER-FENV** nello

; slot della struttura 'namespace' dello spazio che si sta abbandonando,

; sostituendolo con la forma lisp dello spazio obiettivo, aggiornando anche la

; variabile globale **ACTUAL-SPACE**.



(defun CHANGE-SPACE (nome)

  (declare (special **ACTUAL-SPACE** **USER-FENV** **NAME-SPACES**))

  (if (gethash nome **NAME-SPACES**)

      (progn

	(if **ACTUAL-SPACE**

	    (setf (namespace-space (gethash **ACTUAL-SPACE** **NAME-SPACES**)) **USER-FENV**)

	  )

	(setf **USER-FENV** (namespace-space (gethash nome **NAME-SPACES**))

	      **ACTUAL-SPACE** nome

	      )

	T

	)

    nil

    )

)



; Nome:      ACTIVE-SERVER

; Sintassi:  ACTIVE-SERVER ( string number) = T

; Status:    External

; Parametri: network stringa contenente l'indirizzo IP di dominio (secondo il

;                    netmask 255.255.255.0)

;            stdport porta a cui connettersi

;

; Questa funzione attiva la modalita' server attiva.

; E' costituita da un loop continuo di tentativi di connessione a tutti gli hosts

; di un dominio alla ricerca di un bridge PLaSM/SHastra attivo.

; Se lo trova inizializza l'ambiente per passare il controllo al bridge remoto.

; Il loop puo' essere interrotto solo dall'utente (se il sistema e' presidiato)

; visto che anche la console viene controllata dal bridge.



(defun ACTIVE-SERVER (&optional network stdport)

  (declare (special **STREAM-SOCK** **OLD-TERM** **ACTUAL-BRIDGE**))

  (let (

	(network (if network network "127.0.0."))

	(stdport (if stdport stdport 49999))

	(valid nil)

	)

    (loop until valid do

      (format t "~A~D~%" network **ACTUAL-BRIDGE**)

      (setq **STREAM-SOCK** (ignore-errors 

			      (socket-connect stdport 

					      (format nil 

						      "~A~D" 

						      network 

						      **ACTUAL-BRIDGE**)

					      )

			      )

	    )

      (if **STREAM-SOCK**

	  (setq **OLD-TERM** *terminal-io*

		*terminal-io* **STREAM-SOCK**

		valid T

		)

	(setq **ACTUAL-BRIDGE** (if (> **ACTUAL-BRIDGE** 254)

				    1

				  (+ **ACTUAL-BRIDGE** 1)

				  )

	      )

	)

      )

    )

  T

)



; Nome:      PASSIVE-SERVER

; Sintassi:  PASSIVE-SERVER ( number) = T

; Status:    External

; Parametri: stdport porta su cui attendere

;

; Questa funzione attiva la modalita' server passiva.

; In questo caso ci si mette in attesa perenne di una connessione proveniente

; da un bridge PLaSM/Shastra che assuma il controllo del server.

; L'attesa puo' essere interrotta solo dall'utente (se il sistema e' presidiato)

; visto che il sistema entra in 'socket wait'.



(defun PASSIVE-SERVER ( &optional stdport )

  (declare (special **OLD-TERM** **SOCK** **STREAM-SOCK**))

  (setq **SOCK** (socket-server

		  (if stdport

		      stdport

		    49999

		    )

		  )

	)

  (socket-wait **SOCK**)

  (setq **STREAM-SOCK** (socket-accept **SOCK**))

  (setq **OLD-TERM** *terminal-io*

	*terminal-io* **STREAM-SOCK**

	)

  T

)



; Nome:      STOP-SERVER

; Sintassi:  STOP-SERVER (&key bool) = nil | T

; Status:    External

; Parametri: quit flag che indica se, una volta chiusa la connessione, e' richiesto

;                 lo shutdown del server LISP.

;

; Questa funzione, normalmente invocata dal bridge che controlla il server,

; chiude la connessione e disabilita' la modalita' server, riattivando di fatto 

; la normale interazione via console.

; Se non era attiva alcuna connessione remota, ritorna 'nil' altrimenti procede

; alla chiusura e ritorna T.



(defun STOP-SERVER ( &key ((:quit quit) nil) )

  (declare (special **OLD-TERM**

		    **SOCK** 

		    **STREAM-SOCK**))

  (if **STREAM-SOCK**

      (progn

	(close **STREAM-SOCK**)

	(if **SOCK** (socket-server-close **SOCK**))

	(setq *terminal-io*  **OLD-TERM**)

	(if quit (quit) T)

	)

    nil

    )

)



; Nome:      UPLOAD-ACTUAL-SPACE

; Sintassi:  UPLOAD-ACTUAL-SPACE (bool) = nil | T

; Status:    External

; Parametri: readable richiede il trasferimento della forma lisp adatto al successivo

;                     caricamento da parte della funzione 'read'

;

; Questa funzione trasferisce lo spazio dei nomi attuale verso il bridge, che provvedera'

; a smistarlo verso un server DB.

; La trasmissione segue un unico protocollo molto semplice:

; <lunghezza dello script><newline><script><newline><forma lisp>



(defun UPLOAD-ACTUAL-SPACE (&key ((:readable readable) nil))

  (declare (special **ACTUAL-SPACE** **NAME-SPACES** **USER-FENV** **STREAM-SOCK**))

  (let (

	(old-var *print-readably*)

	(script)

	)

    (if **STREAM-SOCK**

	(progn

	  (setq *print-readably* readable

		script (prepare-actual-space 

			(namespace-script 

			 (gethash **ACTUAL-SPACE** **NAME-SPACES**)))

		)

	  (if (= (length script) 0)

	      (progn

		(prin1 "Hey guy, are you joking me?")

		nil

		)

	    (progn

	      (format **STREAM-SOCK** "~D~%~S~%~S" (length script) script **USER-FENV**)

	      (setq *print-readably* old-var)

	      T

	      )

	    )

	  )

      nil

      )

    )

)



; Nome:      DOWNLOAD-ACTUAL-SPACE

; Sintassi:  DOWNLOAD-ACTUAL-SPACE (bool) = bool

; Status:    External

; Parametri: compiled indica se il bridge ha a disposizione una forma lisp precompilata

;                     di cui dovremo aspettarci la trasmissione



(defun DOWNLOAD-ACTUAL-SPACE ( &key ((:compiled compiled) nil ))

  (declare (special **STREAM-SOCK**))

  (let (

	fenstream

	)

    (if compiled

	(setq fenstream 

	      (make-string-input-stream (format nil "~A" (read **STREAM-SOCK**))))

      )

    (read-actual-space **STREAM-SOCK** fenstream :compiled compiled)

    )

)



; Nome:      CALLED

; Sintassi:  CALLED (string) = tree

; Status:    Internal

; Parametri: nome nome del simbolo da cui ricavare il grafo delle dipendenze

;

; Questa funzione genera la lista di tutte le funzioni da cui dipende la funzione

; passata per parametro, in forma di grafo.

; E' una funzione interna che viene usata da PRINT-TO.



(defun CALLED (nome)

  (declare (special **USER-FENV**))

  (let (

	(risultato (list nome))

	(chiamate (plasm-function-called-function (gethash nome **USER-FENV**)))

	)

    (if chiamate

	(setq risultato (union risultato (mapcar #'called chiamate)))

      )

  risultato

  )

)



; Nome:      FLATTEN

; Sintassi:  FLATTEN (tree) = list

; Status:    Internal

; Parametri: lista e' una lista di liste e simboli che rappresenta il grafo delle

;                  dipendenze

;

; Questa funzione ignora la natura del grafo, serve solo per trasformare (appiattire)

; una lista di liste e simboli in una lista di soli simboli.

; E' una funzione interna usata da PRINT-TO.



(defun FLATTEN (lista)

  (let (

	(parziale '())

	(testa nil)

	)

    (if lista

	(loop

	  (setq testa (car lista) lista (cdr lista))

	  (if (eq testa nil) (return))

	  (setq parziale (union parziale (if (symbolp testa) (list testa) (FLATTEN testa)))

		)

	  )

      )

    parziale

    )

)



; Nome:      PRINT-TO

; Sintassi:  PRINT-TO (string stream)

; Status:    Internal

; Parametri: nome   e' il nome del simbolo 'root' da cui far partire l'analisi del grafo

;            stream e' lo stream di uscita per il salvataggio degli spazi dei nomi

; Note:      deve essere sottoposta a massiccia revisione in quanto non fa uso della nuova struttura

;            dati



(defun PRINT-TO (nome file)

  (declare (special **USER-FENV**))

  (let (

	(lista (FLATTEN (CALLED nome)))

	)

    (mapcar #'(lambda (x) (prin1 (gethash x **USER-FENV**) file)) lista)

    )

)



; Nome:      SAVE-DEF

; Sintassi:  SAVE-DEF (string string bool bool) = bool

; Status:    External

; Parametri: nome-def      nome della definizione da cui far partire il salvataggio del grafo

;                          delle dipendenze

;            filename      nome del file su cui salvare le definizioni

;            readable      deve essere rileggibile tramite 'read'

;            dont-traverse dobbiamo realmente effettuare il traversal oppure no?

; Note:      deve essere sottoposta a massiccia revisione in quanto non fa uso della nuova struttura

;            dati



(defun SAVE-DEF (nome-def filename &key ((:readable readable) T) ((:dont-traverse dont-traverse) nil))

  (declare (special **USER-FENV**))

  (let (

	(old-var *print-readably*)

	(filename (string-downcase (make-ext filename "def")))

	(*print-readably* readable)

	(res nil)

	)

    (setq res

	  (with-open-file (outfile filename

				   :direction :output

				   :if-exists :overwrite

				   :if-does-not-exist :create)

			  (if (listp nome-def)

			      (if dont-traverse

				  (mapcar #'(lambda (x) (prin1 (gethash x **USER-FENV**) outfile)) nome-def)

				(mapcar #'(lambda (x) (print-to x outfile)) nome-def)

				)

			    (if dont-traverse

				(prin1 (gethash nome-def **USER-FENV**) outfile)

			      (print-to nome-def outfile)

			      )

			    )

			  )

	  )

    (setq *print-readably* old-var)

    res)

)



; Nome:      READ-DEF

; Sintassi:  READ-DEF (string) = bool

; Status:    External

; Parametri: filename      nome del file da cui caricare le definizioni

; Note:      deve essere sottoposta a massiccia revisione in quanto non fa uso della nuova struttura

;            dati



(defun READ-DEF (filename)

  (declare (special **USER-FENV** ))

  (setq filename (make-ext filename "def"))

  (with-open-file (infile filename

			  :direction :input)

		  (let (

			(da-inserire nil)

			)

		    (loop

		     (setq da-inserire (read infile nil nil))

		     (if da-inserire

			 (insert-in-user-fenv da-inserire)

		       (return)

		       )

		     )

		    )

		  )

)



; Nome:      RESOLVE-NAME

; Sintassi:  RESOLVE-NAME

; Status:    Internal

;

; Questa funzione implementa il meccanismo di risoluzione degli spazi dei nomi tramite

; accesso ad un DB Server



(defun RESOLVE-NAME ()

  (download-actual-space)

)



; Nome:      RESOLVE-NAME

; Sintassi:  RESOLVE-NAME (string)

; Status:    Internal

; Parametri: nome Rappresenta il nome del package da risolvere

;

; Questa funzione implementa il meccanismo di risoluzione degli spazi dei nomi tramite

; file system locale.

; Se non e' presente un file con nome <nome>.psm produce un errore.

; Se non e' presente un file con nome <nome>.fen oltre a caricare lo script testuale,

; procede anche alla creazione della forma lisp e, se lo script contiene istruzioni

; di valutazione, procede alla valitazione ed, eventualmente, al calcolo dei complessi.

;

; Questa funzione e' un front-end verso la funzione 'READ-ACRUAL-SPACE' che e' quella

; che procede effettivamente al caricamento del/i file(s).



(defun RESOLVE-NAME (nome)

  (let (

	(psmstream (open (make-ext nome "psm") 

			 :direction :input 

			 :if-does-not-exist nil))

	(fenstream (open (make-ext nome "fen") 

			 :direction :input

			 :if-does-not-exist nil))

	(compiled T)

	)

 

    (if (not psmstream)

	(progn

	  (format t "Error: can't load the script")

	  (return-from resolve-name nil)

	  )

      )



    (if (not fenstream)

	(setq compiled nil)

      )



    (read-actual-space psmstream fenstream :compiled compiled)

    )

)



; Nome:      CREATE-FUNCTOR

; Sintassi:  CREATE-FUNCTOR (list)

; Status:    Internal

; Parametri: arg Deve essere una lista di due simboli: nome dell'alias e nome del package

;

; Questa funzione crea un FUNTORE ovvero una funzione PLaSM che effettua calcoli ma si occupa

; di scambiare temporaneamente gli spazi dei nomi per permettere la valutazione di simboli

; presente in spazi diversi da quello attuale.

;

; Viene usata sia per creare funtori veri e propri, quando in una clausola IMPORT si specifica

; una lista al posto del singolo simbolo, sia quando si genera uno schema di esportazione.

; Nel primo caso si ha

;

; IMPORT:<DesignParameters,<p1,FunctionsToolbox>>;

;

; Il secondo elemento e' una lista che definisce un funtore 'p1' da usare per accedere allo

; spazio dei nomi 'FunctionsToolbox'.

; In questo caso non si usano gli schemi di esportazione.

;

; Per una discussione piu' approfondita su importazione per schemi e/o per funtori, vedere

; la documentazione allegata.



(defun CREATE-FUNCTOR (arg)

  (let (

	(str (make-plasm-function-from-simple-cons

			(list (first arg) '(symb-) `(dofunctor ',(second arg) symb-))

			))

	)

    (setf (plasm-function-file str) (second arg))

    (insert-in-user-fenv str)

    )

)



; Nome:      NEW-TOKEN

; Sintassi:  NEW-TOKEN

; Status:    Internal

; Parametri: nessuno

;

; Questa funzione interna si occupa di inserire due nuove primitive del linguaggio PLaSM,

; IMPORT e SYMEXP.

;

; IMPORT si usa per importare altri spazi dei nomi mentre SYMEXP si usa per dichiarare

; uno schema di esportazione.



(defun NEW-TOKEN ()

  (declare (special **DEFAULT-FENV**))

  (INSERT-IN-DEFAULT-FENV 

   (make-plasm-function-from-simple-cons 

    '(IMPORT (LIST-) (DOIMPORT LIST-))

    )

   )

  (INSERT-IN-DEFAULT-FENV

   (make-plasm-function-from-simple-cons

    '(SYMEXP (LIST-) (DOSYMEXP LIST-))

    )

   )

  T

)



; Nome:      EVAL-APPLIC

; Sintassi:  EVAL-APPLIC (list)

; Status:    Internal Patched

; Parametri: tree

;

; Questa funzione e' identica all'originale PLaSM tranne per le seguenti modifiche:

; l'ordine di valutazione della parte sinistra e della parte destra di una applicazione

; e' stato invertito ma solo per quanto riguarda il codice LISP: dal punto di vista del

; PLaSM non e' cambiato niente.

; Questa modifica si e' resa necessaria per l'uso dei funtori: un costrutto con un funtore

; e' sempre del tipo <funtore>:<simbolo> e nella versione precedente veniva valutato

; prima <simbolo> di <funtore>, causando un errore del tipo 'Attempt to take functional

; value of symbol <simbolo>'.

;

; Analogo discorso per quanto riguarda IMPORT e SYMEXP: l'inteprete cercava di valutare

; come funzioni i nomi dei package, fallendo.

; Nel caso di queste due istruzioni, si risolve disabilitando temporaneamente il valutatore

; facendogli credere che ci troviamo in una DEFinizione, in modo che non cerchi di valutare

; i simboli.

;

; Ovviamente se siamo in una definizione viene prodotto il codice necessario all'attuazione

; del funtore mentre se ci troviamo in una valutazione questo codice viene effettivamente

; eseguito.

;

; Note: proprio il codice di valutazione del funtore e' instabile nell'interfacciamento

; con il valutatore PLaSM, da rivedere.



(defun EVAL-APPLIC (tree )

  (declare (special **DEFINITION-EVALUATING** **USER-FENV**))

  (let* (

	 (expres2-value (eval-expres (pick-up-son-num 1 tree)))

	 (expres1-value nil)

	 (de **DEFINITION-EVALUATING**)

	 fenv

	 tmp-space

	 space

	 )



    (if (or (eq expres2-value 'IMPORT) (eq expres2-value 'SYMEXP))

	(progn

	  (setq **DEFINITION-EVALUATING** T

		expres1-value (eval-expres (pick-up-son-num 3 tree))

		**DEFINITION-EVALUATING** de)

	  (reduce-function-args expres2-value expres1-value)

	  )

      (progn

	(setq fenv (gethash expres2-value **USER-FENV**)

	      tmp-space **ACTUAL-SPACE**

	      space (if fenv (plasm-function-file fenv) nil)

	      )

	(if space

	    (progn

	      (setq expres1-value(second (first (second (pick-up-son-num 3 tree))))) 

	      (if (definition-p)

		  `(dofunctor ',space ',expres1-value)

		(dofunctor space expres1-value)

		)

	      )

	  (progn

	    (setq expres1-value (eval-expres (pick-up-son-num 3 tree)))

	    (reduce-function-args expres2-value expres1-value)

	    )

	  )

	)

      )

    )

)



; Nome:      DOFUNCTOR

; Sintassi:  DOFUNCTOR (symbol symbol) = (dipende dalle funzioni PLaSM)

; Status:    Internal

; Parametri: space indica verso quale spazio dei nomi portarsi

;            symb  indica il simbolo da valutare nello spazio destinazione

;

; Questa funzione e' il cuore di tutto il sistema.

; Nel tentare di valutare un simbolo, se questo non esiste procede al caricamento del

; pacchetto, realizzando di fatto un late binding.

; In ogni caso procede poi alla valutazione del simbolo.

;

; Per maggiori informazioni consultare la documentazione.



(defun DOFUNCTOR ( space symb )

  (declare (special **USER-FENV** **ACTUAL-SPACE**))

  (let (

	(tmp-space **ACTUAL-SPACE**)

	(res nil)

	(funstr nil)

	(pars nil)

	)

    (if (not (change-space space))

	(load-pack space))

    

    (change-space space)



    (if (and (gethash symb **USER-FENV**)

	     (not (member symb (namespace-exports (gethash space **NAME-SPACES**)))))

	(progn

	  (change-space tmp-space)

	  (plasm-error 3003 symb)

	  )

      )



    (setf res

	  (progn

	    (setq funstr (gethash symb **USER-FENV** )

		  pars (plasm-function-formal-pars funstr)

		  )

	    (if pars

		(set-anonymous-function-value funstr) 

	      (set-and-or-return-value symb)

	      )

	    )

	  )

    (change-space tmp-space)

    res

    )

)



; Nome:      ATOMIC-P

; Sintassi:  ATOMIC-P (obj) = bool

; Status:    Internal Patched

; Parametri: obj E' l'oggetto PLaSM da identificare.

;

; Serve per decidere se l'oggetto in questione e' atomico (non ulteriormente processabile).

; Ho aggiunto la condizione che esprime l'atomicita' di un oggetto FUNTORE.



(defun ATOMIC-P (obj)  

  (declare (special flat-parameter-names local-function-names))

  (or (and (listp obj)

	   (eq (first obj) 'DOFUNCTOR)

	   )

      (and (atom obj)

	   (not (symbolp obj))

	   )

      (and (consp obj)

	   (equal 'reduce-function (first obj))

	   )

      (and (boundp 'flat-parameter-names)

	   (member obj flat-parameter-names))

      )

)



; Nome:      EVAL-PLASM

; Sintassi:  EVAL-PLASM (string)

; Status:    Internal Patched

; Parametri: plasm-string Stringa contenente lo script da processare

;

; L'unica modifica effettuata riguarda il ripristino dello spazio LISP utente in caso di errore.

; Prima rimaneva nello space PLASM, confondendo il resolver dei simboli.



(defun EVAL-PLASM (plasm-string)

  (declare (special *PLASM-STRING*))

  (WITHOUT-FLOATING-POINT-UNDERFLOW

   (with-simple-restart

    (abort " PLASM ERROR! " )

    (let (

	  (old-package *package*)

	  )

      (setf *PLASM-STRING* plasm-string)

      (inpackage  plasm)

      (syn-tree-3  *PLASM-STRING*)

      (setf *package* old-package)

      )

    )

   (inpackage user)

   )

  (values)

)



; Nome:      SYN-TREE-3

; Sintassi:  SYN-TREE-3 (string)

; Status:    Internal Patched

; Parametri: script Lo script da valutare in forma di stringa.

;

; L'unica modifica apportata (come indicato nel codice) serve per aggiungere ogni stringa

; immessa nell'apposita lista della struttura NAMESPACE.

; La memorizzazione dello script testuale avviene durante l'immissione e non con un reverse

; engineering, cosa che tra l'altro non so se sarebbe possibile in LISP.

;

; Non e' stato alterato minimamente il funzionamento precedente dell'interprete.



(defun SYN-TREE-3 (script)

  (declare (special *PLASM-READTABLE* **ACTUAL-SPACE** 

		    **NAME-SPACES**))

  (let (

	(punt 0)

        (end (length script))

        *aut-answer*

	oldpunt

	)

    (setq *READTABLE* *PLASM-READTABLE*)

    (loop

      (if (or (not punt) ; punt does not point to anything !

              (and (setq punt (all-not-important-char script punt end))

                   nil) ; the AND always return NIL

              (= punt end)

	      )

	  (return)

	)

      (reset-pred-stack) ; it has a new top level plasm expression

      (setq *aut-answer* (automa script punt end))

      (if (eq 'automa-error (first *aut-answer*))

	  (there-was-an-error *aut-answer*)) ; this call abort the evaluation

; inizio area sottoposta a modifica

      (setq oldpunt punt)

      (setq punt (second *aut-answer*))

      (setf (namespace-script (gethash **ACTUAL-SPACE** **NAME-SPACES**)) 

	    (append (namespace-script (gethash **ACTUAL-SPACE** **NAME-SPACES**)) 

		    (list (subseq script oldpunt punt))))

; fine area sottoposta a modifica

      (eval-script (first (second (first

				   (build-tree-3 (first *aut-answer*)))))))

    (setq *READTABLE* (copy-readtable nil))

    )

)



; Nome:      DOIMPORT

; Sintassi:  DOIMPORT (obj)

; Status:    Internal

; Parametri: arg Simbolo o lista con due simboli

;

; Questa funzione esegue la vera importazione, discerndendo tra funtori e schemi d'importazione.

;

; Nel caso dei simboli, procede al caricamento dello spazio dei nomi e produce uno schema per

; i simboli esportabili.

;

; Nel caso di lista, crea un funtore adeguato e non risolve fisicamente il pacchetto, fino al primo

; utilizzo del funtore (late binding).



(defun DOIMPORT ( arg )

  (cond (

	 (eq arg nil)

	 T

	 )



	(

	 (not (listp arg)) 

	 (plasm-error 3000) 

	 ) 

	

	(

	 T

	 (every #'(lambda (item) (cond (

				       (symbolp item)

				       (load-pack item)

				       (get-scheme item)

				       )

				      

				      (

				       (listp item)

				       (create-functor item)

				       )

				      

				      (

				       T

				       (plasm-error 3001)

				       nil

				       )

				      )

		   )

		arg

		)

	 )

	)

  T

)



; Nome:      DOSYMEXP

; Sintassi:  DOSYMEXP (list)

; Status:    Internal

; Parametri: arg Lista dei simboli da esportare.

;

; La lista dei simboli da esportare viene processata e memorizzata nell'apposito slot della

; struttura NAMESPACE.



(defun DOSYMEXP (arg)

  (declare (special **ACTUAL-SPACE** **NAME-SPACES**))

  (if (not (listp arg))

      (plasm-error 3000))

  (if (not (every #'symbolp arg))

      (plasm-error 3002)

    (setf (namespace-exports (gethash **ACTUAL-SPACE** **NAME-SPACES**)) arg)

    )

  T

)



; Nome:      LOAD-PACK

; Sintassi:  LOAD-PACK (string)

; Status:    Internal

; Parametri: name Nome dello spazio dei nomi da risolvere

;

; Se non esiste, crea un apposito spazio dei nomi che conterra' il pacchetto caricato.

; In ogni caso procede alla risoluzione del pacchetto.

; Se questa fallisce, prima di uscire cancella lo spazio precedentemente creato.



(defun LOAD-PACK (name)

  (declare (special **NAME-SPACES** **ACTUAL-SPACE**))

  (let (

	(oldspace **ACTUAL-SPACE**)

	)

    (if (not (gethash name **NAME-SPACES**))

	(progn

	  (new-space name)

	  (change-space name)

	  (if (not (prog1

		       (resolve-name name)

		     (change-space oldspace)

		     ))

	      (rm-space name)

	    )

	  )

      )

    

    )

  T

)



; Nome:      GET-COPY-DEF

; Sintassi:  GET-COPY-DEF (symbol symbol)

; Status:    Internal

; Parametri: name Nome spazio dei nomi da utilizzare

;            symb Simbolo da cui generare lo schema

;

; Tramite questa funzione si crea nello spazio attuale una funzione PLaSM che e' l'esatta

; copia di quella referenziata ma non contiene ne codice ne complessi poliedrali: e' solo

; una fake-function che racchiude all'interno un funtore.

; Un suo utilizzo e' equivalente a quello di un funtore tranne per il fatto che non richiede

; un ciclo di applicazione ':' quindi e' molto piu' facile da gestire.



(defun GET-COPY-DEF ( name symb )

  (declare (special **NAME-SPACES**))

  (let (

	(tmp (gethash symb (namespace-space (gethash name **NAME-SPACES**))))

	)

    (insert-in-user-fenv

     (make-plasm-function :name (plasm-function-name tmp)

			  :formal-pars (plasm-function-formal-pars tmp)

			  :body `(dofunctor ',name ',symb)

			  :calling-function (plasm-function-calling-function tmp)

			  :called-function (plasm-function-called-function tmp)

			  )

     )

    )

)



; Nome:      GET-SCHEME

; Sintassi:  GET-SCHEME (symbol)

; Status:    Internal

; Parametri: name Nome spazio dei nomi da utilizzare

;

; Questa funzione crea uno schema di esportazione per lo spazio dei nomi referenziato

; e lo inserisce nello spazio attuale.

; Questo consente il referenziamento dei simboli esportati da un pacchetto tramite il

; loro nome, senza specificare un funtore.

; In questo modo si possono referenziare solo simboli esportati (pubblici).



(defun GET-SCHEME (name)

  (declare (special **NAME-SPACES**))

  (let* (

	 (space (gethash name **NAME-SPACES**))

	 (exports (if space (namespace-exports space) nil))

	 )

    (mapc #'(lambda (symb) (get-copy-def name symb)) exports)

    )

)



(export '(SAVE-ACTUAL-SPACE READ-ACTUAL-SPACE NEW-SPACE RM-SPACE

			    CHANGE-SPACE SCAN-SERVER STOP-SERVER

			    UPLOAD-ACTUAL-SPACE DOWNLOAD-ACTUAL-SPACE SAVE-DEF

			    READ-DEF READ-PLASM-SCRIPT))



(new-token)

(new-space 'default)

(change-space 'default)



; Il modello sintattico cambia:

; IMPORT:<elem1,elem2,...,elemn>;

; dove ogni elemento e' un simbolo, nel qual caso diventa il nome del package da

; scaricare usando gli schemi, oppure e' una lista composta da due elementi,

; dove il primo e' l'alias del pacchetto mentre il secondo e' il nome, nel qual

; caso non usiamo gli schemi ma l'indirizzamento diretto simbolo per simbolo.

;

; SYMEXP:<sym1,...,symn>;

; dove con sym-i si intende un nome di funzione plasm precedentemente definito

; ogni successiva istruzione 'SYMEXP' cancella la precedente lista dei simboli

; esportati



