; -*- package: plasm -*-
;=========================================================================
;                    This is the file SYNTAX.LSP
;=========================================================================


(provide 'syntax)

(inpackage plasm)


;------------------------------------------------------------------------
;Some constants for the interpreter enviroment.


(setf *EPS* '*EPS*) 

;------------------------------------------------------------------------
;le funzioni che seguono sono alcune fra le operazioni elementari
;che si eseguono sulle strutture dati di tipo STACK, in questo ambiente
;realizzato tramite una lista. 

(defun TOP-STK (stack)
  (first stack))

(defun PUSH-STK (state stack)
  (cons state stack))

(defun POP-STK (stack &optional (n 1))
  (if (eq n 1)
    (rest stack)
    (pop-stk (rest stack) (1- n))))

;------------------------------------------------------------------------
;sono dei predicati che utilizza l'automa per controllare il valore  
;restituito dalla funzione DECIDI invocata dall'automa stesso.

(defun SPOSTA-P (decidi-answer)
  (eq 'Sp (first decidi-answer)))

(defun APPLICA-P (decidi-answer)
  (eq 'Ap (first decidi-answer)))

(defun ACCETTA-P (decidi-answer)
  (eq 'Ac (first decidi-answer)))

;-----------------------------------------------------------------------
;restituisce *EPS* se il suo argomento e' una lista vuota oppure il car
;che sara' un carattere


(defun CONTEXT (char-list)
  (declare (special *EPS*))
  (if char-list
      (first char-list)
    *EPS*))

;-----------------------------------------------------------------------
;restituisce la parte destra della produzione che nella lista associativa
;*PRODUZIONI* ha come chiave prod-code

(defun DX-SIDE-PROD (prod-code)
  (declare (special *PRODUZIONI*))
  (second (aref *PRODUZIONI* prod-code)))

;--------------------------------------------------------------------------
;restituisce la parte sinistra della produzione che nella lista associativa
;*PRODUZIONI* ha come chiave 'prod-code'

(defun SX-SIDE-PROD (prod-code)
  (declare (special *PRODUZIONI*))
  (first (aref *PRODUZIONI* prod-code)))

;------------------------------------------------------------------------
;dato lo stato dell'automa 'actual-state' e il primo dei tokens  ancora da 
;leggere nella stringa 'context', restituisce una lista composta al massimo
;da 2 elementi: la lista e' ottenuta dalla lista associativa *DECIDI* che
;per ogni stato contiene l'operazione da eseguire in funzione del context

(defun DECIDI (actual-state context)
  (declare (special *DECIDI* *EPS*))
    (let* ((decidi-for-actual-state (aref *DECIDI* actual-state))
	   (do-not-imp-context      (is-only-eps 
					decidi-for-actual-state
					context))
	   (imp-context             (if do-not-imp-context
					nil
					(assoc (cond
						((characterp context) context)
						((eq context *EPS*) *EPS*)
						(T  #\%))
					       decidi-for-actual-state
					       :test 'member))))
      (cond
       (do-not-imp-context (second do-not-imp-context))
       (imp-context        (second imp-context))
       (t                `(decidi ,decidi-for-actual-state)))))


;---------------------------------------------------------------


(defun IS-ONLY-EPS (decidi-for-actual-state context)
  (declare (special *EPS*))
  (let (a-list)
    (cond
     ((and (eq context *EPS*)
	   (equal '(*EPS*) 
		  (first 
		   (setq a-list 
			 (assoc *EPS* decidi-for-actual-state 
				:test #'member)))))
      a-list)
      ((and (eq context #\;)
	    (or a-list 
		(assoc *EPS* decidi-for-actual-state :test #'member)))))))
 
;--------------------------------------------------------------------------
;In seguito all'esecuzione da parte dell'automa della operazione indicata
;dalla funzione DECIDI, AGGIORNA e' chiamata affinche' restituisca allo
;automa, che l'ha chiamata, lo stato in cui deve portarsi trovandosi nello
;stato 'actual-state' ed avendo letto in ingresso 'what-is-in' (puo'
;essere sia un terminale che un non-terminale).

(defun AGGIORNA (actual-state what-is-in)
  (declare (special *AGGIORNA*))
  (let* ((aggiorna-for-actual-state  (aref *AGGIORNA* actual-state))
	 (pair-char-state  (assoc what-is-in aggiorna-for-actual-state)))
       (if pair-char-state
	   (second pair-char-state)
	   `(aggiorna ,aggiorna-for-actual-state
		      ,what-is-in
		      ,actual-state))))

;----------------------------------------------------------------------------
;This function checks if the state 'state' is inadequate. 
;*STATIINADEGUATI* is a costant list that contains automa inadeguate states.



(setq *STATIINADEGUATI-HASHTABLE* 
      (make-hash-table    :test #'eq
			  :size (+ 2 (length *STATIINADEGUATI*))
			  :rehash-size  1 
			  :rehash-threshold  1.0 
			  ))

(mapcar #'(lambda (key)
	    (declare (special *STATIINADEGUATI-HASHTABLE*))
	    (setf (gethash key *STATIINADEGUATI-HASHTABLE*) t))
	*STATIINADEGUATI*)

(defun INADEQUATE-STATE (state)
  (declare (special *STATIINADEGUATI-HASHTABLE*))
   (gethash state *STATIINADEGUATI-HASHTABLE*))


;-------------------------------------------------------------------------
;this symbol has value equal to the number of the production rule
;whose left side is the non-terminal TOKENID

(setq *TOKENID-PROD-NUMBER* 
      (let ((i 1))
	(declare (special *PRODUZIONI*))
	(loop (if (eq 'TOKENID (first (aref *PRODUZIONI* i)))
		(return i)
		(incf i)))))

;-------------------------------------------------------------------------



(defun AUTOMA (script punt end)
  (declare (special *TOKENID-PROD-NUMBER*))
  (let ((op-code nil)
	(token-read nil)
	(context-ret nil)
	(stack '(1)) ; the automa initial state is the number 1 
	(prod-seq '()))
    (loop
      (setq context-ret (lex script punt end))
      (setq op-code 
	    (if (inadequate-state (top-stk stack))
	      (solve-ambiguous-state 
	       (top-stk stack) 
	       (if context-ret
		 (first context-ret)
		 '*EPS*)
	       prod-seq)
	      (decidi (top-stk stack) 
		      (if context-ret
			(if (characterp (first context-ret))
			  (first context-ret)
			  #\%)
			'*EPS*))))
      (cond ((sposta-p op-code)
	     (setq token-read (first context-ret))
	     (setq punt (second  context-ret))
	     (setq stack 
		   (push-stk 
		    (aggiorna (top-stk stack) (if (characterp token-read)
						token-read
						#\%)
			      ) ; go in this state
		    stack)))
	    ((applica-p op-code)
	     (setq prod-seq  (cons (if (= (second op-code) *TOKENID-PROD-NUMBER*)
				       (list *TOKENID-PROD-NUMBER* token-read)
				       (second op-code))
				   prod-seq))
	     (setq stack     (pop-stk stack
				      (length (dx-side-prod
					       (second op-code)))))
	     (setq stack    (push-stk (aggiorna (top-stk stack)
						(sx-side-prod
						 (second op-code)))
				      stack)))
	    ((and (accetta-p op-code) ; is it in a final state ?
		  (or 
		   (not context-ret) ; and no others chars ?
		   (eq #\; (first context-ret)) 
		   ))
	     (return (list (reverse (cons (second op-code) prod-seq ))
			   (second context-ret))))
	    (T ; there is an error
	     (return `(automa-error ,script 
				    ,punt
				    ,end
				    ,(top-stk stack)
				    ,op-code)))))))


(defun AUTOMA (script punt end)
  (declare (special *TOKENID-PROD-NUMBER*))
  (let ((op-code nil)
	(token-read nil)
	(context-ret nil)
	(stack '(1))           ; the automa initial state is the number 1 
	(prod-seq '())
	(no-have-just-token T) ; at beginning a token must be read
	(decidi-context nil))
    (loop
      
      (if no-have-just-token ; the token was read at last loop?
	(progn 
	  (setq context-ret (lex script punt end))
	  (setq decidi-context 
		(cond
		 ((not context-ret) '*EPS*)
		 ((characterp (first context-ret))
		  (first context-ret))
		 (T #\%)))
	  (setq no-have-just-token nil)))
      
      
      (setq op-code 
	    (if (inadequate-state (top-stk stack))
	      (solve-ambiguous-state 
	       (top-stk stack)
	       (if (not context-ret) 
		 '*EPS*
		 (first context-ret)) ;decidi-context
	       prod-seq)
	      (decidi (top-stk stack) decidi-context)))
      (cond ((sposta-p op-code)
	     (setq token-read (first context-ret))
	     (setq punt (second  context-ret))
	     (setq no-have-just-token T) ; a new token must be read from script
	     (setq stack 
		   (push-stk 
		    (aggiorna (top-stk stack) 
			      decidi-context ; when we have a 'sposta-p' code
					     ; 'decidi-context' is an 
					     ; 'aggiorna-context' too
			      ) ; go in this state
		    stack)))
	    ((applica-p op-code)
	     (setq prod-seq  (cons (if (= (second op-code) *TOKENID-PROD-NUMBER*)
				       (list *TOKENID-PROD-NUMBER* token-read)
				       (second op-code))
				   prod-seq))
	     (setq stack     (pop-stk stack
				      (length (dx-side-prod
					       (second op-code)))))
	     (setq stack    (push-stk (aggiorna (top-stk stack)
						(sx-side-prod
						 (second op-code)))
				      stack)))
	    ((and (accetta-p op-code) ; is it in a final state ?
		  (or 
		   (not context-ret) ; and no others chars ?
		   (eq #\; (first context-ret))
		   ))
	     (return (list (reverse (cons (second op-code) prod-seq ))
			   (second context-ret))))
	    (T ; there is an error
	     (return `(automa-error ,script 
				    ,punt
				    ,end
				    ,(top-stk stack)
				    ,op-code)))))))


;------------------------------------------------------------------------
;Given a string belonging to the language, it exists a sequence of rules which
;give us the way to build the syntax tree and we no need of the string.
;There is una corrispondenza biunivoca between a correct string and its rules 
;sequence .

(defun BUILD-TREE-3 (sequence-rules)
  (let ((tree '()))
    (dolist (prod sequence-rules tree)
      (setf tree (give-them-a-father prod tree)))))

;------------------------------------------------------------------------
;It checks if a dx-side of a rule is made of all terminals


(defun ALL-TERMINAL-P (dx-side-list)
  (not (member nil (mapcar #'terminal-p dx-side-list))))

;------------------------------------------------------------------------
;Is elem a terminal ?

(defun TERMINAL-P (elem)
  (declare (special *TERMINALI*))
  (if (member elem *TERMINALI*)
    t))

;------------------------------------------------------------------------




(defun GIVE-THEM-A-FATHER (prod-code a-list-tree)
  (if (consp prod-code) 

    ; prod-code is a pair ('int' 'token') where 'int' is the index for TOKENID syntax
    ; rule (that is the value of symbol *TOKENID-PROD-NUMBER*) and 'token' is the 
    ; token returned by the lexical parsing 
    (acons (list (first prod-code)  (sx-side-prod (first prod-code)))
	   (rest prod-code)
	   a-list-tree)

    (let ((father (sx-side-prod prod-code))
	  (sons   (dx-side-prod prod-code)))
      (if (all-terminal-p sons)
	(acons (list prod-code father)
	       (list sons)
	       a-list-tree)
	(let ((new-a-list-tree a-list-tree)
	      (dx-side-father  '())
	      (non-terminal    nil))
	  (dolist (elem (reverse sons)
			(acons (list prod-code father)
			       (list dx-side-father)
			       new-a-list-tree))
	    (if (terminal-p elem)
	      (setf dx-side-father (cons elem dx-side-father))
	      (progn
		(setf non-terminal (assoc elem new-a-list-tree
					  :test 'member))
		(setf new-a-list-tree (remove (assoc elem new-a-list-tree
						     :test 'member)
					      new-a-list-tree))
		(setf dx-side-father (cons non-terminal
					   dx-side-father))))))))))


;------------------------------------------------------------------------------


(defun SYN-TREE-3 (script)
  (declare (special *PLASM-READTABLE*))
  (let ((punt 0)
	(end (length script))
	*aut-answer*)
    (setq *READTABLE* *PLASM-READTABLE*)
    (loop
      (if (or (not punt) ; punt does not point to anything !
	      (and (setq punt (all-not-important-char script punt end))
		   nil) ; the AND always return NIL
	      (= punt end))
	(return))
      (reset-pred-stack) ; it has a new top level plasm expression
      (setq *aut-answer* (automa script punt end))
      (if (eq 'automa-error (first *aut-answer*)) 
	(there-was-an-error *aut-answer*)) ; this call abort the evaluation
      (setq punt (second *aut-answer*))
      (eval-script (first (second (first 
		      (build-tree-3 (first *aut-answer*)))))))
    (setq *READTABLE* (copy-readtable nil))
    ))



;------------------------------------------------------------------------
;It's called if the function AUTOMA received not a valid transition but
;a form that SYN-TREE fun can use for a simple diagnostic on the script .



(defun THERE-WAS-AN-ERROR (*aut-answer*)
  (format T "~%~%PLASM SYNTAX ERROR! I'm trying to give you a diagnostic!~%~%" )
  (let ((decidi-for-it   (collect-chars (second (sixth *aut-answer*))))
	(script (second *aut-answer*))
	(punt (third *aut-answer*))
	(end (fourth *aut-answer*)))
    
    (format T "~%You have typed this string:~%~%~A" (second *aut-answer*)) 
    
    (format T "~%~%When I expected as first char of the ~
	       remainder string one of these:~%~%")
    (mapcar #'translate-char-from-decidi decidi-for-it)

    (format T "~%~%I met as remainder string:~%~%~A~%~%"
	    (if (= punt end)
	      "No String At All!"
	      (subseq script punt end)))

    (plasm-error 0)))





;--------------------------------------------------------------- 
;restituisce una lista di caratteri senza duplicati

(defun COLLECT-CHARS (a-list)
  (declare (special *EPS*))
  (let ((chars-list '()))
    (setf chars-list (dolist (x a-list chars-list)
		       (setf chars-list
			     (append chars-list (first x)))))
    (setf chars-list (remove-duplicates chars-list))))


;---------------------------------------------------------------

(defun ALL-NOT-IMPORTANT-CHAR (script punt end)
  (loop
    (if (= punt end) (return punt))
    (if (not-important-char script punt) 
      (incf punt)
      (return punt))))

(defun ALL-NOT-IMPORTANT-CHAR (script punt end)
  (loop
    (if (= punt end) (return punt))
    (cond
      ((not-important-char script punt) 
       (incf punt))
      ((begin-remark-p script punt)
       (setf punt (1+ (skip-to-second-% script punt end))))
      (T (return punt)))))




;---------------------------------------------------------------

(defun TRANSLATE-CHAR-FROM-DECIDI (x) 
  (format T "~A "
	  (cond
	   ((symbolp x) "\"No char at all\"")
	   ((char= x #\0) "D(ef)")
	   ((char= x #\1) "D(efault)")
	   ((char= x #\2) "E(nd)")
	   ((char= x #\3) "W(here)")
	   ((char= x #\%) "\"Any Atom\"")
	   (T  x))))
