; -*- package: plasm -*-
;==========================================================================
;
;                           update-plasm-file.lisp File
;
;==========================================================================

(provide 'up_file)

(inpackage plasm)

(export 'update-plasm-code)


(defun UPDATE-PLASM-CODE (&optional (filename nil))
  (declare (special *PATH-TO-FOLDER-PLASM_SCRIPTS*))
  (let* ((in-file (if filename
                    filename
                    (prompt-for-an-existing-file *PATH-TO-FOLDER-PLASM_SCRIPTS*)))
         (out-file (make-pathname  
                    :directory (pathname-directory in-file)
                    :name (pathname-name in-file)
                    :type (concatenate 'string 
                                       (pathname-type in-file)
                                       "-nv")))
         (*ASK* (funcall #+:APPLE #'y-or-n-dialog
                         #+:RS6000 #'y-or-n-p
                         #+:CLISP #'y-or-n-p
                         (format nil 
                                 "Are you translating from the before last ~
                                  version of PLASM?") )))
                         
    (declare (special *ASK*))
    (with-open-file (input-stream in-file
                                  :direction :input
                                  ;:element-type :character
                                  )
      (with-open-file (output-stream  out-file
                                      :direction :output
                                      ;:element-type :character
                                      :if-does-not-exist :create
                                      :if-exists :overwrite)
        (if (not (TRANSLATE-STREAMS input-stream output-stream))
          (error "File ~A is not a correct PLASM buffer!" in-file)
          (format T "Successfull translation of file \"~A\" ~
                     in \"~A\"." in-file out-file))))
    (values)))


(defun TRANSLATE-STREAMS (input-stream output-stream)
  (loop
    (if (third (READ-WRITE-UNTIL-CHAR #\" input-stream output-stream))
      (return t))
    (write-char #\" output-stream)
    (inside-double-quote input-stream output-stream)
    (write-char #\" output-stream)))



(defun INSIDE-DOUBLE-QUOTE (input-stream output-stream)
  (declare (special *ASK*))
  
  (let (returned)
    (loop
      (if (third (READ-WRITE-UNTIL-CHAR #\{ input-stream output-stream t))
        (return nil))
      
      (setq returned (READ-WRITE-UNTIL-CHAR #\} input-stream output-stream nil t nil))
      (if (third returned)
        (return nil))
      
      (let ((builded (coerce-string (fourth returned))))
        (if (or *ASK*
                (funcall #+:APPLE #'y-or-n-dialog
                         #+:RS6000 #'y-or-n-p 
                         #+:CLISP #'y-or-n-p 
                         (format nil "Do i change the string \"~A\" to \"~A\" ?~%"
                                 (format nil "~A~A~A" "{" builded "}")
                                 (format nil "~A~A~A" "(" builded "::TT)"))))

          (progn
            (write-char #\( output-stream)
            (format output-stream builded)
            (format output-stream "::TT) "))
          (progn
            (write-char #\{ output-stream)
            (format output-stream builded)
            (write-char #\} output-stream)))))))



(defun READ-WRITE-UNTIL-CHAR (the-char input-stream output-stream 
                                       &optional (double-quote nil) (read-string nil) (write? t))
  (let (unread-char
        chars-list) 
    (loop
      (setq unread-char (read-char  input-stream nil t))
      (cond
       ((or (eq t unread-char)
            (and double-quote (char= #\" unread-char)))
        
        (return (list input-stream output-stream t ; get the end of stream
                      (if read-string (reverse chars-list))))) 
       ((char= the-char unread-char)
        (return (list input-stream output-stream nil ; do not get the end of stream
                      (if read-string (reverse chars-list)))))
       ((char= #\; unread-char) ; LISP comment line reached
        (read-write-all-line  input-stream output-stream))
       (T 
        (setq chars-list (cons unread-char chars-list))
        (if write? (write-char unread-char output-stream)))))))



(defun READ-WRITE-ALL-LINE  (input-stream output-stream)
  (format output-stream ";~A~%" (read-line input-stream))) 
                         ; LISP comment char (#\;) was just read from input-stream
