;-----------------------------------------------------------------
;
;  Initialization for PLASM
;  
;  Created  13-01-93
;  Modified Version of SimpleXn Initialization 
;  Last changhed: 16-05-1998
;  M. Vicentino
;
;-----------------------------------------------------------------

;-----------------------------------------------------------------------------
#+:APPLE
(load (concatenate 'string (directory-namestring (mac-default-directory))
                   "Library:logical-dir-compatibility.lisp"))
;----------------------------------------------------------------------------


(defpackage plasm)

#+:WIN32 
(setf *FEATURES* (cons 'UNIX *FEATURES*))


; Start Up Function


(if (not (boundp '*IMAGE-DATE*))
  (progn
        (defconstant *IMAGE-DATE* (multiple-value-list (get-decoded-time)))

        (defconstant *ASK-VERSION-STRING* "Please, type a string to identify this PLASM version:")

        (defvar **VERSION-NUMBER 
        (let (read-string)
             (setf read-string 
                   #+:APPLE
                (get-string-from-user  *ASK-VERSION-STRING* 
                                 :initial-string "1.0")
                #-:APPLE
                (progn 
                (format t "~%~%~%~A" *ASK-VERSION-STRING*)
                (read-line))
        )
        (if (string-equal "" read-string)
	        "1.0"
                 read-string)))
))

#+:APPLE
(defun PLASM-welcome ()
   (declare (special *IMAGE-DATE* **VERSION-NUMBER))
   (let* ((current-time *IMAGE-DATE*)
	  (day (nth 3 current-time))
	  (month (nth 4 current-time))
	  (year (nth 5 current-time)))
    (message-dialog  
     (with-output-to-string (stringa)
       (format stringa
               "~%~%
	      Welcome in 
	   **** PLASM ****
   Programming LAnguage for
         Solid Modeling 
          Version: ~A~% 
	  Date   : ~A/~A/~A~%~%"
               
	  **VERSION-NUMBER     day month year))
     :size (make-point 300 200))))







#-:APPLE
(defun PLASM-welcome ()
   (declare (special **VERSION-NUMBER  *IMAGE-DATE*))
   (let* ((current-time *IMAGE-DATE*)
	  (day (nth 3 current-time))
	  (month (nth 4 current-time))
	  (year (nth 5 current-time)))
    (format t
"~%
	      Welcome in 
	   **** PLASM ****
   Programming LAnguage for Solid Modeling
          Version: ~A
          Date   : ~A/~A/~A~%~%"

          **VERSION-NUMBER     day month year)) 
	     (plasm::use-plasm)
)
  


;Please, comment the next line when you compile new code, then remove comment!

(setf *COMPILE-WARNINGS* nil)
(defvar *PLASM-COMPILING* T)
(setf *LOAD-VERBOSE* T)
 


#+:RS6000
(proclaim '(optimize (speed 3) (safety 0) (space 0) (debug 0)))

;---------
;Path of entirely Plasm-code
  
#+:APPLE
(progn
  (message-dialog "It will be asked to select a directory to all PLASM code!")
  (defvar **disk-name 
    (mac-directory-namestring (choose-directory-dialog))
    ))
;-----------------------------------------------------------------------------



(defvar **path-to-all-code
  #+:RS6000
  "~/plasm98/"
  
  #+:UNIX
  #-:WIN32
  ;"~/plasm98/"
  (progn 
    (format t "~%~%~%~A~%~A~%>"
            "Insert the complete path to Plasm e Simplex directory"
            " [for example (no quote) /usr/lib/plasm98/]")
    (read-line))
  
  #+:WIN32
  (progn 
    (format t "~%~%~%~A~%~A~%>"
            "Insert the complete path to Plasm e Simplex directory"
            " [for example (no quote) c:\source\plasm98\]")
    (read-line))
  
  #+:DOS
  "\\clisp\\"
  )


;-----------------------------------------------------------------------
; this is the pathname to the SimpleXn directory --


#+:APPLE
(defvar **simplexn-directory
  (get-string-from-user  "Type path to Simplexn directory:" 
			 :initial-string 
			 (concatenate 'string **disk-name "simplexn")))

#+:UNIX
(defvar **simplexn-directory
  (concatenate 'string **path-to-all-code "simplexn"))

#+:RS6000
(defvar **simplexn-directory
  (concatenate 'string **path-to-all-code "simplexn"))


#+:DOS
(defvar **simplexn-directory
  (concatenate 'string **path-to-all-code "simplexn"))


;--------------------------------------------------------------------------


#+:APPLE
(defvar **plasm-directory
  (get-string-from-user  "Type path to PLASM directory:" 
			 :initial-string 
			 (concatenate 'string **disk-name "plasm")))
#+:UNIX
(defvar **plasm-directory
  (concatenate 'string **path-to-all-code "plasm"))

#+:RS6000
(defvar **plasm-directory
  (concatenate 'string **path-to-all-code "plasm"))


#+:DOS
(defvar **plasm-directory
  (concatenate 'string **path-to-all-code "plasm"))

;--------------------------------------------------------------------------

(defvar **recompile-p
  #+:APPLE 
  (y-or-n-dialog
   "Do you want to compile PLASM and 
substitute the present image on the disk?" :cancel-text nil)
  
  #+:UNIX
  nil
  ;(y-or-n-p  "Do you want to compile PLASM and create a new PLASM image?  ")
  
  
  )


;; *********************** setting paths **************************


#+:RS6000
(setq system:*require-search-list* 
      (append system:*require-search-list*
              '(
                (:newest-do-compile
                 
                 #.(make-pathname :directory **simplexn-directory :type "fasl")
                 #.(make-pathname :directory **simplexn-directory :type "lsp")
                 #.(make-pathname :directory **plasm-directory :type "fasl")
                 #.(make-pathname :directory **plasm-directory :type "lsp")
                 )
                ))
      )

#-:RS6000 
#+:UNIX
(setq *LOAD-PATHS* 
  (append *LOAD-PATHS*  
	  (list (concatenate 'string **simplexn-directory "/")
		(concatenate 'string **plasm-directory "/"))))

#+:DOS
(setq *LOAD-PATHS* 
  (append *LOAD-PATHS*  
	  (list (concatenate 'string **simplexn-directory "\\")
		(concatenate 'string **plasm-directory "\\"))))


#+:RS6000 (load (concatenate 'string **simplexn-directory "/" "simplexn.go"))

#+:APPLE (load (concatenate 'string **simplexn-directory ":" "simplexn.go"))

#-:RS6000 
#+:UNIX (require 'simplexn.go)

#+:DOS (require 'simplexn.go)


;---------

  
#+:RS6000
(setq system:*require-search-list* 
      (append system:*require-search-list*
              '(
                (:newest-do-compile
                 
                 #.(make-pathname :directory
                                  (concatenate 'string **plasm-directory "/p_code") :type "fasl")
                 #.(make-pathname :directory
                                  (concatenate 'string **plasm-directory "/p_code") :type "lsp")
                 #.(make-pathname :directory
                                  (concatenate 'string **simplexn-directory "/bool") :type "lsp")
                 )
                )))

#-:RS6000 
#+:UNIX
(setq *LOAD-PATHS* 
  (append *LOAD-PATHS*  
	  (list (concatenate 'string **plasm-directory "/p_code/")
                (concatenate 'string **simplexn-directory "/bool/"))))

#+:DOS
(setq *LOAD-PATHS* 
  (append *LOAD-PATHS*  
	  (list (concatenate 'string **plasm-directory "\\p_code\\")
        (concatenate 'string **simplexn-directory "\\bool\\"))))


(defun READ-INFO ()
  (with-open-file 
    (infile 
     (concatenate 'string **plasm-directory 
                  #+:APPLE  ":"
                  #+:RS6000 "/"
                  #+:UNIX "/"
                  #+:DOS  "\\"
	          "info")
     :direction :input
     :element-type 'character
     :if-does-not-exist nil)
    (if infile
      (let (char)
        (loop
          (setf char (read-char  infile nil nil))
          (if char 
            (write-char char)
            (return )))))
    (values)))


#+:APPLE
(when **recompile-p
  (setq *lisp-startup-functions*
	(append *lisp-startup-functions* 
		(list 'read-info 'plasm-welcome))))
  


;-----------------------------------------------------------------------------
;    PLASM environment


#+:APPLE
(def-logical-directory "plasm" **plasm-directory)


(defvar **plasm-code-directory
  #+:APPLE  (concatenate 'string **plasm-directory ":p_code")
  #+:RS6000 (concatenate 'string **plasm-directory "/" "p_code") 
  #+:UNIX   (concatenate 'string **plasm-directory "/" "p_code") 
  #+:DOS    (concatenate 'string **plasm-directory "\\" "p_code") )



#+:APPLE
(def-logical-directory "p_code" **plasm-code-directory)



(defvar **plasm-bool-directory
  #+:APPLE  (concatenate 'string **simplexn-directory ":bool")
  #+:RS6000 (concatenate 'string **simplexn-directory "/" "bool") 
  #+:UNIX   (concatenate 'string **simplexn-directory "/" "bool") 
  #+:DOS    (concatenate 'string **simplexn-directory "\\" "bool") )



#+:APPLE
(def-logical-directory "p_bool" **plasm-bool-directory)



#+:APPLE
(setf *module-search-path* 
      (append  
       '("plasm;" "p_code;" )
       *module-search-path*))


;----------------------------------------------------------------

;vrml extension

(require 'p_vrml)

(require 'addpvrml)


;----------------------------------------------------------------

;----------------------------------------------------------------

;animation extension

(require 'p_iv)

;----------------------------------------------------------------


;----------------------------------------------------------------

(require 'p_main)

;----------------------------------------------------------------

;(require 'putfenvn)
(require 'bool)


(use-package 'plasm)

;----------------------------------------------------------------





;------------------------------------------------------------
;     Installing menu in Apple environment


#+:APPLE
(progn
  
  (defvar **plasm-mac-menu-directory
    (concatenate 'string 
	         **plasm-directory 
	         ":p_mac"))
  
  (def-logical-directory "p_mac" **plasm-mac-menu-directory)
  
  (setf *module-search-path* 
        (cons 
	 "p_mac;"
	 *module-search-path*))
  
  (require 'p_menu))


;------------------------------------------------------------

#+:RS6000
(progn
  
  excl:(dolist (x (others *print-pprint-dispatch*))
	 (when (and (functionp (test x))
		    (not (compiled-function-p (test x))))
	   (setf (test x) (compile nil (fn_code (test x))))))
  
  (setf user::*evalhook* #'plasm::new-eval))

;----------------------------------------------------------------

; Start Dumping


(when **recompile-p
  
  #+:APPLE    
  (progn
    (message-dialog 
     "It will be asked to choose a directory and a name for PLASM image!")
    
    (save-application  (make-pathname
		        :directory 
                        (mac-directory-namestring
                         (choose-directory-dialog :directory (mac-default-directory)))
		        :name
                        (get-string-from-user  "Type a name for PLASM image:" 
			                       :initial-string
                                               "mac_plasm"))))
    
  #+:RS6000
  (dumplisp 
   :name **path-to-all-code  "/plasm.image"
   :read-init-file T
   :restart-function   #'(lambda () (read-info) (plasm-welcome)))
  
  #+:CLISP
  (saveinitmem
   (make-pathname 
    :directory
    (pathname-directory 
     (progn 
       (format t "~%~%~%~A~%~A~%>"
               "Insert the complete path to Plasm binary directory"
               "[for example (no quote) /usr/bin/]")
       (read-line)))            
    :name "plasm"
    :type "mem")
   :quiet t
   :init-function 'plasm-welcome)      
  
  )



#-:APPLE


