% errore interprete PLASM: -(3+2)  !!!!! ----------------------%

%--------------------------------------------------------------%
%-- Utility toolbox -------------------------------------------%
%--------------------------------------------------------------%

DEF Q = QUOTE~IF:<IsSeq,ID,LIST>;
DEF ButLast = REVERSE~TAIL~REVERSE;
DEF circle (a::IsReal) (r::IsReal) (n::IsInt) =
  (S:<1,2>:<r,r>~JOIN):(MAP:([cos,sin]~s1):((QUOTE~#:n):(a/n)));
DEF TrunCone (r1,r2,h::IsReal)(n::IsInt) =
  MAP:[ x * cos~s2, x * sin~s2, z ]:
    (QUOTE:<1> * (QUOTE~#:n):(2*PI/n))
WHERE
  x = K:r1 + s1 * ( K:r2 - K:r1 ),
  y = K:0,
  z = s1 * K:h
END;

%--------------------------------------------------------------%
%-- Design parameters -----------------------------------------%
%--------------------------------------------------------------%

DEF rodHeight = 20;
DEF basisRadius = 20;
DEF basisHeight = 2;

%--------------------------------------------------------------%
%-- Geometric model of parts ----------------------------------%
%--------------------------------------------------------------%

DEF halfHinge2D = circle:PI:1:12;
DEF hinge2D = STRUCT:< halfHinge2D, T:<1,2>:<-1,-3>,Q:2 * Q:3 >;
DEF hinge = (MKPOL~UKPOL):(hinge2D * Q:0.5);
DEF DoubleHinge = (MKPOL~UKPOL~STRUCT):<hinge, T:3:1, hinge>;

DEF DoubleHinge = STRUCT:<hinge, T:3:1.2, hinge>;
DEF hbasis = circle:(2*PI):1.2:24 * Q:2;
DEF femaleJoint = XSTRUCT:< COLOR:<1,0,0>,
  XT:3:-5,hbasis, XT:2:0.85,XR:<2,3>:(PI/2),DoubleHinge>;
DEF maleJoint   = XSTRUCT:< COLOR:<1,0,0>,
  XT:3:-5,hbasis, XT:2:0.25,XR:<2,3>:(PI/2),Hinge>;
DEF rod = XSTRUCT:< COLOR:<1,1,1>,
  XT:<1,2>:<SQRT:2/-2,SQRT:2/-2>,CUBOID:<SQRT:2,SQRT:2,rodHeight>
>;
DEF JointedRod = (R:<2,3>:PI:maleJoint TOP rod TOP femaleJoint);
DEF basis = (circle:(2*PI):basisRadius:32 * Q:basisHeight) TOP femaleJoint;
DEF head = STRUCT:<  R:<1,3>:PI, T:3:-33,
  XSTRUCT:< COLOR:<1,1,1>, basis, T:3:0.01, COLOR:<1,0,0>, basis>, T:3:20,
  TrunCone:<4,4,8>:32, T:3:8, cap, T:3:5, maleJoint >
  WHERE
    basis = TrunCone:<20,4,20>:32,
    cap = embed:1:(circle:(2*PI):4:32)
  END;

%--------------------------------------------------------------%
%-- Fixed assembly --------------------------------------------%
%--------------------------------------------------------------%

DEF Luxo (a1,a2,a3::IsReal) =
  XSTRUCT:< COLOR:<1,0,0>, basis, T:3:(basisHeight+5),
           R:<1,3>:a1, JointedRod,
           T:3:(rodHeight+10), R:<1,3>:a2, JointedRod,
           T:3:(rodHeight+10), R:<1,3>:a3, head >;
%
Luxo:<-45,45,45>
%

%--------------------------------------------------------------%
%-- Bezier curves ---------------------------------------------%
%--------------------------------------------------------------%

DEF Fact (n::IsInt) =  *:(CAT:<<1>, 2..n>);
DEF BinCoeff (n,i::IsInt) = Fact:n / (Fact:i * Fact:(n-i));
DEF Bernstein (n::IsInt)(i::IsInt) =
  * ~ [K:(BinCoeff:<n,i>),** ~ [ID,K:i],** ~ [- ~ [K:1,ID],K:(n-i)]] ~ s1;
DEF BernsteinBase (n::IsInt) = AA:(Bernstein:n):(0..n);
DEF Bezier (ControlPoints::IsSeq) = (CONS~AA:(+~AA:*~TRANS)~DISTL):
     <BernsteinBase:degree,(AA:(AA:K)~TRANS):ControlPoints>
WHERE
  degree = LEN:ControlPoints - 1
END;

%--------------------------------------------------------------%
%-- Motion path -----------------------------------------------%
%--------------------------------------------------------------%

DEF CSpath = Bezier:<<0,0,0>,<30,0,-10>,<-20,90,0>,<-60,105,60>>;
DEF Sampling (n::IsIntPos) = (AA:LIST ~ AA:/ ~ DISTR):<0..n,n>;

%--------------------------------------------------------------%
%-- Dinamic animation ------------------------------------------%
%--------------------------------------------------------------%

DEF motion = (PI/180) AA:(AA:*~DISTL)~DISTL ((AA:CSpath):(Sampling:9));
MOVE:Luxo:motion:(1..10)


MOVE:Luxo:((aa:AR~trans):<motion,1..10>)
