;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;	Project..........:	Graduation Thesis
;;	Module...........:	polrecv.lsp
;;	Application......:	BSP Library
;;	Version..........:	1.10
;;
;;	Compiler.........:	Common Lisp (CLISP)
;;	Development by...:	Claudio Baldazzi
;;	Internet.........:	baldazzi@inf.uniroma3.it
;;
;;	Copyright........:	Universita' degli studi di Roma TRE
;;						Dipartimento di Informatica e Automazione
;;
;;	Environment......:	CLISP
;;	Beginning........:	18/10/95
;;	Last update......:	01/09/97
;;
;;	-----------------------------------
;;	Abstract:
;;
;;		Receiving .... C++ -- Lisp
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



(inpackage kernel)



;;;;;;;;;;;;;;
;;
;; Receive polyhedra complex from C++ after boolean operation
;;
(defun receive-pol-complex ()
	(setq is (open fileout :direction :input))		;; Open File 'buffer.out' from c++
	(setq pol (receive-pol is))
	(setq mat (make-array (list (+ 1 (polyhedron-m pol)) (+ 1 (polyhedron-m pol))) :initial-element '0.0))
	(dotimes (i (+ 1 (polyhedron-m pol)))
		(setf (aref mat i i) '1.0)
	)
	(setq comp (make-empty-polyhedra-complex))
	(put-node-in-pol-complex (make-pol-node pol mat) comp)
	(close is)                                                                                                      ;; Close File 'bspthunk.out' from c++
	comp
)


;;;;;;;;;;;;;;
;;
;; Receive polyhedron from file
;; This function create a polyhedron represented on faces
;;
(defun receive-pol (istream)                                                                    
	(setq pol (make-empty-polyhedron))					;; Create an empty polyhedra
	(setq dim (receive-item istream))					;; Receive dimension
	(setf (polyhedron-m pol) dim)                                           
	(setf (polyhedron-n pol) dim)                                           
	(setf (polyhedron-representation pol) 'on-faces)	;; Set representation on faces
	(setf (polyhedron-f pol) (receive-item istream))
	(setf (polyhedron-faces pol) (receive-faces (polyhedron-f pol) istream))
	(setf (polyhedron-kind-of-brick pol) 'convex-cells)
	(setf (polyhedron-c pol) (receive-item istream))
	(setf (polyhedron-convex-cells pol) (receive-cells (polyhedron-c pol) istream))
	pol
)       

;;;;;;;;;;;;;;
;;
;; Receive polyhedron faces
;;
(defun receive-faces(number-hyperplane istream)
	(setq ftable (make-hash-table :test #'equal :size number-hyperplane))
	(dotimes (i number-hyperplane ftable)
		;(put-in-hash-table (make-hash-id 'f i) (list (receive-item istream)) ftable))
		(setq object (receive-item istream))
		;(setf (gethash (make-hash-id 'f i) ftable) object))
		(setf (gethash object ftable) (make-hash-id 'f i)))
	;;table
)

;;;;;;;;;;;;;;
;;
;; Receive polyhedron convex cell
;;
(defun receive-cells(number-cells istream)
	(setq table (make-hash-table :test #'equal :size number-cells))
	(dotimes (i number-cells)
		(setq cell (make-convex-cell (make-hash-id 'c i) '+ '() '() '() '()))
		(setq hplist (receive-item istream))
		(dolist (el hplist)
			(setf (convex-cell-faces cell) 
				;(append (convex-cell-faces cell) (list (make-face-elem el))))
				(append (convex-cell-faces cell) (list (make-face-elem-bis el))))
			(setf (convex-cell-adjacent-convex-cells cell) 
				(append (convex-cell-adjacent-convex-cells cell) (list '0)))
			(setf (convex-cell-external-adjacences cell) 
				(append (convex-cell-external-adjacences cell) (list '())))
			(setf (convex-cell-open-faces cell) 
				(append (convex-cell-open-faces cell) (list '())))
		)
		(setq vertlist (receive-item istream))
		(dolist (vert vertlist)
			(setq vatom '())
			(dolist (hp vert)
				(setq vatom (append vatom (list (make-hash-id 'f hp))))                                 
			)
			(setf (convex-cell-vertices cell) 
				(append (convex-cell-vertices cell) (list vatom)))                              
		)               
		;;(put-in-hash-table cell (make-hash-id 'c i) table)            
		(setf (gethash (make-hash-id 'c i) table) cell)
	)
	table
)

;;;;;;;;;;;;;;
;;
;; Receive an object from input stream
;;
(defun receive-item (istream) 
	(read istream)
)

;;;;;;;;;;;;;;
;;
;; This function create hash table id concatening lett with number i
;;
(defun make-hash-id (lett i) 
	(intern (concatenate 'string (string lett) (integer->string i)))
)

;;;;;;;;;;;;;;
;;
;; This function put value with key key in hash table
;;
(defun put-in-hash-table (key value table)
	;;(setf (gethash value table) key)
	(setf (gethash key table) value)
)

;;;;;;;;;;;;;;
;;
;; This function create a face element
;;
(defun make-face-elem (x)				;; ritorna una lista del tipo (Fx . -) per :FACES
	(setf elem '(f . -))
	(pop elem)
	(push (make-hash-id 'f x) elem)
	elem
)



(defun make-face-elem-bis (x)
	(setq n (pop x))
	(push (make-hash-id 'f n) x)
	x
)




;;;;;;;;;;;;;;
;;
;; EXtension
;; Receive polyhedra complex from buffer.ex after boolean operation
;;

(setq fileoutex 'buffer.ex)

(defun receive-pol-complex-ex-begin ()
	(setq isex (open fileoutex :direction :input))  ;; Open File 'buffer.ex' from c++
	T
)

(defun receive-pol-complex-ex-end ()
	(close isex)  
	T
)

(defun receive-pol-complex-ex (dummy)
	(setq pol (receive-pol isex))
	(setq mat (make-array (list (+ 1 (polyhedron-m pol)) (+ 1 (polyhedron-m pol))) :initial-element '0.0))
	(dotimes (i (+ 1 (polyhedron-m pol)))
		(setf (aref mat i i) '1.0)
	)
	(setq comp (make-empty-polyhedra-complex))
	(put-node-in-pol-complex (make-pol-node pol mat) comp)
	comp
)


