;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;	Project..........:	Graduation Thesis
;;	Module...........:	polsend.lsp
;;	Application......:	BSP Library
;;	Version..........:	1.10
;;
;;	Compiler.........:	Common Lisp (CLISP)
;;	Development by...:	Claudio Baldazzi
;;	Internet.........:	baldazzi@inf.uniroma3.it
;;
;;	Copyright........:	Universita' degli studi di Roma TRE
;;						Dipartimento di Informatica e Automazione
;;
;;	Environment......:	CLISP
;;	Beginning........:	18/10/95
;;	Last update......:	01/09/97
;;
;;	-----------------------------------
;;	Abstract:
;;
;;		Sending .... Lisp -- C++
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(inpackage kernel)


(defun sendobj (pol-complex)
	(setf os (open 'obj.obj :direction :output))
	(format T "~%Sending...")
	(send-item (coerce-pol-complex pol-complex 'on-faces) os)
	(close os)
	(format T "done.~%")
	T
)


(defun sendpols (pol-complex-s)
	(setf os (open filein :direction :output))
	(format T "~%Sending...")
	(sendp (first pol-complex-s) os)
	(sendp (second pol-complex-s) os)
	(close os)
	(format T "done.~%")
	T
)


(defun sendpol (pol-complex)
	(setf os (open filein :direction :output))
	(format T "~%Sending...")
	(sendp pol-complex os)
	(close os)
	(format T "done.~%")
	T
)


(defun sendp (pol-complex os)
	(format os ";;POL-COMPLEX")
	(let*   ((d (get-dpol-complex pol-complex))             ;; d dimensione intrinseca 
			 (n (get-npol-complex pol-complex))             ;; n dimensione spazio di immersione
			 (id-mat (mat-identity (1+ n)))                 ;; matrice identita'
			)
			(send-item d os)
			(send-item n os)
			(mapcar #'(lambda (node) (traversal node id-mat d n os)) 
				(coerce-pol-complex pol-complex 'on-faces))
	)
	(format os "~%#~%")
)


;****************************************************************************
;*                                  traversal                               *
;****************************************************************************
(defun traversal (node old-mat d n os)
	(let    ((pol (get-node-pol-complex node)) 
			(matrix (get-node-matrix node)))
			(if (polyhedron? pol)
				(let ((faces-list (get-flist pol)))
					 (format os "~%;POLYHEDRON")
					 (send-item (polyhedron-m pol) os)
					 (send-item (polyhedron-n pol) os)
					 (send-item (polyhedron-f pol) os)
					 (dolist (elem faces-list)                      ;; send polyhedra hyperplanes
						(send-hyperplane elem os))
					 (format os "~%;")
					 (send-matrix (mat-product-extended matrix old-mat) os)
					 ;;;;(send-matrix (mat-product-extended old-mat matrix) os)
					 (format os "~%;")
					 (send-item (polyhedron-c pol) os)
					 (mapcar #'(lambda (cell) 
						(convert-cell cell os))
						(get-clist pol))
				)
				(mapcar #'(lambda (node) 
					(traversal node (mat-product-extended matrix old-mat) d n os))
					;;;;(traversal node (mat-product-extended old-mat matrix) d n os))  
					pol)
			) 
	)
)


(defun convert-cell (cell os)
	(send-cell (get-cell-faces cell) os)
)


(defun send-hyperplane (hp os)
	(format os "~%~a  " (string->integer (strim (first hp))) os)
	;;(format os "~%~a  " (strim (first hp)) os)
	(dolist (el (first (rest hp)))
		(send-float el os)
		(format os " ")
	)
)

	
(defun send-cell (c os)
	(format os "~%")
	(dolist (el c)
		(format os "~a " (string->integer (strim (first el))))
		;;(format os "~a " (strim (first el)))
		(format os "~a  " (rest el))
	)
)


(defun send-list (l os)
	(format os "~%")
	(dolist (el l)
		(format os "~a " el)
	)
)

		
(defun send-matrix (mat os)
	(dotimes (i (array-dimension mat 0))
		(format os "~%") 
		(dotimes (j (array-dimension mat 1))
			(send-float (aref mat i j) os)
			(format os " ")
		)
	)
)


(defun send-item (item ostream) 
	(print item ostream)
)


(defun send-float (item ostream)
	(format ostream "~16,10E" (coerce item 'double-float)))
	
	
(defun strim (symb)
	(setq alf ":|abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ")
	(string-trim alf (write-to-string symb))
)


