; -*- package: intersect -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package Intersect                                         ;                                
;                                                                           ;
; Date: 20-12-90                                                            ;
;  Version: 1.0                                                             ;
;  Note: in function generate-vertices, a bound on the max number           ;
;         of bases is used; the bound is at the beginning of the file       ;
;                                                                           ;
;  Author: V. Ferrucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;; ATTENZIONE: FUNZIONA SOLO CON NUMERI RATIONAL!!!!

(provide 'intersec)                           ;module name

(defpackage intersect) 
(inpackage intersect) ;package name

(export '(intersect?
          calculate-rectangular-coordinates
          build-constraints
          generate-vertices))
                        
(require 'kernel)
(require 'pl)  ;linear programming package
(require 'matrix)

(use-package 'kernel)
(use-package 'pl)
(use-package 'matrixlib)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(setq $$max-bases 6000) ;the max number of bases enumerated

(defun intersect? (vertices1 vertices2)  
  (let* ((p (length (car vertices1)))
         (d1 (- (length vertices1) 1))
         (d2 (- (length vertices2) 1))       
         (nconstr (+ p 2))
         (nvar (+ d1 d2 2))
         (A (build-constraints vertices1 vertices2 d1 d2 p)))
    (multiple-value-bind (A m n v0)
                         (build-canonical-form A nconstr nvar)
      (if A (values A m n v0) nil))))


(defun calculate-rectangular-coordinates (vertices lambdacoordslist)
  (let* ((p (length (car vertices)))
         (d1+1 (length vertices))
         (V (transpose (make-array (list d1+1 p) 
                                   :initial-contents vertices))))
    (mapcar #'(lambda (vertexlambdacoords)
                (coerce (pre-multiply 
                         V 
                         (coerce vertexlambdacoords 'vector))
                        'list))
            lambdacoordslist)))


(defun build-constraints (vert1 vert2 d1 d2 p)
  (let*((p+1 (+ p 1))
        (d1+1 (+ d1 1))
        (d2+1 (+ d2 1))
        (pointlambda2 (+ d1+1 1))
        (nrows (+ p 2))
        (ncol (+ d1+1 d2+1 1))
        (constraints (make-array (list nrows ncol) 
                                 :element-type 'rational
                                 :initial-element 0))) 
    (do ((vertices vert1)
         (k 1 (+ k 1)))
        ((> k d1+1))
      (store-subcolumn 
       constraints 0 k
       (coerce (car vertices) 'vector)
       p)
      (setf vertices (cdr vertices)))
    (do ((vertices vert2)
         (k pointlambda2 (+ k 1)))
        ((= k ncol))
      (store-subcolumn 
       constraints 0 k
       (change-the-sign (coerce (car vertices) 'vector) p)
       p)
      (setf vertices (cdr vertices)))
    (setf (aref constraints p 0) 1)
    (do ((j 1 (+ j 1)))
        ((= j pointlambda2))
      (setf (aref constraints p j) 1))
    (setf (aref constraints p+1 0) 1)
    (do ((j pointlambda2 (+ j 1)))
        ((= j ncol))
      (setf (aref constraints p+1 j) 1))
    constraints))

(defun neighbours (vj A m n)
  (do ((nlist '())
       (j 1 (+ j 1)))
      ((= j n) nlist)
    (if (not (member j vj))
      (let ((llist (find-l j A m)))
        (if llist 
          (mapc #'(lambda (l)
                    (push (substitute j (nth l vj) vj) nlist))
                llist))))))

(defun find-l (j A m)
  (do* ((l (do ((i1 0 (+ i1 1)))
              ((= i1 m) (return-from find-l nil))
            (if (> (aref A i1 j) 0) (return i1))))
       (min (/ (aref A l 0) (aref A l j)))
       (llist (list l))
       (i (+ l 1) (+ i 1)))
      ((= i m) llist)
   (if (> (aref A i j) 0)
    (let ((q (/ (aref A i 0) (aref A i j))))
     (cond ((< q min) (setf min q llist (list i)))
           ((= q min) (push i llist))
           (t))))))

(defun nearest-node (vi W m+1) ;gives vj,the nearest node to vi in W
  (let* ((vj nil)
         (dj m+1))
    (mapc #'(lambda (v)
              (if (zerop (cdr v))
                (let ((d (count 0 (mapcar #'- (car v) vi) :test #'/=)))
                  (if (< d dj) (setf dj d vj (car v))))))
          W)
    (values vj dj)))

(defun get-BFS (vj A m n)  ;BFS = Basic Feasible Solution
  (do ((BFS (make-array (list (- n 1))
                        :initial-element 0))
       (pointers (coerce vj 'vector))
       (i 0 (+ i 1)))
      ((= i m) (coerce BFS 'list))
    (setf (svref BFS (- (svref pointers i) 1)) (aref A i 0))))

(defun generate-vertices (A m n v0)
  (let* ((m+1 (+ m 1))
         (lastinserted v0)
         (vertices (list (get-BFS v0 A m n)))
         (W (list (cons v0 1)))
         (fathers (make-hash-table :size 100 :test #'equal)))
    (mapc #'(lambda (vi)
              (setf W (acons vi 0 W) (gethash vi fathers) v0))
          (neighbours v0 A m n))   
    (loop 
      (multiple-value-bind (vk d) 
                           (nearest-node lastinserted W m+1)
        (if (or (= d m+1) (> (hash-table-count fathers) $$max-bases))  ;d=m+1 or number of
          (progn (format t "~D" (hash-table-count fathers))
                 (return-from generate-vertices  ;bases enumerated > 50 exit
                              (values vertices (length W)))))
        (rplacd (assoc vk W :test #'equal) 1)    ;flag vk 
        (pushnew (calculate-node lastinserted vk d fathers A m n)
                 vertices :test #'equal)
        (mapc #'(lambda (vi)
                  (unless (assoc vi W :test #'equal) 
                    (setf W (acons vi 0 W) (gethash vi fathers) vk)))
              (neighbours vk A m n))       
        (setq lastinserted vk)))))




(defun from-vi-to-vj (vi vj A m n)
  (mapc #'(lambda (l k) 
           (if (/= l k) (pivot A m n (position l vi) k))) vi vj))

(defun calculate-node (vi vj d fathers A m n)   ; vi->vj
  (if (= d 1)
    (from-vi-to-vj vi vj A m n)
    (follow-the-path (build-path vi vj fathers) A m n))
  (get-BFS vj A m n))

(defun follow-the-path (path A m n)
  (if (cdr path)
    (progn 
      (from-vi-to-vj (first path) (second path) A m n)
      (follow-the-path (cdr path) A m n))))

(defun build-path (vi vj fathers)
  (let ((fromvi (list vi))
        (fromvj (list vj)))
    (loop
      (let ((nextvi (gethash (car fromvi) fathers))
            (nextvj (gethash (car fromvj) fathers)))
        (if (member nextvi fromvj :test #'equal)
          (return-from build-path
                       (append 
                        (reverse fromvi) 
                        (member nextvi fromvj :test #'equal)))
          (push nextvi fromvi))
        (if (member nextvj fromvi :test #'equal)
          (return-from build-path
                       (append
                        (reverse (member nextvj fromvi :test #'equal))
                        fromvj))
          (push nextvj fromvj))))))
