; -*- package: pl -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package Linear Programming                                ;                                
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: V. Ferrucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'pl)                           ;module name

(defpackage pl)  

(inpackage pl)  ;package name

(export '(build-canonical-form
	  pivot))
			
(require 'matrix)

(use-package 'matrixlib)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;; ATTENZIONE: FUNZIONA SOLO CON NUMERI RATIONAL!!!!


;;;; Questa e' la versione con soli vincoli di uguaglianza e senza funzione obiettivo. 
;;;; ATTENZIONE: Non possono comparire variabili non vincolate in segno
;;;; Pone solo la matrice in forma canonica risolvendo il problema artificiale.
;;;; La prima colonna della matrice CONSTRAINTS deve contenere i termini noti (bi).
;;;;
;;;; INPUT :
;;;;       M1          :=       NUMBER OF EQUALITY CONSTRAINTS.
;;;;       P           :=       NUMBER OF VARIABLES
;;;;       CONSTRAINTS :=       MATRIX OF CONTRAINTS
;;;; OUTPUT:  PUTS THE MATRIX IN CANONICAL FORM. GIVES THE MATRIX, ITS DIMENSIONS AND
;;;;          A LIST OF INDICES FOR THE VARIABLES ACTUALLY IN THE BASE

(defun pivot (A m n l k)
  (let ((pivotcolumn (make-array (list m) :element-type 'rational)))
    (dotimes (i m) (setf (svref pivotcolumn i) (aref A i k)))
    (dotimes (j n) (setf (aref A l j) (/ (aref A l j)  (svref pivotcolumn l))))
    (dotimes (i m A) 
      (if (/= i l) 
	(dotimes (j n)
	  (setf (aref A i j) 
		(- (aref A i j) (* (aref A l j) (svref pivotcolumn i)))))))))

(defun simplex-algorithm (A m n pointers)
  (loop (let* ((k (do ((min 0)
		       (k nil)
		       (j 1 (+ j 1)))
		      ((= j n) (if k k (return-from simplex-algorithm t)))
		    (if (< (aref A 0 j) min) (setq min (aref A 0 j) k j))))
	       (l (do* ((l (do ((i1 1 (+ i1 1)))
			       ((= i1 m) (return-from simplex-algorithm nil))
			     (if (> (aref A i1 k) 0) (return i1))))
			(min (/ (aref A l 0) (aref A l k)))
			(i (+ l 1) (+ i 1)))
		       ((= i m) l)
		    (if (> (aref A i k) 0)
		      (if (< (/ (aref A i 0) (aref A i k)) min)
			(setf min (/ (aref A i 0) (aref A i k)) l i))))))
	  (pivot A m n l k)
	  (setf (svref pointers l) k))))

(defun build-canonical-form (CONSTRAINTS m1 p)
  (let* ((m (+ 1 m1))
	(n  (+ 1 p))
	(pointaux n)
	(A (make-array (list m n)
		     :initial-element 0
		     :element-type 'rational
		     :adjustable t))
	(pointers (make-array (list m))))
   ;; REDUCTION OF PROBLEM TO STANDARD FORMAT
   ;; STEP 1.   STORE CONSTRAINTS IN SIMPLEX  TABLEAU A
   (store-submatrix A 1 0 CONSTRAINTS m1 (+ p 1))
   ;;  STEP 4.  INSERT  SLACK VARIABLES ALREADY PERFORMED
   ;; REDUCTION OF PROBLEM TO CANONICAL FORMAT
   ;; STEP  2.  INTRODUCE AUX. VAR. FOR EQ.  CONSTRAINTS 
   (setq n (+ n m1))
   (adjust-array A (list m n) :initial-element 0)
   (do ((j pointaux (+ j 1))
	(i 1 (+ i 1)))
       ((= i m))
     (setf (aref A i j) 1))
   ;; STEP 3.  WRITE OBJ FUNCTION FOR AUXILIARY PROBLEM
   (do ((j pointaux (+ j 1)))
       ((= j n))
    (setf (aref A 0 j) 1))
   ;;   STEP 4.  SUBTRACT ROW[ I ]  FROM ROW[ 0 ], FOR 1 <= I <= M  
   (let (( firstrow (exstract-row A 0 n)))
    (do ((i 1 (+ i 1)))
       ((= i m))
     (nsubtract firstrow (exstract-row A i n) n))
    (store-subrow A 0 0 firstrow n))
   ;; STEP 5.  SET POINTERS TO BASIC VARIABLES IN POINTERS ARRAY
   (do ((j pointaux (+ j 1)))
      ((= j n))
    (if (= (aref A 0 j) 0)
     (do (( i 1 (+ i 1)))
	((= i m))
      (if (= (aref A i j) 1) (setf (svref pointers i) j)))))
   ;; SOLVE AUXILIARY PROBLEM
   ;; STEP 1.  PERFORM SIMPLEX ALGORITHM
   (simplex-algorithm A m n pointers)
   ;; STEP 2.  TEST IF FEASIBLE SOLUTION SET  OF ORIGINAL PROBLEM IS EMPTY 
   (if (/= (aref A 0 0) 0) (return-from build-canonical-form nil))
   ;; STEP 3.  FORCE AUXILIARY VARIABLES TO LEAVE THE BASIS
   (do ((i 1 (+ i 1)))
      ((= i m))
    (if (>= (svref pointers i) pointaux)
     (if (do ((jaux (svref pointers i))
	     (j 1 (+ j 1)))
	    ((= j pointaux) t)
	  (if (and (/= j jaux) (/= (aref A i j) 0))
	  (progn (pivot A m n i j)
	       (setf (svref pointers i) j)
	       (return nil))))
      (progn (delete-row A i m n)
	    (setf (svref pointers i) nil)
	    (setf m (- m 1))
	    (adjust-array A (list m n))))))
   ;; SOLVE ORIGINAL PROBLEM
   ;;   STEP 1.  RESTORE ORIGINAL PROBLEM
   (setq n pointaux)
   (adjust-array A (list m n))
   (values (exstract-submatrix A 1 0 (- m 1) n)  (- m 1) n (coerce (remove nil pointers) 'list))))
