; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; G. Proietti (2/4/1993)
;;;
;;; routines di supporto a Pol-Complex-Faces-limits per il trattamento di
;;; poliedri rappresentati tramite vertici.

(provide 'pol_c_v)

(inpackage kernel)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  calcola i limiti di contenimento di un poliedro sottoposto alla trasformazione
;;;  specificata dalla matrice, sia rappresentato con vertici.
;    Il calcolo e' effettuato solo sulle coordinate specificate, di minimo o massimo
;    a seconda del type-symbol specificato (vedi commenti precedenti)
;    Ritorna la lista dei risultati, posti nello stesso ordine espresso dalla lista
;    di coordinate passata.

(defun calc-limits-pol-on-verts (POL MAT indices type-symbol)
  (LET* ((VERTICI (RECALC-VERTS (GET-VERTICES POL) MAT))
         res
         temp
         )
    (dolist (i indices res)
      (setq temp (mapcar #'(lambda (x) (nth i x)) vertici))
      (cond 
       ((equalp type-symbol 'max) (SETQ res (APPEND res (LIST (GMAX temp)))) )
       ((equalp type-symbol 'min) (SETQ res (APPEND res (LIST (GMIN temp)))) )
       )
      )
    )
  )



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; ricalcola una lista di vertici in coordinate non omogenee in base alla 
;; trasformazione geometrica specificata dalla matrice mat.

; old version

;(defun RECALC-VERTS (verts MAT)
;  (DOTIMES (I (LENGTH verts) verts)
;    (SETF (NTH I verts)
;       (COERCE (MATLIB:MAT-VET-PRODUCT MAT (COERCE (cons 1 (NTH I verts)) 'VECTOR)) 'LIST))
;  )
;)


(defun RECALC-VERTS (verts MAT)
  (let ( (dimspace (car (array-dimensions mat))) ;in effetti e' la dim dello spazio + 1
         (dimobj   (length (car verts))) 
         (temp     nil ) )
    (progn
      (if (> (1- dimspace) dimobj)
        (progn
          (setq temp (make-list (- dimspace dimobj 1) :initial-element 0 ))
          (setq verts (dotimes (i (length verts) verts)
                         (setf (nth i verts) (append (nth i verts) temp))) )
        )
      )
      (setq verts 
            (DOTIMES (I (LENGTH verts) verts)
              (SETF (NTH I verts)
                    ;(COERCE (MATLIB:MAT-VET-PRODUCT MAT (COERCE (cons 1 (NTH I verts)) 'VECTOR)) 'LIST)))
                    (COERCE (MATLIB:MAT-VET-PRODUCT-extended MAT (COERCE (cons 1 (NTH I verts)) 'VECTOR)) 'LIST)))
            )
      (setq verts (mapcar 'cdr verts))
      )
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; preleva, da una lista di vertici nel formato 
;; ( (label (coord) ) .. (label (coord) ) ) le coordinate, ritornando una
;; lista di liste.

(defun GET-VERTICES (POL)
  (MAPCAR #'(LAMBDA (X) (CAR (CDR X))) (GET-VLIST POL))
)
