; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package Kernel                                            ;
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: V. Pascucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require 'f_ext)

(provide 'conv_ext)
(inpackage kernel )   ;package name

(export '(make-convex-cell
	  
	  make-convex-cell-symbol
	  new-convex-cells-table

	  current-convex-cell-counter
	  reset-convex-cell-counter
	  advance-convex-cell-counter
	  advance-convex-cell-counter
	  previous-convex-cell-counter
	  
	  get-cell-identifier
	  get-cell-sign
	  get-cell-faces
	  get-cell-vertices
	  get-cell-internal-adjacences
          get-cell-external-adjacences
	  get-cell-adjacences
          get-cell-open-faces
          get-cell-centroid
          get-cell-cubic-simplexes
	  get-cells
	  get-clist

	  put-cells
	  put-cell

	  build-cube
	  build-simplex
	  ))

(require 'hyperp)
(use-package 'hyperplanes)

(require 'lisp_ext)
(use-package 'lisp-extension)

#+:akcl 
   (load (concatenate 
	  'string user::**simplexn-directory 
	  "/faces-convexes-complexes/new-extrusion.lisp"))

(require 'new_extr)
(use-package 'extrusion)


;-----------------------------------------------------------------------------
; Conversion to double-float for CLISP
;                                             by Marco Conti
;-----------------------------------------------------------------------------

(defun ->double-float (x  
                       &aux ris )
  (if (numberp x) (setq ris (coerce x 'double-float)))
  (if (listp (first x)) 
      (setq ris (mapcar 
                   #'(lambda (y) 
                       (mapcar #'(lambda (z) (coerce z 'double-float)) y)) x)))
  (if (numberp (first x)) 
      (setq ris (mapcar 
                   #'(lambda (y)
                       (coerce y 'double-float)) x)))
 ris)


;-----------------------------------------------------------------------------
; Definition of the structure of the convex-cell
;-----------------------------------------------------------------------------
;(defstruct (convex-cell (:constructor create-convex-cell))
;  (identifier nil :type symbol)
;  (sign '+ :type sign)
;  (faces '() :type list)
;  (vertices '() :type list)
;  (adjacent-convex-cells '() :type convex-cells-list)
;  (internal-adjacences '() :type list))
;-----------------------------------------------------------------------------
; Definition of the structure of the convex-cell extendex with a new
; field "open-faces" to tag the open faces of the cell
;-----------------------------------------------------------------------------
;(defstruct (convex-cell (:constructor create-convex-cell))
;  (identifier nil :type symbol)
;  (sign '+ ;:type sign
;       )
;  (faces '() :type list)
;  ;or a closed face (nil)
;  (vertices '() :type list)
;  (adjacent-convex-cells '(); :type convex-cells-list 
;                        )
;  (internal-adjacences '() :type list) 
;  (open-faces (make-list  (length faces) :initial-element nil)
;             :type list) ;list of booleans required to tag an open face (t)
;)
;(deftype convex-cells-list () '(satisfies convex-cells-list?))
;
;(defun convex-cells-list? (x)
;   (every 'convex-cell-p x))
;
;(defun new-convex-cells-table ()
;  (make-hash-table ))


;-----------------------------------------------------------------------------
; Definition of the structure of the convex-cell extended with new
; fields "external-adjacences" "centroid" and "cubic-simplexes".
; Modified in June 1995 by Conti Marco
;-----------------------------------------------------------------------------
(defstruct (convex-cell (:constructor create-convex-cell))
  (identifier nil :type symbol)
  (sign '+ )
  (faces '() :type list)
  ;or a closed face (nil)
  (vertices '() :type list)
  (adjacent-convex-cells '() )
  (internal-adjacences '() :type list)
  (external-adjacences '() ) 
  (open-faces (make-list  (length faces) :initial-element nil)
	      :type list) ;list of booleans required to tag an open face (t)
  (centroid nil)
  (cubic-simplexes nil))

(deftype convex-cells-list () '(satisfies convex-cells-list?))

(defun convex-cells-list? (x)
   (every 'convex-cell-p x))

(defun new-convex-cells-table ()
  (make-hash-table ))

;-----------------------------------------------------------------------------
; Creates a convex call of given caracteristics
;-----------------------------------------------------------------------------
;(defun make-convex-cell 
;       (identifier
;        sign
;        faces
;        vertices
;        adjacent-convex-cells
;        internal-adjacences)
;  (create-convex-cell 
;   :identifier identifier
;   :sign sign
;   :faces faces
;   :vertices vertices
;   :adjacent-convex-cells adjacent-convex-cells
;   :internal-adjacences internal-adjacences))
;
;-----------------------------------------------------------------------------
; extension with open faces
;-----------------------------------------------------------------------------
;
;(defun make-convex-cell 
;       (identifier
;       sign
;       faces
;       vertices
;       adjacent-convex-cells
;       internal-adjacences
;       &optional open-faces)
;  (let ((open-faces (or open-faces 
;                      (make-list (length faces) :initial-element nil))))
;    (create-convex-cell 
;     :identifier identifier
;     :sign sign
;     :faces faces
;     :vertices vertices
;     :adjacent-convex-cells adjacent-convex-cells
;     :internal-adjacences internal-adjacences
;     :open-faces open-faces)))


;-----------------------------------------------------------------------------
; Extension with "external-adjacences" "centroid" and "cubic-simplexes"
;-----------------------------------------------------------------------------

(defun make-convex-cell 
       (identifier
	sign
	faces
	vertices
	adjacent-convex-cells
	internal-adjacences
       	&optional (external-adjacences nil)
                   open-faces
		  (centroid nil)
		  (cubic-simplexes nil))
  (let ((open-faces (or open-faces 
		       (make-list (length faces) :initial-element nil))))
    (create-convex-cell 
     :identifier identifier
     :sign sign
     :faces faces
     :vertices vertices
     :adjacent-convex-cells adjacent-convex-cells
     :internal-adjacences internal-adjacences
     :external-adjacences external-adjacences
     :open-faces open-faces
     :centroid centroid
     :cubic-simplexes cubic-simplexes)))

;-----------------------------------------------------------------------------
; The character used as prefix in cells symbols
;-----------------------------------------------------------------------------
;(defconstant o-piccolo-on-faces-and-simplices
;(quote #S(KERNEL::POLYHEDRON :M 0
;                      :N 0
;                      :K0 0
;                      :KM 1
;                      :SIGN +
;                      :VLIST NIL
;                      :SLIST ((S0 (+ () ())))
;                      :INFO NIL
;                      :REPRESENTATION KERNEL::ON-FACES
;                      :F 0
;                      :C 0
;                      :KIND-OF-BRICK KERNEL::SIMPLICES
;                      :GENERAL-CELLS NIL)))

;-----------------------------------------------------------------------------
; The character used as prefix in cells symbols
;-----------------------------------------------------------------------------
(defvar $$convex-cell-prefix "C")

;-----------------------------------------------------------------------------
;create the symbol Ci with the given i
; es: i=5 -> C5 
;-----------------------------------------------------------------------------
(defun make-convex-cell-symbol (i)
  (if (< i 0)
    (intern 
     (concatenate 
      'string $$convex-cell-prefix (integer->string i)))
    (intern 
     (concatenate 
      'string $$convex-cell-prefix (integer->string i)))))

;-----------------------------------------------------------------------------
;  counter functions for convex cells
;-----------------------------------------------------------------------------
(defvar $$convex-cell-counter 0)

(defun current-convex-cell-counter ()
  (make-convex-cell-symbol $$convex-cell-counter))

(defun advance-convex-cell-counter ()
  (prog1
    (make-convex-cell-symbol $$convex-cell-counter)
    (setq $$convex-cell-counter (+ $$convex-cell-counter 1))))

(defun reset-convex-cell-counter (&optional (n 0))
  (setq $$convex-cell-counter  n))

(defun previous-convex-cell-counter ()
  (make-convex-cell-symbol (- $$convex-cell-counter 1)))

;-----------------------------------------------------------------------------
; Returns the name of the cell
;-----------------------------------------------------------------------------
(defun get-cell-identifier (cell)
  (convex-cell-identifier cell))

;-----------------------------------------------------------------------------
; Returns the sign of the cell
;-----------------------------------------------------------------------------
(defun get-cell-sign (cell)
  (convex-cell-sign cell))

;-----------------------------------------------------------------------------
; Returns the list of faces of the convex cell
;-----------------------------------------------------------------------------
(defun get-cell-faces (cell)
  (convex-cell-faces cell))

;-----------------------------------------------------------------------------
; Returns the list of vertices of the convex cell
;-----------------------------------------------------------------------------
(defun get-cell-vertices (cell)
  (convex-cell-vertices cell))

;-----------------------------------------------------------------------------
; Returns the list of internal adjacences of the convex cell
;-----------------------------------------------------------------------------
(defun get-cell-internal-adjacences (cell)
  (convex-cell-internal-adjacences cell))

;-----------------------------------------------------------------------------
; Returns the list of adjacences of the convex cell
;-----------------------------------------------------------------------------
(defun get-cell-adjacences (cell)
  (convex-cell-adjacent-convex-cells cell))

;-----------------------------------------------------------------------------
; Returns the list of external-adjacences of the convex cell
;-----------------------------------------------------------------------------
(defun get-cell-external-adjacences (cell)
  (convex-cell-external-adjacences cell))

;-----------------------------------------------------------------------------
; Returns the list of open-faces of the convex cell
;-----------------------------------------------------------------------------
(defun get-cell-open-faces (cell)
  (convex-cell-OPEN-FACES cell))

;-----------------------------------------------------------------------------
; Returns the centroid of the convex cell
;-----------------------------------------------------------------------------
(defun get-cell-centroid (cell)
  (convex-cell-CENTROID cell))

;-----------------------------------------------------------------------------
; Returns the hash table of the cubic simplexes of the convex cell
;-----------------------------------------------------------------------------
(defun get-cell-cubic-simplexes (cell)
  (convex-cell-CUBIC-SIMPLEXES cell))

;-----------------------------------------------------------------------------
; Returns the cell of name cell-id of the polyhedron pol
;-----------------------------------------------------------------------------
(defun get-polyhedron-cell (pol cell-id)
  (gethash cell-id (get-cells pol)))

;-----------------------------------------------------------------------------
; Returns the list of adjacences of the convex cell
;-----------------------------------------------------------------------------
;(defun get-polyhedron-convex-cells (pol)
;  (polyhedron-convex-cells pol))

;-----------------------------------------------------------------------------
; Returns the hash table of the convex cells of the polyhedron pol
;-----------------------------------------------------------------------------
(defun get-cells (pol)
  (polyhedron-convex-cells pol))

;-----------------------------------------------------------------------------
; Returns the list of the convex cells of the polyhedron pol
;-----------------------------------------------------------------------------
(defun get-clist (pol &optional (sorted nil))
  (let ((clist nil))
    (declare (special clist))
    (maphash #'(lambda (name cell)
		(push cell clist))
	     (get-cells pol))
    (if sorted 
      (sort clist #'(lambda (f1 f2)
		     (string< (get-cell-identifier f1)
			      (get-cell-identifier f2))))
      clist)))

;-----------------------------------------------------------------------------
; Returns a convex cell equal to the simplex
;-----------------------------------------------------------------------------
(defun convert-simplex-into-covex-cell (simplex representation)
  (let ((faces (get-vert simplex))
	(adjacences (get-ad simplex))
	(internal-adjacences nil))
    (if (listp (first faces))
      (setf faces (mapcar 'first faces)))
    (dotimes (i (length adjacences))
      (push (but-nth i (copy-tree faces)) internal-adjacences))
    (cond
     ((symbol-equal representation 'on-vertices)
      (make-convex-cell 
			;representation
			(convert-simplex-symbol-into-covex-cell-symbol
			 (get-id simplex))
			(get-sign simplex)
			(nreverse internal-adjacences) ; nil
			(get-vertices-or-faces simplex)
			(mapcar 'convert-simplex-symbol-into-covex-cell-symbol
				adjacences)
			nil ; (nreverse internal-adjacences)
			))
     ((symbol-equal representation 'on-faces)
      (make-convex-cell 
			;representation
			(convert-simplex-symbol-into-covex-cell-symbol
			 (get-id simplex))
			(get-sign simplex)
			(get-vertices-or-faces simplex)
			(nreverse internal-adjacences) ; nil
			(mapcar 'convert-simplex-symbol-into-covex-cell-symbol
				adjacences)
			nil ; (nreverse internal-adjacences)
			)))))

;-----------------------------------------------------------------------------
; Returns a convex cell equal to the simplex
;-----------------------------------------------------------------------------
(defun convert-simplex-symbol-into-covex-cell-symbol
       (simplex-symbol)
  (if (symbolp simplex-symbol)
    (make-symbol 
     (concatenate 
      'string $$convex-cell-prefix
      (string-left-trim "S" (string simplex-symbol))))
    simplex-symbol))

;-----------------------------------------------------------------------------
; Puts the given convex cell in the polyhedron
; returns the modified polyhedron.
;-----------------------------------------------------------------------------
(defun put-cell (cell polyhedron)
  (setf (gethash (get-cell-identifier cell)
		 (get-cells polyhedron))
	cell)
  (set-#c polyhedron)
  polyhedron)

;------------------------------------------------------------------
; put the cells in the polyhedron pol.
; DESTRUCTIVE :
; the polyhedron is modified
; if the optional parameter remove-old is t, the old cells are lost
; returns the modified polyhedron.
;------------------------------------------------------------------
(defun put-cells (cells-list-or-hash-table pol &optional (remove-old nil))
  (if (listp cells-list-or-hash-table)
    (progn
      (when remove-old (setf (polyhedron-convex-cells pol)
			     (new-convex-cells-table)))
      (mapcar #'(lambda (cell)
		 (put-cell cell pol))
	      cells-list-or-hash-table))
    (progn
      (if remove-old 
	(setf (polyhedron-convex-cells pol) cells-list-or-hash-table)
	(maphash #'(lambda (name cell)
		    (put-cell cell pol))
		 cells-list-or-hash-table))))
  (set-#c pol)
  pol)
;-----------------------------------------------------------------------------
; Returns the polyhedron equal to pol but structured in convex cells.
; The representation is based on faces.
;-----------------------------------------------------------------------------
(defun coerce-polyhedron-on-faces-from-simplices-to-convex-cells (pol)
  (let ((new-polyhedron (make-empty-polyhedron 'on-faces 'convex-cells))
	(simplices (get-slist pol)))
    (mapcar #'(lambda (cell)
	       (put-cell cell new-polyhedron))
	    (mapcar #'(lambda (simplex)
		       (convert-simplex-into-covex-cell simplex 'on-faces))
		    simplices))
    (put-faces (get-flist pol) new-polyhedron t)
    new-polyhedron))
    
;-----------------------------------------------------------------------------
; Returns the polyhedron equal to pol but structured in convex cells.
; The representation is based on vertices.
;-----------------------------------------------------------------------------
(defun coerce-polyhedron-on-vertices-from-simplices-to-convex-cells (pol)
  (let ((new-polyhedron (make-empty-polyhedron 'on-vertices 'convex-cells))
	(simplices (get-slist pol)))
    (mapcar #'(lambda (cell)
	       (put-cell cell new-polyhedron))
	    (mapcar #'(lambda (simplex)
		       (convert-simplex-into-covex-cell simplex 'on-vertices))
		    simplices))
    (maphash #'(lambda (name cell)
		(put-cell cell new-polyhedron))
	     (get-cells pol))
    (put-vlist (copy-tree (get-vlist pol)) new-polyhedron)
    new-polyhedron))

;-----------------------------------------------------------------------------
; Returns the polhedron equal to pol but with a representation based on
; faces.
;-----------------------------------------------------------------------------
(defun coerce-polyhedron-on-convex-cells-from-vertices-to-faces (pol)
  (let ((old-cells (get-clist pol))
	(new-cells (new-convex-cells-table))
	(new-faces (new-faces-table))
	new-polyhedron)
    (reset-face-counter)
    (dolist (current-old-cell old-cells)
      (let* ((old-cell-vertices (get-cell-vertices current-old-cell))
	     (cell-adjacences (get-cell-adjacences  current-old-cell))
	     (cell-open-faces (convex-cell-open-faces current-old-cell))
	     (old-cell-faces (get-cell-faces current-old-cell))
	     (new-cell-vertices nil)
	     (new-cell-faces nil)
	     (cell-identifier (get-cell-identifier current-old-cell))
	     (cell-sign (get-cell-sign current-old-cell))
	     (vertices-list (mapcar #'(lambda (vid)
				       (get-vertex vid pol))
				    old-cell-vertices))
	     (vertices-list-list (list vertices-list)))
	(setf (cdr vertices-list-list) vertices-list-list)
	(dotimes (current-adjacent-index (length cell-adjacences))
	  (push 
	   (or 
	    ;            (get-adjacent-face (find-el  (nth current-adjacent-index cell-adjacences)
	    ;                                        new-slist)
	    ;                              cell-identifier
	    ;                              cell-sign)
	    (add-new-face new-faces 
			  (->double-float (mapcar 'find-vcoord 
				  (nth current-adjacent-index old-cell-faces)
				  vertices-list-list))
			  (find-vcoord 
			   (first (set-difference old-cell-vertices
						  (nth current-adjacent-index 
						       old-cell-faces)))
			   ;      (nth current-adjacent-index vertices-list)
			   vertices-list)
			  cell-sign))
	   new-cell-faces))
	(setf new-cell-faces (nreverse new-cell-faces))

	(dotimes (i (length old-cell-vertices))
	  (let ((old-vertex-name (nth i old-cell-vertices))
		(new-vertex nil))
	    (dotimes (j (length old-cell-faces))
	      (if (member old-vertex-name (nth j old-cell-faces))
		(push (car (nth j new-cell-faces)) new-vertex)))
	    (push new-vertex new-cell-vertices)))

	(setf (gethash  cell-identifier new-cells)
	      (make-convex-cell cell-identifier 
				cell-sign
				new-cell-faces
				new-cell-vertices
				cell-adjacences
				nil 
				cell-open-faces))))
    (setf new-polyhedron
	  (make-empty-polyhedron 'on-faces 'convex-cells))
    (put-faces new-faces new-polyhedron t)
    (put-cells new-cells new-polyhedron t)
    new-polyhedron))

;-----------------------------------------------------------------------------
; Returns the polhedron equal to pol but with a representation based on
; vertices.
;-----------------------------------------------------------------------------
(defun coerce-polyhedron-on-convex-cells-from-faces-to-vertices (pol)
  (let ((old-cells (get-clist pol))
	(new-cells (new-convex-cells-table))
	(old-faces-list (get-flist pol))
	(new-vertices nil)
	new-polyhedron)
    (reset-vertex-counter)
    (dolist (current-old-cell old-cells)
      (let* ((old-cell-vertices (get-cell-vertices current-old-cell))
	     (cell-adjacences (get-cell-adjacences  current-old-cell))
	     (old-cell-faces (get-cell-faces current-old-cell))
	     (cell-open-faces (convex-cell-open-faces current-old-cell))
	     (new-cell-vertices nil)
	     (new-cell-faces nil)
	     (cell-identifier (get-cell-identifier current-old-cell))
	     (cell-sign (get-cell-sign current-old-cell))
	     (faces-list (mapcar #'(lambda (fid)
				       (assoc fid old-faces-list)) ;eliminato quote davanti old-faces-list
				    (mapcar 'first old-cell-faces)))
	     (faces-list-list (list faces-list)))
	(setf (cdr faces-list-list) faces-list-list)
	(dotimes (current-vertex-index (length old-cell-vertices))
	  (push 
	    (add-new-vertex new-vertices 
			    (->double-float (mapcar 'find-fcoord  
				    (nth current-vertex-index old-cell-vertices)
				    faces-list-list)))
	   new-cell-vertices))

	(setf new-cell-vertices (nreverse new-cell-vertices)) ;aggiunto per disperazione

	(dotimes (i (length old-cell-faces))
	  (let ((old-face-name  (first (nth i old-cell-faces)))
		(new-face nil))
	    (dotimes (j (length old-cell-vertices))
	      (if (member old-face-name (nth j old-cell-vertices))
		(push (nth j new-cell-vertices) new-face)))
	    (push new-face new-cell-faces)))

	(setf new-cell-faces (nreverse new-cell-faces)) ;aggiunto per disperazione

	(setf (gethash  cell-identifier new-cells)
	      (make-convex-cell cell-identifier 
				cell-sign
				new-cell-faces
				new-cell-vertices
				cell-adjacences
				nil
				cell-open-faces))))
    (setf new-polyhedron
	  (make-empty-polyhedron 'on-vertices 'convex-cells))
    (put-vlist new-vertices new-polyhedron t)
    (put-cells new-cells new-polyhedron t)
    new-polyhedron))
    
;-----------------------------------------------------------------------------
; Returns a polyhedron composed of a convex cell equal to the standard 
; cube of dimension n.
;-----------------------------------------------------------------------------
(defun build-cube (n)
  (declare (special o-piccolo-on-faces-and-convex-cells))
  (let ((new-polyhedron o-piccolo-on-faces-and-convex-cells))
    (dotimes (i n)
      (setf new-polyhedron (general-extrude new-polyhedron)))
    new-polyhedron))


;-----------------------------------------------------------------------------
; Returns a polyhedron composed of a convex cell equal to the standard 
; simplex of dimension n.
;-----------------------------------------------------------------------------
(defun build-simplex (n)
  (if (< n 2)
    (build-cube n)
    (let ((new-polyhedron (make-empty-polyhedron 'on-faces 'convex-cells))
	  (faces-names nil)
	  (vertices nil)
	  (faces nil)
	  (zero-list (make-list n :initial-element 0))
	  face-par)
      (reset-face-counter) 
      (dotimes (i n)
	(setf face-par zero-list)
	(push-nth (1+ i) 1.0 face-par)
	(put-face (make-face (advance-face-counter) face-par) new-polyhedron)
	(push (previous-face-counter) faces-names)
	(push (cons (previous-face-counter) '-) faces))
      (setf face-par (make-list n :initial-element -1))
      (push 1.0 face-par)
      (put-face (make-face (advance-face-counter) 
			   (first (normalize-parameters face-par)))
		new-polyhedron)
      (push (previous-face-counter) faces-names)
      (push (cons (previous-face-counter) '-) faces)
      (dotimes (i (1+ n))
	(push (but-nth i faces-names) vertices))
      (put-cell (make-convex-cell (make-convex-cell-symbol 0)  
				  '+
				  faces
				  vertices
				  (push 0 zero-list)
				  nil)
		new-polyhedron)
	(set-npol new-polyhedron n)
	(set-dpol new-polyhedron n)
	(set-#kd  new-polyhedron)
	(set-#k0  new-polyhedron) 
	(set-#F  new-polyhedron)
	(set-#C  new-polyhedron) 
      new-polyhedron)))

;-----------------------------------------------------------------------------
; The 0-dimensional polyhedron
;-----------------------------------------------------------------------------
(defvar o-piccolo-on-faces-and-convex-cells
    (put-cells (list (make-CONVEX-CELL 'C0
					'+
				       NIL
				       '(())
				       NIL
				       NIL))
	       (make-empty-polyhedron 'on-faces 'convex-cells)
	       t))
