;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Il seg. modulo contiene le funzioni per settare le adiacenze  ;;
;; esterne alle celle, ossia non appartenenti allo stesso        ;;
;; poliedro. Inoltre calcola il valore della funzione sul        ;;
;; centroide di ogni cella.                                      ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(provide 'extadiac)

(inpackage kernel)


;-------------------------------------------------------------------------
; La seg. funzione visita la struttura di complesso poliedrale
; ------------------------------------------------------------------------

(defun Set-External-Adjacences (obj Ctable function)
  (cond ((polyhedron?      obj) (Set-Ext-Ad-Pol obj Ctable function))
        ((polyhedra-complex? obj)
          (mapcar  #'(lambda (nodo) 
            (Set-External-Adjacences nodo Ctable function)) obj))
        ((polyhedral-node? obj)
         (progn
           (make-pol-node 
             (Set-External-Adjacences 
                  (get-node-pol-complex obj) 
                   Ctable 
                   function)
             (get-node-matrix obj)
             (get-node-id obj) ))) ))


;-------------------------------------------------------------------------
; La seg. funzione visita le celle del poliedro dato in input
;-------------------------------------------------------------------------

(defun Set-Ext-Ad-Pol (pol Ctable function)
  (let ((cells (get-cells pol)))
    (maphash #'(lambda (id cell)
      (let* ((adiac (get-cell-adjacences cell))
             (ext-adiac (make-list (length adiac) 
                         :initial-element 0))
             (coord (get-cell-vertices-coord (get-cell-vertices cell) pol)))
        (dotimes (i (length adiac))
          (if (not (equal (nth i adiac) 0))
            (setf (nth i ext-adiac) (list (nth i adiac) 
                                          (gethash (nth i adiac) Ctable)))))
        (setf (convex-cell-EXTERNAL-ADJACENCES cell) ext-adiac)
        (setf (convex-cell-CENTROID cell) 
          ;; (plasm::reduce-function-args function (calculate-centroid coord))
         nil)))
     cells)
                 ;; restituisce il poliedro con le adiacenze esterne 
                 ;; ben settate
  pol ))









