(inpackage kernel)

(require 'mat_C1)

;-------------------------------------------------------------
; La seg. funzione riceve la lista delle coord. di un     
; punto dello spazio e le componenti di un vettore, e      
; restituisce i coeff. dell' equazione del piano passante  
; per il punto, perpendicolare al vettore dato.          
; L'eq. del piano e' rappresentata come:                    
; a*x + b*y + c*z + d = 0                                 
; Point normal plane sono vettori riga.                    
;-------------------------------------------------------------

(defun plane-point-nvector (point vector)  
  (let ((plane (make-array '(4))))
    (setf (aref plane 0) (aref vector 0))
    (setf (aref plane 1) (aref vector 1))
    (setf (aref plane 2) (aref vector 2))
    (setf (aref plane 3) (- (+ (* (aref vector 0) (nth 0 point ))
                               (* (aref vector 1) (nth 1 point ))
                               (* (aref vector 2) (nth 2 point )))))
    plane))


;-------------------------------------------------------------
; La seg. funzione riceve la lista delle coord. di un     
; punto dello spazio ed un vettore di coeff. dell'equaz.   
; di un piano e restituisce la retta per il punto perpend. 
; al piano. La retta e' espressa come intersez. di due piani
; I coeff. di tali equaz. eghuagliate a 0 sono restituiti in 
; un array (2,4)                                                    
;-------------------------------------------------------------


(defun line-point-nplane (point plane)                              
  (let ((line (make-array '(2 4) 
                 :initial-element 0.0d0 
                 :element-type 'double-float)))
    (cond 
      ( (/= (aref plane 2) 0)
          (setf (aref line 0 0) (aref plane 2))
          (setf (aref line 0 2) (- (aref plane 0)))
          (setf (aref line 0 3) (- (* (nth 2 point ) (aref plane 0))
                                   (* (nth 0 point ) (aref plane 2))))
          (setf (aref line 1 1) (aref plane 2))
          (setf (aref line 1 2) (- (aref plane 1)))
          (setf (aref line 1 3) (- (* (nth 2 point ) (aref plane 1))
                                   (* (nth 1 point ) (aref plane 2)))) )
      ( (/= (aref plane 1) 0)
          (setf (aref line 0 0) (aref plane 1))
          (setf (aref line 0 1) (- (aref plane 0)))
          (setf (aref line 0 3) (- (* (nth 1 point ) (aref plane 0))
                                   (* (nth 0 point ) (aref plane 1))))
          (setf (aref line 1 2) (aref plane 1))
          (setf (aref line 1 1) (- (aref plane 2)))
          (setf (aref line 1 3) (- (* (nth 1 point ) (aref plane 2))
                                   (* (nth 2 point ) (aref plane 1)))) )

      ( (/= (aref plane 0) 0)
          (setf (aref line 0 1) (aref plane 0))
          (setf (aref line 0 0) (- (aref plane 1)))
          (setf (aref line 0 3) (- (* (nth 0 point ) (aref plane 1))
                                   (* (nth 1 point ) (aref plane 0))))
          (setf (aref line 1 2) (aref plane 0))
          (setf (aref line 1 0) (- (aref plane 2)))
          (setf (aref line 1 3) (- (* (nth 0 point ) (aref plane 2))
                                   (* (nth 2 point ) (aref plane 0)))) ))
    line))


;-------------------------------------------------------------
; La seg. funzione effettua la proiezione 
; di un punto su un piano secondo la direzione della retta 
; data. Al solito point e' una lista di 3 elem.,      
; plane e' un vettore riga di 4 elem., line e' una matrice     
; (2,4)                                                      
;-------------------------------------------------------------

(defun point-projection (point plane line)
  (let ((coeff (make-array '(3 3)))
        (noti  (make-array '(3 1))))
    (dotimes (i 3)
      (if (= i 0)
          (setf (aref noti i 0) (- (aref plane  3))) 
          (setf (aref noti i 0) (- (aref line  (1- i) 3))))
      (dotimes (j 3)
        (if (= i 0)    
            (setf (aref coeff 0 j) (aref plane  j)) 
            (setf (aref coeff i j) (aref line (1- i) j)))))
    (soluz-sistema coeff noti)))

;-------------------------------------------------------------
; La seg. funzione effettua la proiezione ortogonale
; di un punto su un piano.                                                     
;-------------------------------------------------------------

(defun ortogonal-projection (point plane)
  (let ((coeff (make-array '(3 3)))
        (noti  (make-array '(3 1)))
        (line (line-point-nplane point plane)))
    (dotimes (i 3)
      (if (= i 0)
          (setf (aref noti i 0) (- (aref plane  3))) 
          (setf (aref noti i 0) (- (aref line  (1- i) 3))))
      (dotimes (j 3)
        (if (= i 0)    
            (setf (aref coeff 0 j) (aref plane  j)) 
            (setf (aref coeff i j) (aref line (1- i) j)))))
    (soluz-sistema coeff noti)))


;-------------------------------------------------------------
; La seg. funzione riceve la lista delle coord. di un     
; punto dello spazio ed un vettore rappresentante una 
; direzione e restituisce la retta per il punto parallela 
; alla direzione data. La retta e' espressa come intersez. 
; di due piani
; I coeff. di tali equaz. eghuagliate a 0 sono restituiti in 
; un array (2,4)                                                    
;-------------------------------------------------------------


(defun line-point-par-direction (point direction)                           
  (let ((line (make-array '(2 4) 
                 :initial-element 0.0d0 
                 :element-type 'double-float)))
    (setf (aref line 0 0) (aref direction 2))
    (setf (aref line 0 2) (- (aref direction 0)))
    (setf (aref line 0 3) (- (* (nth 2 point ) (aref direction 0))
                             (* (nth 0 point ) (aref direction 2))))
    (setf (aref line 1 1) (aref direction 2))
    (setf (aref line 1 2) (- (aref direction 1)))
    (setf (aref line 1 3) (- (* (nth 2 point ) (aref direction 1))
                             (* (nth 1 point ) (aref direction 2))))
    line))


;-------------------------------------------------------------
; La seg. funzione effettua l'intersezione tra una retta e 
; un piano. Restituisce il punto di intersezione che e' una 
; lista di 3 elem.      
; Plane e' un vettore riga di 4 elem., line e' una matrice     
; (2,4)                                                      
;-------------------------------------------------------------

(defun line-plane-intersection (plane line)
  (let ((coeff (make-array '(3 3)))
        (noti  (make-array '(3 1))))
    (dotimes (i 3)
      (if (= i 0)
          (setf (aref noti i 0) (- (aref plane  3))) 
          (setf (aref noti i 0) (- (aref line  (1- i) 3))))
      (dotimes (j 3)
        (if (= i 0)    
            (setf (aref coeff 0 j) (aref plane  j)) 
            (setf (aref coeff i j) (aref line (1- i) j)))))
    (soluz-sistema coeff noti)))


;--------------------------------------------------------------
; La seg. funzione controlla se una retta e' parallela ad un
; piano. Al solito, i coeff. di un'equazione di una retta sono 
; memorizzati in una matrice (2,4) e quelli del piano in una
; matrice (1,4).
;--------------------------------------------------------------

(defun line-parallel-plane? (plane line
                             &aux YorN)
  (let ((mat (make-array '(3 3)))
        (eps 1.0d-14))
    (dotimes (i 3)
      (dotimes (j 3)
        (if (= i 0)
          (setf (aref mat 0 j)(aref plane j))
          (setf (aref mat i j)(aref line (1- i) j)))))
    (if (< (det_3 mat) eps)
      (setq YorN T)
      (setq YorN nil))
   YorN))
        
;--------------------------------------------------------------
; La seg. funzione controlla se due vettori sono paralleli.
; I vettori sono rappresentati con array.
;--------------------------------------------------------------

(defun vect-parallel-vect? (vet1 vet2
                            &aux (YorN T))
  (let ((index-zero (comp-zero vet2))
        lista)
    (if (equal index-zero nil) ;Vet2 ha tutte componenti 
                                     ;diverse da 0

     (progn 
       (dotimes (i (first (array-dimensions vet1)))
         (setq lista (append lista (list (/ (aref vet1 i) (aref vet2 i))))))
       (dotimes (i (1- (first (array-dimensions vet1))))
         (if (not (approx-equal (nth i lista) (nth (1+ i) lista)))
           (setq YorN nil))))
                                     ;else Vet2 ha componenti nulle
 
     (progn 
       (let ((ref (- (aref vet1 index-zero) (aref vet2 index-zero))))
         (dotimes (i (1- (first (array-dimensions vet1))))
           (if (not (approx-equal ref (- (aref vet1 i)(aref vet2 i))))
             (setq YorN nil))))))
  YorN))

;--------------------------------------------------------------
; La seg. funzione definisce un'uguaglianza approssimata
;--------------------------------------------------------------

(defun approx-equal (a b &aux YorN)
  (let ((eps 1.0d-4))
    (if (< (abs (- a b)) eps)
      (setq YorN T)
      (setq YorN nil))))


;--------------------------------------------------------------
; La seg. funzione controlla se in un vettore rappresentato
; tramite array e' presente un elemento nullo.
; In caso affermativo restituisce l'indice dell'ultimo
; 0 trovato.
; Nel caso che tutte le sue componenti siano non nulle 
; restituisce nil.
;--------------------------------------------------------------

(defun comp-zero (vet)
  (let ((eps 1.0d-10)
        (index nil))
    (dotimes (i (first (array-dimensions vet)))
      (if (< (abs (aref vet i)) eps)
        (setq index i)))
   index))     

;--------------------------------------------------------------
; La seg. funzione calcola il baricentro di n punti.
; Riceve in input la lista delle coordinate di essi.
;--------------------------------------------------------------

(defun baricenter (list-coord)
  (let* ((n (length list-coord))
         (m (length (first list-coord)))
         (sum (make-list m :initial-element 0.0d0))
         (baric (make-list m :initial-element 0.0d0))
          point
          temp)
    (dotimes (i n)
      (setq point (nth i list-coord))
      (dotimes (j m)
        (setf (nth j sum) (+ (nth j sum) (nth j point)))))
    (dotimes (j m)
      (setf (nth j baric) (/ (nth j sum) n)))
    baric)) 

;--------------------------------------------------------------
;La seg. funzione restituisce T se il coseno dell'angolo fra i 
;due vettori dati in input e' minore di un certo valore fissato. 
;-------------------------------------------------------------

(defun approx-vect-parallel-vect? (vet1 vet2
                                   &aux YorN)
  (if (> (abs (cos-angle vet1 vet2)) 0.98)
    (setq YorN T)
    (setq YorN nil)))

;-------------------------------------------------------------
;La seg. funzione calcola il coseno dell'angolo dei due 
;vettori dati in input.
;-------------------------------------------------------------

(defun cos-angle (vet1 vet2)
  (/ (scalar-prod vet1 vet2)(* (vect-mod vet1)(vect-mod vet2))))





