(provide 'io_ext3.lsp)

(require 'cub_ext)

(inpackage io)

(export '(export-pol-complex-nrb))

;---------------------------------------------------------------------------
; La seg. funzione esporta la rete dei punti di controllo dei simplessi 
; cubici di un poliedro complesso.
;---------------------------------------------------------------------------

(defun export-pol-complex-nrb (pol-complex file-name)
  (let* ((new-file-name file-name) ;(concatenate 'string file-name ".nrb"))
         (stream (open new-file-name :direction :output)))
    (princ "B-spline triangular-surface-patch: output PLASM" stream)
    (Visit-Pol-Complex-nrb pol-complex stream)
    (terpri stream)
    (close stream)))   


;-------------------------------------------------------------------------
; La seg. funzione visita la struttura di complesso poliedrale
; ------------------------------------------------------------------------

(defun Visit-Pol-Complex-nrb (obj stream)
  (cond ((polyhedron?      obj) (Visit-Pol-nrb obj stream))
        ((polyhedra-complex? obj)
          (mapcar  #'(lambda (nodo) (Visit-Pol-Complex-nrb nodo stream)) obj))
        ((polyhedral-node? obj)
         (progn
           (make-pol-node 
             (Visit-Pol-Complex-nrb (get-node-pol-complex obj) stream)
             (get-node-matrix obj)
             (get-node-id obj) ))) ))


;-------------------------------------------------------------------------
; La seg. funzione visita le celle ed i simplessi del poliedro 
; dato in input
;-------------------------------------------------------------------------

(defun Visit-Pol-nrb (pol stream)
  (let ((cells (get-cells pol)))
     (maphash #'(lambda (id-cell cell)
       (let* ((simplexes (get-cell-cubic-simplexes cell)))
         (maphash #'(lambda (id-simplex simplex)
            (print-nrb-on-file simplex stream))
          simplexes)))
      cells)))
           
;-------------------------------------------------------------------------
; La seg. funzione stampa la rete dei punti di controllo di un simplesso 
; su file.
;-------------------------------------------------------------------------

(defun print-nrb-on-file (simplex stream)
  (let ((points-net (kernel::cubic-simplex-BEZIER-NET simplex))
        (rect-net (make-array '(4 4) )))

    (setf (aref rect-net 0 0) 
          (gethash (kernel::make-id-coord-baric '(3 0 0)) points-net))
    (setf (aref rect-net 0 1) 
          (gethash (kernel::make-id-coord-baric '(2 1 0)) points-net))
    (setf (aref rect-net 0 2) 
          (gethash (kernel::make-id-coord-baric '(1 2 0)) points-net))
    (setf (aref rect-net 0 3) 
          (gethash (kernel::make-id-coord-baric '(0 3 0)) points-net))


    (setf (aref rect-net 1 0) 
          (gethash (kernel::make-id-coord-baric '(2 0 1)) points-net))
    (setf (aref rect-net 1 1) 
          (gethash (kernel::make-id-coord-baric '(1 1 1)) points-net))
    (setf (aref rect-net 1 2) 
          (gethash (kernel::make-id-coord-baric '(1 1 1)) points-net))
    (setf (aref rect-net 1 3) 
          (gethash (kernel::make-id-coord-baric '(0 2 1)) points-net))


    (setf (aref rect-net 2 0) 
          (gethash (kernel::make-id-coord-baric '(1 0 2)) points-net))
    (setf (aref rect-net 2 1) (int-point points-net))
    (setf (aref rect-net 2 2) (int-point points-net))
    (setf (aref rect-net 2 3) 
          (gethash (kernel::make-id-coord-baric '(0 1 2)) points-net))
 

    (setf (aref rect-net 3 0) 
          (gethash (kernel::make-id-coord-baric '(0 0 3)) points-net))
    (setf (aref rect-net 3 1) 
          (gethash (kernel::make-id-coord-baric '(0 0 3)) points-net))
    (setf (aref rect-net 3 2) 
          (gethash (kernel::make-id-coord-baric '(0 0 3)) points-net))
    (setf (aref rect-net 3 3) 
          (gethash (kernel::make-id-coord-baric '(0 0 3)) points-net))


    ;scrittura file
     
    ;(princ "B-spline triangular-surface-patch: output PLASM" stream)
    (terpri stream)
    (princ " " stream)
    (princ 4 stream)
    (princ " " stream)
    (princ 4 stream)
    (terpri stream)
    (princ " " stream)
    (princ 4 stream)
    (princ " " stream)
    (princ 4 stream)
    (terpri stream)
    (princ "Control points:" stream)
    (terpri stream)
    (dotimes (i 4)
      (dotimes (j 4)
        (dotimes (k 3)
          (format stream "~16,10F" 
              (coerce (nth k (aref rect-net i j)) 'double-float))
          (princ " " stream))
        (format stream "~16,10F" 1.0d0);coord. omogenea
        (terpri stream))
       (if (not (= i 3))
         (terpri stream)))
     (princ "U-Knots" stream)
     (terpri stream)
     (dotimes (i 4)
       (princ " " stream)
       (format stream "~16,10F" 0.0d0)) 
     (dotimes (i 4)
       (princ " " stream)
       (format stream "~16,10F" 1.0d0))
     (terpri stream)
     (princ "U-parameter-range" stream)
     (terpri stream)
     (format stream "~16,10F" 0.0d0)
     (princ " " stream)
     (format stream "~16,10F" 1.0d0)
     (terpri stream)
     (princ "V-Knots" stream)
     (terpri stream)
     (dotimes (i 4)
       (princ " " stream)
       (format stream "~16,10F" 0.0d0)) 
     (dotimes (i 4)
       (princ " " stream)
       (format stream "~16,10F" 1.0d0))
     (terpri stream)
     (princ "V-parameter-range" stream)
     (terpri stream)
     (format stream "~16,10F" 0.0d0)
     (princ " " stream)
     (format stream "~16,10F" 1.0d0)
     (terpri stream)
     (princ "====== Trimming Curves ======" stream)
     (terpri stream)
     (princ 0 stream)
 
     (princ ".")))
        
       
;-------------------------------------------------------------------------
;La seg. funzione ricava un punto di controllo ausiliario per il
;patch rettangolare degenere.
;-------------------------------------------------------------------------

(defun int-point (points-net)
  (let* ((vet (kernel::points->vect 
                 (gethash (kernel::make-id-coord-baric '(0 0 3)) points-net)
                 (gethash (kernel::make-id-coord-baric '(1 1 1)) points-net)))
         (point1 (gethash (kernel::make-id-coord-baric '(1 0 2)) points-net))
         (point2 (gethash (kernel::make-id-coord-baric '(1 1 1)) points-net))
         (plane (kernel::plane-point-nvector point1 vet)))
         
     (kernel::ortogonal-projection point2 plane)))

    













