

(require 'matlib)

(inpackage kernel)

(provide 'mat_C1)

(export '(det_2
          det_3
          vect-prod
          sostituz-colonna
          soluz-sistema
          scalar-vect-mult
          scalar-prod
          vect-mod
          vect->list
          list->vect
          points->vect
          ))



;-----------------------------------------------------------
; La seg. funzione calcola il determinante di una       
; matrice quadrata (2,2)                                
;-----------------------------------------------------------

(defun det_2 (mat_2)
  (- (* (aref mat_2 0 0) (aref mat_2 1 1)) 
     (* (aref mat_2 0 1) (aref mat_2 1 0))) )

;-----------------------------------------------------------
; La seg. funzione calcola il determinante di una        
; matrice quadrata (3,3) secondo la regola di Sarrus    
;-----------------------------------------------------------

(defun det_3 (mat_3)
  (- (+ (* (aref mat_3 0 0) (aref mat_3 1 1) (aref mat_3 2 2))  
        (* (aref mat_3 0 1) (aref mat_3 1 2) (aref mat_3 2 0))
        (* (aref mat_3 0 2) (aref mat_3 1 0) (aref mat_3 2 1)))
     (+ (* (aref mat_3 0 2) (aref mat_3 1 1) (aref mat_3 2 0))  
        (* (aref mat_3 0 0) (aref mat_3 1 2) (aref mat_3 2 1))
        (* (aref mat_3 0 1) (aref mat_3 1 0) (aref mat_3 2 2)))))

;-----------------------------------------------------------
; La seg. funzione calcola le componenti del vettore     
; risultato del prodotto vettoriale v1 x v2.             
; Esse coincidono con i minori del secondo ordine della  
; matrice ((v11 v12 v13) (v21 v22 v23)) ottenuti         
; cancellando le rispettive colonne e presi con i segni  
; + - +                                                  
;-----------------------------------------------------------


(defun vect-prod (vet1 vet2)
  (let ((prod (make-array '(3)  
               :initial-element 0.0d0))
        (temp (make-array '(2 2)
               :initial-element 0.0d0))
        (v1 (make-array '(3)  
               :initial-element 0.0d0))
        (v2 (make-array '(3)  
               :initial-element 0.0d0)))
    (dotimes (i (length vet1))
      (setf (aref v1 i) (aref vet1 i))
      (setf (aref v2 i) (aref vet2 i))) 
    (dotimes (i 3 prod)
      (setf (aref temp 0 0) (aref v1 (mod (+ i 1) 3)))
      (setf (aref temp 0 1) (aref v1 (mod (+ i 2) 3)))
      (setf (aref temp 1 0) (aref v2 (mod (+ i 1) 3)))
      (setf (aref temp 1 1) (aref v2 (mod (+ i 2) 3)))
      (setf (aref prod i) (det_2 temp)))))

;-----------------------------------------------------------
; La seg. funzione prende un indice di colonna, una      
; matrice ed un vettore colonna e restituisce una        
; matrice la cui i-esima colonna  il vettore dato.      
;-----------------------------------------------------------

(defun sostituz-colonna (i mat colonna)
  (let* ((dim (array-dimensions mat))
         (mat-ris (make-array dim)))
    (dotimes (k (nth 0 dim) mat-ris)
      (dotimes (j (nth 1 dim))  
        (if (= j i)
          (setf (aref mat-ris k i) (aref colonna k 0))
          (setf (aref mat-ris k j) (aref mat k j))))))) 

;-----------------------------------------------------------
; La seg. funzione risolve i sistemi 3 x 3 con il metodo 
; di Cramer. Riceve una la matrice dei coefficienti delle
; incognite ed il vettore colonna che rappresenta i      
; termini noti. Il sistema  scritto nel modo Ax=b       
;-----------------------------------------------------------

(defun soluz-sistema (coeff noti)
  (let ((den (det_3 coeff))
         soluz)  
    (dotimes (i 3 soluz)
      (setq soluz 
        (append soluz 
          (list 
            (/ (det_3 (sostituz-colonna i coeff noti)) den)))))))

;-----------------------------------------------------------      
; La seg. funzione esegue il prodotto tra un vettore ed uno 
; scalare.
;-----------------------------------------------------------

(defun scalar-vect-mult (scal vet)        
  (let* ((dim (array-dimensions vet))
         (vet-ris (make-array dim :element-type 'double-float)))
    (dotimes (i (first dim))
      (setf (aref vet-ris i) (* scal (aref vet i))))
  vet-ris))

;-----------------------------------------------------------
;La seg. funzione trasforma un array unidimensionale in una
;lista.
;-----------------------------------------------------------

(defun vect->list (vect)
  (let* ((n (first (array-dimensions vect))) 
         (lista (make-list n)))
    (dotimes (i n lista)
      (setf (nth i lista) (aref vect i)))))

;-----------------------------------------------------------
;La seg. funzione trasforma una lista in un array 
; unidimensionale.
;-----------------------------------------------------------

(defun list->vect (lista)
  (let* ((n (length lista)) 
         (vect (make-array n :element-type 'double-float)))
    (dotimes (i n vect)
      (setf (aref vect i)(nth i lista)))))


;-----------------------------------------------------------
;La seg. funzione a partire dai punti P1 P2 rappresentati
;come liste di coordinate, restituisce il vettore P1P2 
;rappresentato come array.
;-----------------------------------------------------------

(defun points->vect (P1 P2)
  (let* ((n-coord (length P1))
         (vect (make-array n-coord :element-type 'double-float)))
    (dotimes (i n-coord)
      (setf (aref vect i) (- (nth i P2) (nth i P1))))
   vect))

;-----------------------------------------------------------
;La seg. funzione calcola la norma euclidea (il modulo)
;di un vettore rappr. tramite array.
;-----------------------------------------------------------

(defun vect-mod (vect)
  (let ((norma 0))
    (dotimes (i (first (array-dimensions vect)))
      (setq norma (+ norma (coerce (expt (aref vect i) 2) 'double-float))))
  (sqrt norma)))

;-----------------------------------------------------------
;La seg. funzione ritorna il prodotto scalare di due
;vettori rappr. tramite array.
;-----------------------------------------------------------

(defun scalar-prod (vect1 vect2)
  (let ((prod 0))
    (dotimes (i (first (array-dimensions vect1)))
      (setq prod (+ prod (coerce (* (aref vect1 i) 
                                    (aref vect2 i)) 'double-float))))
   prod))






