;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Il seguente modulo contiene le funzioni per eseguire la ;;
;; fase 2 dell'algoritmo per ottenere le sup. C1           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(inpackage kernel)

;(require 'geometry)

(require 'cub_ext)

(provide 'step2)


;-----------------------------------------------------------
; La seg. funzione ritorna la lista degli spigoli incidenti 
; il vertice dato in input.
; La lista restituita e' costituita da elementi del tipo:
; ((id-simplesso id-cella) . spigolo)
; Si noti che in genere uno spigolo puo' appartenere a due
; celle ed in questo caso avremo due elementi con lo stesso
; spigolo, ma con id diversi.
;-----------------------------------------------------------

(defun incidence-v (vertex cell pol
                    &optional 
                        ;(list-incidence '())
                        (tab-incidence-cell (make-hash-table :test 'equal)))
  (let* ((adiac (get-cell-adjacences cell))
         (ext-adiac (get-cell-external-adjacences cell))
         (cell-id (get-cell-identifier cell))
         (simplexes (get-cubic-simplexes pol cell-id))
          prox-cell
          prox-pol)
      
  (if (not (in-tab? cell-id tab-incidence-cell))    ;;controlla se la cella
                                                    ;;e' gia' stata esaminata
    (progn
      (setf (gethash cell-id tab-incidence-cell) cell)   

                     ;; Ottiene la prossima cella da esaminare
      (dotimes (i (length ext-adiac))
        (let ((edge (nth i (get-cell-faces cell))))

          (if (not (equal (nth i ext-adiac) 0))
            (progn
              (setq prox-pol
                   (second (nth i ext-adiac)))
              (setq prox-cell
                   (get-polyhedron-cell prox-pol (first (nth i ext-adiac)))))
            (if (not (equal (nth i adiac) 0))
              (progn
                (setq prox-pol
                      pol)
                (setq prox-cell
                     (get-polyhedron-cell prox-pol (nth i adiac))))))   

          (if (or (not (equal (nth i ext-adiac) 0))  ;la cella ha adiacenze su 
                  (not (equal (nth i adiac) 0)))     ;  questo lato
            (if (or (equal vertex (first edge))
                    (equal vertex (second edge)))
              (setq tab-incidence-cell 
                    (incidence-v vertex 
                                 prox-cell 
                                 prox-pol 
                                 tab-incidence-cell))))))
           tab-incidence-cell)

tab-incidence-cell ) ))

;-----------------------------------------------------------
;La seg. funzione, dato un elemento ed una lista, controlla
;se esso era gia' presente in essa.
;-----------------------------------------------------------

(defun in-list? (elem list 
                 &aux (YorN nil))
  (dolist (x list YorN)
    (if (equal x elem)
      (setq YorN T))))


;------------------------------------------------------------
;La seg. funzione a partire dall'elemento i-esimo della
;list-incidence-cell, ritorna il vettore che ha come
;come estremi i vertici dello spigolo dell'elemento dato
;------------------------------------------------------------

(defun return-vector (elem)
  (let* ((n-coord (length (second (first elem))))
         (vect (make-array n-coord :element-type 'double-float))
         (v1 (second (first elem)))
         (v2 (second (second elem))))
     (dotimes (i n-coord)
       (setf (aref vect i) (- (nth i v2) (nth i v1))))
   vect))


;------------------------------------------------------------
;La seg. funzione calcola la normale al piano nel vertice V
;------------------------------------------------------------

(defun calculate-normal-point-type1 (vertex tab-incidence-cell)
  (let* ((list-incidence-cell 
            (make-list-incidence-sort vertex tab-incidence-cell))
         (dim (length (second (first (first list-incidence-cell)))))
         (sum (make-array dim 
                               :initial-element 0.0d0
                               :element-type 'double-float))
         (m (length list-incidence-cell)))

    (dotimes (i (1- m))
      (let ((Vi (return-vector (nth i list-incidence-cell)))
            (Vi_1 (return-vector (nth (1+ i) list-incidence-cell)))) 
        (setq sum (vet-sum sum (vect-prod Vi Vi_1)))))

   sum))

;-----------------------------------------------------------
; La seg. funzione calcola le coordinate dei punti di tipo1
; intorno al vertice v, memorizzandoli nei corrispondenti
; simplessi cubici.
; Restituisce la lista degli identificatori delle celle 
; esaminate.
;-----------------------------------------------------------

(defun calculate-point-type1 (vertex cell pol normal
                              &optional (list-cell-id '()))
  (let* ((adiac (get-cell-adjacences cell))
         (ext-adiac (get-cell-external-adjacences cell))
         (cell-id (get-cell-identifier cell))
         (simplexes (get-cubic-simplexes pol cell-id))
         (vertex-coord (second
                         (first
                           (get-coord-vlist
                             (list vertex)
                             (polyhedron-VLIST pol)))))
          prox-cell
          prox-pol)


    (if (not (in-list? cell-id list-cell-id)) ;;controlla se la cella e'
                                              ;;gia' stata esaminata
      (progn
        (maphash #'(lambda (id simpl)       ;scandisce la tabella 
                                            ;dei simplessi cubici
            (dolist (x (cubic-simplex-EDGES simpl))
              (let ((y (get-coord-vlist (orientation vertex x)
                                        (cubic-simplex-VERTICES simpl))))   
                (if (or (equal vertex (first x))
                        (equal vertex (second x)))
                  (progn
                    (setf 
                      (gethash 
                         (make-id-coord-baric
                           (get-coord-baric-point vertex x simpl))
                         (cubic-simplex-BEZIER-NET simpl))
                      (ortogonal-projection 
                        (coord-point-type1 (second (first y)) 
                                           (second (second y)))
                        (plane-point-nvector vertex-coord normal) ))  
                    (setq list-cell-id 
                      (append list-cell-id
                         (list cell-id))))))))
         simplexes)

                     ;; Ottiene la prossima cella da esaminare
      (dotimes (i (length ext-adiac))
        (let ((edge (nth i (get-cell-faces cell))))

          (if (not (equal (nth i ext-adiac) 0))
            (progn 
              (setq prox-pol 
                   (second (nth i ext-adiac)))
              (setq prox-cell 
                   (get-polyhedron-cell prox-pol (first (nth i ext-adiac)))))
            (if (not (equal (nth i adiac) 0))
              (progn 
                (setq prox-pol 
                      pol)
                (setq prox-cell 
                     (get-polyhedron-cell prox-pol (nth i adiac))))))
 
          (if (or (not (equal (nth i ext-adiac) 0))  ;la cella ha adiacenze su 
                  (not (equal (nth i adiac) 0)))     ;  questo lato
            (if (or (equal vertex (first edge))
                    (equal vertex (second edge)))
              (setq list-cell-id
                (calculate-point-type1
                       vertex prox-cell prox-pol normal list-cell-id))))))
       list-cell-id)
 
             ;se la cella era gia' stata esaminata 
             ;restituisce la lista di incidenza stessa
             ;senza fare niente
     list-cell-id)))


;-----------------------------------------------------------
;La seg. funzione e' la main del passo 2 dell'algoritmo.
;Essa setta i punti di tipo 1 intorno ad un vertice.
;-----------------------------------------------------------

(defun set-point-type1 (vertex cell pol)
  (calculate-point-type1 vertex
                         cell
                         pol 
       (calculate-normal-point-type1 vertex 
                                    (incidence-v vertex 
                                                 cell 
                                                 pol)))
  cell)
      

;------------------------------------------------------------
;La seg. funzione orienta uno spigolo in modo che il primo
;punto coincida con vertex
;------------------------------------------------------------

(defun orientation (vertex edge
                    &aux orient-edge)
  (if (equal vertex (first edge))
      (setq orient-edge edge))
  (if (equal vertex (second edge))
      (setq orient-edge (reverse edge)))
   orient-edge)

;------------------------------------------------------------
;La seg. funzione calcola le coordinate del punto da 
;proiettare sul piano passante per il vertice esaminato.
;Le coord. sono ottenute come:
; 2/3*V+1/3*P
;------------------------------------------------------------

(defun coord-point-type1 (V P)
  (vect->list
     (vet-sum (scalar-vect-mult '2/3 (list->vect V)) 
              (scalar-vect-mult '1/3 (list->vect P)))))


;------------------------------------------------------------
;La seg. funzione mi dice se un vettore dato in input ha tutti
;elementi < in valore assoluto di **epsilon
;------------------------------------------------------------

(defun vect-of-zero? (vect &aux (YorN T))
  (dotimes (i (first (array-dimensions vect)))
    (if (> (abs (aref vect i)) **epsilon)
      (setq YorN nil)))
   YorN)


;------------------------------------------------------------
;La seg. funzione mi restituisce la prima coppia di spigoli
;da memorizzare nella lista di incidenza e l'identificatore
;della  prossima cella dalla quale prendere i rimanenti 
;spigoli. Restituisce
;(edge1 edge2 C) se la tabella di incidenza ha piu' di una 
;                entry.
;(edge1 edge2 0) se la tabella di incidenza ha una sola 
;                entry.
;(nil nil C)     se siamo nel caso di loop di spigoli.
;------------------------------------------------------------

(defun get-first-edge (vertex tab-incidence-cell)

  (let ((found nil)
        (result nil)
         id-prox-cell
         id-one-cell
         two-edges)
     (maphash #'(lambda (id cell)
       (if (not found)
         (progn
         (setq id-one-cell id)
         (let ((adiac (get-cell-adjacences cell))
               (ext-adiac (get-cell-external-adjacences cell))
               (edges (get-cell-faces cell))
                temp)
            (setq two-edges nil)
            (setq id-prox-cell nil)
            (dotimes (i (length edges))
               (let ((edge (nth i edges)))

                  (if (or (equal vertex (first edge))
                          (equal vertex (second edge)))
                     (if (and (equal (nth i adiac) 0)
                              (equal (nth i ext-adiac) 0))
                        (progn 
                          (setq found T)
                          (remhash id tab-incidence-cell)
                          (setq two-edges 
                            (append 
                              (list 
                                (edge-coord (orientation vertex edge) cell))
                               two-edges)))
                        (progn
                          (setq two-edges 
                            (append 
                               two-edges 
                              (list 
                                (edge-coord (orientation vertex edge) cell))))
                          (if (not (equal (nth i ext-adiac) 0))
                            (setq id-prox-cell (first (nth i ext-adiac)))
                            (setq id-prox-cell (nth i adiac))))))))))))
      tab-incidence-cell)

    (if (and (equal found T)
             (equal id-prox-cell nil))  ;siamo nel caso di una sola 
                                        ;cella incidente
      (setq id-prox-cell 0))

    (if found 
      (setq result (list (nth 0 two-edges) (nth 1 two-edges) id-prox-cell))
      (setq result (list nil nil id-one-cell)))
   result))    


;------------------------------------------------------------
;La seg. funzione mi restituisce il resto della lista di
;incidenza ordinata.
;------------------------------------------------------------

(defun get-prox-edge (vertex cell-id tab-incidence-cell list-incidence-sort)

  (if (not (equal (hash-table-count tab-incidence-cell) 0))
    (let* ((cell (gethash cell-id tab-incidence-cell))
           (adiac (get-cell-adjacences cell))
           (ext-adiac (get-cell-external-adjacences cell))
           (edges (get-cell-faces cell))
            id-prox-cell)

      (dotimes (i (length edges))
        (let ((edge (nth i edges)))

          (if (or (equal vertex (first edge))
                  (equal vertex (second edge)))

            (if (not (edge-loop? (edge-coord (orientation vertex edge) cell)
                                  list-incidence-sort)) 

                (if (not (existent-edge? 
                              (edge-coord (orientation vertex edge) cell)
                               list-incidence-sort))
                  (progn 
                    (remhash cell-id tab-incidence-cell)
                    (setq list-incidence-sort
                       (append 
                        list-incidence-sort
                       (list 
                         (edge-coord (orientation vertex edge) cell))))

                    (if (not (equal (nth i ext-adiac) 0))
                      (setq id-prox-cell (first (nth i ext-adiac)))
                      (setq id-prox-cell (nth i adiac)))))
              
               (progn
                 (remhash cell-id tab-incidence-cell)
                 (if (not (equal (nth i ext-adiac) 0))
                          (setq id-prox-cell (first (nth i ext-adiac)))
                          (setq id-prox-cell (nth i adiac))))))))

       (setq list-incidence-sort 
           (get-prox-edge vertex 
                          id-prox-cell 
                          tab-incidence-cell 
                          list-incidence-sort)))
                                 ;la tab. hash non ha piu' entry
   (if (not (equal cell-id 0))   ;siamo nel caso in cui bisogna aggiungere 
                                 ;anche il vertice composto dal primo e
                                 ;dall'ultimo elemento della lista.
     (setq list-incidence-sort 
        (append 
            list-incidence-sort
           (list (first list-incidence-sort))))))

   list-incidence-sort)


;------------------------------------------------------------
;La seg. funzione mi restituisce la lista di incidenza
;ordinata a partire dalla tabella delle celle incidenti uno 
;spigolo.
;------------------------------------------------------------

(defun make-list-incidence-sort (vertex tab-incidence-cell)
  (let ((list-incidence-sort nil)
         start-inf)
    (setq start-inf (get-first-edge vertex tab-incidence-cell))
   (if (not (equal (first start-inf) nil))
        (setq list-incidence-sort (get-prox-edge vertex  
                                               (nth 2 start-inf)
                                                tab-incidence-cell
                                               (append 
                                                 (list (nth 0 start-inf))
                                                 (list (nth 1 start-inf)))))
        (setq list-incidence-sort (get-prox-edge vertex  
                                               (nth 2 start-inf)
                                                tab-incidence-cell
                                                nil))) ))


;------------------------------------------------------------
;La seg. funzione mi restituisce il vertice con le sue
;coordinate.
;------------------------------------------------------------

(defun edge-coord (edge cell)
  (let ((simplexes (get-cell-cubic-simplexes cell))
        (ok nil) 
         edge-with-coord)

    (if (not ok)
      (maphash #'(lambda (id simplex)
        (dolist (x (cubic-simplex-EDGES simplex))
          (if (or (equal x edge)
                  (equal x (reverse edge)))        
            (progn
              (setq ok T)
              (setq edge-with-coord
                 (get-coord-vlist edge (cubic-simplex-VERTICES simplex))))))) 
       simplexes))
    edge-with-coord))



;------------------------------------------------------------
;La seg. funzione mi dice se siamo nel caso di loop di
;spigoli.
;------------------------------------------------------------
 
(defun edge-loop? (edge list-edges)
  (equal-edge edge (first list-edges)))



;------------------------------------------------------------
;La seg. funzione mi dice se uno spigolo con le sue 
;coordinate e' gia' presente nella lista di incidenza.
;------------------------------------------------------------

(defun existent-edge? (edge list-edges
                       &aux (YorN nil))
  (dolist (x list-edges YorN)
    (if (equal-edge x edge)
      (setq YorN T))))


;------------------------------------------------------------
;La seg. funzione mi testa l'uguaglianza di due spigoli
;includenti anche le coordinate.
;------------------------------------------------------------

(defun equal-edge (edge1 edge2)
  (and (equal (first (first edge1))
              (first (first edge2)))
       (equal (first (second edge1))
              (first (second edge2)))))




