;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Il seguente modulo contiene le funzioni per eseguire la ;;
;; fase 3 dell'algoritmo per ottenere le sup. C1           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(inpackage kernel)

;(require 'geometry)

(require 'cub_ext)

(provide 'step3)

;--------------------------------------------------------------------------
; La seg. funzione dato lo spigolo di una cella, e la cella stessa, mi
; restituisce il simplesso della cella a cui appartiene lo spigolo dato.
; NOTA che per costruzione, di tali simplessi ne esiste solo uno per
; ogni cella in quanto lo spigolo della cella non conterra' mai il 
; centroide della cella stessa.
;--------------------------------------------------------------------------

(defun get-simplex-from-edge (edge cell)
  (let ((simplexes (get-cell-cubic-simplexes cell))
         simplex)
    (maphash #'(lambda (id simpl)
      (dolist (x (cubic-simplex-EDGES simpl))
        (if (or (equal edge x)
                (equal edge (reverse x)))
          (setq simplex simpl))))
     simplexes)
   simplex))
         
;--------------------------------------------------------------------------
; La seg. funzione controlla se il simplesso era gia' stato esaminato,
; controllando se il punto di coordinate baricentriche 1,1,1 era gia'
; settato
;--------------------------------------------------------------------------

(defun check-point-1_1_1 (simplex
                          &aux (YorN T))
  (let ((tab (cubic-simplex-BEZIER-NET simplex))
        (label (make-id-coord-baric '(1 1 1))))
    (if (equal (gethash label tab) nil)
      (setq YorN nil))
   YorN))

;--------------------------------------------------------------------------
; La seg. funzione e' la main del passo3 dell'algoritmo. Essa setta i 
; punti della Bezier-net di coord. baricentriche 1,1,1 dei simplessi
; della cella in esame e di quelle adiacenti.
;--------------------------------------------------------------------------

(defun set-point-type2 (cell pol)
  (let ((edges (get-cell-faces cell))
        (adiac (get-cell-adjacences cell))
        (ext-adiac (get-cell-external-adjacences cell))
         simp
         simp-adiac
         cell-adiac
         pol-adiac)
    (dotimes (i (length edges))
      (setq simp (get-simplex-from-edge (nth i edges) cell))
      (if (not (check-point-1_1_1 simp))
                                         ;Se il simplesso era gia' stato
                                         ;esaminato, non fa niente

        (if (or (not (equal (nth i ext-adiac) 0))
                (not (equal (nth i adiac) 0))) 
                                               ;La cella ha adiacenze su 
                                               ;questo spigolo

          (progn
            (if (not (equal (nth i ext-adiac) 0))
               (progn 
                 (setq pol-adiac 
                      (second (nth i ext-adiac)))
                 (setq cell-adiac
                      (get-polyhedron-cell pol-adiac (first (nth i ext-adiac))))
                 (setq simp-adiac 
                      (get-simplex-from-edge (nth i edges) cell-adiac)))
               (if (not (equal (nth i adiac) 0))
                 (progn 
                   (setq pol-adiac 
                         pol)
                   (setq cell-adiac
                        (get-polyhedron-cell pol-adiac (nth i adiac)))
                   (setq simp-adiac 
                        (get-simplex-from-edge (nth i edges) cell-adiac)))))
            (calculate-point-type2 (nth i edges)
                                    simp
                                    simp-adiac
                                   (calculate-normal-point-type2 (nth i edges)
                                                                  simp 
                                                                  simp-adiac)))

                                                       
                                                ;La cella non ha adiacenze
                                                ;su questo spigolo

          ;else
          (setf 
             (gethash (make-id-coord-baric '(1 1 1))
                      (cubic-simplex-BEZIER-NET simp))
             (median-point (gethash (make-id-coord-baric '(2 0 1))
                                    (cubic-simplex-BEZIER-NET simp))
                           (gethash (make-id-coord-baric '(0 2 1))
                                    (cubic-simplex-BEZIER-NET simp)))))))))


;-------------------------------------------------------------------------
; La seg. funzione calcola la normale per i punti di tipo2 a partire
; da uno spigolo e dai due simplessi che incidono su esso.
;-------------------------------------------------------------------------

(defun calculate-normal-point-type2 (edge simp1 simp2)
  (let ((C1 (calculate-centroid (cubic-simplex-VERTICES simp1)))
        (C2 (calculate-centroid (cubic-simplex-VERTICES simp2)))
        (R (gethash (make-id-coord-baric
                       (get-coord-baric-point (first edge) edge simp1))
                    (cubic-simplex-BEZIER-NET simp1)))
        (G (gethash (make-id-coord-baric
                       (get-coord-baric-point (second edge) edge simp1))
                    (cubic-simplex-BEZIER-NET simp1)))
         normal)


    (setq normal (vet-sum 
                    (vect-prod (points->vect R C1) (points->vect R G))
                    (vect-prod (points->vect R G) (points->vect R C2))))
   normal))
    

;-------------------------------------------------------------------------
; La seg. funzione calcola le coordinate dei punti di tipo2.
;-------------------------------------------------------------------------

(defun calculate-point-type2 (edge simp1 simp2 normal)
  
  (let ((C1 (calculate-centroid (cubic-simplex-VERTICES simp1)))
        (C2 (calculate-centroid (cubic-simplex-VERTICES simp2)))
        (R (gethash (make-id-coord-baric
                       (get-coord-baric-point (first edge) edge simp1))
                       (cubic-simplex-BEZIER-NET simp1))))
    (setf 
      (gethash
        (make-id-coord-baric '(1 1 1)) (cubic-simplex-BEZIER-NET simp1))
        (ortogonal-projection C1 
                             (plane-point-nvector R normal)))
    (setf 
      (gethash
        (make-id-coord-baric '(1 1 1)) (cubic-simplex-BEZIER-NET simp2))
        (ortogonal-projection C2 
                             (plane-point-nvector R normal)))))   



;-------------------------------------------------------------------------
; La seg. funzione trova il punto medio tra due punti espressi per mezzo 
; delle loro coordinate.
;-------------------------------------------------------------------------

(defun median-point (P1 P2)
  (vect->list
    (vet-sum (scalar-vect-mult '1/2 (list->vect P1))
             (scalar-vect-mult '1/2 (list->vect P2)))))



