;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Il seguente modulo contiene le funzioni per eseguire la ;;
;; fase 4 dell'algoritmo per ottenere le sup. C1           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(inpackage kernel)

;(require 'geometry)

(require 'cub_ext)

(provide 'step4)

         
;--------------------------------------------------------------------------
; La seg. funzione controlla se lo spigolo di un simplesso e' lo spigolo
; contenente il centroide della cella. 
; In input riceve il solo spigolo. 
;--------------------------------------------------------------------------

(defun edge-with-centroid? (edge
                            &aux (YorN nil))
  (if (or (centroid-symbol? (first edge))
          (centroid-symbol? (second edge)))
    (setq YorN T))
   YorN)

;--------------------------------------------------------------------------
; La seg. funzione controlla se il punto di tipo 3 sul simplesso in esame
; e' gia' stato settato.
;--------------------------------------------------------------------------

(defun check-point-type3 (edge simplex
                          &aux (YorN T))
  (let ((tab (cubic-simplex-BEZIER-NET simplex))
         label
         centroid-point)
    (if (centroid-symbol? (first edge))
      (setq centroid-point (first edge))
      (setq centroid-point (second edge)))
    (setq label (make-id-coord-baric 
                  (get-coord-baric-point centroid-point
                                         edge
                                         simplex)))
    (if (equal (gethash label tab) nil)
        (setq YorN nil))
   YorN))


;--------------------------------------------------------------------------
; La seg. funzione e' la main del passo4 dell'algoritmo. Essa setta i 
; punti della Bezier-net di tipo3.
;--------------------------------------------------------------------------

(defun set-point-type3 (cell)
  (let ((simplexes (get-cell-cubic-simplexes cell))
        ;(simp-tab (make-hash-table :test 'equal))
         fun-centroid
         list-points
         point)
    ;(setf (gethash 'ATT! simp-tab) nil)  ;La entry corrispondente ad ATT! 
                                         ;viene settata a T se si verifica
                                         ;nella cella il caso di complanarita'
                                         ;tra il piano per D C C' e la
                                         ;giacitura della cella.
     
    (setq ATT! nil) ;evantualmente poi dichiarare in altro modo

    (maphash #'(lambda (id simp)
      (let ((edges (cubic-simplex-EDGES simp))
            (adiac (cubic-simplex-ADIACENT-CUBIC-SIMPLEXES simp))
             simp-adiac)
        
        (dotimes (i (length edges))
          (if (edge-with-centroid? (nth i edges))
            (if (not (check-point-type3 (nth i edges) simp))
              (progn
                (setq simp-adiac (gethash (nth i adiac) simplexes))
                (setq point
                  (calculate-point-temp-type3 
                     (nth i edges)
                      simp
                      simp-adiac
                     (calculate-normal-point-type3 (nth i edges)
                                                    simp
                                                    simp-adiac)))
                (setq list-points (append list-points (list point)))))))))
     simplexes)

    (setq fun-centroid (baricenter list-points))

    (maphash #'(lambda (id simp)
      (let ((edges (cubic-simplex-EDGES simp))
            (adiac (cubic-simplex-ADIACENT-CUBIC-SIMPLEXES simp))
             simp-adiac)
        (set-centroid simp fun-centroid)  ;setta il valore del centroide
                                          ; nella bezier-net


;;;;;(setq ATT! T)
    
         (if (not ATT! )
          (dotimes (i (length edges))
            (if (edge-with-centroid? (nth i edges))
    
                (progn
                  (setq simp-adiac (gethash (nth i adiac) simplexes))
                  (adjust-point-type3 (nth i edges)
                                          simp
                                          simp-adiac
                                          fun-centroid
                                    (calculate-normal-point-type3 (nth i edges)
                                                    simp
                                                    simp-adiac)))  )))))  
     simplexes)))




;-------------------------------------------------------------------------
; La seg. funzione calcola la normale per i punti di tipo3 a partire
; da uno spigolo contenente avente come vertice il centroide della cella 
; in esame e dai due simplessi che incidono su esso.
;-------------------------------------------------------------------------

(defun calculate-normal-point-type3 (edge simp1 simp2)
  (let ((C1 (gethash (make-id-coord-baric '(1 1 1)) 
                     (cubic-simplex-BEZIER-NET simp1)))
        (C2 (gethash (make-id-coord-baric '(1 1 1)) 
                     (cubic-simplex-BEZIER-NET simp2)))
         D
         normal)
     (if (not (centroid-symbol? (first edge)))
       (setq D (gethash 
                  (make-id-coord-baric
                    (get-coord-baric-point (first edge) edge simp1))
                  (cubic-simplex-BEZIER-NET simp1)))            
       (setq D (gethash 
                  (make-id-coord-baric
                    (get-coord-baric-point (second edge) edge simp1))
                  (cubic-simplex-BEZIER-NET simp1))))
     (setq normal (vect-prod (points->vect D C1) (points->vect D C2)))
   
   normal))
    

;-------------------------------------------------------------------------
; La seg. funzione calcola le coordinate dei punti temporanei che sono
; usati per calcolare la quota del cetroide della cella.
; Restituisce le coord. del punto.
;-------------------------------------------------------------------------

(defun calculate-point-temp-type3 (edge simp1 simp2 normal)
  
  (let ( value
         plane
         line 
         D)
    (if (approx-vect-parallel-vect? normal (cell-normal simp1))
      (setq ATT! T))

    (if (not (centroid-symbol? (first edge)))
       (setq D (gethash (make-id-coord-baric
                          (get-coord-baric-point (first edge) edge simp1))
                        (cubic-simplex-BEZIER-NET simp1)))
       (setq D (gethash (make-id-coord-baric
                          (get-coord-baric-point (second edge) edge simp1))
                        (cubic-simplex-BEZIER-NET simp1))))
    (setq plane (plane-point-nvector D normal))
    (setq line 
      (line-point-par-direction (coord-point-type3 edge simp1)
                                (cell-normal simp1)))
    (setq value (line-plane-intersection plane line ))

    (if (centroid-symbol? (first edge))
      (progn
        (setf 
          (gethash 
            (make-id-coord-baric
               (get-coord-baric-point (first edge) edge simp1))
            (cubic-simplex-BEZIER-NET simp1))
           value)
        (setf 
          (gethash 
            (make-id-coord-baric
               (get-coord-baric-point (first edge) edge simp2))
            (cubic-simplex-BEZIER-NET simp2))
           value))
      (progn    
        (setf 
          (gethash 
            (make-id-coord-baric
               (get-coord-baric-point (second edge) edge simp1))
            (cubic-simplex-BEZIER-NET simp1))
           value)
        (setf 
          (gethash 
            (make-id-coord-baric
               (get-coord-baric-point (second edge) edge simp2))
            (cubic-simplex-BEZIER-NET simp2))
           value)))

  value))


;-------------------------------------------------------------------------
; La seg. funzione calcola la normale alla cella nel suo centroide.
; Essa sfrutta il fatto che il verso di percorrenza di ogni simplesso
; e' preso in modo che il centroide sia l'ultimo di posizione.
;-------------------------------------------------------------------------

(defun cell-normal (simp)
  (let ((vertices (cubic-simplex-VERTICES simp)))
    (vect-prod (points->vect (second (nth 2 vertices)) 
                             (second (nth 0 vertices)))
               (points->vect (second (nth 2 vertices)) 
                             (second (nth 1 vertices))))))

;-------------------------------------------------------------------------
; La seg. funzione ottiene il punto che e' in prossimita' del centroide
; della cella.
;-------------------------------------------------------------------------

(defun coord-point-type3 (edge simp)
  (let ((list-points (get-coord-vlist edge (cubic-simplex-VERTICES simp)))
         point)
    (if (centroid-symbol? (first edge))
      (setq point (vect->list
                    (vet-sum 
                       (scalar-vect-mult '2/3 
                          (list->vect (second (first list-points))))
                       (scalar-vect-mult '1/3 
                          (list->vect (second (second list-points)))))))
                ;il centroide e' il secondo vertice
      (setq point (vect->list
                    (vet-sum 
                       (scalar-vect-mult '1/3 
                          (list->vect (second (first list-points))))
                       (scalar-vect-mult '2/3 
                          (list->vect (second (second list-points))))))))
   point))
      



;-------------------------------------------------------------------------
; La seg. funzione ricalcola le coordinate dei punti di tipo3 in modo
; da renderli complanari.
;-------------------------------------------------------------------------


(defun adjust-point-type3 (edge simp1 simp2 fun-centroid normal)
  
  (let ( value
         direction
         plane
         line 
         D)
    (if (not (centroid-symbol? (first edge)))
       (setq D (gethash (make-id-coord-baric
                          (get-coord-baric-point (first edge) edge simp1))
                        (cubic-simplex-BEZIER-NET simp1)))
       (setq D (gethash (make-id-coord-baric
                          (get-coord-baric-point (second edge) edge simp1))
                        (cubic-simplex-BEZIER-NET simp1))))
    (setq plane (plane-point-nvector D normal)) 
    (setq direction (points->vect
                      (second 
                        (first 
                          (get-coord-vlist (list (second edge))
                                           (cubic-simplex-VERTICES simp1))))
                      (second 
                        (first 
                          (get-coord-vlist (list (first edge))
                                           (cubic-simplex-VERTICES simp1))))))
    (setq line (line-point-par-direction fun-centroid direction))
    (setq value (line-plane-intersection plane line))
    (if (centroid-symbol? (first edge))
      (progn
        (setf 
          (gethash 
            (make-id-coord-baric
               (get-coord-baric-point (first edge) edge simp1))
            (cubic-simplex-BEZIER-NET simp1))
           value)
        (setf 
          (gethash 
            (make-id-coord-baric
               (get-coord-baric-point (first edge) edge simp2))
            (cubic-simplex-BEZIER-NET simp2))
           value))
      (progn    
        (setf 
          (gethash 
            (make-id-coord-baric
               (get-coord-baric-point (second edge) edge simp1))
            (cubic-simplex-BEZIER-NET simp1))
           value)
        (setf 
          (gethash 
            (make-id-coord-baric
               (get-coord-baric-point (second edge) edge simp2))
            (cubic-simplex-BEZIER-NET simp2))
           value)))
  ))


;--------------------------------------------------------------------------
; La seg. funzione setta il valore del centroide nella Bezier-net del
; simplesso in esame al valore di fun-centroid.
; NOTA. Per la scelta dell'orientamento dei vertici del simplesso, il
; vertice corrispondente al centroide della cella ha sempre coordinate
; (0 0 3).
;--------------------------------------------------------------------------

(defun set-centroid (simp fun-centroid)
  (setf (gethash (make-id-coord-baric '(0 0 3))
                 (cubic-simplex-BEZIER-NET simp))
         fun-centroid))















