; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package Kernel                                            ;                                
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: V. Pascucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require 'kernel)

(provide 'f_ext)
(inpackage kernel )   ;package name

(export '(representation?
	  current-face-counter advance-face-counter
	  reset-face-counter previous-face-counter
	  face? face-list? face-symbol?
	  make-face make-face-symbol
	  put-face
	  put-faces 
	  set-fid
	  set-sid
	  set-coord
	  set-sign
	  set-faces
	  
	  set-pol-sign
	  set-npol
	  set-dpol
	  set-#C
	  set-#F
	  set-#kd
	  set-#K0
	  
	  get-#F
	  get-#C
	  get-pol-rep
	  get-pol-brick-type 
	  get-vertices-or-faces
	  get-facescoord
	  get-faces
	  get-flist
	  get-face
	  get-pol-face
	  get-fcoord
	  get-sfaces
	  get-sfacescoord
	  find-fcoord
	  find-sfaces
	  find-same-vertex
	  find-same-plane
	  add-new-face
	  add-new-vertex
	  get-adjacent-face
	  opposit-face
	  new-faces-table
	  
	  coerce-polyhedron))


#+:akcl
   (load (concatenate 
	  'string user::**simplexn-directory 
	  "/faces-convexes-complexes/hyperplanes.lisp"))

(require 'hyperp)
(use-package 'hyperplanes)

(require 'lisp_ext)
(use-package 'lisp-extension)

;;;;  Global declarations and functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(deftype flist () '(satisfies faces-list?))

;-------------------------------------------------------------
; list of the possible kind of objects into witch a polyhedron
; can be decomposed.
; the first one is used as default in functions  that
; require the optional parameter kind-of-bricks 
;-------------------------------------------------------------
(defvar $$kind-of-bricks '(simplices
			   convex-cells))

(defvar $$default-kind-of-brick (first $$kind-of-bricks))

; test if the item is a kind of brick

(defun  kind-of-brick? (item)
  (if (find item $$kind-of-bricks :test 'symbol-equal) t))

(deftype kind-of-brick () '(satisfies kind-of-brick?))

;-------------------------------------------------------------
; list of the possible representations used for a polyhedron
; the firstone is used as default in functions that
; require the optional parameter representation
;-------------------------------------------------------------
(defvar $$representations '(on-vertices
			    on-faces))

(defvar $$default-representation
  (first $$representations))

; test if the item is a kind of representation

(defun  representation? (item)
  (if (find item $$representations :test 'symbol-equal) t))

(deftype representation () '(satisfies representation?))

;-------------------------------------------------------------
; define the new structure of the polyhedron adding 
; the slots :
;       representation
;       flist
;-------------------------------------------------------------
(defstruct (polyhedron (:constructor create-polyhedron))
  (m 0 :type intpos)
  (n 0 :type intpos)
  (k0 0 :type intpos)
  (km 0 :type intpos)
  (sign '+ :type symbol)
  (vlist '() :type vlist)
  (slist '() :type slist)
  (info nil)
  (representation $$default-representation :type representation)
  (faces (make-hash-table :test 'equal)
					;       :type hashtable
	 )      
  (F 0 :type intpos)
  (convex-cells (make-hash-table :test 'equal) 
					; :type hashtable
		)
  (C 0 :type intpos)
  (kind-of-brick $$default-kind-of-brick :type kind-of-brick)
  (general-cells nil))

;-------------------------------------------------------------
; Assgnazioni di valori ai parametri del poliedro 
;-------------------------------------------------------------
(defun set-#K0 (pol &optional (n (length (get-vlist pol))))
  (if (get-type-checking) 
   (if (not (polyhedron? pol)) 
    (error-handler 1 "get-#K0")))
  (setf (polyhedron-k0 pol) n))

(defun set-#kd (pol &optional (n (length (get-slist pol))))
  (if (get-type-checking) 
   (if (not (polyhedron? pol))
    (error-handler 1 "get-#kd")))
  (setf (polyhedron-km pol) n))

(defun set-#F (pol &optional (n (hash-table-count (get-faces pol))))
  (if (get-type-checking) 
   (if (not (polyhedron? pol)) 
    (error-handler 1 "get-#K0")))
  (setf (polyhedron-F pol) n))

(defun set-#C (pol &optional (n (hash-table-count (get-cells pol))))
  (if (get-type-checking) 
   (if (not (polyhedron? pol))
    (error-handler 1 "get-#kd")))
  (setf (polyhedron-C pol) n))

(defun set-dpol (pol n)
  (if (get-type-checking)
   (if (not (polyhedron? pol))
     (error-handler 1 "get-dpol")
     (if (empty-polyhedron? pol)
      (error-handler 2 "get-dpol"))))
  (setf (polyhedron-m pol) n))
			   
(defun set-npol (pol n)
  (if (get-type-checking)
   (if (not (polyhedron? pol))
     (error-handler 1 "get-npol")))
  (setf (polyhedron-n pol) n))

(defun set-pol-sign (pol n)
  (if (get-type-checking)
   (if (not (polyhedron? pol))
     (error-handler 1 "get-pol-sign")))
  (setf (polyhedron-sign pol) n))

;-------------------------------------------------------------
; create an empty polyhedron
; the default representation used is the first of
; the $$representations list
; partitioned into elements of type
; the $$representations list
;-------------------------------------------------------------
(defun make-empty-polyhedron (&optional (representation $$default-representation)
					(kind-of-bricks $$default-kind-of-brick))
  (create-polyhedron
   :representation representation
   :kind-of-brick kind-of-bricks))

; create a list of empty polyhedrons, one for each kind
; of representation

(defvar **all-possible-empty-polyhedrons 
  (mapcar 'make-empty-polyhedron $$representations))

;;;; Counters
;
;  counter functions for faces
;

(defvar $$fcounter 0)

(defun current-face-counter ()
  (make-face-symbol $$fcounter))

(defun advance-face-counter ()
  (prog1
    (make-face-symbol $$fcounter)
    (setq $$fcounter (+ $$fcounter 1))))

(defun reset-face-counter (&optional (n 0))
  (setq $$fcounter  n))

(defun previous-face-counter ()
  (make-face-symbol (- $$fcounter 1)))

;;;;  Predicates
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; tests if x is an empty polyhedron of any kind of representation
(defun empty-polyhedron? (x)
  (if (find x **all-possible-empty-polyhedrons) t))

;------------------------------------------------------------------
; tests if f is a fece, i.e. a list of length 2 with the first
; element is a face symbol and the second is a list of numbers.
;------------------------------------------------------------------
(defun face? (f) 
  (and (length-two? f)
       (face-symbol? (get-id f))
       (every 'numberp (get-coord f))))

(defun simplex? (l) ; da sistemare
  (and (length-two? l)
       (length-three? (second l))
       (simplex-symbol? (get-id l))
       (plus-or-minus? (get-sign l))
       (let ((svert (get-vert l))
	     (sad (get-ad l)))
	 (and (= (length svert) (length sad))
	      (every 'vertex-symbol? svert)
	      (every 'simplex-symbol-or-zero? sad)))))

(defun face-list? (l)
  (every 'face? l))

(defun face-symbol? (x)
  (and (symbolp x)
       (every 'characterp 
	      (coerce (symbol-name x) 'list))))

;;;;  Constructors
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;------------------------------------------------------------------
; create a polyhedron based on the list of vertices (or faces)
;  and the list of simplices
;------------------------------------------------------------------
;(defun make-polyhedron (f-or-v-list slist &optional (representation $$default-representation))
;  (if (get-type-checking) 
;    (if (not 
;         (or (and (vertex-list? f-or-v-list) (simplex-list? slist))
;             (and (face-list? f-or-v-list)   (simplex-list? slist)))) 
;      (error-handler 1 "make-polyhedron")))
;  (cond
;   ((symbol-equal representation 'on-vertices)
;    (create-polyhedron 
;     :m (- (length (get-vert (first slist))) 1)
;     :n (length (get-coord (first f-or-v-list)))
;     :k0 (length f-or-v-list)
;     :km (length slist)
;     :vlist f-or-v-list
;     :slist slist
;     :representation 'on-vertices))
;   ((symbol-equal representation 'on-faces)
;    (create-polyhedron 
;     :m 0
;     :n 0
;     :k0 0
;     :km 0
;     :faces f-or-v-list
;     :slist slist
;     :representation 'on-faces))))

(defun make-polyhedron (faces-or-vertices slist &optional (representation $$default-representation))
  (if (get-type-checking) 
    (if (not 
	 (or (and (vertex-list? faces-or-vertices) (simplex-list? slist))
	     (and (face-list? faces-or-vertices)   (simplex-list? slist)))) 
      (error-handler 1 "make-polyhedron")))
  (cond
   ((symbol-equal representation 'on-vertices)
    (create-polyhedron 
     :m (- (length (get-vert (first slist))) 1)
     :n (length (get-coord (first faces-or-vertices)))
     :k0 (length faces-or-vertices)
     :km (length slist)
     :vlist faces-or-vertices
     :slist slist
     :representation 'on-vertices))
   ((symbol-equal representation 'on-faces)
    (create-polyhedron 
     :m 0
     :n 0
     :k0 0
     :km 0
     :faces faces-or-vertices
     :slist slist
     :representation 'on-faces))))

; create a face 
(defun make-face (fsymb parameters)
  (list fsymb parameters))

;create the symbol Fi with the given i
; es: i=5 -> F5 
(defun make-face-symbol (i)
  (if (< i 0)
    (intern "F0")
    (intern 
     (concatenate 
      'string "F" (integer->string i)))))

;------------------------------------------------------------------
; put the face in the list of the faces of the polyhedron pol.
; DESTRUCTIVE (the polyhedron is modified)
;------------------------------------------------------------------
; ricontrollare il type-checking
(defun put-face (face pol)
  (if (get-type-checking) 
    (if (not (and (face? face) (polyhedron? pol))) 
      (error-handler 1 "put-face")))
  (if (get-pol-face face pol)
    (error-handler 3 "put-face")
    (setf (gethash (second face) 
		   (polyhedron-faces pol))
	  (get-id face)))
  (set-#f pol)
  pol)

;------------------------------------------------------------------
; put the flist (list of faces) in the polyhedron pol.
; DESTRUCTIVE (the old flist is lost)
;------------------------------------------------------------------
; ricontrollare il type-checking
(defun put-faces (faces-list-or-hash-table pol &optional (remove-old nil))
  (if (listp faces-list-or-hash-table)
    (progn
      (when remove-old (setf (polyhedron-faces pol)
			     (new-faces-table)))
      (mapcar #'(lambda (face)
		 (put-face face pol)) faces-list-or-hash-table))
    (progn
      (if remove-old 
	(setf (polyhedron-faces pol) faces-list-or-hash-table)
	(maphash #'(lambda (parameters name)
		    (put-face (list name parameters) pol))
		 faces-list-or-hash-table))))
  (set-#f pol)
  pol)

;------------------------------------------------------------------
;; The following functions are not destructive
;------------------------------------------------------------------

;------------------------------------------------------------------
; create a face equal to f but with name id
;------------------------------------------------------------------
; ricontrollare il type-checking
(defun set-fid (id f)
  (if (get-type-checking) 
    (if (not (and (face-symbol? id) (face? f)))
      (error-handler 1 "set-fid")))
  (list id (get-coord f)))


;------------------------------------------------------------------
; create a simplex equal to s but with name id
;------------------------------------------------------------------
; ricontrollare il type-checking
(defun set-sid (id s)
  (if (get-type-checking) 
    (if (not (and (simplex-symbol? id) (simplex? s)))
      (error-handler 1 "set-sid")))
  (list id (list (get-sign s) (get-vertices-or-faces s) (get-ad s))))

;------------------------------------------------------------------
; sets the coordinates (list of numbers defining the object)
; of a vertex or a face, to the value given in coord
;------------------------------------------------------------------
; ricontrollare il type-checking
(defun set-coord (coord v-or-f)
  (if (get-type-checking) 
    (if (not (and (every 'numberp coord) 
		  (or (vertex? v-or-f)
		      (face? v-or-f)))) 
      (error-handler 1 "set-coord")))
  (list (get-id v-or-f) coord))

;------------------------------------------------------------------
; sets the sign of the simplex s to sign 
;------------------------------------------------------------------
; ricontrollare il type-checking
(defun set-sign (sign s)
  (if (get-type-checking) 
    (if (not (and (plus-or-minus? sign)
		  (simplex? s))) 
      (error-handler 1 "set-svert")))
  (list (get-id s) (list sign (get-vertices-or-faces s) (get-ad s))))

;------------------------------------------------------------------
; sets the list of conses (face . sign) of the simplex s to sfaces 
;------------------------------------------------------------------
; ricontrollare il type-checking
(defun set-faces (sfaces s)
  (if (get-type-checking) 
    (if (not (and (every 'face-symbol? sfaces)
		  (simplex? s))) 
      (error-handler 1 "set-sfaces")))
  (list (get-id s) (list (get-sign s) sfaces (get-ad s))))

;;;;  Selectors
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun get-#F (pol)
  (if (get-type-checking) 
    (if (not (polyhedron? pol)) 
      (error-handler 1 "get-#K0")))
  (polyhedron-F pol))

(defun get-#C (pol)
  (if (get-type-checking) 
    (if (not (polyhedron? pol))
      (error-handler 1 "get-#kd")))
  (polyhedron-C pol))

;------------------------------------------------------------------
; returns the kind of the representation of pol 
;------------------------------------------------------------------
(defun get-pol-rep (pol)
  (if (get-type-checking)
   (if (not (polyhedron? pol))
     (error-handler 1 "get-pol-rep")))
  (polyhedron-representation pol))
;------------------------------------------------------------------
; returns the kind of the type of th basical element of the
; polyhedron 
;------------------------------------------------------------------
(defun get-pol-brick-type (pol)
  (if (get-type-checking)
   (if (not (polyhedron? pol))
     (error-handler 1 "get-pol-rep")))
  (polyhedron-kind-of-brick pol))

;------------------------------------------------------------------
; gets the list of elements definig the shape if the simplex s.
; (don't care elements type)
;------------------------------------------------------------------
(defun get-vertices-or-faces (s)
  (second (second s)))

;------------------------------------------------------------------
; returns the list of coordinates of the faces of s in pol
;------------------------------------------------------------------
(defun get-facescoord (s pol)
  (let ((faces (get-faces s)))
    (mapcar #'(lambda (f sign) 
		(get-fcoord f pol sign))
	    (mapcar 'first faces)
	    (mapcar 'rest faces))))

;------------------------------------------------------------------
; returns the hash table of faces of the polyhedron pol
;------------------------------------------------------------------
(defun get-faces (pol)
  (polyhedron-faces pol))

;------------------------------------------------------------------
; returns the face in the polyhedron pol if name fid
;------------------------------------------------------------------
;(defun get-face (face-name pol)
;  (get-face-from-table-for-name face-name 
;                                (get-faces pol)))
;
;------------------------------------------------------------------
; returns the list of faces of the polyhedron pol
;------------------------------------------------------------------
(defun get-flist (pol &optional (sorted nil))
  (let ((flist nil))
    (maphash #'(lambda (coord name)
		(push (make-face name coord) flist))
	     (get-faces pol))
    (if sorted 
	(sort flist #'(lambda (f1 f2)
			(string< (get-id f1)(get-id f2))))
      flist)))

;------------------------------------------------------------------
; Create a new hash-table that can be used as a table of faces
;------------------------------------------------------------------
(defun new-faces-table ()
  (make-hash-table :test 'equal))

;------------------------------------------------------------------
; Puts the face in the table.
; The table is modified.
; The list of parameters defining the face is used as key.
;------------------------------------------------------------------
(defun put-face-in-table (face table)
  (setf (gethash (second face)
		 table) (get-id face)))

;------------------------------------------------------------------
; Returns the face of the table with name face-name.
;------------------------------------------------------------------
(defun get-face-from-table-for-name (face-name table)
  (catch 'scan-faces
    (maphash #'(lambda (coord name)
		(if (symbol-equal face-name name )
		  (throw (quote scan-faces) (make-face name coord))))
	     table)))

;------------------------------------------------------------------
; Returns the face of the table with parameters face-par.
;------------------------------------------------------------------
(defun get-face-from-table-for-par (face-par table)
  (make-face (gethash face-par table) face-par))

;------------------------------------------------------------------
; returns a copy of the hash table of faces of the polyhedron pol
;------------------------------------------------------------------
(defun copy-faces (pol)
  (let ((faces (new-faces-table)))
    (maphash #'(lambda (par name)
		(setf (gethash par faces) name)) (get-faces pol))
    faces))

;------------------------------------------------------------------
; returns the face with label fid of the polyhedron pol 
;------------------------------------------------------------------
(defun get-pol-face (fid pol)
  (get-face-from-table-for-name fid (get-faces pol)))

;------------------------------------------------------------------
; returns the coordinates of the face with label fid
; of the polyhedron pol 
;------------------------------------------------------------------
(defun get-fcoord (fid pol &optional (sign '+))
  (let ((aux (get-pol-face fid pol)))
    (if aux (if (eq sign '+)
	      (second aux)
	      (mapcar '- (second aux)))
	nil)))

;------------------------------------------------------------------
; returns the list of faces of the simplex with label sid
; of the polyhedron pol 
;------------------------------------------------------------------
(defun get-sfaces (sid pol)
  (let ((aux (get-simplex sid pol)))
    (if aux (second (second aux)) nil)))

;------------------------------------------------------------------
; returns the list of faces' coordinates of the simplex with 
; label sid of the polyhedron pol 
;------------------------------------------------------------------
(defun get-sfacescoord (sid pol)
  (get-facescoord (get-simplex sid pol) pol))

;------------------------------------------------------------------
; returns the coordinates of the face with label fid
; in the list of faces flist 
;------------------------------------------------------------------
(defun find-fcoord (fid flist &optional (sign '+))
  (let ((aux (assoc fid flist :test 'symbol-equal)))
    (if aux (if (eq sign '+)
	      (second aux)
	      (mapcar '- (second aux)))
	nil)))

;------------------------------------------------------------------
; returns the list of faces of the simplex with label sid
; in the list of faces flist 
;------------------------------------------------------------------
(defun find-sfaces (sid slist)
  (let ((aux (assoc sid slist :test 'symbol-equal)))
    (if aux (second (second aux)) nil)))

;------------------------------------------------------------------
; returns the identifier of the vertex coincident with the 
; vertex of coordinates equal to coord-list taken from vlist.
;------------------------------------------------------------------
(defun find-same-vertex (coordlist vlist)
  (some #'(lambda (vertex)
	   (same-vertex coordlist vertex))
	vlist))

;------------------------------------------------------------------
; returns the cons (plane-identifier . sign) of the plane
; coincident with the  plane of parameters parameters-list
; taken from flist.
;------------------------------------------------------------------
(defun find-same-plane (parameters-list flist)
  (some #'(lambda (plane)
	   (same-plane parameters-list plane))
	flist))

;------------------------------------------------------------------
; Creates a new face and put it inthe hash table fhash.
; The parameters of the face are calculated from 
; the ponts specified in plist (list of lists of coordinates)
; The sign of the face is calculated considering the test
; vertex internal to the simplex (convex cell).
; Returns the cons simbol of the new face sign.
;------------------------------------------------------------------
;(defmacro add-new-face (fhash plist &optional (sign '+)(add-anyway nil))
;  `(let* ((new-plane-parameters (calc-plane-parameters ,sign ,plist))
;          (same-plane (and (not ,add-anyway)
;                           (gethash (first new-plane-parameters)
;                                     ,fhash))))
;     (cons
;      (or same-plane
;          (setf (gethash (first new-plane-parameters)
;                          ,fhash)
;                (advance-face-counter)))
;      (rest new-plane-parameters))))

(defmacro add-new-face (fhash plist test-vertex &optional (sign '+)(add-anyway nil))
  `(let* ((new-plane-parameters 
	   (calc-plane-parameters-with-test ,sign ,plist ,test-vertex))
	  (same-plane (and (not ,add-anyway)
			   (gethash (first new-plane-parameters)
				     ,fhash))))
     (cons
      (or same-plane
	  (setf (gethash (first new-plane-parameters)
			  ,fhash)
		(advance-face-counter)))
      (rest new-plane-parameters))))

;------------------------------------------------------------------
; Creates vertex and put it inthe list vlist.
; The coordinates of the vertex are calculated from 
; the parameters of the faces specified in flist (list of lists of
; face parameters)
; Returns the symbol
;------------------------------------------------------------------
(defmacro add-new-vertex (vlist flist &optional (add-anyway nil))
  `(let* ((intersection-point (calc-intersection-point ,flist))
	  (same-vertex (and (not ,add-anyway)
			    (find-same-vertex intersection-point ,vlist))))
     (or same-vertex
	 (progn
	   (push 
	    (make-vertex (advance-vertex-counter)
			 intersection-point)
	    ,vlist)
	   (previous-vertex-counter)))))

;------------------------------------------------------------------
; Returns the identifier of the face of the simplex the-simplex
; adjacent to the simplex of identifier adjacent-simplex
;------------------------------------------------------------------
(defun get-adjacent-face (the-simplex adjacent-simplex adjacent-sign)
  (let* ((flist (get-vertices-or-faces the-simplex))
	 (simplex-sign (get-sign the-simplex))
	 (adjcences (get-ad the-simplex))
	 (the-posistion (position adjacent-simplex adjcences)))
    (if the-posistion
      (if (equal simplex-sign adjacent-sign)
	(opposit-face (nth the-posistion flist))
	(nth the-posistion flist)))))

;------------------------------------------------------------------
; Reverses the sign of the cons (fid . sign)
;------------------------------------------------------------------
(defun opposit-face (face)
  (when face
    (cons (first face) (opposit-sign (rest face)))))

;------------------------------------------------------------------
; This function returns a polyhedron equal to pol but built
; with a representation based on vertices.
;------------------------------------------------------------------
(defun coerce-polyhedron-in-simplices-from-faces-to-vertices (pol)
  (let ((old-flist (get-flist pol))
	(old-slist (get-slist pol))
	(new-slist nil)
	(one-time-found-simplices nil)
	(two-times-found-simplices nil)
	(already-built-simplices nil)
	(new-vlist nil))
    (reset-vertex-counter 0)
    (do ((current-old-simplex (pop old-slist)
			      (find-el (or (pop two-times-found-simplices)
					   (pop one-time-found-simplices))
				       old-slist)))
	((not current-old-simplex))
      (let* ((faces-names
	      (mapcar 'first (get-vertices-or-faces current-old-simplex)))
	     (adjaceces (get-ad current-old-simplex))
	     (vertices nil)
	     (simplex-identifier (get-id current-old-simplex))
	     (coord-list (mapcar #'(lambda (fid)
				    (find-fcoord fid old-flist))
				 faces-names))
	     (first-adjacent nil)
	     (first-adjacent-position 0)
	     (second-adjacent nil))
	
	(push simplex-identifier already-built-simplices)
	
	(dotimes (i (length adjaceces))
	  (let* ((adjacent-name (nth i adjaceces)))
	    (when (not (numberp adjacent-name))
	    (if (member adjacent-name already-built-simplices)
	      (if first-adjacent
		(when (not second-adjacent)
		  (let ((adjacent (find-el adjacent-name new-slist)))
		    (setf second-adjacent 
			  (but-nth (position simplex-identifier
					     (get-ad adjacent))
				   (get-vert adjacent)))))
		(progn 
		  (let ((adjacent (find-el adjacent-name new-slist)))
		    (setf first-adjacent-position i
			  first-adjacent 
			  (but-nth 
			   (position simplex-identifier
				     (get-ad adjacent))
			   (get-vert adjacent))))))
	      (when (not (member adjacent-name two-times-found-simplices))
		(if (member adjacent-name one-time-found-simplices)
		  (progn (setf one-time-found-simplices
			       (delete adjacent-name one-time-found-simplices))
			 (push adjacent-name two-times-found-simplices))
		  (push  adjacent-name one-time-found-simplices)))))))
	(if (setf vertices first-adjacent)
	  (push-nth first-adjacent-position
		    (or (nth first-adjacent-position second-adjacent)
			(add-new-vertex new-vlist 
					(but-nth first-adjacent-position coord-list) t))
		    vertices)
	  (dotimes (current-adjacent-index 
		    (length adjaceces)
		    (setf vertices (nreverse vertices)))
	    (push
	     (add-new-vertex new-vlist 
			     (but-nth current-adjacent-index coord-list) t)
	     vertices)))
	(push (make-simplex simplex-identifier 
			    (get-sign current-old-simplex)
			    vertices
			    adjaceces)
	      new-slist)))
    (make-polyhedron new-vlist new-slist 'on-vertices)))

;------------------------------------------------------------------
; This function returns a polyhedron equal to pol but built
; with a representation based on faces
;------------------------------------------------------------------
(defun coerce-polyhedron-in-simplices-from-vertices-to-faces (pol)
  (let ((old-vlist (get-vlist pol))
	(old-slist (get-slist pol))
	(new-slist nil)
	(new-fhash (new-faces-table)))
    (reset-face-counter)
    (dolist (current-old-simplex old-slist)
      (let* ((vertices (get-vertices-or-faces current-old-simplex))
	     (adjaceces (get-ad current-old-simplex))
	     (faces nil)
	     (simplex-identifier (get-id current-old-simplex))
	     (simplex-sign (get-sign current-old-simplex))
	     (coord-list (mapcar #'(lambda (vid)
				    (find-vcoord vid old-vlist))
				 vertices)))
	(dotimes (current-adjacent-index (length adjaceces))
	  (push 
	   (or 
	    (get-adjacent-face (find-el  (nth current-adjacent-index adjaceces)
					new-slist)
			      simplex-identifier
			      simplex-sign)
	    (add-new-face new-fhash 
			  (but-nth current-adjacent-index coord-list)
			  (nth current-adjacent-index coord-list)
			  simplex-sign))
	   faces))
	(push (make-simplex simplex-identifier 
			    simplex-sign
			    (nreverse faces)
			    adjaceces)
	      new-slist)))
    (make-polyhedron new-fhash new-slist 'on-faces)))

;------------------------------------------------------------------
; This function returns a new polyhedron equal to POL but
; built with representation type NEW-REP
;------------------------------------------------------------------
(defun coerce-polyhedron (pol new-rep)
  (let ((old-rep (get-pol-rep pol))
	(old-bricks (get-pol-brick-type pol))
	new-polyhedron)
    (setf new-polyhedron 
	  (cond
	   ((or (symbol-equal old-rep new-rep)
		(symbol-equal old-bricks new-rep)) pol)
	   ((symbol-equal old-bricks 'simplices)
	    (cond
	     ((symbol-equal old-rep 'on-faces)
	      (cond
	       ((symbol-equal new-rep 'on-vertices)
		(coerce-polyhedron-in-simplices-from-faces-to-vertices pol))
	       ((symbol-equal new-rep 'convex-cells)
		(coerce-polyhedron-on-faces-from-simplices-to-convex-cells pol))))
	     ((symbol-equal old-rep 'on-vertices)
	      (cond
	       ((symbol-equal new-rep 'on-faces)
		(coerce-polyhedron-in-simplices-from-vertices-to-faces pol))
	       ((symbol-equal new-rep 'convex-cells)
		(coerce-polyhedron-on-vertices-from-simplices-to-convex-cells pol))))))
	   ((symbol-equal  old-bricks 'convex-cells)
	    (cond
	     ((symbol-equal old-rep 'on-faces)
	      (cond
	       ((symbol-equal new-rep 'on-vertices)
		(coerce-polyhedron-on-convex-cells-from-faces-to-vertices pol))
	       ((symbol-equal new-rep 'simplices)
		(error-handler " Coercion from convex-cells to simplices on faces is currently not supported. "
			       " coerce-polyhedron "))))
	     ((symbol-equal old-rep 'on-vertices)
	      (cond
	       ((symbol-equal new-rep 'on-faces)
		(coerce-polyhedron-on-convex-cells-from-vertices-to-faces pol))
	       ((symbol-equal new-rep 'simplices)
		(error-handler " Coercion from convex-cells to simplices on vertices is currently not supported. "
			       " coerce-polyhedron "))))))))
    (if new-polyhedron
      (progn
	(set-npol new-polyhedron (get-npol pol))
	(set-dpol new-polyhedron (get-dpol pol))
	(set-#kd  new-polyhedron)
	(set-#k0  new-polyhedron) 
	(set-#F  new-polyhedron)
	(set-#C  new-polyhedron) 
	new-polyhedron)
      (error-handler " Bad coercion parameters " " coerce-polyhedron "))))
