; -*- package: kernel -*-

(inpackage kernel)


(defmacro add-new-vertex-9 (vlist flist &optional (add-anyway nil))
  `(let* ((intersection-point (calc-intersection-point ,flist))
          (same-vertex (and (not ,add-anyway)
                            (find-same-vertex-9 intersection-point ,vlist))))
     
     (or same-vertex
         (progn
           (push 
            (make-vertex (advance-vertex-counter)
                         intersection-point)
            ,vlist)
           (previous-vertex-counter)))))

(defun find-same-vertex-9 (coordlist vlist)
  (some #'(lambda (vertex)
           (same-vertex-9 coordlist vertex))
        vlist))

(defmacro almost-equal-9 (number1 number2)
  `(< (abs (- ,number1 ,number2)) 1e-9))

(defun same-vertex-9 (coord-list vertex)
  (if (every #'(lambda (x y)
                (almost-equal-9 x y)) coord-list (get-coord vertex))
    (get-id vertex)))


(defun coerce-polyhedron-on-convex-cells-from-faces-to-vertices (pol)
  (let ((old-cells (get-clist pol))
        (new-cells (new-convex-cells-table))
        (old-faces-list (get-flist pol))
        (new-vertices nil)
        new-polyhedron)
    (reset-vertex-counter)
    (dolist (current-old-cell old-cells)
      (let* ((old-cell-vertices (get-cell-vertices current-old-cell))
             (cell-adjacences (get-cell-adjacences  current-old-cell))
             (open-faces (convex-cell-open-faces current-old-cell))
             (old-cell-faces (get-cell-faces current-old-cell))
             (new-cell-vertices nil)
             (new-cell-faces nil)
             (cell-identifier (get-cell-identifier current-old-cell))
             (cell-sign (get-cell-sign current-old-cell))
             (faces-list (mapcar #'(lambda (fid)
                                       (assoc fid old-faces-list)) ;eliminato quote davanti old-faces-list
                                    (mapcar 'first old-cell-faces)))
             (faces-list-list (list faces-list)))
        (setf (cdr faces-list-list) faces-list-list)
        (dotimes (current-vertex-index (length old-cell-vertices))
          (push 
           (add-new-vertex-9 new-vertices 
                           (mapcar 'find-fcoord  
                                   (nth current-vertex-index old-cell-vertices)
                                   faces-list-list))
           new-cell-vertices))

        (setf new-cell-vertices (nreverse new-cell-vertices)) ;aggiunto per disperazione

        (dotimes (i (length old-cell-faces))
          (let ((old-face-name  (first (nth i old-cell-faces)))
                (new-face nil))
            (dotimes (j (length old-cell-vertices))
              (if (member old-face-name (nth j old-cell-vertices))
                (push (nth j new-cell-vertices) new-face)))
            (push new-face new-cell-faces)))

        (setf new-cell-faces (nreverse new-cell-faces)) ;aggiunto per disperazione

        (setf (gethash  cell-identifier new-cells)
              (make-convex-cell cell-identifier 
                                cell-sign
                                new-cell-faces
                                new-cell-vertices
                                cell-adjacences
                                nil
                                open-faces))))
    (setf new-polyhedron
          (make-empty-polyhedron 'on-vertices 'convex-cells))

    (put-vlist new-vertices new-polyhedron t)
    (put-cells new-cells new-polyhedron t)
    new-polyhedron))
