; -*- package: hyperplanes -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Beneath-Beyond Algorithm :Spazi affini                    ;               
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: G. Proietti                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;(30/7/93)

(provide 'Affines)

(use-package 'matlib)


(inpackage hyperplanes)

(export '(CalcAffineSpace 
          Gen0DimAffineSpace
          ReducedEqnList
          EvalEqnOnPoint
          ApproxEqual 
          Is-Point-Not-In-Affine?))
 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Test delle funzioni :
; N.B. : Essendo la soluzione non univoca, modifiche al codice possono
;        portare a risultati diversi ma egualmente corretti
;
;(calcaffinespace '((1 1 1)) )
;((1 0 0 -1) (0 1 0 -1) (0 0 1 -1))
;(calcaffinespace '((1 1 1) (2 3 4) ))
;((1 -1/2 0 -1/2) (1 0 -1/3 -2/3))
;(calcaffinespace '((1 1 1) (2 3 4) (5 0 1)))
;((-1/8 -1/2 3/8 1/4))
;(calcaffinespace '((1 1 1) (1 3 4) ))
;((1 0 0 -1) (0 1 -2/3 -1/3))
;(calcaffinespace '((1 1 1) (1 1 4) ))
;((1 0 0 -1) (0 1 0 -1))
;(calcaffinespace '((0 0 0) (2 3 4) (5 0 1)))
;((-1/9 -2/3 5/9 0))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;---------------------------------------------------------
;Ricava, a partire da una lista di punti affinemente indipendenti, 
;posta come lista di liste, una lista di equazioni che definiscano 
;il sottospazio affine contenente l'insieme di punti stessi.
(defun CalcAffineSpace (PointList)
  (let ( (EqnList (Gen1DimAffineSpace (car PointList)))
         )
    (dolist (Point (cdr PointList) EqnList)
      (let ( (Index (Is-Point-Not-In-Affine? EqnList Point)))
        (if Index (setq EqnList (ReducedEqnList EqnList Index Point)))
        )
      )
    )
  )

;---------------------------------------------------------
;Genera, a partire da un punto nello spazio Rn, una lista di
;n equazioni di iperpiani individuanti il punto stesso e
;e paralleli ai sottospazi di dimensione n-1 identificati dagli
; assi.
; es :     (Gen0DimAffineSpace  '( 1 3 ))
;
; produce   ((1 0 -1) (0 1 -3))
;

(defun Gen0DimAffineSpace (Point)
  (let ( (len (length Point) )
         Result
         Eqn
         )    
    (dotimes (i len Result)
      (setq Eqn (make-list len :initial-element 0))
      (setf (nth i Eqn) 1)
      (setq Result (append Result (list (append Eqn (list (- (nth i Point)))))))
      )
    )
  )

;---------------------------------------------------------
;Data una lista di equazioni specificanti iperpiani che individuano
;, con la loro intersezione, un sottospazio affine ed un
;punto esterno a tale sottospazio, calcola una nuova lista di eq.
;definente il sottospazio affine contenente anche il punto.
;Essa lo calcola facendo ruotare opportunamente gli iperpiani,
;eliminando dall'insieme l'iperpiano di indice specificato nella
;lista di equazioni.
(defun ReducedEqnList (EqnList PurgePlaneIndex Point)
  (let ( (PurgePlane (nth PurgePlaneIndex EqnList))
         )
;(print PurgePlaneIndex)
;(print EqnList)
;(print Point)
    (setq EqnList (Lisp-Extension:but-nth PurgePlaneIndex EqnList))
    (mapcar #'(lambda (P) (Plane PurgePlane P Point)) EqnList)
    )
  )


;    (dolist (eqn EqnList NewEqnList)
;      (if (ApproxEqual (EvalEqnOnPoint FirstPlane Point) 0)
;        (progn
;          (setq NewEqnList (cons FirstPlane NewEqnList))
;          (setq FirstPlane eqn)
;          )
;        (setq NewEqnList (cons (Plane FirstPlane eqn Point) NewEqnList))
;        )
;      )
;    )
;  )
     

;(defun ReducedEqnList (EqnList Point)
;  (mapcar #'(lambda (x) (Plane x (car EqnList) Point)) (cdr EqnList))
;  )

;Calcola, a partire dal fascio di iperpiani individuato
;dalla coppia di iperpiani forniti, l'equazione del piano
;passante per il punto dato.
;Se uno dei piani e' NIL, viene calcolato il piano parallelo
;all'altro e passante per il punto dato.
(defun Plane( Plane1 Plane2 Point)
  (let ( (a (EvalEqnOnPoint Plane1 Point))
         (b (EvalEqnOnPoint Plane2 Point))
         Param
         )
    (cond ( (null Plane1)    ;Piano risultato // a Plane2
            (append    
             (butlast Plane2) 
             (list (- (EvalEqnOnPoint (append (butlast Plane2) '(0)) Point )))))
          ( (null Plane2)    ;Piano risultato // a Plane1
            (append    
             (butlast Plane1) 
             (list (- (EvalEqnOnPoint (append (butlast Plane1) '(0)) Point )))))
          ( (ApproxEqual a 0) Plane1)
          ( (ApproxEqual b 0) Plane2)
          (      T    
                 (progn
                   (setq Param (- (/ a b)))
                   (mapcar #'(lambda (x y) (+ x (* y Param))) Plane1 Plane2)
                   )
                 )
          )
    )
  )


;Valuta il valore dell'equazione sul punto fornito
;,considerando l'ultimo elemento di Eqn come il termine noto,
;riportato al primo membro.
(defun EvalEqnOnPoint (Eqn Point)
  (if Eqn
    (matlib:Vet-Product 
     (coerce Eqn 'vector) 
     (coerce (append Point '(1)) 'vector))
    nil
    )
  )

;Confronta due numeri, ritornando True se sono uguali a meno di
;un errore numerico di entita' pari a 1e-9
; Successivamente modificato ad 1e-6 in seguito a problemi
; della funzione di esportazione verso roblan nel caso di
; facce quasi complanari.

(defun ApproxEqual (a b)
  (< (abs (- a b)) 1e-6)
  )

;
;Data una lista di piani definenti uno spazio affine ed un punto,
;ritorna nil se il punto appartiene allo spazio affine, altrimenti
;ritorna l'indice della prima delle equazioni violate dal punto.
;Es (IS-POINT-NOT-IN-AFFINE? '((1 0 0 -1)(0 1 0 -1)) '(1 1 1))
;   NIL
;   (IS-POINT-NOT-IN-AFFINE? '((1 0 0 -1)(0 1 0 -1)) '(1 0 1))
;   (0 1 0 -1)
;
(defun Is-Point-Not-In-Affine? ( Affine Point )
  (dotimes (i (length Affine) nil)
    (if (not (ApproxEqual 0 (EvalEqnOnPoint (nth i Affine) Point))) 
      (return-from Is-Point-Not-In-Affine? i)
      )
    )
  )
