; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: MKPOL : Funzioni principali                               ;               
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: G. Proietti                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;(30/7/1993)
;

(provide 'BuildPC )

(inpackage kernel)

(export '(Build-Pol-Complex))

(require 'Affines)
(require 'Coerenza)
(require 'GenPol)
(require 'Beneath)
(require 'Lex_Sort)
(require 'IncGraph)
(require 'Not_Pyr)
(require 'Pyr)
(require 'Gram_S)
(require 'Face_DB)

;----------------------------------------------------------------------
;Costruisce, a partire dalle liste di vertici, celle e poliedri forniti
;un comlesso poliedrale, eseguedo, opzionalmente, la costruzione del complesso 
;rappresentato tramite facce ed una serie di controlli
;preventivi.
;Per default la costruzione del complesso e' on-vertices. Si puo' ottenerlo 'on-faces
;specificando 'on-faces il parametro opzionale rep.
;I controlli sono attivati ponendo a 'check il parametro check.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun Build-Pol-Complex ( vlist clist plist &optional rep checks)
  (cond
   ((null checks)) ;non fai niente
   ((or (symbol-equal checks 'check))
    ;Verifiche correttezza sintattica
    (if (not (Check-Lists vlist clist plist)) (return-from Build-Pol-Complex nil))
    )     
   )
  (Generate-PolComplex vlist clist plist rep)
  )

;----------------------------------------------------------------------
;Esegue una serie di controlli sintattici sulle liste di vertici, celle
;e poliedri per verificare la coerenza delle informazioni contenute nelle
;stesse.
;Quest'insieme di controlli non fanno riferimento alla conformazione geometrica
;dell'oggetto che si vuole definire.
;Si verifica che :
; - Tutti i vertici abbiano lo stesso numero di coordinate
; - Gli indici individuanti i vertici e le celle siano tutti positivi
; - Tutti i vertici e le celle specificate nelle liste di appartenenza 
;   siano definiti nelle rispettive liste

(defun Check-Lists (vlist clist plist)
  (cond
   ((VerifyVerticesDimension vlist (length (car vlist))) 
    (print "Vertice con numero di coordinate errato")
    nil)

   ((VerifyPositive clist) 
    (print "Indice vertice negativo")
    nil);Indici vertici positivi

   ((VerifyPositive plist)  ;Indici celle positivi
    (print "Indice cella negativo")
    nil)

   ((VerifyIndices  clist vlist) ;Tutti vertici esistenti
    (print "Vertice non fornito")
    nil)

   ((VerifyIndices  plist clist) ;Tutte celle esistenti
    (print "Cella non fornita")
    nil)
   
   ( T  T )  ;altrimenti ok
   )
  )

