; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Proietti Giuseppe  (30/7/93)
;

(provide 'Coerenza )


(inpackage kernel)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Funzioni di verifica coerenza tra i dati 
; lista dei vertici, lista vertici dele celle e
; lista celle dei poliedri.

;
; Lista dei vertici : ( (coord0) (coord1) ...... (coordn) )
; Lista celle       : ( (ind. cella 0) (ind. cella 1) ..... (ind. cella k) )
; Lista poliedri    : ( (celle pol 0) (celle pol 1)  ..... (celle pol. m) )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Verifica che nella lista di vertici tutti i punti possiedano lo
;stesso numero di coordinate passato nella variabile dim.
;Ritorna nil se cio' avviene, altrimenti T .
(defun VerifyVerticesDimension (vlist dim)
  (dolist (elem vlist nil)
    (if (/= (length elem) dim) (return-from VerifyVerticesDimension T)
        )
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Verifica che non vi siano indici negativi nella lista di liste
;Ritorna nil se non vi sono elementi negativi, altrimenti T
(defun VerifyPositive( lista )
  (dolist (el lista nil)
    (if (not (null (find-if #'(lambda (x) (< x 0)) el)) ) 
      (return-from VerifyPositive T))
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Verifica che di tutti gli indici contenuti nella lista di liste lista1
; esista lelemento corrispondente indicato nella lista2
; ATT: in lista2 NON CI DEVONO ESSERE INDICI NEGATIVI.
; Ritorna nil se le liste sono coerenti, altrimenti T.
(defun VerifyIndices( lista1 lista2)
  (dolist (el lista1 nil)
    (mapcar #'(lambda (x)
               (if (null (nth x lista2)) (return-from VerifyIndices T))
               )
            el
            )
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Verifica che tutti gli elementi di lista1 appartengano almeno ad
; un elemento di lista2
; Ritorna nil se vi e' un elemento di lista1 non menzionato in nessun
; elemento della lista2 : non usata
(defun VerifyAppartenence( lista1 lista2 )
  (let ( (aux (make-list (length lista1) :initial-element 0))
         )
    (dolist (el1 lista2)
      (dolist (el2 el1)
        (setf (nth el2 aux) 1)
        )
      )
    (return-from VerifyAppartenence (null (find 0 aux)))
    )
  )
