;-*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Beneath-Beyond Algorithm :Facce per ricostruzione poliedri;               
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: G. Proietti                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;(20/9/93)

(provide 'Face_Db )

(inpackage kernel)

;Il database delle facce e' una a-list  cosi' organizzata:
; ( FaceID (( vertici ) ( coeff. )) ) 
;
;P. Es. ( (F0 ((v0 v2) (0 1 1) sign)) .... (F5 ((v11 v8) (3 2 1)) sign) )


;A partire da una lista di simboli di vertici e da un database, ritorna una
;lista di liste specificante, per ciascun vertice, gli identificatori delle
;facce incidenti su di esso.
(defun Transform-VIDs-in-List-Of-List-Of-FIDs (VIDList DataBase)
  (mapcar #'(lambda (x)
             (Transform-VID-in-List-Of-FIDs x DataBase)
             )
          VIDList
          )
  )

(defun Transform-VID-in-List-Of-FIDs (VID DataBase)
  (let ( Res )
    (dolist (Elem DataBase Res)
      (if (find VID (car (second Elem)))
        (setq Res (append Res (list (first Elem))))
        )
      )
    )
  )


;Crea DataBase delle facce vuoto.
(defun Make-Empty-Face-DataBase ()
  (list nil)
  )

;Ritorna l'identificatore della faccia passante per i vertici di simboli specificati nella
;VertList.
;Se questa faccia gia' esisteva, ritorna l'id. della faccia esistente, con segno invesro,
; altrimenti ne ritorna
;uno nuovo ed aggiunge la nuova faccia la database.
(defun Make-a-Face-and-Update-DataBase ( VertList Coeff Sign DataBase )
  (let ( (FID (Exist-Face-in-DataBase? VertList DataBase))
         )
    (if (null FID)
      (let ( NewDataBaseEntry
             )
        (setq FID (advance-face-counter))
        (setq NewDataBaseEntry (list FID (list VertList Coeff Sign)))
        (if (null (first DataBase))
          (setf (first DataBase) NewDataBaseEntry)
          ;else
          (setf (cdr (last DataBase)) (list NewDataBaseEntry))
          )
        (cons FID Sign)
        )
      ;else
      ;(cons FID (Inversed-Sign Sign))
      (cons FID Sign)
      )
    )
  )

(defun Inversed-Sign (Sign)
  (if (lisp-extension::Symbol-Equal '+ Sign)
    (return-from Inversed-Sign '-)
    ;else
    (return-from Inversed-Sign '+)
    )
  )
  

;Controlla, nel database, se vi e' gia' una faccia a cui siano associati i vertici
;associati a Face.
;IMPORTANTE: i test sono eseguiti esclusivamente sugli identificatori dei vertici
;Se trovata, ritorna l'identificatore della faccia passante per i medesimi vertici,
;altrimenti nil.
(defun Exist-Face-in-DataBase? (ListOfVerts DataBase)
  (dolist (Element DataBase nil)
    (if (Equal-Elements-List? (car (second Element)) ListOfVerts)
      (return-from Exist-Face-in-DataBase? (first Element))
      )
    )
  )

;Confronto insiemistico tra' due liste
;Confronta che le due liste abbiano gli stessi elementi, anche se disposti
;in ordine diverso.
;Confronta prima le due lunghezze. Se queste sono diverse, le liste sono diverse.
;Se le lunghezze sono uguali, calcola gli elementi comuni alle due, e, se sono in
;numero pari alla lunghezza, ritorna true.
(defun Equal-Elements-List? (Lista1 Lista2)
  (let ( (Lung1 (length Lista1))
         )
    (if (= Lung1 (length Lista2)) ;Se le lunghezze sono uguali
      (= Lung1 (length (intersection Lista1 Lista2)))
      ;else
      nil
      )
    )
  )
