; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Costruzione di complessi da vertici                       ;               
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: G. Proietti                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; (22/9/93)

(provide 'Lex_Sort)

(require 'affines)

(use-package 'hyperplanes)


(inpackage kernel)

;------------------------------------------
;Esegue un sort della lista di cui ciascun elemento e' costituito
;dalla coppia label del vertice  e coordinate dello stesso.

(defun Lex-Sort-with-Labels ( lista )
  (sort lista 'Lex-Cmp-with-Labels)
  )

;--------------------------------------
; Data una lista ordinata lessicograficamente, ritorna una lista
; depurata da tutti i doppioni.

(defun Purge-Duplicates (lista)
  (let ( res
         (len  (length lista))
         )
    (setf (cdr (last lista)) lista)
    (dotimes (i len res)
      (if  
        (not (Lexicographic-Equal? (nth i lista) (nth (1+ i) lista))) 
        (setq res (append res (list (nth i lista))))
        )
      )
    )
  )


;--------------------------------------
(defun Lex-Equal? ( lista1 lista2 )
  (let ( (lista1 (second lista1))
         (lista2 (second lista2))
         ) 
    (Equal-Coords? lista1 lista2)
    )
  )


;--------------------------------------
; funzione che confronta lessicograficamente due liste e
; ritorna true se il primo e' minore del secondo, altrimenti false
;
(defun Lexicographic-Cmp ( lista1 lista2 )
  (dotimes (i (length lista1))
    (cond  
     ((> (nth i lista1) (nth i lista2)) (return-from Lexicographic-Cmp nil))
     ((< (nth i lista1) (nth i lista2)) (return-from Lexicographic-Cmp T))
     )
    )
  (return-from Lexicographic-Cmp nil)
  )
  

;--------------------------------------
; Come sopra, ma per liste dotate di indici
(defun Lex-Cmp-with-Labels (lista1 lista2)
  (Lexicographic-Cmp (second lista1) (second lista2))
  )

;---------------------------------------
;Elimina vertici uguali da lista ordinata lessicograficamente
;(old version)
;(defun Lex-Purge-Equal-Verts (ListaVert)
;  (let ((len (length ListaVert))
;        ResultList
;        )
;    (if (<= len 1)
;      ListaVert
;      ;else
;      (progn
;        (setq ResultList (list (car ListaVert)))
;        (dotimes (i (1- len) ResultList)
;          (if (not (Equal-Coords? (second (nth i ListaVert))
;                                  (second (nth (1+ i) ListaVert))
;                                  )
;                   )
;            (setq ResultList (append ResultList (list (nth (1+ i) ListaVert))))
;            )
;          )
;        )
;      )
;    )
;  )


;--------------------------------------
; Data una lista ordinata lessicograficamente, ritorna una lista
; depurata da tutti i doppioni ed una lista associativa indicante le
; corrispondenze tra' vertici cancellati e vertici non cancellati.

(defun Lex-Purge-Equal-Verts (lista)
  (let ( (res (list (car lista)))
         (len  (1- (length lista)))
         Correspondence-List
         (TestVert (car lista))
         )
    (setq lista (cdr lista))
    (dotimes (i len (values res Correspondence-List))
      (let ( (ActualVert (nth i lista)) )
        (if  
          (not (Lex-Equal? ActualVert TestVert))
          (progn
            (setq res (append res (list ActualVert)))
            (setq TestVert ActualVert)
            )
          ;else
          (setq Correspondence-List (cons (list (first ActualVert)
                                                (first TestVert)
                                                )
                                          Correspondence-List
                                          )
                )
          )
        )
      )
    )
  )


    
(defun Equal-Coords? (Coords1 Coords2)
  (dotimes (i (length Coords1) T)
    (if (not (ApproxEqual (nth i Coords1) (nth i Coords2)))
      (return-from Equal-Coords? nil))
    )
  )
