; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Costruzione di complessi da vertici                       ;               
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: G. Proietti                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; (30/8/93)

(provide 'Not_Pyr)

(require 'Incgraph)
(require 'Affines)

(use-package 'hyperplanes)

(inpackage kernel)
        
;Aggiornamento non piramidale per l'algoritmo di Beneath-Beyond
(defun Not-Pyr-Update (Convex Point)
  (let ( (FaceList (Coloring Convex Point)))
    (ChangeGraph Convex Point FaceList)
    (ResetColor   (ConvexHull-Graph Convex))
    (ResetInduced (ConvexHull-Graph Convex))
    )
  )

; Aggiornamento non piramidale del grafo di incidenza
; del poliedro.
(defun ChangeGraph (Convex Point FaceList)
  (let ( (Graph       (ConvexHull-Graph Convex))
         (Root        (ConvexHull-Root  Convex))
         (Empty       (ConvexHull-Empty Convex))
         (DimPol      (ConvexHull-Dim   Convex))
         (PointCoord  (second Point))
         (Intern      (ConvexHull-InternalPoint Convex))
         Ind
         FaceColor
         Dim
         )
    ;Nodo Vuoto
    (SetInduced Graph Empty (NewFace Graph))
    (SetVertex Graph (ConvexHull-Empty Convex) Point)
    
    (dotimes (j (1+ DimPol))
      (setq Dim (1- j))
      (dolist (Face (nth j FaceList))
        (setq FaceColor (GetColor Graph Face))
        (if (or (EqualColor? (Purple) FaceColor) (EqualColor? (Brown) FaceColor))
          (progn
            (setq Ind (SetInduced Graph Face (NewFace Graph)))
            (if (= Dim -1) (SetVertex Graph Ind Point))
            (if (= Dim (- DimPol 2))
              (let* ((SupPlane1 (GetHyperplane Graph (first (GetSuperfaces Graph Face))))
                     (SupPlane2 (GetHyperplane Graph (second (GetSuperfaces Graph Face))))
                     (NewPlane  (SignedPlane SupPlane1 SupPlane2 PointCoord Intern))
                     )
                (SetHyperPlane Graph Ind NewPlane)
                (Creat-Super-Sub-Face-Rel Graph Root Ind)
                )
              )
            (Creat-Super-Sub-Face-Rel Graph Ind Face)
            (dolist (sub (GetSubFaces Graph Face))
              (if (GetInduced Graph sub)
                (Creat-Super-Sub-Face-Rel Graph Ind (GetInduced Graph sub))
                )
              )
            )
          )
        (if (EqualColor? (Orange) FaceColor)
          (progn
            (dolist (sup (GetSuperFaces Graph Face))
              (if (EqualColor? (GetColor Graph sup) (Yellow))
                (dolist (sub (GetSubFaces Graph Face))
                  (if (EqualColor? (GetColor Graph sub) (Brown))
                    (Creat-Super-Sub-Face-Rel Graph sup (GetInduced Graph sub))
                    )
                  )
                )
              )
            ;La cancellazione delle facce dalle liste non e' implementata
            (DeleteFace Graph Face) ;Cancellazione faccia dal grafo
            )
          )
        (if (EqualColor? (Red) FaceColor)
          (DeleteFace Graph Face) ;Cancellazione faccia dal grafo
          )
        )
      )
    )
  )
          
;Colorazione 
(defun Coloring (Convex Point)
  (let ( (Graph       (ConvexHull-Graph Convex))
         (PointCoord  (second Point))
         (Root        (ConvexHull-Root Convex))
         (FaceList    (make-list (- (ConvexHull-Dim Convex) 1) :initial-element nil))
         Ld-1
         Ld-2
         temp
         )
    (dolist (FacetID (GetSubfaces Graph Root) temp)
      (setq temp (Depth-First-Search Graph FacetID PointCoord 
                          Ld-1
                          Ld-2
                          )
            )
      (setq Ld-1 (second temp))
      (setq Ld-2 (first temp))
      )
    (setq FaceList (append FaceList temp))
    (Find-and-Color-All-Remaining-Faces Convex FaceList)
    FaceList
    )
  )

; Colorazione di tutte le facce e di tutti 

(defun Depth-First-Search (Graph FacetID PointCoord Ld-1 Ld-2)
  (let ( (FacetColor (GetColor Graph FacetID)) )
    (if (EqualColor? (Uncolor) FacetColor)
      (progn 
        (setq Ld-1 (cons FacetID Ld-1))  ;Se la facet non e' colorata aggiungi a lista ! DIVERSO DA LIBRO !!!   
        (setq FacetColor (ColorFacet Graph FacetID PointCoord)) ;e colorala
        )
      )
    (if (not (EqualColor? (Blue) FacetColor))
      (let ( (SubFaces (GetSubfaces Graph FacetID))
             )
        (dolist (sub SubFaces (list Ld-2 Ld-1))
          (let* ( (sup (car (remove FacetID (GetSuperfaces Graph sub))) )
                  temp )
            (if (EqualColor? (Uncolor) (GetColor Graph sup))
              (progn
                (setq temp (Depth-First-Search Graph sup PointCoord Ld-1 Ld-2))
                (setq Ld-1 (second temp))
                (setq Ld-2 (first temp))
                )
              )
            (if (EqualColor? (Uncolor) (GetColor Graph sub))
              (progn
                (setq Ld-2 (cons sub Ld-2))
                (SetColor Graph sub (MergeColors 
                                     (GetColor Graph sup) 
                                     (GetColor Graph FacetID)
                                     )
                          )
                )
              )
            )
          )
        )
      (list Ld-2 Ld-1)
      )
    )
  )

;Lista = ( L-1 L0 L1 L2 ..... Ld-2 Ld-1 )

(defun Find-and-Color-All-Remaining-Faces ( Convex Lista )
  (let ( (Graph       (ConvexHull-Graph Convex))
         (Dim         (- (ConvexHull-Dim Convex) 3))
         i
         )
    (dotimes (j (+ Dim 2) Lista)
      (setq i (- Dim j))
      (dolist (Face (nth (+ i 2) Lista) )
        (dolist (Sub (GetSubfaces Graph Face))
          (if (EqualColor? (GetColor Graph Sub) (UnColor))
            (setf (nth (1+ i) Lista) (cons Sub (nth (1+ i) Lista)))
            )
          (SetColor Graph Sub (MergeColors (GetColor Graph Sub) 
                                           (GetColor Graph Face)
                                           )
                    )
          )
        )
      )
    )
  )

; Colorazione di una Facet rispetto al punto di coordinate passato
; ritorna il colore calcolato
(defun ColorFacet (Graph FaceID PointCoord)
  (let* ( (Plane (GetHyperPlane Graph FaceID))
          (Value (EvalEqnOnPoint Plane PointCoord))
          )
    (SetColor Graph FaceID 
              (cond ( (ApproxEqual Value 0)  (Yellow))
                    ( (<           Value 0)  (Blue))
                    ( (>           Value 0)  (Red))
                    )
              )
    )
   )
