; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Costruzione di complessi da vertici                       ;               
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: G. Proietti                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; (30/8/93)

(provide 'Pyr)

(require 'Incgraph)
(require 'Not_Pyr)
(require 'Affines)

(use-package 'hyperplanes)

(inpackage kernel)

; Funzioni per il trattamento del caso di incremento "piramidale" del
; poliedro, cioe' del caso in cui il nuovo punto non si trovi nello 
; spazio affine contenente il poliedro.


;Aggiornamento piramidale
; Convex : Struttura ConvexHull
; Index  : Indice di iperpiano nel guscio affine base della piramide
; Point  : lista ( IDVertice (coordinate))

(defun Pyr-Update (Convex Index Point)
  (SetVertex (ConvexHull-Graph Convex) (ConvexHull-Empty Convex) Point)
  (DuplicateGraph Convex (ConvexHull-Root Convex) (ConvexHull-Dim Convex))
  (setf (ConvexHull-Dim Convex) (1+ (ConvexHull-Dim Convex)))
  (if (= 0 (ConvexHull-Dim Convex)) 
    (progn
      (setf (ConvexHull-AffineSpace Convex) (Gen0DimAffineSpace (second Point)))
      (SetVertex (ConvexHull-Graph Convex) (ConvexHull-Root Convex) Point )
      )
    ;else
    (progn
      (RecalcFacetsWithHyperPlane Convex Index (second Point))
      (RecalcAffine Convex Index (second Point))
      )
    )
  (ResetInduced (ConvexHull-Graph Convex))
  )

;Ricalcolo del nuovo spazio affine, ottenuto a partire dal vecchio, che,
; privato dell'equazione di indice Index, che non contiene il punto
; di coordinate specificate in PointCoords, viene ricavato considerando
; i fasci di piani individuati dal piano scartato preso in coppia a turno
; con ciascuno degli altri piani, imponendo il passaggio per il punto dato.
(defun RecalcAffine (Convex Index PointCoords)
  (setf (ConvexHull-AffineSpace Convex) 
        (ReducedEqnList (ConvexHull-AffineSpace Convex) Index PointCoords))
  )

;Esegue il ricalcolo di tutte le facet del convesso, in modo da
; ottenere le facet del nuovo convesso a cui e' stato agiunto il
; nuovo punto.
; Per tutte le vecchie facet (individuate dal fatto che hanno il
; campo induced non nullo) calcola il piano appartenente al fascio
; di piani individuato dal piano della vecchia facet e dal piano
; in posizione Index nella lista di equazioni definenti
; lo spazio affine, che passi per il punto appena aggiunto e setta
; il piano della nuova facet a questo.
; Unica eccezione il nodo specificante la faccia aggiunta come induced
; della faccia a dimensione piena, che corrisponde alla faccia della
; base della piramide, che avra' associato proprio l'iperpiano
; scartato dall'insieme dello spazio affine.
(defun RecalcFacetsWithHyperPlane (Convex Index PointCoords)
  (let* ( (G      (ConvexHull-Graph Convex))
          (Root   (ConvexHull-Root Convex))
          (HyperPlane  (nth Index (ConvexHull-AffineSpace Convex)))
          (IntPoint (ConvexHull-InternalPoint Convex))
          (Facets (GraphNode-Sub (gethash Root G)))
          )
    (dolist (Facet Facets)
      (if (GetInduced G Facet)
        (SetHyperPlane G Facet 
                       (SignedPlane (GetHyperPlane G Facet) HyperPlane PointCoords IntPoint))
        ;else
        (SetHyperPlane G Facet (if (> (EvalEqnOnPoint HyperPlane IntPoint) 0)
                                 (mapcar #'- HyperPlane)
                                 HyperPlane
                                 )
                       )
        )
      )
    )
  )

(defun SignedPlane (Plane1 Plane2 PointCoord NegativePointCoord)
  (let ( (P (hyperplanes::Plane Plane1 Plane2 PointCoord))
         )
    (if (> (EvalEqnOnPoint P NegativePointCoord) 0)
      (mapcar #'- P)
      P
      )
    )
  )


; Funzione per la duplicazione del grafo di incidenza del poliedro.
; Opera una visita in profondita'. Aggiorna l'identificatore Empty,
; che specifica il nodo indicante l'insieme vuoto a dimensione -1 e
; lascia setati i campi induced per la successiva sistemazione delle facet.
; Si occupa anche dell'aggiustamento dei vertici, escluso l'ultimo
; immesso, che dovra' essere posto nel nodo empty del vecchio grafo,
; quindi prima della chiamata a questa funzione.
(defun DuplicateGraph (G Face Dim)
  (let ( (Graph (ConvexHull-Graph G)))    
    (if (not (GetInduced Graph Face))
      (let ( (Ind (NewFace Graph))
             (SubFaces (GraphNode-Sub (gethash Face Graph)) )
             )
        (if (= Dim 0)   ; 0-Faccia : Vertice
          (progn
            (SetVertex Graph Ind (copy-tree (GetVertex Graph Face)))
            (SetVertex Graph Face '(nil nil))
            )
          )
        (if (= Dim -1) (setf (ConvexHull-Empty G) Ind)) ;Ultimo nodo
        (SetInduced Graph Face Ind)
        (mapcar #'(lambda (x) (Creat-Super-Sub-Face-Rel Graph Ind
                                                        (DuplicateGraph G x (1- Dim))
                                                        )
                   )
                SubFaces
                )
        (Creat-Super-Sub-Face-Rel Graph Face Ind)
        Ind
        )
      ;else
      (GetInduced Graph Face)  ; !!!!! DIVERSO DA COME RIPORTATO SUL LIBRO !!!!!
      )                        ; Nel libro Sub-Sup-face sse Induced = nil
    )  
  )

;Test
;(reset-graphnode-counter)
;(setq b (Beneath-Beyond '((v0 (0 0 0)) (v1 (0 1 0)) (v2 (1 0 0)) ( V3 (1 1 0)) ) ))
;(ColorFacet (convexHull-graph b) 'N5 '(1 1 0))
;(printconvexhull b)
;(Coloring b '( V3 (1 1 0)) )   ;Risultato ((N7) (N4 N8) (N2 N3 N5))
;
;(Coloring b '( V3 (0 2 0)) )   ;Risultato ((N7) (N4 N8 N6) (N2 N3 N5))
;(GetVerticesOfFacets b)
;(print b)

;(setq b (Beneath-Beyond '((v0 (0 0 0)) (v1 (1 0 0)) (v2 (0 1 0)) ) ))
;(printconvexhull b)

;(setq b (build-cell '( (V0 (1 1 0)) (V1 (1 3 0)) (V2 (5 1 0)) (V3 (5 3 0))(V5 (5 2 1))) 2))
;(setq b (Beneath-Beyond '( (V0 (1 1 0)) (V1 (1 3 0)) (V2 (5 1 0)) (V5 (5 2 1)) (V3 (5 3 0))) ))


;(setq a (genclist '( (V0 (1 1 0)) (V1 (1 3 0)) (V2 (5 1 0)) (V3 (5 3 0))
; (V4 (1 2 1)) (V5 (5 2 1)))
;'( (0 1 4 5) (0 1 3 2 5)) ) )

;(build-cell '( (V0 (1 1 0)) (V1 (1 3 0)) (V2 (5 1 0)) (V3 (5 3 0))
;(V5 (5 2 1))) 2)
