; -*- package: extrusion -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ; 
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package extrusion                                         ;
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: V. Pascucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Questo modulo estende le funzionalita` del package extrusion al caso di ;
; poliedri definiti su celle convesse a facce.                              ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;(import  'COMMON-LISP-USER::INTERSECTION-OF-EXTRUSIONS )
(provide 'int_extr)
(inpackage extrusion)                   ;package name

(require 'new_extr)
(require 'boole)
(require 'gen_c_ex)
(require 'io_ext2)
(require 'f_ext)
(require 'g_int)

;(unintern 'COMMON-LISP-USER::INTERSECTION-OF-EXTRUSIONS 'COMMON-LISP-USER)
;(shadow 'INTERSECTION-OF-EXTRUSIONS)
(export '(product
	  intersection-of-extrusions))

(use-package 'kernel)

;-------------------------------------------------------------------------------------------
; Product of n polyhedra.
; Representation basen on faces.
; input: - all the parameters (but the first polyhedron) are optional, and have to be polyhedra
;          whoose representation is face based, and the basic elements
;          are convex cells.
;          If if the inpu polyhedra have a representation different
;          to the one required they will bw converted.
; output:- one polyhedron decomposed in convex cells with face based representation.
;-------------------------------------------------------------------------------------------
(defun product (polyhedron &rest other-polyhedra)
  ;; if polyhedron is not represented with faces and convex cells it will be converted
  (setf polyhedron  (coerce-polyhedron
	      (coerce-polyhedron polyhedron 'on-faces) 'convex-cells))
  (dolist (other-polyhedron other-polyhedra polyhedron)
    (setf polyhedron  
	  (pol-on-faces-and-convexes-product-2 
	   polyhedron 
	   ;; if other polyhedron is not represented with faces and convex cells it will be converted
	   (coerce-polyhedron
	      (coerce-polyhedron other-polyhedron 'on-faces) 
	      'convex-cells)))))

;-------------------------------------------------------------------------------------------
; product of two polyhedra with representation basen on faces:
; input: - the two polyhedra must be polyhedra
;          whoose representation is face based, and the basic elements
;          are convex cells.
; output: - one polyhedron decomposed in convex cells with face based representation.
;-------------------------------------------------------------------------------------------
(defun pol-on-faces-and-convexes-product-2 (pol1 pol2)
  (let* ((new-polyhedron (make-empty-polyhedron 'on-faces 'convex-cells))
	 (old-faces1 (get-flist pol1))
	 (old-faces2 (get-flist pol2))
	 (new-faces nil)
	 (new-old-faces-names-alist-1 nil)
	 (new-old-faces-names-alist-2 nil)
	 (old-cells1 (get-cells pol1))
	 (old-cells2 (get-cells pol2))
	 (new-old-cells-names-alist nil)    
	 ;;codimension of each original polyhedron in the product space.
	 (codim1 (1- (length (get-coord (first old-faces2)))))
	 (codim2 (1- (length (get-coord (first old-faces1)))))
	 (zero-covector1 (make-list codim1 :initial-element 0))
	 (zero-covector2 (make-list codim2 :initial-element 0)))
    ;.......................
    ;;resetcounters
    (reset-convex-cell-counter)
    (reset-face-counter)
    ;.......................
    ;;create new faces list and the two association list for the 
    ;;correspondance from old to new names.
    (setf new-faces 
	  (append
	   (mapcar #'(lambda (the-face)
		       (advance-face-counter)
		       (setf new-old-faces-names-alist-1
			     (acons (get-id the-face)       ;old name
				    (current-face-counter)  ;new name
				    new-old-faces-names-alist-1))
		       (make-face (current-face-counter)
				  (append (get-coord the-face) zero-covector1)))
		   old-faces1)
	   (mapcar #'(lambda (the-face)
		       (advance-face-counter)
		       (setf new-old-faces-names-alist-2
			     (acons (get-id the-face)      ;old name
				    (current-face-counter) ;new name
				    new-old-faces-names-alist-2))
		       (make-face (current-face-counter)
				  (append (list (first (get-coord the-face))) 
					  zero-covector2
					  (rest (get-coord the-face)))))
		   old-faces2)))
    (mapcar 'put-face new-faces (never-ending-list new-polyhedron))
    ;(put-faces new-faces new-polyhedron t)
    ;;association list of product cell name and the cons of
    ;;cell factors names
    (maphash 
     #'(lambda (the-cell-1-name dummy)
	 (declare (ignore dummy))
	 (maphash 
	  #'(lambda (the-cell-2-name dummy)
	      (declare (ignore dummy))
	      (setf new-old-cells-names-alist
		    (acons (cons the-cell-1-name the-cell-2-name ) ;cell factors names
			   (advance-convex-cell-counter) ;cell product name
			   new-old-cells-names-alist)))
	  old-cells2))
     old-cells1)
    (maphash 
     #'(lambda (the-cell-1-name the-cell-1)
	 (maphash 
	  #'(lambda (the-cell-2-name the-cell-2)
	      (put-cell 
	       (make-convex-cell  (cdr (assoc  (cons the-cell-1-name the-cell-2-name)
					     new-old-cells-names-alist
					     :test 'equal))
				 '+
				  (append                               ;the new faces list
					  (sublis new-old-faces-names-alist-1  ;is the union of the
						 (if (eq '- (get-cell-sign the-cell-1))
						   (mapcar 'opposit-face (get-cell-faces the-cell-1))
						    (get-cell-faces the-cell-1)))  ;two old faces lists
					  (sublis new-old-faces-names-alist-2  ;after the appropriate
						 (if (eq '- (get-cell-sign the-cell-2))
						   (mapcar 'opposit-face (get-cell-faces the-cell-2))
						   (get-cell-faces the-cell-2)))) ;transformation
				 ;;The vertices list is the proct of the list of
				 ;;vertices of each cell.
				 ;;Each resulting vertex is the union of a vertex
				 ;;of the first cell and another of the second.
				  (let ((new-cell-vertices nil))
				   (mapcar 
				    #'(lambda (cell-1-vertex)
					(mapcar 
					 #'(lambda (cell-2-vertex)
					     (push (append cell-1-vertex cell-2-vertex)
						   new-cell-vertices))
					 (sublis new-old-faces-names-alist-2
						 (get-cell-vertices the-cell-2))))
				    (sublis new-old-faces-names-alist-1
					    (get-cell-vertices the-cell-1)))
				   new-cell-vertices)
				 
				  (append
				  (sublis new-old-cells-names-alist
					  (mapcar
					   #'(lambda(adjacent-cell-name)
					       (if (numberp adjacent-cell-name)
						 adjacent-cell-name
						 (cons adjacent-cell-name the-cell-2-name)))
					   (get-cell-adjacences the-cell-1)) 
					  :test 'equal)
				  (sublis new-old-cells-names-alist
					  (mapcar
					   #'(lambda (adjacent-cell-name)
					       (if (numberp adjacent-cell-name)
						 adjacent-cell-name
						 (cons the-cell-1-name adjacent-cell-name)))
					   (get-cell-adjacences the-cell-2))
					  :test 'equal))
				 nil)
	       new-polyhedron))
	  old-cells2))
     old-cells1)
    ;............................................
    ; Reset all polyhedron attributes
    (progn
      (set-npol new-polyhedron (+ (get-npol pol1)(get-npol pol2)))
      (set-dpol new-polyhedron (+ (get-dpol pol1)(get-dpol pol2)))
      (set-#kd  new-polyhedron)
      (set-#k0  new-polyhedron) 
      (set-#F  new-polyhedron)
      (set-#C  new-polyhedron) 
      new-polyhedron)))

;-------------------------------------------------------------------------------------------
; return a new hashtable of faces reduced by the rule Xi=Xj
; the resulting coincident faces are store only one time
; and the name corrispondence of the discarded faces are
; stored in an association list:  
; input: - faces hashtable;
;        - index of the first coincident axe;
;        - index of the second coincident axe;
; output: - reduced faces hash table;
;         - association list of pairs (old-desappeared-name . new-coicident-face-name).
;-------------------------------------------------------------------------------------------
(defun reduce-faces-hash-table (old-faces-hash-table first-axe-index second-axe-index)
  (let* ((new-faces (new-faces-table))
	 (old-new-discarded-faces-names '()))
    (maphash
     #'(lambda (face-parameters face-identifier)
	 (setf face-parameters (copy-tree face-parameters))
	 (cond 
	  ((zerop (nth second-axe-index face-parameters))
	   (setf face-parameters (but-nth second-axe-index face-parameters)))
	  ((zerop (nth first-axe-index face-parameters))
	   (setf (nth first-axe-index face-parameters)
		 (nth second-axe-index face-parameters))
	   (setf face-parameters (but-nth second-axe-index face-parameters)))
	  (t (error-handler "non null paramerter values correspong to both axes indices" 
			    "reduce-faces-hash-table" )))
	 (if (gethash face-parameters new-faces)
	   (push (cons face-identifier (gethash face-parameters new-faces))
		 old-new-discarded-faces-names)
	   (setf (gethash face-parameters new-faces)
		 face-identifier)))
     old-faces-hash-table)
    (list new-faces old-new-discarded-faces-names)))

;-----------------------------------------------------------
;-----------------------------------------------------------
(defun remove-empty-cells (pol old-removed-faces-alist)
  (let ((cells-hash-table (get-cells pol)))
    (maphash 
     #'(lambda (cell-name cell)
	 (when (empy-cell-p cell pol old-removed-faces-alist)
	   (remhash cell-name cells-hash-table)))
     cells-hash-table)))

;------------------------------------------------------------
; remove cells with 2 coincident faces with opposit sign
; and adjust adjacent cells
;------------------------------------------------------------
(defun remove-degenerations (pol old-removed-faces-alist)
  (let ((cells-hash-table (get-cells pol))
	(cells-to-be-removed nil))
    (maphash 
     #'(lambda (cell-name cell)
	 (let* ((faces (get-cell-faces cell))
		(adjacences (get-cell-adjacences cell))
		(vertices (get-cell-vertices cell)))
	   (dolist (old-new-cell-cons old-removed-faces-alist)
	     ;test for each couple of faces
	     ;that are coincident after
	     ;the identification of the
	     ;coordinates
	     (let ((f1 (find (car old-new-cell-cons) faces :key 'car))
		   (f2 (find (cdr old-new-cell-cons) faces :key 'car)))
	       (when (and f1 f2)
		 (if (eq (cdr f1) (cdr f2))
		   ;; if the faces have the same sign one of them is to be removed
		   (let ((face-position (position f1 faces)))
		     (setf faces (but-nth face-position faces))
		     (setf adjacences (but-nth face-position adjacences))
		     (setf vertices 
			   (mapcar 'set-difference vertices  
				   (never-ending-list (list (car f1))))))
		   ;; if the faces have opposit sign the cell is to be removed
		   ;; and the djacences have to be bypassed
		   (let* ((f1-adjacent-cell 
			   (gethash (nth (position f1 faces) adjacences) 
				    cells-hash-table))
			  (f1-position-in-adjacent-cell 
			   (when f1-adjacent-cell
			     (position (car f1) (get-cell-faces f1-adjacent-cell)
				       :key 'car)))
			  (f2-adjacent-cell
			   (gethash (nth (position f2 faces) adjacences) 
				    cells-hash-table))
			  (f2-position-in-adjacent-cell
			   (when f2-adjacent-cell
			     (position (car f2) (get-cell-faces f2-adjacent-cell)
				       :key 'car))))
		     (push cell-name cells-to-be-removed)
		     ;by-pass the adjacence to the two adjacent cells
		     (when f1-position-in-adjacent-cell
		       (setf (nth f1-position-in-adjacent-cell
				  (get-cell-adjacences f1-adjacent-cell))
			     (if f2-adjacent-cell 
			       (get-cell-identifier f2-adjacent-cell)
			       0)))
		     (when f2-position-in-adjacent-cell
		       (setf (nth f2-position-in-adjacent-cell
				  (get-cell-adjacences f2-adjacent-cell))
			     (if f1-adjacent-cell
			       (get-cell-identifier f1-adjacent-cell)
			       0))))))))
	       (setf faces (sublis old-removed-faces-alist faces)
		     vertices (sublis old-removed-faces-alist vertices)
		     )
	       (setf (gethash cell-name cells-hash-table)
		     (make-convex-cell cell-name
				       (get-cell-sign cell)
				       faces
				       vertices
				       adjacences
				       nil))))
     cells-hash-table)
    (dolist (cell cells-to-be-removed)
      (remhash cell cells-hash-table))))

;--------------------------------------------------------------------
; This function removes the reduntant faces of a convex cell.
; A face of a convex cell is redundant if it (the corresponding
; inequality) can be removed from the system defining the cell
; without modifying the cell. 
; 
;--------------------------------------------------------------------
(defun remove-redundant-faces-from-cell (cell pol)
  (let* ((cell-faces (copy-tree (get-cell-faces cell)))
	 (redundant-faces-indices nil))
    (dotimes (i (length cell-faces))
      (when (equalp 
	     (let* ((objfun (mapcar '- (get-fcoord (car (nth i cell-faces))
					pol
					(cdr (nth i cell-faces)))))
		    (cell-faces (but-nth i cell-faces))
		    (equations 
		     (mapcar 'get-fcoord
			     (mapcar 'car cell-faces)
			     (never-ending-list pol)
			     (mapcar 'rest cell-faces)))
		    (neq    (length equations))
		    (dim    (- (length (car equations)) 1) )
		    (arg    (construct-canonical-tableau neq dim equations)))
	       (resolvesim neq dim objfun (duplicate-arg neq dim arg)))
	     (let* ((equations 
		     (mapcar 'get-fcoord
			     (mapcar 'car cell-faces)
			     (never-ending-list pol)
			     (mapcar 'rest cell-faces)))
		    (neq    (length equations))
		    (dim    (- (length (car equations)) 1) )
		    (objfun (mapcar '- (get-fcoord (car (nth i cell-faces))
					pol
					(cdr (nth i cell-faces)))))
		    (arg    (construct-canonical-tableau neq dim equations)))
	       (resolvesim neq dim objfun (duplicate-arg neq dim arg))))
	(push i redundant-faces-indices)))
    (when redundant-faces-indices 
      (let* ((faces (get-cell-faces cell))
	     (adjacences (get-cell-adjacences cell))
	     (vertices (get-cell-vertices cell)))
	(dotimes (i (length redundant-faces-indices))
	  (let* ((face-position (nth i redundant-faces-indices))
		 (f1 (nth face-position faces)))
	    (setf faces (but-nth face-position faces))
	    (setf adjacences (but-nth face-position adjacences))
	    (setf vertices 
		  (mapcar 'set-difference vertices  
			  (never-ending-list (list (car f1)))))))
	(setf (gethash (get-cell-identifier cell) (get-cells pol))
	      (make-convex-cell (get-cell-identifier cell)
				(get-cell-sign cell)
				faces
				vertices
				adjacences
				nil))))))

; Copia della funzione precedente
; Si modifica il test di ridondanza delle facce  della cella

; Si calcolano i valori numerici dei vertici per verificare poi
; la loro posizione rispetto all'iperpiano Xi=Xj.

(defun remove-redundant-faces-from-cell (cell pol)
  (let* ((cell-faces (copy-tree (get-cell-faces cell)))
	 (numeric-faces  (mapcar 'get-fcoord (mapcar 'first cell-faces)
				       (never-ending-list pol)))
	 (redundant-faces-indices nil)
	 (symbolic-vertices (get-cell-vertices cell))
	 (numeric-vertices (mapcar 'calc-intersection-point
				   (mapcar #'(lambda (symbolic-vertex)
				       (mapcar 'get-fcoord symbolic-vertex
					   (never-ending-list pol))) symbolic-vertices)))
	 (positions (pprint (mapcar 
		     #'(lambda (face-coords)
				    (mapcar #'(lambda (vertex-coords)
				     (apply '+ (mapcar '* face-coords (cons 1.0 vertex-coords))))
				 numeric-vertices)) numeric-faces))))
    (io::print-convex-cell cell)
    (pprint symbolic-vertices)
    (pprint numeric-vertices)
    (dotimes (i (length cell-faces))
      (when (equalp 
	     (let* ((objfun (mapcar '- (get-fcoord (car (nth i cell-faces))
					pol
					(cdr (nth i cell-faces)))))
		    (cell-faces (but-nth i cell-faces))
		    (equations 
		     (mapcar 'get-fcoord
			     (mapcar 'car cell-faces)
			     (never-ending-list pol)
			     (mapcar 'rest cell-faces)))
		    (neq    (length equations))
		    (dim    (- (length (car equations)) 1) )
		    (arg    (construct-canonical-tableau neq dim equations)))
	       (resolvesim neq dim objfun (duplicate-arg neq dim arg)))
	     (let* ((equations 
		     (mapcar 'get-fcoord
			     (mapcar 'car cell-faces)
			     (never-ending-list pol)
			     (mapcar 'rest cell-faces)))
		    (neq    (length equations))
		    (dim    (- (length (car equations)) 1) )
		    (objfun (mapcar '- (get-fcoord (car (nth i cell-faces))
					pol
					(cdr (nth i cell-faces)))))
		    (arg    (construct-canonical-tableau neq dim equations)))
	       (resolvesim neq dim objfun (duplicate-arg neq dim arg))))
	(push i redundant-faces-indices)))
    (when redundant-faces-indices 
      (let* ((faces (get-cell-faces cell))
	     (adjacences (get-cell-adjacences cell))
	     (vertices (get-cell-vertices cell)))
	(dotimes (i (length redundant-faces-indices))
	  (let* ((face-position (nth i redundant-faces-indices))
		 (f1 (nth face-position faces)))
	    (setf faces (but-nth face-position faces))
	    (setf adjacences (but-nth face-position adjacences))
	    (setf vertices 
		  (mapcar 'set-difference vertices  
			  (never-ending-list (list (car f1)))))))
	(setf (gethash (get-cell-identifier cell) (get-cells pol))
	      (make-convex-cell (get-cell-identifier cell)
				(get-cell-sign cell)
				faces
				vertices
				adjacences
				nil))))))

;------------------------------------------------------------
; 
;------------------------------------------------------------
(defun rebuild-vertices (cell pol)
  (let* ((faces (mapcar 'first (get-cell-faces cell)))
	 (old-vertices (get-cell-vertices cell))
	 (new-vertices nil)
	 (dim (1- (length (get-fcoord (car faces) pol)))))
    (dolist (first-old-vertex old-vertices)
      (dolist (second-old-vertex old-vertices)
	(when (not (eq first-old-vertex second-old-vertex))
	  (let ((resulting-vertex 
		 (intersection  first-old-vertex second-old-vertex)))
	    (when (and resulting-vertex
		       ;*********
		       (not (member resulting-vertex
				    new-vertices
				    :test 'subsetp))
		       ;*******
		       (= dim
			  (new-pl:calc-matrix-rank 
			   0 (1- (length resulting-vertex))
			   1 dim
			   (mapcar 'get-fcoord
					  resulting-vertex
					  (never-ending-list pol)))))
	      (pushnew resulting-vertex new-vertices :test 'equalp))))))
    (setf (gethash (get-cell-identifier cell) (get-cells pol))
	  (make-convex-cell (get-cell-identifier cell)
			    (get-cell-sign cell)
			    (get-cell-faces cell)
			    new-vertices
			    (get-cell-adjacences cell)
			    nil))))

;------------------------------------------------------------
; Si ricostruiscono i vertici rappresentati implicitamente
; come lista di facce (tutte e sole quelle facce a cui 
; appartiene).
; Presa una coppia di vertici tali che:
;    - siano situati in posizione opposta rispeto 
;      all'iperpiano Xi=Xj;
;    - la loro intersezione (come insiemi di simbili) sia
;      un sottoinsieme delle facce non-ridondanti;
;    - la loro intersezione (come insiemi di simbili) non sia
;      un sottoinsieme di un'altro vertice calcolato nello 
;      stesso modo;
;  allora la loro intersezione (come insiemi di simbili) e`
;  un vertice della nuova cella.
;------------------------------------------------------------
(defun rebuild-vertices (cell pol 
			      vertices_in_+_halfspase 
			      vertices_in_=_halfspase 
			      vertices_in_-_halfspase
			      rename-faces-pairs)
  ;(print 'RICOSTRUZIONE_VERTICI_INIZIO)
  ;(io::print-convex-cell cell)
  ;(pprint vertices_in_+_halfspase)
  ;(pprint vertices_in_=_halfspase)
  ;(pprint vertices_in_-_halfspase)
  ;(pprint 'inizio-ricostruzione-vertici******************************)
  (let* ((faces (mapcar 'first (get-cell-faces cell)))
	 (new-vertices nil))
    ;scansione dei vertici che giacciono sull'iperpiano
    ;Xi=Xj.
    (dolist (new-vertex vertices_in_=_halfspase)
      (push (intersection faces new-vertex ) new-vertices))
    ;(print '---1)
    ;(pprint new-vertices)
    ;doppia scansione per visitare i vertici
    ;che si trovano dai due lati opposti dell'iperpiano Xi=Xj
    (dolist (first-old-vertex vertices_in_-_halfspase)
      (dolist (second-old-vertex vertices_in_+_halfspase)
	(let* ((resulting-vertex 
		(intersection faces 
			     (intersection  first-old-vertex 
					    second-old-vertex))))
	  ;(print 'new_vertex)
	  ;(print resulting-vertex)
	  (setf new-vertices
		(remove-if #'(lambda (x)
			     (subsetp x resulting-vertex))  new-vertices  ))
	  
	  (pushnew resulting-vertex new-vertices :test 'subsetp )
)))
	;(print '---1)
	;(pprint new-vertices)

    (setf (gethash (get-cell-identifier cell) (get-cells pol))
	  (make-convex-cell (get-cell-identifier cell)
			    (get-cell-sign cell)
			    (sublis rename-faces-pairs (get-cell-faces cell))
			    (sublis rename-faces-pairs new-vertices)
			    (get-cell-adjacences cell)
			    nil))))




;------------------------------------------------------------
;------------------------------------------------------------
(defun remove-empty-cell-references-from-adjacent-cells (pol)
  (let ((cells (get-cells pol)))
    (maphash 
     #'(lambda (cell-name cell)
	 (let ((adjacences (get-cell-adjacences)))
	   (dolist (adjacent-cell adjacences)
	     (when (not (gethash adjacent-cell cells)) ; the adjacent cell does not exis
	       )
     cells))))))

;-------------------
; Returns t if the cell "cell" of the polyhedron "pol" is empty
;-------------------
(defun empy-cell-p (cell pol old-removed-faces-alist)
  ;(setf (kernel::convex-cell-faces cell) (sublis old-removed-faces-alist (get-cell-faces cell)))
  (let* ((cell-faces  (sublis old-removed-faces-alist (get-cell-faces cell)))
	 (equations 
	  (mapcar 'get-fcoord
		  (mapcar 'car cell-faces)
		  (never-ending-list pol)
		  (mapcar 'rest cell-faces)))
	 (neq    (length equations))
	 (dim    (- (length (car equations)) 1) )
	 (arg (store neq dim equations) ))
    (not (pl::build-canonical-form arg neq (+ dim neq 1)))))

;-------------------
(defun identify-coords (pol i j)
  (let* (;(pol (product pol1 pol2))
	 (list-of-new-faces-and-names-conversion-alist
	  (reduce-faces-hash-table (get-faces pol) i j))
	 (new-faces (first list-of-new-faces-and-names-conversion-alist))
	 (names-conversion-alist
	  (second list-of-new-faces-and-names-conversion-alist)))
    ;(io::printpol  pol)
    ;(pprint names-conversion-alist)
    (put-faces  new-faces pol t)
   
    ;(io::printpol pol)
    (remove-empty-cells  pol names-conversion-alist)

    ;(io::printpol pol)
    (remove-degenerations pol names-conversion-alist)

    ;(io::printpol pol)
    (maphash #'(lambda (c cell)
		 (remove-redundant-faces-from-cell 
		  cell pol))
	     (get-cells pol))
    (maphash #'(lambda (c cell)
		 (rebuild-vertices 
		  cell pol))
	     (get-cells pol))
    pol))


(defun identify-coords (pol i j)
  (when  pol
  (let* ((polyhedron-cells (get-clist pol))
	 remove-cells
	 ;nuovo database delle facce
	 (new-faces (new-faces-table))
	 ;::::::::::::::::::::::::::::::::::::::::::::::::::::::
	 ; Si calcolano le coppie di facce che siano coincidenti 
	 ; sull'iperpiano Xi=Xj.
	 (coencident-faces-pairs (coincident-faces-pairs (get-faces pol) i j new-faces))
	 )
    ;(pprint '+++++++++++++++++++++++++++++++++++++++)
    ;(pprint coencident-faces-pairs)
    ;(pprint '+++++++++++++++++++++++++++++++++++++++)
    (dolist (cell polyhedron-cells)
      ;(print '---------------------------)
      ;(print (get-cell-identifier cell))
      ;(print '---------------------------)
      (let* ((cell-faces (copy-tree (get-cell-faces cell)))
	     (cell-adjacences (get-cell-adjacences cell))
	     (symbolic-faces  (mapcar 'first cell-faces))
	     ; In the faces symbols substitution you obviously
	     ; have to take into account only the faces bilonging to 
	     ; the current face
	     (rename-faces-pairs 
	      (remove-if #'(lambda (faces-pair)
			     (subsetp   (list (car faces-pair)(cdr faces-pair)) symbolic-faces)) 
			 coencident-faces-pairs))
	     (coencident-faces-pairs 
	      (remove-if-not #'(lambda (faces-pair)
				 (subsetp   (list (car faces-pair)(cdr faces-pair)) symbolic-faces)) 
			     coencident-faces-pairs))
	     
	     
	     (redundant-faces-indices nil)
	     (redundant-faces nil)
	     (vertices_in_+_halfspase nil)
	     (vertices_in_=_halfspase nil)
	     (vertices_in_-_halfspase nil)
	     (symbolic-vertices (get-cell-vertices cell))
	     (numeric-vertices 
	      (mapcar 'calc-intersection-point
		      (mapcar #'(lambda (symbolic-vertex)
				  (mapcar 'get-fcoord symbolic-vertex
					  (never-ending-list pol))) symbolic-vertices)))
	     (positions  (mapcar #'(lambda (vertex-coords)
				    (let ((position
					   (- (nth (1- i) vertex-coords)
					      (nth (1- j) vertex-coords))))
				      (cond ((< (abs position) (* 1  **epsilon)) '=)
					    ((plusp position)
					     '+)
					    ((minusp position) '-)
					    (t '=))))
				numeric-vertices)))
    ;(pprint '+++++++++++++++++++++++++++++++++++++++)
    ;(pprint coencident-faces-pairs)
    ;(pprint rename-faces-pairs)
    ;(pprint '+++++++++++++++++++++++++++++++++++++++)

	(if (and (member '+ positions)
		 (member '- positions))
	  (progn 
	    (dotimes (i (length positions))
	      (if (symbol-equal (nth i positions) '-)
		(push (nth i symbolic-vertices)  vertices_in_-_halfspase)
		(if (symbol-equal (nth i positions) '=)
		  (push (nth i symbolic-vertices)  vertices_in_=_halfspase)
		  (push (nth i symbolic-vertices)  vertices_in_+_halfspase))))
	    ;:::::::::::::::::::::::::::::::::::::::::::::::::::::
	    ; Si eliminano le facce ridondanti secondo la regola
	    ; che considera ridondanti tutte le facce che 
	    ; che abbiano tutti i vertici dallo stesso lato
	    ; di tale faccia (il prodotto scalare della faccia
	    ; per i vertici sia sempre >=0 o sempre <=0
	    (dolist (face-sign cell-faces)
	      (let  ((face (car face-sign))
		     (signs nil))
		(dotimes (i (length numeric-vertices))
		  (when (member face (nth i symbolic-vertices))
		    (push  (nth i positions) signs)
		    ))
		;(print face)
		;(print signs)
		(when (or (every #'(lambda (x)(and (symbolp x)
						   (or (symbol-equal x '+)
						       (symbol-equal x '=)))) signs)
			  (every #'(lambda (x)(and (symbolp x)
						   (or (symbol-equal x '-)
						       (symbol-equal x '=)))) signs))
		  (push (position face-sign cell-faces) redundant-faces-indices)
		  (push face redundant-faces)
		  )
		;(print face) 
		;(print signs)
		signs
		))

	    ;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
	    ; Si deve reinserire una faccia di ogni coppia di facce la cui interseziene
	    ; appartenga al sottospazio Xi=Xj.
	    ;:::::::DA FARE ::::::::
	    ; Le due facce devono avere uguale segno, altrimenti la cella e` vuota.
	    ; Se le due facce si chiamano Fa e Fb e si decide (arbitrariamente)
	    ; di reinserire Fa che risuta adiacete alla cella Ca, allora
	    ; la cella adiacente su Fa deve divetare la cella adiacete a Ca su Fb. 
	    ;(print 'SISTEMAZIONE_COPPIA_DI_FACCE_COINCIDENTI)
	    (dolist (faces-pair coencident-faces-pairs)
	      ;(print 'faces-pair)
	      ;(print faces-pair)
	      (when (and (subsetp (list (car faces-pair)(cdr faces-pair))
					 symbolic-faces)
			 (or  (cdr faces-pair)
			     t)
			 (not (equal  (position (cdr faces-pair) cell-faces :key 'car)
				      (position (first faces-pair) cell-faces :key 'car))))
		;(pprint 'TEST_SUPERATO)
		 ; La seconda faccia della coppia e` quella che deve
		 ; rimanere nella cella
		(setf 
		 redundant-faces-indices 
		 (but-nth (position (cdr faces-pair) redundant-faces)
			  redundant-faces-indices)
		 redundant-faces 
		 (but-nth (position (cdr faces-pair) redundant-faces)
			  redundant-faces))
		(if (and (not (numberp (nth (position (cdr faces-pair) cell-faces :key 'car)
					    cell-adjacences)))
			 (not (numberp (nth (position (first faces-pair) cell-faces :key 'car)
					    cell-adjacences))))
		  (let* ((adjacent-cell (kernel::get-polyhedron-cell
					 pol (nth (position (cdr faces-pair) cell-faces :key 'car)
						  cell-adjacences)))
			 (other-cell-faces (get-cell-faces adjacent-cell))
			 (other-cell-adjacences (get-cell-adjacences adjacent-cell))
			 )
		    (setf (nth (position (cdr faces-pair) cell-faces :key 'car)
			       cell-adjacences)
			  (nth (position (first faces-pair) other-cell-faces :key 'car)
			       other-cell-adjacences)))
		  (setf (nth (position (cdr faces-pair) cell-faces :key 'car)
			       cell-adjacences) 0))))
	    ;:::::::::::::::::::::::::::::::::::::::::::::
	    ; essendo gia` note le facce che coincidono a
	    ; coppie, si puo` sostituire la linea di comando
	    ; successiva con la chiamata che sfrutti tali
	    ; proprieta`.
	    ;(reduce-faces-hash-table (get-faces pol) i j)
	    
	    (when t ;redundant-faces-indices 
	      (let* ((faces cell-faces);(faces (get-cell-faces cell))
		     (adjacences cell-adjacences);(adjacences (get-cell-adjacences cell))
		     (vertices (get-cell-vertices cell)))
		;(pprint 'RIMOZIONE_FACCE_RIDONDANTI)
		;(pprint (list faces adjacences))
		;;_______________________________________________________
		;; rimozione dalla cella dei simboli di facce ridondanti
		(dotimes (i (length redundant-faces-indices))
		  (let* ((face-position (nth i redundant-faces-indices)))
		    (setf faces (but-nth face-position faces))
		    (setf adjacences (but-nth face-position adjacences))
		    ))
		;(pprint (list faces adjacences))
		(setf cell (make-convex-cell (get-cell-identifier cell)
					     (get-cell-sign cell)
					     (sublis coencident-faces-pairs faces)
					     (sublis coencident-faces-pairs vertices)
					     adjacences
					     nil)
		      
		      (gethash (get-cell-identifier cell) (get-cells pol)) cell)
		(rebuild-vertices cell pol
				  ;vertices_in_+_halfspase
				  ;vertices_in_=_halfspase
				  ;vertices_in_-_halfspase          
				  ;(sublis coencident-faces-pairs vertices_in_+_halfspase)
				  vertices_in_+_halfspase
				  (sublis coencident-faces-pairs vertices_in_=_halfspase)
				  ;vertices_in_=_halfspase
				  ;(sublis coencident-faces-pairs vertices_in_-_halfspase)  
				  vertices_in_-_halfspase          
				  rename-faces-pairs))))
	  (push (get-cell-identifier cell) remove-cells))))
    (if remove-cells 
      (dolist (cell remove-cells)
	(remhash cell (get-cells pol))))
    (put-faces  new-faces pol t)
    (set-#F  pol)
    (set-#C  pol) 
    (if (zerop (get-#C pol))
      nil 
      pol))))
;-----------------------------------------------------------------------
; Individua l'insieme di coppie di facce che risultano coincidenti
; a seguito dell'applicazione del'equazione Xi=Xj.
; Gli indici i e j che individuano l'iperpiano Xi=Xj, oltre 
; alla tavola hash delle facce del poliedro stesso,
; vengono forniti come parametri di ingresso. 
;------------------------------------------------------------------------
;; N.B.
;; Il test sarebbe piu` efficiente se fosse anticipato 
;; prima del prodotto fra poliedri.
;; In particolare si dovrebbero scandire solo le facce 
;; del primo poliedro argomento e confrontarle con le facce del secondo.
;; Verificare se e quando e` vero. 
;;-----------------------------------------------------------------------
(defun coincident-faces-pairs
       (faces-hash-table i j new-faces)
  (let* ((coincident-faces-pairs nil)
	 ;(new-faces (new-faces-table))
	 (coincident-faces-pairs1 nil)
	 (faces-list nil))
    
    (maphash #'(lambda (face-parameters face-identifier)
		 (let ((new-parameters (but-nth j face-parameters)))
		   (if (zerop (nth i face-parameters))
		     (setf (nth i new-parameters) (nth j face-parameters)))
		   (if 
		     (gethash new-parameters new-faces)
		     ;(find-same-plane new-parameters faces-list)
		     ;test
		     (push (cons face-identifier (gethash new-parameters new-faces))
			   coincident-faces-pairs1)
		     (progn 
		       (setf (gethash new-parameters new-faces) face-identifier)
		       (push (list face-identifier new-parameters) 
			     faces-list))))
		 
		 (when (zerop (nth i face-parameters))
		   (let ((new-par (copy-list face-parameters))
			 other-face)
		     (setf (nth i new-par) (nth j new-par)
			   (nth j new-par) 0
			   other-face (gethash new-par faces-hash-table))
		     (if (and other-face
			      (not (symbol-equal face-identifier
						 other-face)))
		       (push (cons face-identifier other-face)
			     coincident-faces-pairs)))))
	     faces-hash-table)
     coincident-faces-pairs1
    ))



(defun intersection-of-extrusions (pol1 pol2 coincident-coord-list )
  (let ((pol (product pol1 pol2))
	(n1 (get-npol pol1)))
    (dotimes (i (length coincident-coord-list))
      ; adjust rest of indices
      (dotimes (j (- (length coincident-coord-list) i 1))
	;(when (< (first (nth i coincident-coord-list))
	;         (first (nth (+ 1 i j) coincident-coord-list)))
	;  (decf (first (nth (+ 1 i j) coincident-coord-list))))
	(when (< (second (nth i coincident-coord-list))
		 (second (nth (+ 1 i j) coincident-coord-list)))
	  (decf (second (nth (+ 1 i j) coincident-coord-list)))))
      (setf pol (identify-coords pol 
				 (first (nth i coincident-coord-list))
				 (+ n1 (second (nth i coincident-coord-list))))))
    
    (progn
      (set-npol pol (+ (get-npol pol1)(get-npol pol2) 
				  (- (length coincident-coord-list))))
      (set-dpol pol (+ (get-dpol pol1)(get-dpol pol2)
				  (- (length coincident-coord-list))))
      (set-#kd  pol)
      (set-#k0  pol) 
      (set-#F  pol)
      (set-#C  pol) 
      pol)))




(defun intersection-of-extrusions (pol1 pol2 coincident-coord-list )
  (let ((new-par-mode coincident-coord-list)
	old-par-mode
	coincident-coord-list)
    (setf old-par-mode 
	  (mapcar 'list (first new-par-mode)(second new-par-mode)))
    (setf coincident-coord-list 
	  (remove-if #'(lambda (coord-pair) (some 'zerop coord-pair))
		    old-par-mode))
  (let ((pol (product pol1 pol2))
	(n1 (get-npol pol1)))
    (dotimes (i (length coincident-coord-list))
      ; adjust rest of indices
      (dotimes (j (- (length coincident-coord-list) i 1))
	;(when (< (first (nth i coincident-coord-list))
	;         (first (nth (+ 1 i j) coincident-coord-list)))
	;  (decf (first (nth (+ 1 i j) coincident-coord-list))))
	(when (< (second (nth i coincident-coord-list))
		 (second (nth (+ 1 i j) coincident-coord-list)))
	  (decf (second (nth (+ 1 i j) coincident-coord-list)))))
      (setf pol (identify-coords pol 
				 (first (nth i coincident-coord-list))
				 (+ n1 (second (nth i coincident-coord-list))))))
    
    (if pol (progn
      (set-npol pol (+ (get-npol pol1)(get-npol pol2) 
				  (- (length coincident-coord-list))))
      (set-dpol pol (+ (get-dpol pol1)(get-dpol pol2)
				  (- (length coincident-coord-list))))
      (set-#kd  pol)
      (set-#k0  pol) 
      (set-#F  pol)
      (set-#C  pol) 
      pol)))))


















;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Funzioni per calcolare il box di contenimento di un convesso
;;;
;;;  scritte da G. Proietti   16/1/1993
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  CONVEX-BOX
;;;
;;;   funzione che ritorna, dato un poliedro espresso tramite intersezione
;;;   di sottospazi, il box di contenimento dello stesso
;;;
;;;  INPUT
;;;
;;;   equations := lista di liste contenente i coefficienti delle disequazioni
;;;                nel formato ( (b1 a11 a12 ... a1n) (b2 a21 a22 ... a2n).. )
;;;
;;;  OUTPUT
;;;
;;;   lista contenente gli estremi superiore ed inferiore del box di conteni-
;;;   -mento del poliedro nel formato ( (cmax1 ... cmaxn) (cmin1 ... cminn) )

(defun convex-box (equations)
  (let* ((neq    (length equations) )
	 (dim    (- (length (car equations)) 1) )
	 (objfun (make-list (+ dim 1) :initial-element 0))
	 (plus   (make-list dim ))
	 (minus  (make-list dim ))
	 (arg    (construct-canonical-tableau neq dim equations)))
       (dotimes (i dim (list plus minus) )
	 (progn
	   (setf (nth (+ i 1) objfun) 1)
	   (setf (nth i minus) (nth i (resolvesim neq dim objfun (duplicate-arg neq dim arg))) )
	   (setf (nth (+ i 1) objfun) -1)
	   (setf (nth i plus) (nth i (resolvesim neq dim objfun (duplicate-arg neq dim arg))) )
	   (setf (nth (+ i 1) objfun) 0)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; ritorna una lista contenente un tableau in forma canonica per il
;;; con la riga dei coefficienti della funzione obiettivo vuota e
;;; pronta per la memorizzazione dei suddetti coefficienti,il numero di
;;; righe, di colonne ed un vettore specificante le variabili di base.
;;;
;;;
;;; INPUT :
;;;
;;;         neq           :=  numero delle disuguaglianze ( <= ) costituenti i vincoli
;;;         dim           :=  dimensione dello spazio ( numero delle variabili )
;;;         objfun        :=  lista dei coefficienti della funzione obiettivo
;;;                           ( t.noto coeff1 .... coeffn )
;;; OUTPUT
;;;
;;;  ( #2A( .. tableau ..) nrighe ncol #( .. base .. ) )

(defun construct-canonical-tableau (neq dim equations)
  (let ( (arg (store neq dim equations) ) )
    (progn
      (setq arg  (multiple-value-list (pl::build-canonical-form arg neq (+ dim neq 1))))
      (setf (first arg) (copy-arry neq dim (first arg)) )
      (setf (second arg) (+ 1 (second arg)) )
      (setf (fourth arg) (let ( (vec (make-array (+ neq 1) ) ) )
			   (dotimes (i neq vec)
			     (setf (aref vec (+ i 1)) (nth i (fourth arg))))))
      (return-from construct-canonical-tableau arg))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; ritorna un duplicato della lista contenente il tableau con tableau e
;;; variabili di base nuovamente memorizzate per salvare gli originali
;;; dai side-effects del calcolo del simplesso

(defun duplicate-arg (neq dim arg)
  (let ( (newarg1 (make-array (list (+ neq 1) (+ dim neq 2)) :initial-element 0))
	 (newarg4 (make-array (+ neq 1)))
	 (arg1 (first arg))
	 (arg4 (fourth arg)))
    (list
      (dotimes (i (+ neq 1) newarg1)
	(dotimes (j (+ dim neq 2))
	  (setf (aref newarg1 i j) (aref arg1 i j))
	)
      )
      (second arg)
      (third  arg)
      (dotimes (i neq newarg4)
	(setf (aref newarg4 (+ i 1)) (aref arg4 (+ i 1)))
      )
    )
  )
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  ritorna il punto soluzione del problema specificato, sotto forma di
;;;  lista di razionali

(defun resolvesim (neq dim objfun arg )
    (progn
      (setf (first arg) (store-objective-fun dim (first arg) objfun))
      (report-canonical  neq dim (first arg) (fourth arg))
      (if (pl::simplex-algorithm (first arg) (second arg) (third arg) (fourth arg))
	(let* ( (result (make-list dim :initial-element 0) )
		(pos (search (list (+ dim 1)) (fourth arg) ) )
		(w   (if pos (aref (first arg) pos 0) 0) ) )
	  (progn
	    (do* ( (i 1 (+ i 1) ) )
		 ( (> i neq) )
	      (let ( (n (aref (fourth arg) i) ) )
		(progn
		  (if (<= n dim )
		    (setf (nth (- n 1) result )  (aref (first arg) i 0) )
		  )
		)
	      )
	    )
	    (dotimes (j dim)
	      (setf (nth j result) (- (nth j result) w)))
	    (return-from resolvesim result)
	  )
	)
	(return-from resolvesim nil)
      )
    )
)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; memorizza i coefficienti della funzione obiettivo, trasformati in rational,
;;; nella prima riga del tableau fornito

(defun store-objective-fun (dim mat objfun)
  (progn
    (dotimes (i (+ 1 dim) mat)
      (setf (aref mat 0 i) (rationalize (nth i objfun)))
    )
    (add-w 0 dim mat)
    (return-from store-objective-fun mat)
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Crea un array dove memorizzare i coefficienti dei vincoli del problema
;;; per il successivo utilizzo della funzione build-canonical-form
;;; Da notare che inverte il segno del termine noto

(defun store  (neq dim equaz)
  (let ((mat (make-array (list neq (+ dim neq 2)) :initial-element 0)))
    (dotimes (i neq mat)
      (let ((a (nth i equaz)))
	(progn
	  (setf (nth 0 a) (- (nth 0 a)) )
	  (if (>= (car a) 0)
	    (progn
	      (dotimes (j (+ dim 1) )
		(setf (aref mat i j) (rationalize (nth j a)))
	      )
	      (add-w i dim mat)
	      (setf (aref mat i (+ dim i 2)) 1)
	    )
	    (progn
	      (dotimes (j (+ dim 1) )
		(setf (aref mat i j) (rationalize (- (nth j a))))
	      )
	      (add-w i dim mat)
	      (setf (aref mat i (+ dim i 2)) -1)
	    )
	  )
	)
      )
    )
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Aggiunge il coefficiente della variabile necessaria a gestire
;;; le variabili non vincolate in segno nella posizione dim+1
;;;
;;; INPUT : row  riga sulla quale calcolare ed aggiungere il coefficiente
;;;         dim  dimensione dello spazio
;;;         mat  matrice

(defun add-w (row dim mat)
  (let ((w 0))
    (progn
      (dotimes (j dim)
	(setq w (+ w (aref mat row (+ j 1)))))
      (setf (aref mat row (+ dim 1)) (- w) )
    )
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Copia la matrice in una nuova con una riga in piu ( corrispondente
;;; alla riga 0 ) dove poter memorizzare i coefficienti della f. obiettivo

(defun copy-arry (neq dim mat)
  (let ((newmat (make-array (list (+ neq 1) (+ dim neq 2)) :initial-element 0)))
    (dotimes (i neq newmat)
      (dotimes (j (+ dim neq 2) newmat)
	(setf (aref newmat (+ i 1) j) (aref mat i j))
      )
    )
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Riporta il tableau in forma canonica a partire dal tableau a cui
;;; siano aggiunti i coefficienti della f. obiettivo, eseguendo una serie
;;; di operazioni di pivot in corrispondenza delle variabili di base
;;; che corrispondano a coefficienti di costo non nulli

(defun report-canonical (neq dim mat base)
  (dotimes (i neq mat)
    (if (/= 0 (aref mat 0 (aref base (+ i 1)) ) )
      (pl::pivot mat (+ neq 1) (+ dim neq 2) (+ i 1) (aref base (+ i 1)))
    )
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Converte una lista generalizzata di rational nella corrispondente
;;; lista di float

(defun ratio2float (lista)
  (let ( (a (car lista)) )
    (cond ( (null a) nil)
	  ( (atom a) (cons (coerce a 'float) (ratio2float (cdr lista))))
	  (    t     (cons (ratio2float a) (ratio2float (cdr lista))))
    )
  )
)
