; -*- package: io -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ; 
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package Input&Output                                      ;  
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: V. Pascucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;                                                                           ;
; Modificato per ottenere la stampa anche delle adiacenze                   ;
; esterne e dei simplessi cubici di una cella.                              ;
;                                                                           ;
;                                             by  Marco Conti               ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(provide 'io_ext)

(defpackage io)
(inpackage io)

          

(export '(print-pol-node
	  print-pol-complex
	  read-pol-complex
	  save-pol-complex))

(require 'conv_ext)
(require 'lisp_ext)
(require 'dvrlib)
(require 'io)

(use-package 'lisp-extension)
(use-package 'dvrlib)
(use-package 'kernel)


;-----------------------------------------------------------------------------
; Test se una lista e' tutta nil e quindi non va stampata
;-----------------------------------------------------------------------------
(defun list-null? (lista
		   &aux (YorN T))
  (if (not (null lista))
    (dolist (x lista)
      (if (not (equal x nil))
	(setq YorN NIL))))
  YorN)

;-----------------------------------------------------------------------------
; Stampa di una faccia nel fomato adatto per la printpol
;-----------------------------------------------------------------------------
(defun print-face (face &optional (level 0))
  (print-newline-and-white-chars level)
  (princ (get-id face))
  (princ " : ")
  (princ (get-coord face))
  (values))

;-----------------------------------------------------------------------------
; Stampa di una lista di facce nel fomato adatto per la printpol
;-----------------------------------------------------------------------------
(defun printflist (flist &optional (level 0))
  (mapcar 'print-face flist (never-ending-list level)))


;-----------------------------------------------------------------------------
; Estensione della printpol al caso di poliedri composti anche
; di celle convesse e con rappresentazione basata sulle facce
;-----------------------------------------------------------------------------
(defun printpol (pol &optional (printalot t)(level 0))
  (if (polyhedron? pol)
    (let ((representation (get-pol-rep pol))
	  (kind-of-bricks (get-pol-brick-type pol)))
      (print-newline-and-white-chars level)
      (princ "Intrinsic dimension : ") (princ (get-dpol pol)) (print-newline-and-white-chars level)
      (princ "E-space dimension : ") (princ (get-npol pol)) (print-newline-and-white-chars level)
      (princ "Sign : ") (princ (get-pol-sign pol)) (print-newline-and-white-chars level)
      (cond
       ((lisp-extension:symbol-equal representation 'on-vertices)
	(princ "Vertices number : ") (princ (get-#K0 pol)) (print-newline-and-white-chars level)
	(cond
	 ((lisp-extension:symbol-equal kind-of-bricks 'simplices)
	  (princ "Simplices number : ")  (princ (get-#kd pol)) (print-newline-and-white-chars level))
	 ((lisp-extension:symbol-equal kind-of-bricks 'convex-cells)
	  (princ "Convex cells number : ")  (princ (get-#C pol)) (print-newline-and-white-chars level))))
       ((lisp-extension:symbol-equal representation 'on-faces)
	(princ "Faces number : ")  (princ (get-#F pol)) (print-newline-and-white-chars level)
	(cond
	 ((lisp-extension:symbol-equal kind-of-bricks 'simplices)
	  (princ "Simplices number : ")  (princ (get-#kd pol)) (print-newline-and-white-chars level))
	 ((lisp-extension:symbol-equal kind-of-bricks 'convex-cells)
	  (princ "Convex cells number : ")  (princ (get-#C pol)) (print-newline-and-white-chars level)))))
      (if printalot (progn
		      (printvlist (get-vlist pol))
		      (printflist (get-flist pol t) level)
		      (printslist (get-slist pol))
		      (printclist (get-clist pol t) level)
		      (print-newline-and-white-chars level)
		      (values))))
    (error-handler 1 "printpol")))

;--------------------------------------------------------------------------------
; Stampa di una lista di celle convesse nel fomato adatto al printpol
;--------------------------------------------------------------------------------
(defun printclist (clist &optional (level 0))
  (mapcar 'print-convex-cell clist (never-ending-list t)(never-ending-list level)))

;-----------------------------------------------------------------------------
; Stampa di una cella convessa nel fomato adatto al printpol
;-----------------------------------------------------------------------------
(defun print-convex-cell (cell &optional (one-row t) (level 0))
  (if one-row
    (progn
      (print-newline-and-white-chars level)
      (princ (get-cell-identifier cell))(princ " : ")
      (princ (get-cell-sign cell))
      (if (not (list-null? (get-cell-faces cell)))
	(princ (get-cell-faces cell)))
      (princ " ")
      (if (not (list-null? (get-cell-vertices cell)))
	(princ (get-cell-vertices cell)))
      (print-newline-and-white-chars (+ level 6))
      (princ (get-cell-adjacences cell))
      (princ "  ")
      (princ (show-external-adjacences (get-cell-external-adjacences cell)))
      (if (not (list-null? (get-cell-internal-adjacences cell)))
	(princ (get-cell-internal-adjacences cell)))
      (if (not (list-null? (kernel::get-cell-open-faces cell)))
	(princ (kernel::get-cell-open-faces cell)))
      (if (not (list-null? (kernel::get-cell-centroid cell)))
	(progn
          (print-newline-and-white-chars level)
          (princ "F(centroide)  ")
          (princ (kernel::get-cell-centroid cell))))
      (if (not (equal nil (kernel::get-cell-cubic-simplexes cell)))
	(print-cubic-simplexes (kernel::get-cell-cubic-simplexes cell) level)))
    (progn 
      (print-newline-and-white-chars level)
      (princ "Cell name ..........: ")(princ (get-cell-identifier cell))(print-newline-and-white-chars level)
      (princ "Sign ...............: ")(princ (get-cell-sign cell))(print-newline-and-white-chars level)
      (if (not (list-null? (get-cell-faces cell)))
        (progn
	  (princ "Faces ..............: ")(princ (get-cell-faces cell))(print-newline-and-white-chars level)))
      (if (not (list-null? (get-cell-vertices cell)))
	(progn
          (princ "Vertices ...........: ")(princ (get-cell-vertices cell))(print-newline-and-white-chars level)))
      (princ "Adjacent cells .....: ")(princ (get-cell-adjacences cell))(print-newline-and-white-chars level)
      (princ "Internal adjacences : ")
      (princ (get-cell-internal-adjacences cell)))))

;-----------------------------------------------------------------------------
; Visualizzazione delle adiacenze esterne
;-----------------------------------------------------------------------------

(defun show-external-adjacences (ext-adiac)
  (let ((show-list '()))
    (dolist (x ext-adiac)
      (if (or (equal x 0) (equal x NIL))
        (setq show-list (append show-list (list x)))
        (setq show-list (append show-list (list (list (car x) 'pol))))))
   show-list))

;-----------------------------------------------------------------------------
; Stampa della lista dei simplessi cubici
;-----------------------------------------------------------------------------

(defun print-cubic-simplexes (simp-table &optional (level 0))
 (print-newline-and-white-chars (+ level 4))
 (princ "Cubic-Simplexes")
 (print-newline-and-white-chars (+ level 4))
 (princ "---------------")
 (maphash #'(lambda (id simpl)
     (print-newline-and-white-chars (+ level 6))
     (princ id)
     (princ " : ")
     (princ (kernel::cubic-simplex-EDGES simpl))
     (print-newline-and-white-chars (+ level 6))
     (princ (kernel::cubic-simplex-VERTICES simpl))
     (maphash #'(lambda (i elem)
         (print-newline-and-white-chars (+ level 10))
         (princ i)
         (princ " ")
         (princ elem))
       (kernel::cubic-simplex-BEZIER-NET simpl)))
 simp-table))

  
;-----------------------------------------------------------------------------
; Stampa di un complesso poliedrale
;-----------------------------------------------------------------------------
(defun print-newline-and-white-chars (&optional (white-chars 0))
  (terpri)
  (dotimes (i white-chars) (princ " ")))

;-----------------------------------------------------------------------------
; Stampa di un complesso poliedrale
;-----------------------------------------------------------------------------
(defun print-pol-complex (pol-complex &optional (printalot t) (level 0))
  (if (polyhedron? pol-complex)
    (printpol pol-complex printalot level)
    (progn 
      (print-newline-and-white-chars level)
      (princ "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
      (print-newline-and-white-chars level)
      (princ " Start level ")
      (princ level) 
      (print-newline-and-white-chars level)
      (princ "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
      (mapcar 'print-pol-node pol-complex (never-ending-list printalot) (never-ending-list level))
      (print-newline-and-white-chars level)
      (princ "-----------------------------------------------------------")
      (print-newline-and-white-chars level)
      (princ " End level ")
      (princ level)
      (print-newline-and-white-chars level)
      (princ "-----------------------------------------------------------")))
  (values))
  

;-----------------------------------------------------------------------------
; Stampa di un nodo di un complesso poliedrale
;-----------------------------------------------------------------------------
(defun print-pol-node (pol-node &optional (printalot t) (level 0))
  (print-newline-and-white-chars level)
  (princ " Node : ")
  (princ (get-node-id pol-node))
  (print-newline-and-white-chars level)
  (princ " Matrix : ")
  (princ (get-node-matrix pol-node))
  (print-pol-complex (get-node-pol-complex pol-node) printalot (1+ level))
  (values))
  

;-----------------------------------------------------------------------------
; Estensione della savepol al caso di poliedri conteneti anche
; hashtables di facce e celle convesse.
; Il cambiamento principale rispetto alla versione precedente
;  data dal fatto che, (a causa della presenza delle hashtables)
; invece di salvare una stampa del polyhedro si salva una forma
; lisp la cui valutazione restituisca un polyhedro uguale a
; quello fornito in input.
;-----------------------------------------------------------------------------
(defun savepol (pol &optional (filename nil))
  (if (polyhedron? pol)
    (let ((input-data filename)
	  (filename (if filename filename
			(prompt-for-a-new-file))))
      (if (and (not input-data)
	       (not (string-equal (pathname-type filename) "pol")))
	(setq filename (concatenate 'string (pathname-host  filename)
				    (pathname-device filename)
				    (pathname-directory filename)
				    (pathname-name filename)
				    ".pol")))
      (with-open-file (outfile filename
			       :direction :output
			       :if-exists :overwrite
			       :if-does-not-exist :create)
	(let ((cells-list (get-clist pol))
	      (faces-list (get-flist pol))
	      (cells (get-cells pol))
	      (faces (get-faces pol)))
	  (setf (kernel::polyhedron-faces pol) nil)
	  (setf (kernel::polyhedron-convex-cells pol) nil)
	  (princ " (let ((pol 
 " outfile)
	  (prin1 pol outfile)
	  (princ " )) 
 (put-cells
 '" outfile)
	  (prin1 cells-list outfile)
	  
	  (princ " pol t) 
  (put-faces 
 '" outfile)
	  (prin1 faces-list outfile)
	  
	  (princ " pol t) pol)
 " outfile)
	  (put-faces faces pol t)
	  (put-cells cells pol t))
	t))
    (error-handler 1 "save-pol")))

;-----------------------------------------------------------------------------
; Modificato per leggere anche i poliedri contenenti hash tables di funzioni e
; celle convesse e salvati dalla nuova savepol.
; La modifica consiste nel valutare l'aggetto letto da disco prima di assegnarlo
; alla variabile pol.
;-----------------------------------------------------------------------------
(defun readpol (&optional(filename nil))
  (if (not filename) (setq filename (prompt-for-an-existing-file)))
      (with-open-file (infile filename
			      :direction :input)
	(let ((pol (eval (read infile))))
	  (if (polyhedron? pol)
	    pol
	    (error-handler 6 "readpol")))))

;-----------------------------------------------------------------------------
; Legge un complesso poliedrale da file
;-----------------------------------------------------------------------------
(defun read-pol-complex (&optional(filename nil))
  (if (not filename) (setq filename (prompt-for-an-existing-file)))
      (with-open-file (infile filename
			      :direction :input)
	(let ((pol-complex (eval (read infile))))
	  (if (polyhedra-complex? pol-complex)
	    pol-complex
	    (error-handler 6 "readpol")))))

;-----------------------------------------------------------------------------
; Salva un complesso poliedrale su file
;-----------------------------------------------------------------------------
(defun save-pol-complex (pol-complex &optional (filename nil))
  (if (polyhedra-complex?  pol-complex)
    (let ((input-data filename)
	  (filename (if filename filename
			(prompt-for-a-new-file ))))
      (if (and (not input-data)
	       (not (string-equal (pathname-type filename) "polc")))
	(setq filename
	      #+:APPLE
	      (make-pathname :host (pathname-host filename)
			     :device (pathname-device filename)
			     :directory (pathname-directory filename)
			     :name (pathname-name filename)
			     :type ".polc")
	       #+:CLISP
	      (make-pathname (concatenate 'string (pathname-host  filename)
					  (pathname-device filename)
					  (pathname-directory filename)
					  (pathname-name filename)
					  ".plc"))
	      #+:RS6000
	      (make-pathname (concatenate 'string (pathname-host  filename)
					  (pathname-device filename)
					  (pathname-directory filename)
					  (pathname-name filename)
					  ".polc"))))
      (with-open-file (outfile filename
			       :direction :output
			       :if-exists :overwrite
			       :if-does-not-exist :create)
	(print-pol-complex-to-file pol-complex outfile)
	t))
    (error-handler 1 "save-pol-complex")))


;-----------------------------------------------------------------------------
; Stampa sull'output definito da outfile una forma lisp la cui valutazione
; restitusca il nodo poliedrale fornito in input
;-----------------------------------------------------------------------------
(defun print-pol-node-to-file (pol-node outfile)
  (if (arrayp (first pol-node))
    (progn
      (princ " (list " outfile)
      (prin1 (first pol-node) outfile)
      (print-pol-complex-to-file (second pol-node) outfile)
      (princ " '" outfile)
      (prin1 (third pol-node) outfile)
      (princ " ) " outfile))
    (progn
      (princ " (list " outfile)
      (print-pol-complex-to-file (first pol-node) outfile)
      (prin1 (second pol-node) outfile)
      (princ " '" outfile)
      (prin1 (third pol-node) outfile)
      (princ " ) " outfile))))


;-----------------------------------------------------------------------------
; Stampa sull'output definito da outfile una forma lisp la cui valutazione
; restitusca il complesso poliedrale fornito in input
;-----------------------------------------------------------------------------
(defun print-pol-complex-to-file (pol-complex outfile)
  (if (polyhedron? pol-complex)
    (let* ((pol pol-complex)
	   (cells-list (get-clist pol))
	   (faces-list (get-flist pol))
	   (cells (get-cells pol))
	   (faces (get-faces pol)))
      (setf (kernel::polyhedron-faces pol) nil)
      (setf (kernel::polyhedron-convex-cells pol) nil)
      (princ " (let ((pol 
 " outfile)
      (prin1 pol outfile)
      (princ " )) 
 (put-cells
 '" outfile)
      (prin1 cells-list outfile)
      
      (princ " pol t) 
  (put-faces 
 '" outfile)
      (prin1 faces-list outfile)
      (princ " pol t) pol)
 " outfile)
      (put-faces faces pol t)
      (put-cells cells pol t)
      t)
    (progn
      (princ " (list " outfile)
      (mapcar 'print-pol-node-to-file 
	      pol-complex
	      (never-ending-list outfile))
      (princ " ) " outfile))))





