; -*- package: io -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;           Universita' di Roma "La Sapienza"                              
;
;       Dipartimento di Informatica e Sistemistica                         
;
;            Via Buonarroti 12 - 00185 Roma                                
;
;                                                                          
;
;                      CAD Group                                           
;                    
;                                                                          
;
; Project: SimpleXn                                                        
;
; Identification: Package Input&Output                                     
;                                
;                                                                          
;
;  Version: 1.0                                                            
;
;                                                                          
;
;  Author: V. Pascucci
;  Changed: M. Vicentino (Export to a generic stream, 21/03/1998)                                                     
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;

(provide 'io_ext2)
(INPACKAGE IO)
(export '(export-pol-complex
	  export-pol-complex-structure
	  export-polyhedron
	  export-pol-complex-roblan))
(require 'matlib_e)
(use-package 'matlib)
#+:APPLE (use-package 'ccl)

;esportazione di un complesso poliedrale come struttura minerva
(defun export-pol-complex-structure (pol-complex &optional pathname 
						 filename)
  (if (symbol-equal (get-pol-complex-rep pol-complex) 'on-faces)
    (setf pol-complex (coerce-pol-complex pol-complex 'on-vertices)))
  (let ((exported-pol-complex nil))
    (declare (special exported-pol-complex))
    (execute-export-pol-complex-structure pol-complex pathname filename)))

(defun execute-export-pol-complex-structure (pol-complex &optional pathname
						 filename )
  (declare (special exported-pol-complex))
  (when (null pathname)
    (setf pathname (prompt-for-a-new-file "new-structure" "")
	  filename (file-namestring pathname)
	  pathname (concatenate 'string (namestring pathname) ":"))
    (reset-polyhedral-node-counter)
    (create-file pathname :if-exists nil))
      (with-open-file (outfile (concatenate 'string
				    pathname filename ".STR")
			     :direction :output
			     :if-exists :supersede
			     :if-does-not-exist :create)
	(princ 
	 "  -199     ; Ascii structure code
;
" outfile)
(princ (length pol-complex) outfile)
(princ " ; No. of structure objects
;
; Objects list...
" outfile)
	
  (mapcar #'(lambda(node)
	      (cond
	       ((polyhedron? (get-node-pol-complex node))
		(princ 
"; object
 1.000000000e+0    0.000000000e+0    0.000000000e+0    0.000000000e+0   
 0.000000000e+0    1.000000000e+0    0.000000000e+0    0.000000000e+0   
 0.000000000e+0    0.000000000e+0    1.000000000e+0    0.000000000e+0   
 0.000000000e+0    0.000000000e+0    0.000000000e+0    1.000000000e+0   
polyhedron
"outfile)
		(advance-polyhedral-node-counter)
		(princ (string (current-polyhedral-node-counter)) outfile)
		(princ "
" outfile)
		(export-pol-complex (list node)
				    (concatenate 'string
						 pathname 
						 (string (current-polyhedral-node-counter))
						 ".POL")))
	       (t
		(let* ((mat (get-node-matrix node))
		       (pol (get-node-pol-complex node))
		       (n (array-dimension mat 0)))
		  (princ "; object
" outfile)
		  (dotimes (i n)
		    (dotimes (j n)
		      (format outfile " ~20,10E  "  (aref mat i j)))
		    (terpri outfile ))
		  (princ "
structure
"outfile)
		  (advance-polyhedral-node-counter)
		  (princ (string (current-polyhedral-node-counter)) outfile)
		  (execute-export-pol-complex-structure pol
						pathname
				    (string (current-polyhedral-node-counter)))))))
		
	  pol-complex))
(values)) 
(defun execute-export-pol-complex-structure (pol-complex &optional pathname

							 filename )
  (declare (special exported-pol-complex))
(if (symbol-equal (get-pol-complex-rep pol-complex) 'on-faces)
    (setf pol-complex (coerce-pol-complex pol-complex 'on-vertices)))
  (when (null pathname)
    (setf pathname (prompt-for-a-new-file "new-structure" "")
	  filename (file-namestring pathname)
	  pathname (concatenate 'string (namestring pathname) ":"))
    (reset-polyhedral-node-counter)
    (create-file pathname :if-exists :supersede))
  (with-open-file (outfile (concatenate 'string
					pathname filename ".STR")
			   :direction :output
			   :if-exists :supersede
			   :if-does-not-exist :create)
    (princ 
     "  -199     ; Ascii structure code
;
" outfile)
    (princ (length pol-complex) outfile)
    (princ " ; No. of structure objects
;
; Objects list..." outfile)
    
    (mapcar #'(lambda(node)
		(let* ((mat (get-node-matrix node))
		       (pol (get-node-pol-complex node))
		       (n (array-dimension mat 0))
		       name)
		  (princ "
; object
" outfile)
		  (dotimes (i 4)
		    (dotimes (j 4)
		      (let ((h (mod (1+ i) 4))
			    (k (mod (1+ j) 4)))
			(format outfile " ~20,10E  "  
				(if (and (< h n) (< k n))
				  (aref mat k h )
				  (if (= h k) 1 0))
			)))
		    (terpri outfile ))
		  (cond
		   ((polyhedron? pol)
		    (princ "polyhedron
"outfile)
		    (princ 
		     (cdr (or (assoc pol exported-pol-complex :test #'equalp)
			      (car (push (cons pol
					       (progn 
						 (advance-polyhedral-node-counter)
						 (setf name 
						       (string (current-polyhedral-node-counter)))
						 (export-pol-complex 
						  ; (list node)
						  (list (make-pol-node pol (mat-identity 4)))
						  (concatenate
						   'string
						   pathname 
						   name
						   ".POL"))
						 name
						 ))
					 exported-pol-complex))))
		     outfile))
		   (t
		    (princ "structure
"outfile)
		    (princ 
		     (cdr (or (assoc pol exported-pol-complex :test #'equalp)
			      (car (push (cons pol
					       (progn 
						 (advance-polyhedral-node-counter)
						 (setf name 
						       (string (current-polyhedral-node-counter)))
						 (execute-export-pol-complex-structure
						  pol
						  pathname
						  name)
						 name
						 ))
					 exported-pol-complex))))
		     outfile)))))
	    pol-complex))
  (values))

;esportazione di un complesso poliedrale come struttura roblan
(defun export-pol-complex-roblan (pol-complex &optional pathname level)
  (setf pathname 
	(if pathname
	  (let* ((dir-name (directory-namestring pathname))
		 (name-string (file-namestring pathname))
		 (empty-name (string= dir-name 
				      #+:APPLE  ""
				      ;;#+:RS6000 "./"
				      #+:DOS ""
				      #+:UNIX "./"
				      ))
		 (full-dir (if (and (not empty-name)
				    #+:APPLE (directoryp dir-name)
				    #+:RS6000 (XTK::directoryp dir-name)
				    #+:CLISP   (directoryp dir-name)
				    )
			     dir-name)))
	    (if (string= name-string "")
	      (prompt-for-a-new-file full-dir)
	      (make-pathname  :directory full-dir
			      :name name-string)))
	  (prompt-for-a-new-file 
	   "untitled" ".pol")))


  (if (symbol-equal (get-pol-complex-rep pol-complex) 'on-faces)
    (setf pol-complex (coerce-pol-complex pol-complex 'on-vertices)))
  (let ((exported-pol-complex nil)
	filename)
    (declare (special exported-pol-complex))
    (setf filename (file-namestring pathname)
	  pathname (concatenate 'string (namestring pathname) ":"))
      (reset-polyhedral-node-counter)
      (create-file pathname :if-exists nil)
    (with-open-file (outfile (concatenate 'string
					  pathname filename ".rl")
			     :direction :output
			     :if-exists :supersede
			     :if-does-not-exist :create)
(princ "
FUN POL LOAD (STG name)
f := OPEN(STG name, 'R')
p := POL READ(f)
CLOSE(f)
RETURN (POL p)
ENDFUN LOAD

MAIN

" outfile )      (execute-export-pol-complex-roblan 
		  (coerce-pol-complex pol-complex 'on-vertices) outfile 
		  (current-polyhedral-node-counter) pathname filename 
		  (if (numberp level) level))
      
      (princ "
SLIDE (PC0, VF_DIM_VIEW+VF_LOCK)
END.
" outfile )
(values))))
(defun execute-export-pol-complex-roblan (pol-complex outfile name pathname
filename &optional level)
  (declare (special exported-pol-complex))
  (let ((pol-complex-elements   
	 (mapcar #'(lambda(node)
		     (let* ((mat (get-node-matrix node))
			    (pol (get-node-pol-complex node))) 
		       
		       (cons (if (mat-identity? mat) nil
			      (progn 
			       (advance-polyhedral-node-counter)
			       (let ((name (string (current-polyhedral-node-counter))))
				 (export-matrix-roblan mat name outfile)
				 name)))
			     (cdr (or (assoc pol exported-pol-complex :test #'equalp)
				      (car (push (cons pol
						       (progn 
							 (advance-polyhedral-node-counter)
							 (let ((name (string (current-polyhedral-node-counter))))
							   (if (or (polyhedron? pol)
								   (and level (not (plusp level))))                                
                    
							     (export-pol-roblan pol outfile pathname name filename)
							     (execute-export-pol-complex-roblan pol outfile name pathname
filename (if level
															    (1- level))))
							   name)))
						 exported-pol-complex)))))))
		 pol-complex)))
    (princ name outfile)
    (princ ":= STR <" outfile)
    (dolist (data (butlast pol-complex-elements))
      (let ((mat (car data))
	    (name (cdr data)))
	(if mat (progn
		  (princ " STR <MAT " outfile)
		  (princ mat outfile)
		  (princ ", STR " outfile)
		  (princ name outfile)
		  (princ ">," outfile))
	    (progn
	      (princ " STR " outfile)
	      (princ name outfile)
	      (princ "," outfile)))))
    (let ((data (car (last pol-complex-elements))))
      (let ((mat (car data))
	    (name (cdr data)))
	(if mat (progn
		  (princ " STR <MAT " outfile)
		  (princ mat outfile)
		  (princ ", STR " outfile)
		  (princ name outfile)
		  (princ "> > " outfile))
	    (progn
	      (princ " STR " outfile)
	      (princ name outfile)
	      (princ " > " outfile)))))
    (princ "
" outfile)))
(defun export-matrix-roblan (mat name &optional outfile)
  (let ((n (array-dimension mat 0)))
    (princ name outfile)
    (princ ":=[" outfile)
    (dotimes (i 4)
      (princ "[" outfile)
      (dotimes (j 4)
	(let* ((h (mod (1+ i) 4))
	       (k (mod (1+ j) 4))
	       (num (if (and (< h n) (< k n))
		      (aref mat k h )
		      (if (= h k) 1 0))))
	  (if (zerop num)(princ 0 outfile)
	      (format outfile "~15,15F " num ))
	  (if (not (eq j 3))(princ "," outfile))))
      (princ "]" outfile))
    (princ "] 
" outfile)))



 
(defun export-pol-complex (pol-complex &optional filename)
  (if (symbol-equal (get-pol-complex-rep pol-complex) 'on-faces)
    (setf pol-complex (coerce-pol-complex pol-complex 'on-vertices)))
  (let* ((n (get-npol-complex pol-complex))
	 (id-mat (mat-identity (1+ n)))
	 (vertices (make-hash-table))
	 (faces (make-hash-table)))
    (with-open-file (outfile (or filename 
				 (prompt-for-a-new-file))
			     :direction :output
			     :if-exists :supersede
			     :if-does-not-exist :create)
      
      (mapcar #'(lambda (node)
		  (get-pol-complex-vertices-and-faces node id-mat vertices faces))
	      pol-complex)
      (format outfile 
	      "  -200     ; Ascii polyhedron code
;
")
      (prin1 (+ 3 (hash-table-count vertices)) outfile)
      (format outfile 
	      "     ; No. of polyhedron points
;
; Points list...
")
      
      (dotimes (i (hash-table-count vertices))
	(mapcar #'(lambda (number)
		   (prin1 number outfile)(princ "  " outfile))
		(gethash (1+ i) vertices))
	(terpri outfile))
      
      (mapcar #'(lambda (number)
		 (prin1 number outfile)(princ "  " outfile))
	      (mapcar '+ '(0.1 0 0)(gethash 1 vertices)))
      (terpri outfile)
      (mapcar #'(lambda (number)
		 (prin1 number outfile)(princ "  " outfile))
	      (mapcar '+ '(0 0.1 0)(gethash 1 vertices)))
      (terpri outfile)
      (mapcar #'(lambda (number)
		 (prin1 number outfile)(princ "  " outfile))
	      (mapcar '+ '(0 0 0.1)(gethash 1 vertices)))
      (terpri outfile)
      (princ
       ";
;
" outfile)
      (prin1 (hash-table-count faces) outfile)
      (princ "    ; No. of polyhedron triangles
;
; Triangles list...
" outfile)
      (dotimes (i (hash-table-count faces))
	(mapcar #'(lambda (number)
		   (prin1 number outfile)(princ "  " outfile))
		(gethash (1+ i) faces))
	(terpri outfile))
      )))

(defun export-pol-roblan (pol outfile pathname name filename1)
  (advance-polyhedral-node-counter)
  (let* ((filename (concatenate 'string pathname name)))
    (export-pol-complex (list (make-pol-node pol (mat-identity 4)))
filename)
    (princ name outfile)
    (princ ":= STR < POL LOAD('" outfile)(princ filename1 outfile)(princ
"/" outfile)(princ name outfile)(princ "') >
" outfile)))
;versione per stampare piu' cifre !
(defun export-pol-complex (pol-complex &optional filename)
  (if (symbol-equal (get-pol-complex-rep pol-complex) 'on-faces)
    (setf pol-complex (coerce-pol-complex pol-complex 'on-vertices)))
  (let* ((n (get-npol-complex pol-complex))
	 (id-mat (mat-identity (1+ n)))
	 (vertices (make-hash-table))
	 (faces (make-hash-table)))
    (with-open-file (outfile (or filename 
				 (prompt-for-a-new-file))
			     :direction :output
			     :if-exists :supersede
			     :if-does-not-exist :create)
      
      (mapcar #'(lambda (node)
		  (get-pol-complex-vertices-and-faces node id-mat vertices faces))
	      pol-complex)
      (format outfile 
	      "  -200     ; Ascii polyhedron code
;
")
      (prin1 (+ 3 (hash-table-count vertices)) outfile)
      (format outfile 
	      "     ; No. of polyhedron points
;
; Points list...
")
      
      (dotimes (i (hash-table-count vertices))
	(mapcar #'(lambda (number)
		   (format outfile "~15,15F" number )(princ "  " outfile))
		(gethash (1+ i) vertices))
	(terpri outfile))
      
      (mapcar #'(lambda (number)
		 (format outfile "~15,15F" number )(princ "  " outfile))
	      (mapcar '+ '(0.1 0 0)(gethash 1 vertices)))
      (terpri outfile)
      (mapcar #'(lambda (number)
		 (format outfile "~15,15F" number )(princ "  " outfile))
	      (mapcar '+ '(0 0.1 0)(gethash 1 vertices)))
      (terpri outfile)
      (mapcar #'(lambda (number)
		 (format outfile "~15,15F" number)(princ "  " outfile))
	      (mapcar '+ '(0 0 0.1)(gethash 1 vertices)))
      (terpri outfile)
      (princ
       ";
;
" outfile)
      (prin1 (hash-table-count faces) outfile)
      (princ "    ; No. of polyhedron triangles
;
; Triangles list...
" outfile)
      (dotimes (i (hash-table-count faces))
	(mapcar #'(lambda (number)
		   (prin1  number outfile)(princ "  " outfile))
		(gethash (1+ i) faces))
	(terpri outfile))
      )))

(defun export-pol-complex (pol-complex &optional filename)
  (if (symbol-equal (get-pol-complex-rep pol-complex) 'on-faces)
    (setf pol-complex (coerce-pol-complex pol-complex 'on-vertices)))
  (let* ((n (get-npol-complex pol-complex))
	 (id-mat (mat-identity (1+ n)))
	 (vertices (make-hash-table))
	 (faces (make-hash-table)))
    (with-open-file (outfile (or filename 
				 (prompt-for-a-new-file))
			     :direction :output
			     :if-exists :supersede
			     :if-does-not-exist :create)
      
      (mapcar #'(lambda (node)
		  (get-pol-complex-vertices-and-faces node id-mat vertices faces))
	      pol-complex)
      (format outfile 
	      "  -200     ; Ascii polyhedron code
;
")
      (prin1 (+ 3 (hash-table-count vertices)) outfile)
      (format outfile 
	      "     ; No. of polyhedron points
;
; Points list...
")
      
      (dotimes (i (hash-table-count vertices))
	(mapcar #'(lambda (number)
		   (format outfile "~15,15F" number )(princ "  " outfile))
		(subseq (append (gethash (1+ i) vertices) '(0 0 0)) 0 3))
	(terpri outfile))
      
      (mapcar #'(lambda (number)
		 (format outfile "~15,15F" number )(princ "  " outfile))
	      (mapcar '+ '(0.1 0 0)
		      (subseq (append (gethash 1 vertices)'(0 0 0)) 0 3)
		      ))
      (terpri outfile)
      (mapcar #'(lambda (number)
		  (format outfile "~15,15F" number )(princ "  " outfile))
	      (mapcar '+ '(0 0.1 0)
		      (subseq (append (gethash 1 vertices)'(0 0 0)) 0 3)))
      (terpri outfile)
      (mapcar #'(lambda (number)
		 (format outfile "~15,15F" number)(princ "  " outfile))
	      (mapcar '+ '(0 0 0.1)
		      (subseq (append (gethash 1 vertices)'(0 0 0)) 0 3)))
      (terpri outfile)
      (princ
       ";
;
" outfile)
      (prin1 (hash-table-count faces) outfile)
      (princ "    ; No. of polyhedron triangles
;
; Triangles list...
" outfile)
      (dotimes (i (hash-table-count faces))
	(mapcar #'(lambda (number)
		   (prin1  number outfile)(princ "  " outfile))
		(gethash (1+ i) faces))
	(terpri outfile))
      )))
;aggiungere aggiustamento dimensioni matrici
(defun get-pol-complex-vertices-and-faces (node old-mat vertices faces)
  (let ((pol (get-node-pol-complex node))
	(matrix (get-node-matrix node)))
    (if (polyhedron? pol)
      (get-pol-vertices-and-faces pol (mat-product-extended old-mat matrix)
vertices faces)
      (mapcar `(lambda (node)
		 (get-pol-complex-vertices-and-faces 
		  node 
		  ,(mat-product-extended old-mat matrix)
		  ,vertices ,faces))
	      pol))))

;aggiungere aggiustamento dimensioni matrici
(defun get-pol-complex-vertices-and-faces (node old-mat vertices faces)
  (let ((pol (get-node-pol-complex node))
	(matrix (get-node-matrix node)))
    (if (polyhedron? pol)
      (get-pol-vertices-and-faces pol (mat-product-extended old-mat matrix)
vertices faces)
      (mapcar  #'(lambda (node)
		 (get-pol-complex-vertices-and-faces 
		  node 
		  (mat-product-extended old-mat matrix)
		  vertices faces))
	      pol))))

(defun get-pol-vertices-and-faces (pol mat vertices faces)
  (let* ((n (get-npol pol))
	 (nn (array-dimension mat 0))
	 (zeros (make-list (- nn n 1) :initial-element 0))
	 (vertices-list (get-vlist pol))
	 (cells-list (get-clist pol))
	 (vert-num (hash-table-count vertices))
	 (face-num (hash-table-count faces))
	 (verts-adiac (make-hash-table :test 'equal))
	 )
    
    (cond
     ((= n 1)
      (mapcar #'(lambda (cell)
		  (let* ((cell-vertices (get-cell-vertices cell))
			 (v1 (make-array 
			      nn
			      :initial-contents           
			      (append '(1)
				      (find-vcoord (first cell-vertices) vertices-list)
				      zeros)))
			 (v2 (make-array 
			      nn
			      :initial-contents           
			      (append '(1)
				      (find-vcoord (second cell-vertices) vertices-list)
				      zeros))))
		    (setf (gethash (incf vert-num) vertices)
			  (rest (coerce (mat-vet-product-extended
					 mat v1) 'list)))
		    (setf (gethash (incf vert-num) vertices)
			  (rest (coerce (mat-vet-product-extended
					 mat v2) 'list)))
		    (setf (gethash (incf face-num) faces) 
			  (list vert-num
				(1- vert-num)
				vert-num 0 0 0))
		    
		    
		    ))
	      cells-list))
     ((= n 2)
      (mapcar #'(lambda (cell)
		  (let* ((old-vert-num (1+ vert-num))
			 (old-face-num (1+ face-num))
			 (cell-vertices (get-cell-vertices cell))
			 (cells-num (length cell-vertices))
			 (cell-vert-coord 
			  (mapcar #'(lambda (vid)
				      (find-vcoord vid vertices-list))
				  cell-vertices))
			 (center (let ((x 0)(y 0))
				   (mapcar #'(lambda (xy)
					       (setf x (+ x (first xy))
						     y (+ y (second xy))))
					   cell-vert-coord)
				   (list (/ x cells-num) (/ y cells-num))))
			 (new-coord (mapcar #'(lambda (coords)
						(list (- (first coords)
							 (first center))
						      (- (second coords)
							 (second center))))
					    cell-vert-coord))
			 (phases (mapcar #'(lambda (coords)
					     (phase (complex (first coords)(second coords))))
					 new-coord))
			 (phases&coord (mapcar 'cons phases cell-vert-coord))
			 (sorted-coord 
			  (mapcar #'cdr (sort phases&coord #'< :key #'first)))
			 
			 
			 ; a questo punto i vertici della cella ( 2D ! ) sono ordinati in
			 ; base alla fase rispetto al centro della cella
			 
			 (world-coords
			  (mapcar #'(lambda (vet)
				      (mat-vet-product-extended
				       mat vet))
				  (mapcar #'(lambda (coord)
					      (make-array 
					       nn
					       :initial-contents           
					       (append '(1)
						       coord
						       zeros)))
					  sorted-coord)))
			 (world-center (mat-vet-product-extended
					
					mat 
					(make-array 
					 nn
					 :initial-contents           
					 (append '(1)
						 center
						 zeros))))
			 )
		    ;Pone nella tab. hash le coordinate del centro della cella
		    (setf (gethash (incf vert-num) vertices)
			  (rest (coerce world-center 'list)))
		    ;pone nella tavola hash tutti i vertici
		    (dotimes (i cells-num)
		      (setf (gethash (incf vert-num) vertices)
			    (rest (coerce (nth i world-coords) 'list))))
		    
		    
		    ;ciclo di gestione delle adiacenze
		    
		    (dotimes (i cells-num)
		      (setf (gethash (list
				      (gethash (+ old-vert-num 1 (mod (+ i 1) cells-num)) vertices)
				      (gethash (+ old-vert-num 1  i ) vertices)
				      )
				     verts-adiac)
			    (1+ (+ i face-num)))
		      )
		    
		    ;pone nella tab. hash dei triangoli tutti i triangoli ottenuti da
		    ; punto centrale + facet. cells-num e' il numero di vertici.
		    (dotimes (i cells-num)
		      (setf (gethash (incf face-num) faces) 
			    (list old-vert-num
				  (+ old-vert-num 1  i )
				  (+ old-vert-num 1 (mod (+ i 1) cells-num))
				  
				  (+ old-face-num (mod (- i 1) cells-num))
				  
				  (let ( (cella (gethash (list
							  (gethash (+ old-vert-num 1  i ) vertices)
							  (gethash (+ old-vert-num 1 (mod (+ i 1) cells-num)) vertices)
							  )
							 verts-adiac))
					 )
				    ; la linea di programma commentata che segue
				    ; sostituiva la successiva
				    ;(if (null cella)
				    (if (or (null cella) (null (gethash cella faces)))
				      0
				      ;else aggiusta adiacenza all'indietro e poni questa alla cella corretta
				      (progn
					(setf (nth 4  (gethash cella faces)) face-num)
					cella))
				    )
				  
				  (+ old-face-num (mod (+ i 1) cells-num))
				  ))
		      )
		    ))
	      
	      cells-list)))))

;this version avoids to triangolate a triangle convex cell.
(defun get-pol-vertices-and-faces (pol mat vertices faces)
  (let* ((n (get-npol pol))
	 (nn (array-dimension mat 0))
	 (zeros (make-list (- nn n 1) :initial-element 0))
	 (vertices-list (get-vlist pol))
	 (cells-list (get-clist pol))
	 (vert-num (hash-table-count vertices))
	 (face-num (hash-table-count faces))
	 (verts-adiac (make-hash-table :test 'equal))
	 )
    
    (cond
     ((= n 0)
      (mapcar #'(lambda (cell)
		  (let* ((cell-vertices (get-cell-vertices cell))
			 (v1 (make-array 
			      nn
			      :initial-contents           
			      (append '(1)
				      (find-vcoord (first cell-vertices) vertices-list)
				      zeros)))
			 
			 )
		    (setf (gethash (incf vert-num) vertices)
			  (rest (coerce (mat-vet-product-extended
					 mat v1) 'list)))
		    
		    (setf (gethash (incf face-num) faces) 
			  (list vert-num
				vert-num
				vert-num 0 0 0))))
	      cells-list))
     ((= n 1)
      (mapcar #'(lambda (cell)
		  (let* ((cell-vertices (get-cell-vertices cell))
			 (v1 (make-array 
			      nn
			      :initial-contents           
			      (append '(1)
				      (find-vcoord (first cell-vertices) vertices-list)
				      zeros)))
			 (v2 (make-array 
			      nn
			      :initial-contents           
			      (append '(1)
				      (find-vcoord (second cell-vertices) vertices-list)
				      zeros))))
		    (setf (gethash (incf vert-num) vertices)
			  (rest (coerce (mat-vet-product-extended
					 mat v1) 'list)))
		    (setf (gethash (incf vert-num) vertices)
			  (rest (coerce (mat-vet-product-extended
					 mat v2) 'list)))
		    (setf (gethash (incf face-num) faces) 
			  (list vert-num
				(1- vert-num)
				vert-num 0 0 0))
		    
		    
		    ))
	      cells-list))
     
     ((= n 2)
      (mapcar #'(lambda (cell)
		  (if (= (length (get-cell-vertices cell)) 3)
		    
		    (let* ((old-vert-num (1+ vert-num))
			   (cell-vertices (get-cell-vertices cell))
			   (cells-num (length cell-vertices))
			   (cell-vert-coord 
			    (mapcar #'(lambda (vid)
					(find-vcoord vid vertices-list))
				    cell-vertices))
			   (center (let ((x 0)(y 0))
				     (mapcar #'(lambda (xy)
						 (setf x (+ x (first xy))
						       y (+ y (second xy))))
					     cell-vert-coord)
				     (list (/ x cells-num) (/ y cells-num))))
			   (new-coord (mapcar #'(lambda (coords)
						  (list (- (first coords)
							   (first center))
							(- (second coords)
							   (second center))))
					      cell-vert-coord))
			   (phases (mapcar #'(lambda (coords)
					       (phase (complex (first coords)(second coords))))
					   new-coord))
			   (phases&coord (mapcar 'cons phases cell-vert-coord))
			   (sorted-coord 
			    (mapcar #'cdr (sort phases&coord #'< :key #'first)))
			   ; a questo punto i vertici della cella ( 2D ! ) sono ordinati in
			   ; base alla fase rispetto al centro della cella
			   (world-coords
			    (mapcar #'(lambda (vet)
					(mat-vet-product-extended
					 mat vet))
				    (mapcar #'(lambda (coord)
						(make-array 
						 nn
						 :initial-contents           
						 (append '(1)
							 coord
							 zeros)))
					    sorted-coord))))
		      ;;   ;Pone nella tab. hash le coordinate del centro della cella
		      ;;   (setf (gethash (incf vert-num) vertices)
		      ;;         (rest (coerce world-center 'list)))
		      ;pone nella tavola hash tutti i vertici
		      (dotimes (i cells-num)
			(setf (gethash (incf vert-num) vertices)
			      (rest (coerce (nth i world-coords) 'list))))
		      
		      
		      ;ciclo di gestione delle adiacenze
		      
		      (dotimes (i cells-num)
			(setf (gethash (list
					(gethash (+ old-vert-num (mod (+ i 1) cells-num)) vertices)
					(gethash (+ old-vert-num i ) vertices)
					)
				       verts-adiac)
			      (list (1+  face-num) i)))
		      
		      ;pone nella tab. hash dei triangoli tutti i triangoli ottenuti da
		      ; punto centrale + facet. cells-num e' il numero di vertici.
		      ;; (dotimes (i cells-num)
		      (dotimes (i 1)
			(setf (gethash (incf face-num) faces) 
			      (list old-vert-num
				    (+ old-vert-num 1  i )
				    (+ old-vert-num 1 (mod (+ i 1) cells-num))
				    
				    (let* ((cella-face (gethash (list
								 (gethash (+ old-vert-num i ) vertices)
								 (gethash (+ old-vert-num (mod (+ i 1) cells-num)) vertices))
								verts-adiac))
					   (cella (first cella-face))
					   (face  (second cella-face)))
				      ; la linea di programma commentata che segue
				      ; sostituiva la successiva
				      ;(if (null cella)
				      (if (or (null cella) (null (gethash cella faces)))
					0
					;else aggiusta adiacenza all'indietro e poni questa alla cella corretta
					(progn
					  (setf (nth (+ 3 face) (gethash cella faces)) face-num)
					  cella)))
				    (let* ((cella-face (gethash (list
								 (gethash (+ old-vert-num i 1) vertices)
								 (gethash (+ old-vert-num (mod (+ i 2) cells-num)) vertices))
								verts-adiac))
					   (cella (first cella-face))
					   (face  (second cella-face)))
				      ; la linea di programma commentata che segue
				      ; sostituiva la successiva
				      ;(if (null cella)
				      (if (or (null cella) (null (gethash cella faces)))
					0
					;else aggiusta adiacenza all'indietro e poni questa alla cella corretta
					(progn
					  (setf (nth (+ 3 face) (gethash cella faces)) face-num)
					  cella)))
				    (let* ((cella-face (gethash (list
								 (gethash (+ old-vert-num i 2) vertices)
								 (gethash (+ old-vert-num (mod (+ i 3) cells-num)) vertices))
								verts-adiac))
					   (cella (first cella-face))
					   (face  (second cella-face)))
				      ; la linea di programma commentata che segue
				      ; sostituiva la successiva
				      ;(if (null cella)
				      (if (or (null cella) (null (gethash cella faces)))
					0
					;else aggiusta adiacenza all'indietro e poni questa alla cella corretta
					(progn
					  (setf (nth (+ 3 face) (gethash cella faces)) face-num)
					  cella)))
				    ; 0 0 
				    ))))                    
		    (let* ((old-vert-num (1+ vert-num))
			   (old-face-num (1+ face-num))
			   (cell-vertices (get-cell-vertices cell))
			   (cells-num (length cell-vertices))
			   (cell-vert-coord 
			    (mapcar #'(lambda (vid)
					(find-vcoord vid vertices-list))
				    cell-vertices))
			   (center (let ((x 0)(y 0))
				     (mapcar #'(lambda (xy)
						 (setf x (+ x (first xy))
						       y (+ y (second xy))))
					     cell-vert-coord)
				     (list (/ x cells-num) (/ y cells-num))))
			   (new-coord (mapcar #'(lambda (coords)
						  (list (- (first coords)
							   (first center))
							(- (second coords)
							   (second center))))
					      cell-vert-coord))
			   (phases (mapcar #'(lambda (coords)
					       (phase (complex (first coords)(second coords))))
					   new-coord))
			   (phases&coord (mapcar 'cons phases cell-vert-coord))
			   (sorted-coord 
			    (mapcar #'cdr (sort phases&coord #'< :key #'first)))
			   
			   
			   ; a questo punto i vertici della cella ( 2D ! ) sono ordinati in
			   ; base alla fase rispetto al centro della cella
			   
			   (world-coords
			    (mapcar #'(lambda (vet)
					(mat-vet-product-extended
					 mat vet))
				    (mapcar #'(lambda (coord)
						(make-array 
						 nn
						 :initial-contents           
						 (append '(1)
							 coord
							 zeros)))
					    sorted-coord)))
			   (world-center (mat-vet-product-extended
					  
					  mat 
					  (make-array 
					   nn
					   :initial-contents           
					   (append '(1)
						   center
						   zeros))))
			   )
		      ;Pone nella tab. hash le coordinate del centro della cella
		      (setf (gethash (incf vert-num) vertices)
			    (rest (coerce world-center 'list)))
		      ;pone nella tavola hash tutti i vertici
		      (dotimes (i cells-num)
			(setf (gethash (incf vert-num) vertices)
			      (rest (coerce (nth i world-coords) 'list))))
		      
		      
		      ;ciclo di gestione delle adiacenze
		      
		      (dotimes (i cells-num)
			(setf (gethash (list
					(gethash (+ old-vert-num 1 (mod (+ i 1) cells-num)) vertices)
					(gethash (+ old-vert-num 1  i ) vertices)
					)
				       verts-adiac)
			      (list (1+ (+ i face-num)) 1))
			)
		      
		      ;pone nella tab. hash dei triangoli tutti i triangoli ottenuti da
		      ; punto centrale + facet. cells-num e' il numero di vertici.
		      (dotimes (i cells-num)
			(setf (gethash (incf face-num) faces) 
			      (list old-vert-num
				    (+ old-vert-num 1  i )
				    (+ old-vert-num 1 (mod (+ i 1) cells-num))
				    
				    (+ old-face-num (mod (- i 1) cells-num))
				    
				    (let* ((cella-face (gethash 
						       (list
							(gethash (+ old-vert-num 1  i ) vertices)
							(gethash (+ old-vert-num 1 (mod (+ i 1) cells-num)) vertices))
						       verts-adiac))
					  (cella (first cella-face))
					  (face  (second cella-face)))
				      ; la linea di programma commentata che segue
				      ; sostituiva la successiva
				      ;(if (null cella)
				      (if (or (null cella) (null (gethash cella faces)))
					0
					;else aggiusta adiacenza all'indietro e poni questa alla cella corretta
					(progn
					  (setf (nth (+ 3 face)  (gethash cella faces)) face-num)
					  cella)))
				    (+ old-face-num (mod (+ i 1) cells-num))))))))
	      
	      cells-list)))))


;----------------------------------------------------
; Export a single polyhedron.
; Creates a dummy polyhedral-complex composed only
; of the single input polyhedron, then exports
; the polyhedral complex.
;---------------------------------------------------
(defun export-polyhedron (polyhedron &optional filename)        
  (let ((pol-complex '())) 
    (put-node-in-pol-complex (make-pol-node polyhedron) pol-complex)
    (if filename (export-pol-complex  (skeleton pol-complex 2) filename )
	(export-pol-complex  (skeleton pol-complex 2)))))


;---------------------------------------------------------------------------

;    unredundant export version 


(defvar epsilon 1.0d-4 ;kernel::**epsilon
		      )

; Crea la lista della rappresentazione triangolo-alata a partire
; dalle hash-table dei vertici e delle facce.
(defun CREA-W-REP (vertices faces)
  (list (crea-lista vertices)
	(crea-lista faces)))

(defun INSERT-INDICE (lista)
  (let ((i 0))
    (mapcar #'(lambda (x) (list x (incf i))) lista)))


(defun AGGIORNA-INDICI (triangoli v) 
  (let* ((half (/ (length (first triangoli)) 2)))
    (declare (special half))
    (mapcar #'(lambda (a-triangle)
		(append
		 (mapcar #'(lambda (index) (aref v index))
			 (subseq a-triangle 0 half))
		 (subseq a-triangle half)))
	    triangoli)))

(defun CONFRONTA1 (x y)
  (< (abs (- x y)) epsilon))

(defun CONFRONTA (x y)
  (cond ((not x))
	((confronta1 (first x) (first y))
	 (confronta (rest x) (rest y)))))


(defun RIPULISCI (lista-w-rep)
  (let* ((triangoli (second lista-w-rep))
	 (vertici-ord-ind (sort (insert-indice (first lista-w-rep))
				'vert-min
				:key 'car))
	 (v (make-array (+ 1 (length vertici-ord-ind))
			:element-type 'integer)))
    (list (new-lista vertici-ord-ind v)
	  (aggiorna-indici triangoli v))))


(defun VERT-MIN (x y)
  (cond ((confronta1 (car x) (car y))
	 (cond ((null (cdr x)) nil)
	       (t (vert-min (cdr x) (cdr y)))))
	((< (car x) (car y)))))

(defun NEW-LISTA (lista-vert-ord v)
  (setf (aref v (second (first lista-vert-ord))) 1)
  (cons (first (first  lista-vert-ord))
	(pulisci-crea-v (rest lista-vert-ord) v (first lista-vert-ord) 1)))

(defun PULISCI-CREA-V (lista-vert-ord v elemento-prec indice)
  (cond ((null lista-vert-ord) nil)
	((confronta (car elemento-prec) (caar lista-vert-ord))
	 (setf (aref v (cadar lista-vert-ord)) indice)
	 (pulisci-crea-v (cdr lista-vert-ord) v elemento-prec indice))
	(t  (setf (aref v (cadar lista-vert-ord)) (+ 1 indice))
	    (cons (caar lista-vert-ord)                 
		  (pulisci-crea-v 
		   (cdr lista-vert-ord) v (car lista-vert-ord) (+ 1 indice))))))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;versione TAIL RECURSIVE

(defun NEW-LISTA (lista-vert-ord v)
  (setf (aref v (second (first lista-vert-ord))) 1)
  (cons (first (first  lista-vert-ord))
	(pulisci-crea-v (rest lista-vert-ord) v (first lista-vert-ord) 1 nil)))


(defun PULISCI-CREA-V (lista-vert-ord v elemento-prec indice res)
  (cond ((null lista-vert-ord) (reverse res))
	((confronta (car elemento-prec) (caar lista-vert-ord))
	 (setf (aref v (cadar lista-vert-ord)) indice)
	 (pulisci-crea-v (cdr lista-vert-ord) v elemento-prec indice res))
	(t  (setf (aref v (cadar lista-vert-ord)) (+ 1 indice))               
	    (pulisci-crea-v 
	     (cdr lista-vert-ord) 
	     v 
	     (car lista-vert-ord)
	     (+ 1 indice) 
	     (cons (caar lista-vert-ord) res)))))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Crea una lista a partire da un hash-table in modo tail recursive.

(defun CREA-LISTA (t-hash)
  (crea-lista1 t-hash (hash-table-count t-hash) nil))


(defun CREA-LISTA1 (t-hash n-elem res)
  (if (= n-elem 0)
    res
    (crea-lista1 t-hash (- n-elem 1) (cons (gethash n-elem t-hash) res))))



(defun EXPORT-POL-COMPLEX (pol-complex &optional filename)
  (if (symbol-equal (get-pol-complex-rep pol-complex) 'on-faces)
    (setf pol-complex (coerce-pol-complex pol-complex 'on-vertices)))
  (let* ((n (get-npol-complex pol-complex))
	 (id-mat (mat-identity (1+ n)))
	 (vertices (make-hash-table))
	 (faces (make-hash-table)))
    (with-open-file (outfile (or filename 
				 (prompt-for-a-new-file))
			     :direction :output
			     :if-exists :supersede
			     :if-does-not-exist :create)
      
      (mapcar #'(lambda (node)
		  (get-pol-complex-vertices-and-faces node id-mat vertices faces))
	      pol-complex)
      
      
      
      (let ((clear-w-rep (ripulisci (crea-w-rep vertices faces))))
	
	;
	;  Scrittura del file.
	;
	
	(format outfile 
		"  -200     ; Ascii polyhedron code
;
")
	(prin1 (+ 3 (length (first clear-w-rep))) outfile)
	(format outfile 
		"     ; No. of polyhedron points
;
; Points list...
")
	(mapcar #'(lambda (vertice)
		    (mapcar #'(lambda (number)
				(format outfile "~15,15F " number ))
			    (subseq (append vertice '(0 0 0)) 0 3))
		    (terpri outfile))
		(first clear-w-rep))
	
	(anti-bomb-minerva outfile (gethash 1 vertices))
	
	(princ
	 ";
;
" outfile)
	(prin1 (length (second clear-w-rep)) outfile)
	(princ "    ; No. of polyhedron triangles
;
; Triangles list...
" outfile)
	
	(mapcar #'(lambda (triangolo)
		    (mapcar #'(lambda (number)
				(format outfile "~A  " number))
			    triangolo)
		    (terpri outfile))
		(second clear-w-rep))  
	
	
	()
	))))

; crea tre punti affinemente indipendenti al punto point
; serve per evitare che MINERVA bombi nel caso di oggetti piani

(defun ANTI-BOMB-MINERVA (outfile point) 
  (let ((fun #'(lambda (number)
		 (format outfile "~15,15F  " number )))
	(3d-point (subseq (append point '(0 0 0)) 0 3)))
    (mapcar fun (mapcar '+ '(0.1 0 0) 3d-point))
    (terpri outfile)
    (mapcar fun (mapcar '+ '(0 0.1 0) 3d-point))
    (terpri outfile)
    (mapcar fun (mapcar '+ '(0 0 0.1) 3d-point))
    (terpri outfile)))


(defun AGGIORNA-ALL-INDICI (triangoli v) 
  (mapcar #'(lambda (a-triangle)
	      (mapcar #'(lambda (index) (aref v index))
		      a-triangle))
	  triangoli))

(defun RIPULISCI-MAKE-POL (lista-w-rep)
  (let* ((triangoli (second lista-w-rep))
	 (vertici-ord-ind (sort (insert-indice (first lista-w-rep))
				'vert-min
				:key 'car))
	 (v (make-array (+ 1 (length vertici-ord-ind))
			:element-type 'integer
			:adjustable t)))
    (list (new-lista vertici-ord-ind v)
	  (aggiorna-all-indici triangoli v)
	  (third lista-w-rep))))


(defun remove-omo-coord (vertex) 
  (let ((omo-coord (first vertex)))
   (mapcar #'(lambda (x) (/ x omo-coord)) (rest vertex))))

(defun normalize-omo-coord (vertex) 
  (let ((omo-coord (aref  vertex 0)))
    (dotimes (i  (length vertex) (progn (setf (aref vertex 0) 1 ) vertex))
      (setf (aref   vertex i)
	(/ (aref   vertex i) omo-coord)))))
	     
;this version avoids to triangolate a triangle convex cell.
(defun get-pol-vertices-and-faces (pol mat vertices faces)
  (let* ((n (get-npol pol))
	 (nn (array-dimension mat 0))
	 (zeros (make-list (- nn n 1) :initial-element 0))
	 (vertices-list (get-vlist pol))
	 (cells-list (get-clist pol))
	 (vert-num (hash-table-count vertices))
	 (face-num (hash-table-count faces))
	 (verts-adiac (make-hash-table :test 'equal))
	 )
    
    (cond
     ((= n 0)
      (mapcar #'(lambda (cell)
		  (let* ((cell-vertices (get-cell-vertices cell))
			 (v1 (make-array 
			      nn
			      :initial-contents           
			      (append '(1)
				      (find-vcoord (first cell-vertices) vertices-list)
				      zeros)))
			 
			 )
		    (setf (gethash (incf vert-num) vertices)
			  (rest (coerce (normalize-omo-coord (mat-vet-product-extended
					 mat v1)) 'list)))
		    
		    (setf (gethash (incf face-num) faces) 
			  (list vert-num
				vert-num
				vert-num 0 0 0))))
	      cells-list))
     ((= n 1)
      (mapcar #'(lambda (cell)
		  (let* ((cell-vertices (get-cell-vertices cell))
			 (v1 (make-array 
			      nn
			      :initial-contents           
			      (append '(1)
				      (find-vcoord (first cell-vertices) vertices-list)
				      zeros)))
			 (v2 (make-array 
			      nn
			      :initial-contents           
			      (append '(1)
				      (find-vcoord (second cell-vertices) vertices-list)
				      zeros))))
		    (setf (gethash (incf vert-num) vertices)
			  (rest  (coerce (normalize-omo-coord (mat-vet-product-extended
					 mat v1)) 'list)))
		    (setf (gethash (incf vert-num) vertices)
			  (rest (coerce (normalize-omo-coord (mat-vet-product-extended
					 mat v2)) 'list)))
		    (setf (gethash (incf face-num) faces) 
			  (list vert-num
				(1- vert-num)
				vert-num 0 0 0))
		    
		    
		    ))
	      cells-list))
     
     ((= n 2)
      (mapcar #'(lambda (cell)
		  (if (= (length (get-cell-vertices cell)) 3)
		    
		    (let* ((old-vert-num (1+ vert-num))
			   (cell-vertices (get-cell-vertices cell))
			   (cells-num (length cell-vertices))
			   (cell-vert-coord 
			    (mapcar #'(lambda (vid)
					(find-vcoord vid vertices-list))
				    cell-vertices))
			   (center (let ((x 0)(y 0))
				     (mapcar #'(lambda (xy)
						 (setf x (+ x (first xy))
						       y (+ y (second xy))))
					     cell-vert-coord)
				     (list (/ x cells-num) (/ y cells-num))))
			   (new-coord (mapcar #'(lambda (coords)
						  (list (- (first coords)
							   (first center))
							(- (second coords)
							   (second center))))
					      cell-vert-coord))
			   (phases (mapcar #'(lambda (coords)
					       (phase (complex (first coords)(second coords))))
					   new-coord))
			   (phases&coord (mapcar 'cons phases cell-vert-coord))
			   (sorted-coord 
			    (mapcar #'cdr (sort phases&coord #'< :key #'first)))
			   ; a questo punto i vertici della cella  [2D !] sono ordinati in
			   ; base alla fase rispetto al centro della cella
			   (world-coords
			    (mapcar #'(lambda (vet)  (normalize-omo-coord
					(mat-vet-product-extended
					 mat vet)))
				    (mapcar #'(lambda (coord)
						(make-array 
						 nn
						 :initial-contents           
						 (append '(1)
							 coord
							 zeros)))
					    sorted-coord))))
		      ;pone nella tavola hash tutti i vertici
		      (dotimes (i cells-num)
			(setf (gethash (incf vert-num) vertices)
			      (rest (coerce (nth i world-coords) 'list))))
		      
		      
		      ;ciclo di gestione delle adiacenze
		      
		      (dotimes (i cells-num)
			(setf (gethash (list
					(gethash (+ old-vert-num (mod (+ i 1) cells-num)) vertices)
					(gethash (+ old-vert-num i ) vertices)
					)
				       verts-adiac)
			      (list (1+  face-num) i)))
		      
		      ;pone nella tab. hash dei triangoli tutti i triangoli ottenuti da
		      ; punto centrale + facet. cells-num e' il numero di vertici.
		      (dotimes (i 1)
			(setf (gethash (incf face-num) faces) 
			      (list old-vert-num
				    (+ old-vert-num 1  i )
				    (+ old-vert-num 1 (mod (+ i 1) cells-num))
				    
				    (let* ((cella-face (gethash (list
								 (gethash (+ old-vert-num i ) vertices)
								 (gethash (+ old-vert-num (mod (+ i 1) cells-num)) vertices))
								verts-adiac))
					   (cella (first cella-face))
					   (face  (second cella-face)))
				      ; la linea di programma commentata che segue
				      ; sostituiva la successiva
				      (if (or (null cella) (null (gethash cella faces)))
					0
					;else aggiusta adiacenza all'indietro e poni questa alla cella corretta
					(progn
					  (setf (nth (+ 3 face) (gethash cella faces)) face-num)
					  cella)))
				    (let* ((cella-face (gethash (list
								 (gethash (+ old-vert-num i 1) vertices)
								 (gethash (+ old-vert-num (mod (+ i 2) cells-num))
									  vertices))
								verts-adiac))
					   (cella (first cella-face))
					   (face  (second cella-face)))
				      ; la linea di programma commentata che segue
				      ; sostituiva la successiva
				      (if (or (null cella) (null (gethash cella faces)))
					0
					;else aggiusta adiacenza all'indietro e poni questa alla cella corretta
					(progn
					  (setf (nth (+ 3 face) (gethash cella faces)) face-num)
					  cella)))
				    (let* ((cella-face (gethash (list
								 (gethash (+ old-vert-num i 2) vertices)
								 (gethash (+ old-vert-num (mod (+ i 3) cells-num)) 
								  vertices))
								verts-adiac))
					   (cella (first cella-face))
					   (face  (second cella-face)))
				      ; la linea di programma commentata che segue
				      ; sostituiva la successiva
				      (if (or (null cella) (null (gethash cella faces)))
					0
					;else aggiusta adiacenza all'indietro e poni questa alla cella corretta
					(progn
					  (setf (nth (+ 3 face) (gethash cella faces)) face-num)
					  cella)))
				    ; 0 0 
				    ))))                    
		    (let* ((old-vert-num (1+ vert-num))
			   (old-face-num (1+ face-num))
			   (cell-vertices (get-cell-vertices cell))
			   (cells-num (length cell-vertices))
			   (cell-vert-coord 
			    (mapcar #'(lambda (vid)
					(find-vcoord vid vertices-list))
				    cell-vertices))
			   (center (let ((x 0)(y 0))
				     (mapcar #'(lambda (xy)
						 (setf x (+ x (first xy))
						       y (+ y (second xy))))
					     cell-vert-coord)
				     (list (/ x cells-num) (/ y cells-num))))
			   (new-coord (mapcar #'(lambda (coords)
						  (list (- (first coords)
							   (first center))
							(- (second coords)
							   (second center))))
					      cell-vert-coord))
			   (phases (mapcar #'(lambda (coords)
					       (phase (complex (first coords)(second coords))))
					   new-coord))
			   (phases&coord (mapcar 'cons phases cell-vert-coord))
			   (sorted-coord 
			    (mapcar #'cdr (sort phases&coord #'< :key #'first)))
			   
			   
			   ; a questo punto i vertici della cella [ 2D !] sono ordinati in
			   ; base alla fase rispetto al centro della cella
			   
			   (world-coords
			    (mapcar #'(lambda (vet)   (normalize-omo-coord
					(mat-vet-product-extended
					 mat vet)))
				    (mapcar #'(lambda (coord)
						(make-array 
						 nn
						 :initial-contents           
						 (append '(1)
							 coord
							 zeros)))
					    sorted-coord)))
			   (world-center  (normalize-omo-coord (mat-vet-product-extended
					  
					  mat 
					  (make-array 
					   nn
					   :initial-contents           
					   (append '(1)
						   center
						   zeros)))))
			   )
		      ;Pone nella tab. hash le coordinate del centro della cella
		      (setf (gethash (incf vert-num) vertices)
			    (rest (coerce world-center 'list)))
		      ;pone nella tavola hash tutti i vertici
		      (dotimes (i cells-num)
			(setf (gethash (incf vert-num) vertices)
			      (rest (coerce (nth i world-coords) 'list))))
		      
		      
		      ;ciclo di gestione delle adiacenze
		      
		      (dotimes (i cells-num)
			(setf (gethash (list
					(gethash (+ old-vert-num 1 (mod (+ i 1) cells-num)) vertices)
					(gethash (+ old-vert-num 1  i ) vertices)
					)
				       verts-adiac)
			      (list (1+ (+ i face-num)) 1))
			)
		      
		      ;pone nella tab. hash dei triangoli tutti i triangoli ottenuti da
		      ; punto centrale + facet. cells-num e' il numero di vertici.
		      (dotimes (i cells-num)
			(setf (gethash (incf face-num) faces) 
			      (list old-vert-num
				    (+ old-vert-num 1  i )
				    (+ old-vert-num 1 (mod (+ i 1) cells-num))
				    
				    (+ old-face-num (mod (- i 1) cells-num))
				    
				    (let* ((cella-face (gethash 
						       (list
							(gethash (+ old-vert-num 1  i ) vertices)
							(gethash (+ old-vert-num 1 (mod (+ i 1) cells-num))
								 vertices))
						       verts-adiac))
					  (cella (first cella-face))
					  (face  (second cella-face)))
				      ; la linea di programma commentata che segue
				      ; sostituiva la successiva
				      (if (or (null cella) (null (gethash cella faces)))
					0
					;else aggiusta adiacenza all'indietro e poni questa alla cella corretta
					(progn
					  (setf (nth (+ 3 face)  (gethash cella faces)) face-num)
					  cella)))
				    (+ old-face-num (mod (+ i 1) cells-num))))))))
	      
	       cells-list)))))



;-----------------------------------------------------------------

(defun EXPORT-POL-COMPLEX (pol-complex 
                           &rest streams
                           )
  (if (or (not streams)
          (= 1 (length streams)))
    ; then streams stands for "export it on a file!"
    (with-open-file (outfile (or (first streams) 
                                 (prompt-for-a-new-file))
                             :direction :output
                             :if-exists :supersede
                             :if-does-not-exist :create)
      (export-to-stream pol-complex outfile))
    ; then streams stands for "export it on standard output!"
  ;  (with-open-stream (outfile *standard-output*)
  ;      (export-to-stream pol-complex outfile))
     (let ((outstring (make-string-output-stream)))
	(export-to-stream pol-complex outstring)
	(format t "~A" (get-output-stream-string outstring)))
    ))


  

(defun EXPORT-TO-STREAM (pol-complex out-stream)
  (if (symbol-equal (get-pol-complex-rep pol-complex) 'on-faces)
    (setf pol-complex (coerce-pol-complex pol-complex 'on-vertices)))
  (let* ((n (get-npol-complex pol-complex))
	 (id-mat (mat-identity (1+ n)))
	 (vertices (make-hash-table))
	 (faces (make-hash-table)))
    
      (mapcar #'(lambda (node)
		  (get-pol-complex-vertices-and-faces node id-mat vertices faces))
	      pol-complex)
      
      
      
      (let ((clear-w-rep (ripulisci (crea-w-rep vertices faces))))
	
	;
	;  Scrittura del file.
	;
	
	(format out-stream 
		"  -200     ; Ascii polyhedron code
;
")
	(prin1 (+ 3 (length (first clear-w-rep))) out-stream)
	(format out-stream 
		"     ; No. of polyhedron points
;
; Points list...
")
	(mapcar #'(lambda (vertice)
		    (mapcar #'(lambda (number)
				(format out-stream "~15,15F " number ))
			    (subseq (append vertice '(0 0 0)) 0 3))
		    (terpri out-stream))
		(first clear-w-rep))
	
	(anti-bomb-minerva out-stream (gethash 1 vertices))
	
	(princ
	 ";
;
" out-stream)
	(prin1 (length (second clear-w-rep)) out-stream)
	(princ "    ; No. of polyhedron triangles
;
; Triangles list...
" out-stream)
	
	(mapcar #'(lambda (triangolo)
		    (mapcar #'(lambda (number)
				(format out-stream "~A  " number))
			    triangolo)
		    (terpri out-stream))
		(second clear-w-rep))  
	
	
	(values)
	)))
