; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package Kernel, Module join                               ;                                
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: V. Pascucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'join)                              ;module name

(inpackage kernel)   ;package name
(use-package 'lisp-extension)
(import 'plasm::join-pol-complex)
(import 'plasm::unmake-pol-complex)
(export '(join-pol-complex
          unmake-pol-complex))

(defun join-pol-complex (&rest pol-complexes)
  (let (vertices 0-skeletons v-indices)
    (dolist (pol-complex pol-complexes)
      (push (skeleton pol-complex 0) 0-skeletons))
    (dolist (0-skeleton   0-skeletons)
      (setf vertices 
            (append vertices
                    (collect-0-skeleton-vertices 0-skeleton))))
    (dotimes (i (length vertices))
      (push i v-indices))
    (build-pol-complex vertices (list v-indices) '((0)))))


(defun join-pol-complex (&rest pol-complexes)
  (let (vertices 0-skeletons v-indices)
    (dolist (pol-complex pol-complexes)
      (push (skeleton pol-complex 0) 0-skeletons))
    (dolist (0-skeleton   0-skeletons)
      (setf vertices 
            (append vertices
                    (collect-0-skeleton-vertices
                     (coerce-pol-complex 0-skeleton 'on-vertices)))))
    (dotimes (i (length vertices))
      (push i v-indices))
    (build-pol-complex vertices (list v-indices) '((0)))))



(defun collect-0-skeleton-vertices (0-skeleton &optional previous-mat)
  (if (polyhedron? 0-skeleton)
    (list (cdr (coerce  (mat-get-a-column (mat-inv  previous-mat) 0) 'list)))
    (apply 'concatenate
           (cons 'list
                 (mapcar #'(lambda (pol-instance)
                             (collect-0-skeleton-vertices 
                              (get-node-pol-complex  pol-instance)
                              (or (and previous-mat
                              (mat-product-extended  previous-mat
                               (get-node-matrix pol-instance) t))
                                  (get-node-matrix pol-instance))))
                         0-skeleton)))))


(defun collect-0-skeleton-vertices (0-skeleton &optional previous-mat)
  (if (polyhedron? 0-skeleton)
    (list (cdr (coerce  (mat-get-a-column   previous-mat 0) 'list)))
    (apply 'concatenate
           (cons 'list
                 (mapcar #'(lambda (pol-instance)
                             (collect-0-skeleton-vertices 
                              (get-node-pol-complex  pol-instance)
                              (or (and previous-mat
                                       (mat-product-extended  
                                        previous-mat
                                        (get-node-matrix pol-instance) 
                                         t))
                                  (get-node-matrix pol-instance))))
                         0-skeleton)))))

;-------------------------------------------------
; Prende in ingresso un complesso poliedrale e
; restituisce in uscita una lista di tre liste:
;  - lista dei vertici;
;  - lista delle celle;
;  -lista dei poliedri.
; Ogni vertice e` una lista di numeri.
; Ogni cella e` una lista di indici di vertice 
;  (secondo l'ordine della lista precedente.
; Ogni poliedro e` una lista di indici di 
; cella (secondo l'ordine della lista precedente).
;
; Input : polyhedral complex;
;
; Output: lista specificata sopra.
;------------------------------------------------- 
(defun unmake-pol-complex (pol-complex)
  (let  ((pol-complex  (coerce-pol-complex pol-complex 'on-vertices)))
    (execute-unmake-pol-complex pol-complex 
                                (mat-identity (1+ (get-npol-complex  pol-complex))))))

;-------------------------------------------------
; Prende in ingresso un complesso poliedrale 
; rappresentato a vertici con tutte le matrici
; di trasformazione pari all'identita` e
; restituisce in uscita una lista di tre liste:
;  - lista dei vertici;
;  - lista delle celle;
;  -lista dei poliedri.
; Ogni vertice e` una lista di numeri.
; Ogni cella e` una lista di indici di vertice 
;  (secondo l'ordine della lista precedente.
; Ogni poliedro e` una lista di indici di 
; cella (secondo l'ordine della lista precedente).
;
; Input : complesso poliedrale;
;         dimensione dello spazio di immersione;
;
; Output: lista specificata sopra.
;------------------------------------------------- 
(defun execute-unmake-pol-complex (pol-complex transf-matrix)
  (if (polyhedron? pol-complex)
    (unmake-pol pol-complex transf-matrix)
    (reduce #'append-mkpol-args (mapcar #'execute-unmake-pol-complex 
                                      (mapcar #'get-node-pol-complex pol-complex)
                                      (mapcar #'(lambda (pol-instance)
                                                  (mat-product-extended
                                                   transf-matrix 
                                                   (get-node-matrix pol-instance)
                                                   t))
                                                  pol-complex)))))

;------------------------------------------------- 
; Prende in ingresso due liste che costituiscono ognuna
; un set di tre parametri di ingresso della funzione
; MKPOL  e restituisce un unico set di tre parametri
; adatto alla MKPOL e tale che il risultato creato
; dalla MKPOL sia l'unione dei risultati che la
; stessa avrebbe generato se applicata ai due
; set di parametri iniziali.
;------------------------------------------------- 
(defun append-mkpol-args (arg1 arg2)
  (let* ((vert1 (first arg1))
         (vert2 (first arg2))
         (length-vert1 (length vert1))
         (cells1 (second arg1))
         (cells2 (second arg2))
         (length-cells1 (length cells1))
         (pols1 (third arg1))
         (pols2 (third arg2))
         (verts (append vert1 vert2))
         (cells (append cells1 
                        (mapcar #'(lambda (cell)
                                    (mapcar #'(lambda (vert-index) 
                                                (+ vert-index length-vert1)) cell))
                                cells2)))
         (pols (append pols1 (mapcar #'(lambda (pol)
                                         (mapcar #'(lambda (cell-index)
                                                     (+ cell-index length-cells1)) pol))
                                     pols2))))
    (list verts cells pols)))

;(append-mkpol-args 
; '((( 1  2  3)( 4  5  6)( 7  8  9))((1 2)(2 0)(0 1))((1 2)(2 0))) 
; '(((10 11 12)(13 14 15)(16 17 18))((1 2)(2 0)(0 1))((1 2)(2 0))))
         
        
;-------------------------------------------------
; Prende in ingresso un  poliedro 
; rappresentato a vertici e
; restituisce in uscita una lista di tre liste:
;  - lista dei vertici;
;  - lista delle celle;
;  -lista dei poliedri.
; Ogni vertice e` una lista di numeri.
; Ogni cella e` una lista di indici di vertice 
;  (secondo l'ordine della lista precedente.
; Ogni poliedro e` una lista di indici di 
; cella (secondo l'ordine della lista precedente).
;
; Input : complesso poliedrale;
;         dimensione dello spazio di immersione;
;
; Output: lista specificata sopra.
;------------------------------------------------- 
(defun unmake-pol (pol  transf-matrix )
  (let* ((vertices (get-vlist pol))
         (cells (get-clist pol))
         (new-vertices (mapcar #'(lambda (vertex)
                                   (cdr (coerce 
                                    (mat-vet-product-extended
                                     transf-matrix 
                                     (coerce (append '(1) 
                                                     (second vertex))
                                             'array) t)
                                    'list)))
                               vertices))
         (new-cells (mapcar #'(lambda (cell)
                                (mapcar  #'(lambda (v-symbol)
                                             (position v-symbol vertices :key 'first))
                                         (get-cell-vertices cell)))
                            cells))
         (polyhedron (from-to 0 (1- (length cells)))))
    (list new-vertices new-cells (list polyhedron))))
    

;(unmake-pol-complex (build-pol-complex-cube 2))
;(unmake-pol-complex (build-pol-complex-cube 3))


;(unmake-pol-complex
; (build-pol-complex '((1 0 1) (1 2 3) (2 3 4)) '((0 1) (1 2) (2 0)) '((0))))

;(unmake-pol-complex
; (build-pol-complex '((0 0 0) (0 0 1) (0 1 0) (1 0 0) (0 1 1 )) '((0 1 2 3) (1 2 3 4) ) '((0 1)) ))

;(unmake-pol-complex
; (build-pol-complex '((0 0 0) (0 0 1) (0 1 0) (1 0 0) (0 1 1 ) (1 0 1)) '((0 1 2 3) (1 2 3 4) (2 3 4 5) ) '((0 1) (2))))
