;-*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Mapping su complessi : Valutazione trasf. affini          ;               
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: G. Proietti                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;Aggiunta: Valutazione di complessi 'on-vertices
; ( 29/09/93)

; (23/4/94) La funzione Pol-Eval e' stata modificata per tener conto dei
;           cambiamenti di segno inseriti dalla funzione di normalizzazione
;           dei coefficienti degli iperpiani.

(inpackage kernel)

(provide 'evalmap)

(require 'matlib)
(require 'hyperp)
(require 'lisp_ext)

(use-package 'matlib)
(use-package 'hyperplanes)
(use-package 'lisp-extension)

(export '(
	  Push-Down-Transf-Pol-Complex
	  Push-Down-Transf-Pol-Complex-2
	  Put-Clist))

;
;Input :
;        Pol-Complex  :Il complesso poliedrale 
;  (opt.)       mat   :per compatibilita' verso il basso
;Output : il complesso poliedrale rivalutato.

(defun Push-Down-Transf-Pol-Complex (obj &optional (check T) mat)
  (if check 
    (if (= (get-npol-complex obj) (get-dpol-complex obj))
      (Push-Down-Transf-Pol-Complex-2 obj mat)
      ;else
      (progn
	(format t "Uncorrect input for PUSH-DOWN-TRANSF-POL-COMPLEX.~%")
	(format t "Polyhedra dimension minor than Embedding Space Dimension.~%")
	(format t "Use nil check setting to force operation execution. In this case~%")
	(format t "a probably uncorrect complex representation will be produced.~%")
	nil)
      )
   ;else
    (Push-Down-Transf-Pol-Complex-2 obj mat)
    )
  )

(defun Push-Down-Transf-Pol-Complex-2 (Pol-Complex &optional mat)
  (if (lisp-extension::symbol-equal (get-pol-complex-rep Pol-Complex) 'on-faces)
    (Push-Down-Transf-Pol-Complex-On-Faces Pol-Complex mat)
    ;else
    (Push-Down-Transf-Pol-Complex-On-Vertices Pol-Complex mat)
    )
  )

;
;Input :
;        obj  :Il complesso poliedrale rappresentato a facce
;  (opt) mat  :Matrice di trasformazione da applicare al complesso
;
;Output : il complesso poliedrale rivalutato.

(defun Push-Down-Transf-Pol-Complex-On-Faces (obj &optional mat)
  (cond ( (polyhedron?      obj) (Pol-Eval obj mat))
	( (polyhedra-complex? obj)
	  (mapcar  #'(lambda (nodo) (Push-Down-Transf-Pol-Complex-On-Faces nodo mat)) obj)
	  )
	( (polyhedral-node? obj)
	  (progn
	    (if (null mat) 
	      (setq mat (get-node-matrix obj))
	      ;else
	      (setq mat  (mat-product (get-node-matrix obj) mat) )
	      )
	    (make-pol-node 
	     (Push-Down-Transf-Pol-Complex-On-Faces (get-node-pol-complex obj) mat)
	     (mat-identity (car (array-dimensions mat)))
	     (get-node-id obj)
	     )
	    )
	  )
	)
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; ritorna le disequazioni che definiscono una cella
;;; sotto forma di lista associativa (simbolo (coefficienti))

(defun get-disequations (cell pol)
  (let ((listaeq    nil)
	(facce      nil)
	(segno      nil))
    (setq facce (get-cell-faces cell))
    (dotimes (i (length facce) listaeq)
      (setq segno (cdr (nth i facce)) )
      (setq listaeq (cons (list 
			   (car (nth i facce)) 
			   (get-fcoord (car (nth i facce)) pol segno ))
			  listaeq ) )
      )
    )
  )


;----------------------------------------------------------------------
;Valuta un poliedro rappresentato tramite facce in base alla matrice
;passata in mat.
;
;(defun Pol-Eval (pol mat)
;  (let* ( (facce (get-flist pol))
;          (celle (get-clist pol))
;          newface
;          Mod-Signs
;          newpol
;        )
;    (print 'zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz)
;    (pprint facce)
;    ;ricalcolo delle facce in base alla matrice specificata
;    (if (not (null mat))
;      (setq facce (mapcar 
;                   #'(lambda (faccia) 
;                       (progn
;                         (setq newface 
;                               (normalize-parameters
;                                (coerce 
;                                 (matlib::vet-mat-product 
;                                  (coerce (second faccia) 'vector) 
;                                  mat ) 'list)))
;                         
;                         (list 
;                          (first faccia)
;                          (mapcar #'(lambda (par) (if (<= (abs par) 1e-10) 
;                                                    0
;                                                    par
;                                                    ))
;                                  (car newface))
;                          (cdr newface)
;                          )))
;                   facce
;                   )) 
;      )
;    (pprint facce)
;
;    ;Ora in facce ci sono le facce nella forma (symbol coeffs sign) 
;
;    ;Ricavo la lista delle facce che hanno cambiato segno rospetto a prima
;    (setq Mod-Signs (remove nil 
;                            (mapcar #'(lambda (face)
;                                        (if (equal (last face) '(-))
;                                          (car face)
;                                          ))
;                                    facce))
;          )
;    (pprint facce)
;
;    ;Depuro la lista dai segni
;    (setq facce (mapcar #'(lambda (face)
;                            (list (first face)
;                                  (second face)
;                                  ))
;                        facce))
;    (pprint facce)
;
;    ;Aggiorno segni delle facce nella lista celle
;    (mapcar #'(lambda (cella)
;                (setf (convex-cell-faces cella)
;                      (invert-Faces-Signs Mod-Signs
;                                          (convex-cell-faces cella)
;                                          )))
;            celle
;            )
;    (pprint facce)
;                                          
;    (setq newpol (make-empty-polyhedron 'on-faces))
;    (setq newpol (put-clist celle newpol))
;    (set-npol newpol (get-npol pol))
;    (pprint 'LLLLLLLLLLLLLLLL)
;    (put-faces (print facce) newpol)
;    (set-dpol newpol (get-dpol pol))
;    (set-#kd  newpol)
;    (set-#k0  newpol)
;    (set-#F  newpol)
;    (set-#C  newpol)
;    (setf (kernel::polyhedron-KIND-OF-BRICK newpol)  'EXTRUSION::CONVEX-CELLS)
;    (setf (kernel::polyhedron-REPRESENTATION newpol) 'EXTRUSION::ON-FACES)
;    (io:printpol newpol)
;    newpol
;    )
;  
;  )


;----------------------------------------------------------------------
;Valuta un poliedro rappresentato tramite facce in base alla matrice
;passata in mat.
;
(defun Pol-Eval (pol mat)
  (let* ((facce (get-flist pol))
	 (celle (get-clist pol))
	 (new-faces  (new-faces-table))
	 equal-faces 
	 newface
	 Mod-Signs
	 newpol
	 )
    ;ricalcolo delle facce in base alla matrice specificata
    (if (not (null mat))
      (setq facce (mapcar 
		   #'(lambda (faccia) 
		       (progn
			 (setq newface 
			       (normalize-parameters
				(coerce 
				 (matlib::vet-mat-product 
				  (coerce (second faccia) 'vector) 
				  mat ) 'list)))
			 
			 (list 
			  (first faccia)
			  (mapcar #'(lambda (par) (if (<= (abs par) 1e-10) 
						    0
						    par
						    ))
				  (car newface))
			  (cdr newface)
			  )))
		   facce
		   )) 
      )
    ;Ora in facce ci sono le facce nella forma (symbol coeffs sign) 

    ;Ricavo la lista delle facce che hanno cambiato segno rospetto a prima
    (setq Mod-Signs (remove nil 
			    (mapcar #'(lambda (face)
					(if (equal (last face) '(-))
					  (car face)
					  ))
				    facce))
	  )
    ;Depuro la lista dai segni
    (setq facce (mapcar #'(lambda (face)
			    (list (first face)
				  (second face)
				  ))
			facce))
    (dolist (face facce)
      (let ((old-face (gethash (second face) new-faces)))
	(if old-face 
	    (push (cons  (first face) old-face) equal-faces)
	  (setf (gethash (second face) new-faces)
	    (first face)))))

    ;Aggiorno segni delle facce nella lista celle
    (mapcar #'(lambda (cella)
		(setf (convex-cell-faces cella)
		      (invert-Faces-Signs Mod-Signs
					  (convex-cell-faces cella)
					  ))
		(setf (convex-cell-faces cella)
		  (nsublis equal-faces (convex-cell-faces cella)))
		(setf (convex-cell-vertices cella)
		  (nsublis equal-faces (convex-cell-vertices cella))))
	    celle
	    )
   
    
    (setq newpol (make-empty-polyhedron 'on-faces))
    (setq newpol (put-clist celle newpol))
    (set-npol newpol (get-npol pol))
    (put-faces new-faces  newpol t)
    (set-dpol newpol (get-dpol pol))
    (set-#kd  newpol)
    (set-#k0  newpol)
    (set-#F  newpol)
    (set-#C  newpol)
    (setf (kernel::polyhedron-KIND-OF-BRICK newpol)  'EXTRUSION::CONVEX-CELLS)
    (setf (kernel::polyhedron-REPRESENTATION newpol) 'EXTRUSION::ON-FACES)
    newpol
    )
  
  ) 


(defun Invert-Faces-Signs (Faces-to-invert Faces)
  (mapcar #'(lambda (face)
	      (if (find (car face) Faces-to-invert)
		(cons (car face)
		      (if (equal (cdr face) '+)
			'-
		       ;else
			'+)
		      )
		face)
	      )
	  Faces
	  )
  )

;------------------------------------------------------------------
;Ritorna un poliedro a cui sono state aggiunte le celle nella lista
(defun put-clist (clist pol)
  (dolist (cell clist pol)
    (setq pol (put-cell cell pol))
    )
  )

;
;Input :
;        obj  :Il complesso poliedrale rappresentato a vertici
;  (opt) mat  :Matrice di trasformazione da applicare al complesso
;
;Output : il complesso poliedrale rivalutato.

(defun Push-Down-Transf-Pol-Complex-On-Vertices (obj &optional mat)
  (cond ( (polyhedron?      obj) (Pol-Eval-On-Vertices obj mat))
	( (polyhedra-complex? obj)
	  (mapcar  #'(lambda (nodo) (Push-Down-Transf-Pol-Complex-On-Vertices nodo mat)) obj)
	  )
	( (polyhedral-node? obj)
	  (progn
	    (if (null mat) 
	      (setq mat (get-node-matrix obj))
	      ;else
	      (setq mat  (mat-product-extended mat (get-node-matrix obj)) )
	      )
	    (make-pol-node 
	     (Push-Down-Transf-Pol-Complex-On-Vertices (get-node-pol-complex obj) mat)
	     (mat-identity (car (array-dimensions mat)))
	     (get-node-id obj)
	     )
	    )
	  )
	)
  )

;----------------------------------------------------------------------
;Valuta un poliedro rappresentato tramite vertici in base alla matrice
;passata in mat.
;
(defun Pol-Eval-On-Vertices (pol mat)
  (let* ( (Vertici (get-vlist pol))
	  (celle (get-clist pol))
	  newpol
	  )
    ;ricalcolo i vertici in base alla matrice specificata
    (if (not (null mat))
      (setq vertici (mapcar 
		   #'(lambda (vertice) 
		      (list 
		       (first vertice)
		       (cdr (coerce (matlib::mat-vet-product 
				mat
				(coerce (cons 1 (second vertice)) 'vector) 
				) 
			       'list))
		       )
		      )
		   vertici
		   )
	    )
      )
    (setq newpol (make-empty-polyhedron 'on-vertices))
    (setq newpol (put-clist celle newpol))
    (set-npol newpol (get-npol pol))
    (put-vlist vertici newpol)
    (set-dpol newpol (get-dpol pol))
    (set-#kd  newpol)
    (set-#k0  newpol)
    (set-#F  newpol)
    (set-#C  newpol)
    (setf (kernel::polyhedron-KIND-OF-BRICK newpol)  'EXTRUSION::CONVEX-CELLS)
    newpol
    )
  )


;---------------------------------------------
; Questa funzione prende un complesso poliedrale le cui
; trasformazioni siano tutte identita' e 
; ne altera le matrici di trasformazione, sostituendo
; a tutte le matrici delle matrici identita' di dimensione
; pari a NewDim+1.
; Att.! Non controlla che le matrici siano delle identita'
; e quindi non deve essere applicato a complessi aventi
; tali matrici non identita', pena l' "appiattimento" di
; tutte le trasformazioni.

(defun Modify-Dimension-Of-Pol-Complex (PolCompl NewDim)
  (let ( (LastDim (1- (car (array-dimensions
			(Get-Node-Matrix (car PolCompl))))))
	 )
    (if (/= (1- LastDim) NewDim)
      (Modify-Matrix-Dimensions PolCompl (1+ NewDim))
      )
    )
  )

(defun Modify-Matrix-Dimensions (Obj NewDim)
    (cond 
     ( (polyhedron?      obj) (progn 
				(set-dpol obj (1- NewDim))
				obj
				))
     ( (polyhedra-complex? obj)
       (mapcar  #'(lambda (nodo) (Modify-Matrix-Dimensions nodo NewDim)) obj)
       )
     ( (polyhedral-node? obj)
       (progn
	 (make-pol-node 
	  (Modify-Matrix-Dimensions (get-node-pol-complex obj) NewDim)
	  (make-idmat NewDim)
	  (get-node-id obj)
	  )
	 )
       )
     )
    )
