;-*- package:kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Mapping su complessi : Rimozione parti degeneri           ;               
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: G. Proietti                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;(1/10/93)

;Migliorata la gestione della ricostruzione delle adiacenze.

(inpackage kernel)

(provide 'Garbage)

(defun Remove-Degenerate-Cells-Pol-Complex (obj)
  (cond ( (polyhedron?      obj) 
          (Remove-Degenerate-Cells-And-Rebuild-Adiacens obj))
        ( (polyhedra-complex? obj)
          (mapcar  #'(lambda (nodo) 
                       (Remove-Degenerate-Cells-Pol-Complex nodo)) obj)
          )
        ( (polyhedral-node? obj)
          (progn
            (make-pol-node 
             (Remove-Degenerate-Cells-Pol-Complex (get-node-pol-complex obj))
             (get-node-matrix obj)
             (get-node-id obj)
             )
            )
          )
        )
  )


(defun Remove-Degenerate-Cells-And-Rebuild-Adiacens (pol)
  (Clear-Adiacens pol)

  (let* ((dim      (Get-npol pol)) ;Dimensione intrinseca poliedro
         (DimSpace (Get-dpol pol)) ;Dimensione spazio d'immersione
         (vertlist (Get-vlist pol))
         (CellsHashTab (kernel::polyhedron-convex-cells pol))
         CellsList
         NewVerts

;Aggiunta durante il porting su RS6000 (28/1/94)
         (OldVerts (copy-tree vertlist))

        )
    ; ***** Eliminazione delle celle degeneri *****
    (maphash #'(lambda (x y)
                (if (not (Test-Cell y vertlist dim))
                  (remhash x CellsHashTab)
                  ))
             CellsHashTab)

    ; ***** Rimappatura vertici ******
    (setq NewVerts 
          (multiple-value-list 
           (kernel::Lex-Purge-Equal-Verts
            (kernel::Lex-Sort-with-Labels vertlist))
           )
          ) 

    (Remap-Vertices CellsHashTab (second NewVerts))

    ; ***** Ricostruzione adiacenze *****
    (setq CellsList (get-clist Pol))
    (Rebuild-adiacens CellsList 'on-vertices)
    (clrhash (kernel::polyhedron-Convex-Cells pol)) 
    (put-clist CellsList Pol)

;agggiunta (28/1/94)
    (setf (kernel::polyhedron-vlist pol) OldVerts)
    pol
;
    )
  )


(defun Clear-Adiacens (Pol)
  (maphash 
   #'(lambda (x y)
       (setq l (length (kernel::Convex-Cell-adjacent-convex-cells y)))
       (setf (kernel::Convex-Cell-adjacent-convex-cells y)
             (make-list l :initial-element 0))
       )
   (kernel::polyhedron-convex-cells Pol)
   )
  )


(defun Remap-Vertices (CellsHashTab VertsAList)
  (let ( VertSeq  ;Lista vertici di una cella
         FaceSeq  ;Lista facce di una cella
         )
    (maphash #'(lambda (x y)
                 (setq VertSeq (kernel::Convex-Cell-Vertices y))
                 (setq VertSeq (multi-substitute VertsAList VertSeq))
                 (setf (kernel::Convex-Cell-Vertices y) VertSeq)

                 (setq FaceSeq (kernel::Convex-Cell-Faces y))
                 (setq FaceSeq (mapcar 
                                #'(lambda (x)
                                   (multi-substitute VertsAList x))
                                FaceSeq)
                       )
                 (setf (kernel::Convex-Cell-Faces y) FaceSeq)
                 )
             CellsHashTab
             )
    )
  )

(defun Multi-Substitute (Subs Seq)
  (dolist (el Subs Seq) 
    (setq Seq 
          (substitute (second el) (first el) Seq)))
  )

; Versione vecchia non funziona correttamente

;(defun Remove-Degenerate-Cells-And-Rebuild-Adiacens (pol)
;  ;Attenzione : Le celle devono essere simplessi.
;  (let ((dim      (Get-npol pol)) ;Dimensione intrinseca poliedro
;        (DimSpace (Get-dpol pol)) ;Dimensione spazio d'immersione
;        (vertlist (Get-vlist pol))
;        (CellsHashTab (kernel::polyhedron-convex-cells pol))
;        (CellsAList (mapcar
;                     #'(lambda (TheCell) 
;                         (list (CONVEX-CELL-IDENTIFIER TheCell)
;                               TheCell
;                               )) 
;                     (get-clist pol))
;                    )
;        NewVerts
;        )
;    ;(print CellsAList)
;    ;Provvedo, adesso, ad individuare i vertici ripetuti piu' volte,
;    ; ottenendo una lista ripulita dei doppioni ed una lista
;    ; associativa che descrive le corrispondenze con i vertici
;    ; rimossi.
;    (setq NewVerts 
;          (multiple-value-list 
;           (kernel::Lex-Purge-Equal-Verts
;            (kernel::Lex-Sort-with-Labels vertlist))
;           )
;          ) 
;    ;Pongo ad essere la lista ripulita dei vertici come la lista dei
;    ;vertici del poliedro
;    (setf (kernel::polyhedron-vlist Pol) (first NewVerts))
;    ;Ripulisco le celle ad tutti i vertici eliminati, sostituendoli con
;    ; le uniche copie mantenute.
;    (Reassign-Verts-To-Cells CellsAList (second NewVerts))
;
;    ;Ora scandisco tutte le celle del poliedro ed eseguo il test sulla
;    ;loro dimensione intrinseca. Se questa e' inferiore a quella
;    ;intrinseca del poliedro, esse vengono scartate,
;    (dolist (TheCell CellsAList)
;      (if (not (Test-Cell (second TheCell) vertlist dim))
;        (progn
;          (pprint TheCell)
;          (print CellsAList)
;          (Remove-a-Cell-Rebuilding-Adiacens (car TheCell) CellsHashTab CellsAList)
;          (print CellsAList)
;          )
;        )
;      )
;    ;Ora ho solo celle a dimensione pari a quella intrinseca del poliedro
;
;    ;Ricostruisco le adiacenze tra' le celle, utilizzando i nuovi
;    ; vertici.
;    
;    ;Da eliminare, perche trovato il bug.
;    ;(dolist (x Cells)
;    ;  (setf (kernel::Convex-Cell-Sign x) '+))
;
;    ;(print CellsAList)
;
;    ;Poni le celle nel poliedro, eliminando tutte le vecchie celle.
;    pol
;    )
;  )
  

(defun Remove-a-Cell-Rebuilding-Adiacens (CellID CellsHashTab CellsAList)
  (let* ( (ToRemove (gethash CellID CellsHashTab))
          (AdiacensID
           (remove 0
                   (kernel::convex-cell-ADJACENT-CONVEX-CELLS ToRemove))
           )
          AdiacentCells
          )

    (setq AdiacentCells (mapcar #'(lambda (x) 
                                    (second (assoc 
                                             x 
                                             CellsAList
                                             :test 'lisp-extension::Symbol-equal 
                                             ))) 
                                AdiacensID ))   
     (pprint AdiacentCells)
    ;Rimuovi adiacenza con CellID
    (mapcar #'(lambda (Cell)
                (setf (kernel::convex-cell-ADJACENT-CONVEX-CELLS Cell)
                      (substitute 0 CellID 
                                  (kernel::convex-cell-ADJACENT-CONVEX-CELLS Cell)
                                  ))) AdiacentCells)
    (pprint AdiacentCells)
    (kernel::rebuild-adiacens AdiacentCells 'on-vertices)
    (pprint AdiacentCells)
    (remhash CellID CellsHashTab)
    )
  )

 

;--------------------------------------------------------
; Esegue un test sulla cella convessa passata in Cell , utilizzando v-list
; per ricavare i vertici, per verificare che la sua dimensione non sia
; inferiore alla dimensione specificata in Pol-Dim. Se le due dimensioni
; sono uguali, ritorna la cella, altrimenti ritorna nil.
(defun Test-Cell (Cell v-list Pol-Dim)
  (let* ( (vertici (mapcar #'(lambda (x) (second (assoc x v-list)))
                           (get-cell-vertices Cell)
                           ))
          (1Vert  (first Vertici))
          (TraslVerts (mapcar #'(lambda (x)
                                 (mapcar #'(lambda(y z) 
                                            (- y z)
                                            ) x 1Vert)
                                 ) (cdr Vertici)
                              )
                      )
          )
    (if (= Pol-Dim (new-pl::calc-matrix-rank 0 (1- (length TraslVerts))
                                             0 (1- (length (car TraslVerts)))
                                             TraslVerts))
      ;then
      (progn
        Cell)
        )
      )
  )


(defun ReduceVerts (Pol)
  (let ((Vertici (get-vlist Pol))
        NewVerts
        )
    (setq NewVerts 
          (multiple-value-list 
           (kernel::Lex-Purge-Equal-Verts
            (kernel::Lex-Sort-with-Labels Vertici))
           )
          )
    (setf (kernel::polyhedron-vlist Pol) (first NewVerts))
    (Reassign-Verts-To-Cells Pol (second NewVerts))
    )
  )

(defun Reassign-Verts-To-Cells (Cells Verts-AList)
  (dolist (Cell Cells)
    (Reassign-Verts-To-Single-Cell (second Cell) Verts-AList)
    )
  )


(defun Reassign-Verts-To-Single-Cell (Cell Verts-AList)
  (setf (kernel::Convex-Cell-Faces Cell)
         (mapcar #'(lambda (x)
                    (mapcar #'(lambda (y)
                               (let ( (Nuovo (assoc y Verts-AList) ))
                                 (if Nuovo
                                   (second Nuovo)
                                   y
                                   )
                                 )
                               ) x )) (kernel::Convex-Cell-Faces Cell) )
         )
  (setf (kernel::Convex-Cell-Vertices Cell)
        (mapcar #'(lambda (y)
                   (let ( (Nuovo (assoc y Verts-AList)) )
                     (if Nuovo
                       (second Nuovo)
                       y
                       )
                     )
                   ) (kernel::Convex-Cell-Vertices Cell) )
        )
  Cell
  )

;--------------------------------
;Funziona
;
;(defun test (espr)
;  (let ( (u 3)
;         (v 5)
;         temp)
;    (setf temp (eval `(function (lambda ,'(u v) ,espr))))
;    (apply temp '(3 5))
;    )
;  )
;--------------------------------
