; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Mapping su complessi : Main function                      ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: G. Proietti                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; (16/4/94) Con i nuovi parametri la dimensione del nuovo spazio di immersione
;           e' stata posta pari al numero di funzioni fornite, e non piu' pari
;           al numero di funzioni -1 (in precedenza la lista di funzioni
;           conteneva anche la lista dei parametri).
;
; (Luglio 1995) Modifica per la costruzione delle adiacenze esterne ad una
;               cella settando un puntatore al poliedro a cui la cella
;               adiacente appartiene. Per questo motivo e' richiesto
;               che le funzioni che modificano i poliedri dopo questo modulo
;               lavorino in maniera distruttiva direttamente sulle locazioni
;               di memoria dei poliedri stessi. In pratica non si devono
;               effettuare delle copie delle strutture, ma agire direttamente 
;               su esse. Questo serve per mantenere la coerenza tra i 
;               puntatori settati per le adiacenze esterne ed i poliedri
;               modificati.
;               Le adiacenze esterne vengono settate nel modulo "extadiac.lsp".
;
;                                                 
;  by Marco Conti
;
;;;;;;;;;;;;;;;;;

 
(inpackage kernel)

(provide 'mapping)

(export '(Map-On-Pol-Complex) )

(require 'N_Triang)
(require 'NewMap)
(require 'TComplex)
(require 'Evalmap)
(require 'Garbage)
(require 'Split_P)

(require 'extadiac)




(defun Embedding-Dimension-only-for-Mapping (Obj)
  (cond 
   ( (polyhedron?      obj) (get-dpol Obj)
     )
   ( (polyhedra-complex? obj)
     (Embedding-Dimension-only-for-Mapping (get-node-pol-complex  (first obj)))
     )
   )
  )



;(defun Map-On-Pol-Complex (Functions 
;                           Pol-Complex 
;                           &optional (Repr 'on-faces) (verb nil) )
;  
;  (if verb (format T "Coercing On-Vertices... "))
;  (let ( (Newpc (coerce-pol-complex Pol-Complex 'on-vertices))
;         (NewDim (length Functions))
;         )
;    (if verb (format T "Triangulation... "))
;    (setq Newpc (Triangulate-Pol-Complex Newpc))
;    (if verb (format T "Evaluation... "))
;    (setq Newpc (Push-Down-Transf-Pol-Complex-2 Newpc))
;    (if verb (format T "Mapping... "))
;    (setq Newpc (Mapper::Map-Functions-On-Pol-Complex Functions Newpc))
;    (if verb (format T "Removing Degeneration... "))
;    (setq Newpc (Remove-Degenerate-Cells-Pol-Complex Newpc))
;    (if verb (format T "Embedding... ")) 
;    (setq Newpc (Modify-Dimension-Of-Pol-Complex Newpc NewDim))
;    (if verb (format T "Splitting Polyhedrons... "))
;    (setq Newpc (Split-Pol-Complex Newpc))
;    (if (lisp-extension::Symbol-Equal Repr 'On-Faces)
;      (progn
;        (if verb (format T "Coercing On-Faces... "))
;        (Coerce-Pol-Complex Newpc 'On-Faces)
;        )
;      ;else
;      Newpc
;      )
;    )
;  )

;---------------------------------------------
; new version: the following is called on a plasm function symbol and
; a polyhedral complex. The previous one was called on a sequence of plasm
; function and a polyhedral complex. The number of mapping functions is provided
; by the lisp function "plasm::how-many-function-for-this-cons?": this checks
; that symbol function can stand for a cons of plasm functions

;(defun Map-On-Pol-Complex (Functions 
;                           Pol-Complex 
;                           &optional (Repr 'on-faces) (verb nil) )
;  (if verb (format T "Coercing On-Vertices... "))
;  (let ( (Newpc (coerce-pol-complex Pol-Complex 'on-vertices))
;         NewDim
;         )
;    (if verb (format T "Triangulation... "))
;    (setq Newpc (Triangulate-Pol-Complex Newpc))
;    (if verb (format T "Evaluation... "))
;    (setq Newpc (Push-Down-Transf-Pol-Complex-2 Newpc))
;    (if verb (format T "Mapping... "))
;    (setq Newpc (Mapper::Map-Functions-On-Pol-Complex Functions Newpc))
;    (if verb (format T "Removing Degeneration... "))
;    (setq Newpc (Remove-Degenerate-Cells-Pol-Complex Newpc))
;    (if verb (format T "Embedding... ")) 
;    (setq Newpc (Modify-Dimension-Of-Pol-Complex Newpc NewDim))
;    (if verb (format T "Splitting Polyhedrons... "))
;    (setq Newpc (Split-Pol-Complex Newpc))
;    (if (lisp-extension::Symbol-Equal Repr 'On-Faces)
;      (progn
;        (if verb (format T "Coercing On-Faces... "))
;        (Coerce-Pol-Complex Newpc 'On-Faces)
;        )
;      ;else
;      Newpc
;      )
;    )
;  )
;



;(Luglio 1995) Versione che setta anche le adiacenze esterne

 
(defun Map-On-Pol-Complex (Functions 
                           Pol-Complex)
  
  (let ((Cells-table (make-hash-table :test 'equal)) 
         Newpc 
         NewDim)
    (if (not (lisp-extension::symbol-equal (get-pol-complex-rep pol-complex)
                                           'On-Vertices))
      (setq Newpc (coerce-pol-complex Pol-Complex 'on-vertices))
      (setq Newpc Pol-Complex))

    (setq Newpc (Triangulate-Pol-Complex Newpc))

    (setq Newpc (Push-Down-Transf-Pol-Complex-2 Newpc))

    (setq Newpc (Mapper::Map-Functions-On-Pol-Complex Functions Newpc))

    (setq Newpc (Remove-Degenerate-Cells-Pol-Complex Newpc))

    (setq NewDim (Embedding-Dimension-only-for-Mapping Newpc))

    (setq Newpc (Modify-Dimension-Of-Pol-Complex Newpc NewDim))

    (setq Newpc (Split-Pol-Complex Newpc Cells-table))
  ;cells-table e' modificata dai side-effect della funzione
  ;precedente

    (setq Newpc (Set-External-Adjacences Newpc Cells-table Functions)) 

     Newpc))


