; -*- package: newtriangulation -*-
;-------------------------------------------------------------------------------
;           Universita' di Roma "La Sapienza"
;       Dipartimento di Informatica e Sistemistica
;            Via Buonarroti 12 - 00185 Roma
;
;                      CAD Group                 
; 
; Project: SimpleXn
; Identification: Package Triangulation                           
; Date: 27-12-90
; Note: Simplicial Decomposition, Von Hohenbalken Algorithm
;       with optional build-adjacencies flag
;             
; Version: 1.0
;              
; Author: F. Bernardini
;-------------------------------------------------------------------------------
; Modificata

;(28/9/93) La funzione convex-list-to-pol produce, ora, un poliedro contenente
;          celle convesse, costituite da triangoli, e non piu' un complesso
;          simpliciale.
;(30/9/93) Ora lo schema del convesso viene ricostruito utilizzando il grafo di
;          incidenza del politopo costituito cal convesso, ricostruito mediante
;          il metodo Beneath-Beyond.

(provide 'n_triang)                              ;module name

(inpackage newtriangulation)


(export '(convex-list-to-pol
          build-ad-pol
          svertcoord-list-list-to-pol
          convex-intersection-test
          convex-intersection))
                        
(require 'kernel)
(require 'matlib)
(require 'PL)
(require 'intersec)

(use-package 'kernel)
(use-package 'matlib)
(use-package 'PL)
(use-package 'intersect)


;-------------------------------------------------------------------------------
; Von Hohenbalken algorithms
;-------------------------------------------------------------------------------

(defun simplicial-complex (scheme)
  "Build the simplicial complex associated with the scheme"
  (let ((Z ())
        (d (1+ (length scheme)))
        (x (caaar scheme)))
    (dolist (face (first (last scheme)) Z)
      (unless (member x face)
        (if (= (length face) d)
          (push (cons x face) Z)
          (setf Z (append 
                   (mapcar #'(lambda (s) (cons x s))
                           (simplicial-complex 
                            (extract-scheme face (butlast scheme)))) 
                   Z)))))))
            

(defun extract-scheme (face scheme)
  "Extract the sub-scheme of a face."
  (mapcar
   #'(lambda (j-faces)
       (let ((sel-faces ()))
         (dolist (f j-faces sel-faces)
           (when (subsetp f face) (push f sel-faces)))))
   scheme))


;-------******************************-------------------------
;Versione nuova : le seguenti 3 routines, insieme con quelle
; dell'algoritmo Beneath-Beyond, ricostruiscono lo schema
; corretto del convesso.
;Probabilmente le funzioni precedente possono essere
; tutte buttate. 

(defun convex-scheme (vcoordlist d)
  (let ( (i -1)
         )
    (lisp-extension::butlast 
     (Reordinate-Faces
      (kernel::Get-Scheme-Of-Convex
       (kernel::Beneath-Beyond 
        (mapcar #'(lambda (point) (list (setq i (1+ i)) point ))
                vcoordlist)
        nil))
      )
     )
    )
  )

(defun Reordinate-Faces (ListOfListFaces)
  (mapcar #'(lambda (ListFaces)
              (Extract-Ordered-Faces
               ListFaces
               0))
          ListOfListFaces)
  )

(defun Extract-Ordered-Faces (FaceList ID)
  (let ( SelFaces
         Remaining
         )
    (if (not (null FaceList))
      (progn
        (dolist (el FaceList)
          (if (find ID el)
            (setq SelFaces (append SelFaces (list el)))
            ;else
            (setq Remaining (append Remaining (list el)))
            )
          )
        (append SelFaces (Extract-Ordered-Faces Remaining (1+ ID)))
        )
      ;else
      nil
      )
    )
  )


;-------------------------------------------------------------------------------
; Polyhedron reconstruction
;-------------------------------------------------------------------------------

(defun convex-list-to-vlist (convex-list)
  "Build vlist and the translation table."
  (let ((vlist ())
        (vtable ())
        (i 0))
    (dolist (convex convex-list)
      (let ((vtable-slot ()))
        (dolist (vertex convex)
          (let ((j (exist-vertex? vertex vlist)))
            (unless j
              (setf j i)
              (setf i (1+ i))
              (setf vlist 
                    (pushlast 
                     (make-vertex (make-vertex-symbol j) vertex) 
                     vlist)))
            (setf vtable-slot (pushlast j vtable-slot))))
        (setf vtable (pushlast (coerce vtable-slot 'vector) vtable))))
    (values vlist vtable)))


(defun exist-vertex? (vertex vlist)
  (position vertex (mapcar #'second vlist) :test #'same-vertex))


(defun same-vertex (v1 v2)
  (every #'(lambda (c1 c2) (< (abs (- c1 c2)) **epsilon)) v1 v2))


(defun translate-vid (scheme vtable-slot)
  "Rename vertices"
  (mapcar
   #'(lambda (s-faces)
       (mapcar
        #'(lambda (face)
            (mapcar #'(lambda (vid) (svref vtable-slot vid)) face))
        s-faces))
   scheme))


(defun sc-to-slist (sc)
  "Transform a simplicial complex in a slist"
  (let ((i -1)
        (d+1 (length (first sc))))
    (mapcar #'(lambda (s) 
                (make-simplex 
                 (make-simplex-symbol (setq i (1+ i)))
                 '+
                 (mapcar #'(lambda (c) 
                             (make-vertex-symbol c)) 
                         (sort (copy-seq s) #'<))
                 (make-list d+1 :initial-element 0)))
            sc)))


(defun extract-faces (sc faces)
  "Extract faces from a simplicial complex"
  (let* ((d+1 (length (first sc)))
         (s-faces (mapcan
                   #'(lambda (vl)
                       (let ((fl ()))
                       (dotimes (i d+1 fl)
                         (push (remove-pos i vl) fl))))
                   (copy-seq sc))))
    (mapcar
     #'(lambda (f)
         (let ((fl ()))
           (dolist (s-face s-faces fl)
             (when (subsetp s-face f)
               (push s-face fl)))))
     faces)))


(defun svertcoord-list-list-to-pol (svertcoord-list-list &optional (build-ad-flag t))
;  (format t "~%Number of convexes : ~D" (length svertcoord-list-list))
;  (format t "~%Calculating convexes...")  
  (let* ((i 0)
         (convex-list (mapcar #'(lambda (svl)
                                  ;(format t "~D  " (setf i (1+ i)))
                                  (convex-intersection svl))
                              svertcoord-list-list))
         (dim-list (mapcar 
                    #'dimension 
                    (mapcar #'vcoordlist-to-vertmat convex-list)))
         (max-dim (reduce #'max dim-list)))
    (if (= max-dim 1)                                        ;if dim = 1
      (segment-list-to-pol convex-list)                      ; a 1-polyhedron
      (convex-list-to-pol                                    ;transform in a pol
       (remove nil (mapcar #'(lambda (c d)                   ;regularization
                               (if (= d max-dim) c nil))
                           convex-list dim-list))
       max-dim build-ad-flag))))


(defun segment-list-to-pol (convex-list)
  (multiple-value-bind
    (vlist vtable) (convex-list-to-vlist convex-list)
    (let* ((sc (mapcar 
                #'(lambda (v) (list (svref v 0) 
                                    (svref v 1)))
                vtable)))
      ;(format t "~%Reconstructing polyhedron...~%")
      (build-ad-pol vlist (sc-to-slist sc)))))


(defun convex-list-to-pol (convex-list d &optional (build-ad-flag t))
  ;(format t "~%Calculating schemes of convexes...")
  (let (scheme-list)
    (setq scheme-list (convex-list-to-scheme-list convex-list d))
    ;(print convex-list)
    ;(format t "~%Triangulating convexes...")
    (convex-scheme-list-to-pol convex-list scheme-list d build-ad-flag)))


(defun convex-list-to-scheme-list (convex-list d)
  (let ((i 0)) 
    (mapcar #'(lambda (c)
                ;(format t "~D  " (setf i (1+ i)))
                (convex-scheme c d))
            convex-list)))


(defun convex-scheme-list-to-pol (convex-list scheme-list  Pdim &optional (build-ad-flag t))
  "Transform a list of convexes in a polyhedron." 
  (multiple-value-bind
    (vlist vtable) (convex-list-to-vlist convex-list)
    (let* ((sc ())
           (scheme-table 
            (map 'vector #'translate-vid scheme-list vtable))
           (face-table (map 'vector #'lastitem scheme-table)))
      (do* ((others (coerce 
                     (vet-interval (array-dimension scheme-table 0)) 
                     'list))
            (next (list 0))
            (i (or (car next) (car others)) (or (car next) (car others)))
            (sc-i ())
            (fc-i ()))
           ((null i))
        (setf others (remove i others))
        (setf next (cdr next))
        (setf sc-i (simplicial-complex (svref scheme-table i)))
        (setf fc-i (extract-faces sc-i (svref face-table i)))
        (setf sc (append sc sc-i))
        (dolist (j others)
          (let ((h 0))
            (dolist (f1 (svref face-table i))
              (let ((k 0))
                (dolist (f2 (svref face-table j))
                  (when (subsetp f1 f2)
                    (setf next (adjoin j next))
                    (let ((newfaces 
                           (append 
                            (remove-pos k (lastitem (svref scheme-table j)))
                            (nth h fc-i))))
                      (setf (first (last (svref scheme-table j))) newfaces)
                      (setf (svref face-table j) newfaces)))
                  (setq k (1+ k))))
              (setq h (1+ h)))))
        ;(format t "~D  " i)
        )
      ;(format t "~%Reconstructing polyhedron...~%")
      (if build-ad-flag
        (build-ad-pol vlist (sc-to-slist sc) Pdim)
        (kernel::Make-Polyhedron-On-Vertices-And-Convex-Cells vlist (sc-to-slist sc) Pdim)))))


(defun build-ad-pol (vlist slist Pdim)
  "Build a polyhedron, reconstructing the adjacencies 
and signs of simplices."
  (let* ((d+1 (length (get-vert (first slist))))
         (ns (length slist))
         (faces (make-array (list ns)
                            :initial-contents 
                            (mapcar
                             #'(lambda (vl)
                                 (let ((fl ()))
                                   (dotimes (i d+1 (reverse fl))
                                     (push (remove-pos i vl) fl))))
                             (mapcar #'get-vert slist))))
         (admat (make-array (list ns d+1)
                            :element-type 'integer
                            :initial-element -1))
         (signvet (make-array (list ns)
                              :element-type 'integer
                              :initial-element 0)))
    (do* ((others (coerce (vet-interval ns) 'list))
          (next (list 0))
          (i (or (car next) (car others)) (or (car next) (car others))))
         ((null i))
      (setf others (remove i others))
      (setf next (cdr next))
      (when (zerop (svref signvet i)) (setf (svref signvet i) 1))
      (dolist (j others)
        (let ((h 0))
          (dolist (f1 (svref faces i))
            (let ((k 0))
              (dolist (f2 (svref faces j))
                (when (equalp f1 f2)
                  (setf next (adjoin j next))
                  (if (minusp (aref admat i h))
                    (setf (aref admat i h) j)
                    (error-handler 0 "build-ad-pol"))
                  (if (minusp (aref admat j k))
                    (setf (aref admat j k) i)
                    (error-handler 0 "build-ad-pol"))
                  (if (zerop (svref signvet j))
                    (setf (svref signvet j) 
                          (- (* (expt -1 h) (expt -1 k) (svref signvet i))))))
                (setq k (1+ k))))
            (setq h (1+ h))))))
    (kernel::Make-Polyhedron-On-Vertices-And-Convex-Cells
     vlist
     (let ((i -1))
       (mapcar #'(lambda (s)
                   (setf i (1+ i))
                   (make-simplex
                    (get-id s)
                    (if (plusp (svref signvet i)) '+ '-)
                    (get-vert s)
                    (let ((sad ()))
                      (dotimes (j d+1 sad)
                        (let ((ad (aref admat i j)))
                          (setf sad 
                                (pushlast 
                                 (if (minusp ad) 
                                   0 
                                   (make-simplex-symbol ad)) sad)))))))
               slist)) Pdim)))

;-------------------------------------------------------------------------------
; Utilities
;-------------------------------------------------------------------------------

(defun nextitem (item lista)
  (let ((l (length lista))
        (p (position item lista)))
    (if p (nth (mod (1+ p) l) lista) (nth (mod 2 l) lista))))


(defun pushlast (item lista)
  (append lista (list item)))


(defun lastitem (l)
  (first (last l)))


(defun remove-pos (pos lista)
  (append (subseq lista 0 pos) (subseq lista (1+ pos))))


(defun print-scheme (F)
  (let ((d 1))
    (dolist (G F)
      (format t "~%~D-Facce : ~D~%" d (length G))
      (dolist (H G)
        (dolist (j H)
          (format t "~D  " j))
        (format t "~%"))
      (setq d (1+ d)))))
      

(defun vcoordlist-to-vertmat (vcoordlist)
  (let* ((m (length vcoordlist))                             ;number of vertices
         (n (length (first vcoordlist))))                    ;E-space dimension
    (mat-trans (make-array (list m n)                        ;matrix of vertices
                           :element-type 'real
                           :initial-contents vcoordlist))))


;-------------------------------------------------------------------------------
; Intersection of Convexes
;-------------------------------------------------------------------------------

; Nota Bene: in the following functions, all vertices must have the
; same number of coordinates.

(defun convex-intersection-test (VCLL)        ;VCLL (Vertex Coord. List List)
  (let ((num-convex (length VCLL)))           ; is a list of convexes
    (if (= num-convex 1)                      ; each convex is a list of vert.
      t                                       ; each vertex is a list of coord.
      (let* ((A (build-initial-matrix VCLL))        
             (nrows (array-dimension A 0))
             (ncols (array-dimension A 1)))
        (if (build-canonical-form A nrows (1- ncols)) t nil)))))


(defun convex-intersection (VCLL)              ;VCLL (Vertex Coord. List List)
  (let* ((A (build-initial-matrix VCLL))       ; is a list of convexes
         (nrows (array-dimension A 0))         ; each convex is a list of vert.
         (ncols (array-dimension A 1))         ; each vertex is a list of coord.
         (vert1 (car VCLL))
         (nvert1 (length vert1)))
    (multiple-value-bind 
      (A m n v0) (build-canonical-form A nrows (1- ncols))
        (calculate-rectangular-coordinates 
         vert1 
         (mapcar 
          #'(lambda (l) (subseq l 0 nvert1)) 
          (generate-vertices A m n v0))))))


(defun build-initial-matrix (VCLL)             ;VCLL (Vertex Coord. List List)
  (let* ((num-convex (length VCLL))            ; is a list of convexes
         (dim (length (caar VCLL)))            ; each convex is a list of vert.
         (num-vert-list (mapcar #'length VCLL)); each vertex is a list of coord.
         (nrows (- (* (1+ dim) num-convex) dim))
         (ncols (1+ (apply '+ num-vert-list)))
         (offs (* dim (1- num-convex)))
         (A (make-array (list nrows ncols)
                        :element-type 'real
                        :initial-element 0)))
    (do* ((s 0 (1+ s))
          (temp VCLL (cdr temp))
          (V1 (first temp) (first temp))
          (V2 (second temp) (second temp))
          (posr 0 (+ dim posr))
          (posc 1 (+ nv posc))
          (nv-list num-vert-list (cdr nv-list))
          (nv (first nv-list) (first nv-list)))
         ((= s (1- num-convex)) (fill-with-one A (+ s offs) posc nv))
      (copy-mat+ V1 A posr posc)
      (copy-mat- V2 A posr (+ nv posc))
      (fill-with-one A (+ s offs) posc nv))
    A))


(defun copy-mat+ (vlist mat posr posc)
  (let ((i posr)
        (j posc))
    (dolist (v vlist)
      (dolist (c v)
        (setf (aref mat i j) c)
        (setq i (1+ i)))
      (setq i posr)
      (setq j (1+ j)))))


(defun copy-mat- (vlist mat posr posc)
  (let ((i posr)
        (j posc))
    (dolist (v vlist)
      (dolist (c v)
        (setf (aref mat i j) (- c))
        (setq i (1+ i)))
      (setq i posr)
      (setq j (1+ j)))))

(defun fill-with-one (mat posr posc nv)
  (dotimes (i nv)
    (setf (aref mat posr (+ posc i)) 1))
  (setf (aref mat posr 0) 1))
