;;; -*- package: mapper -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package Parser                                            ;
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: V. Pascucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;Versione del 23/10/93 Corretta gestione poliedri subdimensionali

;(8/4/94) modificata per gestire ingresso funzioni fornite direttamente 
;         da plasm

;(11/4/94) Eliminate funzioni relative al parsing di espressioni matematiche,
;          in quanto la loro valutazione e' stata spostata all'interno di Plasm.


(provide 'newmap)

(inpackage mapper)

(export '(Map-Functions-On-Pol-Complex
          MAP-FUNCTIONS
          ))

(require 'kernel)
(use-package 'kernel)

(require 'dvrlib)
(use-package 'dvrlib)

(require 'extrusio)
(use-package 'extrusion)

;*******************************************************************************
;*********************** Funzioni modificate per eseguire    *******************
;*********************** il mapping sui complessi poliedrali *******************
;*******************************************************************************

(defun Map-Functions-On-Pol-Complex (Functions Obj)
  (cond 
   ( (polyhedron?      obj) (Map-Functions Functions Obj)
     )
   ( (polyhedra-complex? obj)
     (mapcar  #'(lambda (nodo) (Map-Functions-On-Pol-Complex Functions nodo)) obj)
     )
   ( (kernel::polyhedral-node? obj)
     (progn
       (kernel::make-pol-node 
        (Map-Functions-On-Pol-Complex Functions (get-node-pol-complex obj))
        (get-node-matrix obj)
        (get-node-id obj)
        )
       )
     )
   )
  )

;----------------------------------------------------
;
; NUOVA VERSIONE. fun contiene una funzione.
;
;----------------------------------------------------

;(defun MAP-FUNCTIONS (fun pol)
 ; (let* ( (Old-vlist (get-vlist pol))
  ;        (New-vlist         
   ;        (mapcar #'(lambda (vert)
    ;                   (list
     ;                   (first vert)
      ;                  (plasm::reduce-function-of-one-arg fun (second x)))
       ;                Old-vlist)))
        ;  )
    ;Dimensione del nuovo spazio di immersione del poliedro
;    (set-dpol Pol (length (second (car New-vlist)))) ;Per sbrigarsi facciamo cosi'
 ;   (setf (kernel::polyhedron-vlist Pol) New-vlist)
  ;  Pol
    ;)
  ;)

;----------------------------------------------------
;
; NUOVA VERSIONE. fun-list contiene una lista di funzioni.
;
; questo perche' la funzione Map-On-Pol-Complex chiamata precedentemente
; a questa determina la nuova dimensione del nuovo complesso dal numero
; delle funzioni che agiscono sui singoli vertici 
;----------------------------------------------------

;(defun MAP-FUNCTIONS (fun-list pol)
 ; (let* ((Old-vlist (get-vlist pol))
  ;       (New-vlist         
   ;       (mapcar #'(lambda (vert)
    ;                  (list
     ;                  (first vert)
      ;                 (plasm::reduce-function-of-one-arg  
       ;                 (plasm::reduce-function-of-one-arg 'plasm::cons fun-list)
        ;                (second vert))))
         ;         Old-vlist)))
          
    ;Dimensione del nuovo spazio di immersione del poliedro
;    (set-dpol Pol (length (second (car New-vlist)))) ;Per sbrigarsi facciamo cosi'
 ;   (setf (kernel::polyhedron-vlist Pol) New-vlist)
  ;  Pol))


;---------------------------------------
; this version works with a plasm function symbol and a polyhedron

(defun MAP-FUNCTIONS (fun-symbol pol)
  (let* ((Old-vlist (get-vlist pol))
         
         (New-vlist         
          (mapcar #'(lambda (vert)
                      (list
                       (first vert)
                       (plasm::reduce-function-of-one-arg  
                        fun-symbol
                        (second vert))))
                  Old-vlist)))
          
    ;Dimensione del nuovo spazio di immersione del poliedro
    (set-dpol Pol (length (second (car New-vlist)))) ;Per sbrigarsi facciamo cosi'
    (setf (kernel::polyhedron-vlist Pol) New-vlist)
    Pol))

(defun MAP-FUNCTIONS (fun-symbol pol)
  (let* ((Old-vlist (get-vlist pol))
         
         (New-vlist         
          (mapcar #'(lambda (vert)
                      (list
                       (first vert)
                       (plasm::reduce-function-args  
                        fun-symbol
                        (second vert))))
                  Old-vlist)))
          
    ;Dimensione del nuovo spazio di immersione del poliedro
    (set-dpol Pol (length (second (car New-vlist)))) ;Per sbrigarsi facciamo cosi'
    (setf (kernel::polyhedron-vlist Pol) New-vlist)
    Pol))
