;-*- package:kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;
; Project: SimpleXn                                                         ;
; Identification: Mapping su complessi : Trattamento dimensione non piena   ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: G. Proietti                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; (23/4/94) modificata la funzione Make-Matrix-For-Map-From-Standard-Simplex.
;           Ora, nel caso di simplessi di dimensione intrinseca nulla viene
;           ritornata la corretta matrice di traslazione, saltando il calcolo
;           della matrice di trasformazione generica in quanto non necessaria.
;
;;;
;
; (Luglio 1995) Modificata la funzione Make-Pol-Node-From-One-Simplex-Cell
;               in modo che non butti piu' via le adiacenze. 
;               Esse infatti verranno usate per la costruzione delle adiacenze
;               esterne delle celle.
;
;    by Marco Conti
;
;;;;;;;;;;;;;;;;;;;


(inpackage kernel)

(provide 'Split_P)

;-----------------------------------------
;
(defun Split-Pol-Complex (Obj
                          &optional Ctable)
  (cond 
   ( (polyhedron?      obj) 
     ;Se dim. spazio di immersione maggiore della dim. intrinseca
     (if (> (get-dpol obj) (get-npol obj))  
       (Split-Polyhedron obj Ctable)
       ;else
       obj
       )
     )
   ( (polyhedra-complex? obj)
     (mapcar  #'(lambda (nodo) (Split-Pol-Complex nodo Ctable)) obj)
     )
   ( (polyhedral-node? obj)
     (progn
       (make-pol-node 
        (Split-Pol-Complex (get-node-pol-complex obj) Ctable)
        (get-node-matrix obj)
        (get-node-id obj)
        )
       )
     )
   )
  )



;-----------------------------------------------
;Dato un poliedro rappresentato in celle convesse,
;triangolato, ossia con le celle consistenti in 
;simplessi ed avente dimensione non piena, lo 
;divide in tanti nodi poliedrali quante sono le
;sue celle.
(defun Split-Polyhedron (Pol 
                         &optional Ctable)
  (let ( (VList (Get-VList Pol)) )
    (mapcar #'(lambda (Cell)
                (Make-Pol-Node-From-One-Simplex-Cell Cell VList Ctable)
                ) (Get-CList Pol) )
    )
  )


;------------------------------------------------------------
;Data una cella convessa Cell costituente un simplesso di dimensione
;intrinseca pari a d, immersa in uno spazio di dimensione n>=d,
;un database di vertici Vlist contenente i vertici della cella,
;ritorna il nodo poliedrale costituito da un poliedro contenente
;l'unica cella convessa ridotta ad un simplesso standard di 
;dimensione d e dalla matrice di immersione che mappa il
;simplesso nella cella originaria.
;
(defun Make-Pol-Node-From-One-Simplex-Cell (Cell VList
                                            &optional Ctable)
  (let* (
         (CellVertsID (get-cell-vertices Cell))
         (DimPol (1-  (length CellVertsID)))
         (CellsVerts  (mapcar #'(lambda (x) (assoc x VList)) 
                             CellVertsID)) ;Lista associativa con vertici della cella
         (CellVertsCoords   (mapcar #'second CellsVerts))
         (pol (make-empty-polyhedron 'on-vertices 'convex-cells))
         StdSimplexVcoords
         StdSimplexVerts
         Coord
         Matrix
         pol-node
         )

    (setq Cell (Copy-Cell Cell))
    ;Calcolo le coordinate dei vertici del simplesso standard
    (setq StdSimplexVcoords
          (cons (make-list DimPol :initial-element 0) ;primo vertice: Origine
                (dotimes (i DimPol StdSimplexVcoords)
                  (setq Coord (make-list DimPol :initial-element 0))
                  (setf (nth i Coord) 1)
                  (setq StdSimplexVcoords
                        (append StdSimplexVcoords
                                (list Coord)))
                  ))
          )

    ;Etichetto i vertici con le label contenute nella cella iniziale
    (setq StdSimplexVerts (mapcar #'(lambda (x y) (list x y)) 
                                  CellVertsID
                                  StdSimplexVcoords))

    ;Calcolo la matrice di trasformazione per l'immersione nello spazio
    (setq Matrix  (Make-Matrix-For-Map-From-Standard-Simplex CellVertsCoords))

    ;Costruzione effettiva poliedro
    (put-vlist StdSimplexVerts pol)

    ;rimuovi vecchie adiacenze tra' le celle
    ;(setf (kernel::Convex-Cell-ADJACENT-CONVEX-CELLS Cell)
    ;      (make-list (length (kernel::Convex-Cell-ADJACENT-CONVEX-CELLS Cell))
    ;                 :initial-element 0)
    ;      )

    (setq pol (put-cell Cell pol))
    (set-npol pol DimPol)  
    (set-dpol pol DimPol)
    (set-#kd  pol)
    (set-#k0  pol)
    (set-#F  pol)
    (set-#C  pol)

    ;costruzione della tabella delle adiacenze
    (maphash #'(lambda (id cell)
      (setf (gethash id Ctable) pol))
     (get-cells pol)) 

    ;costruzione del nodo poliedrale
    (setq pol-node (make-pol-node pol Matrix 
                     (advance-polyhedral-node-counter)))
    pol-node
    )
  )


(defun Copy-Cell (Cell)
  (make-convex-cell (kernel::convex-cell-identifier Cell)  
                    (kernel::convex-cell-sign       Cell)
                    (copy-list (kernel::convex-cell-FACES       Cell))
                    (copy-list (kernel::convex-cell-vertices    Cell))
                    (copy-list (kernel::convex-cell-ADJACENT-CONVEX-CELLS Cell))
                    nil)
  )








;------------------------------------------------------------
;Dato un insieme di n+1 punti in uno spazio di dimensione
;d >= n, ritorna la matrice di trasformazione affine che
;mappa i vertici del simplesso standard a dimensione n nei
;vertici passati. l'origine verra' mappata nel primo punto,
;mentre il vertice sulla coordinata i-esima verra' mappato
;nell' i-esimo punto ( i=1 ... n ) 
;
(defun Make-Matrix-For-Map-From-Standard-Simplex ( VertList )
  (let ( (TranslVerts (mapcar #'(lambda (Vert)
                                  (mapcar (function (lambda 
                                              (x y) (- x y)
                                              )) Vert (car VertList))
                                  ) (cdr VertList))
                      ) ; n-1 vertici con l'origine nel primo
         (TranslMat   (kernel::My-Make-Translation-Matrix 
                       (length (car VertList))
                       (car VertList)
                       )
                      )
         OrtoBase
         Base
         Mat
         )
    ;Caso di dimensione intrinseca nulla:
    (if (null TranslVerts) 
      TranslMat
     ;else
      (progn
        (setq OrtoBase (kernel::Ortogonal-base-Of-Space TranslVerts))
        (setq Base (append TranslVerts (nthcdr (length TranslVerts) OrtoBase)))
        (setq Mat (kernel::listoflist-to-mat 
                   (kernel::Extend-Matrix-First-Row-and-Col
                    (new-pl:transpose-matrix 
                     Base)))
              )
        
        ;(list TranslVerts TranslMat)
        
        (matlib::Mat-Product TranslMat Mat)
        )
      )
    )
  )
