; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Mapping su complessi : Triangolazione di complessi        ;               
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: G. Proietti                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; (8/4/94) G. Proietti : esclusa la triangolazione di oggetti di dimensione
;                        0 od 1 (che sono gia' trangolati di per se stessi).
;                        La modifica e' stata fatta nella Triangulate-polyhedron
;
;------------------------------------------------------------------------------

(inpackage kernel)

(provide 'TComplex)

(require 'n_triang) 

(defun Triangulate-Pol-Complex (obj)
  (cond ( (polyhedron?      obj) (Triangulate-Polyhedron obj))
        ( (polyhedra-complex? obj)
          (mapcar  #'(lambda (nodo) (Triangulate-Pol-Complex nodo)) obj)
          )
        ( (polyhedral-node? obj)
          (progn
            (make-pol-node 
             (Triangulate-Pol-Complex (get-node-pol-complex obj))
             (get-node-matrix obj)
             (get-node-id obj)
             )
            )
          )
        )
  )



;-------------------------------------------------------------------------------
; Triangolazione di un singolo poliedro. Qualora la dimensione intrinseca dello
; stesso sia 0 od 1 , la triangolazione e' superflua e non viene eseguita.
; Questa modifica e' stata inserita l' 8/4/94 in seguito a problemi nella funzione
; di estrazione degli schemi dei convessi nel caso di dimensione 0 od 1.

(defun Triangulate-Polyhedron ( Pol )
  (if (> (get-dpol Pol) 1)
    (let ( (celle     (get-clist Pol))
           vcelle  
           (vertici   (get-vlist Pol))
           )
      ;Prelievo dei vertici delle celle (specificati simbolicamente)
      (setq vcelle (mapcar #'convex-cell-vertices celle))
      ;Converto i simboli in vertici
      (setq vcelle (mapcar #'(lambda (cell)
                               (mapcar #'(lambda (vert)
                                           (second (assoc vert vertici))
                                           ) cell )
                               ) vcelle )
            )
      (NewTriangulation::Convex-List-to-Pol vcelle (polyhedron-M Pol) )
      )
    ;else non devo triangolare e ritorno una copia del poliedro
    (copy-tree Pol)
    )
  )


;Dato un poliedro specificato da una lista di vertici etichettati e da una lista 
; di simplessi, ritorna il poliedro costituito da celle convesse, ciascuna
;costituita da un simplesso.
(defun Make-Polyhedron-On-Vertices-And-Convex-Cells (Vlist Slist Pdim)
  (let ( (NewPolyhedron (kernel::Make-Empty-Polyhedron 'on-vertices)) 
         )
    (put-cells 
     (mapcar #'(lambda (Simpl)
                (convert-simplex-into-covex-cell Simpl 'on-vertices)) Slist)
     NewPolyhedron )

    (setf (polyhedron-vlist NewPolyhedron) Vlist)
    
    (set-npol NewPolyhedron Pdim)  
    (set-dpol NewPolyhedron Pdim)
    (set-#kd  NewPolyhedron)
    (set-#k0  NewPolyhedron)
    (set-#F  NewPolyhedron)
    (set-#C  NewPolyhedron)
    (setf (polyhedron-KIND-OF-BRICK NewPolyhedron)  'CONVEX-CELLS)
    NewPolyhedron
    )
  )
