; -*- package: extrusion -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package extrusion                                         ;
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: V. Pascucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Questo modulo estende le funzionalit del package extrusion al caso di 
; poliedri definiti su celle convesse a facce.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require 'extrusio)
(provide 'new_extr)


(inpackage extrusion)  ;package name

(export '(extrude-pol-on-convexes-and-faces
          linear-extrude-pol-on-convexes-and-faces
          general-extrude))

(require 'hyperp)
(use-package 'hyperplanes)

(require 'lisp_ext)
(use-package 'lisp-extension)

;(require 'conv_ext)
(use-package 'kernel)

;-------------------------------------------------------------------------------------------
; estrusione lineare di un complesso a celle convesse basato sulle facce:
;   - pol   poliedro da estrudere 
;   - list-of-samples  lista di liste;
;                      ogni lista interna definisce un intervallo "pieno" lungo la direzione
;                      di estrusione;
;                      ciascuna lista puo` contenere 2 o piu` valori che definiscono la 
;                      coordinata (n+1)-esima degli iperpiani che tagliano in slices
;                      il poliedro estruso.
;-------------------------------------------------------------------------------------------
(defun extrude-pol-on-convexes-and-faces (pol &optional (list-of-samples '((0 1))))
  (let* ((new-polyhedron (make-empty-polyhedron 'on-faces 'convex-cells))
         (old-faces (get-flist pol))
         faces 
         (convex-cells (get-cells pol))
         (new-cells (new-convex-cells-table))
         (cells-number (hash-table-count convex-cells))
         (old-cells-names-list (let ((names-list nil))
                                 (declare (special names-list))
                                 (maphash #'(lambda (name xxx)
                                             (push name names-list))
                                          convex-cells)
                                 names-list))
         (e_space_dim (max 0 (- (length (get-coord (first old-faces))) 1)))
         (zero-list (make-list e_space_dim :initial-element 0)))
    
    (reset-convex-cell-counter)
    (reset-face-counter)
    
    (setf faces                                      ;new faces parameters
          (mapcar                                    ;accirding to new
           #'(lambda (the-face)                       ;e-space dimension 
              (make-face (get-id the-face) 
                         (append (get-coord the-face) 
                                 (list 0))))
           old-faces))
    
    (do ((samples (pop list-of-samples)(pop list-of-samples)))
        ((null samples)
         new-polyhedron)
      
      
      (let* ((new-cells-names nil)
             (new-faces-names nil))
        
        (dotimes (i (1- (length samples)))               ;definition of new
          (push '() new-cells-names)                     ;cells names
          (dotimes (j cells-number)
            (push  (advance-convex-cell-counter)
                   (first new-cells-names))))
        
        (dotimes (i (length samples))
          (let ((the-face (normalize-parameters
                           (append (list (pop samples))
                                   zero-list
                                   '(-1)))))
            (push (cons (do ()
                            ((not (get-pol-face (advance-face-counter) pol))
                             (previous-face-counter))) (cdr the-face))
                  new-faces-names)                          ;add new faces
            (push                                           ;according to the
             (make-face                                     ;values of the
              (first (first new-faces-names))               ;list samples
              (first the-face))
             faces)))
        
        (do* ((top-face-name (pop new-faces-names)(pop new-faces-names))
              (bottom-face-name (first new-faces-names)(first new-faces-names))
              (top-adjacents nil (pop new-cells-names))
              (cells-names (first new-cells-names)(first new-cells-names))
              (old-cells-names old-cells-names-list old-cells-names-list)
              (names-matching (mapcar 'cons old-cells-names cells-names)
                              (mapcar 'cons old-cells-names cells-names))
              (bottom-adiacents (second new-cells-names)(second new-cells-names)))
             ((not new-faces-names))
          
          (do* ((top-adjacent-name (or (pop top-adjacents) 0)
                                   (or (pop top-adjacents) 0))
                (old-cell-name (pop old-cells-names)(pop old-cells-names))
                (old-cell (gethash old-cell-name convex-cells)
                          (gethash old-cell-name convex-cells))
                (old-cell-sign (when old-cell (get-cell-sign old-cell))
                               (when old-cell (get-cell-sign old-cell)))
                (new-cell-name (pop cells-names)(pop cells-names))
                (bottom-adjacent-name (or (pop bottom-adiacents) 0)
                                      (or (pop bottom-adiacents) 0)))
               
               ((not old-cell-name))
            (setf (gethash new-cell-name new-cells)
                  (make-convex-cell new-cell-name
                                    old-cell-sign
                                    (append (list (cons (first bottom-face-name)
                                                        (if (equal '+ (cdr bottom-face-name))
                                                          old-cell-sign
                                                          (opposit-sign
                                                           old-cell-sign)))
                                                  (cons (first top-face-name) 
                                                        (if (equal '+ (cdr top-face-name))
                                                          (opposit-sign
                                                           old-cell-sign)
                                                          old-cell-sign)))
                                            (get-cell-faces old-cell))
                                    (append (mapcar `(lambda (x)
                                                       (append (list (quote ,(first bottom-face-name))) x))
                                                    (get-cell-vertices old-cell))
                                            (mapcar `(lambda (x)
                                                       (append (list (quote ,(first top-face-name))) x))
                                                    (get-cell-vertices old-cell)))
                                    (append (list bottom-adjacent-name
                                                  top-adjacent-name) 
                                            (sublis names-matching
                                                    (get-cell-adjacences old-cell)
                                                    :test 'symbol-equal))
                                    nil))))
        (put-faces faces new-polyhedron)
        (put-cells new-cells new-polyhedron)))
    (progn
      (set-npol new-polyhedron (1+ (get-npol pol)))
      (set-dpol new-polyhedron (1+ (get-dpol pol)))
      (set-#kd  new-polyhedron)
      (set-#k0  new-polyhedron) 
      (set-#F  new-polyhedron)
      (set-#C  new-polyhedron) 
      new-polyhedron)))

;-------------------------------------------------------------------------------------------
; estrusione lineare di un complesso a celle convesse basato sulle facce:
;   - pol   poliedro da estrudere 
;   - list-of-samples  lista di liste;
;                      ogni lista interna definisce un intervallo "pieno" lungo la direzione
;                      di estrusione;
;                      ciascuna lista puo` contenere 2 o piu` valori che definiscono la 
;                      coordinata (n+1)-esima degli iperpiani che tagliano in slices
;                      il poliedro estruso.
;   - direction-vector lista di reali (vettore che definisce la direzione di estrusione).
;-------------------------------------------------------------------------------------------
(defun linear-extrude-pol-on-convexes-and-faces (pol &optional (list-of-samples '((0 1)))
                                                     direction-vector)
  (let* ((new-polyhedron (make-empty-polyhedron 'on-faces 'convex-cells))
         (old-faces (get-flist pol))
         faces 
         (convex-cells (get-cells pol))
         (new-cells (new-convex-cells-table))
         (cells-number (hash-table-count convex-cells))
         (old-cells-names-list (let ((names-list nil))
                                 (declare (special names-list))
                                 (maphash #'(lambda (name xxx)
                                             (push name names-list))
                                          convex-cells)
                                 names-list))
         (e_space_dim (max 0 (- (length (get-coord (first old-faces))) 1)))
         (zero-list (make-list e_space_dim :initial-element 0))
         (direction-vector (if (null direction-vector)
                             (append zero-list '(1))
                             (mapcar `(lambda (x)
                                        (/ x ,(first (last direction-vector))))
                                     direction-vector)))
         (shearing-vector (append '(0) (mapcar '- (butlast direction-vector))))
         )
    
    (reset-convex-cell-counter)
    (reset-face-counter)
    
    (setf faces                                      ;new faces parameters
          (mapcar                                    ;accirding to new
           `(lambda (the-face)                       ;e-space dimension 
              (make-face (get-id the-face) 
                         (first
                          (normalize-parameters
                           (append (get-coord the-face) 
                                   (list (apply '+ (mapcar '* (get-coord the-face)
                                                           (quote ,shearing-vector)))))))))
           old-faces))
    
    (do ((samples (pop list-of-samples)(pop list-of-samples)))
        ((null samples)
         new-polyhedron)
      
      
      (let* ((new-cells-names nil)
             (new-faces-names nil))
        
        (dotimes (i (1- (length samples)))               ;definition of new
          (push '() new-cells-names)                     ;cells names
          (dotimes (j cells-number)
            (push  (advance-convex-cell-counter)
                   (first new-cells-names))))
        
        (dotimes (i (length samples))
          (let ((the-face (normalize-parameters
                           (append (list (pop samples))
                                   zero-list
                                   '(-1)))))
            (push (cons (do ()
                            ((not (get-pol-face (advance-face-counter) pol))
                             (previous-face-counter))) (cdr the-face))
                  new-faces-names)                          ;add new faces
            (push                                           ;according to the
             (make-face                                     ;values of the
              (first (first new-faces-names))               ;list samples
              (first the-face))
             faces)))
        
        (do* ((top-face-name (pop new-faces-names)(pop new-faces-names))
              (bottom-face-name (first new-faces-names)(first new-faces-names))
              (top-adjacents nil (pop new-cells-names))
              (cells-names (first new-cells-names)(first new-cells-names))
              (old-cells-names old-cells-names-list old-cells-names-list)
              (names-matching (mapcar 'cons old-cells-names cells-names)
                              (mapcar 'cons old-cells-names cells-names))
              (bottom-adiacents (second new-cells-names)(second new-cells-names)))
             ((not new-faces-names))
          
          (do* ((top-adjacent-name (or (pop top-adjacents) 0)
                                   (or (pop top-adjacents) 0))
                (old-cell-name (pop old-cells-names)(pop old-cells-names))
                (old-cell (gethash old-cell-name convex-cells)
                          (gethash old-cell-name convex-cells))
                (old-cell-sign (when old-cell (get-cell-sign old-cell))
                               (when old-cell (get-cell-sign old-cell)))
                (new-cell-name (pop cells-names)(pop cells-names))
                (bottom-adjacent-name (or (pop bottom-adiacents) 0)
                                      (or (pop bottom-adiacents) 0)))
               
               ((not old-cell-name))
            (setf (gethash new-cell-name new-cells)
                  (make-convex-cell new-cell-name
                                    old-cell-sign
                                    (append (list (cons (first bottom-face-name)
                                                        (if (equal '+ (cdr bottom-face-name))
                                                          old-cell-sign
                                                          (opposit-sign
                                                           old-cell-sign)))
                                                  (cons (first top-face-name) 
                                                        (if (equal '+ (cdr top-face-name))
                                                          (opposit-sign
                                                           old-cell-sign)
                                                          old-cell-sign)))
                                            (get-cell-faces old-cell))
                                    (append (mapcar `(lambda (x)
                                                       (append (list (quote ,(first bottom-face-name))) x))
                                                    (get-cell-vertices old-cell))
                                            (mapcar `(lambda (x)
                                                       (append (list (quote ,(first top-face-name))) x))
                                                    (get-cell-vertices old-cell)))
                                    (append (list bottom-adjacent-name
                                                  top-adjacent-name) 
                                            (sublis names-matching
                                                    (get-cell-adjacences old-cell)
                                                    :test 'symbol-equal))
                                    nil))))
        (put-faces faces new-polyhedron)
        (put-cells new-cells new-polyhedron)))
    (progn
      (set-npol new-polyhedron (1+ (get-npol pol)))
      (set-dpol new-polyhedron (1+ (get-dpol pol)))
      (set-#kd  new-polyhedron)
      (set-#k0  new-polyhedron) 
      (set-#F  new-polyhedron)
      (set-#C  new-polyhedron) 
      new-polyhedron)))

(defun linear-extrude-pol-on-convexes-and-faces (pol &optional (list-of-samples '((0 1)))
                                                     direction-vector)
  (let* ((new-polyhedron (make-empty-polyhedron 'on-faces 'convex-cells))
         (old-faces (get-flist pol))
         faces 
         (convex-cells (get-cells pol))
         (new-cells (new-convex-cells-table))
         (cells-number (hash-table-count convex-cells))
         (old-cells-names-list (let ((names-list nil))
                                 (declare (special names-list))
                                 (maphash #'(lambda (name xxx)
                                             (push name names-list))
                                          convex-cells)
                                 names-list))
         (e_space_dim (max 0 (- (length (get-coord (first old-faces))) 1)))
         (zero-list (make-list e_space_dim :initial-element 0))
         (direction-vector (if (null direction-vector)
                             (append zero-list '(1))
                             (mapcar #'(lambda (x)
                                        (/ x (first (last direction-vector))))
                                     direction-vector)))
         (shearing-vector (append '(0) (mapcar '- (butlast direction-vector))))
         )
    
    (reset-convex-cell-counter)
    (reset-face-counter)
    
    (setf faces                                      ;new faces parameters
          (mapcar                                    ;accirding to new
           #'(lambda (the-face)                       ;e-space dimension 
              (make-face (get-id the-face) 
                         (first
                          (normalize-parameters
                           (append (get-coord the-face) 
                                   (list (apply '+ (mapcar '* (get-coord the-face)
                                                           shearing-vector))))))))
           old-faces))
    
    (do ((samples (pop list-of-samples)(pop list-of-samples)))
        ((null samples)
         new-polyhedron)
      
      
      (let* ((new-cells-names nil)
             (new-faces-names nil))
        
        (dotimes (i (1- (length samples)))               ;definition of new
          (push '() new-cells-names)                     ;cells names
          (dotimes (j cells-number)
            (push  (advance-convex-cell-counter)
                   (first new-cells-names))))
        
        (dotimes (i (length samples))
          (let ((the-face (normalize-parameters
                           (append (list (pop samples))
                                   zero-list
                                   '(-1)))))
            (push (cons (do ()
                            ((not (get-pol-face (advance-face-counter) pol))
                             (previous-face-counter))) (cdr the-face))
                  new-faces-names)                          ;add new faces
            (push                                           ;according to the
             (make-face                                     ;values of the
              (first (first new-faces-names))               ;list samples
              (first the-face))
             faces)))
        
        (do* ((top-face-name (pop new-faces-names)(pop new-faces-names))
              (bottom-face-name (first new-faces-names)(first new-faces-names))
              (top-adjacents nil (pop new-cells-names))
              (cells-names (first new-cells-names)(first new-cells-names))
              (old-cells-names old-cells-names-list old-cells-names-list)
              (names-matching (mapcar 'cons old-cells-names cells-names)
                              (mapcar 'cons old-cells-names cells-names))
              (bottom-adiacents (second new-cells-names)(second new-cells-names)))
             ((not new-faces-names))
          
          (do* ((top-adjacent-name (or (pop top-adjacents) 0)
                                   (or (pop top-adjacents) 0))
                (old-cell-name (pop old-cells-names)(pop old-cells-names))
                (old-cell (gethash old-cell-name convex-cells)
                          (gethash old-cell-name convex-cells))
                (old-cell-sign (when old-cell (get-cell-sign old-cell))
                               (when old-cell (get-cell-sign old-cell)))
                (new-cell-name (pop cells-names)(pop cells-names))
                (bottom-adjacent-name (or (pop bottom-adiacents) 0)
                                      (or (pop bottom-adiacents) 0)))
               
               ((not old-cell-name))
            (setf (gethash new-cell-name new-cells)
                  (make-convex-cell new-cell-name
                                    old-cell-sign
                                    (append (list (cons (first bottom-face-name)
                                                        (if (equal '+ (cdr bottom-face-name))
                                                          old-cell-sign
                                                          (opposit-sign
                                                           old-cell-sign)))
                                                  (cons (first top-face-name) 
                                                        (if (equal '+ (cdr top-face-name))
                                                          (opposit-sign
                                                           old-cell-sign)
                                                          old-cell-sign)))
                                            (get-cell-faces old-cell))
                                    (append (mapcar #'(lambda (x)
                                                       (append (list (first bottom-face-name)) x))
                                                    (get-cell-vertices old-cell))
                                            (mapcar #'(lambda (x)
                                                       (append (list (first top-face-name)) x))
                                                    (get-cell-vertices old-cell)))
                                    (append (list bottom-adjacent-name
                                                  top-adjacent-name) 
                                            (sublis names-matching
                                                    (get-cell-adjacences old-cell)
                                                    :test 'symbol-equal))
                                    nil))))
        (put-faces faces new-polyhedron)
        (put-cells new-cells new-polyhedron)))
    (progn
      (set-npol new-polyhedron (1+ (get-npol pol)))
      (set-dpol new-polyhedron (1+ (get-dpol pol)))
      (set-#kd  new-polyhedron)
      (set-#k0  new-polyhedron) 
      (set-#F  new-polyhedron)
      (set-#C  new-polyhedron) 
      new-polyhedron)))

;-------------------------------------------------------------------------------------------
; estrusione lineare di un complesso a celle convesse basato sulle facce:
;   - pol   poliedro da estrudere 
;   - list-of-samples  lista di liste;
;                      ogni lista interna definisce un intervallo "pieno" lungo la direzione
;                      di estrusione;
;                      ciascuna lista puo` contenere 2 o piu` valori che definiscono la 
;                      coordinata (n+1)-esima degli iperpiani che tagliano in slices
;                      il poliedro estruso.
;   - direction-vector lista di reali (vettore che definisce la direzione di estrusione).
;-------------------------------------------------------------------------------------------
(defun general-extrude (pol &optional (list-of-samples '((0 1))) direction-vector)
  (let ((rep (get-pol-rep pol))
        (bricks (get-pol-brick-type pol))
        new-polyhedron)
    (setf new-polyhedron 
          (cond
           ((symbol-equal bricks 'simplices)
            (cond
             ((symbol-equal rep 'on-faces)
              (error-handler " Extrusion of simplices on faces is not currently supported. "
                               " general-extrude "))
             ((symbol-equal rep 'on-vertices)
              (error-handler " Extrusion of simplices on vertices is not currently supported. "
                               " general-extrude "))))
           ((symbol-equal bricks 'convex-cells)
            (cond
             ((symbol-equal rep 'on-faces)
              (linear-extrude-pol-on-convexes-and-faces pol list-of-samples direction-vector))
             ((symbol-equal rep 'on-vertices)
              (coerce-polyhedron
               (linear-extrude-pol-on-convexes-and-faces
                (coerce-polyhedron pol 'on-faces)
                list-of-samples direction-vector)
               'on-vertices))))))
    (if new-polyhedron
      new-polyhedron
      (error-handler " Bad extrusion parameters " " general-extrude "))))

;-------------------------------------------------------------------------------------------
; product of polyhedra with representation basen on faces:
; input: - all the parameters are optional, and have to be polyhedra
;          whoose representation is face based, and the basic elements
;          are convex cells.
; output: - one polyhedron decomposed in convex cells with face based representation.
;-------------------------------------------------------------------------------------------
(defun pol-on-faces-and-convexes-product (&rest polyhedra))
  

