; -*- package: new-pl -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package new-pl                                            ;               
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: V. Pascucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; SI SOTTOLINEA CHE IN QUESTO PACKAGE LE MATRICI NON SONO MEMORIZZATE
; TRAMMITE ARRAYS MA COME LISTE DI LISTE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'new_pl)
(defpackage  new-pl)

(inpackage new-pl)

(export '(reduce-matrix
	  calc-matrix-rank
	  transpose-matrix
	  identity-matrix
	  swap-rows))

(require 'lisp_ext) 
(use-package 'lisp-extension) 

(require 'kernel)
;;(import '(kernel:**epsilon)) ;;PRODUCE CONFLITTO !!!!!!!!!!!!!
(setq **epsilon 1e-13)


;-----------------------------------------------------------------------------
; Scambia le righe di indici first-row second-row della matrice a
;-----------------------------------------------------------------------------
(defun swap-rows (first-row second-row a)
  (swap (nth first-row a)
       (nth second-row a)))

;-----------------------------------------------------------------------------
; Somma alla riga di indice row della matrice matrix, la riga pivot-row
; moltiplicata per il coefficiente coeff.
;-----------------------------------------------------------------------------
(defun add-pivot-row*coeff-to-row (pivot-row
				   coeff
				   row
				   matrix)
  (setf (nth row matrix)
       (mapcar '+
	      (nth row matrix)
	      (mapcar `(lambda (x)
		       (* x ,coeff))
		    (nth pivot-row matrix)))))

(defun add-pivot-row*coeff-to-row (pivot-row
				   coeff
				   row
				   matrix)
  (setf (nth row matrix)
       (mapcar '+
	      (nth row matrix)
	      (mapcar #'(lambda (x)
		       (* x coeff))
		    (nth pivot-row matrix)))))

;-----------------------------------------------------------------------------
; moltiplica la riga di indice row della matrice matrix
;-----------------------------------------------------------------------------
(defun multiply-row-for-coeff (row
			       coeff
			       matrix)
  (setf (nth row matrix)
       (mapcar `(lambda (x)
		 (* x ,coeff))
	      (nth row matrix))))

(defun multiply-row-for-coeff (row
			       coeff
			       matrix)
  (setf (nth row matrix)
       (mapcar #'(lambda (x)
		 (* x coeff))
	      (nth row matrix))))

;-----------------------------------------------------------------------------
; Applica un passo dell'algoritmo del simplesso sulla matrice a considerando
; come elemento pivot quello di indici pivot-rowv pivot-column.
; La riga dell'elemento di pivot viene divisa per il valore dell'elemento
; di pivot stesso che quindi dopo vale 1.
; Le righe vengono combinate in modo tale che poi gli elementi della stessa
; colonna dell'elemento di pivot siano tutti nulli.
;-----------------------------------------------------------------------------
(defun pivot (pivot-row
	      pivot-column
	      a)
  (let ((pivot-element*-1 (- (nth pivot-column (nth pivot-row a)))))
   (dotimes (row (length a))
    (when (/= row pivot-row)
     (add-pivot-row*coeff-to-row pivot-row
				(/ (nth pivot-column (nth row a))
			      pivot-element*-1)
				row
				a)))
   (multiply-row-for-coeff pivot-row
			  (/ (- pivot-element*-1))
			  a)))

;-----------------------------------------------------------------------------
; Cerca l'elemento di pivot nelle sottomatrice specificata 
; da min-row max-row min-col max-col della matrice a.
; L'elemnto di pivot  quello di valure assoluto massimo.
;-----------------------------------------------------------------------------
(defun find-pivot-element (min-row
			   max-row
			   min-col
			   max-col
			   a)
  (let* ((sub-matrix (subseq (mapcar `(lambda (row)
				       (subseq row ,min-col ,(1+ max-col))) a)
			    min-row
			    (1+ max-row)))
	(min-el-list (mapcar #'(lambda (row)
				 (apply (quote min) row)) sub-matrix))
	(min-el (apply 'min min-el-list))
	
	(max-el-list (mapcar #'(lambda (row)
			      (apply (quote max) row)) sub-matrix))
	(max-el (apply 'max max-el-list)))
   (if (< (abs min-el) (abs max-el))
    (let* ((row (position max-el max-el-list))
	  (col (position max-el (nth row sub-matrix))))
     (list (+ row min-row)
	  (+ col min-col)))
    (let* ((row (position min-el min-el-list))
	  (col (position min-el (nth row sub-matrix))))
     (list (+ row min-row)
	  (+ col min-col))))))

(defun find-pivot-element (min-row
			   max-row
			   min-col
			   max-col
			   a)
  (let* ((sub-matrix (subseq (mapcar #'(lambda (row)
				       (subseq row min-col (1+ max-col))) a)
			    min-row
			    (1+ max-row)))
	(min-el-list (mapcar #'(lambda (row)
				 (apply (quote min) row)) sub-matrix))
	(min-el (apply 'min min-el-list))
	
	(max-el-list (mapcar #'(lambda (row)
			      (apply (quote max) row)) sub-matrix))
	(max-el (apply 'max max-el-list)))
   (if (< (abs min-el) (abs max-el))
    (let* ((row (position max-el max-el-list))
	  (col (position max-el (nth row sub-matrix))))
     (list (+ row min-row)
	  (+ col min-col)))
    (let* ((row (position min-el min-el-list))
	  (col (position min-el (nth row sub-matrix))))
     (list (+ row min-row)
	  (+ col min-col))))))

;--------------------------------------------------------------------------------
; Implementazion edell'algoritmo del simplesso per ridurre in forma 
; canonica una matrice a (input).
; La forma canonica consiste nella riduzione di n (input)
; colonne come n differenti colonne della matrice identit, 
; anche se non ordinate.
; Gli indici di input min-row max-row min-col indicano la sottomtrice
; di a su cui l'algoritmo andr ad operare nella fase di ricerca dell'elemento
; di pivot.
; Una volta individuato l'elemento di pivot, la sua riga verr portata in testa 
; alla suttomatrice di ricerca del pivot per poi aumentare di 1 il valore
; di min-row prima di iniziare la ricerca dell'elemeto susccessivo.
; La matrice a  rappresentata da una lista di liste di numeri di cu le liste 
; interne rappresentano le righe della matrice.
; La matrice a viene alterata direttamente per essere portata nella forma 
; canonica descritta, mentre viene restituito in output la lista degli indici di 
; colonna dei termini di valore unitario che compaiono nelle prime n righe 
; della suttomatrice di a definita in input tramite i parametri 
; min-row max-row min-col max-col.
;
;  es : (setf a '((1 2 3 4 5)
;                 (2 3 4 5 6)
;                 (2 4 3 2 1)
;                 (4 5 7 6 7)))
;
;       (NEW-PL:REDUCE-MATRIX 1 3 1 4 2 a) -> (2 4)
;
;       a -> ((-3/7  12/7  0 6/7 0)
;             ( 5/7  23/14 1 4/7 0) 
;             (-1/7 -13/14 0 2/7 1)
;             ( 0    2     0 1   0))
;--------------------------------------------------------------------------------
(defun reduce-matrix (min-row max-row min-col max-col n a)
  (let ((pivot-col-indices-list nil))
    (dotimes (i n)
      (let* ((x-y (find-pivot-element min-row max-row min-col max-col a))
	     (row-index (first x-y))
	     (col-index (second x-y)))
	(push col-index pivot-col-indices-list)
	(pivot row-index col-index a)
	(swap-rows min-row row-index a)
	(incf min-row)))
    (nreverse pivot-col-indices-list)))

;--------------------------------------------------------------------------------
; Funzione che restituisce il rango della sottomatrice della matrice a
; specificata dagli indici min-row max-row min-col max-col.
; Si richiede che la matrice a sia memorizzata in una lista di liste.
; La matrice a viene alterata durante il calcolo.
;--------------------------------------------------------------------------------
(defun calc-matrix-rank (min-row max-row min-col max-col a)
  (declare (special **epsilon))
  (let ((n (1+ (min (- max-row min-row) (- max-col min-col))))
	(rank 0))
    (dotimes (i n rank)
      (let* ((x-y (find-pivot-element min-row max-row min-col max-col a))
	     (row-index (first x-y))
	     (col-index (second x-y)))
	(if (< (abs (nth col-index (nth row-index a))) **epsilon)
	  (return rank))
	(pivot row-index col-index a)
	(swap-rows min-row row-index a)
	(incf min-row)
	(incf rank)))))

;--------------------------------------------------------------------------------
; Costruisce la matrice trasposta di quella data in input
;--------------------------------------------------------------------------------
(defun transpose-matrix (matrix)
  (let (transposed-matrix
	(n (length (first matrix))))
  (dotimes (i n)
    (push (mapcar `(lambda (row)(nth ,(- n 1 i) row)) matrix)
	  transposed-matrix))
  transposed-matrix))


(defun transpose-matrix (matrix)
  (let (transposed-matrix
	(n (length (first matrix))))
  (dotimes (i n)
    (push (mapcar #'(lambda (row)(nth (- n 1 i) row)) matrix)
	  transposed-matrix))
  transposed-matrix))

;--------------------------------------------------------------------------------
; Costruisce la matrice identit di dimensione n 
;--------------------------------------------------------------------------------
(defun identity-matrix (n)
  (let (matrix)
    (dotimes (i n)
      (push (make-list n :initial-element 0)
	    matrix))
    (dotimes (i n)
      (setf (nth i (nth i matrix)) 1))
    matrix))












