; -*- package: bound -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification:  Package Boundary                                         ;                                
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: V. Ferrucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'bound)     ;module name

(defpackage bound)  ;package name
(inpackage bound)

(export '(boundary get-admat))

(require 'kernel)

(use-package 'kernel)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar $$addata '()) ;this variable mantains
                      ; adjacency arrays

(defun calc-admat-of-s (m)
  (do ((admat (make-array (list (+ m 1) m) :element-type 'integer))
       (i 0 (+ i 1)))
      ((= i (+ m 1))  admat)
    (do ((j i (+ j 1)))
        ((= j m))
      (setf (aref admat i j) (+ j 1))
      (setf (aref admat (+ j 1) i) i))))

(defun get-admat (m)
  (let ((aux (assoc m $$addata)))
    (if aux       ;if already calculated
      (cadr aux)  ;just recall
      (let ((admat (calc-admat-of-s m)))   ;else calculate
        (push (list m admat) $$addata )    ;and store
        admat))))

(defun make-s-adtable (p m m+1) ;conversion from array
  (do                           ;to an offsetted table
    ((offsetp (* p m+1))        ;in the form ((3.#(0 3 2))..)
     (admat (get-admat m))
     (adtable '() 
              (append adtable 
                      (list 
                       (cons 
                        (+ i offsetp) 
                        (do ((row '() 
                                  (append row (list 
                                               (+ (aref admat i j) 
                                                  offsetp))))
                             (j 0 (+ j 1)))
                            ((= j m) (coerce row 'vector)))))))
     (i 0 (+ i 1)))
    ((= i m+1) adtable)))

(defun reindex&merge (bstable adtable)  ;finally eliminates leaks
  (let ((transtable                       ;in the numbering
         (do* ((len (length adtable))
               (transt (make-hash-table :size len))
               (i 0 (+ i 1)))
              ((= i len) transt)
           (setf  (gethash (car (nth i adtable)) transt) 
                  (make-simplex-symbol i)))))
    (reset-simplex-counter 0)
    (mapcar #'(lambda (adinfo)
                (let ((faceinfo (gethash (car adinfo) bstable)))
                  (make-simplex 
                   (advance-simplex-counter)
                   (car faceinfo)
                   (cadr faceinfo)
                   (map 'list 
                        #'(lambda (ad)
                            (gethash ad transtable))
                        (cdr adinfo)))))
            adtable)))

(defun work (adtable h k)  ;sewing operation
  (let* ((wholerowh (assoc h adtable))
         (wholerowk (assoc k adtable))
         (rowh (cdr wholerowh))
         (rowk (cdr wholerowk)))
    (map 'list
         #'(lambda (i j)
             (let* ((wholerow1 (assoc i adtable))
                    (wholerow2 (assoc j adtable))
                    (row1 (cdr wholerow1))
                    (row2 (cdr wholerow2))
                    (posh (position h row1))
                    (posk (position k row2)))
               (rplacd wholerow1
                       (progn (setf (svref row1 posh) j) row1))
               (rplacd wholerow2 
                       (progn (setf (svref row2 posk) i) row2))))
         rowh
         rowk)
    (setq adtable (remove wholerowh adtable))
    (setq adtable (remove wholerowk adtable))
    adtable))

(defun boundary (pol)  ;main function
  (let* ((S (get-#kd pol))
        (m (get-dpol pol))
        (m+1 (+ m 1))
        (slist (get-slist pol))
        (bstable (make-hash-table :size (* S 2)))
        (alreadyinserted (make-array (list S) :element-type 'bit
                                   :initial-element 0))
        (adtable '())
        (A (extract-admat slist)))
   (do ((p 0 (+ p 1)))
      ((= p S))
    (if (= (sbit alreadyinserted p) 0)
     (progn
      (setq adtable (append adtable (make-s-adtable p m m+1)))
      (setf (sbit alreadyinserted p) 1)))
    (do ((h 0 (+ h 1)))
       ((= h m+1))
     (cond ((= (aref A p h) -99))
          ((= (aref A p h) -1) 
         (multiple-value-bind 
         (sphsign sphvert) 
         (get-face (find-el (make-simplex-symbol p) slist) h) 
         (setf (gethash (+ h (* p m+1)) bstable) (list sphsign sphvert))))
          (t (let* ((q (aref A p h))
                  (k (do ((k 0 (+ k 1)))
                      ((= (aref A q k) p) k))))
             (if (= (sbit alreadyinserted q) 0)
             (progn
             (setq adtable (append adtable (make-s-adtable q m m+1)))
             (setf (sbit alreadyinserted q) 1)))
             (setq adtable (work adtable (+ h (* p m+1)) (+ k (* q m+1))))
             (setf (aref A q k) -99))))))
   (if (= (hash-table-count bstable) 0)
    **the-empty-polyhedron
    (make-polyhedron
    (get-vlist pol)
    (reindex&merge bstable adtable)))))
