; -*- package: extrusion -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package Extrude&Revolve                                   ;
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: V. Ferrucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'extrusio)                      ;module name

(defpackage extrusion)  ;package name
;(in-package 'extrusion :nicknames '("ex"))  ;package name
(inpackage extrusion) ;package name

(export '(set-extrude set-revolve get-extrude get-revolve 
          extrude s-extrude revolve build-grid build-grid1))

(require 'kernel)

(use-package 'kernel)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar $$exdata '())  ;data for extrude operations
(defvar $$redata '())  ;data for revolve operations

;;;;  Data input and output for extrude and revolve

(defun set-extrude (l)
  (cond ((not (consp l)) (error-handler 1 "set-extrude"))
        ((every 'consp l) (progn 
                            (do ((len (length l))
                                 (lenfirst (length (first l)))
                                 (i 0 (+ i 1)))
                                ((= i len))
                              (if (or (notevery 'numberp (nth i l))
                                      (/= (length (nth i l)) (+ i lenfirst)))
                                (error-handler 1 "set-extrude")))
                            (setq  $$exdata l)))
        (t (error-handler 1 "set-extrude"))))

;;;; The function validates user input to set-revolve

(defun validate-revolve-data (l k)
  (or (null l)
      (and (let ((currdata (car l)))
             (and (consp currdata)
                  (= (length currdata) 4)
                  (every 'numberp currdata)
                  (and (plusp (second currdata))
                       (plusp (third currdata))
                       (plusp (fourth currdata)))
                  (and (/= (third currdata) 
                           (fourth currdata)))
                  (<= (max (third currdata) (fourth currdata)) k)))
           (validate-revolve-data (cdr l) (+ k 1)))))

(defun set-revolve (l)
  (if (and (consp l)
           (validate-revolve-data (list (car l)) 100000) ;if just one element          
           (validate-revolve-data                 ;a large number
            (cdr l) 
            (+ 1 (max (third (car l)) (fourth (car l)))))); note the +1
    (setq  $$redata l)
    (error-handler 1 "set-revolve")))

;;;; The following two function are needed only because the variables
;;;; $$exdata and $$redata are internal

(defun get-extrude ()  $$exdata)

(defun get-revolve ()  $$redata)

;;;; Functions for extrude

(defun extrude (pol)
  (cond  ((empty-polyhedron? pol) (error-handler 2 "extrude"))
         ((not (polyhedron? pol)) (error-handler 1 "extrude"))
         ((not $$exdata) (error-handler 10 "extrude"))
         ((/= (get-npol pol) (- (length (first $$exdata)) 1)) 
          (error-handler 7 "extrude"))
         (t 
          (do ((m (get-dpol pol) (+ m 1))
               (expolvlist (get-vlist pol) 
                           (calculate-expolvlist expolvlist i))
               (expolslist (get-slist pol) (do-extrude expolslist i m))
               (i 0 (+ i 1)))
              ((= i (length $$exdata)) 
               (make-polyhedron expolvlist expolslist))))))

(defun calculate-expolvlist (vlist ind)
  (let ((vlistand0 (merge&append0 vlist)))
    (append vlistand0
            (mapcar #'(lambda (v) 
                        (extrude-vertex v ind))
                    vlistand0))))

(defun extrude-vertex (v ind)
  (list (image-symbol (get-id v) ind)
        (mapcar #'+ (get-coord v) (nth ind $$exdata))))

;-------------------------------------------------------------------------
;-------------------------------------------------------------------------
;
;       Functions for s-extrude :added by Sind-Vic in sept 91
;   Questa funzione esegue una estrusione lineare in  nsteps passi di
;   un poliedro pol.Restituisce un poliedro.

(defun s-extrude (pol nsteps)
  (cond  ((empty-polyhedron? pol) (error-handler 2 "s-extrude"))
         ((not (polyhedron? pol)) (error-handler 1 "s-extrude"))
         (t (do-s-extrude pol nsteps))))


(defun do-s-extrude (pol nsteps) ;does just one s-extrude operation 
  (let* ((n (get-npol pol))
         (S (get-#kd pol))
         (m (get-dpol pol))
         (m+1 (+ m 1))
         (S*m+1 (* S m+1))
         (mergedvlist (merge&append0 (get-vlist pol)))
         (vlistimage 
              (tr-image-of-vlist mergedvlist n (/  1 nsteps) 0))
         (expolvlist (append mergedvlist vlistimage))
         (expolslist (do-extrude (get-slist pol) 0 m)))
    (if (= nsteps 1) 
      (make-polyhedron expolvlist expolslist)
      (do* ((image1 mergedvlist)
            (image2 vlistimage)
            (thecomplex expolslist)
            (k 1 (+ k 1)))
           ((= k nsteps) 
            (make-polyhedron 
             expolvlist 
             (resolve-intercomplexes-ad expolslist nsteps S m m+1)))
        (let ((nextimage2 
               (tr-image-of-vlist image2 n (/ 1 nsteps)  k)))
          (setq thecomplex 
                (make-iso image1 image2 nextimage2 thecomplex S*m+1))
          (setq image1 image2)
          (setq image2 nextimage2)
          (setq expolvlist (append expolvlist image2))
          (setq expolslist (append expolslist thecomplex))))))) 

;----------------------------------------------------------------------------------
;    Questa funzione esegue una traslazione dei vertici contenuti nella lista vlist
;    nella loro n-esima componente di una quantita' pari ad inc.
;

(defun tr-image-of-vlist (vlist n inc ind)
  (mapcar #'(lambda (v) 
              (make-vertex 
               (image-symbol (get-id v) ind)
               (let* ((vcoord (coerce (get-coord v) 'vector))) ;converted to vector
                 (setf (svref vcoord n) (+ (svref vcoord n) inc))
                 (coerce vcoord 'list))))
          vlist))


;
;---------------------------------------------------------------------------------
;---------------------------------------------------------------------------------


;;;; Functions for revolve

(defun revolve (pol)
  (let ((n (get-npol pol)))
    (cond  ((empty-polyhedron? pol) (error-handler 2 "revolve"))
           ((not (polyhedron? pol)) (error-handler 1 "revolve"))
           ((< n 2) (error-handler 8 "revolve"))
           ((not $$redata) (error-handler 10 "revolve"))
           ((< n (max (third (first $$redata)) 
                      (fourth (first $$redata)))) 
            (error-handler 7 "revolve"))
           (t (do ((currpol pol)
                   (i 0 (+ i 1)))
                  ((= i (length $$redata)) currpol)
                (let* ((currparlist (nth i $$redata))
                       (nstep (second currparlist))
                       (alfa (* (/ (car currparlist) nstep) **pi/180))
                       (i (- (third currparlist) 1))   ;note the -1
                       (j (- (fourth currparlist) 1))) 
                  (setq currpol                        
                        (do-revolve currpol alfa nstep i j))))))))


(defun do-revolve (pol alfa nstep i j) ;does just one revolve operation 
  (let* ((n (get-npol pol))
         (S (get-#kd pol))
         (m (get-dpol pol))
         (m+1 (+ m 1))
         (S*m+1 (* S m+1))
         (cosalfa (cos alfa))
         (sinalfa (sin alfa))
         (mergedvlist (merge&append0 (get-vlist pol)))
         (vlistimage 
          (r-image-of-vlist mergedvlist n alfa cosalfa sinalfa i j 0))
         (expolvlist (append mergedvlist vlistimage))
         (expolslist (do-extrude (get-slist pol) 0 m)))
    (if (= nstep 1) 
      (make-polyhedron expolvlist expolslist)
      (do* ((image1 mergedvlist)
            (image2 vlistimage)
            (thecomplex expolslist)
            (k 1 (+ k 1)))
           ((= k nstep) 
            (make-polyhedron 
             expolvlist 
             (resolve-intercomplexes-ad expolslist nstep S m m+1)))
        (let ((nextimage2 
               (r-image-of-vlist image2 n alfa cosalfa sinalfa i j k)))
          (setq thecomplex 
                (make-iso image1 image2 nextimage2 thecomplex S*m+1))
          (setq image1 image2)
          (setq image2 nextimage2)
          (setq expolvlist (append expolvlist image2))
          (setq expolslist (append expolslist thecomplex)))))))


(defun r-image-of-vlist (vlist n alfa cosalfa sinalfa i j ind)
  (mapcar #'(lambda (v) 
              (make-vertex 
               (image-symbol (get-id v) ind)
               (let* ((vcoord (coerce (get-coord v) 'vector)) ;converted 
                      (xi (svref vcoord i))                   ;to vector
                      (xj (svref vcoord j)))
                 (setf (svref vcoord i) 
                       (let ((newxi (- (* xi cosalfa) (* xj sinalfa))))
                         (if (< (abs newxi) **epsilon) 0 newxi)))
                 (setf (svref vcoord j) 
                       (let ((newxj (+ (* xi sinalfa) (* xj cosalfa))))
                         (if (< (abs newxj) **epsilon) 0 newxj)))
                 (setf (svref vcoord n) (+ (svref vcoord n) alfa))
                 (coerce vcoord 'list))))
          vlist))

(defun make-iso (image1 image2 nextimage2 inputslist offset)
  (let ((table (build-table image1 image2 nextimage2 inputslist offset)))
    (mapcar #'(lambda (s) 
                (make-simplex
                 (second (assoc (get-id s) table))
                 (get-sign s)
                 (mapcar #'(lambda (vid) (second (assoc vid table)))
                         (get-vert s))
                 (mapcar #'(lambda (ad)
                             (if (eql ad 0) 0 (second (assoc ad table))))
                         (get-ad s))))
            inputslist)))

(defun build-table (image1 image2 nextimage2 inputslist offset)
  (let ((table '()))
    (mapc #'(lambda (v1 v2)
              (push (list (get-id v1) (get-id v2)) table))
          image1 
          image2)
    (mapc #'(lambda (v1 v2)
              (push (list (get-id v1) (get-id v2)) table))
          image2 
          nextimage2)
    (mapc #'(lambda (s)
              (let ((sid (get-id s)))
                (push 
                 (list sid 
                       (make-simplex-symbol (+ (get-index sid) offset)))
                 table)))
          inputslist)
    table))

(defun resolve-intercomplexes-ad (slist k S m m+1)
  (do* ((end (* S (- k 1)))
        (admat (extract-admat slist))
        (i 0 (+ i 1))
        (p m (+ (* i m+1 ) m))
        (q (* m+1 S) (* m+1 (+ i S))))
       ((= i end) (update-ad admat slist))
    (setf (aref admat p 0) q)
    (setf (aref admat q m+1) p)))

;;;; Common functions

(defun merge&append0 (vlist)
  (mapcar #'(lambda (v) 
              (list (image-symbol (get-id v) -1) (add-1dim (get-coord v))))
          vlist))

(defun image-symbol (sym ind)
  (let ((symstring (symbol-name sym)))
   (intern 
   (concatenate 'string symstring (integer->string (+ ind 1))))))

(defun do-extrude (slist ind m)
  (let ((exslist '())
       (m+1 (+ m 1)))
   (reset-simplex-counter 0)
   (mapc 
   #'(lambda (s) 
       (let ((svert (mapcar #'(lambda (sid) (image-symbol sid -1))
                            (get-vert s))))
       (setq
       exslist 
       (append exslist 
              (generate-chain svert svert '() (get-sign s) ind m m+1)))))
   slist)
   (update-ad 
   (calculate-ad 
   (extract-admat slist) 
   (extract-admat exslist)
   m m+1) 
   exslist)))

(defun calculate-ad (admat exadmat m m+1)
  (flet ((where-to-write 
         (offset h)
         (do ((i (if (= h 0) (+ offset 1) offset))
            (j (if (= h 0) m h))
            (l '())
            (k 0 (+ 1 k)))
           ((= k m) l)
         (setq l (append l (list (list i j))))
         (setq i (if (= j 1) (+ i 2) (+ i 1)))
         (setq j (if (= j 1) m (- j 1)))))
        (what-to-write 
       (l)
       (mapcar #'first l))
        (write-ad 
       (what where)
       (mapc #'(lambda (indices content)
                (setf (aref exadmat 
                         (first indices)
                         (second indices)) content))
            where
            what))) 
   (do ((i 0 (+ i 1)))
      ((= i (array-dimension admat 0)) exadmat)
    (do ((hi 0 (+ hi 1)))
       ((= hi m+1))
     (let ((ad (aref admat i hi))
          (offseti (* i m+1)))
      (cond ((= ad -9))
           ((= ad -1)
          (write-ad (make-list m :initial-element -1)
                  (where-to-write offseti hi)))
           (t (let* ((q ad)
                   (hq (do ((h 0 (+ h 1)))
                        ((eql (aref admat q h) i) h)))
                   (offsetq (* q m+1))
                   (li (where-to-write offseti hi))
                   (lq (where-to-write offsetq hq)))
              (write-ad (what-to-write li) lq)
              (write-ad (what-to-write lq) li)
              (setf (aref admat q hq) -9)))))))))


(defun generate-chain (svert l chain sign ind m m+1)
  (if (not (endp svert))
   (let ((ad-vect (make-array (+ 1 m+1) :initial-element '*))
        (exvert (append l (list (image-symbol (car l) ind))))
        (currlen (length svert)))
    (setf (svref ad-vect m+1) (previous-simplex-counter))
    (setf (svref ad-vect 0) (progn
                             (advance-simplex-counter)
                             (current-simplex-counter)))
    (if (= currlen m+1) (setf (svref ad-vect m+1) 0))
    (if (= currlen 1) (setf (svref ad-vect 0) 0))
    (generate-chain (cdr svert) 
                   (cdr exvert)
                   (append 
                 chain 
                 (list 
               (let ((s (make-simplex
                      (previous-simplex-counter) 
                      sign
                      exvert
                      (coerce ad-vect 'list))))
              (if (oddp (* m currlen))
              (reverse-sign s)
              s))))
                   sign
                   ind
                   m
                   m+1))
   chain))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Versione modificata da Valerio Pascucci il 21-1-91.
;;;; La funzione da chiamare e' build-grid con un numero a piacere di liste 
;;;; del tipo (min max numeropassi) o di numeri.
;;;; Nel caso in cui il parametro passato sia un numero 
;;;; (oppure numeropassi = 0) non viene effettuata nessuna estrusione,
;;;; ma si aggiunge semplicemente come ultima coordinata dei vertici
;;;; il numero passato (oppure min).
;;;; La funzione restituisce un poliedro.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun build-grid (&rest parlists)
  (let ((thegrid (make-polyhedron '((v0 ())) '((s0 (+ (v0) (0))))))
        (i 0))
    (dotimes (i (length parlists))
      (if (numberp (nth i parlists))
        (setf (nth i parlists)
              (list (nth i parlists)(nth i parlists) 0))))
    (dolist (this parlists)
      (format t "~%Extrusion...~D" (setq i (1+ i)))
      (setq thegrid 
            (if (= 0 (third this))
              (merge-pol thegrid)
              (extrude-grid 
               thegrid
               (third this)
               (/ (- (second this) (first this)) (third this))))))
    (translate thegrid (mapcar #'car parlists))))


;;; Questa funzione e' simile alla r-image-of-vlist

(defun t-image-of-vlist (vlist thestep ind)
  (mapcar #'(lambda (v) 
              (make-vertex 
               (image-symbol (get-id v) ind)
               (let ((vcoord (get-coord v)))
                 (append 
                  (butlast vcoord) (list (+ (car (last vcoord)) thestep))))))
          vlist))


;;; Questa funzione e' simile alla do-revolve

(defun extrude-grid (pol nstep thestep)
  (let*  ((S (get-#kd pol))
          (m (get-dpol pol))
          (m+1 (+ m 1))
          (S*m+1 (* S m+1))
          (mergedvlist (merge&append0 (get-vlist pol)))
          (vlistimage (t-image-of-vlist mergedvlist thestep 0))
          (expolvlist (append mergedvlist vlistimage))
          (expolslist (do-extrude (get-slist pol) 0 m)))
    (if (= nstep 1) 
      (make-polyhedron expolvlist expolslist)
      (do* ((image1 mergedvlist)
            (image2 vlistimage)
            (thecomplex expolslist)
            (i 1 (+ i 1)))
           ((= i nstep) 
            (make-polyhedron expolvlist (resolve-intercomplexes-ad expolslist nstep S m m+1)))
        (let ((nextimage2 (t-image-of-vlist image2 thestep i)))
          (setq thecomplex (make-iso image1 image2 nextimage2 thecomplex S*m+1))
          (setq image1 image2)
          (setq image2 nextimage2)
          (setq expolvlist (append expolvlist image2))
          (setq expolslist (append expolslist thecomplex)))))))






