; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package Kernel                                            ;
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: V. Ferrucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;Attenzione: error-handler e' stato modificato.
;La nuova versione accetta come primo argomento
;un error-code  (come prima) oppure una stringa.
;Fausto Bernardini 25-9-90

(provide 'kernel)                              ;module name

;; The new version of allgro common lisp for IBM risc 6000
;; requires the definition of a new package before the first
;; time it is used.
(defpackage kernel)
;(in-package 'kernel :nicknames '("ker"))   ;package name
(inpackage kernel) ;package name

(export '(intpos
          **the-empty-polyhedron
          **pi/180
          **epsilon
          initialize
          set-type-checking
          reset-type-checking
          get-type-checking
          error-handler
          current-simplex-counter advance-simplex-counter
          reset-simplex-counter   previous-simplex-counter
          current-vertex-counter  advance-vertex-counter
          reset-vertex-counter    previous-vertex-counter
          current-counter
          advance-counter
          reset-counter
          previous-counter
          length-two?
          length-three?
          string-intpos?
          empty-polyhedron?
          plus-or-minus?
          simplex-symbol-or-zero?
          polyhedron?
          vertex?
          simplex?
          vertex-list?
          simplex-list?
          vertex-symbol?
          simplex-symbol?
          make-polyhedron make-empty-polyhedron
          make-simplex make-vertex make-vertex-symbol make-simplex-symbol
          reverse-sign
          put-vertex put-simplex put-pol-sign 
          put-vlist put-slist put-info reset-info
          set-vid set-sid
          set-coord set-sign  set-vert set-ad 
          get-#K0   get-#kd get-dpol get-npol get-pol-sign get-info
          get-dsim 
          get-id
          get-coord 
          get-index
          get-sign get-vert get-vertcoord get-ad
          get-vlist  get-slist get-vertex get-simplex
          get-vcoord 
          get-svert get-svertcoord get-sad 
          find-el 
          find-vcoord 
          find-ssign find-svert find-sad))

(require 'functlib)

;;;;  Global declarations and functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(deftype intpos () `(integer 0 ,most-positive-fixnum))

(defun vertex-list? (l)
  (every 'vertex? l))

(defun simplex-list? (l)
  (every 'simplex? l))

(deftype vlist () '(satisfies vertex-list?))

(deftype slist () '(satisfies simplex-list?))

(defstruct (polyhedron (:constructor create-polyhedron))
  (m 0 :type intpos)
  (n 0 :type intpos)
  (k0 0 :type intpos)
  (km 0 :type intpos)
  (sign '+ :type symbol)
  (vlist '() :type vlist)
  (slist '() :type slist)
  (info nil))

(defun make-empty-polyhedron ()
  (create-polyhedron))

(defvar **the-empty-polyhedron (make-empty-polyhedron))
(defvar **pi/180 (/ pi 180))
(defvar  **epsilon 0.1d-12)

(defvar $$type-checking nil)
(defvar $$counter 0)
(defvar $$scounter 0)
(defvar $$vcounter 0)


(defun initialize ()
  (reset-type-checking)
  (reset-simplex-counter 0)
  (reset-vertex-counter 0)
  (reset-counter 0)
  t)

(defun error-handler (errorcode errorsource)
  (if (stringp errorcode)
    (format t "~%~%ERROR in ~A: ~A" errorsource errorcode)
    (case errorcode 
      (1 (format t "~%~%ERROR in ~A: Bad type argument" errorsource))
      (2 (format t "~%~%ERROR in ~A: Function not defined for the empty polyhedron" errorsource))
      (3 (format t "~%~%ERROR in ~A: Already present, can't insert" errorsource))
      (4 (format t "~%~%ERROR in ~A: Simplex must be fully instanced" errorsource))
      (5 (format t "~%~%ERROR in ~A: Wrong size arguments" errorsource))
      (6 (format t "~%~%ERROR in ~A: Content of file is not proper" errorsource))
      (7 (format t "~%~%ERROR in ~A: Sweeping data are not coherent with E-space dimension of the input" errorsource))
      (8 (format t "~%~%ERROR in ~A: E-space dimension of the input must be at least 2" errorsource))
      (9 (format t "~%~%ERROR in ~A: Can convert only (2)-polyhedra merged in R3" errorsource))
      (10 (format t "~%~%ERROR in ~A: Must first set sweeping data" errorsource))
      (t (format t "~%~%ERROR: Unknown error code to errorhandler from : ~A" errorsource))))
  (break))

(defun set-type-checking ()
  (setf  $$type-checking t))

(defun reset-type-checking ()
  (setf  $$type-checking nil))

(defun get-type-checking ()
  $$type-checking)

;;;; Counters

(defun current-simplex-counter ()
  (make-simplex-symbol $$scounter))

(defun advance-simplex-counter ()
  (prog1
    (make-simplex-symbol $$scounter)
    (setq $$scounter (+ $$scounter 1))))

(defun reset-simplex-counter (n)
  (setq $$scounter  n))

(defun previous-simplex-counter ()
  (make-simplex-symbol (- $$scounter 1)))

(defun current-vertex-counter ()
  (make-vertex-symbol $$vcounter))

(defun advance-vertex-counter ()
  (prog1
    (make-vertex-symbol $$vcounter)
    (setq $$vcounter (+ $$vcounter 1))))

(defun reset-vertex-counter (&optional (n 0))
  (setq $$vcounter  n))

(defun previous-vertex-counter ()
  (make-vertex-symbol (- $$vcounter 1)))

(defun current-counter ()
  $$counter)

(defun advance-counter ()
  (setq $$counter (+ $$counter 1))
  (- $$counter 1))

(defun reset-counter (n)
  (setq $$counter n))

(defun previous-counter ()
  (- $$counter 1))


;;;;  Predicates
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun length-two? (l)
  (and (listp l)
       (= (list-length l) 2)))

(defun length-three? (l)
  (and (listp l)
       (= (list-length l) 3)))

(defun string-intpos? (s)
  (and (stringp s)
       (every 'digit-char-p (coerce s 'list))))

(defun empty-polyhedron? (x)
  (equal x **the-empty-polyhedron))

(defun plus-or-minus? (x)
  (and (symbolp x)
       (or (equal x '+)
           (equal x '-))))

(defun simplex-symbol-or-zero? (x)
  (or (simplex-symbol? x)
      (equal x '*)
      (zerop x)))

(defun polyhedron? (thing)
  (polyhedron-p thing))

(defun vertex? (l)
  (and (length-two? l)
       (vertex-symbol? (get-id l))
       (every 'numberp (get-coord l))))

(defun simplex? (l)
  (and (length-two? l)
      (length-three? (second l))
      (simplex-symbol? (get-id l))
      (plus-or-minus? (get-sign l))
      (let ((svert (get-vert l))
           (sad (get-ad l)))
       (and (= (length svert) (length sad))
           (every 'vertex-symbol? svert)
           (every 'simplex-symbol-or-zero? sad)))))

(defun vertex-symbol? (x)
  (and (symbolp x)
       (every 'characterp 
                      (coerce (symbol-name x) 'list))))

(defun simplex-symbol? (x)
  (and (symbolp x)
       (let ((stringx (symbol-name x)))
         (and (char= (char stringx 0) #\S)
              (string-intpos? 
               (subseq stringx 1 (length stringx)))))))

;;;;  Constructors
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun make-polyhedron (vlist slist)
  (if (get-type-checking) 
   (if (not (and (vertex-list? vlist) (simplex-list? slist))) 
    (error-handler 1 "make-polyhedron")))
  (create-polyhedron 
  :m (- (length (get-vert (first slist))) 1)
  :n (length (get-coord (first vlist)))
  :k0 (length vlist)
  :km (length slist)
  :vlist vlist
  :slist slist))

(defun make-vertex (vsymb coord)
  (list vsymb coord))

(defun make-simplex (ssymb sign svert sad)
  (list ssymb (list sign svert sad)))

(defun make-vertex-symbol (i)
  (if (< i 0)
   (intern "V0")
   (intern 
   (concatenate 
   'string "V" (integer->string i)))))

(defun make-simplex-symbol (i)
  (if (< i 0)
   (intern "S0")
   (intern 
   (concatenate 
   'string "S" (integer->string i)))))

(defun reverse-sign (s)
  (if (get-type-checking) 
   (if (not (simplex? s))
    (error-handler 1 "reverse-sign")))
  (let ((sign (get-sign s)))
   (if (equal sign '-)
    (set-sign '+ s)
    (set-sign '- s))))

;; The following functions beginning with "put-" are destructive

(defun put-pol-sign (sign pol)
  (if (get-type-checking) 
    (if (not (and (plus-or-minus? sign) (polyhedron? pol))) 
      (error-handler 1 "put-pol-sign")))
  (setf (polyhedron-sign pol) sign))

(defun put-vertex (v pol)
  (if (get-type-checking) 
   (if (not (and (vertex? v) (polyhedron? pol))) 
    (error-handler 1 "put-vertex")))
  (if (get-vertex v pol)
   (error-handler 3 "put-vertex")
   (setf (polyhedron-vlist pol)
        (append (polyhedron-vlist pol) (list v)))))

(defun put-simplex (s pol)
  (if (get-type-checking) 
   (if (not (and (simplex? s) (polyhedron? pol))) 
    (error-handler 1 "put-simplex")))
  (if (get-simplex s pol)
   (error-handler 3 "put-simplex")
   (setf (polyhedron-slist pol)
        (append (polyhedron-slist pol) (list s)))))

(defun put-vlist (vlist pol &optional (replace-old nil))
  (if (get-type-checking) 
    (if (not (and (vertex-list? vlist) (polyhedron? pol))) 
      (error-handler 1 "put-vlist")))
  (if replace-old
    (setf (polyhedron-vlist pol) 
          vlist)
    (setf (polyhedron-vlist pol) 
          (append (polyhedron-vlist pol) vlist))))

(defun put-slist (slist pol &optional (replace-old nil))
  (if (get-type-checking) 
    (if (not (and (simplex-list? slist) (polyhedron? pol))) 
      (error-handler 1 "put-slist")))
  (if replace-old
    (setf (polyhedron-slist pol) 
          slist)
    (setf (polyhedron-slist pol) 
          (append (polyhedron-slist pol) slist))))

(defun put-info (pol key value)
  (if (get-type-checking) 
    (if (not (polyhedron? pol))
      (error-handler 1 "put-info")))
  (let ((info (polyhedron-info pol)))
    (if (assoc key info)
      (rplacd (assoc key info) value)
      (setf (polyhedron-info pol) (push (list key value) info)))))

(defun reset-info (pol)
  (if (get-type-checking) 
    (if (not (polyhedron? pol))
      (error-handler 1 "put-info")))
  (setf (polyhedron-info pol) nil))


;; The following functions are not destructive

(defun set-vid (id v)
  (if (get-type-checking) 
   (if (not (and (vertex-symbol? id) (vertex? v)))
    (error-handler 1 "set-vid")))
  (list id (get-coord v)))

(defun set-sid (id s)
  (if (get-type-checking) 
   (if (not (and (simplex-symbol? id) (simplex? s)))
    (error-handler 1 "set-sid")))
  (list id (list (get-sign s) (get-vert s) (get-ad s))))

(defun set-coord (vcoord v)
  (if (get-type-checking) 
   (if (not (and (every 'numberp vcoord) (vertex? v))) 
    (error-handler 1 "set-vcoord")))
  (list (get-id v) vcoord))

(defun set-sign (sign s)
  (if (get-type-checking) 
   (if (not (and (plus-or-minus? sign)
                (simplex? s))) 
    (error-handler 1 "set-svert")))
  (list (get-id s) (list sign (get-vert s) (get-ad s))))

(defun set-vert (svert s)
  (if (get-type-checking) 
   (if (not (and (every 'vertex-symbol? svert)
                (simplex? s))) 
    (error-handler 1 "set-svert")))
  (list (get-id s) (list (get-sign s) svert (get-ad s))))


(defun set-ad (sad s)
  (if (get-type-checking) 
   (if (not (and (every 'simplex-symbol-or-zero? sad) 
                (simplex? s))) 
    (error-handler 1 "set-sad")))
  (list (get-id s) (list (get-sign s) (get-vert s) sad)))

;;;;  Selectors
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun get-#K0 (pol)
  (if (get-type-checking) 
   (if (not (polyhedron? pol)) 
    (error-handler 1 "get-#K0")))
  (polyhedron-k0 pol))

(defun get-#kd (pol)
  (if (get-type-checking) 
   (if (not (polyhedron? pol))
    (error-handler 1 "get-#kd")))
  (polyhedron-km pol))

(defun get-dpol (pol)
  (if (get-type-checking)
   (if (not (polyhedron? pol))
     (error-handler 1 "get-dpol")
     (if (empty-polyhedron? pol)
      (error-handler 2 "get-dpol"))))
  (polyhedron-m pol))
                           
(defun get-npol (pol)
  (if (get-type-checking)
   (if (not (polyhedron? pol))
     (error-handler 1 "get-npol")))
  (polyhedron-n pol))

(defun get-pol-sign (pol)
  (if (get-type-checking)
   (if (not (polyhedron? pol))
     (error-handler 1 "get-pol-sign")))
  (polyhedron-sign pol))

(defun get-info (pol key)
  (if (get-type-checking)
    (if (not (polyhedron? pol))
      (error-handler 1 "get-info")))
  (second (assoc key (polyhedron-info pol))))

(defun get-dsim (s)
  (if (get-type-checking)
    (if (not (simplex? s))
     (error-handler 1 "get-dsim")
     (if (null (get-vert s))
      (error-handler 4 "get-dsim"))))
  (- (length (get-vert s)) 1))

(defun get-id (l)
  (first l))

(defun get-coord (v)
  (second v))

(defun get-index (ssymb)
      (let ((str (symbol-name ssymb)))
        (string->integer (subseq str 1 (length str)))))

(defun get-sign (s)
  (first (second s)))

(defun get-vert (s)
  (second (second s)))

(defun get-vertcoord (s pol)
  (mapcar #'(lambda (v) (get-vcoord v pol)) (get-vert s)))

(defun get-ad (s)
  (third (second s)))

(defun get-vlist (pol)
  (polyhedron-vlist pol))

(defun get-slist (pol)
  (polyhedron-slist pol))

(defun get-vertex (vid pol)
  (assoc vid (get-vlist pol)))

(defun get-simplex (sid pol)
  (assoc sid (get-slist pol)))
  
(defun get-vcoord (vid pol)
  (let ((aux (get-vertex vid pol)))
    (if aux (second aux) nil)))

(defun get-ssign (sid pol)
  (let ((aux (get-simplex sid pol)))
    (if aux (first (second aux)) nil)))

(defun get-svert (sid pol)
  (let ((aux (get-simplex sid pol)))
    (if aux (second (second aux)) nil)))

(defun get-svertcoord (sid pol)
  (mapcar #'(lambda (vid) (get-vcoord vid pol)) (get-svert sid pol)))

(defun get-sad (sid pol)
  (let ((aux (get-simplex sid pol)))
    (if aux (third (second aux)) nil)))

(defun find-el (el l)
  (assoc el l))

(defun find-vcoord (vid vlist)
  (let ((aux (assoc vid vlist)))
    (if aux (second aux) nil)))

(defun find-ssign (sid slist)
  (let ((aux (assoc sid slist)))
    (if aux (first (second aux)) nil)))

(defun find-svert (sid slist)
  (let ((aux (assoc sid slist)))
    (if aux (second (second aux)) nil)))


(defun find-sad (sid slist)
  (let ((aux (assoc sid slist)))
    (if aux (third (second aux)) nil)))
