; -*- package: matlib -*-
;-------------------------------------------------------------------------------
;           Universita' di Roma "La Sapienza"
;       Dipartimento di Informatica e Sistemistica
;            Via Buonarroti 12 - 00185 Roma
;
;                      CAD Group                 
; 
; Project: SimpleXn
; Identification: Package Matlib                         
; Date: 19-9-90
; Note: Matrix Functions Library
;             
; Version: 1.0
; Last modify: 12/6/92
;              
; Author: F. Bernardini
;--------------------------------------------------------------------------------

(provide 'matlib)        ;module name
			 ;package name
(defpackage  matlib)
(inpackage   matlib)


(export '(bincoeff             ;functions: alphabetical order
	  fact
	  mat-copy
	  mat-cut-a-column
	  mat-det
	  mat-diff
	  mat-gen-jacobian
	  mat-get-a-column
	  mat-get-columns
	  mat-inv 
	  mat-product 
	  mat-scalar-product
	  mat-sel-columns
	  mat-set-a-column
	  mat-sum
	  mat-sum-columns
	  mat-to-vet
	  mat-trans
	  mat-vet-mult 
	  mat-vet-product
	  mat-vet-solve
	  vet-copy 
	  vet-diff
	  vet-divide 
	  vet-grade-up 
	  vet-interval
	  vet-mat-product 
	  vet-max-position
	  vet-min-position
	  vet-mult 
	  vet-positions 
	  vet-product
	  vet-scalar-product
	  vet-sum 
	  vet-to-mat))

#+:akcl
  (load (concatenate 'string user::**simplexn-directory 
		     "/core/kernel.lisp"))

(use-package 'kernel)

;#+:APPLE
;(deftype real () `(double-float))

;--------------------------------------------------------------------------------
; Matrices
;--------------------------------------------------------------------------------

(defun mat-get-a-column (mat col)
  (let* ((d (array-dimension mat 0))
	 (vet (make-array (list d)
			  :element-type 'real)))
    (dotimes (i d) (setf (aref vet i) (aref mat i col)))
    vet))


(defun mat-get-columns (mat cols)
  (let* ((C (coerce cols 'vector))
	 (n (array-dimension mat 0))
	 (m (array-dimension C 0))
	 (new-mat (make-array (list n m) :element-type 'real)))
    (dotimes (i n new-mat)
      (dotimes (j m)
	(setf (aref new-mat i j) (aref mat i (aref C j)))))))


(defun mat-set-a-column (mat col vet)
  (let ((d (array-dimension mat 0)))
    (dotimes (i d vet) (setf (aref mat i col) (aref vet i)))))


(defun mat-cut-a-column (mat k)
  (do* ((n (array-dimension mat 0))
	(m-1 (1- (array-dimension mat 1)))
	(new-mat (make-array (list n m-1) :element-type 'real))
	(j 0 (1+ j))
	(h (if (< j k) j (1+ j)) (if (< j k) j (1+ j))))
       ((= j m-1) new-mat)
    (dotimes (i n) 
      (setf (aref new-mat i j) (aref mat i h)))))


(defun mat-sum-columns (mat)
  (let* ((d (array-dimension mat 0))
	 (res (make-array (list d)
			   :element-type 'real)))
    (dotimes (i d res)
      (setf (aref res i) 0)
      (dotimes (j (array-dimension mat 1))
	(setf (aref res i) (+ (aref res i) (aref mat i j)))))))


(defun mat-copy (mat)
  (let* ((n (array-dimension mat 0))
	 (m (array-dimension mat 1))
	 (new-mat (make-array (list n m)
			      :element-type 'real)))
	 (dotimes (i n new-mat) 
	   (dotimes (j m)
	     (setf (aref new-mat i j) (aref mat i j))))))


(defun mat-trans (A)
  (let* ((n (array-dimension A 0))
	 (m (array-dimension A 1))
	 (At (make-array (list m n)
			 :element-type 'real)))
    (dotimes (i n At)
      (dotimes (j m)
	(setf (aref At j i) (aref A i j))))))


(defun mat-scalar-product (A s)
  (let* ((n (array-dimension A 0))
	 (m (array-dimension A 1))
	 (As (make-array (list n m)
			 :element-type 'real)))
    (dotimes (i n As)
      (dotimes (j m)
	(setf (aref As i j) (* (aref A i j) s))))))


(defun mat-diff (A B)
  (let* ((n (array-dimension A 0))
	 (m (array-dimension A 1))
	 (C (make-array (list n m)
			:element-type 'real)))
    (dotimes (i n C)
      (dotimes (j m)
	(setf (aref C i j) (- (aref A i j) (aref B i j)))))))


(defun mat-sum (A B)
  (let* ((n (array-dimension A 0))
	 (m (array-dimension A 1))
	 (C (make-array (list n m)
			:element-type 'real)))
    (dotimes (i n C)
      (dotimes (j m)
	(setf (aref C i j) (+ (aref A i j) (aref B i j)))))))


(defun mat-product (A B)
  (let* ((n (array-dimension A 0))
	 (p (array-dimension A 1))
	 (m (array-dimension B 1))
	 (C (make-array (list n m)
			:element-type 'real
			:initial-element 0.0d0)))
    (dotimes (i n C)
      (dotimes (j m)
	(dotimes (k p) 
	  (setf (aref C i j) 
		(+ (aref C i j) (* (aref A i k) (aref B k j)))))))))


(defun mat-inv (mat)
  (let* ((n (array-dimension mat 0))
	 (2n (* 2 n))
	 (A (make-array (list n n)
			:element-type 'real))
	 (B (make-array (list n 2n)
			:element-type 'real
			:initial-element 0.0d0)))
    (dotimes (i n)
      (dotimes (j n) (setf (aref B i j) (aref mat i j)))
      (setf (aref B i (+ n i)) 1))
    (dotimes (k n)
      (when (zerop (aref B k k))
	(do* ((i k (1+ i)))
	     ((or (= i n) (not (zerop (aref B i k))))
	      (if (= i n) (error-handler "Det = 0" "mat-inv") 
		  (do ((j k (1+ j))) 
		      ((= j 2n)) 
		    (setf (aref B k j) (+ (aref B k j) (aref B i j))))))))
      (let ((d (aref B k k))) 
	(dotimes (j 2n) (setf (aref B k j) (/ (aref B k j) d))))
      (dotimes (j n)
	(unless (= j k)
	  (do ((d (aref B j k))
	       (h k (1+ h)))
	      ((= h 2n))
	    (setf (aref B j h) (- (aref B j h) (* d (aref B k h))))))))
    (dotimes (i n A)
      (dotimes (j n) (setf (aref A i j) (aref B i (+ n j)))))))


(defun mat-det (A)
  (do*
    ((n (array-dimension A 0))
     (n-1 (- n 1))
     (zero nil)
     (i (- n 2) (- i 1)))
    ((or zero (< i 0)) (if zero
			 0
			 (do ((j 0 (+ j 1))
			      (res 1 (* res (aref A j j))))
			     ((= j n) res))))
    (do
      ((h n-1 (- h 1))
       (k 0))
      ((or zero (= h i)))
      (when (zerop (aref A h h))
	(setf k (do
		  ((j 0 (+ j 1)))
		  ((or (> j h) (not (zerop (aref A h j)))) j)))
	(if (> k h)
	  (setf zero t)
	  (dotimes (j n)
	    (setf (aref A j h) (+ (aref A j h) (aref A j k))))))
      (unless zero
	(do
	  ((d (/ (aref A i h) (aref A h h)))
	   (j 0 (+ j 1)))
	  ((= j n))
	  (setf (aref A i j) (- (aref A i j) (* d (aref A h j)))))))))


(defun mat-gen-jacobian (X)
  (let*
    ((m (array-dimension X 0))
     (n (array-dimension X 1))
     (m-1 (1- m))
     (n-m (- n m))
     (C (make-array (1+ m))))
    (dotimes (j m) (setf (aref C j) j))
    (setf (aref C m) n)
    (do ((tot (expt (mat-det (mat-get-columns X C)) 2) 
	      (+ tot (expt (mat-det (mat-get-columns X C)) 2))))
      ((= (aref C 0) n-m) (sqrt tot))
      (do* ((i m-1 (1- i))
	    (r (1+ (aref C i)) (1+ (aref C i)))
	    (s (aref C m) (aref C (1+ i))))
	   ((< r s) (do ((j i (1+ j))
		       (k r (1+ k)))
		      ((= j m))
		    (setf (aref C j) k)))))))

(defun mat-gen-jacobian (X)
  (let*
    ((m (array-dimension X 0))
     (n (array-dimension X 1))
     (m-1 (1- m))
     (n-m (- n m))
     (C (make-array (1+ m))))
    (dotimes (j m) (setf (aref C j) j))
    (setf (aref C m) n)
    (do ((tot (expt (mat-det 
		     (mat-get-columns X 
	    (butlast (coerce C 'list)))) 2) 
	      (+ tot (expt (mat-det 
			    (mat-get-columns X 
	   (butlast (coerce C 'list)))) 2))))
	((= (aref C 0) n-m) (sqrt tot))
      (do* ((i m-1 (1- i))
	    (r (1+ (aref C i)) (1+ (aref C i)))
	    (s (aref C m) (aref C (1+ i))))
	   ((< r s) (do ((j i (1+ j))
			 (k r (1+ k)))
			((= j m))
		      (setf (aref C j) k)))))))


;--------------------------------------------------------------------------------
; Vectors
;--------------------------------------------------------------------------------

(defun vet-scalar-product (vet s)
  (let* ((d (array-dimension vet 0))
	 (res (make-array (list d)
			   :element-type 'real)))
    (dotimes (i d res)
	(setf (aref res i) (* (aref vet i) s)))))


(defun vet-diff (p q)
  (let* ((d (array-dimension p 0))
	 (res (make-array (list d)
			   :element-type 'real)))
    (dotimes (i d res)
	(setf (aref res i) (- (aref p i) (aref q i))))))


(defun vet-sum (p q)
  (let* ((d (array-dimension p 0))
	 (res (make-array (list d)
			   :element-type 'real)))
    (dotimes (i d res)
	(setf (aref res i) (+ (aref p i) (aref q i))))))


(defun vet-divide (p q)
  (let* ((d (array-dimension p 0))
	 (res (make-array (list d)
			   :element-type 'real)))
    (dotimes (i d res)
	(setf (aref res i) (/ (aref p i) (aref q i))))))


(defun vet-product (p q)
  (let ((r 0))
    (dotimes (i (array-dimension p 0) r)
      (setf r (+ r (* (aref p i) (aref q i)))))))
 

(defun vet-copy (vet)
  (let* ((d (array-dimension vet 0))
	 (res (make-array (list d)
			   :element-type 'real)))
    (dotimes (i d res)
	(setf (aref res i) (aref vet i)))))


(defun vet-mult (vet1 vet2)
  (map '(vector real) #'* vet1 vet2))


(defun vet-grade-up (vet)
  (let* ((base (vet-interval (array-dimension vet 0)))
	 (lista (map 'vector #'(lambda (x y) (list x y)) vet base)))
    (map 
     '(vector integer) 
     #'second 
     (stable-sort lista #'(lambda (x y) (< (first x) (first y)))))))


(defun vet-interval (n)
  (let ((buf (make-array (list n) :element-type 'integer)))
    (dotimes (i n buf) (setf (aref buf i) i))))


(defun vet-max-position (vet)
  (let ((max (aref vet 0))
	(pos 0))
    (dotimes (j (array-dimension vet 0) pos)
      (when (> (aref vet j) max)
	(setf max (aref vet j))
	(setf pos j)))))


(defun vet-min-position (vet)
  (let ((min (aref vet 0))
	(pos 0))
    (dotimes (j (array-dimension vet 0) pos)
      (when (< (aref vet j) min)
	(setf min (aref vet j))
	(setf pos j)))))


(defun vet-positions (item vet)
  (let ((res ()))
    (dotimes (i (array-dimension vet 0))
      (when (= item (aref vet i)) (push i res)))
    (reverse res)))


;--------------------------------------------------------------------------------
; Miscellanea
;--------------------------------------------------------------------------------

(defun mat-vet-product (mat vet)
  (let* ((d (array-dimension mat 0))
	 (res (make-array (list d)
			    :element-type 'real
			    :initial-element 0.0d0)))
    (dotimes (j (array-dimension vet 0) res)
      (dotimes (i d) 
	(setf (aref res i) 
	      (+ (aref res i) (* (aref vet j) (aref mat i j))))))))
 

(defun vet-mat-product (vet mat)
  (let* ((n (array-dimension mat 1))
	 (res (make-array (list n)
			     :element-type 'real
			     :initial-element 0.0d0)))
    (dotimes (i (array-dimension vet 0) res)
      (dotimes (j n)
	(setf (aref res j) 
	      (+ (aref res j) (* (aref vet i) (aref mat i j))))))))


(defun mat-vet-mult (mat vet)
  (let* ((n (array-dimension mat 0))
	 (m (array-dimension mat 1))
	 (res (make-array (list n m)
			:element-type 'real)))
    (dotimes (i n res)
      (dotimes (j m)
	(setf (aref res i j) (* (aref mat i j) (aref vet j)))))))


(defun mat-vet-solve (mat vet)
  (let ((tmat (mat-trans mat)))
    (mat-to-vet 
     (mat-product 
      (mat-inv (mat-product tmat mat)) 
      (mat-product tmat (vet-to-mat vet))))))


(defun mat-to-vet (mat)
  (let* ((d (array-dimension mat 0))
	 (vet (make-array (list d)
			    :element-type 'real)))
    (dotimes (i d) (setf (aref vet i) (aref mat i 0)))
    vet))


(defun vet-to-mat (vet)
  (let* ((d (array-dimension vet 0))
	 (mat (make-array (list d 1)
			  :element-type 'real)))
    (dotimes (i d mat) (setf (aref mat i 0) (aref vet i)))))


;--------------------------------------------------------------------------------
; Fattoriali
;--------------------------------------------------------------------------------

(defun fact1 (n) (do ((m 1 (+ m 1)) (res 1 (* res m))) ((> m n) res)))

;(defconstant factval 
;   (let ((temp (make-array 50 :element-type 'integer)))
;     (dotimes (i 50 temp)
;      (setf (aref temp i) (fact1 i)))))
(defvar factval 
   (let ((temp (make-array 50 :element-type 'integer)))
     (dotimes (i 50 temp)
      (setf (aref temp i) (fact1 i)))))

(defun fact (n) (aref factval n))

(defun bincoeff (n m) (/ (fact n) (* (fact m) (fact (- n m)))))






