; -*- package: matrixlib -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package Matrixlib                                         ;                                
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: V. Ferrucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'matrix) 

(defpackage  matrixlib)  

(inpackage matrixlib)

(export '(pre-multiply
          transpose
          nsubtract
          subtrac
          exstract-submatrix
          store-submatrix
          store-subrow
          store-subcolumn
          max-on-vect
          sum-rows
          exstract-row
          exstract-column
          exstract-subcolumn
          delete-row
          change-the-sign))

(require 'kernel)

(use-package 'kernel)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun pre-multiply (mat vect)
  (let ((rv (length vect))
        (rm (array-dimension mat 0))
        (cm (array-dimension mat 1)))
    (if (= cm rv)
      (do ((result (make-array rm :element-type 'double-float))
           (i 0 (+ i 1)))
          ((= i rm) result)
        (setf (aref result i)
              (do ((sum 0)
                   (j 0 (+ j 1)))
                  ((= j cm) sum)
                (setq  sum (+ sum (* (aref vect j)
                                     (aref mat i j)))))))
      (error-handler 5 "pre-multiply"))))

(defun transpose (A)
  (do* ((n (array-dimension A 0))
        (m (array-dimension A 1))
        (outmat (make-array (list m n)))
        (i 0 (+ i 1)))
       ((= i n) outmat)
    (do ((j 0 (+ j 1)))
        ((= j m))
      (setf (aref outmat j i) (aref A i j)))))

(defun nsubtract (A B n)    ;A=A-B, replaces A
  (dotimes (j n) (setf (aref A j) (- (aref A j) (aref B j)))))

(defun subtract (A B)    ;A=A-B
  (map 'vector #'- A B))

(defun exstract-submatrix (A i j n m)     ;gives submatrix
  (do ((outmat (make-array (list n m)))   ; n x m starting from Aij
      (i1 0 (+ i1 1)))
     ((= i1 n) outmat)
   (do ((j1 0 (+ j1 1)))
      ((= j1 m))
    (setf (aref outmat i1 j1) (aref A (+ i i1) (+ j j1))))))

(defun store-submatrix (A i j B p q)   ;B p x q in A starting from Aij
  (dotimes (i1 p)
   (dotimes (j1 q)
    (setf (aref A (+ i i1) (+ j j1)) (aref B i1 j1)))))

(defun store-subrow (A i j B q)   ;B 1 x q becomes part 
  (dotimes (j1 q)                 ;of row i in A starting from aij
   (setf (aref A i (+ j j1)) (aref B j1))))

(defun store-subcolumn (A i j B q)   ;B 1 x q becomes part of
  (dotimes (i1 q)                    ; column j in A starting from aij
   (setf (aref A (+ i i1) j) (aref B i1))))

(defun max-on-vect (B n) ; B has n columns
  (do ((max (aref B 0)) 
      (j 0)
      (k 0 (+ k 1)))
     ((= k n) j)
   (if (> (aref B k) max) (setf j k max (aref B k)))))


(defun sum-rows (A m n)  ; sums the m rows of A, starting from row 0
  (do ((B (make-array (list m) :initial-element 0))
      (i 0 (+ i 1)))
     ((= i m) B)
   (dotimes (j n)
    (setf (aref B i) (+ (aref B i) (aref A i j))))))

(defun exstract-row (A i n)  ; exstracts row i 
  (do ((B (make-array (list n)))
      (j 0 (+ j 1)))
     ((= j n) B)
    (setf (aref B j) (aref A i j))))

(defun exstract-column (A j m)  ; exstracts column j 
  (do ((B (make-array (list m)))
      (i 0 (+ i 1)))
     ((= i m) B)
   (setf (aref B i) (aref A i j))))

(defun exstract-subcolumn (A i1 j1 k)  ; subcolumn of length k 
  (do ((B (make-array (list k)))       ; starting from Ai1j1
      (i 0 (+ i 1)))
     ((= i k) B)
   (setf (aref B i) (aref A (+ i i1) j1))))

(defun delete-row (A i m n)  ; deletes row i 
  (do ((k i (+ k 1)))
     ((= k (- m 1)))
   (dotimes (j n)
    (setf (aref A k j) (aref A (+ k 1) j)))))

(defun change-the-sign  (B q)   ;B 1 x q 
  (dotimes (j q B)
   (setf (aref B j) (- (aref B j)))))
