; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package Kernel                                            ;
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: V. Pascucci                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require 'conv_ext)
(provide 'open_set)
(inpackage kernel )   ;package name

(export '(open-cell
	  get-cell-open-faces
	  open-cell
	  open-pol))

;-----------------------------------------------------------------------------
; Returns the list of booleans tagging the open faces of the convex cell
;-----------------------------------------------------------------------------
(defun get-cell-open-faces (cell)
  (convex-cell-open-faces cell))

;-----------------------------------------------------------------------------
; Sets all the faces of a cell to be open
;-----------------------------------------------------------------------------
  (defun open-cell (cell)
    (make-convex-cell
     (get-cell-identifier cell)
     (get-cell-sign cell)
     (get-cell-faces cell)
     (get-cell-vertices cell)
     (get-cell-adjacences cell) 
     (get-cell-internal-adjacences cell)
     (make-list  (length (get-cell-faces cell)) :initial-element t)))

  ;----------------------------------------------------------------------------
  ; Sets all the faces of a polyhedron to be open
  ;----------------------------------------------------------------------------
  (defun open-pol (pol)
      (put-cells   (mapcar 'open-cell (get-clist pol)) (copy-polyhedron pol) t))


  (inpackage extrusion)
  ;----------------------------------------------------------------------------
  ; product of two polyhedra with representation basen on faces:
  ; input: - the two polyhedra must be polyhedra
  ;          whoose representation is face based, and the basic elements
  ;          are convex cells.
  ; output: - one polyhedron decomposed in convex cells with face based representation.
  ;----------------------------------------------------------------------------
  (defun pol-on-faces-and-convexes-product-2 (pol1 pol2)
    (let* ((new-polyhedron (make-empty-polyhedron 'on-faces 'convex-cells))
	   (old-faces1 (get-flist pol1))
	   (old-faces2 (get-flist pol2))
	   (new-faces nil)
	   (new-old-faces-names-alist-1 nil)
	   (new-old-faces-names-alist-2 nil)
	   (old-cells1 (get-cells pol1))
	   (old-cells2 (get-cells pol2))
	   (new-old-cells-names-alist nil)    
	   ;;codimension of each original polyhedron in the product space.
	   (codim1 (1- (length (get-coord (first old-faces2)))))
	   (codim2 (1- (length (get-coord (first old-faces1)))))
	   (zero-covector1 (if (minusp codim1) nil (make-list codim1 :initial-element 0)))
	   (zero-covector2 (if (minusp codim2) nil (make-list codim2 :initial-element 0)))
	   ;(zero-covector1 (if (minusp codim1) nil (make-list codim1 :initial-element 0)))
	   ;(zero-covector2 (if (minusp codim2) nil (make-list codim2 :initial-element 0)))
	   ) 
      ;.......................
      ;;resetcounters
      (reset-convex-cell-counter)
      (reset-face-counter)
      ;.......................
      ;;create new faces list and the two association list for the 
      ;;correspondance from old to new names.
      (setf new-faces 
	    (append
	     (mapcar #'(lambda (the-face)
			 (advance-face-counter)
			 (setf new-old-faces-names-alist-1
			       (acons (get-id the-face)       ;old name
				      (current-face-counter)  ;new name
				      new-old-faces-names-alist-1))
			 (make-face (current-face-counter)
				    (append (get-coord the-face) zero-covector1)))
		     old-faces1)
	     (mapcar #'(lambda (the-face)
			 (advance-face-counter)
			 (setf new-old-faces-names-alist-2
			       (acons (get-id the-face)      ;old name
				      (current-face-counter) ;new name
				      new-old-faces-names-alist-2))
			 (make-face (current-face-counter)
				    (append (list (first (get-coord the-face))) 
					    zero-covector2
					    (rest (get-coord the-face)))))
		     old-faces2)))
      (mapcar 'put-face new-faces (never-ending-list new-polyhedron))
      ;(put-faces new-faces new-polyhedron t)
      ;;association list of product cell name and the cons of
      ;;cell factors names
      (maphash 
       #'(lambda (the-cell-1-name dummy)
	   (declare (ignore dummy))
	   (maphash 
	    #'(lambda (the-cell-2-name dummy)
		(declare (ignore dummy))
		(setf new-old-cells-names-alist
		      (acons (cons the-cell-1-name the-cell-2-name ) ;cell factors names
			     (advance-convex-cell-counter) ;cell product name
			     new-old-cells-names-alist)))
	    old-cells2))
       old-cells1)
      (maphash 
       #'(lambda (the-cell-1-name the-cell-1)
	   (maphash 
	    #'(lambda (the-cell-2-name the-cell-2)
		(put-cell 
		 (make-convex-cell  (cdr (assoc  (cons the-cell-1-name the-cell-2-name)
					       new-old-cells-names-alist
					       :test 'equal))
				   '+
				    (append                               ;the new faces list
					    (sublis new-old-faces-names-alist-1  ;is the union of the
						   (if (eq '- (get-cell-sign the-cell-1))
						     (mapcar 'opposit-face (get-cell-faces the-cell-1))
						      (get-cell-faces the-cell-1)))  ;two old faces lists
					    (sublis new-old-faces-names-alist-2  ;after the appropriate
						   (if (eq '- (get-cell-sign the-cell-2))
						     (mapcar 'opposit-face (get-cell-faces the-cell-2))
						     (get-cell-faces the-cell-2)))) ;transformation
				   ;;The vertices list is the proct of the list of
				   ;;vertices of each cell.
				   ;;Each resulting vertex is the union of a vertex
				   ;;of the first cell and another of the second.
				    (let ((new-cell-vertices nil))
				     (mapcar 
				      #'(lambda (cell-1-vertex)
					  (mapcar 
					   #'(lambda (cell-2-vertex)
					       (push (append cell-1-vertex cell-2-vertex)
						     new-cell-vertices))
					   (sublis new-old-faces-names-alist-2
						   (get-cell-vertices the-cell-2))))
				      (sublis new-old-faces-names-alist-1
					      (get-cell-vertices the-cell-1)))
				     new-cell-vertices)

				    (append
				    (sublis new-old-cells-names-alist
					    (mapcar
					     #'(lambda(adjacent-cell-name)
						 (if (numberp adjacent-cell-name)
						   adjacent-cell-name
						   (cons adjacent-cell-name the-cell-2-name)))
					     (get-cell-adjacences the-cell-1)) 
					    :test 'equal)
				    (sublis new-old-cells-names-alist
					    (mapcar
					     #'(lambda (adjacent-cell-name)
						 (if (numberp adjacent-cell-name)
						   adjacent-cell-name
						   (cons the-cell-1-name adjacent-cell-name)))
					     (get-cell-adjacences the-cell-2))
					    :test 'equal))
				   nil
				   (append (get-cell-open-faces the-cell-1)
					   (get-cell-open-faces the-cell-2)))
		 new-polyhedron))
	    old-cells2))
       old-cells1)
      ;............................................
      ; Reset all polyhedron attributes
      (progn
	(set-npol new-polyhedron (+ (get-npol pol1)(get-npol pol2)))
	(set-dpol new-polyhedron (+ (get-dpol pol1)(get-dpol pol2)))
	(set-#kd  new-polyhedron)
	(set-#k0  new-polyhedron) 
	(set-#F  new-polyhedron)
	(set-#C  new-polyhedron) 
	new-polyhedron)))

  (defun identify-coords (pol i j)
    (when  pol
      (let* ((polyhedron-cells (get-clist pol))
	     remove-cells
	     ;;nuovo database delle facce
	     (new-faces (new-faces-table))
	     ;;::::::::::::::::::::::::::::::::::::::::::::::::::::::
	     ;; Si calcolano le coppie di facce che siano coincidenti 
	     ;; sull'iperpiano Xi=Xj.
	     (coencident-faces-pairs (coincident-faces-pairs (get-faces pol) i j new-faces))
	     )
	;;(pprint '+++++++++++++++++++++++++++++++++++++++)
	;;(pprint coencident-faces-pairs)
	;;(pprint '+++++++++++++++++++++++++++++++++++++++)
    (dolist (cell polyhedron-cells)
      (let* ((cell-faces (copy-tree (get-cell-faces cell)))
	     (cell-open-faces (get-cell-open-faces cell))
	     (cell-adjacences (get-cell-adjacences cell))
	     (symbolic-faces  (mapcar 'first cell-faces))
	     ;; In the faces symbols substitution you obviously
	     ;; have to take into account only the faces bilonging to 
	     ;; the current face
	     (rename-faces-pairs 
	      (remove-if #'(lambda (faces-pair)
			     (subsetp   (list (car faces-pair)(cdr faces-pair)) symbolic-faces)) 
			 coencident-faces-pairs))
	     (coencident-faces-pairs 
	      (remove-if-not #'(lambda (faces-pair)
				 (subsetp   (list (car faces-pair)(cdr faces-pair)) symbolic-faces)) 
			     coencident-faces-pairs))
	     
	     
	     (redundant-faces-indices nil)
	     (redundant-faces nil)
	     (vertices_in_+_halfspase nil)
	     (vertices_in_=_halfspase nil)
	     (vertices_in_-_halfspase nil)
	     (symbolic-vertices (get-cell-vertices cell))
	     (numeric-vertices 
	      (mapcar 'calc-intersection-point
		      (mapcar #'(lambda (symbolic-vertex)
				  (mapcar 'get-fcoord symbolic-vertex
					  (never-ending-list pol))) symbolic-vertices)))
	     (positions  (mapcar #'(lambda (vertex-coords)
				     (let ((position
					    (- (nth (1- i) vertex-coords)
					       (nth (1- j) vertex-coords))))
				       (cond ((< (abs position) (* 1  **epsilon)) '=)
					     ((plusp position)
					      '+)
					     ((minusp position) '-)
					     (t '=))))
				 numeric-vertices)))
					;(pprint '+++++++++++++++++++++++++++++++++++++++)
					;(pprint coencident-faces-pairs)
					;(pprint rename-faces-pairs)
					;(pprint '+++++++++++++++++++++++++++++++++++++++)

	(if (and (member '+ positions)
		 (member '- positions))
	    (progn 
	      (dotimes (i (length positions))
		(if (symbol-equal (nth i positions) '-)
		    (push (nth i symbolic-vertices)  vertices_in_-_halfspase)
		  (if (symbol-equal (nth i positions) '=)
		      (push (nth i symbolic-vertices)  vertices_in_=_halfspase)
		    (push (nth i symbolic-vertices)  vertices_in_+_halfspase))))
	      ;;:::::::::::::::::::::::::::::::::::::::::::::::::::::
	      ;; Si eliminano le facce ridondanti secondo la regola
	      ;; che considera ridondanti tutte le facce che 
	      ;; che abbiano tutti i vertici dallo stesso lato
	      ;; di tale faccia {il prodotto scalare della faccia
	      ;; per i vertici sia sempre >=0 o sempre <=0}
	      (dolist (face-sign cell-faces)
		(let  ((face (car face-sign))
		       (signs nil))
		  (dotimes (i (length numeric-vertices))
		    (when (member face (nth i symbolic-vertices))
		      (push  (nth i positions) signs)
		      ))
					;(print face)
					;(print signs)
		  (when (or (every #'(lambda (x)(and (symbolp x)
						     (or (symbol-equal x '+)
							 (symbol-equal x '=)))) signs)
			    (every #'(lambda (x)(and (symbolp x)
						     (or (symbol-equal x '-)
							 (symbol-equal x '=)))) signs))
		    (push (position face-sign cell-faces) redundant-faces-indices)
		    (push face redundant-faces)
		    )
					;(print face) 
					;(print signs)
		  signs
		  ))

					;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
					; Si deve reinserire una faccia di ogni coppia di facce la cui interseziene
					; appartenga al sottospazio Xi=Xj.
					;:::::::DA FARE ::::::::
					; Le due facce devono avere uguale segno, altrimenti la cella e` vuota.
					; Se le due facce si chiamano Fa e Fb e si decide (arbitrariamente)
					; di reinserire Fa che risuta adiacete alla cella Ca, allora
					; la cella adiacente su Fa deve divetare la cella adiacete a Ca su Fb. 
					;(print 'SISTEMAZIONE_COPPIA_DI_FACCE_COINCIDENTI)
	      (dolist (faces-pair coencident-faces-pairs)
					;(print 'faces-pair)
					;(print faces-pair)
		(when (and (subsetp (list (car faces-pair)(cdr faces-pair))
				    symbolic-faces)
			   (or  (cdr faces-pair)
				t)
			   (not (equal  (position (cdr faces-pair) cell-faces :key 'car)
					(position (first faces-pair) cell-faces :key 'car))))
					;(pprint 'TEST_SUPERATO)
		  (let* ((face1-position (position (cdr faces-pair) cell-faces :key 'car))
			 (face2-position (position (car faces-pair) cell-faces :key 'car))
			 (open-tag (or (nth face1-position cell-open-faces)
				       (nth face1-position cell-open-faces))))
		    (setf (nth face1-position cell-open-faces) open-tag)
		    (setf (nth face2-position cell-open-faces) open-tag))
					; La seconda faccia della coppia e` quella che deve
					; rimanere nella cella
		  (setf 
		      redundant-faces-indices 
		    (but-nth (position (cdr faces-pair) redundant-faces)
			     redundant-faces-indices)
		    redundant-faces 
		    (but-nth (position (cdr faces-pair) redundant-faces)
			     redundant-faces))
		  (if (and (not (numberp (nth (position (cdr faces-pair) cell-faces :key 'car)
					      cell-adjacences)))
			   (not (numberp (nth (position (first faces-pair) cell-faces :key 'car)
					      cell-adjacences))))
		      (let* ((adjacent-cell (kernel::get-polyhedron-cell
					     pol (nth (position (cdr faces-pair) cell-faces :key 'car)
						      cell-adjacences)))
			     (other-cell-faces (get-cell-faces adjacent-cell))
			     (other-cell-adjacences (get-cell-adjacences adjacent-cell))
			     )
			(setf (nth (position (cdr faces-pair) cell-faces :key 'car)
				   cell-adjacences)
			  (nth (position (first faces-pair) other-cell-faces :key 'car)
			       other-cell-adjacences)))
		    (setf (nth (position (cdr faces-pair) cell-faces :key 'car)
			       cell-adjacences) 0))))
					;:::::::::::::::::::::::::::::::::::::::::::::
					; essendo gia` note le facce che coincidono a
					; coppie, si puo` sostituire la linea di comando
					; successiva con la chiamata che sfrutti tali
					; proprieta`.
					;(reduce-faces-hash-table (get-faces pol) i j)
	    
	      (when t                   ;redundant-faces-indices 
		(let* ((faces cell-faces) ;(faces (get-cell-faces cell))
		       (open-faces cell-open-faces)
		       (adjacences cell-adjacences) ;(adjacences (get-cell-adjacences cell))
		       (vertices (get-cell-vertices cell)))
					;(pprint 'RIMOZIONE_FACCE_RIDONDANTI)
					;(pprint (list faces adjacences))
		  ;;_______________________________________________________
		  ;; rimozione dalla cella dei simboli di facce ridondanti
		  (dotimes (i (length redundant-faces-indices))
		    (let* ((face-position (nth i redundant-faces-indices)))
		      (setf faces (but-nth face-position faces))
		      (setf open-faces (but-nth face-position open-faces))
		      (setf adjacences (but-nth face-position adjacences))
		      ))
					;(pprint (list faces adjacences))
		  (setf cell (make-convex-cell (get-cell-identifier cell)
					       (get-cell-sign cell)
					       (sublis coencident-faces-pairs faces)
					       (sublis coencident-faces-pairs vertices)
					       adjacences
					       nil
					       open-faces)
		      
			(gethash (get-cell-identifier cell) (get-cells pol)) cell)
		  (rebuild-vertices cell pol
					;vertices_in_+_halfspase
					;vertices_in_=_halfspase
					;vertices_in_-_halfspase          
					;(sublis coencident-faces-pairs vertices_in_+_halfspase)
				    vertices_in_+_halfspase
				    (sublis coencident-faces-pairs vertices_in_=_halfspase)
					;vertices_in_=_halfspase
					;(sublis coencident-faces-pairs vertices_in_-_halfspase)  
				    vertices_in_-_halfspase          
				    rename-faces-pairs))))
	  (push (get-cell-identifier cell) remove-cells))))
    (if remove-cells 
      (dolist (cell remove-cells)
	(remhash cell (get-cells pol))))
    (put-faces  new-faces pol t)
    (set-#F  pol)
    (set-#C  pol) 
    (if (zerop (get-#C pol))
      nil 
      pol))))


(defun rebuild-vertices (cell pol 
			      vertices_in_+_halfspase 
			      vertices_in_=_halfspase 
			      vertices_in_-_halfspase
			      rename-faces-pairs)
  ;(print 'RICOSTRUZIONE_VERTICI_INIZIO)
  ;(io::print-convex-cell cell)
  ;(pprint vertices_in_+_halfspase)
  ;(pprint vertices_in_=_halfspase)
  ;(pprint vertices_in_-_halfspase)
  ;(pprint 'inizio-ricostruzione-vertici******************************)
  (let* ((faces (mapcar 'first (get-cell-faces cell)))
	 (new-vertices nil))
    ;scansione dei vertici che giacciono sull'iperpiano
    ;Xi=Xj.
    (dolist (new-vertex vertices_in_=_halfspase)
      (push (intersection faces new-vertex ) new-vertices))
    ;(print '---1)
    ;(pprint new-vertices)
    ;doppia scansione per visitare i vertici
    ;che si trovano dai due lati opposti dell'iperpiano Xi=Xj
    (dolist (first-old-vertex vertices_in_-_halfspase)
      (dolist (second-old-vertex vertices_in_+_halfspase)
	(let* ((resulting-vertex 
		(intersection faces 
			     (intersection  first-old-vertex 
					    second-old-vertex))))
	  ;(print 'new_vertex)
	  ;(print resulting-vertex)
	  (setf new-vertices
		(remove-if #'(lambda (x)
			     (subsetp x resulting-vertex))  new-vertices  ))
	  
	  (pushnew resulting-vertex new-vertices :test 'subsetp )
)))
	;(print '---1)
	;(pprint new-vertices)

    (setf (gethash (get-cell-identifier cell) (get-cells pol))
	  (make-convex-cell (get-cell-identifier cell)
			    (get-cell-sign cell)
			    (sublis rename-faces-pairs (get-cell-faces cell))
			    (sublis rename-faces-pairs new-vertices)
			    (get-cell-adjacences cell)
			    nil
			    (get-cell-open-faces cell)))))


(inpackage kernel)

(defun polyhedron-skeleton (pol k)
  (declare (special konwn-skeletons))
  
  (cdr (or (assoc pol konwn-skeletons :test #'equalp)
	   (car (push (cons pol
			    (if (<= (get-dpol pol) k) pol
				(let* ((n (get-npol pol))
				       (n-k (- n k))
				       (faces (get-flist pol))
				       (cells (get-clist pol))
				       cells-groups-list
				       faces-not-adjacent)
				  (dolist (cell cells)
				    (let* ((cell-faces (get-cell-faces cell))
					   (cell-open-faces (get-cell-open-faces cell))
					   (cell-vertices (get-cell-vertices cell))
					   (cell-adjacences (get-cell-adjacences cell))
					   (max  (length cell-faces))
					   (indices nil))
				      (dotimes (i n-k)
					(push i indices))
				      (do ((indices indices)
					   (all-bundary-faces t t)
					   (test t (incf-indices indices max)))
					  ((not test))
					(dotimes (i n-k)
					  (if (or (symbolp (nth (nth i indices) cell-adjacences))
						  (nth (nth i indices) cell-open-faces))
					    (setf all-bundary-faces nil)))
					(when all-bundary-faces
					  (let ((faces-names   ; facce di base del sottospazio 
						 (mapcar 'first 
							 (mapcar 'nth indices (never-ending-list cell-faces))))
						all-vertices   ; insieme dei vertici appartenenti al sottospazio
						(all-faces     ; insieme di tutte le facce che generano il sottospazio 
						 (mapcar 'first cell-faces))
						all-faces-par  ; parametri delle facce che generano il sottospazio
						other-faces    ; altre facce che intersecano il sottospazio
						boundary-faces ; facce che delimitano il sottospazio
						)
					    (dotimes (i (length cell-adjacences))
					      (if (or (symbolp (nth i cell-adjacences))
						      (nth i cell-open-faces))
						(setf faces-not-adjacent
						      (adjoin (first (nth i cell-faces)) faces-not-adjacent))))
					    (dolist (vertex cell-vertices)
					      (if (subsetp faces-names vertex)
						(push vertex all-vertices)))
					    (when all-vertices
					      (dolist (vertex all-vertices)
						(setf all-faces (intersection all-faces vertex)))
					      (setf all-vertices 
						    (mapcar 'set-difference
							    all-vertices
							    (never-ending-list all-faces)))
					      (dolist (vertex all-vertices)
						(setf other-faces (union other-faces vertex)))
					      (setf all-faces-par
						    (mapcar 'get-coord
							    (mapcar 'assoc 
								    all-faces
								    (never-ending-list faces)
								    (never-ending-list  :test)
								    (never-ending-list  #'equalp))))
					      (when (= n-k
						       (calc-matrix-rank 0 (1- (length all-faces))
									 0 n
									 (copy-tree all-faces-par)))
						(dolist (boundary-face other-faces)
						  (let (boundary-face-vertices
							(common-faces other-faces))
						    (dolist (vertex all-vertices)
						      (if (member boundary-face vertex)
							(push vertex boundary-face-vertices)))
						    (dolist (vertex boundary-face-vertices)
						      (setf common-faces
							    (intersection common-faces vertex)))
						    (if (and (= 1 (length common-faces)) 
							     ; calcolo superfluo
							     ;(= (1+ n-k)
							     ;   (calc-matrix-rank 
							     ;    0 (length all-faces)
							     ;    0 n
							     ;    (copy-tree 
							     ;     (cons 
							     ;      (get-coord
							     ;       (assoc (first common-faces)
							     ;              faces :test #'equalp))
							     ;      all-faces-par)))) 
							     )
						      (setf boundary-faces
							    (adjoin (first common-faces) boundary-faces)))))
						(setf all-vertices 
						      (mapcar 'intersection
							      all-vertices
							      (never-ending-list boundary-faces)))
						(setf cells-groups-list
						      (put-new-cell-in-list
						       all-faces-par
						       boundary-faces
						       all-vertices
						       cells-groups-list
						       n-k
						       n
						       cell-faces
						       (get-cell-sign cell))))))))))
				  (reset-polyhedral-node-counter)
				  (mapcar 'calc-node-from-cells-group 
					  cells-groups-list
					  (never-ending-list faces)
					  (never-ending-list n-k)
					  (never-ending-list faces-not-adjacent)))))
		      konwn-skeletons))))
  
  )






; dato di prova
;(setf c (list (make-pol-node (setf b (let ((pol '#S(KERNEL::POLYHEDRON :M 2 :N 2 :K0 0 :KM 0 :SIGN + :VLIST NIL :SLIST NIL :INFO NIL :REPRESENTATION EXTRUSION::ON-FACES :FACES NIL :F 4 :CONVEX-CELLS NIL :C 1 :KIND-OF-BRICK EXTRUSION::CONVEX-CELLS :GENERAL-CELLS NIL) ))  (put-cells '(#S(KERNEL::CONVEX-CELL :IDENTIFIER C0 :SIGN + :FACES ((F2 . -) (F3 . -) (F0 . -) (F1 . -)) :VERTICES ((F2 F0) (F2 F1) (F3 F0) (F3 F1)) :ADJACENT-CONVEX-CELLS (0 0 0 0) :INTERNAL-ADJACENCES NIL :OPEN-FACES (t NIL NIL NIL))) pol t)   (put-faces '((F0 (0 1.0 0)) (F3 (0.7071067811865 0 -0.7071067811865)) (F2 (0 0 1.0)) (F1 (0.7071067811865 -0.7071067811865 0))) pol t) pol)))))


;;-----------------------------------------------------------------------------
;; Returns a polyhedron composed of a convex cell equal to the standard 
;; cube of dimension n.
;;-----------------------------------------------------------------------------
(defun build-cube (n &optional extremes)
  (declare (special o-piccolo-on-faces-and-convex-cells))
  (let* ((number (- (/ (sqrt 2.0d0))))
	 (empty-face1 (make-list (1+ n) :initial-element 0))
	 (empty-face2 (make-list n :initial-element 0))
	 (pol (make-empty-polyhedron 'on-faces 'CONVEX-CELLS))
	 (faces nil)
	 (faces-extr nil)
	 (vertices '(()))
	 (adj (make-list (* 2 n) :initial-element 0)))
    (push (- number) empty-face2)
    (dotimes (i n)
      (let ((f1-name (make-face-symbol (* 2 i)))
	    (f1-pars (copy-list empty-face1))
	    (f2-name (make-face-symbol (1+ (* 2 i))))
	    (f2-pars (copy-list empty-face2))
	    (extr (pop extremes)))
	(setf (nth (1+ i) f1-pars) 1)
	(setf (nth (1+ i) f2-pars) number)
	(put-face (make-face f1-name f1-pars) pol)
	(put-face (make-face f2-name f2-pars) pol)
	
	(push (cons f1-name '-) faces)
	(push (cons f2-name '-) faces)
	(push (eq 0 (car extr)) faces-extr)
	(push (eq 0 (cadr extr)) faces-extr)
	(setf vertices  (apply  #'append
				(mapcar #'(lambda (x) (list (cons f1-name x)(cons f2-name x)))
					vertices)))))
    (put-cell (make-convex-cell 'c0 '+ faces vertices adj nil faces-extr)
	      pol)
    (set-npol pol n)
    (set-dpol pol n)
    pol))

(defun power-pol-for-pol (base exponent)
  (declare (special known-powers))
  (if (= 0 (get-npol exponent))
    base
    (cdr (or (assoc (cons base exponent) known-powers :test #'equalp)
	     (car (push (cons (cons base exponent)
			 (product  base exponent))
		   known-powers))))))


;-----------------------------------------------------------------------------
; Creazione di un complesso poliedrale formato da un solo cubo standard
;-----------------------------------------------------------------------------
(defun build-pol-complex-cube (n &optional extremes)
  (let ((the-cube (build-cube n extremes))
	(the-complex (make-empty-polyhedra-complex)))
    (put-node-in-pol-complex (make-pol-node the-cube)
			     the-complex)
    the-complex))

;-----------------------------------------------------------------------------
; Returns the list of the convex cells of the polyhedron pol
;-----------------------------------------------------------------------------
(defun get-clist (pol &optional (sorted nil))
  (let ((clist nil))
    (declare (special clist))
    (maphash #'(lambda (name cell)
		 (if (convex-cell-p cell)(push cell clist)))
	     (get-cells pol))
    (if sorted 
      (sort clist #'(lambda (f1 f2)
		     (string< (get-cell-identifier f1)
			      (get-cell-identifier f2))))
      clist)))


;;-------------------------------------------------------
;; Thi function returns a square grid 3x3 without the
;; internal square of the grid.
;;
;;   .----.---.---.
;;   |    |   |   |          
;;   .----.---.---.
;;   |    |xxx|   | dy2     
;;   .----.---.---.
;;   |    |   |   | dy1              
;;   .----.---.---.
;;  V  dx1 dx2 dx3
;;
;; The cell marked as xxx is not present in the grid.
;; The lower left vertex V has coordinates (x0, y0)
;; The three sides along the X axes are dx1, dx2, dx3 long
;; The three sides along the Y axes are dy1, dy2, dx3 long
;;--------------------------------------------------------

(defun  square-hole (args)
  (let ((x0 (first args)) (dx1 (second args)) (dx2 (third args))   (dx3 (fourth args)) 
	(y0 (fifth args)) (dy1 (sixth args))  (dy2 (seventh args)) (dy3 (eighth args)))
    (let (
	  (xx x0)(yy y0)
	  (a dx1) 
	  (b dx2)
	  (c dx3) 
	  (d dy1)
	  (e dy2)
	  (f dy3))
      (list  (list '#2A((1 0.0d0 0.0d0) (0.0d0 1 0.0d0) (0.0d0 0.0d0 1)) 
		   (let ((pol
			   (make-empty-polyhedron 'on-vertices 'CONVEX-CELLS)
;                         '#S(POLYHEDRON :M 2 :N 2 :K0 16 :KM 0 :SIGN +
;                                       :VLIST nil
;                                       :SLIST NIL :INFO NIL :REPRESENTATION ON-VERTICES :FACES NIL :F 0
;                                       :CONVEX-CELLS NIL :C 9 :KIND-OF-BRICK CONVEX-CELLS :GENERAL-CELLS NIL)
			  )) 
		     (put-vlist     `((V15 (,(+ xx a b c) ,(+ yy d e f))) 
				      (V11 (,(+ xx a b)   ,(+ yy d e f))) 
				      (V7  (,(+ xx a)     ,(+ yy d e f))) 
				      (V3  (,xx           ,(+ yy d e f)))
				      (V14 (,(+ xx a b c) ,(+ yy d e))) 
				      (V10 (,(+ xx a b)   ,(+ yy d e))) 
				      (V6  (,(+ xx a)     ,(+ yy d e))) 
				      (V2  (,xx           ,(+ yy d e)))
				      (V13 (,(+ xx a b c) ,(+ yy d))) 
				      (V12 (,(+ xx a b c) ,yy)) 
				      (V9  (,(+ xx a b)   ,(+ yy d))) 
				      (V8  (,(+ xx a b)   ,yy))
				      (V0  (,xx           ,yy)) 
				      (V4  (,(+ xx a)     ,yy))
				      (V1  (,xx           ,(+ yy d)))
				      (V5  (,(+ xx a)     ,(+ yy d)))) pol t)
		     (put-cells
		      (list (make-convex-cell  
			     'C8 
			     +
			     '((V11 V15) (V14 V15) (V10 V11) (V14 V10))
			     '(V10 V14 V11 V15)
			     '(0 0 C7 C5) NIL
			     '(NIL NIL NIL NIL))
			    (make-convex-cell 'C7 +
					      '((V7 V11) (V10 V11) (V6 V7) (V10 V6))
					      '(V6 V10 V7 V11)
					      '(0 C8 C6 0) NIL
					      '(NIL NIL NIL t))
			    (MAKE-CONVEX-CELL 'C6 +
					      '((V3 V7) (V6 V7) (V2 V3)(V6 V2))
					      '(V2 V6 V3 V7)
					      '(0 C7 0 C3) NIL
					      '(NIL NIL NIL NIL))
			    (make-CONVEX-CELL 'C5 +
					      '((V10 V14) (V13 V14) (V9 V10) (V13 V9))
					      '(V9 V13 V10 V14)
					      '(C8 0 0 C2) NIL
					      '(NIL NIL t NIL))
			    ;;              '#S(CONVEX-CELL :IDENTIFIER C4 :SIGN +
			    ;;                 :FACES
			    ;;                 ((V6 V10) (V9 V10) (V5 V6)
			    ;;                  (V9 V5))
			    ;;                 :VERTICES (V5 V9 V6 V10)
			    ;;                 :ADJACENT-CONVEX-CELLS (C7 C5 C3 C1)
			    ;;                 :INTERNAL-ADJACENCES NIL :OPEN-FACES (NIL NIL NIL NIL))
			    (make-CONVEX-CELL 'C3 '+
					      '((V2 V6) (V5 V6) (V1 V2) (V5 V1))
					      '(V1 V5 V2 V6)
					      '(C6 0 0 C0) NIL
					      '(NIL t NIL NIL))
			    (make-CONVEX-CELL 'C2 +
					      '((V9 V13) (V12 V13) (V8 V9) (V12 V8))
					      '(V8 V12 V9 V13)
					      '(C5 0 C1 0) NIL
					      '(NIL NIL NIL NIL))
			    (make-CONVEX-CELL 'C1 :+
					      '((V5 V9) (V8 V9) (V4 V5) (V8 V4))
					      '(V4 V8 V5 V9)
					      '(0 C2 C0 0) NIL
					      '(t NIL NIL NIL))
			    (make-CONVEX-CELL 'C0 +
					      '((V1 V5) (V4 V5) (V0 V1) (V4 V0))
					      '(V0 V4 V1 V5)
					      '(C3 C1 0 0) NIL
					      '(NIL NIL NIL NIL))) pol t) 
		     (put-faces 
		      'NIL pol t)
		       (set-npol pol 2)
		       (set-dpol pol 2)
		       (set-#kd  pol)
		       (set-#k0  pol) 
		       (set-#F  pol)
		       (set-#C  pol) 
 pol)
		   'PC0 )))))

;;(KERNEL:setpol a  (kernel::square-hole 0 1 1 1 0 1 1 1))
;;(KERNEL:setpol b (KERNEL:build-pol-complex-cube 1 ))
;;(io:export-pol-complex (KERNEL:skeleton (KERNEL:power a b) 2) "prova.min")
;;(eval-plasm "export_to:'prova.min':(square_hole:<0,1,1,1,0,1,1,1>  (@2~*) quote:<1>)")

(defun polygon (lengths-and-angles)
  (when (functionp (car (last lengths-and-angles)) ) 
    (if (=  1 (apply (car (last lengths-and-angles)) '(1)))
	(setf (car (last lengths-and-angles))'+)
      (setf (car (last lengths-and-angles)) '- ) ))
  (if (not (subsetp   (last lengths-and-angles) '(+ -)))
      (setf (cdr (last lengths-and-angles)) '(+)))
  (let ((current-complex nil)
	(vertices-list '((0 0)))
	(open-close-mark-list nil)
	(center '(0 0))
	(current-angle 0)
	(current-direction '(1 0)))
    (dolist (current-data (butlast lengths-and-angles))
      (if (listp current-data)
	  (setf current-angle (+ current-angle (first current-data))
		current-direction (let ((angle (* pi (/ current-angle 180))))
				    (list (cos angle) (sin angle))))
	(let ((length (abs current-data)))
	  (if (minusp current-data)  (push '- open-close-mark-list) (push '+ open-close-mark-list))
	  (push (list (+ (* length (car  current-direction)) (caar  vertices-list))
		      (+ (* length (cadr current-direction)) (cadar vertices-list)))
		vertices-list)
	  (incf (car center) (caar vertices-list ))
	  (incf (cadr center) (cadar vertices-list )))))
					;(PPRINT CENTER)
    (let ((n  (length  vertices-list)))
      (setf (car center) (/ (car center) n)
	    (cadr center) (/ (cadr center) n)))
					;(PPRINT CENTER)
    (push (car (last vertices-list)) vertices-list)
    (push (car (last lengths-and-angles)) open-close-mark-list)
    (dolist (current-mark open-close-mark-list)
      (let ((x0 (car (first vertices-list ))) (y0 (cadr (first vertices-list )))
	    (x1 (car (second vertices-list )))(y1 (cadr (second vertices-list )))
	    (x2 (car center))                 (y2 (cadr center)))
	(let ((current-node  
	       (list '#2A((1 0.0d0 0.0d0) (0.0d0 1 0.0d0) (0.0d0 0.0d0 1))
		     ;; simplesso chiuso
		     (let ((pol 
			    (make-empty-polyhedron   'on-vertices
						     'convex-cells)))
		       (put-vlist    `((V0 (,x0 ,y0)) 
				       (V1 (,x1 ,y1)) 
				       (V2 (,x2 ,y2)) ;; centro del poligono
				       ) pol t)
		       (put-cells 
			(if (equal '+ current-mark) 
			    (list (make-CONVEX-CELL 'C3 +
						    '((V0 V1) (V1 V2) (V2 V0))
						    '(V0 V1 V2 )
						    '(0 0 0) NIL
						    '(NIL t t)))        
			  (list (make-CONVEX-CELL 'C3 +
						  '((V0 V1) (V1 V2) (V2 V0))
						  '(V0 V1 V2 )
						  '(0 0 0) NIL
						  '(T t t))))
			pol)
		       (set-npol pol 2)
		       (set-dpol pol 2)
		       (set-#kd  pol)
		       (set-#k0  pol) 
		       (set-#F  pol)
		       (set-#C  pol) 
		       pol)
		     (advance-polyhedral-node-counter)) 
	       ))
	  (push current-node current-complex)))
      (pop vertices-list))
    current-complex))


;;(KERNEL:setpol a  (kernel::polygon '(1 (90) 1 (90) 1 (90) )))
;;(KERNEL:setpol b (KERNEL:build-pol-complex-cube 1 ))
;;(io:export-pol-complex (KERNEL:skeleton (KERNEL:power a b) 2) "prova.min")

;;(io:export-pol-complex (KERNEL:skeleton (KERNEL:power (kernel::polygon '(0.5 -0.5 (90) 1 (90) 1 (90) 1)) (KERNEL:build-pol-complex-cube 1)) 2) "prova.min")
;;(eval-plasm "export_to:'prova.min':(polygon:<1,<90>,-1,<90>,1,2,<90>,-> (@2~*) quote:<1>)")


