; -*- package: kernel -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package Kernel                                            ;                                
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: G. Proietti                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require 'conv_ext)

(provide 'extpatch)
(inpackage kernel)


(require 'matlib_e)
(use-package 'matlib)

(require 'new_pl)
(use-package 'new-pl)

(format t "Patch su Polyhedra-Complex-Ext ~%")

;-----------------------------------------------------------------------------
; Calcola il nodo poliedrale dato dalla trasformazione affine che posiziona
; il sottospazio Rk in Rn e dal poliedro Pk,k composto dalle celle date.
; Il gruppo di celle  una lista di 3 elementi:
;   1) matrice (lista di liste) dei parametri delle facce che generano
;      il sottospazio;
;   2) lista dei nomi di tutte le facce usate in tutte le celle;
;   3) lista dei parametri che definiscono ciascuna cella; 
;    I parametri di definizione di una cella sono a loro volta forniti
;    in una lista di tre elementi:
;       1) lista dei nomi delle facce di bordo della cella;
;       2) lista dei vertici (indicati come lista delle facce che li generano)
;          di bordo della cella;
;       3) lista delle coppie (nome_faccia . segno) che delimitano la cella
;          considerata con segno positivo.
;-----------------------------------------------------------------------------
(defun calc-node-from-cells-group (cells-group pol-faces n-k faces-not-adjacent)

  (reset-face-counter) 
  (reset-convex-cell-counter) 
  (let* ((transposed-matrix (calc-node-transposed-matrix (first cells-group)))
         (n (length transposed-matrix))
         (faces (copy-tree (mapcar 'assoc 
                                   (second cells-group)
                                   (never-ending-list pol-faces)
                                   (never-ending-list  :test)
                                   (never-ending-list  #'equalp))))
         faces-to-invert
         new-face-pars
         (matrix (mat-inv (make-array (list n n)
                                      :initial-contents 
                                      (transpose-matrix transposed-matrix))))
         (cells-list (mapcar #'(lambda (cell)
                                (push (advance-convex-cell-counter)
                                      cell))(third cells-group)))
         (new-polyhedron (make-empty-polyhedron 'on-faces 'convex-cells)))
    (dolist (face faces)
      (setf new-face-pars (normalize-parameters 
                           (butlast  
                            (mapcar `(lambda (row) 
                                       (apply '+ 
                                              (mapcar '* row 
                                                      (quote ,(get-coord face)))))
                                    transposed-matrix)
                            n-k))
            (second face) (first new-face-pars))
      (when (eql '+ (cdr new-face-pars))
        (push (first face) faces-to-invert)))
    (dolist (cell cells-list)
      (setf (fourth cell)
            (mapcar #'(lambda (face)
                        (if (member (first face) faces-to-invert)
                          (opposit-face face)
                          face))
                    (fourth cell)))
      (push 
       (mapcar `(lambda (face)
                  (if (member face (quote ,faces-not-adjacent))
                    (dolist (cell (quote ,cells-list) 0)
                      (if (and (not (eq (first cell)
                                        (quote ,(first cell))))
                               (member face (second cell)))
                        (return (first cell))))
                    0))
               (second cell))
       (cddr cell)))
    (dolist (cell cells-list)
      (put-cell (make-convex-cell (first cell)
                                  '+
                                  (fifth cell)
                                  (fourth cell)
                                  (third cell)
                                  nil)
                new-polyhedron))

    (setq faces (reduce-Redundant-Faces-from-Cells-List (get-clist New-Polyhedron) faces))

    (put-faces faces new-polyhedron)
    (set-npol new-polyhedron (- n n-k 1))
    (set-dpol new-polyhedron (- n n-k 1))
    (set-#kd  new-polyhedron)
    (set-#k0  new-polyhedron) 
    (set-#F  new-polyhedron)
    (set-#C  new-polyhedron) 
    (make-pol-node new-polyhedron
                   matrix
                   (advance-polyhedral-node-counter))))


(defun calc-node-from-cells-group (cells-group pol-faces n-k faces-not-adjacent)

  (reset-face-counter) 
  (reset-convex-cell-counter) 
  (let* ((transposed-matrix (calc-node-transposed-matrix (first cells-group)))
         (n (length transposed-matrix))
         (faces (copy-tree (mapcar 'assoc 
                                   (second cells-group)
                                   (never-ending-list pol-faces)
                                   (never-ending-list  :test)
                                   (never-ending-list  #'equalp))))
         faces-to-invert
         new-face-pars
         (matrix (mat-inv (make-array (list n n)
                                      :initial-contents 
                                      (transpose-matrix transposed-matrix))))
         (cells-list (mapcar #'(lambda (cell)
                                (push (advance-convex-cell-counter)
                                      cell))(third cells-group)))
         (new-polyhedron (make-empty-polyhedron 'on-faces 'convex-cells)))
    (dolist (face faces)
      (setf new-face-pars (normalize-parameters 
                           (butlast  
                            (mapcar #'(lambda (row) 
                                       (apply '+ 
                                              (mapcar '* row 
                                                      (get-coord face))))
                                    transposed-matrix)
                            n-k))
            (second face) (first new-face-pars))
      (when (eql '+ (cdr new-face-pars))
        (push (first face) faces-to-invert)))
    (dolist (cell cells-list)
      (setf (fourth cell)
            (mapcar #'(lambda (face)
                        (if (member (first face) faces-to-invert)
                          (opposit-face face)
                          face))
                    (fourth cell)))
      (push 
       (mapcar #'(lambda (face)
                  (if (member face faces-not-adjacent)
                    (dolist (cell cells-list 0)
                      (if (and (not (eq (first cell)
                                        (first cell)))
                               (member face (second cell)))
                        (return (first cell))))
                    0))
               (second cell))
       (cddr cell)))
    (dolist (cell cells-list)
      (put-cell (make-convex-cell (first cell)
                                  '+
                                  (fifth cell)
                                  (fourth cell)
                                  (third cell)
                                  nil)
                new-polyhedron))

    (setq faces (reduce-Redundant-Faces-from-Cells-List (get-clist New-Polyhedron) faces))

    (put-faces faces new-polyhedron)
    (set-npol new-polyhedron (- n n-k 1))
    (set-dpol new-polyhedron (- n n-k 1))
    (set-#kd  new-polyhedron)
    (set-#k0  new-polyhedron) 
    (set-#F  new-polyhedron)
    (set-#C  new-polyhedron) 
    (make-pol-node new-polyhedron
                   matrix
                   (advance-polyhedral-node-counter))))

(defun Reduce-Redundant-Faces-from-Cells-List (c-list Faces)
  (let ( NewFaces )
    (setq NewFaces 
          (multiple-value-list 
           (kernel::Lex-Purge-Equal-Verts
            (kernel::Lex-Sort-with-Labels Faces))
           )
          )
    (Reassign-Faces-To-Cells c-list (second NewFaces))
    (first NewFaces)
    )
  )

(defun Reassign-Faces-To-Cells (Cells Faces-AList)
  (dolist (Cell Cells)
    (Reassign-Faces-To-Single-Cell Cell Faces-AList)
    )
  )


(defun Reassign-Faces-To-Single-Cell (Cell Faces-AList)
  (setf (kernel::Convex-Cell-Vertices Cell)
         (mapcar '(lambda (x)
                    (mapcar '(lambda (y)
                               (let ( (Nuovo (assoc y Faces-AList) ))
                                 (if Nuovo
                                   (second Nuovo)
                                   y
                                   )
                                 )
                               ) x )) (kernel::Convex-Cell-Vertices Cell) )
         )
  (setf (kernel::Convex-Cell-Faces Cell)
        (mapcar '(lambda (y)
                   (let ( (Nuovo (assoc (car y) Faces-AList)) )
                     (if Nuovo
                       (cons
                        (second Nuovo)
                        (cdr y)
                        )
                       ;else
                       y
                       )
                     )
                   ) (kernel::Convex-Cell-Faces Cell) )
        )
  Cell
  )

(defun Reassign-Faces-To-Single-Cell (Cell Faces-AList)
  (setf (kernel::Convex-Cell-Vertices Cell)
         (mapcar #'(lambda (x)
                    (mapcar #'(lambda (y)
                               (let ( (Nuovo (assoc y Faces-AList) ))
                                 (if Nuovo
                                   (second Nuovo)
                                   y
                                   )
                                 )
                               ) x )) (kernel::Convex-Cell-Vertices Cell) )
         )
  (setf (kernel::Convex-Cell-Faces Cell)
        (mapcar #'(lambda (y)
                   (let ( (Nuovo (assoc (car y) Faces-AList)) )
                     (if Nuovo
                       (cons
                        (second Nuovo)
                        (cdr y)
                        )
                       ;else
                       y
                       )
                     )
                   ) (kernel::Convex-Cell-Faces Cell) )
        )
  Cell
  )
