; -*- package: matlib -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;           Universita' di Roma "La Sapienza"                               ;
;       Dipartimento di Informatica e Sistemistica                          ;
;            Via Buonarroti 12 - 00185 Roma                                 ;
;                                                                           ;
;                      CAD Group                                            ;                    
;                                                                           ;
; Project: SimpleXn                                                         ;
; Identification: Package Matlib                                            ;                                
;                                                                           ;
;  Version: 1.0                                                             ;
;                                                                           ;
;  Author: G. Proietti                                                      ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(inpackage matlib)

(provide 'matinvp)

(format t "Patch su matlib:Mat-Inv")

;--------------------------------------------------------------------
; Questa funzione va a sostituire la funzione l'omonima funzione nel 
; package matlib ed esegue l'inversione della matrice per mezzo di
; operazioni di pivot di Gauss-Jordan.
; Il metodo consiste nel considerare una matrice doppia con, in prima
; posizione la matrice da invertire ed in seconda una matrice identita'.
; Trasformando, mediante operazioni di pivoting, la prima sottomatrice in
; una matrice identita', si ottiene, nella seconda sottomatrice,
; la matrice inversa.

; Per il suo funzionamento richiede la presenza del package new-pl, 
; contenente la funzione di scambio delle righe della matrice e delle due
; funzioni che eseguono la trasformazione da rappresentazione mediante
; array a liste di liste della matrice e viceversa.

(defun Mat-Inv ( Mat )
  (let* (Mat-Lista 
         Ext-Mat
         N-Rows
         PivotElement
         )
    ;Converte in lista di liste
    (setq Mat-Lista (Mat-to-ListOfList Mat))
    (setq N-Rows (length Mat-Lista))
    ;Crea la matrice estesa
    (setq Ext-Mat (mapcar #'(lambda (x y) (append x y)) 
                          Mat-Lista
                          (new-pl::identity-matrix N-Rows)
                          )
          )
    ; Per ogni colonna cerca un elemento di pivot
    (dotimes (i N-Rows)
      (setq PivotElement (New-PL::find-pivot-element i (1- N-Rows) i i Ext-Mat))
      (if (/= (first PivotElement) i)
        ; Se l'elemento non si trova sulla diagonale scambia le
        ; righe in modo da portarvelo.
        (new-pl::Swap-Rows (first PivotElement) i Ext-Mat)
        )
      (New-PL::pivot i i Ext-Mat)
      )

    ;Eliminazione prima sottomatrice
    (setq Ext-Mat (mapcar '(lambda (x) (nthcdr N-Rows x)) Ext-Mat))
    ;Riconverte in array
    (ListOfList-to-Mat (mapcar '(lambda (x) 
                                  (mapcar '(lambda (y) 
                                             (coerce y 'float)) 
                                          x))
                               Ext-Mat))
    )
  )

(defun Mat-Inv ( Mat )
  (let* (Mat-Lista 
         Ext-Mat
         N-Rows
         PivotElement
         )
    ;Converte in lista di liste
    (setq Mat-Lista (Mat-to-ListOfList Mat))
    (setq N-Rows (length Mat-Lista))
    ;Crea la matrice estesa
    (setq Ext-Mat (mapcar #'(lambda (x y) (append x y)) 
                          Mat-Lista
                          (new-pl::identity-matrix N-Rows)
                          )
          )
    ; Per ogni colonna cerca un elemento di pivot
    (dotimes (i N-Rows)
      (setq PivotElement (New-PL::find-pivot-element i (1- N-Rows) i i Ext-Mat))
      (if (/= (first PivotElement) i)
        ; Se l'elemento non si trova sulla diagonale scambia le
        ; righe in modo da portarvelo.
        (new-pl::Swap-Rows (first PivotElement) i Ext-Mat)
        )
      (New-PL::pivot i i Ext-Mat)
      )

    ;Eliminazione prima sottomatrice
    (setq Ext-Mat (mapcar #'(lambda (x) (nthcdr N-Rows x)) Ext-Mat))
    ;Riconverte in array
    (ListOfList-to-Mat (mapcar #'(lambda (x) 
                                  (mapcar #'(lambda (y) 
                                             (coerce y 'float)) 
                                          x))
                               Ext-Mat))
    )
  )

;--------------------------------------------------------------
;Le seguenti due funzioni si occupano della conversione di
;rappresentazione di matrici da array a lista di liste e
;viceversa.

(defun Mat-To-ListOfList ( mat )
  (let* ( (dims (array-dimensions mat))
          (nRows (first dims))
          (nCols (second dims))
          Res
          )
    (dotimes (i nRows Res)
      (setq Res 
            (append Res 
                    (list 
                     (let ( (Temp (make-list nCols)) )
                       (dotimes (j nCols Temp) 
                         (setf (nth j Temp) (aref mat i j))
                         )
                       )
                     )
                    )
            )
      )
    )
  )

(defun ListOfList-To-Mat ( LOL )
  (let* ( (nRows (length LOL))
          (nCols (length (car LOL)))
          (Res   (make-array (list nRows nCols)))
          Temp
          )
    (dotimes (i nRows Res)
      (setq Temp (nth i LOL) )
      (dotimes (j nCols) 
        (setf (aref Res i j) (nth j Temp))
        )
      )
    )
  )
